// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file Error.h
 * Declarations of error reporting functions.
 */
#ifndef MM_ERROR_H
#define MM_ERROR_H

#include "ansidecl.h"

#ifdef __cplusplus
#include <cstdlib>
#include <cassert>
#include <cstdio>

extern "C" {
#else
#include <assert.h>
#include <stdio.h>
#endif


/** @name Not-implemented error functions
 * Error and assert functions that are called when some statement 
 * can't be interpreted, because it was not implemented.
 */
/*@{*/
#define ni_error(format, ...) \
      do_ni_error (__FILE__, __FUNCTION__, __LINE__, format, ## __VA_ARGS__)

#define ni_assert(condition, format, ...) \
      do_ni_assert (condition, __FILE__, __FUNCTION__, __LINE__, format, ## __VA_ARGS__)

extern void do_ni_error (const char* file, const char* function, int line, 
                         const char* format, ...) ATTRIBUTE_NORETURN ATTRIBUTE_PRINTF(4, 5);
extern void do_ni_assert (bool condition, const char* file, const char* function, int line, 
                          const char* format, ...) ATTRIBUTE_PRINTF(5, 6);
/*@}*/

/** @name Runtime error reporting functions
 * Functions and macros that are called when program, that is beeing interpreted,
 * tried to do some operation operation with undefined behaviour.
 */
/*@{*/
extern void rt_error (const char* format, ...) 
                         ATTRIBUTE_NORETURN ATTRIBUTE_PRINTF(1, 2);
extern void rt_assert (bool condition, const char* format, ...) 
                          ATTRIBUTE_PRINTF(2, 3);
/*@}*/


/**
 * Writes a message to the standard error output.
 * The message is always printed.
 */
#define LOG_INFO(format, ...) do { fprintf (stderr, format, ## __VA_ARGS__); fprintf (stderr, "\n"); } while (0)

#ifndef NDEBUG
/**
 * Writes a message to the standard error output.
 * The message is printed only if the @c NDEBUG macro is not defined
 */
#define LOG_DEBUG(format, ...) do { fprintf (stderr, format, ## __VA_ARGS__); fprintf (stderr, "\n"); } while (0)
#else
#define LOG_DEBUG(format, ...) do {} while (0)
#endif

#define WARNING(format, ...) do { fprintf (stderr, "Warning: "); fprintf (stderr, format, ## __VA_ARGS__); fprintf (stderr, "\n"); } while (0)

#ifdef __cplusplus
}
#endif


#endif 

