// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file AbstractDecl.h
 * Definition of abstract base for declarations.
 */
#ifndef GIT_ABSTRACT_DECL_H
#define GIT_ABSTRACT_DECL_H

#include "Decl.h"
#include "Serialization.h"

namespace giterator { namespace detail {

    
/**
 * Base class for all classes representing declarations.
 * This class implements all methods in @c Decl class.
 */
class AbstractDecl : public Decl {

  private:
    /*! unique ID of this declaration */
    unsigned int m_uid;
    
    /*! name of this declaration */
    std::string m_name;
    
    /*! type of this declaration */
    const Type* m_type;
    
  public:
    /**
     * Constructor, creates a new declaration.
     * Empty constructor exists for serialization purposes,
     * use nonempty for normal use.
     */
    AbstractDecl () : m_type(NULL) {}
    /**
     * Constructor, creates a new declaration.
     * 
     * @param uid unique ID of the new declaration
     * @param name name of the new declaration
     * @param type type of the new declaration
     */
    AbstractDecl (unsigned int uid, const std::string& name, const Type* type)
    : m_uid (uid), m_name (name), m_type (type) {}
    
    /**
     * Destructor.
     */
    ~AbstractDecl () {}
    
    /**
     * Obtains unique ID of this declaration.
     *
     * @returns unique ID of this declaration.
     */
    unsigned int getUid () const
    { return m_uid; }
    
    /**
     * Obtains name of this declaration.
     * For example declaration:
     *
     * @return name of this declaration
     */
    std::string getName () const
    { return m_name; }
    
    /**
     * Return type of this declaration.
     *
     * @return type of this declaration
     */
    const Type* getType () const
    { return m_type; }

    /** Checks if the declaration is the same declaration as
	 * another declaration. Declarations are equal, when their
	 * unique IDs are equal
	 *
	 * @param decl declaration to be checked for equality
	 * @return true if the declarations are equal, false otherwise
	 */
    bool isSame(const Decl& decl) const
    {
    	return getUid() == decl.getUid();
    }

    friend class boost::serialization::access;

    /**
	 * Serialization function, saving the declaration
	 * into archive.
	 *
	 * @param ar archive to save the declaration to
	 * @param int version of the archive
	 */
	template<class Archive>
	void save(Archive & ar, const unsigned int) const
	{
		ar << m_uid;
		ar << m_name;
		ar << m_type;
	}

	/**
	 * Serialization function, loading the declaration
	 * from archive.
	 *
	 * @param ar archive to load the declaration from
	 * @param version of the archive
	 */
	template<class Archive>
	void load(Archive & ar, const unsigned int)
	{
		ar >> m_uid;
		ar >> m_name;
		Type* type;
		ar >> type;
		m_type = (const Type*) type;
	}

	/**
	 * Serialization function, handling saving and loading
	 * the declaration into the serialization archive.
	 *
	 * @param ar archive to save/load the declaration to/from
	 * @param version of the archive
	 */
	template<class Archive>
	void serialize(Archive & ar, const unsigned int version)
	{
		boost::serialization::split_member(ar, *this, version);
	}
};


}}

#endif /* GIT_ABSTRACT_DECL_H */
