// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file AbstractType.h
 * Definition of abstract base for types.
 */
#ifndef GIT_ABSTRACT_TYPE_H
#define GIT_ABSTRACT_TYPE_H

#include "Type.h"
#include "Serialization.h"

#include <boost/functional/hash.hpp>

namespace giterator { namespace detail {
    
/**
 * Base class for all classes representing types.
 * This class implements all methods in @c Type class.
 */
class AbstractType : public Type {

  private:
    /*! name of the type */
    std::string m_name;
    
    /*! size of the type in bits */
    size_t m_bitSize;

  public:
    /**
     * Constructor, creates a new type.
     * Empty constructor exists for serialization purposes,
     * use nonempty for normal use.
     */
    AbstractType() {}
    /**
     * Constructor, creates a new type.
     * 
     * @param name name of the new type
     * @param bitSize size of the values of the new type in bits
     */
    AbstractType(const std::string& name, size_t bitSize) 
        : m_name (name), m_bitSize (bitSize) {}

    /**
     * Returns the size of the values of this type in bits.
     * 
     * @return the size of the values of this type in bits
     */
    size_t getBitSize() const 
    { return m_bitSize; }

    /**
     * Returns the name of this type.
     * 
     * @return the name of this type
     */
    std::string getName() const 
    { return m_name; }
    
    /**
     * Returns the hash of this type.
     * Used for memory state equivalency checking purposes.
     *
     * @returns hash representing this type.
     */
    mmodule::Hash getHash () const 
    { 
        return boost::hash<std::string> () (getName ());
    }

    /**
	 * Serialization function, handling saving and loading
	 * the type into the serialization archive.
	 *
	 * @param ar archive to save/load the type to/from
	 * @param version of the archive
	 */
    template<class Archive>
	void serialize(Archive & ar, const unsigned int)
	{
    	ar & m_name;
    	ar & m_bitSize;
	}
};

}}

#endif  /* GIT_ABSTRACT_TYPE_H */

