// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file ArrayConstructorOperand.h
 * Operand, that is a constructor of array.
 */
#ifndef GIT_ARRAY_CONSTRUCTOR_OPERAND_H
#define GIT_ARRAY_CONSTRUCTOR_OPERAND_H

#include "ConstructorOperand.h"

#include <map>
#include <boost/serialization/serialization.hpp>
#include <boost/ptr_container/ptr_vector.hpp>
#include <boost/ptr_container/serialize_ptr_vector.hpp>


namespace giterator {

class IntegerConstant;

/**
 * The operand that is an array initializer.
 * For example in: 'int arr1[2][2]={{1,2}, {3,4}};'
 * {{1,2}, {3,4}} is the initializer.
 * These operands get created from GIMPLE @c CONSTRUCTOR
 * tree nodes, that have @c TREE_TYPE @c ARRAY_TYPE.
 * This class stores initialization value for each array index
 * that will be initialized.
 * GIMPLE @c CONSTRUCTOR nodes can define initialization value
 * for a range of indexes of an array. In this case a copy of
 * initialization value gets created for each index in the range
 * and it is stored to @c ValueOwners array to be properly destructed.
 *
 * @see create_git_operand_from_tree() in gimplexx-brideg.c to see how array constructor
 * operands get created from GIMPLE operands
 */
class ArrayConstructorOperand : public ConstructorOperand {
  private:
	/*! type of the container for mapping array indexes to initialization values. */
	typedef std::map<const IntegerConstant*, const Operand*> ArrayConstructors;
	/*! type of the container to store automatic pointers for duplicated
	 * initialization values. */
	typedef boost::ptr_vector<Operand> ValueOwners;

	/*! Field mapping array indexes to their initialization values.*/
	ArrayConstructors m_constructors;
	/*! Array holding auto pointers of copies of duplicate initialization values,
	 * for their proper destruction. */
	ValueOwners m_valueOwners;
  public:
	/** iterator for index->initializer mapper
	 * @see m_constructors
	 */
	typedef ArrayConstructors::const_iterator ArrayConstructorsIterator;

	/**
	 * Constructor, creates a new array constructor operand.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
    ArrayConstructorOperand () {}

    /**
	 * Constructor, creates a new array constructor operand.
	 *
	 * @param type the constructed type
	 */
    ArrayConstructorOperand (const Type* type)
    : ConstructorOperand(type) {}

    virtual void accept (Visitor& v) const
    { v.visit (*this); }

    /**
     * Obtain the array constructors held by this operand.
     *
     * @return pair of values - beginning of constructors map, end of constructors map
     */
    std::pair<ArrayConstructorsIterator, ArrayConstructorsIterator>
	constructors () const
	{ return std::make_pair (m_constructors.begin (), m_constructors.end ()); }

    /**
     * Add a initialization value for a specific index to the constructor map.
     *
     * @param index index in the array for the initialization value
     * @param value the initialization value for specified index
     */
    void addConstructor(const IntegerConstant* index, const Operand* value)
    {
    	m_constructors[index] = value;
    }

    /**
     * Store an initialization value with no value owner to value owners
     * maintained in this class, for the operand to be properly destructed
     *
     * @param operand the initializer with no value owner
     */
    void addValueOwner(std::auto_ptr<Operand> operand)
    {
    	m_valueOwners.push_back(operand);
    }

    friend class boost::serialization::access;
	template<class Archive>
	void save(Archive & ar, const unsigned int) const
	{
		ar << boost::serialization::base_object<detail::TypeOperand>(*this);
		size_t size = m_constructors.size();
		ar << size;
		for (ArrayConstructorsIterator i1 = m_constructors.begin(); i1 != m_constructors.end(); ++i1)
		{
			ar << i1->first;
			Operand* op = (Operand*) i1->second;
			ar << op;
		}
		ar << m_valueOwners;
	}

	template<class Archive>
	void load(Archive & ar, const unsigned int)
	{
		ar >> boost::serialization::base_object<detail::TypeOperand>(*this);

		size_t size;
		ar >> size;
		for (size_t i = 0; i < size; ++i)
		{
			IntegerConstant* ic;
			ar >> ic;
			Operand* op;
			ar >> op;
			m_constructors[ic] = op;
		}
		ar >> m_valueOwners;
	}

	template<class Archive>
	void serialize(Archive & ar, const unsigned int version)
	{
		boost::serialization::void_cast_register(static_cast<ArrayConstructorOperand *>(NULL), static_cast<Operand *>(NULL));
		boost::serialization::split_member(ar, *this, version);
	}
};

}

#endif /* GIT_ARRAY_CONSTRUCTOR_OPERAND_H */
