// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file ArrayRefOperand.h
 * Operand, that is an element of array.
 */
#ifndef GIT_ARRAY_REF_OPERAND_H
#define GIT_ARRAY_REF_OPERAND_H

#include "Operand.h"
#include "ArrayType.h"

#include <memory>
#include <boost/serialization/serialization.hpp>

namespace giterator {
    
/**
 * The operand that is an element of an array.
 * It has two sub operands: the array and the index. Type of this
 * operand is the type of the elements of the array.
 * These operands get created from GIMPLE @c ARRAY_REF
 * tree nodes.
 *
 * @see create_git_operand_from_tree in gimplexx-bridge.c to see how array element
 * operands get created from GIMPLE operands
 */
class ArrayRefOperand : public Operand {
  private:
    /*! operand representing the array */
    std::auto_ptr<Operand> m_array;
    /*! operand representing the index to the array */
    std::auto_ptr<Operand> m_index;
      
  public:
    /**
	 * Constructor, creates a new array operand.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
    ArrayRefOperand () {}
    /**
     * Constructor, creates new array operand.
     * 
     * @param array operand representing the array
     * @param index operand representing the index
     */
    ArrayRefOperand (std::auto_ptr<Operand> array, std::auto_ptr<Operand> index) :
        m_array (array), m_index (index) {}
    
    /**
     * Returns the operand of the array.
     * 
     * @return the operand of the array
     */
    const Operand* getArray () const 
    { return m_array.get (); }
    
    /** 
     * Returns the operand of the index to the array.
     * 
     * @return the operand of the index to the array
     */
    const Operand* getIndex () const
    { return m_index.get (); }
    
    virtual void accept (Visitor& v) const
    { v.visit (*this); }
    
    /**
     * Obtain the type of this operand - the type of the elements of the array.
     *
     * @return type of the elements of this array.
     */
    const Type* getType () const
    { 
        const ArrayType* arrayType = dynamic_cast<const ArrayType*> (m_array->getType ());
        assert (arrayType);
        return arrayType->getElementType ();
    }

    friend class boost::serialization::access;
	template<class Archive>
	void save(Archive & ar, const unsigned int) const
	{
		Operand* array = m_array.get();
		ar << array;
		Operand* index = m_index.get();
		ar << index;
	}

	template<class Archive>
	void load(Archive & ar, const unsigned int)
	{
		Operand* array;
		ar >> array;
		m_array = std::auto_ptr<Operand> (array);
		Operand* index;
		ar >> index;
		m_index = std::auto_ptr<Operand> (index);
	}

	template<class Archive>
	void serialize(Archive & ar, const unsigned int version)
	{
		boost::serialization::void_cast_register(static_cast<ArrayRefOperand *>(NULL), static_cast<Operand *>(NULL));
		boost::serialization::split_member(ar, *this, version);
	}
};

}

#endif /* GIT_ARRAY_REF_OPERAND_H */
