// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file ArrayType.h
 * Class for representing the type of an array.
 */
#ifndef GIT_ARRAY_TYPE_H
#define GIT_ARRAY_TYPE_H

#include "AbstractType.h"

#include <boost/serialization/serialization.hpp>

namespace giterator { 
    
/**
 * Class representing the type of an array.
 * It stores the type of the elements of an array.
 * This type gets created from GIMPLE @c ARRAY_TYPE
 * tree nodes.
 *
 * @see create_git_type_from_tree() in gimplexx-bridge.c
 * to see how array types get created from GIMPLE types
 */
class ArrayType : public detail::AbstractType {
  private:
    /*! type of the array elements */
    const Type* m_elementType;
    
  public:
    /**
	 * Constructor, creates a new array type.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
    ArrayType() : detail::AbstractType() {}

    /**
     * Constructor, creates a new array type.
     * 
     * @param name name of the type
     * @param elementType type of the elements of the new array type
     * @param bitSize size of the new type in bits
     */
    ArrayType (const std::string& name, const Type* elementType, 
               size_t bitSize)
    : AbstractType (name, bitSize), m_elementType (elementType)
	{}
    
    /**
     * Returns the type of array elements.
     * 
     * @return the type of array elements
     */
    const Type* getElementType () const
    { return m_elementType; }
    
    /**
     * Compare this type with another type.
     *
     * @param type type to be compared
     * @return true if the types are the same (both are
     * array types and type of their elements is the same),
     * false otherwise
     */
    bool isSame (const Type& type) const
    {
		const ArrayType* arrayType = dynamic_cast<const ArrayType*> (&type);
			if (!arrayType)
				return false;

			return m_elementType->isSame (*(arrayType->m_elementType)) &&
					getBitSize () == arrayType->getBitSize ();
    }

    friend class boost::serialization::access;
	template<class Archive>
	void save(Archive & ar, const unsigned int) const
	{
		ar << boost::serialization::base_object<AbstractType>(*this);
		ar << m_elementType;
	}

	template<class Archive>
	void load(Archive & ar, const unsigned int)
	{
		ar >> boost::serialization::base_object<AbstractType>(*this);
		Type* type;
		ar >> type;
		m_elementType = (const Type*) type;
	}

	template<class Archive>
	void serialize(Archive & ar, const unsigned int version)
	{
		boost::serialization::void_cast_register(static_cast<ArrayType *>(NULL), static_cast<TreeProduct *>(NULL));
		boost::serialization::void_cast_register(static_cast<ArrayType *>(NULL), static_cast<Type *>(NULL));
		boost::serialization::split_member(ar, *this, version);
	}
};

}

#endif /* GIT_ARRAY_TYPE_H */
