// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file AssignStmt.h
 * Class representing GIMPLE_ASSIGN statement.
 */
#ifndef GIT_ASSIGN_STATEMENT_H
#define GIT_ASSIGN_STATEMENT_H

#include "Stmt.h"
#include "Operand.h"
#include "Operation.h"
#include "Serialization.h"

#include <boost/noncopyable.hpp>
#include <memory>

namespace giterator {

/**
 * Class representing @c GIMPLE_ASSIGN statement.
 *
 * GIMPLE_ASSIGN looks like this:
 * GIMPLE_ASSIGN <OPERATION, LHS, RHS1[, RHS2]>.
 *
 * It is a representation of an assignment statement:
 * LHS = RHS1 OPERATION RHS2.
 *
 * LHS is the operand on the left-hand side of the assignment. It must be a lvalue.
 *
 * RHS1 is the first operand on the right-hand side of the assignment. It must always be
 * present.
 *
 * RHS2 is the second operand on the right-hand side of the assignment. It exists only for
 * binary operations.
 *
 * OPERATION is unary or binary operations to be performed on the right-hand side operations.
 *
 * In GIMPLE, operation tree nodes that may appear on the right-hand side are accepted by
 * function @c get_gimple_rhs_class(). It returns @c GIMPLE_BINARY_RHS for binary operations,
 * @c GIMPLE_UNARY_RHS for unary operations, @c GIMPLE_SINGLE_RHS if left-hand side is not
 * a gimple register ( according to @c is_gimple_reg() function). In this case, the operation
 * will be no operation (just a simple assignment).
 * Finally for operands, that are not valid right-hand side operations, @c get_gimple_rhs_class()
 * returns @c GIMPLE_INVALID_RHS.
 *
 * @see gimple_assign_to_git_statement() in gimplexx-bridge.c to see how an assign statement
 * gets created from GIMPLE tree node.
 */
class AssignStmt : public Stmt, private boost::noncopyable {
  private:
    /*! operand on the left hand side */
    std::auto_ptr<Operand> m_lhs;
    
    /*! operation on the right hand side */
    std::auto_ptr<Operation> m_operation;

  public:
    /**
	 * Constructor, creates a new assign statement.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
    AssignStmt() {}

    /**
	 * Constructor, creates a new assignement statement.
     * 
     * @param lhs operand on the left hand side; it must be an lvalue
     * @param operation operation on the right hand side
     *
     * @see Operation
     */
    AssignStmt (std::auto_ptr<Operand> lhs, std::auto_ptr<Operation> operation)
    	: m_lhs (lhs), m_operation (operation) {}

    void accept (Stmt::Visitor& visitor) const
    { visitor.visit (*this); }

    /**
     * Returns the operand on the left hand side of this statement.
     * 
     * @return the operand on the left hand side of this statement
     */
    const Operand* getLhs () const 
    { return m_lhs.get (); }

    /**
     * Return the operation on the right hand side of the statement.
     * 
     * @return the operation on the right hand side of the statement
     */
    const Operation* getOperation () const
    { return m_operation.get (); }

    friend class boost::serialization::access;
	template<class Archive>
	void save(Archive & ar, const unsigned int) const
	{
		Operand* lhs = m_lhs.get();
		ar << lhs;
		Operation* operation = m_operation.get();
		ar << operation;
	}

	template<class Archive>
	void load(Archive & ar, const unsigned int)
	{
		Operand* lhs;
		ar >> lhs;
		m_lhs = std::auto_ptr<Operand> (lhs);
		Operation* operation;
		ar >> operation;
		m_operation = std::auto_ptr<Operation> (operation);
	}

	template<class Archive>
	void serialize(Archive & ar, const unsigned int version)
	{
		boost::serialization::void_cast_register(static_cast<AssignStmt *>(NULL), static_cast<Stmt *>(NULL));
		boost::serialization::split_member(ar, *this, version);
	}
};

}

#endif /* GIT_ASSIGN_STATEMENT_H */

