// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file Block.h
 * Class representing block of GIMPLE statements.
 */
#ifndef GIT_BLOCK_H
#define GIT_BLOCK_H

#include "Stmt.h"
#include "Edge.h"
#include "ConstructionHelpers.h"

#include "../Error.h"
#include "../git/CTypes.h"

#include <boost/ptr_container/ptr_vector.hpp>
#include <boost/noncopyable.hpp>
#include <vector>
#include <algorithm>
#include <memory>
#include <boost/checked_delete.hpp>
#include <boost/serialization/serialization.hpp>
#include <boost/ptr_container/serialize_ptr_vector.hpp>


namespace giterator {
namespace detail {
    
typedef bool (Edge::*FlagGetter) () const;
    
/**
 * Helper functor testing whether the edge has a specific flag set.
 */
class HasFlagPred_ {
    FlagGetter m_flagGetter;
  public:
    HasFlagPred_ (FlagGetter flagGetter) : m_flagGetter (flagGetter) {}
    
    bool operator() (const Edge& e)
    { return (e.*m_flagGetter) (); }
};

}

/**
 * Class that represents block of GIMPLE statements.
 * Blocks of GIMPLE++ statements get created from
 * GIMPLE @c basic_blocks structures.
 * These structures represent a straight sequence of
 * code with only one entry point and only one exit.
 * Apart from statements in the basic block, this class
 * stores also successor and predecessor edges to next
 * and previous basic blocks.
 * List of Block classes represents the whole control-flow
 * graph of the program.
 *
 * @see basic_block_to_git_block() in gimplexx-bridge.c to see
 * how to create a block from GIMPLE basic block.
 */
class Block : private boost::noncopyable {
  private:
    /*! type of the container holding all statements in the block */
    typedef boost::ptr_vector<Stmt> Stmts;
    /*! type of the container holding all edges to successors */
    typedef boost::ptr_vector<Edge> SuccEdges;
    /*! type of the container holding all edges to predecessors */
    typedef std::vector<const Edge*> PredEdges;
    
    /*! statements in this block */
    Stmts m_stmts;
    
    /**
     * edges to successors.
     * This edges are owned by this object.
     */
    SuccEdges m_succEdges;
    
    /**
     * edges to predecessors.
     * This edges are NOT owned by this object
     * (they are owned as successors of another block).
     */
    PredEdges m_predEdges;
      
  public:
    /*! iterator through the statements in the block */
    typedef Stmts::const_iterator StmtIterator;
    /*! iterator through the edges to the successors of this block */
    typedef SuccEdges::const_iterator SuccEdgesIterator;
    /*! iterator through the edges to the predecessors of this block */
    typedef PredEdges::const_iterator PredEdgesIterator;
        
  protected:    
    /**
     * Adds an edge to the predecessor.
     * 
     * @param edge edge to the predecessor
     */
    void addPredEdge (const Edge* edge)
    { m_predEdges.push_back (edge); }

    /**
     * Adds an edge to the successor.
     * 
     * @param edge edge to the successor
     */
    void addSuccEdge (std::auto_ptr<Edge> edge)
    { m_succEdges.push_back (edge.release ()); }    

    /**
     * Returns one and only edge from the edges to the successors that has enabled the given flag.
     * 
     * @param f the pointer to member getter that returns the value of the flag
     * @return edge from the edges to the successors that has enabled flag @c f.
     *      This block must contain exactly one such edge.
     */
    const Edge* getSingleSuccEdgeWithFlag (detail::FlagGetter f) const
    {
        detail::HasFlagPred_ hasFlagPred (f);
        assert (std::count_if (m_succEdges.begin (), m_succEdges.end (), hasFlagPred) == 1);
    
        return &*std::find_if (m_succEdges.begin (), m_succEdges.end (), hasFlagPred);
    }
    
    /**
     * Adds statement at the end of this block.
     * 
     * @param stmt statement
     */
    void addStmt (std::auto_ptr<Stmt> stmt)
    { m_stmts.push_back (stmt.release ()); }

    
  public:
    /**
	 * Constructor, creates a new block.
	 */
    Block () {}
    /**
     * Returns the range of edges to the successors of this block.
     * 
     * @return the range of edges to the successors of this block
     */
    std::pair<SuccEdgesIterator, SuccEdgesIterator>
    succEdges () const
    { return std::make_pair (m_succEdges.begin (), m_succEdges.end ()); }
    
    /**
     * Returns the range of edges to the predecessors of this block.
     * 
     * @return the range of edges to the predecessors of this block
     */
    std::pair<PredEdgesIterator, PredEdgesIterator>
    predEdges () const
    { return std::make_pair (m_predEdges.begin (), m_predEdges.end ()); }
    
    /**
     * Returns the edge for the true branch.
     * The true branch edge leads to the block, that is executed when the @c GIMPLE_COND statement 
     * at the end of this block succeeds.
     * 
     * @return the edge to the true branch
     */
    const Edge* getTrueBranchEdge () const
    {
        return getSingleSuccEdgeWithFlag (&Edge::isTrueBranch);
    }
    
    /**
     * Returns the edge for the false branch.
     * The false branch edge leads to the block, that is executed when the @c GIMPLE_COND statement 
     * at the end of this block fails.
     * 
     * @return the edge to the false branch
     */
    const Edge* getFalseBranchEdge () const
    {
        return getSingleSuccEdgeWithFlag (&Edge::isFalseBranch);
    }
      
    /**
     * Returns the fallthru edge.
     * The fallthru edge leads to the block, that is executed after this block.
     * 
     * @return the fallthru edge
     */
    const Edge* getFallthruEdge () const
    {
        return getSingleSuccEdgeWithFlag (&Edge::isFallthru);
    }
    
    /**
     * Returns the range of statements in this block.
     * 
     * @return the range of statements in this block
     */
    std::pair<StmtIterator, StmtIterator>
    stmts () const
    { return std::make_pair (m_stmts.begin (), m_stmts.end ()); }
    
    friend void detail::blockAddSuccEdge (Block* block, Edge* edge);
    friend void detail::blockAddPredEdge (Block* block, Edge* edge);
    friend void detail::blockAddStmt (Block* block, Stmt* stmt);
  public:
    friend class boost::serialization::access;
	template<class Archive>
	void save(Archive & ar, const unsigned int) const
	{
		ar <<  m_stmts;
		ar << m_succEdges;
		size_t size = m_predEdges.size();
		ar << size;
		for(PredEdgesIterator i1 = m_predEdges.begin(); i1 != m_predEdges.end(); ++i1)
		{
			ar << (*i1);
		}
	}

	template<class Archive>
	void load(Archive & ar, const unsigned int)
	{
		ar >>  m_stmts;
		ar >> m_succEdges;
		size_t size;
		ar >> size;
		m_predEdges.resize(size);
		Edge* edge;
		for (size_t i = 0; i < size; ++i)
		{
			ar >> edge;
			m_predEdges[i] = (const Edge*) edge;
		}
	}

	template<class Archive>
	void serialize(Archive & ar, const unsigned int version)
	{
		boost::serialization::split_member(ar, *this, version);
	}
};

}

#endif /* GIT_BLOCK_H */
