// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file CallStmt.h
 * Class representing GIMPLE_CALL statement.
 */

#ifndef GIT_CALL_STMT_H
#define GIT_CALL_STMT_H

#include "Stmt.h"
#include "Operand.h"

#include <boost/noncopyable.hpp>
#include <boost/ptr_container/ptr_vector.hpp>
#include <boost/serialization/serialization.hpp>
#include <boost/ptr_container/serialize_ptr_vector.hpp>
#include <memory>
#include <algorithm>

namespace giterator { 

/**
 * Class representing @c GIMPLE_CALL statement.
 * GIMPLE_CALL looks like this:
 * GIMPLE_CALL <FN, LHS, ARG1, ..., ARGN>
 * It represents a function call.
 *
 * FN is the called function, LHS is an operand specifying where
 * the return value will be stored (it may be NULL for void functions).
 * ARG1, ... ARG2 are the function arguments. They must be all
 * accepted by @c is_gimple_operand() function.
 *
 * @see gimple_call_to_git_stmt() in gimplexx-bridge.c to see
 * how call statement gets created from GIMPLE call statements
 */
class CallStmt : public Stmt, private boost::noncopyable {
  private:
    /*! type of the container with parameters for the call */
    typedef boost::ptr_vector<Operand> Parameters;
      
    /*! the function that is called by this statement */
    std::auto_ptr<Operand> m_function;
    /*! the operand where to store the returned value */
    std::auto_ptr<Operand> m_lhs;
    
    /*! parameters for the call */
    Parameters m_parameters;
  public:
    /*! type of the iterator through parameters of the call */
    typedef Parameters::const_iterator ParameterIterator;

    /**
	 * Constructor, creates a new call statement.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
    CallStmt () {}
    /**
     * Constructor, creates a new call statement.
     * 
     * @param function function that is called by the statement
     * @param lhs operand where is the returned value stored; this must be lvalue
     */
    CallStmt (std::auto_ptr<Operand> function, std::auto_ptr<Operand> lhs) 
		: m_function (function), m_lhs (lhs) {}

    void accept (Stmt::Visitor& visitor) const
    { visitor.visit (*this); }

    /**
     * Returns the range of parameters of this call.
     * 
     * @return the range of parameters of this call
     */
    std::pair<ParameterIterator, ParameterIterator>
    parameters () const
    { return std::make_pair (m_parameters.begin (), m_parameters.end ()); }
    
    /**
     * Returns the function that is called by this call.
     * 
     * @return the function that is called by this call
     */
    const Operand* getFunction () const 
    { return m_function.get (); }
    
    /**
     * Returns the operand where is stored the returned value.
     * 
     * @return the operand where is stored the returned value
     */
    const Operand* getLhs () const
    { return m_lhs.get (); }
    
    /**
     * Adds parameter to this call.
     * After the initialization of this object, the number of parameters passed in by calling
     * this method and number of parameters of the function @c getFunction() must be the same.
     * 
     * @param param operand that is added to the parameters of this call
     */
    void addParameter (std::auto_ptr<Operand> param)
    { 
        m_parameters.push_back (param); 
    }

    friend class boost::serialization::access;
	template<class Archive>
	void save(Archive & ar, const unsigned int) const
	{
		Operand* function = m_function.get();
		ar << function;
		Operand* lhs = m_lhs.get();
		ar << lhs;
		ar << m_parameters;
	}

	template<class Archive>
	void load(Archive & ar, const unsigned int)
	{
		Operand* function;
		ar >> function;
		m_function = std::auto_ptr<Operand> (function);
		Operand* lhs;
		ar >> lhs;
		m_lhs = std::auto_ptr<Operand> (lhs);
		ar >> m_parameters;
	}

	template<class Archive>
	void serialize(Archive & ar, const unsigned int version)
	{
		boost::serialization::void_cast_register(static_cast<CallStmt *>(NULL), static_cast<Stmt *>(NULL));
		boost::serialization::split_member(ar, *this, version);
	}
};

}

#endif /* GIT_CALL_STMT_H */

