// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file ComponentRefOperand.h
 * Class representing operand that is an field of struct/union.
 */
#ifndef GIT_COMPONENT_REF_OPERAND_H
#define GIT_COMPONENT_REF_OPERAND_H

#include "Operand.h"

#include <memory>
#include <boost/serialization/serialization.hpp>

namespace giterator {

/**
 * The operand that is an field of struct/union.
 * These operands get created from GIMPLE @c COMPONENT_REF
 * nodes.
 * This operand is a @c FieldDecl.
 *
 * @see create_git_operand_from_tree to see how field of struct/union
 * operands get created from GIMPLE operands
 */
class ComponentRefOperand : public Operand {
  private:
    /*! operand representing the struct/union */
    std::auto_ptr<Operand> m_record;
    /*! operand representing the field of struct/union */
    std::auto_ptr<Operand> m_field;

  public:
    /**
	 * Constructor, creates a new operand that is an field of struct/union.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
    ComponentRefOperand () {}
    /**
     * Constructor, creates a new operand that is an field of struct/union.
     *
     * @param record structure/union, where the field is situated
     * @param field field represented by this operand
     */
    ComponentRefOperand (std::auto_ptr<Operand> record, std::auto_ptr<Operand> field) :
        m_record (record), m_field (field) {}

    /**
     * Returns the operand of the array.
     *
     * @return the operand of the array
     */
    const Operand* getRecord () const
    { return m_record.get (); }

    /**
     * Returns the operand of the index to the array.
     *
     * @return the operand of the index to the array
     */
    const Operand* getField () const
    { return m_field.get (); }

    virtual void accept (Visitor& v) const
    { v.visit (*this); }

    /**
     * Get type of the field represented by this operand.
     *
     * @return type of the field represented by this operand.
     */
    const Type* getType () const
    {
        return m_field->getType ();
    }

    friend class boost::serialization::access;
	template<class Archive>
	void save(Archive & ar, const unsigned int) const
	{
		Operand* record = m_record.get();
		ar << record;
		Operand* field = m_field.get();
		ar << field;
	}

	template<class Archive>
	void load(Archive & ar, const unsigned int)
	{
		Operand* record;
		ar >> record;
		m_record = std::auto_ptr<Operand> (record);
		Operand* field;
		ar >> field;
		m_field = std::auto_ptr<Operand> (field);
	}

	template<class Archive>
	void serialize(Archive & ar, const unsigned int version)
	{
		boost::serialization::void_cast_register(static_cast<ComponentRefOperand *>(NULL), static_cast<Operand *>(NULL));
		boost::serialization::split_member(ar, *this, version);
	}
};

}

#endif /* GIT_COMPONENT_REF_OPERAND_H */
