// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file CondStmt.h
 * Class representing GIMPLE_COND statement.
 */
#ifndef GIT_COND_STATEMENT_H
#define GIT_COND_STATEMENT_H

#include "Stmt.h"
#include "Operation.h"
#include "Serialization.h"

#include <boost/noncopyable.hpp>
#include <memory>

namespace giterator { 
    
/**
 * Class representing @c GIMPLE_COND statement.
 * GIMPLE_COND looks like this:
 * GIMPLE_COND <COND_CODE, OP1, OP2, TRUE_LABEL, FALSE_LABEL>
 *  It represents the conditional jump:
 *
 *  if (OP1 COND_CODE OP2)
 *  	goto TRUE_LABEL
 *  else
 *  	goto FALSE_LABEL
 *
 * COND_CODE is the operation used as the comparison.
 *
 * OP1 and OP2 are the operands used in the comparison.  They must be
 * accepted by @c is_gimple_operand() function.
 *
 * @c TRUE_LABEL and @c FALSE_LABEL are the @c LABEL_DECL
 * nodes used as the jump target for the comparison.
 *
 * @see gimple_cond_to_git_stmt() to see how conditional statements
 * get created from GIMPLE conditional statements.
 */
class CondStmt : public Stmt, private boost::noncopyable {
  private:
    /*! the operation that evaluates the condition */
    std::auto_ptr<Operation> m_oper;

  public:
    /**
	 * Constructor, creates a new conditional statement.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
    CondStmt() {}
    /**
     * Constructor, creates a new conditional statement.
     * 
     * @param oper operation that evaluates the condition
     */
    CondStmt (std::auto_ptr<Operation> oper)
    	: m_oper (oper) {}

    void accept (Stmt::Visitor& visitor) const
    { visitor.visit (*this); }

    /**
     * Returns the operation that evaluates the condition.
     * 
     * @return the operation that evaluates the condition
     */
    const Operation* getOperation () const 
    { return m_oper.get (); }

    friend class boost::serialization::access;
	template<class Archive>
	void save(Archive & ar, const unsigned int) const
	{
		Operation* oper = m_oper.get();
		ar << oper;
	}

	template<class Archive>
	void load(Archive & ar, const unsigned int)
	{
		Operation* oper;
		ar >> oper;
		m_oper = std::auto_ptr<Operation> (oper);
	}

	template<class Archive>
	void serialize(Archive & ar, const unsigned int version)
	{
		boost::serialization::void_cast_register(static_cast<CondStmt *>(NULL), static_cast<Stmt *>(NULL));
		boost::serialization::split_member(ar, *this, version);
	}
};

}


#endif /* GIT_COND_STATEMENT_H */

