// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file ConstDecl.h
 * Class representing declaration of enumeration constant.
 */

#ifndef GIT_CONST_DECL_H
#define GIT_CONST_DECL_H

#include "AbstractDecl.h"
#include "Type.h"

#include <boost/serialization/serialization.hpp>

namespace giterator {

/**
 * Class representing declaration of enumeration constant.
 * Enumeration constants get created from GIMPLE @c CONST_DECL
 * nodes.
 * These are integer constants with enumeral type.
 *
 * @see create_git_decl_from_tree to see how declarations of
 * enumeration constants get created from GIMPLE declarations
 */
class ConstDecl : public detail::AbstractDecl {
  private:
	/*! Flag determining whether this declaration is local or global.*/
	bool m_isLocal;
  public:
	/**
	 * Constructor, creates a new constant declaration.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	ConstDecl () : detail::AbstractDecl() {}
	/**
	 * Constructor, creates a new constant declaration.
	 *
     * @param uid unique ID of the new declaration
     * @param name name of the new declaration
     * @param type type of the new declaration
     * @param isLocal whether this is local declaration (true),
     * or global declaration (false)
     */
    ConstDecl (unsigned int uid, const std::string& name, const Type* type, bool isLocal) :
                detail::AbstractDecl (uid, name, type), m_isLocal (isLocal) {}

    /**
     * Find out whether this is local or global declaration.
     *
     * @return true if this is local declaration, false otherwise
     */
    bool getLocal () const
    { return m_isLocal; }

    void accept (Visitor& visitor) const
    { visitor.visit (*this); }

    friend class boost::serialization::access;
	template<class Archive>
	void save(Archive & ar, const unsigned int) const
	{
		ar << boost::serialization::base_object<AbstractDecl>(*this);
		ar << m_isLocal;
	}

	template<class Archive>
	void load(Archive & ar, const unsigned int)
	{
		ar >> boost::serialization::base_object<AbstractDecl>(*this);
		ar >> m_isLocal;
	}

	template<class Archive>
	void serialize(Archive & ar, const unsigned int version)
	{
		boost::serialization::void_cast_register(static_cast<ConstDecl *>(NULL), static_cast<TreeProduct *>(NULL));
		boost::serialization::void_cast_register(static_cast<ConstDecl *>(NULL), static_cast<Decl *>(NULL));
		boost::serialization::split_member(ar, *this, version);
	}
};

} /* GIT_CONST_DECL_H */

#endif
