// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file ConstructionHelpers.cpp
 * Definitions of helper functions that allow C code to construct
 * GIMPLE++ objects.
 */
#include "ConstructionHelpers.h"

#include "Gimple.h"
#include "GimpleHelpers.h"

#include "TreeProduct.h"
#include "Function.h"
#include "Type.h"

#include "Decls.h"
#include "Types.h"
#include "Constants.h"
#include "Operations.h"
#include "Operands.h"
#include "Stmts.h"


#include <memory>
#include <sstream>

namespace giterator { namespace detail {

#define INT_NAME_BASE "int"
#define REAL_NAME_BASE "real"
#define PTR_NAME_BASE "ptr"
#define BOOL_NAME_BASE "bool"
#define FUNC_NAME_BASE "func"
#define ENUM_NAME_BASE "enum"
#define RECORD_NAME_BASE "record"
#define ARRAY_NAME_BASE "arr"
#define VOID_NAME_BASE "void"

/**
 * Returns the name of a declaration.
 * 
 * @param name name of the declaration taken from GIMPLE tree
 * @param uid unique ID of the declaration
 */
static std::string
getDeclName (const char* name, unsigned int uid)
{
    if (name)
        return name;
    else
    {
        std::stringstream strStream;
        strStream << "D." << uid;
        return strStream.str ();
    }
}


/**
 * Creates @c std::auto_ptr<T>.
 * 
 * @tparam T a type of the value 
 * @param t pointer to @c T that will be stored in the new auto pointer
 * @return the new auto pointer
 */
template <typename T>
static std::auto_ptr<T>
makeAP (T* t)
{
    return std::auto_ptr<T> (t);
}

static git_gimple_construction
getConstructionInstance() {
    static std::auto_ptr<GimpleConstruction> construction;

    if (!construction.get()) {
        construction = makeAP(new GimpleConstruction());
        construction->gimple = new Gimple();
        construction->gimple_helpers = new GimpleHelpers();
        construction->function_helpers = NULL;
    }

    return construction.get();
}

git_gimple_construction
beginFunctionConstruction() {
    GimpleConstruction* construction = getConstructionInstance();

    if (construction->function_helpers) {
        delete construction->function_helpers;
        construction->function_helpers = NULL;
    }

    construction->function_helpers = new FunctionHelpers();

    return construction;
}

extern void
endFunctionConstruction() {
    GimpleConstruction* construction = getConstructionInstance();

    if (construction->function_helpers) {
        delete construction->function_helpers;
        construction->function_helpers = NULL;
    }
}

extern git_gimple
getConstructionResult() {
    return getConstructionInstance()->gimple;
}

void
endConstruction() {
    GimpleConstruction* construction = getConstructionInstance();

    delete construction->gimple;
    construction->gimple = NULL;
    delete construction->gimple_helpers;
    construction->gimple_helpers = NULL;
    if (construction->function_helpers) {
        delete construction->function_helpers;
        construction->function_helpers = NULL;
    }
}

/* Declarations */

git_decl 
createFunctionDecl (git_gimple_helpers gim_help, unsigned int uid, const char* name, const_git_type type)
{
    std::string declName = getDeclName (name, uid);
    return new FunctionDecl (uid, declName, type, gim_help->getFunctionCode(declName));
}

git_decl
createVarDecl (unsigned int uid, const char* name, const_git_type type, bool local)
{
    std::string declName = getDeclName (name, uid);

    if (typeid(*type) == typeid(const RecordType))
    	return new RecordDecl (uid, declName, type, local);
    else
    	return new VarDecl (uid, declName, type, local);
}

git_decl
createRecordDecl (unsigned int uid, const char* name, const_git_type type, bool local)
{
	std::string declName = getDeclName (name, uid);
	return new RecordDecl (uid, declName, type, local);
}

git_decl
createParamDecl (unsigned int uid, const char* name, const_git_type type)
{
    std::string declName = getDeclName (name, uid);
    return new ParamDecl (uid, declName, type);
}

git_decl
createLabelDecl (unsigned int uid, const char* name, const_git_type type)
{
    std::string declName = getDeclName (name, uid);
    return new LabelDecl (uid, declName, type);
}

git_decl
createResultDecl (unsigned int uid, const char* name, const_git_type type)
{
    std::string declName = getDeclName (name, uid);
    return new ResultDecl (uid, declName, type);
}

git_decl
createTypeDecl (unsigned int uid, const char* name, const_git_type type, bool local)
{
    std::string declName = getDeclName (name, uid);
    return new TypeDecl (uid, declName, type, local);
}

git_decl
createFieldDecl (unsigned int uid, const char* name, const_git_type type)
{
    std::string declName = getDeclName (name, uid);
    return new FieldDecl (uid, declName, type);
}

git_decl
createConstDecl (unsigned int uid, const char* name, const_git_type type, bool local)
{
    std::string declName = getDeclName (name, uid);
    return new ConstDecl (uid, declName, type, local);
}

/* Function */

git_function
createFunction (const_git_type type)
{
    const FunctionType* funcType = dynamic_cast<const FunctionType*> (type);
    assert (funcType);
    
    return new Function (funcType);
}

void
functionAddLocalDecl (git_function function, const_git_decl decl)
{
    function->addLocalDecl (decl);
}

void
functionAddParamDecl (git_function function, const_git_decl paramDecl)
{
    function->addParamDecl (paramDecl);
}

void 
functionAddBlock (git_function function, git_function_helpers func_help, basic_block basicBlock, git_block block)
{
    func_help->addBlock (basicBlock, block);
    function->addBlock(std::auto_ptr<Block> (block));
}

void
functionAddBlock2 (git_function function, git_block block)
{
	function->addBlock(std::auto_ptr<Block> (block));
}

void 
functionSetFirstBlock (git_function function, git_function_helpers func_help, basic_block basicBlock)
{
	func_help->setFirstBlock (basicBlock);
	function->setFirstBlock (func_help->getFirstBlock());
}

void functionSetFirstBlock2 (git_function function, git_block block)
{
	function->setFirstBlock (block);
}

git_block
functionGetBlock (git_function_helpers func_help, basic_block basicBlock)
{
    return func_help->getBlock (basicBlock);
}

git_block functionGetFirstBlock(git_function function)
{
	return function->getFirstBlock();
}

git_block
createBlock (void)
{
    return new Block ();
}

void
blockAddStmt (git_block block, git_stmt stmt)
{
    block->addStmt (std::auto_ptr<Stmt> (stmt));
}

void
blockAddPredEdge (git_block block, git_edge edge)
{
    block->addPredEdge (edge);
}

void
blockAddSuccEdge (git_block block, git_edge edge)
{
    block->addSuccEdge (std::auto_ptr<Edge>(edge));
}

git_edge
createEdge (git_block src, git_block dest, bool isFallthru, bool isTrueBranch, bool isFalseBranch)
{
    return new Edge (src, dest, isFallthru, isTrueBranch, isFalseBranch);
}

git_type
createIntegerType (size_t bitSize, bool isUnsigned, int precision)
{
	std::stringstream sstream;
	sstream << isUnsigned << INT_NAME_BASE << bitSize << precision;

    return new IntegerType (sstream.str(), bitSize, isUnsigned, precision);
}

git_type
createRealType(size_t bitSize, int precision)
{
	std::stringstream sstream;
	sstream << REAL_NAME_BASE << bitSize << precision;
	return new RealType(sstream.str(), bitSize, precision);
}

git_type
createPtrType (size_t bitSize, const_git_type pointingToType)
{
	std::stringstream sstream;
	sstream << PTR_NAME_BASE << bitSize << pointingToType->getName();
	LOG_DEBUG ("%s", sstream.str().c_str());

    return new PtrType (sstream.str(), bitSize, pointingToType);
}

git_type
createVoidType ()
{
    return new VoidType (VOID_NAME_BASE);
}

git_type
createBoolType (size_t bitSize)
{
	std::stringstream sstream;
	sstream << BOOL_NAME_BASE << bitSize;

    return new BoolType (sstream.str(), bitSize);
}

git_type 
createFunctionType (git_gimple_helpers gim_help, const char* name, const_git_type returnValue)
{
	std::string funcTypeName;
	if (name == NULL)
	{
		std::stringstream sstream;
		sstream << FUNC_NAME_BASE << gim_help->getFuncNum() << returnValue->getName();
		funcTypeName = sstream.str();
	}
	else
		funcTypeName = name;

    return new FunctionType (funcTypeName, returnValue);
}

void
functionTypeAddParameter (git_type funcType, const_git_type argType)
{
    FunctionType* type = dynamic_cast<FunctionType*> (funcType);
    assert (type);
    type->addParameterType (argType);

}

git_type createEnumType (git_gimple_helpers gim_help, const char* name, size_t bitSize, bool isUnsigned, int precision)
{
	std::string enumName;
	if (name == NULL)
	{
		std::stringstream sstream;
		sstream << isUnsigned << ENUM_NAME_BASE << gim_help->getEnumNum() << bitSize << precision;
		enumName = sstream.str();
	}
	else
		enumName = name;

	return new EnumType(enumName, bitSize, isUnsigned, precision);
}
void enumTypeAddConstant(git_type enumType, const char* eName, const_git_constant eConstant)
{
	EnumType* type = dynamic_cast<EnumType*> (enumType);
	assert (type);
	type->addConstant(eName, eConstant);
}

git_type createRecordType (git_gimple_helpers gim_help, const char* name, size_t bitSize)
{
	std::string recName;
	if (name == NULL)
	{
		std::stringstream sstream;
		sstream << RECORD_NAME_BASE << gim_help->getRecordNum() << bitSize;
		recName = sstream.str();
	}
	else
		recName = name;

	return new RecordType(recName, bitSize);
}

void recordTypeAddField(git_type recordType, const_git_decl field)
{
	RecordType* type = dynamic_cast<RecordType*> (recordType);
	assert (type);

	type->addField(field->getName(), field);
}

void recordTypeAddLocalDecl(git_type recordType, const_git_decl localDecl)
{
	RecordType* type = dynamic_cast<RecordType*> (recordType);
	assert (type);

	type->addLocalDecl(localDecl->getName(), localDecl);
}

git_type 
createArrayType (size_t bitSize, const_git_type elementType)
{
	std::stringstream sstream;
	sstream << ARRAY_NAME_BASE << bitSize << elementType->getName();

    return new ArrayType (sstream.str(), elementType, bitSize);
}

/* Statements */

void
callStmtAddArg (git_stmt callStmt, git_operand param)
{
    CallStmt* stmt = dynamic_cast<CallStmt*> (callStmt);
    assert (stmt);

    stmt->addParameter (std::auto_ptr<Operand> (param));
}

git_stmt
createAssignStmt (git_operand lhs, git_operation operation)
{
    return new AssignStmt (std::auto_ptr<Operand> (lhs), std::auto_ptr<Operation> (operation));
}

git_stmt
createCondStmt (git_operation boolOperation)
{
    return new CondStmt (std::auto_ptr<Operation> (boolOperation));
}

git_stmt
createCallStmt (git_operand fn, git_operand lhs)
{
    return new CallStmt (std::auto_ptr<Operand>(fn), std::auto_ptr<Operand>(lhs));
}

git_stmt
createReturnStmt (git_operand retval)
{
    return new ReturnStmt (std::auto_ptr<Operand> (retval));
}

git_stmt
createGotoStmt (git_operand dest)
{
	return new GotoStmt(std::auto_ptr<Operand> (dest));
}
git_stmt
createLabelStmt (git_operand label)
{
	return new LabelStmt(std::auto_ptr<Operand> (label));
}
git_stmt
createNopStmt (void)
{
	return new NopStmt();
}
git_stmt
createPhiStmt (git_operand result)
{
	return new PhiStmt(std::auto_ptr<Operand> (result));
}

void
phiStmtAddArg (git_stmt phiStmt, git_operand arg)
{
	PhiStmt* stmt = dynamic_cast<PhiStmt*> (phiStmt);
	assert (stmt);

	stmt->addArgument(std::auto_ptr<Operand> (arg));
}

git_stmt
createSwitchStmt (git_operand index)
{
	return new SwitchStmt(std::auto_ptr<Operand> (index));
}

void
switchStmtAddLabel(git_stmt switchStmt, git_operand label)
{
	SwitchStmt* stmt = dynamic_cast<SwitchStmt*> (switchStmt);
	assert(stmt);

	stmt->addLabel(std::auto_ptr<Operand> (label));
}

/* Constants */

git_constant
createIntCst (const_git_type type, uint64_t value)
{
    return new IntegerConstant (type, value);
}

git_constant
createEnumCst (const_git_type type, uint64_t value)
{
    return new EnumeralConstant (type, value);
}

git_constant
createRealCst (const_git_type type, long double value)
{
    return new RealConstant (type, value);
}

git_constant
createStringCst (const_git_type type, int , const char* str)
{
    // TODO Fix reading strings that are not terminated by '\0' or contain '\0';
    return new StringConstant (type, str);
}

/* Operands */

git_operand
createDeclOperand (const_git_decl decl)
{
    return new DeclOperand (decl);
}

git_operand
createAddrOperand (const_git_type type, git_operand op)
{
    return new AddrOperand (type, std::auto_ptr<Operand> (op));
}

git_operand
createLabelExprOperand (const_git_decl decl, git_operand lower, git_operand upper, bool isDefault, bool isRange)
{
	if (isRange)
		return new LabelExprOperand (decl, std::auto_ptr<Operand> (lower), std::auto_ptr<Operand> (upper), isDefault);
	else
		return new LabelExprOperand (decl, std::auto_ptr<Operand> (lower), isDefault);
}

git_operand
createIndirectRefOperand (const_git_type type, git_operand op)
{
    return new IndirectRefOperand (type, std::auto_ptr<Operand> (op));
}

git_operand
createArrayRefOperand (git_operand array, git_operand index)
{
    return new ArrayRefOperand (std::auto_ptr<Operand> (array), std::auto_ptr<Operand> (index));
}

git_operand
createComponentRefOperand (git_operand record, git_operand field)
{
	return new ComponentRefOperand (std::auto_ptr<Operand> (record), std::auto_ptr<Operand> (field));
}

git_operand
createCstOperand (const_git_constant cst)
{
    return new CstOperand (cst);
}

git_operand
createRecordConstructorOperand (const_git_type type)
{
	return new RecordConstructorOperand (type);
}

git_operand
createArrayConstructorOperand (const_git_type type)
{
	return new ArrayConstructorOperand (type);
}

void
recordConstructorOperandAddField(git_operand recordOperand, const_git_decl decl, git_operand operand)
{
	RecordConstructorOperand* rco = dynamic_cast<RecordConstructorOperand*> (recordOperand);
	assert(rco);
	const FieldDecl* field = dynamic_cast<const FieldDecl *> (decl);
	assert(field);

	rco->addConstructor(field, std::auto_ptr<Operand> (operand));
}

void
arrayConstructorOperandAddField(git_operand arrayOperand, const_git_constant index, git_operand operand)
{
	ArrayConstructorOperand* aco = dynamic_cast<ArrayConstructorOperand*> (arrayOperand);
	assert(aco);
	const IntegerConstant* ic = dynamic_cast<const IntegerConstant*> (index);
	assert(index);

	aco->addConstructor(ic, operand);
	aco->addValueOwner(std::auto_ptr<Operand> (operand));
}

void
createAndAddConstantsBetween(git_gimple gimple, git_operand arrayOperand, git_operand operand, git_constant lower, git_constant upper)
{
	ArrayConstructorOperand* aco = dynamic_cast<ArrayConstructorOperand*> (arrayOperand);
	assert(aco);
	const IntegerConstant* low = dynamic_cast<const IntegerConstant*> (lower);
	assert(low);
	const IntegerConstant* up = dynamic_cast<const IntegerConstant*> (upper);
	assert(up);

	aco->addConstructor(low, operand);
	aco->addConstructor(up, operand);

	while (uint64_t val = low->getValue() != up->getValue())
	{
		++val;

		IntegerConstant* ic = new IntegerConstant(low->getType(), val);
		reportConstant(gimple, ic);

		aco->addConstructor(ic, operand);
	}

	std::auto_ptr<Operand> op(operand);
	aco->addValueOwner(std::auto_ptr<Operand> (operand));
}

/* Operations */

/**
 * Helper function that creates GIMPLE++ unary operation.
 * 
 * @param returnType return type of the operation
 * @param op operand of the unary operation
 * @return new GIMPLE++ operation
 */
template <class Op>
static git_operation
makeOp1 (const_git_type returnType, git_operand op)
{
    return new Op (returnType, makeAP (op));
}


/**
 * Helper function that creates GIMPLE++ binary operation.
 * 
 * @param returnType return type of the operation
 * @param op1 first operand of the operation
 * @param op2 second operand of the operation
 * @return new GIMPLE++ operation
 */
template <class Op>
static git_operation
makeOp2 (const_git_type returnType, git_operand op1, git_operand op2)
{
    return new Op (returnType, makeAP (op1), makeAP (op2));
}

git_operation createNegateOperation (const_git_type return_type, git_operand op)
{
    return makeOp1<NegateOperation> (return_type, op);
}

git_operation createAbsOperation (const_git_type return_type, git_operand op)
{
    return makeOp1<AbsOperation> (return_type, op);
}

git_operation createFixTruncOperation(const_git_type return_type, git_operand op)
{
    return makeOp1<FixTruncOperation> (return_type, op);
}

git_operation createFloatOperation(const_git_type return_type, git_operand op)
{
    return makeOp1<FloatOperation> (return_type, op);
}

git_operation createBitNotOperation(const_git_type return_type, git_operand op)
{
    return makeOp1<BitNotOperation> (return_type, op);
}

git_operation createTruthNotOperation(const_git_type return_type, git_operand op)
{
    return makeOp1<TruthNotOperation> (return_type, op);
}

git_operation createParenOperation(const_git_type return_type, git_operand op)
{
    return makeOp1<ParenOperation> (return_type, op);
}

git_operation createNonLvalueOperation(const_git_type return_type, git_operand op)
{
    return makeOp1<NonLvalueOperation> (return_type, op);
}

git_operation createNoOperation (const_git_type return_type, git_operand op)
{
    return makeOp1<NoOperation> (return_type, op);
}

git_operation createConvertOperation (const_git_type return_type, git_operand op)
{
    return makeOp1<ConvertOperation> (return_type, op);
}

git_operation createPlusOperation (const_git_type return_type, git_operand op1, git_operand op2)
{
    return makeOp2<PlusOperation> (return_type, op1, op2);
}

git_operation createMinusOperation (const_git_type return_type, git_operand op1, git_operand op2)
{
    return makeOp2<MinusOperation> (return_type, op1, op2);
}

git_operation createMultOperation (const_git_type return_type, git_operand op1, git_operand op2)
{
    return makeOp2<MultOperation> (return_type, op1, op2);
}

git_operation createRdivOperation (const_git_type return_type, git_operand op1, git_operand op2)
{
    return makeOp2<RdivOperation> (return_type, op1, op2);
}

git_operation createTruncDivOperation (const_git_type return_type, git_operand op1, git_operand op2)
{
    return makeOp2<TruncDivOperation> (return_type, op1, op2);
}

git_operation createFloorDivOperation (const_git_type return_type, git_operand op1, git_operand op2)
{
    return makeOp2<FloorDivOperation> (return_type, op1, op2);
}

git_operation createCeilDivOperation (const_git_type return_type, git_operand op1, git_operand op2)
{
    return makeOp2<CeilDivOperation> (return_type, op1, op2);
}

git_operation createRoundDivOperation (const_git_type return_type, git_operand op1, git_operand op2)
{
    return makeOp2<RoundDivOperation> (return_type, op1, op2);
}

git_operation createTruncModOperation (const_git_type return_type, git_operand op1, git_operand op2)
{
    return makeOp2<TruncModOperation> (return_type, op1, op2);
}

git_operation createFloorModOperation (const_git_type return_type, git_operand op1, git_operand op2)
{
    return makeOp2<FloorModOperation> (return_type, op1, op2);
}

git_operation createCeilModOperation (const_git_type return_type, git_operand op1, git_operand op2)
{
    return makeOp2<CeilModOperation> (return_type, op1, op2);
}

git_operation createRoundModOperation (const_git_type return_type, git_operand op1, git_operand op2)
{
    return makeOp2<RoundModOperation> (return_type, op1, op2);
}

git_operation createExactDivOperation (const_git_type return_type, git_operand op1, git_operand op2)
{
    return makeOp2<ExactDivOperation> (return_type, op1, op2);
}

git_operation createMinOperation (const_git_type return_type, git_operand op1, git_operand op2)
{
    return makeOp2<MinOperation> (return_type, op1, op2);
}

git_operation createMaxOperation (const_git_type return_type, git_operand op1, git_operand op2)
{
    return makeOp2<MaxOperation> (return_type, op1, op2);
}

git_operation createLshiftOperation (const_git_type return_type, git_operand op1, git_operand op2)
{
    return makeOp2<LshiftOperation> (return_type, op1, op2);
}

git_operation createRshiftOperation (const_git_type return_type, git_operand op1, git_operand op2)
{
    return makeOp2<RshiftOperation> (return_type, op1, op2);
}

git_operation createLrotateOperation (const_git_type return_type, git_operand op1, git_operand op2)
{
    return makeOp2<LrotateOperation> (return_type, op1, op2);
}

git_operation createRrotateOperation (const_git_type return_type, git_operand op1, git_operand op2)
{
    return makeOp2<RrotateOperation> (return_type, op1, op2);
}

git_operation createBitIorOperation(const_git_type return_type, git_operand op1, git_operand op2)
{
    return makeOp2<BitIorOperation> (return_type, op1, op2);
}

git_operation createBitXorOperation(const_git_type return_type, git_operand op1, git_operand op2)
{
    return makeOp2<BitXorOperation> (return_type, op1, op2);
}

git_operation createBitAndOperation(const_git_type return_type, git_operand op1, git_operand op2)
{
    return makeOp2<BitAndOperation> (return_type, op1, op2);
}

git_operation createTruthAndOperation(const_git_type return_type, git_operand op1, git_operand op2)
{
    return makeOp2<TruthAndOperation> (return_type, op1, op2);
}

git_operation createTruthOrOperation(const_git_type return_type, git_operand op1, git_operand op2)
{
    return makeOp2<TruthOrOperation> (return_type, op1, op2);
}

git_operation createTruthXorOperation(const_git_type return_type, git_operand op1, git_operand op2)
{
    return makeOp2<TruthXorOperation> (return_type, op1, op2);
}

git_operation createLtOperation (const_git_type return_type, git_operand op1, git_operand op2)
{
    return makeOp2<LtOperation> (return_type, op1, op2);
}

git_operation createLeOperation (const_git_type return_type, git_operand op1, git_operand op2)
{
    return makeOp2<LeOperation> (return_type, op1, op2);
}

git_operation createGtOperation (const_git_type return_type, git_operand op1, git_operand op2)
{
    return makeOp2<GtOperation> (return_type, op1, op2);
}

git_operation createGeOperation (const_git_type return_type, git_operand op1, git_operand op2)
{
    return makeOp2<GeOperation> (return_type, op1, op2);
}

git_operation createEqOperation (const_git_type return_type, git_operand op1, git_operand op2)
{
    return makeOp2<EqOperation> (return_type, op1, op2);
}

git_operation createNeOperation (const_git_type return_type, git_operand op1, git_operand op2)
{
    return makeOp2<NeOperation> (return_type, op1, op2);
}

git_operation createOrderedOperation (const_git_type return_type, git_operand op1, git_operand op2)
{
    return makeOp2<OrderedOperation> (return_type, op1, op2);
}

git_operation createUnorderedOperation (const_git_type return_type, git_operand op1, git_operand op2)
{
    return makeOp2<UnorderedOperation> (return_type, op1, op2);
}

git_operation createUnLtOperation (const_git_type return_type, git_operand op1, git_operand op2)
{
    return makeOp2<UnLtOperation> (return_type, op1, op2);
}

git_operation createUnLeOperation (const_git_type return_type, git_operand op1, git_operand op2)
{
    return makeOp2<UnLeOperation> (return_type, op1, op2);
}

git_operation createUnGtOperation (const_git_type return_type, git_operand op1, git_operand op2)
{
    return makeOp2<UnGtOperation> (return_type, op1, op2);
}

git_operation createUnGeOperation (const_git_type return_type, git_operand op1, git_operand op2)
{
    return makeOp2<UnGeOperation> (return_type, op1, op2);
}

git_operation createUnEqOperation (const_git_type return_type, git_operand op1, git_operand op2)
{
    return makeOp2<UnEqOperation> (return_type, op1, op2);
}

git_operation createLtGtOperation (const_git_type return_type, git_operand op1, git_operand op2)
{
    return makeOp2<LtGtOperation> (return_type, op1, op2);
}

git_operation createWidenSumOperation (const_git_type return_type, git_operand op1, git_operand op2)
{
    return makeOp2<WidenSumOperation> (return_type, op1, op2);
}

git_operation createWidenMultOperation (const_git_type return_type, git_operand op1, git_operand op2)
{
    return makeOp2<WidenMultOperation> (return_type, op1, op2);
}

/* Misc */

const_git_decl
get_git_decl_from_tree_cache (git_gimple_helpers gim_help, tree tree)
{
    return gim_help->getFromTreeCache<Decl> (tree);
}

const_git_constant 
get_git_constant_from_tree_cache (git_gimple_helpers gim_help, tree tree)
{
    return gim_help->getFromTreeCache<Constant> (tree);
}

const_git_type 
get_git_type_from_tree_cache (git_gimple_helpers gim_help, tree tree)
{
    return gim_help->getFromTreeCache<Type> (tree);
}

void 
add_git_decl_to_tree_cache (git_gimple gimple, git_gimple_helpers gim_help, tree tree, git_decl decl)
{
	gim_help->addToTreeCache (tree, decl);
	gimple->addToTreeCache(makeAP<TreeProduct> (decl));
}

void
add_git_type_to_tree_cache (git_gimple gimple, git_gimple_helpers gim_help, tree tree, git_type type)
{
	gim_help->addToTreeCache (tree, type);
	gimple->addToTreeCache(makeAP<TreeProduct> (type));
}

void
add_git_constant_to_tree_cache (git_gimple gimple, git_gimple_helpers gim_help, tree tree, git_constant cst)
{
	gim_help->addToTreeCache (tree, cst);
	gimple->addToTreeCache(makeAP<TreeProduct> (cst));
}

void 
reportConstant (git_gimple gimple, git_constant constant)
{
    gimple->reportCst (std::auto_ptr<Constant> (constant));
}

void 
reportFunction (git_gimple gimple, const_git_decl funcDecl, git_function func)
{
    const FunctionDecl* fd = dynamic_cast<const FunctionDecl*> (funcDecl);
    assert (fd);
    
    gimple->reportFuncDef (fd, std::auto_ptr<Function> (func));
}

void 
reportMain (git_gimple gimple, const_git_decl mainDecl)
{
    const FunctionDecl* fd = dynamic_cast<const FunctionDecl*> (mainDecl);
    assert (fd);
    
    gimple->reportMain (fd);
}

void
reportDecl (git_gimple gimple, git_decl decl)
{
    gimple->reportDecl (std::auto_ptr<Decl> (decl));
}

bool
checkFuncNameExists(git_gimple_helpers gim_help, const char* name)
{
	return gim_help->checkFuncNameExists(name);
}
void
addFuncName(git_gimple_helpers gim_help, const char* name)
{
	gim_help->addFuncName(name);
}

void
setGlobalsResolved(git_gimple_helpers gim_help)
{
	gim_help->setGlobalsResolved();
}

bool
getGlobalsResolved(git_gimple_helpers gim_help)
{
	return gim_help->getGlobalsResolved();
}

void
addDefinedType (git_gimple_helpers gim_help, const char* typeName, git_type type)
{
	gim_help->addDefinedType (typeName, type);
}

git_type
getDefinedType (git_gimple_helpers gim_help, const char* typeName)
{
	return gim_help->getDefinedType (typeName);
}

void
addDefinedType2 (git_gimple_helpers gim_help, tree typeTree, git_type type)
{
	gim_help->addDefinedType2 (typeTree, type);
}

git_type
getDefinedType2 (git_gimple_helpers gim_help, tree typeTree)
{
	return gim_help->getDefinedType2 (typeTree);
}

bool
isRecursive(git_gimple_helpers gim_help)
{
	return gim_help->isRecursive();
}
void setRecursive(git_gimple_helpers gim_help, bool recursive)
{
	gim_help->setRecursive(recursive);
}

enum function_code getFunctionCode(git_gimple_helpers gim_help, const char* name)
{
	return gim_help->getFunctionCode(name);
}

void setConstructor(git_gimple_helpers gim_help, const_git_decl decl, git_operand op)
{
	gim_help->setConstructor(decl, op);
}

git_operand getConstructor(git_gimple_helpers gim_help, const_git_decl decl)
{
	return gim_help->getConstructor(decl);
}

const_git_type
declGetType(const_git_decl decl)
{
	return decl->getType();
}

}}
