// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file ConstructionHelpers.h
 * Declarations of helper functions that allow
 * C code to construct GIMPLE++ objects.
 */
#ifndef GIT_CONSTRUCTION_HELPERS_H
#define GIT_CONSTRUCTION_HELPERS_H

#include <stdint.h>
#include <stdlib.h>

#include "CTypes.h"
#include "FunctionInfo.h"

#ifdef __cplusplus

#include "gimple_types.h"

extern "C" {
namespace giterator { namespace detail {

#else

#include "../tree.h"
#include "../function.h"
#include "../basic-block.h"
                 
#endif /* __cplusplus */

/**
 * Structure containing everything that is needed
 * to convert GIMPLE representation of a program
 * to GIMPLE++ representation.
 */
struct GimpleConstruction {
	/*! GIMPLE++ representation holder */
    git_gimple gimple;
    /*! Helper class needed for GIMPLE++ construction */
    git_gimple_helpers gimple_helpers;
    /*! Helper class needed for construction of functions */
    git_function_helpers function_helpers;
};

/*! type for the constructions structure usable from C code */
typedef struct GimpleConstruction* git_gimple_construction;

/* NOTE: All these 4 functions share a common static instance of the
  GimpleConstruction and they are ment to be called from GCC GIMPLE++
  serializer. */

/**
 * Initialize fields needed for function construction
 * in a static instance of @c GimpleConstruction.
 *
 * @return initialized instance of @c GimpleConstruction
 */
extern git_gimple_construction beginFunctionConstruction(void);

/**
 * Clean up the common static instance of @c GimpleConstruction
 * after a function construction.
 */
extern void endFunctionConstruction(void);

/**
 * Obtain a constructed GIMPLE++ representation of a program from
 * the static instance of @c GimpleConstruction.
 *
 * @return constructed GIMPLE++ representation of a program
 */
extern git_gimple getConstructionResult(void);

/**
 * Clean up the static instance of @c GimpleConstruction
 * after constructing GIMPLE++ representation of a program.
 */
extern void endConstruction(void);

/* Declarations */

/**
 * Creates GIMPLE++ declaration object representing function declaration.
 *
 * @param gim_help
 * @param uid unique ID of this declaration
 * @param name name of this declaration
 * @param type type of this declaration
 * @return GIMPLE++ declaration object
 *
 * @see giterator::FunctionDecl
 */
git_decl createFunctionDecl (git_gimple_helpers gim_help, unsigned int uid, const char* name, const_git_type type);

/**
 * Creates GIMPLE++ declaration object representing variable declaration.
 *
 * @param uid unique ID of this declaration
 * @param name name of this declaration
 * @param type type of this declaration
 * @param local whether this is local (true) or global (false) declaration
 * @return GIMPLE++ declaration object
 *
 * @see giterator::VarDecl
 */
git_decl createVarDecl (unsigned int uid, const char* name, const_git_type type, bool local);

/**
 * Creates GIMPLE++ declaration object representing struct/union declaration.
 *
 * @param uid unique ID of this declaration
 * @param name name of this declaration
 * @param type type of this declaration
 * @param local whether this is local (true) or global (false) declaration
 * @return GIMPLE++ declaration object
 *
 * @see giterator::RecordDecl
 */
git_decl createRecordDecl (unsigned int uid, const char* name, const_git_type type, bool local);

/**
 * Creates GIMPLE++ declaration object representing function parameter declaration.
 *
 * @param uid unique ID of this declaration
 * @param name name of this declaration
 * @param type type of this declaration
 * @return GIMPLE++ declaration object
 *
 * @see giterator::ParamDecl
 */
git_decl createParamDecl (unsigned int uid, const char* name, const_git_type type);


/**
 * Creates GIMPLE++ declaration object representing label declaration.
 *
 * @param uid unique ID of this declaration
 * @param name name of this declaration
 * @param type type of this declaration
 * @return GIMPLE++ declaration object
 *
 * @see giterator::LabelDecl
 */
git_decl createLabelDecl (unsigned int uid, const char* name, const_git_type type);

/**
 * Creates GIMPLE++ declaration object representing result declaration (value returned from function).
 *
 * @param uid unique ID of this declaration
 * @param name name of this declaration
 * @param type type of this declaration
 * @return GIMPLE++ declaration object
 *
 * @see giterator::ResultDecl
 */
git_decl createResultDecl (unsigned int uid, const char* name, const_git_type type);

/**
 * Creates GIMPLE++ declaration object representing type declaration (typedef in C).
 *
 * @param uid unique ID of this declaration
 * @param name name of this declaration
 * @param type type of this declaration
 * @param local whether this is local (true) or global (false) declaration
 * @return GIMPLE++ declaration object
 *
 * @see giterator::TypeDecl
 */
git_decl createTypeDecl (unsigned int uid, const char* name, const_git_type type, bool local);

/**
 * Creates GIMPLE++ declaration object representing struct/union field declaration.
 *
 * @param uid unique ID of this declaration
 * @param name name of this declaration
 * @param type type of this declaration
 * @return GIMPLE++ declaration object
 *
 * @see giterator::FieldDecl
 */
git_decl createFieldDecl (unsigned int uid, const char* name, const_git_type type);

/**
 * Creates GIMPLE++ declaration object representing enumeration constant declaration.
 *
 * @param uid unique ID of this declaration
 * @param name name of this declaration
 * @param type type of this declaration
 * @param local whether this is local (true) or global (false) declaration
 * @return GIMPLE++ declaration object
 *
 * @see giterator::ConstDecl
 */
git_decl createConstDecl (unsigned int uid, const char* name, const_git_type type, bool local);

/* Function */

/**
 * Creates GIMPLE++ function object representing a function definition.
 *
 * @param type type of the function that is being created.
 * @return GIMPLE++ function object
 *
 * @see giterator::Function
 */
git_function createFunction (const_git_type type);

/**
 * Adds local variable into GIMPLE++ function.
 *
 * @param function GIMPLE++ function object
 * @param decl GIMPLE++ declaration object representing local variable declaration
 *
 * @see giterator::Function
 */
void functionAddLocalDecl (git_function function, const_git_decl decl);

/**
 * Adds parameter declaration to the list of parameters of GIMPLE++ function.
 *
 * @param function GIMPLE++ function object
 * @param decl GIMPLE++ declaration object representing parameter declaration
 *
 * @see giterator::Function
 */
void functionAddParamDecl (git_function function, const_git_decl decl);

/**
 * Adds block of statements into GIMPLE++ function.
 *
 * @param function GIMPLE++ function object
 * @param func_help helper class for this function construction
 * @param basicBlock GIMPLE @c basic_block structure of the added block
 * @param block GIMPLE++ block created from @c basicBlock
 *
 * @see giterator::Function
 */
void functionAddBlock (git_function function, git_function_helpers func_help, basic_block basicBlock, git_block block);
void functionAddBlock2 (git_function function, git_block block);

/**
 * Sets first block in a GIMPLE++ function.
 * The execution of the function starts in the first block.
 *
 * @param function GIMPLE++ function object
 * @param func_help helper class for this function construction
 * @param basicBlock GIMPLE @c basic_block sturcture representing the first block
 *
 * @see giterator::Function
 */
void functionSetFirstBlock (git_function function, git_function_helpers func_help, basic_block basicBlock);
void functionSetFirstBlock2 (git_function function, git_block block);

/**
 * Returns GIMPLE++ block for a given GIMPLE @c basic_block structure.
 *
 * @param func_help helper class for this function construction
 * @param basicBlock GIMPLE @c basic_block structure
 * @return GIMPLE++ block created from @c basicBlock
 *
 * @see giterator::Function
 */
git_block functionGetBlock (git_function_helpers func_help, basic_block basicBlock);

/**
 * Returns first block in a GIMPLE++ function.
 * The execution of the function starts in the first block.
 *
 * @param function GIMPLE++ function object
 * @return first block of the function, where the execution of function starts.
 *
 * @see giterator::Function
 */
git_block functionGetFirstBlock(git_function function);


/**
 * Creates a new GIMPLE++ block of statements.
 *
 * @return new GIMPLE++ block
 *
 * @see giterator::Block
 */
git_block createBlock (void);

/**
 * Adds a statement into a GIMPLE++ block.
 *
 * @param block GIMPLE++ block
 * @param stmt GIMPLE++ statement to add into @c block
 *
 * @see giterator::Block
 */
void blockAddStmt (git_block block, git_stmt stmt);

/**
 * Adds an edge to the predecessors list of GIMPLE++ block.
 *
 * @param block GIMPLE++ block
 * @param edge GIMPEL++ edge
 *
 * @see giterator::Block
 */
void blockAddPredEdge (git_block block, git_edge edge);

/**
 * Adds an edge to the successors list of GIMPLE++ block.
 *
 * @param block GIMPLE++ block
 * @param edge GIMPEL++ edge
 *
 * @see giterator::Block
 */
void blockAddSuccEdge (git_block block, git_edge edge);

/**
 * Creates a GIMPLE++ edge.
 *
 * @param src GIMPLE++ block this edge leads from
 * @param dest GIMPLE++ block this edge leads to
 * @param isFallthru @c true if the new edge is a fallthru edge, @c false otherwise
 * @param isTrueBranch @c true if the new edge is a true branch edge, @c false otherwise
 * @param isFalseBranch @c true if the new edge is a false branch edge, @c false otherwise
 * @return GIMPLE++ edge
 *
 * @see giterator::Edge
 */
git_edge createEdge (git_block src, git_block dest, 
                     bool isFallthru, bool isTrueBranch, bool isFalseBranch);

/* Types */

/**
 * Creates GIMPLE++ type object representing integer type.
 *
 * @param bitSize size of the type in bits
 * @param isUnsigned @c true if the new type is unsigned, @c false otherwise
 * @param precision the precision of the type in bits
 * @return new GIMPLE++ type object
 *
 * @see giterator::IntegerType
 */
git_type createIntegerType (size_t bitSize, bool isUnsigned, int precision);

/**
 * Creates GIMPLE++ type object representing real type.
 *
 * @param bitSize size of the type in bits
 * @param precision the precision of the type in bits
 * @return new GIMPLE++ type object
 *
 * @see giterator::RealType
 */
git_type createRealType(size_t bitSize, int precision);

/**
 * Creates GIMPLE++ type object representing pointer type.
 *
 * @param bitSize size of the type in bits
 * @param pointingToType the type to which the newly created pointer type points to
 * @return new GIMPLE++ type object
 *
 * @see giterator::PtrType
 */
git_type createPtrType (size_t bitSize, const_git_type pointingToType);

/**
 * Creates GIMPLE++ type object representing @c void type.
 *
 *
 * @return new GIMPLE++ type object representing @c void type
 *
 * @see giterator::VoidType
 */
git_type createVoidType ();

/**
 * Creates GIMPLE++ type object representing @c bool type.
 *
 * @param bitSize size of the type in bits
 * @return new GIMPLE++ type object
 *
 * @see giterator::BoolType
 */
git_type createBoolType (size_t bitSize);

/**
 * Creates GIMPLE++ type object representing function type.
 *
 * @param gim_help @c GimpleHelpers object to assign a name to unnamed function type
 * @param name name of the type
 * @param returnValue the type of the function return value
 * @return new GIMPLE++ type object
 *
 * @see giterator::FunctionType
 */
git_type createFunctionType (git_gimple_helpers gim_help, const char* name, const_git_type returnValue);

/**
 * Adds a parameter into the list of parameters in the function type.
 *
 * @param funcType GIMPLE++ type object representing function type
 * @param argType a type of the parameter to add to the type
 *
 * @see giterator::FunctionType
 */
void functionTypeAddParameter (git_type funcType, const_git_type argType);

/**
 * Creates GIMPLE++ type object representing enumeral type.
 *
 * @param gim_help @c GimpleHelpers object to assign a name to unnamed enum type
 * @param name name of the type
 * @param bitSize size of the type in bits
 * @param isUnsigned @c true if all fields are unsigned, @c false otherwise
 * @param precision the precision of the type in bits
 * @return new GIMPLE++ type object
 *
 * @see giterator::EnumType
 */
git_type createEnumType (git_gimple_helpers gim_help, const char* name, size_t bitSize, bool isUnsigned, int precision);

/**
 * Add an enumeration constant into a GIMPLE++ enureral type object.
 *
 * @see giterator::EnumType
 */
void enumTypeAddConstant(git_type enumType, const char* eName, const_git_constant eConstant);

/**
 * Creates GIMPLE++ type object representing struct or union type.
 *
 * @param gim_help @c GimpleHelpers object to assign a name to unnamed record type
 * @param name name of the type
 * @param bitSize size of the type in bits
 * @return new GIMPLE++ type object
 *
 * @see giterator::RecordType
 */
git_type createRecordType (git_gimple_helpers gim_help, const char* name, size_t bitSize);

/**
 * Add a field declaration into a struct/union GIMPLE++ object.
 *
 * @see giterator::RecordType
 */
void recordTypeAddField(git_type recordType, const_git_decl field);

/**
 * Add a loacl declaration into a struct/union GIMPLE++ object.
 *
 * @see giterator::RecordType
 */
void recordTypeAddLocalDecl(git_type recordType, const_git_decl localDecl);

/**
 * Creates GIMPLE++ type object representing array type.
 *
 * @param bitSize size of the type in bits
 * @param elementType type of array elements
 * @return new GIMPLE++ type object
 *
 * @see giterator::ArrayType
 */
git_type createArrayType (size_t bitSize, const_git_type elementType);

/* Statements */

/**
 * Creates GIMPLE++ statement representing GIMPLE_ASSIGN statement.
 *
 * @param lhs operand on the left hand side of the assignment
 * @param operation operation on the right hand side of the assignment
 * @return new GIMPLE++ statement
 *
 * @see giterator::AssignStmt
 */
git_stmt createAssignStmt (git_operand lhs, git_operation operation);

/**
 * Creates GIMPLE++ statement representing GIMPLE_COND statement.
 *
 * @param boolOperation operation evaluating the condition
 * @return new GIMPLE++ statement
 *
 * @see giterator::CondStmt
 */
git_stmt createCondStmt (git_operation boolOperation);

/**
 * Adds an argument into a call statement.
 *
 * @param callStmt GIMPLE++ call statement where the argument will be added
 * @param param operand representing the argument
 *
 * @see giterator::CallStmt
 */
void callStmtAddArg (git_stmt callStmt, git_operand param);

/**
 * Creates GIMPLE++ statement representing GIMPLE_CALL statement.
 *
 * @param fn operand representing the pointer to the function to be called
 * @param lhs operand where will be stored the returned value
 * @return new GIMPLE++ statement
 *
 * @see giterator::CallStmt
 */
git_stmt createCallStmt (git_operand fn, git_operand lhs);

/**
 * Creates GIMPLE++ statement representing GIMPLE_RETURN statement.
 *
 * @param retval the operand representing return value
 * @return new GIMPLE++ statement
 * @see giterator::ReturnStmt
 */
git_stmt createReturnStmt (git_operand retval);

/**
 * Creates GIMPLE++ statement representing GIMPLE_GOTO statement.
 *
 * @param dest operand representing destination of the goto statement
 * @return new GIMPLE++ statement
 *
 * @see giterator::GotoStmt
 */
git_stmt createGotoStmt (git_operand dest);

/**
 * Creates GIMPLE++ statement representing GIMPLE_LABEL statement.
 *
 * @param label operand representing the label
 * @return new GIMPLE++ statement
 *
 * @see giterator::LabelStmt
 */
git_stmt createLabelStmt (git_operand label);

/**
 * Creates GIMPLE++ statement representing GIMPLE_NOP statement.
 *
 * @return new GIMPLE++ statement
 *
 * @see giterator::NopStmt
 */
git_stmt createNopStmt (void);

/**
 * Creates GIMPLE++ statement representing GIMPLE_PHI statement.
 *
 * @param result operand representing the result of the phi statement
 * @return new GIMPLE++ statement
 *
 * @see giterator::PhiStmt
 */
git_stmt createPhiStmt (git_operand result);

/**
 * Add an argument into a phi statement.
 *
 * @param phiStmt GIMPLE++ phi statement where the argument will be added
 * @param arg operand representing the argument
 *
 * @see giterator::PhiStmt
 */
void phiStmtAddArg (git_stmt phiStmt, git_operand arg);

/**
 * Creates GIMPLE++ statement representing GIMPLE_SWITCH statement.
 *
 * @param index operand representing the index of switch statement
 * @return new GIMPLE++ statement
 *
 * @see giterator::SwitchStmt
 */
git_stmt createSwitchStmt (git_operand index);

/**
 * Add a label into a switch statement.
 *
 * @param switchStmt GIMPLE++ switch statement where the label will be added
 * @param label operand representing the label
 *
 * @see giterator::SwitchStmt
 */
void switchStmtAddLabel(git_stmt switchStmt, git_operand label);

/* Constants */

/**
 * Creates GIMPLE++ constant representing integer constant.
 *
 * @param type type of the constant
 * @param value raw value of the constant
 * @return new GIMPLE++ constant
 *
 * @see giterator::IntegerConstant
 */
git_constant createIntCst (const_git_type type, uint64_t value);

/**
 * Creates GIMPLE++ constant representing enumeral constant.
 *
 * @param type type of the constant
 * @param value raw value of the constant
 * @return new GIMPLE++ constant
 *
 * @see giterator::EnumeralConstant
 */
git_constant createEnumCst (const_git_type type, uint64_t value);

/**
 * Creates GIMPLE++ constant representing real constant.
 *
 * @param type type of the constant
 * @param value raw value of the constant
 * @return new GIMPLE++ constant
 *
 * @see giterator::RealConstant
 */
git_constant createRealCst (const_git_type type, long double value);

/**
 * Creates GIMPLE++ constant representing string constant.
 *
 * @param type type of the constant
 * @param strLen length of the constant
 * @param str the value of the constant
 * @return new GIMPLE++ constant
 *
 * @see giterator::StringConstant
 */
git_constant createStringCst (const_git_type type, int strLen, const char* str);

/* Operands */

/**
 * Creates GIMPLE++ operand representing declaration.
 *
 * @param decl declaration represented by the operand
 * @return new GIMPLE++ operand
 *
 * @see giterator::DeclOperand
 */
git_operand createDeclOperand (const_git_decl decl);

/**
 * Creates GIMPLE++ operand representing address operand.
 *
 * @param type type of the operand
 * @param op operand which address the new operand represents
 * @return new GIMPLE++ operand
 *
 * @see giterator::AddrOperand
 */
git_operand createAddrOperand (const_git_type type, git_operand op);

/**
 * Creates GIMPLE++ operand representing label expression operand.
 *
 * @param decl declaration representing the label
 * @param lower operand representing the label expression for regular
 * and default case labels or lower bound of range case label
 * @param upper operand representing upper bound of range case label. Valid only for range labels
 * @param isDefault whether this is the default case label
 * @param isRange whether this is a range case label
 * @return new GIMPLE++ operand
 *
 * @see giterator::LabelExprOperand
 */
git_operand createLabelExprOperand (const_git_decl decl, git_operand lower, git_operand upper, bool isDefault, bool isRange);

/**
 * Creates GIMPLE++ operand representing indirect reference operand.
 *
 * @param type type of the operand
 * @param op operand which address the new operand represents
 * @return new GIMPLE++ operand
 *
 * @see giterator::IndirectRefOperand
 */
git_operand createIndirectRefOperand (const_git_type type, git_operand op);
/**
 * Creates GIMPLE++ operand representing member of array.
 *
 * @param array oparand representing the array
 * @param index operand representing the index into the arrray
 * @return operand representing the value stored in the array @c array at index @c index
 *
 * @see giterator::ArrayRefOperand
 */
git_operand createArrayRefOperand (git_operand array, git_operand index);

/**
 * Creates GIMPLE++ operand representing dereference operand.
 *
 * @param record operand representing the record part of the component ref
 * @param field operand representing the field part of the component ref
 * @return new GIMPLE++ operand
 *
 * @see giterator::ComponentRefOperand
 */
git_operand createComponentRefOperand (git_operand record, git_operand field);

/**
 * Creates GIMPLE++ operand representing constant.
 *
 * @param cst GIMPLE++ constant
 * @return new GIMPLE++ operand representing constant @c constant
 *
 * @see giterator::CstOperand
 */
git_operand createCstOperand (const_git_constant cst);

/**
 * Creates GIMPLE++ operand representing struct/union initializer.
 *
 * @param type GIMPLE++ type of the struct/union
 * @return new GIMPLE++ operand representing struct/union initializer
 *
 * @see giterator::RecordConstructorOperand
 */
git_operand createRecordConstructorOperand (const_git_type type);

/**
 * Add a field initializer into a GIMPLE++ RecordConstructorOperand
 *
 * @param recordOperand operand where to add the field initializer
 * @param decl field declaration, that is to be initialized with the added initializer
 * @param operand operand representing initializer initializing field decl @c decl
 *
 * @see giterator::RecordConstructorOperand
 */
void recordConstructorOperandAddField(git_operand recordOperand, const_git_decl decl, git_operand operand);

/**
 * Creates GIMPLE++ operand representing array initializer.
 *
 * @param type GIMPLE++ type of the array
 * @return new GIMPLE++ operand representing array initializer
 *
 * @see giterator::ArrayConstructorOperand
 */
git_operand createArrayConstructorOperand (const_git_type type);

/**
 * Add an initializer into a GIMPLE++ ArrayConstructorOperand.
 *
 * @param arrayOperand operand where to add the initializer
 * @param index index in the array, that is to be initialized with the added initializer
 * @param operand operand representing initializer initializing index @c index
 *
 * @see giterator::ArrayConstructorOperand
 */
void arrayConstructorOperandAddField(git_operand arrayOperand, const_git_constant index, git_operand operand);

/**
 * Create and add initializer into an ArrayConstructorOperand for each index in a range of index values.
 * The operand representing the initializer will get duplicated for each index in the range.
 *
 * @param gimple constructed GIMPLE++ representation of program (to report new constants)
 * @param arrayOperand operand where to add the initializers
 * @param operand operand representing initializer initializing all indexes between lower and upper
 * @param lower lower bound of the initialized indexes
 * @param upper upper bound of the initialized indexes
 *
 * @see giterator::ArrayConstructorOperand
 */
void createAndAddConstantsBetween(git_gimple gimple, git_operand arrayOperand, git_operand operand, git_constant lower, git_constant upper);

/* Operations */

  /* Unary */

/**
 * Creates GIMPLE++ representation of unary negation operation.
 *
 * @param return_type the return type of this operation
 * @param op operand to perform the operation on
 * @return constructed GIMPLE++ unary operation
 *
 * @see giterator::NegateOperation
 */
git_operation createNegateOperation (const_git_type return_type, git_operand op);

/**
 * Creates GIMPLE++ representation of absolute value operation.
 *
 * @param return_type the return type of this operation
 * @param op operand to perform the operation on
 * @return constructed GIMPLE++ unary operation
 *
 * @see giterator::AbsOperation
 */
git_operation createAbsOperation (const_git_type return_type, git_operand op);

/**
 * Creates GIMPLE++ representation of conversion of real to fixed point by truncation.
 *
 * @param return_type the return type of this operation
 * @param op operand to perform the operation on
 * @return constructed GIMPLE++ unary operation
 *
 * @see giterator::FixTruncOperation
 */
git_operation createFixTruncOperation(const_git_type return_type, git_operand op);

/**
 * Creates GIMPLE++ representation of conversion of an integer to a real.
 *
 * @param return_type the return type of this operation
 * @param op operand to perform the operation on
 * @return constructed GIMPLE++ unary operation
 *
 * @see giterator::FloatOperation
 */
git_operation createFloatOperation(const_git_type return_type, git_operand op);

/**
 * Creates GIMPLE++ representation of bitwise not operation.
 *
 * @param return_type the return type of this operation
 * @param op operand to perform the operation on
 * @return constructed GIMPLE++ unary operation
 *
 * @see giterator::BitNotOperation
 */
git_operation createBitNotOperation(const_git_type return_type, git_operand op);

/**
 * Creates GIMPLE++ representation of logical negation operation.
 *
 * @param return_type the return type of this operation
 * @param op operand to perform the operation on
 * @return constructed GIMPLE++ unary operation
 *
 * @see giterator::TruthNotOperation
 */
git_operation createTruthNotOperation(const_git_type return_type, git_operand op);

/**
 * Creates GIMPLE++ representation of re-association barrier for floating point operations.
 *
 * @param return_type the return type of this operation
 * @param op operand to perform the operation on
 * @return constructed GIMPLE++ unary operation
 *
 * @see giterator::ParenOperation
 */
git_operation createParenOperation(const_git_type return_type, git_operand op);

/**
 * Creates GIMPLE++ representation of guaranteed non lvalue operation.
 *
 * @param return_type the return type of this operation
 * @param op operand to perform the operation on
 * @return constructed GIMPLE++ unary operation
 *
 * @see giterator::NonLvalueOperation
 */
git_operation createNonLvalueOperation(const_git_type return_type, git_operand op);

/**
 * Creates GIMPLE++ representation of no operation (result is the operand).
 *
 * @param return_type the return type of this operation
 * @param op operand to perform the operation on
 * @return constructed GIMPLE++ unary operation
 *
 * @see giterator::NoOperation
 */
git_operation createNoOperation (const_git_type return_type, git_operand op);

/**
 * Creates GIMPLE++ representation of conversion operation.
 *
 * @param return_type the return type of this operation
 * @param op operand to perform the operation on
 * @return constructed GIMPLE++ unary operation
 *
 * @see giterator::ConvertOperation
 */
git_operation createConvertOperation (const_git_type return_type, git_operand op);

/**
 * Creates GIMPLE++ representation of conversion of a pointer value to a pointer
 * to a different address space.
 *
 * @param return_type the return type of this operation
 * @param op operand to perform the operation on
 * @return constructed GIMPLE++ unary operation
 *
 * @see giterator::AddrSpaceConvertOperation
 */
git_operation createAddrSpaceConvertOperation (const_git_type return_type, git_operand op);

/**
 * Creates GIMPLE++ representation of conversion of a fixed-point value to an integer,
 * a real, or a fixed-point value. Or conversion of a fixed-point value from
 * an integer, a real, or a fixed-point value.
 *
 * @param return_type the return type of this operation
 * @param op operand to perform the operation on
 * @return constructed GIMPLE++ unary operation
 *
 * @see giterator::FixedConvertOperation
 */
git_operation createFixedConvertOperation(const_git_type return_type, git_operand op);

  /* Binary */

/**
 * Creates GIMPLE++ representation of binary addition.
 *
 * @param return_type return type of this operation
 * @param op1 first operand of this operation
 * @param op2 second operand of this operation
 * @return constructed GIMPLE++ binary operation
 *
 * @see giterator::PlusOperation
 */
git_operation createPlusOperation (const_git_type return_type, git_operand op1, git_operand op2);

/**
 * Creates GIMPLE++ representation of binary substraction.
 *
 * @param return_type return type of this operation
 * @param op1 first operand of this operation
 * @param op2 second operand of this operation
 * @return constructed GIMPLE++ binary operation
 *
 * @see giterator::MinusOperation
 */
git_operation createMinusOperation (const_git_type return_type, git_operand op1, git_operand op2);

/**
 * Creates GIMPLE++ representation of binary multiplication.
 *
 * @param return_type return type of this operation
 * @param op1 first operand of this operation
 * @param op2 second operand of this operation
 * @return constructed GIMPLE++ binary operation
 *
 * @see giterator::MultOperation
 */
git_operation createMultOperation (const_git_type return_type, git_operand op1, git_operand op2);

/**
 * Creates GIMPLE++ representation of binary division for real result.
 *
 * @param return_type return type of this operation
 * @param op1 first operand of this operation
 * @param op2 second operand of this operation
 * @return constructed GIMPLE++ binary operation
 *
 * @see giterator::RdivOperation
 */
git_operation createRdivOperation (const_git_type return_type, git_operand op1, git_operand op2);

/**
 * Creates GIMPLE++ representation of division for integer result that rounds
 * the quotient toward zero.
 *
 * @param return_type return type of this operation
 * @param op1 first operand of this operation
 * @param op2 second operand of this operation
 * @return constructed GIMPLE++ binary operation
 *
 * @see giterator::TruncDivOperation
 */
git_operation createTruncDivOperation (const_git_type return_type, git_operand op1, git_operand op2);

/**
 * Creates GIMPLE++ representation of binary division for integer result
 * that rounds toward minus infinity.
 *
 * @param return_type return type of this operation
 * @param op1 first operand of this operation
 * @param op2 second operand of this operation
 * @return constructed GIMPLE++ binary operation
 *
 * @see giterator::FloorDivOperation
 */
git_operation createFloorDivOperation (const_git_type return_type, git_operand op1, git_operand op2);

/**
 * Creates GIMPLE++ representation of binary division for integer result
 * that rounds the quotient toward infinity.
 *
 * @param return_type return type of this operation
 * @param op1 first operand of this operation
 * @param op2 second operand of this operation
 * @return constructed GIMPLE++ binary operation
 *
 * @see giterator::CeilDivOperation
 */
git_operation createCeilDivOperation (const_git_type return_type, git_operand op1, git_operand op2);

/**
 * Creates GIMPLE++ representation of binary division for integer result
 * that rounds toward nearest integer.
 *
 * @param return_type return type of this operation
 * @param op1 first operand of this operation
 * @param op2 second operand of this operation
 * @return constructed GIMPLE++ binary operation
 *
 * @see giterator::RoundDivOperation
 */
git_operation createRoundDivOperation (const_git_type return_type, git_operand op1, git_operand op2);

/**
 * Creates GIMPLE++ representation of binary remainder that rounds
 * the quotient toward zero.
 *
 * @param return_type return type of this operation
 * @param op1 first operand of this operation
 * @param op2 second operand of this operation
 * @return constructed GIMPLE++ binary operation
 *
 * @see giterator::TruncModOperation
 */
git_operation createTruncModOperation (const_git_type return_type, git_operand op1, git_operand op2);

/**
 * Creates GIMPLE++ representation of binary remainder that rounds
 * the quotient toward minus infinity.
 *
 * @param return_type return type of this operation
 * @param op1 first operand of this operation
 * @param op2 second operand of this operation
 * @return constructed GIMPLE++ binary operation
 *
 * @see giterator::FloorModOperation
 */
git_operation createFloorModOperation (const_git_type return_type, git_operand op1, git_operand op2);

/**
 * Creates GIMPLE++ representation of binary remainder that rounds
 * the quotient toward infinity.
 *
 * @param return_type return type of this operation
 * @param op1 first operand of this operation
 * @param op2 second operand of this operation
 * @return constructed GIMPLE++ binary operation
 *
 * @see giterator::CeilModOperation
 */
git_operation createCeilModOperation (const_git_type return_type, git_operand op1, git_operand op2);

/**
 * Creates GIMPLE++ representation of binary remainder that rounds
 * the quotient toward nearest integer.
 *
 * @param return_type return type of this operation
 * @param op1 first operand of this operation
 * @param op2 second operand of this operation
 * @return constructed GIMPLE++ binary operation
 *
 * @see giterator::RoundModOperation
 */
git_operation createRoundModOperation (const_git_type return_type, git_operand op1, git_operand op2);

/**
 * Creates GIMPLE++ representation of binary division which is
 * not supposed to need rounding.
 *
 * @param return_type return type of this operation
 * @param op1 first operand of this operation
 * @param op2 second operand of this operation
 * @return constructed GIMPLE++ binary operation
 *
 * @see giterator::ExactDivOperation
 */
git_operation createExactDivOperation (const_git_type return_type, git_operand op1, git_operand op2);

/**
 * Creates GIMPLE++ representation of minimum value of the two values.
 *
 * @param return_type return type of this operation
 * @param op1 first operand of this operation
 * @param op2 second operand of this operation
 * @return constructed GIMPLE++ binary operation
 *
 * @see giterator::MinOperation
 */
git_operation createMinOperation (const_git_type return_type, git_operand op1, git_operand op2);

/**
 * Creates GIMPLE++ representation of maximum value of the two values.
 *
 * @param return_type return type of this operation
 * @param op1 first operand of this operation
 * @param op2 second operand of this operation
 * @return constructed GIMPLE++ binary operation
 *
 * @see giterator::MaxOperation
 */
git_operation createMaxOperation (const_git_type return_type, git_operand op1, git_operand op2);

/**
 * Creates GIMPLE++ representation of logical left shift if done on an
 * unsigned type, arithmetic left shift if done on a signed type.
 *
 * @param return_type return type of this operation
 * @param op1 first operand of this operation - to be shifted
 * @param op2 second operand of this operation - number of bits to shift by
 * @return constructed GIMPLE++ binary operation
 *
 * @see giterator::LshiftOperation
 */
git_operation createLshiftOperation (const_git_type return_type, git_operand op1, git_operand op2);

/**
 * Creates GIMPLE++ representation of logical right shift if done on an
 * unsigned type, arithmetic right shift if done on a signed type.
 *
 * @param return_type return type of this operation
 * @param op1 first operand of this operation - to be shifted
 * @param op2 second operand of this operation - number of bits to shift by
 * @return constructed GIMPLE++ binary operation
 *
 * @see giterator::RshiftOperation
 */
git_operation createRshiftOperation (const_git_type return_type, git_operand op1, git_operand op2);

/**
 * Creates GIMPLE++ representation of logical left rotation if done on an
 * unsigned type, arithmetic left shift if done on a signed type.
 *
 * @param return_type return type of this operation
 * @param op1 first operand of this operation - to be rotated
 * @param op2 second operand of this operation - number of bits to rotate by
 * @return constructed GIMPLE++ binary operation
 *
 * @see giterator::LrotateOperation
 */
git_operation createLrotateOperation (const_git_type return_type, git_operand op1, git_operand op2);

/**
 * Creates GIMPLE++ representation of logical right rotation if done on an
 * unsigned type, arithmetic left right if done on a signed type.
 *
 * @param return_type return type of this operation
 * @param op1 first operand of this operation - to be rotated
 * @param op2 second operand of this operation - number of bits to rotate by
 * @return constructed GIMPLE++ binary operation
 *
 * @see giterator::RrotateOperation
 */
git_operation createRrotateOperation (const_git_type return_type, git_operand op1, git_operand op2);

/**
 * Creates GIMPLE++ representation of bitwise inclusive or.
 *
 * @param return_type return type of this operation
 * @param op1 first operand of this operation
 * @param op2 second operand of this operation
 * @return constructed GIMPLE++ binary operation
 *
 * @see giterator::BitIorOperation
 */
git_operation createBitIorOperation(const_git_type return_type, git_operand op1, git_operand op2);

/**
 * Creates GIMPLE++ representation of bitwise exclusive or.
 *
 * @param return_type return type of this operation
 * @param op1 first operand of this operation
 * @param op2 second operand of this operation
 * @return constructed GIMPLE++ binary operation
 *
 * @see giterator::BitXorOperation
 */
git_operation createBitXorOperation(const_git_type return_type, git_operand op1, git_operand op2);

/**
 * Creates GIMPLE++ representation of bitwise and operation.
 *
 * @param return_type return type of this operation
 * @param op1 first operand of this operation
 * @param op2 second operand of this operation
 * @return constructed GIMPLE++ binary operation
 *
 * @see giterator::BitAndOperation
 */
git_operation createBitAndOperation(const_git_type return_type, git_operand op1, git_operand op2);

/**
 * Creates GIMPLE++ representation of logical and operation.
 *
 * @param return_type return type of this operation
 * @param op1 first operand of this operation
 * @param op2 second operand of this operation
 * @return constructed GIMPLE++ binary operation
 *
 * @see giterator::TruthAndOperation
 */
git_operation createTruthAndOperation(const_git_type return_type, git_operand op1, git_operand op2);

/**
 * Creates GIMPLE++ representation of logical or operation.
 *
 * @param return_type return type of this operation
 * @param op1 first operand of this operation
 * @param op2 second operand of this operation
 * @return constructed GIMPLE++ binary operation
 *
 * @see giterator::TruthOrOperation
 */
git_operation createTruthOrOperation(const_git_type return_type, git_operand op1, git_operand op2);

/**
 * Creates GIMPLE++ representation of logical xor operation.
 *
 * @param return_type return type of this operation
 * @param op1 first operand of this operation
 * @param op2 second operand of this operation
 * @return constructed GIMPLE++ binary operation
 *
 * @see giterator::TruthXorOperation
 */
git_operation createTruthXorOperation(const_git_type return_type, git_operand op1, git_operand op2);

/**
 * Creates GIMPLE++ representation of less than operation.
 *
 * @param return_type return type of this operation
 * @param op1 first operand of this operation
 * @param op2 second operand of this operation
 * @return constructed GIMPLE++ binary operation
 *
 * @see giterator::LtOperation
 */
git_operation createLtOperation (const_git_type return_type, git_operand op1, git_operand op2);

/**
 * Creates GIMPLE++ representation of less than or equal.
 *
 * @param return_type return type of this operation
 * @param op1 first operand of this operation
 * @param op2 second operand of this operation
 * @return constructed GIMPLE++ binary operation
 *
 * @see giterator::LeOperation
 */
git_operation createLeOperation (const_git_type return_type, git_operand op1, git_operand op2);

/**
 * Creates GIMPLE++ representation of greater than operation.
 *
 * @param return_type return type of this operation
 * @param op1 first operand of this operation
 * @param op2 second operand of this operation
 * @return constructed GIMPLE++ binary operation
 *
 * @see giterator::GtOperation
 */
git_operation createGtOperation (const_git_type return_type, git_operand op1, git_operand op2);

/**
 * Creates GIMPLE++ representation of greater than or equal operation.
 *
 * @param return_type return type of this operation
 * @param op1 first operand of this operation
 * @param op2 second operand of this operation
 * @return constructed GIMPLE++ binary operation
 *
 * @see giterator::GeOperation
 */
git_operation createGeOperation (const_git_type return_type, git_operand op1, git_operand op2);

/**
 * Creates GIMPLE++ representation of equal operation.
 *
 * @param return_type return type of this operation
 * @param op1 first operand of this operation
 * @param op2 second operand of this operation
 * @return constructed GIMPLE++ binary operation
 *
 * @see giterator::EqOperation
 */
git_operation createEqOperation (const_git_type return_type, git_operand op1, git_operand op2);

/**
 * Creates GIMPLE++ representation of not equal operation.
 *
 * @param return_type return type of this operation
 * @param op1 first operand of this operation
 * @param op2 second operand of this operation
 * @return constructed GIMPLE++ binary operation
 *
 * @see giterator::NeOperation
 */
git_operation createNeOperation (const_git_type return_type, git_operand op1, git_operand op2);

/**
 * Creates GIMPLE++ representation of non-trapping ordered operation.
 *
 * @param return_type return type of this operation
 * @param op1 first operand of this operation
 * @param op2 second operand of this operation
 * @return constructed GIMPLE++ binary operation
 *
 * @see giterator::OrderedOperation
 */
git_operation createOrderedOperation(const_git_type return_type, git_operand op1, git_operand op2);

/**
 * Creates GIMPLE++ representation of non-trapping unordered operation.
 *
 * @param return_type return type of this operation
 * @param op1 first operand of this operation
 * @param op2 second operand of this operation
 * @return constructed GIMPLE++ binary operation
 *
 * @see giterator::UnorderedOperation
 */
git_operation createUnorderedOperation(const_git_type return_type, git_operand op1, git_operand op2);

/**
 * Creates GIMPLE++ representation of unordered less than operation.
 *
 * @param return_type return type of this operation
 * @param op1 first operand of this operation
 * @param op2 second operand of this operation
 * @return constructed GIMPLE++ binary operation
 *
 * @see giterator::UnLtOperation
 */
git_operation createUnLtOperation (const_git_type return_type, git_operand op1, git_operand op2);

/**
 * Creates GIMPLE++ representation of unordered less than or equal operation.
 *
 * @param return_type return type of this operation
 * @param op1 first operand of this operation
 * @param op2 second operand of this operation
 * @return constructed GIMPLE++ binary operation
 *
 * @see giterator::UnLeOperation
 */
git_operation createUnLeOperation (const_git_type return_type, git_operand op1, git_operand op2);

/**
 * Creates GIMPLE++ representation of unordered greater than operation.
 *
 * @param return_type return type of this operation
 * @param op1 first operand of this operation
 * @param op2 second operand of this operation
 * @return constructed GIMPLE++ binary operation
 *
 * @see giterator::UnGtOperation
 */
git_operation createUnGtOperation (const_git_type return_type, git_operand op1, git_operand op2);

/**
 * Creates GIMPLE++ representation of unordered greater than or equal operation.
 *
 * @param return_type return type of this operation
 * @param op1 first operand of this operation
 * @param op2 second operand of this operation
 * @return constructed GIMPLE++ binary operation
 *
 * @see giterator::UnGeOperation
 */
git_operation createUnGeOperation (const_git_type return_type, git_operand op1, git_operand op2);

/**
 * Creates GIMPLE++ representation of unordered equal operation.
 *
 * @param return_type return type of this operation
 * @param op1 first operand of this operation
 * @param op2 second operand of this operation
 * @return constructed GIMPLE++ binary operation
 *
 * @see giterator::UnEqOperation
 */
git_operation createUnEqOperation (const_git_type return_type, git_operand op1, git_operand op2);

/**
 * Creates GIMPLE++ representation of unordered less than or greater than operation.
 *
 * @param return_type return type of this operation
 * @param op1 first operand of this operation
 * @param op2 second operand of this operation
 * @return constructed GIMPLE++ binary operation
 *
 * @see giterator::LtGtOperation
 */
git_operation createLtGtOperation (const_git_type return_type, git_operand op1, git_operand op2);

/**
 * Creates GIMPLE++ representation of range operation.
 *
 * @param return_type return type of this operation
 * @param op1 first operand of this operation
 * @param op2 second operand of this operation
 * @return constructed GIMPLE++ binary operation
 *
 * @see giterator::RangeOperation
 */
git_operation createRangeOperation (const_git_type return_type, git_operand op1, git_operand op2);

/**
 * Creates GIMPLE++ representation of widening summation.
 *
 * @param return_type return type of this operation
 * @param op1 first operand of this operation
 * @param op2 second operand of this operation
 * @return constructed GIMPLE++ binary operation
 *
 * @see giterator::WidenSumOperation
 */
git_operation createWidenSumOperation (const_git_type return_type, git_operand op1, git_operand op2);

/**
 * Creates GIMPLE++ representation of widening multiplication.
 *
 * @param return_type return type of this operation
 * @param op1 first operand of this operation
 * @param op2 second operand of this operation
 * @return constructed GIMPLE++ binary operation
 *
 * @see giterator::WidenMultOperation
 */
git_operation createWidenMultOperation (const_git_type return_type, git_operand op1, git_operand op2);


/**
 * Gets GIMPLE++ declaration from tree cache based on its GIMPLE @c tree representation.
 *
 * @param gim_help @c GimpleHelpers object to get the declaration from
 * @param tree GIMPLE tree for which we are looking for GIMPLE++ equivalent.
 * @return GIMPLE++ declaration that was created from tree @c tree
 */
const_git_decl get_git_decl_from_tree_cache (git_gimple_helpers gim_help, tree tree);

/**
 * Gets GIMPLE++ constant from tree cache based on its GIMPLE @c tree representation.
 *
 * @param gim_help @c GimpleHelpers object to get the constant from
 * @param tree GIMPLE tree for which we are looking for GIMPLE++ equivalent.
 * @return GIMPLE++ constant that was created from tree @c tree
 */
const_git_constant get_git_constant_from_tree_cache (git_gimple_helpers gim_help, tree tree);

/**
 * Gets GIMPLE++ type from tree cache based on its GIMPLE @c tree representation.
 *
 * @param gim_help @c GimpleHelpers object to get the type from
 * @param tree GIMPLE tree for which we are looking for GIMPLE++ equivalent.
 * @return GIMPLE++ type that was created from tree @c tree
 */
const_git_type get_git_type_from_tree_cache (git_gimple_helpers gim_help, tree tree);


/**
 * Adds GIMPLE++ declaration into tree cache.
 *
 * @param gimple @c Gimple object holding the constructed representation
 * @param gim_help @c GimpleHelpers object to add the declaration to.
 * @param tree GIMPLE tree that was used to create the declaration.
 * @param decl GIMPLE++ declaration that was created from tree @c tree
 */
void add_git_decl_to_tree_cache (git_gimple gimple, git_gimple_helpers gim_help, tree tree, git_decl decl);

/**
 * Adds GIMPLE++ type into tree cache.
 *
 * @param gimple @c Gimple object holding the constructed representation
 * @param gim_help @c GimpleHelpers object to add the type to.
 * @param tree GIMPLE tree that was used to create the type.
 * @param type GIMPLE++ type that was created from tree @c tree
 */
void add_git_type_to_tree_cache (git_gimple gimple, git_gimple_helpers gim_help, tree tree, git_type type);

/**
 * Adds GIMPLE++ constant into tree cache.
 *
 * @param gimple @c Gimple object holding the constructed representation
 * @param gim_help @c GimpleHelpers object to add the constant to.
 * @param tree GIMPLE tree that was used to create the constant.
 * @param cst GIMPLE++ constant that was created from tree @c tree
 */
void add_git_constant_to_tree_cache (git_gimple gimple, git_gimple_helpers gim_help, tree tree, git_constant cst);

/**
 * Reports a GIMPLE++ constant that was encountered during GIMPLE++ construction.
 *
 * @param gimple @c Gimple object holding the constructed representation
 * @param constant GIMPLE++ constant encountered during construction
 *
 * @see giterator::Gimple::reportCst
 */
void reportConstant (git_gimple gimple, git_constant constant);

/**
 * Reports a GIMPLE++ function that was encountered during GIMPLE++ construction.
 *
 * @param gimple @c Gimple object holding the constructed representation
 * @param funcDecl GIMPLE++ declaration of the function @c func
 * @param func GIMPLE++ function definition encountered during construction
 *
 * @see giterator::Gimple::reportFuncDef
 */
void reportFunction (git_gimple gimple, const_git_decl funcDecl, git_function func);

/**
 * Reports program entry point function that was encountered during GIMPLE++ construction.
 *
 * @param gimple @c Gimple object holding the constructed representation
 * @param mainDecl GIMPLE++ declaration of the program entry point function
 * encountered during construction
 *
 * @see giterator::Gimple::reportMain
 */
void reportMain (git_gimple gimple, const_git_decl mainDecl);

/**
 * Reports a GIMPLE++ declaration that was encountered during GIMPLE++ construction.
 *
 * @param gimple @c Gimple object holding the constructed representation
 * @param decl GIMPLE++ declaration encountered during construction
 *
 * @see giterator::Gimple::reportDecl
 */
void reportDecl (git_gimple gimple, git_decl decl);


/**
 * Find out whether function of specified name was already processed.
 *
 * @param gim_help @c GimpleHelpers object used for the check
 * @param name examined function name
 * @return true if function of name @name was already processed earlier,
 * false otherwise.
 */
bool checkFuncNameExists(git_gimple_helpers gim_help, const char* name);

/**
 * Add a function name to the list of already processed functions.
 *
 * @param gim_help @c GimpleHelpers object used to store the name
 * @param name function name to store
 */
void addFuncName(git_gimple_helpers gim_help, const char* name);

/**
 * Set a flag informing that global declarations were already processed.
 *
 * @param gim_help @c GimpleHelpers object used to set the flag
 */
void setGlobalsResolved(git_gimple_helpers gim_help);

/**
 * Find out whether the global declarations were already processed.
 *
 * @param gim_help @c GimpleHelpers object used to find out
 * @return true if the globals were processed, false otherwise
 */
bool getGlobalsResolved(git_gimple_helpers gim_help);

/**
 * Add a type to the list of already processed types.
 * Used to check for recursive definitions in structures (like linked lists).
 *
 * @param gim_help @c GimpleHelpers object used to store the type
 * @param typeName name of the type to store
 * @param type the stored type
 */
void addDefinedType (git_gimple_helpers gim_help, const char* typeName, git_type type);

/**
 * Check if type of specified name was already processed and if it was, return it.
 *
 * @param gim_help @c GimpleHelpers object used for the check
 * @param typeName name of the checked type
 * @return already processed type of name @c typeName,
 * or NULL if it was not processed yet
 */
git_type getDefinedType (git_gimple_helpers gim_help, const char* typeName);

/**
 * Add a type to the list of already processed types.
 * Used to check for recursive definitions in structures (like linked lists).
 *
 * @param gim_help @c GimpleHelpers object used to store the type
 * @param typeTree GIMPLE representation of the type to store
 * @param type the stored type
 */
void addDefinedType2 (git_gimple_helpers gim_help, tree typeTree, git_type type);

/**
 * Check if type of specified name was already processed and if it was, return it.
 *
 * @param gim_help @c GimpleHelpers object used for the check
 * @param typeTree GIMPLE representation of the checked type
 * @return already processed type of GIMPLE tree type @c typeTree
 * or NULL if it was not processed yet
 */
git_type getDefinedType2 (git_gimple_helpers gim_help, tree typeTree);

/**
 * Find out whether recursive type is being processed.
 *
 * @param gim_help @c GimpleHelpers object that stores the flag
 * @return true of recursive type is being processed,
 * false otherwise
 */
bool isRecursive(git_gimple_helpers gim_help);

/**
 * Set recursive flag determining that recursive type is being processed (like linked list).
 *
 * @param gim_help @c GimpleHelpers object used to store the flag
 * @param recursive value of the flag
 */
void setRecursive(git_gimple_helpers gim_help, bool recursive);

/**
 * Store a constructor (initializer) for constructed GIMPLE++ declaration.
 *
 * @param gim_help @c GimpleHelpers object used to store the initializer
 * @param decl the declaration that has the initializer @c op
 * @param op operand representing initilizer for declaration @c decl
 */
void setConstructor(git_gimple_helpers gim_help, const_git_decl decl, git_operand op);

/**
 * Get a constructor (initializer) of a GIMPLE++ declaration.
 *
 * @param gim_help @c GimpleHelpers object used to get the initializer
 * @param decl the declaration for which to get the initializer
 * @return operand representing initilizer for declaration @c decl,
 *  or NULL if it has on initializer
 */
git_operand getConstructor(git_gimple_helpers gim_help, const_git_decl decl);

/**
 * Get a type of a GIMPLE++ declaration.
 *
 * @param decl declaration to get the type
 * @return GIMPLE++ type of declaration @c decl
 */
const_git_type declGetType(const_git_decl decl);

#ifdef __cplusplus
}}

}
#endif

#endif /* GIT_CONSTRUCTION_HELPERS_H */
