// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file CstOperand.h
 * Ooperand representing a constant.
 */
#ifndef GIT_CST_OPERAND_H
#define GIT_CST_OPERAND_H

#include "Operand.h"
#include "Constant.h"

#include <boost/serialization/serialization.hpp>

namespace giterator { 
    
/**
 * Operand representing a constant.
 * Constant operands get created from these GIMPLE nodes,
 * that satisfy @c CONSTANT_CLASS_P macro. They are
 * regular GIMPLE constants, that get converted
 * to GIMPLE++ constants and then wrapped in this operand
 * class.
 *
 * @see create_git_operand_from_tree to see how constant
 * operands get created from GIMPLE constants
 */
class CstOperand : public Operand {
  private:    
    /*! the constant represented by this operand */
    const Constant* m_cst;
    
  public:
    /**
	 * Constructor, creates a new constant operand.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
    CstOperand () {}
    /**
     * Constructor, creates a new constant operand.
     * 
     * @param cst the constant that will be represented by the new operand
     */
    CstOperand (const Constant* cst) : m_cst (cst) {}
    
    /**
     * Returns the constant represented by this operand.
     * 
     * @return the constant represented by this operand
     */
    const Constant* getConstant () const 
    { return m_cst; }    
    
    virtual void accept (Visitor& v) const
    { v.visit (*this); }
    
    /**
     * Returns the type of the constant represented by this operand.
     *
     * @return type of the constant represented by this operand
     */
    const Type* getType () const
    { return m_cst->getType (); }

    friend class boost::serialization::access;
	template<class Archive>
	void save(Archive & ar, const unsigned int) const
	{
		ar << m_cst;
	}

	template<class Archive>
	void load(Archive & ar, const unsigned int)
	{
		Constant* cst;
		ar >> cst;
		m_cst = (const Constant*) cst;
	}

	template<class Archive>
	void serialize(Archive & ar, const unsigned int version)
	{
		boost::serialization::void_cast_register(static_cast<CstOperand *>(NULL), static_cast<Operand *>(NULL));
		boost::serialization::split_member(ar, *this, version);
	}
};

}

#endif /* GIT_CST_OPERAND_H */
