// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file Decl.h
 * Base class for classes representing declarations.
 */
#ifndef GIT_DECL_H
#define GIT_DECL_H

#include "TreeProduct.h"

#include <string>

namespace giterator { 

class Type;
class FunctionDecl;
class ParamDecl;
class VarDecl;
class RecordDecl;
class FieldDecl;
class LabelDecl;
class ResultDecl;
class TypeDecl;
class ConstDecl;


/**
 * Interface for classes representing declarations.
 * Declarations get created form GIMPLE *_DECL nodes.
 * Subclasses of this interface represent concrete declarations.
 * Each ..._DECL node has DECL_NAME, DECL_UID and TREE_TYPE.
 * Information from these fields is stored for every Decl subclass.
 *
 * @see create_git_decl_from_tree() in gimplexx_bridge.c
 * to see how declarations get created from GIMPLE constants.
 */
class Decl : public detail::TreeProduct {
  public:
    /**
     * Visitor class for sub-classes of @c Decl class, that are meant to be used 
     * by the user.
     */
    struct Visitor {
        virtual ~Visitor () {}
        
        virtual void visit (const FunctionDecl& decl)     = 0;
        virtual void visit (const ParamDecl& decl)        = 0;
        virtual void visit (const VarDecl& decl)          = 0;
        virtual void visit (const RecordDecl& decl)		  = 0;
        virtual void visit (const FieldDecl& decl)		  = 0;
        virtual void visit (const LabelDecl& decl)        = 0;
        virtual void visit (const ResultDecl& decl)       = 0;
        virtual void visit (const TypeDecl& decl)         = 0;
        virtual void visit (const ConstDecl& decl)        = 0;
    };
    
  public:
    /**
     * Destructor.
     */
    virtual ~Decl () {}
    
    /**
     * Returns unique ID of this declaration.
     * This ID is obtained from GIMPLE using @c DECL_UID() macro.
     * 
     * @return unique ID of this declaration
     */
    virtual unsigned int getUid () const = 0;
    
    /**
     * Returns name of the declaration.
     * The name is obtained from GIMPLE using @c DECL_NAME() macro.
     * 
     * @return name of the declaration
     */
    virtual std::string getName () const = 0;
    
    /**
     * Returns type of the declaration.
     * The type is obtained from GIMPLE using @c TREE_TYPE() macro.
     * @return type of the declaration
     */
    virtual const Type* getType () const = 0;

    /** Checks if the declaration is the same declaration as
     * another declaration. Declarations are equal, when their
     * unique IDs are equal.
     *
     * @param decl declaration to be checked for equality
     * @return true if the declarations are equal, false otherwise
     */
    virtual bool isSame(const Decl& decl) const = 0;

    /**
     * Calls @c visitor.accept().
     * Classes implementing this interface implement this method to call @c visitor.accept()
     * with @c *this as a parameter.
     * 
     * @param visitor visitor on which to call the @c accept method
     */
    virtual void accept (Visitor& visitor) const = 0;
};

}

#endif /* GIT_DECL_H */
