// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file DeclOperand.h
 * Operand representing a declaration.
 */
#ifndef GIT_DECL_OPERAND_H
#define GIT_DECL_OPERAND_H

#include "Operand.h"
#include "Decl.h"
#include "Serialization.h"

namespace giterator { 

/**
 * Operand representing a declaration.
 * Declaration operands get created from these GIMPLE nodes,
 * that satisfy @c DECL_P macro. They are
 * regular GIMPLE declarations, that get converted
 * to GIMPLE++ declarations and then wrapped in this operand
 * class.
 *
 * @see create_git_operand_from_tree to see how declaration
 * operands get created from GIMPLE operands
 */
class DeclOperand : public Operand {
  private:
    /*! the declaration represented by this operand */
    const Decl* m_decl;
      
  public:
    /**
	 * Constructor, creates a new declaration operand.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
    DeclOperand () {}
    /**
     * Constructor, creates a new declaration operand.
     * 
     * @param decl the declaration that will be represented by the new operand
     */
    DeclOperand (const Decl* decl) : m_decl (decl) {}
    
    /**
     * Returns the declaration represented by this operand.
     * 
     * @return the declaration represented by this operand
     */
    const Decl* getDecl () const 
    { return m_decl; }
    
    void accept (Visitor& v) const
    { v.visit (*this); }
    
    /**
	 * Returns the type of the declaration represented by this operand.
	 *
	 * @return type of the declaration represented by this operand
	 */
    const Type* getType () const
    { return m_decl->getType (); }

    friend class boost::serialization::access;
	template<class Archive>
	void save(Archive & ar, const unsigned int) const
	{
		ar << m_decl;
	}

	template<class Archive>
	void load(Archive & ar, const unsigned int)
	{
		Decl* decl;
		ar >> decl;
		m_decl = (const Decl*) decl;
	}

	template<class Archive>
	void serialize(Archive & ar, const unsigned int version)
	{
		boost::serialization::void_cast_register(static_cast<DeclOperand *>(NULL), static_cast<Operand *>(NULL));
		boost::serialization::split_member(ar, *this, version);
	}
};

}

#endif /* GIT_DECL_OPERAND_H */
