// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file DecoratingOperand.h
 * Base class for operands that decorate some other operand.
 */
#ifndef GIT_DECORATING_OPERAND_H
#define GIT_DECORATING_OPERAND_H

#include "Operand.h"

#include <boost/serialization/serialization.hpp>
#include <memory>


namespace giterator { namespace detail {

/**
 * Base class for all operands that only decorate some other operand,
 * like indirect reference operands ( *op in C) or address operands
 * ( &op in C ).
 * This class stores the actual operand that it decorates.
 * Concrete operands are subclasses of this class.
 *
 * @see giterator::AddrOperand, giterator::IndirectRefOperand
 */
class DecoratingOperand : public Operand {
    private:
        /*! type of this operand */
        const Type* m_type;
        
        /*! the decorated operand */
        std::auto_ptr<Operand> m_subOper;
    
    public:
        /**
		 * Constructor, creates a new decorationg operand.
		 * Empty constructor exists for serialization purposes,
		 * use nonempty for normal use.
		 */
        DecoratingOperand () {}
        /**
         * Constructor, creates a new decorationg operand.
         * 
         * @param type type of this operand
         * @param subOper the decorated operand
         */
        DecoratingOperand (const Type* type, std::auto_ptr<Operand> subOper) : 
            m_type (type), m_subOper (subOper) {}
    
        /**
         * Returns the decorated operand.
         * 
         * @return the decorated operand
         */
        const Operand* getSubOperand () const 
        { return m_subOper.get (); }
        
        /**
         * Returns the type of this operand (not the type of the suboperand!).
         * 
         * @return the type of this operand
         */
        const Type* getType () const
        { return m_type; }

		friend class boost::serialization::access;
		template<class Archive>
		void save(Archive & ar, const unsigned int) const
		{
			ar << m_type;
			Operand* subOper = m_subOper.get();
			ar << subOper;
		}

		template<class Archive>
		void load(Archive & ar, const unsigned int)
		{
			Type* type;
			ar >> type;
			m_type = (const Type*) type;
			Operand* subOper;
			ar >> subOper;
			m_subOper = std::auto_ptr<Operand> (subOper);
		}

		template<class Archive>
		void serialize(Archive & ar, const unsigned int version)
		{
			boost::serialization::split_member(ar, *this, version);
		}
};

}}

#endif /* GIT_DECORATING_OPERAND_H */
