// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file Edge.h
 * Class representing an edge between blocks.
 */
#ifndef GIT_EDGE_H
#define GIT_EDGE_H

#include "Serialization.h"

namespace giterator {
    
class Block;

/**
 * Class representing an edge between blocks.
 * Edges in GIMPLE represent possible control flow transfers from the end
 * of some basic block to the start of another basic block.
 * Edges between GIMPLE blocks are represented via GIMPLE @c edge structure.
 * Edge class stores pointers to source and destination block as well as
 * flags extracted from GIMPLE @c edge structure.
 * 
 * @see giterator::Block
 * @see add_git_edge in gimplexx-bridge.c to see how edges get created from
 * GIMPLE edges.
 */
class Edge {
    
  private:    
    /*! source block of the edge */
    const Block* m_src;
    /*! destination block of the edge */
    const Block* m_dest;
    
    /** @name Flags
     * Edge flags. 
     */
    //@{
    /** flag if this edge is a fallthru edge
     * Edge is fallthru edge when the block may continue execution
     * to the following one without branching.
     */
    bool m_fallthru : 1;
    /** flag if this edge is a true branch
     * Edge is true when it is representing control flow for true result
     * of conditional statement.
     */
    bool m_trueBranch : 1;
    /*! flag if this edge is a false branch
     * Edge is true when it is representing control flow for false result
     * of conditional statement.
     */
    bool m_falseBranch : 1;
    //@}
  public:
    /**
	 * Constructor, creates a new edge.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	Edge () {}
    /**
     * Constructor, creates a new edge.
     * 
     * @param src source of the edge
     * @param dest destination of the edge
     * @param fallthru value of the fallthru flag
     * @param trueBranch value of the true branch
     * @param falseBranch value of the false branch
     */
    Edge (const Block* src, const Block* dest, bool fallthru, bool trueBranch, bool falseBranch) 
        : m_src (src), m_dest (dest), m_fallthru (fallthru), 
          m_trueBranch (trueBranch), m_falseBranch (falseBranch) {}
    
    /**
     * Returns the source block of this edge.
     * 
     * @return the source block of this edge
     */
    const Block* getSrc () const
    { return m_src; }
    
    /**
     * Returns the destination of this edge.
     * 
     * @return the destination of this edge
     */
    const Block* getDest () const
    { return m_dest; }
    
    
    /**
     * Checks if this edge is a fallthru edge.
     * 
     * @return @c true if this edge is a fallthru edge, @c false otherwise
     */
    bool isFallthru () const
    { return m_fallthru; }
    
    /**
     * Checks if this edge is a true branch edge.
     * 
     * @return @c true if this edge is a true branch edge, @c false otherwise
     */
    bool isTrueBranch () const
    { return m_trueBranch; }
    
    /**
     * Checks if this edge is a false branch edge.
     * 
     * @return @c true if this edge is a false branch edge, @c false otherwise
     */
    bool isFalseBranch () const
    { return m_falseBranch; }

    friend class boost::serialization::access;
	template<class Archive>
	void save(Archive & ar, const unsigned int) const
	{
		ar << m_src;
		ar << m_dest;
		bool fallthru = m_fallthru;
		bool trueBranch = m_trueBranch;
		bool falseBranch = m_falseBranch;
		ar << fallthru;
		ar << trueBranch;
		ar << falseBranch;
	}

	template<class Archive>
	void load(Archive & ar, const unsigned int)
	{
		Block* src;
		Block* dest;
		ar >> src;
		ar >> dest;
		m_src = (const Block*) src;
		m_dest = (const Block*) dest;
		bool fallthru;
		bool trueBranch;
		bool falseBranch;
		ar >> fallthru;
		ar >> trueBranch;
		ar >> falseBranch;
		m_fallthru = fallthru;
		m_trueBranch = trueBranch;
		m_falseBranch = falseBranch;
	}

	template<class Archive>
	void serialize(Archive & ar, const unsigned int version)
	{
		boost::serialization::split_member(ar, *this, version);
	}
};

}

#endif /* GIT_EDGE_H */
