// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file EnumType.h
 * Class representing an enumeral type.
 */
#ifndef GIT_ENUM_TYPE_H
#define GIT_ENUM_TYPE_H

#include "AbstractType.h"

#include "IntegerType.h"

#include <map>
#include <string>
#include <cassert>
#include <boost/serialization/serialization.hpp>
#include <boost/serialization/map.hpp>

namespace giterator {

/**
 * Class representing enumeral type.
 * It stores all the enumeration constants of this type.
 * This type gets created from GIMPLE @c ENUMERAL_TYPE
 * tree nodes.
 *
 * @see create_git_type_from_tree to see how enumeral types
 * get created from GIMPLE types
 */
class EnumType : public detail::AbstractType {
  private:
	typedef std::map<const std::string, const Constant*> Constants;
	typedef Constants::const_iterator ConstantsIterator;
    /*! type precision in bits */
    size_t m_bitPrecision;
    /*! flag whether all fields are unsigned */
    bool m_isUnsigned;

    /*! enumeration constants for this type */
    Constants m_constants;

  public:
    /**
	 * Constructor, creates a new enumeral type.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
    EnumType() : detail::AbstractType() {}
    /**
     * Constructor, creates a new enumeral type.
     *
     * @param name name of the type
     * @param bitSize size of the type in bits
     * @param isUnsigned @c true if all fields are unsigned
     * @param bitPrecision the presision of the new type
     */
    EnumType (std::string& name,
			  size_t bitSize,
			  bool isUnsigned,
			  size_t bitPrecision)
        : detail::AbstractType (name, bitSize),
          m_bitPrecision (bitPrecision), m_isUnsigned (isUnsigned)
    {}

    /**
     * Returns the bit precision of this type.
     *
     * @return the precision of this type in bits
     */
    size_t getBitPrecision() const
    {
      return m_bitPrecision;
    }

    /**
     * Checks if this type is unsigned.
     *
     * @return @c true if this type is unsigned, @c false otherwise
     */
    size_t isUnsigned () const
    {
      return m_isUnsigned;
    }

    /**
     * Add an enumeration constants to constants for this type.
     *
     * @param name name of the new constant
     * @param constant GIMPLE++ representation of the new constant
     */
    void addConstant(std::string name, const Constant* constant)
    {
    	assert(!containsConstant(name));
    	m_constants[name] = constant;
    }

    /**
     * Get an enumeration constant from constants for this type
     * based on the name of the constant.
     *
     * @param name name of the constant to get
     * @return pointer to the constant with name @c name
     */
    const Constant* getConstant(std::string name) const
    {
    	assert(containsConstant(name));
    	return m_constants.find(name)->second;
    }

    /**
     * Check whether this type contains enumeration constant with
     * specified name.
     *
     * @param name name of the constant to check
     * @return true if this type contains constant with name @c name,
     * false otherwise
     */
    bool containsConstant(std::string name) const
    {
    	return m_constants.find(name) != m_constants.end();
    }

    /**
	 * Compare this type with another type.
	 *
	 * @param type type to be compared
	 * @return true if the types are the same (they are enumeral an have the same constants
	 * or the compared type is integer type), false otherwise
	 */
    bool isSame (const Type& type) const
    {
    	const IntegerType* intType = dynamic_cast<const IntegerType*>(&type);
    	if (intType)
    	{
    		return getBitSize() == intType->getBitSize()
    				&& getBitSize() == intType->getBitSize()
				    && getBitPrecision() == intType->getBitPrecision()
				    && isUnsigned() == intType->isUnsigned();
    	}
    	else
    	{

    		const EnumType* enumType = dynamic_cast<const EnumType*>(&type);
    		if (enumType == 0)
    			return false;

    		return valuesEqualWith(enumType)
    				&& getBitSize() == enumType->getBitSize()
					&& getBitPrecision() == enumType->getBitPrecision()
					&& isUnsigned() == enumType->isUnsigned();
    	}
    }

    friend class boost::serialization::access;
	template<class Archive>
	void save(Archive & ar, const unsigned int) const
	{
		ar << boost::serialization::base_object<AbstractType>(*this);
		ar << m_bitPrecision;
		ar << m_isUnsigned;

		size_t size =  m_constants.size();
		ar << size;
		for (ConstantsIterator i1 = m_constants.begin(); i1 != m_constants.end(); ++i1)
		{
			std::string name = i1->first;
			const Constant* constant = i1->second;
			ar << name;
			ar << constant;
		}

	}

	template<class Archive>
	void load(Archive & ar, const unsigned int)
	{
		ar >> boost::serialization::base_object<AbstractType>(*this);
		ar >> m_bitPrecision;
		ar >> m_isUnsigned;

		size_t size;
		ar >> size;
		for (size_t i = 0; i < size; ++i)
		{
			std::string name;
			Constant* constant;
			ar >> name;
			ar >> constant;
			m_constants[name] = (const Constant*) constant;
		}
	}

	template<class Archive>
	void serialize(Archive & ar, const unsigned int version)
	{
		boost::serialization::void_cast_register(static_cast<EnumType *>(NULL), static_cast<TreeProduct *>(NULL));
		boost::serialization::void_cast_register(static_cast<EnumType *>(NULL), static_cast<Type *>(NULL));
		boost::serialization::split_member(ar, *this, version);
	}
  private:

	bool valuesEqualWith(const EnumType* type) const
	{
		for (ConstantsIterator i1 = m_constants.begin(); i1 != m_constants.end(); ++i1)
		{
			if (!type->containsConstant(i1->first))
				return false;

			if (type->getConstant(i1->first) != i1->second)
				return false;
		}

		return true;
	}
};

}

#endif /* GIT_ENUM_TYPE_H */

