// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file Function.h
 * Class representing function definition.
 */
#ifndef GIT_FUNCTION_H
#define GIT_FUNCTION_H

#include "Block.h"
#include "../git/Decl.h"

#include <vector>
#include <boost/noncopyable.hpp>
#include <boost/ptr_container/ptr_vector.hpp>
#include <boost/serialization/serialization.hpp>
#include <boost/ptr_container/serialize_ptr_vector.hpp>

namespace giterator { 
    
class FunctionType;
class Decl;

/**
 * Class representing function definition.
 * Function definition consists of the type of the function,
 * blocks of function statements (control-flow graph of the
 * function), parameters and local declarations of the
 * function. Simply everything, that is needed for simulation
 * of running the function.
 * Functions get created from GIMPLE @c function structures.
 *
 * @see gb_process_cfun() in gimplexx-bridge.c to see how
 * functions get created from GIMPLE functions
 */
class Function : private boost::noncopyable {
  public:
    /*! type of the container with local declarations */
    typedef std::vector<const Decl*> Declarations;
    /*! type of the container for this function blocks */
    typedef boost::ptr_vector<Block> Blocks;

  private:
    /*! type of this function */
    const FunctionType* m_type;

    /*! blocks of statements that contain the body of this function */
    Blocks m_blocks;

    /*! the first block of this function */
    Block* m_firstBlock;
    
    /*! parameter declarations */
    Declarations m_paramDecl;
    
    /*! local declarations (variables, functions, ...). */
    Declarations m_localDecl;

  public:
    /**
	 * Constructor, creates a new function.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	Function () : m_type (NULL), m_firstBlock (NULL) {}
    /**
     * Constructor, creates a new function.
     * 
     * @param type type of the function
     * @param function function structure representing this function in GIMPLE
     */
    Function (const FunctionType* type) :
        m_type (type), m_firstBlock (NULL) {}

    /**
     * Returns the type of this function.
     * 
     * @return the type of this function
     */
    const FunctionType* getFunctionType () const 
    { return m_type; }
    
    /**
	 * Adds the block to this function.
	 *
	 * @param block GIMPLE++ block created from GIMPLE @c basic_block
	 */
	void addBlock (std::auto_ptr<Block> block)
	{
		m_blocks.push_back(block);
	}

    /**
	 * Sets the first block of this function, where
	 * the execution starts.
	 *
	 * @param block block to be set as the first block
	 */
	void setFirstBlock (Block* block)
	{
		m_firstBlock = block;
	}

	/**
	 * Returns the first block of this function.
	 * The execution of the function starts at the first block.
	 *
	 * @return the first block of this function
	 */
	Block* getFirstBlock () const
	{ return (m_firstBlock); }

	const Blocks& getBlocks() const
	{ return m_blocks; }
    
    /**
     * Returns parameter declarations of this function.
     * 
     * @return parameter declarations of this function
     */
    const Declarations& getParamDeclarations () const
    { return m_paramDecl; }
    
    /**
     * Returns local declarations in this function.
     * 
     * @return local declarations in this function
     */
    const Declarations& getLocalDeclarations () const
    { return m_localDecl; }
  protected:
    /**
     * Adds local declaration for this function.
     * 
     * @param decl declaration to add to the list of local declarations
     */
    void addLocalDecl (const Decl* decl)
    {
        m_localDecl.push_back (decl);
    }
    
    /**
     * Adds parameter declaration for this function.
     * 
     * @param decl declaration to add into the parameters list
     */
    void addParamDecl (const Decl* decl)
    {
        m_paramDecl.push_back (decl);
    }

  private:
    friend void detail::functionAddLocalDecl (Function* function, const Decl* decl);
    friend void detail::functionAddParamDecl (Function* function, const Decl* decl);
    friend void functionAddBlock (git_function function, git_function_helpers func_help, basic_block basicBlock, git_block block);
    friend void functionSetFirstBlock (git_function function, git_function_helpers func_help, basic_block basicBlock);
  public:
    friend class boost::serialization::access;
	template<class Archive>
	void save(Archive & ar, const unsigned int) const
	{
		ar << m_type;
		ar << m_blocks;
		ar << m_firstBlock;
		size_t size = m_paramDecl.size();
		ar << size;
		for(Declarations::const_iterator i1 = m_paramDecl.begin(); i1 != m_paramDecl.end(); ++i1)
		{
			ar << (*i1);
		}
		size = m_localDecl.size();
		ar << size;
		for(Declarations::const_iterator i1 = m_localDecl.begin(); i1 != m_localDecl.end(); ++i1)
		{
			ar << (*i1);
		}
	}

	template<class Archive>
	void load(Archive & ar, const unsigned int)
	{
		FunctionType* type;
		ar >> type;
		m_type = (const FunctionType*) type;
		ar >> m_blocks;
		ar >> m_firstBlock;
		size_t size;
		ar >> size;
		m_paramDecl.resize(size);
		Decl* decl;
		for (size_t i = 0; i < size; ++i)
		{
			ar >> decl;
			m_paramDecl[i] = (const Decl*) decl;
		}
		ar >> size;
		m_localDecl.resize(size);
		for (size_t i = 0; i < size; ++i)
		{
			ar >> decl;
			m_localDecl[i] = (const Decl*) decl;
		}
	}

	template<class Archive>
	void serialize(Archive & ar, const unsigned int version)
	{
		boost::serialization::split_member(ar, *this, version);
	}
};

}

#endif /* GIT_FUNCTION_H */

