// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file FunctionDecl.h
 * Class representing function declaration.
 */
#ifndef GIT_FUNCTION_DECL_H
#define GIT_FUNCTION_DECL_H

#include "AbstractDecl.h"
#include "FunctionType.h"
#include "FunctionInfo.h"
#include "CTypes.h"

#include <cassert>
#include <boost/serialization/serialization.hpp>

namespace giterator { 
    
/**
 * Class representing function declaration.
 * Apart from base declaration information inherited
 * from @c AbstractDecl, function declarations stores
 * if it is a declaration of a builtin (standard) function
 * and if yes, what kind of builtin function (printf? malloc? ...)
 * it is.
 * FunctionDecls get created from GIMPLE @c FUNCTION_DECL nodes.
 *
 * @see tree_to_git_decl() in gimplexx-bridge.c to see
 * how function declarations get created from GIMPLE
 * declarations.
 */
class FunctionDecl : public detail::AbstractDecl {
  private:
    /** 
     * The code of function, determining what kind of function it is.
     *
     */
	enum function_code m_code;
    /**
     * Flag specifying if this function is a built-in (or standard) function or not.
     * Built-in functions are, e.g., @c malloc(), @c free(), @c puts(), ... .
     * Standard fucntions are e.g. @c strtod, ... .
     */
    bool m_isBuiltin : 1;
    
  public:
    /**
	 * Constructor, creates a new function declaration.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
    FunctionDecl () : detail::AbstractDecl() {}
    /**
     * Constructor, creates a new function declaration.
     * 
     * @param uid unique ID of the declaration
     * @param name function name
     * @param type type of the function (NOT the return type)
     * @param isBuiltin flag specifying if the new declaration is a declaration of
     *      built-in (or standard) function or not
     * @param code the code of the built-in (standard) function
     */
    FunctionDecl (unsigned int uid, const std::string& name, 
                  const Type* type, enum function_code code)
        : AbstractDecl (uid, name, type), m_code (code), m_isBuiltin (code != NOT_BUILT_IN_FUNC)
    	{}
    
    /**
     * Return the type of this function (that is of sub-type @c FunctionType).
     * 
     * @return the type of this function
     */
    const FunctionType* getFunctionType () const
    {
        const FunctionType* funcType = dynamic_cast<const FunctionType*> (getType ());
        assert (funcType);
        
        return funcType;
    }
    
    /**
     * Returns the code of the function that is declared by this declaration.
     * 
     * @return the code of the function that is declared by this declaration
     */
    enum function_code getCode () const
    { return m_code; }
    
    /**
     * Returns if this declaration is a declaration of built-in (standard) function.
     * 
     * @return @c true if this declaration is a declaration of built-in (standard) function,
     *      @c false otherwise
     */
    bool isBuiltin () const
    { return m_isBuiltin; }

    void accept (Visitor& visitor) const
    { visitor.visit (*this); }

    friend class boost::serialization::access;
	template<class Archive>
	void save(Archive & ar, const unsigned int) const
	{
		ar << boost::serialization::base_object<AbstractDecl>(*this);
		enum function_code code = m_code;
		ar << code;
		bool isBuiltin = m_isBuiltin;
		ar << isBuiltin;
	}

	template<class Archive>
	void load(Archive & ar, const unsigned int)
	{
		ar >> boost::serialization::base_object<AbstractDecl>(*this);
		enum function_code code;
		ar >> code;
		m_code = code;
		bool isBuiltin;
		ar >> isBuiltin;
		m_isBuiltin = isBuiltin;
	}

	template<class Archive>
	void serialize(Archive & ar, const unsigned int version)
	{
		boost::serialization::void_cast_register(static_cast<FunctionDecl *>(NULL), static_cast<TreeProduct *>(NULL));
		boost::serialization::void_cast_register(static_cast<FunctionDecl *>(NULL), static_cast<Decl *>(NULL));
		boost::serialization::split_member(ar, *this, version);
	}
};

}

#endif /* GIT_FUNCTION_DECL_H */
