// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file FunctionType.h
 * Class representing type of a function.
 */
#ifndef GIT_FUNCTIONTYPE_H
#define GIT_FUNCTIONTYPE_H

#include "AbstractType.h"

#include <vector>
#include <algorithm>
#include <boost/serialization/serialization.hpp>

namespace giterator { 
    
/**
 * Class representing the type of a function.
 * Function type stores return type of the function as well as
 * types of parameters of the function.
 * The type of a function is never used directly in the program,
 * only the pointer to the function is used, therefore the 
 * method @c getBitSize() must not be called on the objects of this class.
 * This type gets created from GIMPLE @c FUNCTION_TYPE
 * tree nodes.
 *
 * @see create_git_type_from_tree in gimplexx-bridge.c to see how function types
 * get created from GIMPLE types
 */
class FunctionType : public detail::AbstractType {
  private:
    /*! type of the container holding the types of the parameters of the function */
    typedef std::vector<const Type*> ParameterTypes;
    
  public:
    /*! type of the iterator through the types of the parameters of the function */
    typedef ParameterTypes::const_iterator ParamTypeIter;

  private:
    /*! return type of the functions of this type */
    const Type* m_returnType;
    
    /*! types of parameters of the functions of this type */
    ParameterTypes m_parameterTypes;

  public:
    /**
	 * Constructor, creates a new function type.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
    FunctionType() : detail::AbstractType() {}
    /**
     * Constructor, creates a new function type.
     * 
     * @param name name of the type
     * @param returnType return type of the function
     */
    FunctionType (std::string& name, const Type* returnType) :
          detail::AbstractType (name, 0), m_returnType (returnType) {}

    /**
     * Get size of the function in bits
     * The size of every function is 8 bits.
     *
     * @returns size of the function in bits.
     */
    size_t getBitSize () const 
    { return 8; }

    /**
     * Returns the type of return value of the functions of this type.
     * 
     * @return the type of return value of the functions of this type
     */
    const Type* getReturnType () const 
    { return m_returnType; }
    
    /**
     * Returns the range of parameter types of the functions of this type.
     * 
     * @return the range of parameter types of the functions of this type
     */
    std::pair<ParamTypeIter, ParamTypeIter> getParameterTypes () const
    { return std::make_pair (m_parameterTypes.begin (), m_parameterTypes.end ()); }

    /**
     * Adds a type of a parameter to this function type.
     * 
     * @param type the type of the parameter that should be added to the 
     *      list of parameters of the functions of this type
     */
    void addParameterType (const Type* type) 
    {
      m_parameterTypes.push_back (type);
    }

    /**
	 * Compare this type with another type.
	 *
	 * @param type type to be compared
	 * @return true if the types are the same
	 * (they have the same return value and parameters),
	 * false otherwise
	 */
    bool isSame (const Type& type) const
    {
        const FunctionType* funcType = dynamic_cast<const FunctionType*> (&type);
        if (!funcType)
            return false;
        
        
        if (!m_returnType->isSame (*(funcType->m_returnType)) ||
                m_parameterTypes.size () != funcType->m_parameterTypes.size ())
            return false;
        
        return std::equal (m_parameterTypes.begin (), m_parameterTypes.end (), 
                            funcType->m_parameterTypes.begin (), compareTypes);
    }

    friend class boost::serialization::access;
	template<class Archive>
	void save(Archive & ar, const unsigned int) const
	{
		ar << boost::serialization::base_object<AbstractType>(*this);
		ar << m_returnType;
		size_t size = m_parameterTypes.size();
		ar << size;
		for (ParamTypeIter i1 = m_parameterTypes.begin(); i1 != m_parameterTypes.end(); ++i1)
		{
			ar << (*i1);
		}
	}

	template<class Archive>
	void load(Archive & ar, const unsigned int)
	{
		ar >> boost::serialization::base_object<AbstractType>(*this);
		Type* type;
		ar >> type;
		m_returnType = (const Type*) type;
		size_t size;
		ar >> size;
		m_parameterTypes.resize(size);
		for (size_t i = 0; i < size; ++i)
		{
			ar >> type;
			m_parameterTypes[i] = (const Type*) type;
		}
	}

	template<class Archive>
	void serialize(Archive & ar, const unsigned int version)
	{
		boost::serialization::void_cast_register(static_cast<FunctionType *>(NULL), static_cast<TreeProduct *>(NULL));
		boost::serialization::void_cast_register(static_cast<FunctionType *>(NULL), static_cast<Type *>(NULL));
		boost::serialization::split_member(ar, *this, version);
	}
};

}

#endif  /* GIT_FUNCTIONTYPE_H */

