/** @file GccBridge.c
 * This file contains helper C functions used by @c Gimple class to access 
 * the GIMPLE representation of the program.
 * The @c Gimple class can not access the C structures directly, because the GIMPLE
 * can not be access from C++.
 */
#include "GccBridge.h"

#include "../../config.h"
#include "../../system.h"
#include "../../coretypes.h"
#include "../../tm.h"
#include "../../real.h"

#include "../../tree-flow.h"
#include "../../cgraph.h"
#include "../../langhooks.h"

#include "ConstructionHelpers.h"
#include "../Error.h"

#include <stdlib.h>

/**
 * Transforms GIMPLE tree to correspoinding GIMPLE++ object.
 * 
 * @param gimple pointer to @c Gimple class
 * @param type type of the GIMPLE++ object to create (@c git_type, @c git_decl, ...).
 * @param node tree to transform
 */
#define TREE_TO_GIT(gimple, gim_help, type, node)                         \
do {                                                            \
const_ ## type c_type_ = (const_ ## type) get_ ## type ## _from_tree_cache (gim_help, node);    \
if (c_type_)                                                      \
    return c_type_;                                               \
else                                                            \
{                                                               \
    type type_ = create_ ## type ## _from_tree (gimple, gim_help, node);       \
    add_ ## type ## _to_tree_cache (gimple, gim_help, node, type_);       \
    return type_;                                               \
}                                                               \
} while (0)
                 
                 

static const_git_type tree_to_git_type (git_gimple gimple, git_gimple_helpers gim_help, tree node);
static const_git_constant tree_to_git_constant (git_gimple gimple, git_gimple_helpers gim_help, tree node);
static const_git_decl tree_to_git_decl (git_gimple gimple, git_gimple_helpers gim_help, tree node);


static git_type create_git_type_from_tree (git_gimple gimple, git_gimple_helpers gim_help, tree node);
static git_decl create_git_decl_from_tree (git_gimple gimple, git_gimple_helpers gim_help, tree node);
static git_constant create_git_constant_from_tree (git_gimple gimple, git_gimple_helpers gim_help, tree node);

/**
 * Checks if @c decl is a declaration of main function.
 * 
 * @param decl tree representing declaration of function
 */
static bool
is_main_function (tree decl)
{
    return DECL_NAME (decl) != NULL_TREE
            && MAIN_NAME_P (DECL_NAME (decl));
}


/**
 * Creates GIMPLE++ type object from tree.
 * 
 * @param gimple pointer to @c Gimple object
 * @param node tree representing a GIMPLE type
 * @return GIMPLE++ type object. The returned object is NOT cached.
 */
static git_type
create_git_type_from_tree (git_gimple gimple, git_gimple_helpers gim_help, tree node)
{
	const char* type_name;

	if (TREE_CODE (node) == RECORD_TYPE || TREE_CODE (node) == UNION_TYPE || TREE_CODE (node) == ENUMERAL_TYPE)
		type_name = TYPE_NAME (node) ? IDENTIFIER_POINTER (TYPE_NAME (node)) : "<unknown_type_name>";
	else
		type_name = TYPE_NAME (node) ? IDENTIFIER_POINTER (DECL_NAME (TYPE_NAME (node))) : "<unknown_type_name>";
    
    node = TYPE_MAIN_VARIANT (node);
    
    HOST_WIDE_INT bit_size;
    switch (TREE_CODE (node))
    {
        case INTEGER_TYPE:
            bit_size = double_int_to_shwi (TREE_INT_CST (TYPE_SIZE (node)));
            return createIntegerType (type_name, bit_size,
                                      TYPE_UNSIGNED (node), TYPE_PRECISION (node));

        case REAL_TYPE:
        	bit_size = double_int_to_shwi (TREE_INT_CST (TYPE_SIZE (node)));
        	return createRealType(type_name, bit_size, TYPE_PRECISION (node));

        case POINTER_TYPE:
            bit_size = double_int_to_shwi (TREE_INT_CST (TYPE_SIZE (node)));
            return createPtrType (type_name, bit_size, 
                                  tree_to_git_type (gimple, gim_help, TREE_TYPE (node)));

        case VOID_TYPE:
            gcc_assert (node == void_type_node);
            return createVoidType ();
            
        case BOOLEAN_TYPE:
            bit_size = double_int_to_shwi (TREE_INT_CST (TYPE_SIZE (node)));
            return createBoolType (type_name, bit_size);
        
        case FUNCTION_TYPE:
        {
            tree ret_type_tree = TREE_TYPE (node);
            const_git_type ret_type = ret_type_tree ? tree_to_git_type (gimple, gim_help, ret_type_tree) :
                    tree_to_git_type (gimple, gim_help, void_type_node);
            
            git_type func_type = createFunctionType (type_name, ret_type);
            
            tree arg_type;
            for (arg_type = TYPE_ARG_TYPES (node); arg_type; arg_type = TREE_CHAIN (arg_type)) 
                functionTypeAddParameter (func_type, tree_to_git_type (gimple, gim_help, TREE_VALUE (arg_type)));
            return func_type;
        }
        
        case ARRAY_TYPE:
        {
            bit_size = double_int_to_shwi (TREE_INT_CST (TYPE_SIZE (node)));
            const_git_type element_type = tree_to_git_type (gimple, gim_help, TREE_TYPE (node));
            return createArrayType (type_name, bit_size, element_type);
        }

        case ENUMERAL_TYPE:
        {
        	bit_size = double_int_to_shwi (TREE_INT_CST (TYPE_SIZE (node)));

        	git_type enum_type = createEnumType(type_name, bit_size,
                    TYPE_UNSIGNED (node), TYPE_PRECISION (node));

        	tree value;
        	for (value = TYPE_VALUES(node); value != NULL; value = TREE_CHAIN(value))
        	{
				const char* e_name = IDENTIFIER_POINTER (TREE_PURPOSE(value));

				const_git_constant e_value_raw = tree_to_git_constant(gimple, gim_help, (tree) TREE_VALUE(value));
				uint64_t e_value = integerConstantGetValue(e_value_raw);

				enumTypeAddValue(enum_type, e_name, e_value);
        	}

        	return enum_type;
        }

        case RECORD_TYPE:
        case UNION_TYPE:
        {
        	git_type record_type = getDefinedType(gim_help, type_name);
			if (record_type != 0) // handling recursive definitions;
				return record_type;

        	bit_size = double_int_to_shwi (TREE_INT_CST (TYPE_SIZE (node)));
    		record_type = createRecordType(type_name, bit_size);
    		addDefinedType(gim_help, type_name, record_type);

        	tree field;
        	for (field = TYPE_FIELDS(node); field != NULL; field = TREE_CHAIN(field))
        	{

        		const_git_decl decl = tree_to_git_decl(gimple, gim_help, (tree) field);

        		// FIELD_DECLs - fields of struct
        		// VAR_DECLs, TYPE_DECLs and CONST_DECL
        		// record-scope variables, types and enumerators.
        		if (TREE_CODE(field) == FIELD_DECL)
        			recordTypeAddField(record_type, decl);
        		else
        			recordTypeAddLocalDecl(record_type, decl);
        	}

        	return record_type;
        }

        default:
            ni_error ("get type for `%s'", tree_code_name [TREE_CODE (node)]);
    }
}


/**
 * Returns GIMPLE++ object representing the type corresponding to the type represented
 * by supplied tree.
 * 
 * @param gimple pointer to @c Gimple object
 * @param node tree representing a GIMPLE type
 * @return GIMPLE++ type object. The returned object is cached.
 */
static const_git_type
tree_to_git_type (git_gimple gimple, git_gimple_helpers gim_help, tree node)
{
    gcc_assert (node);
    gcc_assert (TYPE_P (node));

    TREE_TO_GIT (gimple, gim_help, git_type, node);
}


/**
 * Creates GIMPLE++ declaration object from tree.
 * 
 * @param gimple pointer to @c Gimple object
 * @param node tree representing a GIMPLE declaration
 * @return GIMPLE++ declaration object. The returned object is NOT cached.
 */
static git_decl
create_git_decl_from_tree (git_gimple gimple, git_gimple_helpers gim_help, tree node)
{
    git_decl ret = 0;

    tree name_tree = DECL_NAME (node);
    const char* name = name_tree ? IDENTIFIER_POINTER (name_tree) : 0;

    unsigned int uid = DECL_UID (node);
    
    tree type_tree = TREE_TYPE (node);
	gcc_assert (type_tree);

	const_git_type type = tree_to_git_type (gimple, gim_help, type_tree);

	// if the declaration is global
    bool local = !(TREE_STATIC (node) || DECL_EXTERNAL (node));

    switch (TREE_CODE (node)) 
    {
        case VAR_DECL:
        	// These nodes represent variables with namespace or block scope,
        	// as well as static data members. The DECL_SIZE and DECL_ALIGN are analogous
        	// to TYPE_SIZE and TYPE_ALIGN. For a declaration, you should always use
        	// the DECL_SIZE and DECL_ALIGN rather than the TYPE_SIZE and TYPE_ALIGN
        	// given by the TREE_TYPE, since special attributes may have been applied
        	// to the variable to give it a particular size and alignment. You may use
        	// the predicates DECL_THIS_STATIC or DECL_THIS_EXTERN to test whether
        	// the storage class specifiers static or extern were used to declare a variable.
        	// If this variable is initialized (but does not require a constructor),
        	// the DECL_INITIAL will be an expression for the initializer. The initializer
        	// should be evaluated, and a bitwise copy into the variable performed.
        	// If the DECL_INITIAL is the error_mark_node, there is an initializer,
        	//but it is given by an explicit statement later in the code; no bitwise copy is required.
        	// GCC provides an extension that allows either automatic variables, or global variables,
        	// to be placed in particular registers. This extension is being used for a particular
        	// VAR_DECL if DECL_REGISTER holds for the VAR_DECL, and if DECL_ASSEMBLER_NAME
        	// is not equal to DECL_NAME. In that case, DECL_ASSEMBLER_NAME is the name of
        	//the register into which the variable will be placed.
            ret = createVarDecl (uid, name, type, local);
            break;
        case PARM_DECL:
        	// Used to represent a parameter to a function. Treat these nodes similarly
        	// to VAR_DECL nodes. These nodes only appear in the DECL_ARGUMENTS for a FUNCTION_DECL.
        	// The DECL_ARG_TYPE for a PARM_DECL is the type that will actually be used
        	// when a value is passed to this function. It may be a wider type than
        	// the TREE_TYPE of the parameter; for example, the ordinary type might be short while
        	// the DECL_ARG_TYPE is int.
            ret = createParamDecl (uid, name, type);
            break;
        case FUNCTION_DECL:
        	// DECL_NAME - This macro returns the unqualified name of the function, as an IDENTIFIER_NODE.
        	// For an instantiation of a function template, the DECL_NAME is the unqualified name of the template,
        	// not something like f<int>. The value of DECL_NAME is undefined when used on a constructor,
        	// destructor, overloaded operator, or type-conversion operator, or any function that is implicitly
        	// generated by the compiler. See below for macros that can be used to distinguish these cases.
        	// DECL_ASSEMBLER_NAME - This macro returns the mangled name of the function, also an IDENTIFIER_NODE.
        	// This name does not contain leading underscores on systems that prefix all identifiers with underscores.
        	// The mangled name is computed in the same way on all platforms; if special processing is required
        	// to deal with the object file format used on a particular platform, it is the responsibility
        	// of the back end to perform those modifications. (Of course, the back end should not modify
        	// DECL_ASSEMBLER_NAME itself.)
        	// Using DECL_ASSEMBLER_NAME will cause additional memory to be allocated (for the mangled name of the entity)
        	// so it should be used only when emitting assembly code. It should not be used within the optimizers
        	// to determine whether or not two declarations are the same, even though some of the existing optimizers
        	// do use it in that way. These uses will be removed over time.
        	// DECL_ARGUMENTS - This macro returns the PARM_DECL for the first argument to the function.
        	// Subsequent PARM_DECL nodes can be obtained by following the TREE_CHAIN links.
        	// DECL_RESULT - This macro returns the RESULT_DECL for the function.
        	// DECL_SAVED_TREE - This macro returns the complete body of the function.
        	// TREE_TYPE - This macro returns the FUNCTION_TYPE or METHOD_TYPE for the function.
        	// DECL_INITIAL - A function that has a definition in the current translation unit
        	// will have a non-NULL DECL_INITIAL. However, back ends should not make use of
        	// the particular value given by DECL_INITIAL.
        	// It should contain a tree of BLOCK nodes that mirrors the scopes that variables are bound
        	// in the function. Each block contains a list of decls declared in a basic block,
        	// a pointer to a chain of blocks at the next lower scope level, then a pointer to
        	// the next block at the same level and a backpointer to the parent BLOCK or FUNCTION_DECL.
            ret = createFunctionDecl (uid, name, type, DECL_BUILT_IN (node),
                                       (enum builtin_function_code) DECL_FUNCTION_CODE (node));
            break;
        case LABEL_DECL:
			// These nodes are used to represent labels in function bodies.
        	ret = createLabelDecl (uid, name, type);
        	break;
        case RESULT_DECL:
			// These nodes represent the value returned by a function.
			// When a value is assigned to a RESULT_DECL, that indicates that the value should be returned,
			//  via bitwise copy, by the function.
			// You can use DECL_SIZE and DECL_ALIGN on a RESULT_DECL, just as with a VAR_DECL.
        	ret = createResultDecl (uid, name, type);
        case TYPE_DECL:
			// These nodes represent typedef declarations.
			// The TREE_TYPE is the type declared to have the name given by DECL_NAME.
			// In some cases, there is no associated name.
            ret = createTypeDecl (uid, name, type, local);
            break;
		case FIELD_DECL:
			//These nodes represent non-static data members.
			//The DECL_SIZE and DECL_ALIGN behave as for VAR_DECL nodes.
			//The position of the field within the parent record is specified
			//by a combination of three attributes. DECL_FIELD_OFFSET is the position,
			// counting in bytes, of the DECL_OFFSET_ALIGN-bit sized word containing
			// the bit of the field closest to the beginning of the structure.
			//DECL_FIELD_BIT_OFFSET is the bit offset of the first bit of the field within this word;
			//this may be nonzero even for fields that are not bit-fields,
			//since DECL_OFFSET_ALIGN may be greater than the natural alignment of the field's type.
			//If DECL_C_BIT_FIELD holds, this field is a bit-field.
			//In a bit-field, DECL_BIT_FIELD_TYPE also contains the type that was originally specified for it,
			//while DECL_TYPE may be a modified type with lesser precision, according to the size of the bit field.
			ret = createFieldDecl (uid, name, type);
			break;
		case CONST_DECL:
			// These nodes are used to represent enumeration constants.
			// The value of the constant is given by DECL_INITIAL which will be an INTEGER_CST
			// with the same type as the TREE_TYPE of the CONST_DECL, i.e., an ENUMERAL_TYPE.
			ret = createConstDecl(uid, name, type, local);
			break;
        default:
            ni_error ("declaration `%s'", tree_code_name[TREE_CODE (node)]);
    }

    reportDecl (gimple, ret);
    return ret;
}


/**
 * Returns GIMPLE++ declaration object for specified tree.
 * 
 * @param gimple pointer to @c Gimple object
 * @param node tree representing a GIMPLE declaration
 * @return GIMPLE++ declaration object. The returned object is cached.
 */
static const_git_decl
tree_to_git_decl (git_gimple gimple, git_gimple_helpers gim_help, tree node)
{
    gcc_assert (node);
    gcc_assert (DECL_P (node));

    TREE_TO_GIT (gimple, gim_help, git_decl, node);
}


/**
 * Creates GIMPLE++ constant object from tree.
 * 
 * @param gimple pointer to @c Gimple object
 * @param node tree representing a GIMPLE constant
 * @return GIMPLE++ constant object. The returned object is NOT cached.
 */
static git_constant
create_git_constant_from_tree (git_gimple gimple, git_gimple_helpers gim_help, tree node)
{
    const_git_type type = tree_to_git_type (gimple, gim_help, TREE_TYPE (node));
    git_constant ret = 0;

    switch (TREE_CODE (node))
    {
        case INTEGER_CST:
        {
            unsigned HOST_WIDE_INT low = TREE_INT_CST_LOW (node);
            HOST_WIDE_INT high = TREE_INT_CST_HIGH (node);
            uint64_t value = ((uint64_t)low) | (((uint64_t)high) << (sizeof(unsigned HOST_WIDE_INT)*8));
            
            ret = createIntCst (type, value);
            break;
        }
        case STRING_CST:
        {
            int str_len = TREE_STRING_LENGTH (node);
            const char* str = TREE_STRING_POINTER (node);
            ret = createStringCst (type, str_len, str);
            break;
        }
        case REAL_CST:
        {
        	REAL_VALUE_TYPE value = TREE_REAL_CST (node);

        	// TODO: size
        	char buf[128];
        	real_to_decimal_for_mode(&buf, &value, sizeof(buf), 0, 1, TYPE_MODE(TREE_TYPE(node)));
        	ret = createRealCst (type, (long double) atof((const char*) &buf));
        	break;
        }
        default:
            ni_error ("get constant for `%s'", tree_code_name[TREE_CODE (node)]);
    }
    
    reportConstant (gimple, ret);
    return ret;
}


/**
 * Returns GIMPLE++ constant object for specified tree.
 * 
 * @param gimple pointer to @c Gimple object
 * @param node tree representing a GIMPLE constant
 * @return GIMPLE++ constant object. The returned object is cached.
 */
static const_git_constant
tree_to_git_constant (git_gimple gimple, git_gimple_helpers gim_help, tree node)
{
    TREE_TO_GIT (gimple, gim_help, git_constant, node);
}


/**
 * Creates GIMPLE++ operand object from tree.
 * 
 * @param gimple pointer to @c Gimple object
 * @param node tree representing a GIMPLE expression
 * @return GIMPLE++ operand object. The returned object is NOT cached.
 */
static git_operand
create_git_operand_from_tree (git_gimple gimple, git_gimple_helpers gim_help, tree node)
{
    if (!node)
        return 0;

    tree typeTree = TREE_TYPE (node);
    const_git_type type = tree_to_git_type (gimple, gim_help, typeTree);

    // We need to extract the node from SSA_NAME node.
	if (TREE_CODE (node) == SSA_NAME)
    {
    	node = SSA_NAME_VAR(node);
    }

	if (TREE_CODE (node) == VIEW_CONVERT_EXPR)
	{
    	node = TREE_OPERAND (node, 0);
	}

    if (DECL_P (node))  /* node is declaration */
    {
        const_git_decl decl = tree_to_git_decl (gimple, gim_help, node);
        return createDeclOperand (decl);
    }
    
    if (CONSTANT_CLASS_P (node))
    {
        return createCstOperand (tree_to_git_constant (gimple, gim_help, node));
    }

    switch (TREE_CODE (node))
    {
        case ADDR_EXPR:
        {
            tree sub_node = TREE_OPERAND (node, 0);
            return createAddrOperand (type, create_git_operand_from_tree (gimple, gim_help, sub_node));
        }
        case CASE_LABEL_EXPR:
		{
			const_git_decl decl = tree_to_git_decl (gimple, gim_help, CASE_LABEL(node));

			if (CASE_LOW(node) == NULL_TREE) // default case label
			{
				return createLabelExprOperand (decl, 0, 0, true, false);
			}
			else if (CASE_HIGH(node) == NULL_TREE) // non-default case label
			{
				git_operand value =
						create_git_operand_from_tree(gimple, gim_help, (tree) CASE_LOW(node));
				return createLabelExprOperand (decl, value, 0, false, false);
			}
			else // non-default range case label
			{
				git_operand value_lower =
						create_git_operand_from_tree(gimple, gim_help, (tree) CASE_LOW(node));
				git_operand value_upper =
						create_git_operand_from_tree(gimple, gim_help, (tree) CASE_HIGH(node));
				return createLabelExprOperand (decl, value_lower,
						value_upper,
						false,
						true);
			}
		}
        case INDIRECT_REF:
        {
            tree sub_node = TREE_OPERAND (node, 0);
            return createIndirectRefOperand (type, create_git_operand_from_tree (gimple, gim_help, sub_node));
        }

        case ARRAY_REF:
        {
            git_operand array = create_git_operand_from_tree (gimple, gim_help, TREE_OPERAND (node, 0));
            git_operand index = create_git_operand_from_tree (gimple, gim_help, TREE_OPERAND (node, 1));
            return createArrayRefOperand (array, index);
        }

        case COMPONENT_REF:
        {
        	git_operand record = create_git_operand_from_tree (gimple, gim_help, TREE_OPERAND (node, 0));
        	git_operand field = create_git_operand_from_tree (gimple, gim_help, TREE_OPERAND (node, 1));
        	return createComponentRefOperand (record, field);
        }

        case CONSTRUCTOR:
        {
        	git_operand constructorOperand;

        	unsigned HOST_WIDE_INT idx;
        	tree purpose, value;

        	if (TREE_CODE(TREE_TYPE(node)) == RECORD_TYPE || TREE_CODE(TREE_TYPE(node)) == UNION_TYPE)
        	{
        		constructorOperand = createRecordConstructorOperand(type);

        		FOR_EACH_CONSTRUCTOR_ELT (CONSTRUCTOR_ELTS (node), idx, purpose, value)
        		{
        			git_operand decl = create_git_operand_from_tree(gimple, gim_help, (tree) purpose);
					git_operand operand = create_git_operand_from_tree(gimple, gim_help, (tree) value);
        			recordConstructorOperandAddField(constructorOperand, decl, operand);
        		}

        		return constructorOperand;
        	}
        	else if (TREE_CODE(TREE_TYPE(node)) == ARRAY_TYPE)
        	{
        		constructorOperand = createArrayConstructorOperand(type);

        		FOR_EACH_CONSTRUCTOR_ELT (CONSTRUCTOR_ELTS (node), idx, purpose, value)
				{
					if (TREE_CODE(purpose) == INTEGER_CST)
					{
						git_operand cst = create_git_operand_from_tree(gimple, gim_help, (tree) purpose);
						git_operand operand = create_git_operand_from_tree(gimple, gim_help, (tree) value);

						arrayConstructorOperandAddField(constructorOperand, cst, operand);
					}
					else if (TREE_CODE(purpose) == RANGE_EXPR)
					{
						git_operand lower = create_git_operand_from_tree(gimple, gim_help, (tree) TREE_OPERAND (purpose, 0));
						git_operand upper = create_git_operand_from_tree(gimple, gim_help, (tree) TREE_OPERAND (purpose, 1));
						git_operand operand = create_git_operand_from_tree(gimple, gim_help, (tree) value);

						arrayConstructorOperandAddRangeField(constructorOperand, lower, upper, operand);
					}
					else if (purpose == NULL_TREE)
					{
						//TODO: If the TREE_PURPOSE is NULL_TREE, then the initializer
						//is for the next available array element.
						ni_error ("wierd constructor");
						//git_operand operand = create_git_operand_from_tree(gimple, gim_help, (tree) value);
					}
				}
        	}
        	else
        		gcc_assert(false);

        	return constructorOperand;
        }
        default:
            ni_error ("get operand for `%s'", tree_code_name[TREE_CODE (node)]);
    }
}


/**
 * Creates GIMPLE++ operation object.
 * 
 * @param gimple pointer to @c Gimple object
 * @param rhs_code the code of the operation
 * @param tree_op1 tree representing expression used as the first operand
 * @param tree_op2 tree representing expression used as the second operand
 * @param lhs_type GIMPLE++ type object representing return type
 * @return GIMPLE++ operation object
 */
static git_operation
create_operation (git_gimple gimple, git_gimple_helpers gim_help, enum tree_code rhs_code,
                  tree tree_op1, tree tree_op2, const_git_type lhs_type)
{
    git_operand op1 = create_git_operand_from_tree (gimple, gim_help, tree_op1);
    git_operand op2 = create_git_operand_from_tree (gimple, gim_help, tree_op2);
    
    if (get_gimple_rhs_class (rhs_code) == GIMPLE_SINGLE_RHS)
        return createNoOperation (lhs_type, op1);
    
    switch (rhs_code)
    {
        case NOP_EXPR:
            return createNoOperation (lhs_type, op1);
        case CONVERT_EXPR:
            return createConvertOperation (lhs_type, op1);
            
        case PLUS_EXPR:
            return createPlusOperation (lhs_type, op1, op2);
        case MINUS_EXPR:
            return createMinusOperation (lhs_type, op1, op2);
        case TRUNC_DIV_EXPR:
            return createDivOperation (lhs_type, op1, op2);
        case TRUNC_MOD_EXPR:
            return createModOperation (lhs_type, op1, op2);
        case MULT_EXPR:
            return createMultOperation (lhs_type, op1, op2);
        
        case POINTER_PLUS_EXPR:
            return createPlusOperation (lhs_type, op1, op2);
        
        case LT_EXPR:
            return createLtOperation (lhs_type, op1, op2);
        case LE_EXPR:
            return createLeOperation (lhs_type, op1, op2);
        case GT_EXPR:
            return createGtOperation (lhs_type, op1, op2);
        case GE_EXPR:
            return createGeOperation (lhs_type, op1, op2);
        case EQ_EXPR:
            return createEqOperation (lhs_type, op1, op2);
        case NE_EXPR:
            return createNeOperation (lhs_type, op1, op2);

        case MIN_EXPR:
        	return createMinOperation (lhs_type, op1, op2);
        case MAX_EXPR:
        	return createMaxOperation (lhs_type, op1, op2);
        case ABS_EXPR:
        	return createAbsOperation (lhs_type, op1);
        
        default:
            ni_error ("operation `%s'", tree_code_name[rhs_code]);
    }
}


/**
 * Creates GIMPLE++ statement from GIMPLE_ASSIGN statement.
 * 
 * @param gim pointer to @c Gimple object
 * @param stmt GIMPLE statement
 * @return GIMPLE++ statement
 */
static git_stmt
gimple_assign_to_git_stmt (git_gimple gim, git_gimple_helpers gim_help, gimple stmt)
{
    tree lhs_tree = gimple_assign_lhs (stmt);
    const_git_type lhs_type = 0;
    if (lhs_tree)
        lhs_type = tree_to_git_type (gim, gim_help, TREE_TYPE (lhs_tree));
    
    git_operation operation = create_operation (gim, gim_help, gimple_assign_rhs_code (stmt),
            gimple_assign_rhs1 (stmt), gimple_assign_rhs2 (stmt), lhs_type);
    git_operand lhs = create_git_operand_from_tree (gim, gim_help, lhs_tree);
    
    return createAssignStmt (lhs, operation);
}


/**
 * Creates GIMPLE++ statement from GIMPLE_COND statement.
 * 
 * @param gim pointer to @c Gimple object
 * @param stmt GIMPLE statement
 * @return GIMPLE++ statement
 */
static git_stmt
gimple_cond_to_git_stmt (git_gimple gim, git_gimple_helpers gim_help, gimple stmt)
{
    git_operation operation = create_operation (gim, gim_help, gimple_cond_code (stmt),
            gimple_cond_lhs (stmt), gimple_cond_rhs (stmt), 0);
    
    return createCondStmt (operation);
}


/**
 * Creates GIMPLE++ statement from GIMPLE_CALL statement.
 * 
 * @param gim pointer to @c Gimple object
 * @param stmt GIMPLE statement
 * @return GIMPLE++ statement
 */
static git_stmt
gimple_call_to_git_stmt (git_gimple gim, git_gimple_helpers gim_help, gimple stmt)
{
    git_operand lhs = create_git_operand_from_tree (gim, gim_help, gimple_call_lhs (stmt));
    git_operand fn = create_git_operand_from_tree (gim, gim_help, gimple_call_fn (stmt));
    
    git_stmt git_stmt = createCallStmt (fn, lhs);
    
    unsigned i;
    for (i = 0; i < gimple_call_num_args (stmt); i++)
    {
        git_operand param = create_git_operand_from_tree (gim, gim_help, gimple_call_arg (stmt, i));
        callStmtAddArg (git_stmt, param);
    }
    
    return git_stmt;
}

static git_stmt
gimple_goto_to_git_stmt (git_gimple gim, git_gimple_helpers gim_help, gimple stmt)
{
	git_operand dest = create_git_operand_from_tree (gim, gim_help, gimple_goto_dest(stmt));
	git_stmt git_stmt = createGotoStmt (dest);

	return git_stmt;
}

static git_stmt
gimple_label_to_git_stmt (git_gimple gim, git_gimple_helpers gim_help, gimple stmt)
{
	git_operand label = create_git_operand_from_tree(gim, gim_help, gimple_label_label(stmt));
	git_stmt git_stmt = createLabelStmt (label);

	return git_stmt;
}

static git_stmt
gimple_nop_to_git_stmt ()
{
	git_stmt git_stmt = createNopStmt ();

	return git_stmt;
}

static git_stmt
gimple_phi_to_git_stmt (git_gimple gim, git_gimple_helpers gim_help, gimple stmt)
{
	git_operand result = create_git_operand_from_tree(gim, gim_help, gimple_phi_result(stmt));

	git_stmt git_stmt = createPhiStmt (result);

	unsigned int i;
	for (i = 0; i < gimple_phi_num_args (stmt); ++i)
	{
		git_operand arg = create_git_operand_from_tree(gim, gim_help, gimple_phi_arg_def(stmt, i));
		phiStmtAddArg(git_stmt, arg);
	}

	return git_stmt;
}

static git_stmt
gimple_switch_to_git_stmt (git_gimple gim, git_gimple_helpers gim_help, gimple stmt)
{
	git_operand index = create_git_operand_from_tree(gim, gim_help, gimple_switch_index(stmt));

	git_stmt git_stmt = createSwitchStmt (index);

	unsigned int i;
	for (i = 0; i < gimple_switch_num_labels (stmt); i++)
	{
		tree label_raw = gimple_switch_label(stmt, i);
		gcc_assert(label_raw != NULL_TREE);
		git_operand label = create_git_operand_from_tree(gim, gim_help, label_raw);
		switchStmtAddLabel(git_stmt, label);
	}

	return git_stmt;
}

/**
 * Creates GIMPLE++ statement from GIMPLE_RETURN statement.
 * 
 * @param gim pointer to @c Gimple object
 * @param stmt GIMPLE statement
 * @return GIMPLE++ statement
 */
static git_stmt
gimple_return_to_git_stmt (git_gimple gim, git_gimple_helpers gim_help, gimple stmt)
{
    tree retval_tree = gimple_return_retval (stmt);
    git_operand retval = create_git_operand_from_tree (gim, gim_help, retval_tree);
    return createReturnStmt (retval);
}


/**
 * Creates GIMPLE++ statement from GIMPLE statement.
 * 
 * @param gim pointer to @c Gimple object
 * @param stmt GIMPLE statement
 * @return GIMPLE++ statement created from @c stmt
 */
static git_stmt
gimple_to_git_stmt (git_gimple gim, git_gimple_helpers gim_help, gimple stmt)
{
    gcc_assert (stmt);
    enum gimple_code code = gimple_code (stmt);

    LOG_DEBUG ("creating statement `%s' in %s at %zu", gimple_code_name[code],
             LOCATION_FILE (gimple_location (stmt)), LOCATION_LINE (gimple_location (stmt)));
    
    switch (code)
    {
        case GIMPLE_ASSIGN:
        	// GIMPLE_ASSIGN <SUBCODE, LHS, RHS1[, RHS2]> represents the assignment
        	//   statement
        	//   LHS = RHS1 SUBCODE RHS2.
        	//   SUBCODE is the tree code for the expression computed by the RHS of the
        	//   assignment.  It must be one of the tree codes accepted by
        	//   get_gimple_rhs_class.  If LHS is not a gimple register according to
        	//   is_gimple_reg, SUBCODE must be of class GIMPLE_SINGLE_RHS.
        	//   LHS is the operand on the LHS of the assignment.  It must be a tree node
        	//   accepted by is_gimple_lvalue.
        	//   RHS1 is the first operand on the RHS of the assignment.  It must always be
        	//   present.  It must be a tree node accepted by is_gimple_val.
        	//   RHS2 is the second operand on the RHS of the assignment.  It must be a tree
        	//   node accepted by is_gimple_val.  This argument exists only if SUBCODE is
        	//   of class GIMPLE_BINARY_RHS.
            return gimple_assign_to_git_stmt (gim, gim_help, stmt);

        case GIMPLE_COND:
        	// GIMPLE_COND <COND_CODE, OP1, OP2, TRUE_LABEL, FALSE_LABEL>
        	//   represents the conditional jump:
        	//   if (OP1 COND_CODE OP2) goto TRUE_LABEL else goto FALSE_LABEL
        	//   COND_CODE is the tree code used as the comparison predicate.  It
        	//   must be of class tcc_comparison.
        	//   OP1 and OP2 are the operands used in the comparison.  They must be
        	//   accepted by is_gimple_operand.
        	//   TRUE_LABEL and FALSE_LABEL are the LABEL_DECL nodes used as the
        	//   jump target for the comparison.
            return gimple_cond_to_git_stmt (gim, gim_help, stmt);
           
        case GIMPLE_CALL:
        	// GIMPLE_CALL <FN, LHS, ARG1, ..., ARGN[, CHAIN]> represents function
        	//   calls.
        	//   FN is the callee.  It must be accepted by is_gimple_call_addr.
        	//   LHS is the operand where the return value from FN is stored.  It may
        	//   be NULL.
        	//   ARG1 ... ARGN are the arguments.  They must all be accepted by
        	//   is_gimple_operand.
        	//    CHAIN is the optional static chain link for nested functions.
            return gimple_call_to_git_stmt (gim, gim_help, stmt);
        case GIMPLE_RETURN:
        	// GIMPLE_RETURN <RETVAL> represents return statements.
        	//   RETVAL is the value to return or NULL.  If a value is returned it
        	//   must be accepted by is_gimple_operand.
            return gimple_return_to_git_stmt (gim, gim_help, stmt);
        case GIMPLE_GOTO:
        	// GIMPLE_GOTO <TARGET> represents unconditional jumps.
        	//   TARGET is a LABEL_DECL or an expression node for computed GOTOs.
        	return gimple_goto_to_git_stmt (gim, gim_help, stmt);
        case GIMPLE_LABEL:
        	// GIMPLE_LABEL <LABEL> represents label statements.  LABEL is a
			//   LABEL_DECL representing a jump target.
        	return gimple_label_to_git_stmt (gim, gim_help, stmt);
        case GIMPLE_NOP:
        	// GIMPLE_NOP represents the "do nothing" statement.
        	return gimple_nop_to_git_stmt ();
        case GIMPLE_PHI:
        	// GIMPLE_PHI <RESULT, ARG1, ..., ARGN> represents the PHI node
        	//   RESULT = PHI <ARG1, ..., ARGN>
        	//   RESULT is the SSA name created by this PHI node.
        	//   ARG1 ... ARGN are the arguments to the PHI node.  N must be
        	//   exactly the same as the number of incoming edges to the basic block
        	//   holding the PHI node.  Every argument is either an SSA name or a
        	//   tree node of class tcc_constant.
        	return gimple_phi_to_git_stmt (gim, gim_help, stmt);
        case GIMPLE_SWITCH:
        	// GIMPLE_SWITCH <INDEX, DEFAULT_LAB, LAB1, ..., LABN> represents the
        	//   multiway branch:
        	//   switch (INDEX)
        	//   {
        	//     case LAB1: ...; break;
        	//     ...
        	//     case LABN: ...; break;
        	//     default: ...
        	//   }
        	//   INDEX is the variable evaluated to decide which label to jump to.
        	//   DEFAULT_LAB, LAB1 ... LABN are the tree nodes representing case labels.
        	//   They must be CASE_LABEL_EXPR nodes.
        	return gimple_switch_to_git_stmt (gim, gim_help, stmt);
        case GIMPLE_PREDICT:
        	// Ignored
        	return gimple_nop_to_git_stmt (gim, gim_help, stmt);
        default:
            ni_error ("create gimple statement with code `%s'", gimple_code_name[code]);
    }
}


/**
 * Creates GIMPLE++ block from GIMPLE basic block.
 * 
 * @param gim pointer to @c Gimple object
 * @param bb GIMPLE basic block
 * @return GIMPLE++ block created from @c bb
 */
static git_block
basic_block_to_git_block (git_gimple gim, git_gimple_helpers gim_help, basic_block bb)
{
    git_block block = createBlock ();
    gimple_stmt_iterator gsi;
    for (gsi = gsi_start_bb (bb); !gsi_end_p (gsi); gsi_next (&gsi))
    {
        git_stmt stmt = gimple_to_git_stmt (gim, gim_help, gsi_stmt (gsi));
        blockAddStmt (block, stmt);
    }
    return block;
}


/**
 * Adds GIMPLE edge into GIMPLE++ representation.
 * 
 * @param func function which blocks the edge connects
 * @param e GIMPLE edge
 */
static void
add_git_edge (git_function_helpers func_help, edge e)
{
    unsigned fallthru       = (e->flags & EDGE_FALLTHRU) != 0;
    unsigned true_branch    = (e->flags & EDGE_TRUE_VALUE) != 0;
    unsigned false_branch   = (e->flags & EDGE_FALSE_VALUE) != 0;

    if ((e->flags & (~(EDGE_TRUE_VALUE | EDGE_FALSE_VALUE | EDGE_FALLTHRU))) != 0)
        ni_error ("edge flag, flags (%u)", e->flags);
    
    git_block src = functionGetBlock (func_help, e->src);
    git_block dest = functionGetBlock (func_help, e->dest);
    
    git_edge edge = createEdge (src, dest, fallthru, true_branch, false_branch);
    
    blockAddSuccEdge (src, edge);
    blockAddPredEdge (dest, edge);
}

void
gb_resolve_globals(git_gimple gim, git_gimple_helpers gim_help)
{
	if (getGlobalsResolved (gim_help))
		return;
	
	struct varpool_node* vn;

	for (vn = gb_varpool_nodes (); vn; vn = gb_varpool_next(vn))
	{
		gb_varpool_decl (gim, gim_help, vn);
	}

	setGlobalsResolved (gim_help);
}

void
gb_end_construction () {
    endConstruction();
}

git_gimple
gb_get_construction_result () {
    return getConstructionResult();
}


/**
 * Transforms currently compiled GIMPLE function into GIMPLE++ representation and adds it
 * into supplied gimple object.
 * 
 * @param gim pointer to @c Gimple object
 */

void
gb_process_cfun ()
{
	//fprintf(stderr, "ENTRY: gb_process_cfun\n");

	git_gimple_construction construction = beginFunctionConstruction();
	git_gimple gim = construction->gimple;
	git_gimple_helpers gim_help = construction->gimple_helpers;
	git_function_helpers func_help = construction->function_helpers;

	//fprintf(stderr, "1 %p, %p, %p, %p\n", construction, gim, gim_help, func_help);

	gb_resolve_globals (gim, gim_help);

    struct cgraph_node *node;
    tree fn_decl_tree;

    for (node = cgraph_nodes; node; node = node->next)
    {
        fn_decl_tree = node->decl;
        const_git_decl func_decl = tree_to_git_decl (gim, gim_help, fn_decl_tree);
        const_git_type fn_type = tree_to_git_type (gim, gim_help, TREE_TYPE (fn_decl_tree));
        struct function* func = DECL_STRUCT_FUNCTION (fn_decl_tree);

		if (!func)
		{
			continue;
		}

		/* TODO - check for duplicate function defs */
		const char* name = IDENTIFIER_POINTER (DECL_ASSEMBLER_NAME (fn_decl_tree));

		if (checkFuncNameExists(gim_help, name) == 1)
			continue;
		else
		{
			addFuncName(gim_help, name);
		}

		git_function git_func = createFunction (fn_type);

		tree arg_decl_tree;
		for (arg_decl_tree = DECL_ARGUMENTS (fn_decl_tree); arg_decl_tree;
			 arg_decl_tree = TREE_CHAIN (arg_decl_tree))
		{
			tree unit_size_tree = TYPE_SIZE_UNIT (TREE_TYPE (arg_decl_tree));
			gcc_assert (unit_size_tree);

			const_git_decl arg_decl = tree_to_git_decl (gim, gim_help, arg_decl_tree);
			functionAddParamDecl (git_func, arg_decl);
		}


		// Create declarations
		tree decl_list;
		for (decl_list = func->local_decls; decl_list; decl_list = TREE_CHAIN (decl_list))
		{
			tree local_decl_tree = TREE_VALUE (decl_list);
			const_git_decl local_decl = tree_to_git_decl (gim, gim_help, local_decl_tree);
			functionAddLocalDecl (git_func, local_decl);
		}

		// Create blocks
		basic_block bb;
		FOR_EACH_BB_FN (bb, func)
		{
			git_block block = basic_block_to_git_block (gim, gim_help, bb);
			functionAddBlock (git_func, func_help, bb, block);
		}

		functionSetFirstBlock (git_func, func_help, func->cfg->x_entry_block_ptr->next_bb);

		// Create call graph
		FOR_EACH_BB_FN (bb, func)
		{
			edge e;
			edge_iterator ei;
			FOR_EACH_EDGE (e, ei, bb->succs)
			{
			if (e->dest != func->cfg->x_exit_block_ptr)
				add_git_edge (func_help, e);
			}
		}


		reportFunction (gim, func_decl, git_func);
		if (is_main_function (fn_decl_tree))
			reportMain (gim, func_decl);
    }
    endFunctionConstruction();
}



/**
 * Gets GIMPLE++ type object for @c bool type.
 * 
 * @param gim pointer to @c Gimple object
 * @return GIMPLE++ type object for @c bool type
 */
const_git_type 
gb_get_bool_type (git_gimple gim, git_gimple_helpers gim_help)
{
    return tree_to_git_type (gim, gim_help, boolean_type_node);
}


/**
 * Gets GIMPLE++ type object for @c void* type.
 * 
 * @param gim pointer to @c Gimple object
 * @return GIMPLE++ type object for @c void* type
 */
const_git_type 
gb_get_void_ptr_type (git_gimple gim, git_gimple_helpers gim_help)
{
    return tree_to_git_type (gim, gim_help, ptr_type_node);
}


/**
 * Returns the pointer to the first node in the pool of global variables.
 * 
 * @return pointer to the first node in the pool of global variables
 */
struct varpool_node*
gb_varpool_nodes (void)
{
    //gcc_assert (varpool_nodes);
    return (struct varpool_node*) varpool_nodes;
}

/**
 * Returns the pointer to the the next node in the pool of global variables.
 * 
 * @return pointer to the the next node in the pool of global variables
 */
struct varpool_node*
gb_varpool_next (struct varpool_node* vn)
{
    gcc_assert (vn);
    struct varpool_node* node = (struct varpool_node*)vn;
    
    //while (node && !node->needed)
        node = node->next;
    
    return (struct varpool_node*) node;
}


/**
 * Returns the GIMPLE++ declaration object for the specified node in the pool of global variables.
 * 
 * @return GIMPLE++ declaration object for the specified node in the pool of global variables
 */
const_git_decl 
gb_varpool_decl (git_gimple gim, git_gimple_helpers gim_help, struct varpool_node* vn)
{
    gcc_assert (vn);
    struct varpool_node* node = (struct varpool_node*)vn;
    
    return tree_to_git_decl (gim, gim_help, node->decl);
}
