// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file Gimple.cpp
 * Definitions for class holding GIMPLE++ representation of program.
 */
#include "Gimple.h"
#include "Constant.h"
#include "FunctionDecl.h"
#include "Type.h"
#include "BoolType.h"
#include "VoidType.h"
#include "Function.h"

#include "ConstructionHelpers.h"
#include "GimpleHelpers.h"

#include <algorithm>
#include <typeinfo>
#include <fstream>

namespace giterator { 
    
/** @namespace giterator::detail 
 * This namespace contains internal classes, types and functions of GIMPLE Iterator.
 * The user does not need to use anything from this namespace.
 */
using namespace detail;

Gimple::Gimple() : m_mainFunc(NULL), m_listener(NULL)
{
}
    
Gimple::Gimple (Listener* listener) : m_mainFunc (NULL), m_listener (listener)
{
}

Gimple::~Gimple ()
{
    // Explicitly free the Functions. Other parts are managed by the m_treeCache
    for (functionsIterator i = m_functions.begin(); i != m_functions.end(); ++i)
    {
        delete i->second;
    }
}

const FunctionDecl*
Gimple::entryPoint ()
{
	return m_mainFunc;
}

template <typename T>
static std::auto_ptr<T>
makeAP (T* t)
{
    return std::auto_ptr<T> (t);
}

const Type* 
Gimple::getBoolType () 
{
	Type* retType = new BoolType ("bool8", 8);
	addToTreeCache(std::auto_ptr<TreeProduct> (dynamic_cast<TreeProduct*> (retType)));
	return retType;
}

void
Gimple::addToTreeCache (std::auto_ptr<detail::TreeProduct> obj)
{
	m_treeCache.push_back(obj);
}

void
Gimple::doReportCst_ (const Constant* constant)
{
    if (m_listener)
        m_listener->onCst (constant);
}

void 
Gimple::reportCst (std::auto_ptr<Constant> constant)
{
    m_constants.push_back(constant.get());
    doReportCst_(constant.get());
    constant.release();
}

void
Gimple::doReportFuncDef_ (const FunctionDecl* decl, const Function* func)
{
    if (m_listener)
       m_listener->onFuncDef (decl, func);
}

void 
Gimple::reportFuncDef (const FunctionDecl* decl, std::auto_ptr<Function> func)
{
    m_functions[decl] = func.get();
    doReportFuncDef_(decl, func.get());
    func.release();
}

void
Gimple::doReportDecl_ (const Decl* decl)
{
    if (m_listener)
        m_listener->onDecl (decl);
}

void
Gimple::doAllocGlobalArea_()
{
	if (m_listener)
		m_listener->allocGlobalArea();
}

void 
Gimple::reportDecl (std::auto_ptr<Decl> decl)
{
    m_declarations.push_back(decl.get());
    doReportDecl_(decl.get());
    decl.release();
}

void
Gimple::reportMain (const FunctionDecl* mainDecl)
{
    m_mainFunc = mainDecl;
}

void
Gimple::deserialize_routine()
{
    for (constantsConstIterator i1 = m_constants.begin(); i1 != m_constants.end(); ++i1) {
        doReportCst_(*i1);
    }

    for (declarationsConstIterator i1 = m_declarations.begin(); i1 != m_declarations.end(); ++i1) {
        doReportDecl_(*i1);
    }

    for (functionsConstIterator i1 = m_functions.begin(); i1 != m_functions.end(); ++i1) {
        doReportFuncDef_(i1->first, i1->second);
    }

    reportMain(m_mainFunc);
    doAllocGlobalArea_();
}

}
