// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file Gimple.h
 * Declaration of class holding GIMPLE++ representation of program.
 */
#ifndef GIT_GIMPLE_H
#define GIT_GIMPLE_H

#include "ObjectCache.h"
#include "ConstructionHelpers.h"

#include "git_fwd.h"

#include <memory>
#include <vector>
#include <map>

#include <boost/ptr_container/ptr_vector.hpp>
#include <boost/noncopyable.hpp>
#include <boost/serialization/serialization.hpp>
#include <boost/serialization/vector.hpp>
#include <boost/ptr_container/serialize_ptr_vector.hpp>

/**
 * Namespace of the GIMPLE Iterator, containing classes for converting GIMPLE representation
 * to GIMPLE++ representation.
 * This namespace contains all classes, types and functions that are intended to be used by the user.
 * Helper classes, types and functions can be found in namespace @c giterator::detail.
 */
namespace giterator {

/**
 * Class holding GIMPLE++ representation of the program.
 * This is the main class of GIMPLE Iterator.
 * To initialize this class, the @c processCfun() method in gimplexx-bridge.c must be called
 * for every compiled function, where conversion from GIMPLE @c tree structures to
 * GIMPLE++ objects start.
 * Every time a type, declaration, constant, statement, operation or operand occures,
 * the appropriate subclass of class Type, Decl, Constant, Stmt, Operation or Operand
 * holding it’s GIMPLE++ representation gets created, and is stored into the cache
 * of created object inside this class. Every time a function, declaration or constant
 * is processed, it is reported to the listener provided in the constructor.
 * Implementation of @c Listener - @c GimpleListener in namespace @c ginterpreter
 * is used to initialize memory structures of simulated program.
 *
 * @see ginterpreter::GimpleListener
 */
class Gimple : private boost::noncopyable {
  public:
	/*! type of the container with GIMPLE++ objects */
	typedef boost::ptr_vector <detail::TreeProduct> TreeCache;
	/*! type of the container with GIMPLE++ constants */
	typedef std::vector<Constant*> Constants;
	/*! type of the container with GIMPLE++ declarations */
	typedef std::vector<Decl*> Declarations;
	/*! type of the container with GIMPLE++ functions */
	typedef std::map<const FunctionDecl*, Function*> Functions;
	/*! type of the iterator GIMPLE++ constants container */
	typedef std::vector<Constant*>::const_iterator constantsConstIterator;
	/*! type of the iterator GIMPLE++ declarations container */
	typedef std::vector<Decl*>::const_iterator declarationsConstIterator;
	/*! type of the iterator with GIMPLE++ function definitions */
	typedef std::map<const FunctionDecl*, Function*>::const_iterator functionsConstIterator;
	/*! type of the iterator with GIMPLE++ function definitions */
	typedef std::map<const FunctionDecl*, Function*>::const_iterator functionsIterator;

    /**
     * The listener that is used to report constants, declarations and function definitions
     * encountered by @c Gimple class during the construction of GIMPLE++.
     */
    struct Listener {
        virtual ~Listener () {}
        
        virtual void onCst (const Constant* cst) = 0;
        virtual void onDecl (const Decl* decl) = 0;
        virtual void onFuncDef (const FunctionDecl* decl, const Function* func) = 0;
        virtual void allocGlobalArea () = 0;
    };

  private:
    /**
     * Cache of GIMPLE++ representations of everything,
     * that is produced from @c tree in GIMPLE.
     * Similar cache is maintained in class @c GimpleHelpers,
     * which is a map that stores GIMPLE @c tree as a tree
     * and a GIMPLE++ @c TreeProduct as a value.
     *
     * @see giterator::GimpleHelpers
     */
    TreeCache m_treeCache;
    
    /*! the declaration of program entry point function */
    const FunctionDecl* m_mainFunc;
    
    /** the listener supplied in constructor, used to report constants,
     * declarations and function definitions encountered.
     */
    Listener* m_listener;

	/*! list of constants managed by this class. Only for serialization/deserialization purposes. */
	Constants m_constants;
	/*! list of declarations managed by this class. Only for serialization/deserialization purposes. */
	Declarations m_declarations;
	/*! list of functions managed by this class. Only for serialization/deserialization purposes. */
	Functions m_functions;
  public:

    /**
	 * Constructor, creates a new GIMPLE++ representation holder class.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
    Gimple();

    /**
     * Constructor, creates a new GIMPLE++ representation holder class.
     *
     * @param listener listener that is used to notify the user about encountered
     *      declarations, function definitions and constants
     */
    Gimple (Listener* listener);

    /**
     * Destructor, performs cleanup.
     */
    ~Gimple ();
    
    /**
     * Returns declaration of program entry point function.
     * 
     * @return declaration of program entry point function,
     * or NULL, if no declaration was set
     */
    const FunctionDecl* entryPoint ();
    
    /**
     * Obtain GIMPLE++ representation of a boolean type.
     *
     * @return GIMPLE++ representation of a boolean type
     */
    const Type* getBoolType ();
  protected:

	/**
	 * Reports constant. Assume pointer ownership and report to the listener.
	 *
	 * @param constant constant to report
	 */
	void reportCst (std::auto_ptr<Constant> constant);

	/**
	 * Reports function definition. Assume pointer ownership and report to the
	 * listener.
	 *
	 * @param decl the declaration of the function
	 * @param func the definition of the function
	 */
	void reportFuncDef (const FunctionDecl* decl, std::auto_ptr<Function> func);

	/**
	 * Reports program entry point function encountered.
	 *
	 * @param mainDecl the declaration of program entry point function
	 */
	void reportMain (const FunctionDecl* mainDecl);

	/**
	 * Reports the declaration. Assume pointer ownership and report to the
	 * listener.
	 *
	 * @param decl declaration to be reported
	 */
	void reportDecl (std::auto_ptr<Decl> decl);
    
    /**
	 * Adds an object into cache of GIMPLE++ objects.
	 *
	 * @param obj object to be added to the tree cache
	 */
    void addToTreeCache (std::auto_ptr<detail::TreeProduct> obj);

    /**
     * Reports constant to the listener.
     *
     * @param constant constant to report
     */
    void doReportCst_ (const Constant* constant);

    /**
	 * Reports function to the listener.
	 *
	 * @param decl the declaration of the function
	 * @param func the definition of the function
	 */
    void doReportFuncDef_ (const FunctionDecl* decl, const Function* func);

    /**
     * Reports the declaration to the listener.
     *
     * @param decl declaration to be reported
     */
    void doReportDecl_ (const Decl* decl);

    /**
     * Alloc area for global variables after the end of GIMPLE processing.
     * During GIMPLE processing, size of global area is comupted inside
     * GlobalsManager, to which every global declaration is reported.
     * After processing of GIMPLE, size required for global area is known,
     * so this function can be called.
     *
     * @see GlobalsManager::allocArea()
     */
    void doAllocGlobalArea_();

    friend const_git_decl detail::get_git_decl_from_tree_cache (git_gimple_helpers gim_help, tree tree);
    friend const_git_constant detail::get_git_constant_from_tree_cache (git_gimple_helpers gim_help, tree tree);
    friend const_git_type detail::get_git_type_from_tree_cache (git_gimple_helpers gim_help, tree tree);
    friend void detail::add_git_decl_to_tree_cache (
            git_gimple gimple, git_gimple_helpers gim_help, tree tree, git_decl decl);
    friend void detail::add_git_type_to_tree_cache (
            git_gimple gimple, git_gimple_helpers gim_help, tree tree, git_type type);
    friend void detail::add_git_constant_to_tree_cache (
            git_gimple gimple, git_gimple_helpers gim_help, tree tree, git_constant cst);

    friend void detail::reportConstant (git_gimple gimple, git_constant constant);
    friend void detail::reportFunction (git_gimple gimple, const_git_decl funcDecl,
                                        git_function func);
    friend void detail::reportMain (git_gimple gimple, const_git_decl mainDecl);
    friend void detail::reportDecl (git_gimple gimple, git_decl decl);

    friend void add_to_tree_cache (Gimple* gimple, detail::TreeProduct* tp);

  public:

    friend class boost::serialization::access;
    template<class Archive>
    void save(Archive & ar, const unsigned int) const
    {
    	ar << m_treeCache;
    	ar << m_mainFunc;
    	ar << m_constants;;
    	ar << m_declarations;

    	size_t size = m_functions.size();
    	ar << size;
    	for (functionsConstIterator i1 = m_functions.begin(); i1 != m_functions.end(); ++i1)
    	{
    		ar << i1->first;
    		ar << i1->second;
    	}
    }

    template<class Archive>
    void load(Archive & ar, const unsigned int)
    {
    	ar >> m_treeCache;
    	FunctionDecl* mainFunc;
    	ar >> mainFunc;
    	m_mainFunc = (const FunctionDecl*) mainFunc;

    	ar >> m_constants;
    	ar >> m_declarations;

    	size_t size;
    	ar >> size;
    	for (size_t i = 0; i < size; ++i)
    	{
    		FunctionDecl* fd;
    		Function* f;
    		ar >> fd;
    		ar >> f;
    		m_functions[fd] = f;
    	}
    }

    template<class Archive>
    void serialize(Archive & ar, const unsigned int version)
    {
    	boost::serialization::split_member(ar, *this, version);
    }

	/**
	 * Do necessary operations needed after deserializing @c Gimple data
	 * structures - call reporting functions on the listener.
	 */
	void deserialize_routine();
};

}

#endif /* GIT_GIMPLE_H */

