// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file GimpleHelpers.h
 * Helper classes for creating GIMPLE++ representation of a program.
 */

#ifndef GIT_GIMPLE_HELPERS_H
#define GIT_GIMPLE_HELPERS_H

#include "TreeProduct.h"
#include "Block.h"
#include "FunctionInfo.h"
#include "gimple_types.h"

#include <boost/noncopyable.hpp>

#include <vector>
#include <map>
#include <memory>

namespace giterator
{

/**
 * Helper class for creating GIMPLE++ representation of a program.
 * It holds those data during the construction of GIMPLE++ representation
 * in gimplexx-bridge.c, that are needed outside the scope of a single function.
 * This class is accessed from gimplexx-bridge.c through functions
 * in ConstructionHelpers.cpp.
 * This class exists in one instance during the whole program GIMPLE++
 * representation creation.
 */
class GimpleHelpers : private boost::noncopyable {
private:
	/*! type of the container with GIMPLE++ objects. */
	typedef std::map <tree, detail::TreeProduct* > TreeCache;
	/*! type of the container with function names. */
	typedef std::map <const char*, bool> FuncNames;
	/*! type of the container with types and their names. */
	typedef std::map <tree, Type*> DefinedTypes;
	/*! type of the container with functions and their codes .*/
	typedef std::map<std::string, enum function_code> FunctionCodes;
	/*! type of the container with declarations and their constructors. */
	typedef std::map<const Decl*, Operand*> Constructors;

	/**
	 * Cache of everything, that is produced from @c tree in GIMPLE.
	 * The keys used in the cache are the @c tree values, while the values are
	 * the objects representing the same thing in GIMPLE++.
	 */
	TreeCache m_treeCache;

	/*! Function names encountered during construction. */
	FuncNames m_funcNames;

	/*!Types encountered during construction */
	DefinedTypes m_definedTypes;

	/**
	 * Codes for builtin and standard functions.
	 *
	 * @see enum function_code
	 */
	FunctionCodes m_functionCodes;

	/*!	Whether global variables were already processed. */
	bool m_globalsResolved;

	/*! Whether currently processed declaration is recursive. */
	bool m_isRecursive;

	/*! Constructors for those declarations, that have any. */
	Constructors m_constructors;

	/*! Counter for function types, to assign unique name to function type. */
	int m_funcNum;
	/*! Counter for enumeral types, to assign unique name to enumeral type. */
	int m_enumNum;
	/*! Counter for record types, to assign unique name to record type. */
	int m_recordNum;
	/*! Counter for record types, to assign unique name to void type. */
	int m_voidNum;
public:
	/**
	 * Constructor, creates new helpers class.
	 * Initializes map for function codes
	 */
	GimpleHelpers ()
	: m_globalsResolved(false), m_isRecursive(false),
	  m_funcNum(0), m_enumNum(0), m_recordNum(0), m_voidNum(0)
	{
		createFunctionCodes();
	}

	/**
	 * Adds an object into tree cache.
	 *
	 * @param tree the GIMPLE @c tree representation of the added object
	 * @param obj the object representing the same thins as @c tree does, but in GIMPLE++
	 */
	void addToTreeCache (tree tree, detail::TreeProduct* obj)
	{
		//assert (m_treeCache.find (tree) == m_treeCache.end ());
		if (m_treeCache.find (tree) == m_treeCache.end ())
			m_treeCache[tree] = obj;
	}

	/**
	 * Gets an object from tree cache.
	 * The returned value is casted to the required type.
	 *
	 * @tparam T type of the object that should be returned from the cache
	 * @param tree the key to the cache - GIMPLE @c tree representation of the object to get
	 */
	template <typename T>
	const T* getFromTreeCache (tree tree) const
	{
		TreeCache::const_iterator it = m_treeCache.find (tree);
		if (it != m_treeCache.end ())
		{
			const detail::TreeProduct* tp = it->second;
			return static_cast<const T*> (tp);
		}
		else
			return 0;
	}

	/**
	 * Add a function name into already processed functions.
	 *
	 * @param funcName name of the added function
	 */
	void addFuncName(const char* funcName)
	{
		assert(m_funcNames.find (funcName) == m_funcNames.end ());
		m_funcNames[funcName] = 1;
	}

	/**
	 * Check whether a function of a specified name was already
	 * processed.
	 *
	 * @param funcName checked function name
	 * @return true if function with name @c name was already
	 * processed, false otherwise
	 */
	bool checkFuncNameExists(const char* funcName) const
	{
		if (m_funcNames.find (funcName) != m_funcNames.end ())
			return true;
		return false;
	}

	/**
	 * Set variable informing that global variables were
	 * already processed to avoid their duplicate processing.
	 */
	void setGlobalsResolved ()
	{ m_globalsResolved = true;	}

	/**
	 * Check whether global variables were already processed.
	 *
	 * @return true if global variables were already processed,
	 * false otherwise
	 */
	bool getGlobalsResolved () const
	{ return m_globalsResolved; }

	/**
	 * Add a type with specified name to already defined types.
	 * Used to handle recursive declarations (like linked list elements).
	 *
	 * @param typeName name of the added type
	 * @param type the added type with name @c typeName
	 */
	void addDefinedType(const char* typeName, Type* type)
	{
		/*assert(m_definedTypes.find (typeName) == m_definedTypes.end ());
		m_definedTypes[typeName] = type;*/
	}

	/**
	 * Get a type of a specified name from already defined types.
	 * Used to handle recursive declarations (like linked list elements).
	 *
	 * @param typeName name of the type to get
	 * @return pointer to the type with name @c typeName, if such type
	 * was already added, NULL otherwise.
	 */
	Type* getDefinedType(const char* typeName) const
	{
		/*DefinedTypes::const_iterator i1 = m_definedTypes.find (typeName);
		if ( i1 != m_definedTypes.end ())
			return i1->second;
		else*/
			return NULL;
	}

	/**
	 * Add a type to already defined types.
	 * Used to handle recursive declarations (like linked list elements).
	 *
	 * @param typeTree GIMPLE representation of the added type
	 * @param type the added type with GIMPLE tree type @c typeTree
	 */
	void addDefinedType2(tree typeTree, Type* type)
	{
		assert(m_definedTypes.find (typeTree) == m_definedTypes.end ());
		m_definedTypes[typeTree] = type;
	}

	/**
	 * Get a type from already defined types.
	 * Used to handle recursive declarations (like linked list elements).
	 *
	 * @param typeTree GIMPLE representation of the type to get
	 * @return pointer to the type with GIMPLE tree type @c typeTree, if such type
	 * was already added, NULL otherwise.
	 */
	Type* getDefinedType2(tree typeTree) const
	{
		DefinedTypes::const_iterator i1 = m_definedTypes.find (typeTree);
		if ( i1 != m_definedTypes.end ())
			return i1->second;
		else
			return NULL;
	}

	/**
	 * Check whether currently processed declaration is recursive
	 * (e.g. linked list element).
	 *
	 * @return true if currently processed declaration is recursive,
	 * 	false otherwise
	 */
	bool isRecursive() const
	{
		return m_isRecursive;
	}

	/**
	 * Set a variable informing that currently processed declaration is recursive
	 * (e.g. linked list element).
	 *
	 * @param recursive value of the variable (true - recursive, false - not recursive)
	 */
	void setRecursive(bool recursive)
	{
		m_isRecursive = recursive;
	}

	/**
	 * Get a function code for a function with specified name.
	 *
	 * @param name name of function to get a code for.
	 * @return code of that function if it is a builtin or
	 * standard function, or NOT_BUILT_IN_FUNC otherwise.
	 */
	enum function_code getFunctionCode(const std::string& name)
	{
		if (m_functionCodes.find(name) != m_functionCodes.end())
			return m_functionCodes.at(name);
		else
			return NOT_BUILT_IN_FUNC;
	}

	/**
	 * Add a constructor (initializer) for specified declaration to the
	 * constructors map.
	 *
	 * @param decl declaration with constructor @c op
	 * @param op costructor of declaration @c decl
	 */
	void setConstructor(const Decl* decl, Operand* op)
	{
		assert(m_constructors.find(decl) == m_constructors.end());
		m_constructors[decl] = op;
	}

	/**
	 * Get a constructor (initializer) for specified declaration
	 * from the constructors map.
	 *
	 * @param decl declaration for which to get tje constructor
	 * @return constructor for declaration @c decl if there is
	 * any defined, or NULL otherwise
	 */
	Operand* getConstructor(const Decl* decl)
	{
		if(m_constructors.find(decl) == m_constructors.end())
			return NULL;

		return m_constructors[decl];
	}

    /**
     * Obtain the value of counter of function types for assigning
     * unique names to unnamed function types (types, that have no
     * name set in GIMPLE tree).
     *
     * @return value of counter of funtion types with unassigned name
     *
     * @see create_git_type_from_tree() in gimplexx-bridge.c
     */
    int getFuncNum()
	{ return m_funcNum++; }

    /**
	 * Obtain the value of counter of enumeral types for assigning
	 * unique names to unnamed enumeral types (types, that have no
	 * name set in GIMPLE tree).
	 *
	 * @return value of counter of enumeral types with unassigned name
	 *
	 * @see create_git_type_from_tree() in gimplexx-bridge.c
	 */
	int getEnumNum()
	{ return m_enumNum++; }

	/**
	 * Obtain the value of counter of record types for assigning
	 * unique names to unnamed record types (types, that have no
	 * name set in GIMPLE tree).
	 *
	 * @return value of counter of record types with unassigned name
	 *
	 * @see create_git_type_from_tree() in gimplexx-bridge.c
	 */
	int getRecordNum()
	{ return m_recordNum++; }

	/**
	 * Obtain the value of counter of void types for assigning
	 * unique names to these types.
	 *
	 * @return value of counter of void types
	 *
	 * @see create_git_type_from_tree() in gimplexx-bridge.c
	 */
	int getVoidNum()
	{ return m_voidNum++; }
private:
	/**
	 * Initialize data structure for obtaining function codes with
	 * values from FunctionInfo.h.
	 */
	void createFunctionCodes()
	{
		m_functionCodes[BUILT_IN_PRINTF_NAME] = GCC_BUILT_IN_PRINTF;
		m_functionCodes[BUILT_IN_PRINTF_NAME2] = GCC_BUILT_IN_PRINTF;
		m_functionCodes[BUILT_IN_PUTS_NAME] = GCC_BUILT_IN_PUTS;
		m_functionCodes[BUILT_IN_PUTS_NAME2] = GCC_BUILT_IN_PUTS;
		m_functionCodes[BUILT_IN_PUTCHAR_NAME] = GCC_BUILT_IN_PUTCHAR;
		m_functionCodes[BUILT_IN_PUTCHAR_NAME2] = GCC_BUILT_IN_PUTCHAR;
		m_functionCodes[BUILT_IN_SCANF_NAME] = GCC_BUILT_IN_SCANF;
		m_functionCodes[BUILT_IN_SCANF_NAME2] = GCC_BUILT_IN_SCANF;
		m_functionCodes[BUILT_IN_MALLOC_NAME] = GCC_BUILT_IN_MALLOC;
		m_functionCodes[BUILT_IN_MALLOC_NAME2] = GCC_BUILT_IN_MALLOC;
		m_functionCodes[BUILT_IN_FREE_NAME] = GCC_BUILT_IN_FREE;
		m_functionCodes[BUILT_IN_FREE_NAME2] = GCC_BUILT_IN_FREE;
		m_functionCodes[BUILT_IN_ABORT_NAME] = GCC_BUILT_IN_ABORT;
		m_functionCodes[BUILT_IN_ABS_NAME] = GCC_BUILT_IN_ABS;
		m_functionCodes[BUILT_IN_LABS_NAME] = GCC_BUILT_IN_LABS;
		m_functionCodes[BUILT_IN_LLABS_NAME] = GCC_BUILT_IN_LLABS;
		m_functionCodes[BUILT_IN_ATOF_NAME] = GCC_BUILT_IN_ATOF;
		m_functionCodes[BUILT_IN_ATOI_NAME] = GCC_BUILT_IN_ATOI;
		m_functionCodes[BUILT_IN_ATOL_NAME] = GCC_BUILT_IN_ATOL;
		m_functionCodes[BUILT_IN_ATOLL_NAME] = GCC_BUILT_IN_ATOLL;
		m_functionCodes[BUILT_IN_EXIT_NAME] = GCC_BUILT_IN_EXIT;
		m_functionCodes[BUILT_IN_RAND_NAME] = GCC_BUILT_IN_RAND;
		m_functionCodes[BUILT_IN_RANDOM_NAME] = GCC_BUILT_IN_RANDOM;
		m_functionCodes[BUILT_IN_SRAND_NAME] = GCC_BUILT_IN_SRAND;
		m_functionCodes[BUILT_IN_SRANDOM_NAME] = GCC_BUILT_IN_SRANDOM;
		m_functionCodes[BUILT_IN_STRTOD_NAME] = GCC_BUILT_IN_STRTOD;
		m_functionCodes[BUILT_IN_STRTOLD_NAME] = GCC_BUILT_IN_STRTOLD;
		m_functionCodes[BUILT_IN_STRTOF_NAME] = GCC_BUILT_IN_STRTOF;
		m_functionCodes[BUILT_IN_STRTOL_NAME] = GCC_BUILT_IN_STRTOL;
		m_functionCodes[BUILT_IN_STRTOLL_NAME] = GCC_BUILT_IN_STRTOLL;
		m_functionCodes[BUILT_IN_STRTOUL_NAME] = GCC_BUILT_IN_STRTOUL;
		m_functionCodes[BUILT_IN_STRTOULL_NAME] = GCC_BUILT_IN_STRTOULL;
		m_functionCodes[BUILT_IN_MEMCPY_NAME] = GCC_BUILT_IN_MEMCPY;
		m_functionCodes[BUILT_IN_MEMCPY_NAME2] = GCC_BUILT_IN_MEMCPY;
		m_functionCodes[BUILT_IN_MEMMOVE_NAME] = GCC_BUILT_IN_MEMMOVE;
		m_functionCodes[BUILT_IN_MEMMOVE_NAME2] = GCC_BUILT_IN_MEMMOVE;
		m_functionCodes[BUILT_IN_STRCAT_NAME] = GCC_BUILT_IN_STRCAT;
		m_functionCodes[BUILT_IN_STRCAT_NAME2] = GCC_BUILT_IN_STRCAT;
		m_functionCodes[BUILT_IN_STRNCAT_NAME] = GCC_BUILT_IN_STRNCAT;
		m_functionCodes[BUILT_IN_STRNCAT_NAME2] = GCC_BUILT_IN_STRNCAT;
		m_functionCodes[BUILT_IN_STRCPY_NAME] = GCC_BUILT_IN_STRCPY;
		m_functionCodes[BUILT_IN_STRCPY_NAME2] = GCC_BUILT_IN_STRCPY;
		m_functionCodes[BUILT_IN_STRCNPY_NAME] = GCC_BUILT_IN_STRCNPY;
		m_functionCodes[BUILT_IN_STRCNPY_NAME2] = GCC_BUILT_IN_STRCNPY;
		m_functionCodes[BUILT_IN_STRCMP_NAME] = GCC_BUILT_IN_STRCMP;
		m_functionCodes[BUILT_IN_STRCMP_NAME2] = GCC_BUILT_IN_STRCMP;
		m_functionCodes[BUILT_IN_STRNCMP_NAME] = GCC_BUILT_IN_STRNCMP;
		m_functionCodes[BUILT_IN_STRNCMP_NAME2] = GCC_BUILT_IN_STRNCMP;
		m_functionCodes[BUILT_IN_STRLEN_NAME] = GCC_BUILT_IN_STRLEN;
		m_functionCodes[BUILT_IN_STRLEN_NAME2] = GCC_BUILT_IN_STRLEN;
		m_functionCodes[BUILT_IN_ASSERT_NAME] = GCC_BUILT_IN_ASSERT;
		m_functionCodes[BUILT_IN_ASSERT_NAME2] = GCC_BUILT_IN_ASSERT;
		m_functionCodes[BUILT_IN_ASSERT_NAME3] = GCC_BUILT_IN_ASSERT;
		m_functionCodes[BUILT_IN_SLEEP_NAME] = GCC_BUILT_IN_SLEEP;
	}
};

/**
 * Helper class for creating GIMPLE++ representation of a single function.
 * It holds those data during the construction of GIMPLE++ representation of
 * a single function in gimplexx-bridge.c, that are needed across functions.
 * This class is accessed from gimplexx-bridge.c through functions
 * in ConstructionHelpers.cpp.
 * There is one instance of this class created every time a GIMPLE function
 * is being processed.
 */
class FunctionHelpers {
private:
	/*! type of the container with GIMPLE++ blocks */
	typedef std::map<basic_block, Block* > Blocks;

	/*! blocks of statements that contain the body of the function */
	Blocks m_blocks;

	/*! the first block of the function */
	Block* m_firstBlock;
public:
	/**
	 * Constructor, creates a new helpers class.
	 */
	FunctionHelpers () {}

	/**
	 * Adds the block to this function.
	 *
	 * @param basicBlock GIMPLE representation of basic block
	 * @param block GIMPLE++ block created from @c basicBlock
	 */
	void addBlock (basic_block basicBlock, Block* block)
	{
		m_blocks[basicBlock] =  block;
	}

	/**
	 * Returns block for specified @c basic_block.
	 * @c basic_block is a pointer to structure that represents the block in GIMPLE.
	 *
	 * @param basicBlock GIMPLE representation of the block
	 * @return block that was created from @c basicBlock
	 */
	Block* getBlock (basic_block basicBlock) const
	{
		Blocks::const_iterator it = m_blocks.find (basicBlock);
		assert (it != m_blocks.end ());
		return it->second;
	}

    /**
     * Sets the first block of the processed function.
     * The execution of the function starts at the first block.
     *
     * @param first block holder representing the block,
     * that is the first function block
     */
    void setFirstBlock (basic_block first)
    {
        Blocks::iterator it = m_blocks.find (first);
        assert (it != m_blocks.end ());
        m_firstBlock = it->second;
    }

    /**
     * Returns the first block of the processed function.
     * The execution of the function starts at the first block.
     *
     * @return the first block of the processed function
     */
    Block* getFirstBlock () const
    { return m_firstBlock; }
};

}

#endif /* GIT_GIMPLE_HELPERS_H */
