// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file GotoStmt.h
 * Class representing GIMPLE_GOTO statement.
 */
#ifndef GIT_GOTO_STATEMENT_H
#define GIT_GOTO_STATEMENT_H

#include "Stmt.h"
#include "Operand.h"

#include <boost/serialization/serialization.hpp>
#include <memory>

namespace giterator {

/**
 * Class representing the @c GIMPLE_GOTO statement, which is an unconditional jump.
 *
 * GIMPLE_GOTO looks like this:
 * GIMPLE_GOTO <TARGET>,
 * where TARGET is LABEL_DECL or expression node for computed GOTOs. It determines
 * where to jump.
 *
 * @see gimple_goto_to_git_stmt() in gimplexx-bridge.c to see how
 * a goto statement gets created from GIMPLE tree node.
 * @see giterator::LabelStmt
 */
class GotoStmt : public Stmt {
  private:
	/*! Operand representing destination of the jump */
	std::auto_ptr<Operand> m_dest;
  public:
	/**
	 * Constructor, creates a new goto statement.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	GotoStmt() {}

	/**
	 * Constructor, creates a new goto statement.
	 *
	 * @param dest operand representing destination of the jump
	 */
	GotoStmt(std::auto_ptr<Operand> dest) : m_dest(dest) {}

    void accept (Stmt::Visitor& visitor) const
    { visitor.visit (*this); }

    /**
     * Obtain operand representing destination of the this jump.
     *
     * @returns Operand representing destination of the this jump
     */
    const Operand* getDest () const
	{ return m_dest.get(); }

    friend class boost::serialization::access;
	template<class Archive>
	void save(Archive & ar, const unsigned int) const
	{
		Operand* dest = m_dest.get();
		ar << dest;
	}

	template<class Archive>
	void load(Archive & ar, const unsigned int)
	{
		Operand* dest;
		ar >> dest;
		m_dest = std::auto_ptr<Operand> (dest);
	}

	template<class Archive>
	void serialize(Archive & ar, const unsigned int version)
	{
		boost::serialization::void_cast_register(static_cast<GotoStmt *>(NULL), static_cast<Stmt *>(NULL));
		boost::serialization::split_member(ar, *this, version);
	}
};

}

#endif /* GIT_GOTO_STATEMENT_H */

