// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file IntegerConstant.h
 * Class representing an integer constant.
 */
#ifndef GIT_INTEGER_CONSTANT_H
#define GIT_INTEGER_CONSTANT_H

#include "Constant.h"
#include "Type.h"
#include "Serialization.h"

#include <stdint.h>

namespace giterator { 
    
/**
 * Class representing an integer constant.
 * Integer constants get created from GIMPLE @c INTEGER_CST nodes.
 * The value of integer constant can be retrieved from GIMPLE by
 * <code> ((TREE_INT_CST_HIGH (e) << HOST_BITS_PER_WIDE_INT)
 *         + TREE_INST_CST_LOW (e)) </code>.
 * TREE_INT_CST_HIGH and TREE_INT_CST_LOW return a HOST_WIDE_INT, HOST_BITS_PER_WIDE_INT
 * is at least 32 on all platforms.
 * Value of integer constant can by signed or unsigned, depending on it's type.
 * Char constants are represented as INTEGER_CSTs.
 *
 * @see create_git_constant_from_tree() in gimplexx-bridge.c to
 * see how integer constants get created from GIMPLE constants
 */
class IntegerConstant : public Constant {
  private:
    /*! type of this constant */
    const Type* m_type;
    
    /** raw value of this constant
     * All values are passed in as a @c uint64_t.
     * The precision and the unsigned flag in the @c type specify, how will be
     * the raw value represented.
     */
    uint64_t m_value;
    
  public:

    /**
	 * Constructor, creates a new integer constant.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
    IntegerConstant() {}
    /**
     * Constructor, , creates a new integer constant.
     * 
     * @param type the type of the constant
     * @param value raw value of the constant, retrieved from GIMPLE.
     *
     * @see m_value
     */
    IntegerConstant (const Type* type, uint64_t value) 
        : m_type (type), m_value (value) {}
  
    virtual void accept (Visitor& visitor) const
    { visitor.visit (*this); }
    
    /**
     * Obtain the type of this constant.
     *
     * @return type of this constant.
     */
    const Type* getType () const
    { return m_type; }

    /**
     * Returns the raw value of this constant.
     *
     * @return raw value of this constant
     *
     * @see m_value
     */
    uint64_t getValue () const
    { return m_value; }

	friend class boost::serialization::access;
	template<class Archive>
	void save(Archive & ar, const unsigned int) const
	{
		ar << m_type;
		ar << m_value;
	}

	template<class Archive>
	void load(Archive & ar, const unsigned int)
	{
		Type* type;
		ar >> type;
		m_type = (const Type*) type;
		ar >> m_value;
	}

	template<class Archive>
	void serialize(Archive & ar, const unsigned int version)
	{
		boost::serialization::void_cast_register(static_cast<IntegerConstant *>(NULL), static_cast<TreeProduct *>(NULL));
		boost::serialization::void_cast_register(static_cast<IntegerConstant *>(NULL), static_cast<Constant *>(NULL));
		boost::serialization::split_member(ar, *this, version);
	}
};

}

#endif /* GIT_INTEGER_CONSTANT_H */
