// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file IntegerType.h
 * Class representing integer type.
 */
#ifndef GIT_INTEGER_TYPE_H
#define GIT_INTEGER_TYPE_H

#include "AbstractType.h"

#include <boost/serialization/serialization.hpp>

namespace giterator { 
    
/**
 * Class representing various integer types
 * (char, short, int, long, long long, ...).
 * Enumeration and boolean types are not represented
 * by this class.
 * Integer types get created from GIMPLE @c INTEGER_TYPE
 * nodes. Each such node has @c TYPE_PRECISION determining
 * number of bits used for representation (not the same as @c TYPE_SIZE).
 * @c TYPE_UNSINGED determines whether it is signed or unsigned
 * integer type. These two information are stored in instance
 * variables of the class.
 *
 * @see create_git_type_from_tree() in gimplexx-bridge.c to see
 * how integer types get created from GIMPLE types
 *
 * @see giterator::Type
 */
class IntegerType : public detail::AbstractType {
  private:
    /*! Type precision in bits */
    size_t m_bitPrecision;
    /*! flag determining whether this type is signed or unsinged */
    bool m_isUnsigned;

  public:
    /**
	 * Constructor, creates a new integer type.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
    IntegerType() : detail::AbstractType() {}
    /**
     * Constructor, creates a new integer type.
     * 
     * @param name name of the type
     * @param bitSize size of the type in bits, obtained from GIMPLE @c TYPE_SIZE
     * @param isUnsigned @c true if the new type is unsigned, obtained from GIMPLE @c TYPE_UNSIGNED
     * @param bitPrecision number of bits for representation of the new type,
	 * obtained from GIMPLE @c TYPE_PRECISION
     */
    IntegerType (const std::string& name,
                 size_t bitSize, bool isUnsigned, size_t bitPrecision) 
        : detail::AbstractType (name, bitSize), 
          m_bitPrecision (bitPrecision), m_isUnsigned (isUnsigned) {}

    /**
     * Returns the precision of this type.
     * Bit precision determines the number of bits used for the representation
     * of this type.
     * 
     * @return the precision of this type in bits
     */
    size_t getBitPrecision() const 
    { 
      return m_bitPrecision; 
    }

    /**
     * Checks if this type is unsigned.
     * 
     * @return @c true if this type is unsigned, @c false otherwise
     */
    size_t isUnsigned () const 
    { 
      return m_isUnsigned; 
    }

    /**
	 * Compare this type with another type.
	 *
	 * @param type type to be compared
	 * @return true if the types are the same (they are both integer types, or
	 * compared type is enumeral or boolean), false otherwise
	 */
    bool isSame (const Type& type) const
    {
    	// TODO:
    	return true;
      //if (typeid (type) == typeid (IntegerType) || typeid (type) == typeid (BoolType))
    	//  return true;

      //return true;//getBitSize() == intType->getBitSize()
              //&& getBitPrecision() == intType->getBitPrecision();
              //&& isUnsigned() == intType->isUnsigned();
    }

    friend class boost::serialization::access;
	template<class Archive>
	void serialize(Archive & ar, const unsigned int)
	{
		boost::serialization::void_cast_register(static_cast<IntegerType *>(NULL), static_cast<TreeProduct *>(NULL));
		boost::serialization::void_cast_register(static_cast<IntegerType *>(NULL), static_cast<Type *>(NULL));
		ar & boost::serialization::base_object<AbstractType>(*this);
		ar & m_bitPrecision;
		ar & m_isUnsigned;
	}
};

}

#endif /* GIT_INTEGER_TYPE_H */

