// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file LabelExprOperand.h
 * Class representing a label expression.
 */
#ifndef GIT_LABEL_EXPR_OPERAND_H
#define GIT_LABEL_EXPR_OPERAND_H

#include <memory>

#include "Operand.h"
#include "Decl.h"

#include <stdint.h>
#include <boost/serialization/serialization.hpp>

namespace giterator {

/**
 * The operand representing a label expression - either a case label (range of case labels),
 * or a default label.
 * These operands get created from GIMPLE @c CASE_LABEL_EXPR nodes.
 * If @c CASE_LOW is NULL, it is the default label, if @c CASE_HIGH is NULL, it is a regular
 * case label, if both are not NULL, it is a range of case labels.
 * CASE_LOW and CASE_HIGH are INTEGER_CST nodes. Their type is equal to type of condition
 * expression in switch statement.
 *
 * @see create_git_operand_from_tree in gimplexx-bridge.c to see how label expression
 * operands get created from GIMPLE operands
 */
class LabelExprOperand : public Operand {
  private:
	/*! label declaration for this operand */
	const Decl* m_decl;
	/** operand representing the label expression for regular and default case labels
	 * or lower bound of range case label
	 */
	std::auto_ptr<Operand> m_lower;
	/*! operand representing upper bound of range case label. Valid only for range labels */
	std::auto_ptr<Operand> m_upper;

	/*! Whether this is the default case label */
	bool m_isDefault;
	/*! Whether this is a range case label */
	bool m_isRange;
  public:
	/**
	 * Constructor, creates a new label expression operand.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	LabelExprOperand () {}
    /**
     * Constructor, creates a new label expression operand
     * representing regular or default case label (not range).
     *
     * @param decl label declaration for this operand
     * @param label operand representing the expression
     * @param isDefault whether this is the default label
     */
	LabelExprOperand (const Decl* decl, std::auto_ptr<Operand> label, bool isDefault) :
		m_decl(decl), m_lower(label), m_upper(0), m_isDefault(isDefault), m_isRange(false) {}

	/**
	 * Constructor, creates a new range label expression operand.
	 *
	 * @param decl label declaration for this operand
	 * @param lower operand representing the lower value of the range label
	 * @param upper operand representing the upper value of the range label
	 * @param isDefault whether this is the default label (false)
	 */
	LabelExprOperand (const Decl* decl, std::auto_ptr<Operand> lower, std::auto_ptr<Operand> upper, bool isDefault) :
		m_decl(decl), m_lower(lower), m_upper(upper), m_isDefault(isDefault), m_isRange(true) {}

    void accept (Visitor& v) const
    { v.visit (*this); }

    /**
     * Return type of the expression represented by this operand.
     *
     * @return type type of the expression represented by this operand
     */
    // TODO: decl has no type!
    const Type* getType () const
	{ return m_decl->getType (); }

    /**
     * Obtain a label declaration for this expression.
     *
     * @return label declaration for this expression
     */
    const Decl* getDecl () const
	{ return m_decl; }

    /**
     * Get the expression for regular label or lower bound for range label.
     * Invalid for default label.
     *
     * @return expression for regular label or lower bound for range label
     */
    const Operand* getLower () const
	{ return m_lower.get (); }

    /**
     * Get the expression for upper bound for range label.
     * Invalid for default and regular labels.
     *
     * @return expression for upper bound for range label.
     */
    const Operand* getUpper () const
	{ return m_upper.get (); }

    /**
     * Find out whether this is default label.
     *
     * @return @c true if this is default label, @c false otherwise
     */
    bool getIsDefault () const
    { return m_isDefault; }

    /**
	 * Find out whether this is range label.
	 *
	 * @return @c true if this is range label, @c false otherwise
	 */
    bool getIsRange () const
    { return m_isRange; }

    friend class boost::serialization::access;
	template<class Archive>
	void save(Archive & ar, const unsigned int) const
	{
		ar << m_decl;
		Operand* op = m_lower.get();
		ar << op;
		op = m_upper.get();
		ar << op;
		ar << m_isDefault;
		ar << m_isRange;
	}

	template<class Archive>
	void load(Archive & ar, const unsigned int)
	{
		Decl* decl;
		ar >> decl;
		m_decl = (const Decl*) decl;
		Operand* op;
		ar >> op;
		m_lower = std::auto_ptr<Operand> (op);
		ar >> op;
		m_upper = std::auto_ptr<Operand> (op);
		ar >> m_isDefault;
		ar >> m_isRange;
	}

	template<class Archive>
	void serialize(Archive & ar, const unsigned int version)
	{
		boost::serialization::void_cast_register(static_cast<LabelExprOperand *>(NULL), static_cast<Operand *>(NULL));
		boost::serialization::split_member(ar, *this, version);
	}
};

}

#endif /* GIT_LABEL_EXPR_OPERAND_H */
