// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file LabelStmt.h
 * Class representing GIMPLE_LABEL statement.
 */
#ifndef GIT_LABEL_STATEMENT_H
#define GIT_LABEL_STATEMENT_H

#include "Stmt.h"
#include "Operand.h"

#include <boost/serialization/serialization.hpp>
#include <memory>

namespace giterator {

/**
 * Class representing the @c GIMPLE_LABEL statement - unconditional jump targets.
 *
 * GIMPLE_LABEL looks like this:
 * GIMPLE_LABEL <LABEL> , where LABEL is a LABEL_DECL representing a jump target.
 *
 *
 * @see gimple_label_to_git_stmt() in gimplexx-bridge.c to see how
 * a label statement gets created from GIMPLE tree node.
 * @see giterator::GotoStmt, giterator::LabelDecl
 */
class LabelStmt : public Stmt {
  private:
	/*! Operand representing jump target (a label declaration). */
	std::auto_ptr<Operand> m_label;
  public:
	/**
	 * Constructor, creates a new label statement.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	LabelStmt() {}

	/**
	 * Constructor, creates a new label statement.
	 *
	 * @param label operand representing a jump target (a label declaration).
	 */
	LabelStmt(std::auto_ptr<Operand> label) : m_label(label) {}

    void accept (Stmt::Visitor& visitor) const
    { visitor.visit (*this); }

    /**
     * Obtain an operand representing jump target (a label declaration).
     */
    const Operand* getLabel () const
	{ return m_label.get(); }

    friend class boost::serialization::access;
	template<class Archive>
	void save(Archive & ar, const unsigned int) const
	{
		Operand* label = m_label.get();
		ar << label;
	}

	template<class Archive>
	void load(Archive & ar, const unsigned int)
	{
		Operand* label;
		ar >> label;
		m_label = std::auto_ptr<Operand> (label);
	}

	template<class Archive>
	void serialize(Archive & ar, const unsigned int version)
	{
		boost::serialization::void_cast_register(static_cast<LabelStmt *>(NULL), static_cast<Stmt *>(NULL));
		boost::serialization::split_member(ar, *this, version);
	}
};

}

#endif /* GIT_LABEL_STATEMENT_H */

