// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file ObjectCache.h
 * Cache that maps keys to objects.
 */
#ifndef GIT_OBJECT_CACHE_H
#define GIT_OBJECT_CACHE_H

#include <boost/ptr_container/ptr_map.hpp>
#include <memory>


namespace giterator { namespace detail {
    
/**
 * Cache that maps keys to objects.
 * Cache destroyes overwritten objects.
 * 
 * @tparam Key key to the cache
 * @tparam Value values that are stored in the cache
 */
template <typename Key, typename Value>
class ObjectCache {
  private:
    /*! type of the map holding the cached values */
    typedef boost::ptr_map<Key, Value> ObjectMap;

    /*! map with cached values */
    ObjectMap m_map;


  public:
    /**
     * Finds a value in the cache.
     * 
     * @param key key to lookup
     * @return pointer to the value stored in the cache under the key @c key, 
     *      or 0 if there is no entry for @c key
     */
    const Value* find (Key key) const
    {
      typename ObjectMap::const_iterator it = m_map.find (key);
      if (it == m_map.end ())
        return 0;
      else
        return it->second;
    }

    /**
     * Adds a value into the cache.
     * If there is a value stored under the key @c key, the old value is overwritten 
     * by the new one.
     * 
     * @param key key under which to store the value
     * @param obj value to store
     */
    void add (Key key, std::auto_ptr<Value> obj) 
    {
      typename ObjectMap::iterator it = m_map.find (key);
      if (it != m_map.end ())
        {
	  delete it->second;
	  it->second = obj.release ();
	}
      else
        m_map.insert (std::make_pair (key, obj.release ()));
    }
};

}}

#endif /* GIT_OBJECT_CACHE_H */

