// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file OffsetMapper.h
 * Declaration of class mapping variables to offsets.
 */
#ifndef GIT_OFFSET_MAPPER_H
#define GIT_OFFSET_MAPPER_H

#include "Decl.h"
#include "Type.h"

#include "RecordType.h"
#include "PtrType.h"
#include "ArrayType.h"

#include <typeinfo>
#include <map>


namespace giterator {
    
/**
 * Class mapping variables to offsets.
 * For more complex declarations like structures, offset mapper
 * determines the layout of fields inside the structure.
 * For example an offset mapper for structure:
 * <code> struct point { int x, y; }; </code>
 * where int is of size 4 bytes, therefore the whole struct is of size 8 bytes,
 * would map x to offset 0 and y to offset 4.
 * Offset mapper can be used to specify layout of complex data structures
 * to offsets in memory (Memory Module for example).
 * Also pointers to structures have their offset mappers, so access point->x accesses
 * the right offset. OffsetMapper therefore exist also inside VarDecl class
 * (which represents pointer to sturcure). For simple VarDecls (f.e. int i;),
 * OffsetMapper returns simply allways 0.
 *
 * @see giterator::RecordDecl, giterator::VarDecl, giterator::ParamDecl,
 * giterator::FieldDecl
 */
class OffsetMapper {
  protected:
    /*! type of the map holding the mapping represented by this class */
    typedef std::map<unsigned int, size_t> Offsets;
    
    /** If this mapper is created for simple declaration (like described above),
     * there is no need to maintain the mapping. In that case this variable is false.
     */
    bool m_created;
  public:
    /*! iterator for type of the map holding the mapping represented by this class */
  	typedef Offsets::const_iterator OffsetsIterator;
  private:
    /*! Map holding the mapping of variables to offsets. */
    Offsets m_offsets;
    /*! Size of the mapping - sum of the sizes of mapped variables. */
    size_t m_size;
    
  public:
    /**
     * Constructor, creates a new offset mapper.
     *
     * @param type type for which is the mapper constructed.
     */
    OffsetMapper (const Type* type);

    /**
     * Create a mapper for the specified type.
     * If it is record type, pointer to record type or array of record types,
     * create the map, or set the structure to return allways 0 otherwise.
     * This function is also called in constructor to initialize the mapper.
     *
     * @param type type for which is the mapper constructed
     */
    void create(const Type* type);

    /**
     * Get offset for specified declaration.
     *
     * @param uid unique ID of the declaration for which to get offset.
     */
    size_t getOffset (unsigned int uid) const;

    /**
     * Returns the size of this mapping - sum of sizes of mapped variables.
     */
    size_t getSize () const;
    
  protected:
    /**
     * Constructor, creates a new offset mapper.
     */
    OffsetMapper ();

    /**
     * Add a variable to the mapping of this mapper.
     * Offset of the declaration can be then retrieved with getOffset().
     *
     * @param decl declaration of the variable to add to the mapping.
     *
     * @see getOffset()
     */
    void addVariable (const giterator::Decl* decl);
};

}

#endif 
