// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file Operand.h
 * Base class for all classes representing operands.
 */

#ifndef GIT_OPERAND_H
#define GIT_OPERAND_H

#include "../git/git_fwd.h"

namespace giterator {
    
/**
 * Interface class for all classes representing operands.
 * Expressions in GIMPLE consist of an operation and simple operands for the operation.
 * Operands get created from various GIMPLE nodes, that satisfy
 * function is_gimple_operand(), which checks whether @c gimple_rhs_class
 * of the node is @c GIMPLE_SINGLE_RHS.
 * 
 * @see Operation
 * @see tree_to_git_operand() in gimplexx-bridge.c to see
 * how GIMPLE operands get converted to GIMPLE++ operands.
 */
class Operand {
  public:
    /**
     * Visitor for sub-classes of @c Operand class, 
     * that are meant to be used directly by the user.
     */
    struct Visitor {
        virtual ~Visitor () {}
        
        virtual void visit (const AddrOperand&) = 0;
        virtual void visit (const DeclOperand&) = 0;
        virtual void visit (const LabelExprOperand&) = 0;
        virtual void visit (const IndirectRefOperand&) = 0;
        virtual void visit (const ArrayRefOperand&) = 0;
        virtual void visit (const ComponentRefOperand&) = 0;
        virtual void visit (const CstOperand&) = 0;
        virtual void visit (const ArrayConstructorOperand&) = 0;
        virtual void visit (const RecordConstructorOperand&) = 0;
    };
    
    /**
     * Destructor.
     */
    virtual ~Operand () {}
    
    /**
     * Calls @c visitor.accept().
     * Classes implementing this interface implement this method to call @c visitor.accept()
     * with @c *this as a parameter.
     * 
     * @param v visitor on which to call the @c accept method
     */
    virtual void accept (Visitor& v) const = 0;
    
    /**
     * Returns the type of this operand.
     * 
     * @return the type of this operand
     */
    virtual const Type* getType () const = 0;
};

}

#endif /* GIT_OPERAND_H */

