// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file Operation.h
 * Interface for classes representing operations.
 */
#ifndef GIT_OPERATION_H
#define GIT_OPERATION_H

#include "Operations_fwd.h"

namespace giterator {
    
class Type;
class Operand;

/*
 * NOTE:
 * Parts of these comments were copied from tree.def file in gcc source directory
 * and gcc internal documentation at http://gcc.gnu.org/onlinedocs/gccint
 */


/**
 * Interface for classes representing GIMPLE expressions - further called operations.
 * Operations get created from GIMPLE *_EXPR nodes.
 * Subclasses of this interface represent concrete operations.
 *
 * @see create_operation() in gimplexx_bridge.c
 * to see how operations get created from GIMPLE expressions.
 */
class Operation {
  public:
    /**
     * Visitor class for sub-classes of @c Operation class, that are meant to be used 
     * by the user.
     */
    struct Visitor {
        virtual ~Visitor () {}
        
        virtual void visit (const NegateOperation&) = 0;
        virtual void visit (const AbsOperation&) = 0;
        virtual void visit (const FixTruncOperation&) = 0;
        virtual void visit (const FloatOperation&) = 0;
        virtual void visit (const BitNotOperation&) = 0;
        virtual void visit (const TruthNotOperation&) = 0;
        virtual void visit (const ParenOperation&) = 0;
        virtual void visit (const NonLvalueOperation&) = 0;
        virtual void visit (const NoOperation&) = 0;
        virtual void visit (const ConvertOperation&) = 0;
        virtual void visit (const PlusOperation&) = 0;
        virtual void visit (const MinusOperation&) = 0;
        virtual void visit (const MultOperation&) = 0;
        virtual void visit (const RdivOperation&) = 0;
        virtual void visit (const TruncDivOperation&) = 0;
        virtual void visit (const FloorDivOperation&) = 0;
        virtual void visit (const CeilDivOperation&) = 0;
        virtual void visit (const RoundDivOperation&) = 0;
        virtual void visit (const TruncModOperation&) = 0;
        virtual void visit (const FloorModOperation&) = 0;
        virtual void visit (const CeilModOperation&) = 0;
        virtual void visit (const RoundModOperation&) = 0;
        virtual void visit (const ExactDivOperation&) = 0;
        virtual void visit (const MinOperation&) = 0;
        virtual void visit (const MaxOperation&) = 0;
        virtual void visit (const LshiftOperation&) = 0;
        virtual void visit (const RshiftOperation&) = 0;
        virtual void visit (const LrotateOperation&) = 0;
        virtual void visit (const RrotateOperation&) = 0;
        virtual void visit (const BitIorOperation&) = 0;
        virtual void visit (const BitXorOperation&) = 0;
        virtual void visit (const BitAndOperation&) = 0;
        virtual void visit (const TruthAndOperation&) = 0;
        virtual void visit (const TruthOrOperation&) = 0;
        virtual void visit (const TruthXorOperation&) = 0;
        virtual void visit (const LtOperation&) = 0;
        virtual void visit (const LeOperation&) = 0;
        virtual void visit (const GtOperation&) = 0;
        virtual void visit (const GeOperation&) = 0;
        virtual void visit (const EqOperation&) = 0;
        virtual void visit (const NeOperation&) = 0;
        virtual void visit (const OrderedOperation&) = 0;
        virtual void visit (const UnorderedOperation&) = 0;
        virtual void visit (const UnLtOperation&) = 0;
        virtual void visit (const UnLeOperation&) = 0;
        virtual void visit (const UnGtOperation&) = 0;
        virtual void visit (const UnGeOperation&) = 0;
        virtual void visit (const UnEqOperation&) = 0;
        virtual void visit (const LtGtOperation&) = 0;
        virtual void visit (const WidenSumOperation&) = 0;
        virtual void visit (const WidenMultOperation&) = 0;
    };
    
  public:
    /**
     * Destructor.
     */
    virtual ~Operation () {}
    
    /**
     * Returns the type of the result of this operation.
     * 
     * @return the type of the result of this operation
     */
    virtual const Type* getResultType () const = 0;
    
    /**
     * Calls @c visitor.accept().
     * Classes implementing this interface implement this method to call @c visitor.accept()
     * with @c *this as a parameter.
     * 
     * @param visitor visitor on which to call the @c accept method
     */
    virtual void accept (Visitor& visitor) const = 0;

    /**
     * Find out cardinality of this operation.
     *
     * @return cardinality of operation (unary - 1, binary - 2, etc.)
     */
    virtual int cardinality() const = 0;
};


/**
 * Interface for all unary operations.
 */
struct UnaryOperation : virtual Operation
{
    /**
     * Returns the operand of this operation.
     * 
     * @return the operand of this operation
     */
    virtual const Operand* getOperand () const = 0;
};


/**
 * Interface for all binary operations.
 */
struct BinaryOperation : virtual Operation 
{
    /**
     * Returns the first operand of this operation.
     * 
     * @return the first operand of this operation
     */
    virtual const Operand* getOperand1 () const = 0;
    
    /**
     * Returns the second operand of this operation.
     * 
     * @return the second operand of this operation
     */
    virtual const Operand* getOperand2 () const = 0;
};

}

#endif /* GIT_OPERATION_H */

