// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file Operations.h
 * Header with all class definitions for operations.
 */

#ifndef GIT_OPERATIONS_H
#define GIT_OPERATIONS_H

#include "Operation.h"
#include "Operand.h"

#include <boost/serialization/serialization.hpp>
#include <memory>

namespace giterator { namespace detail {
    
/**
 * Base class for all operations.
 */
class AbstractOperation : virtual public Operation {
  protected:
    /*! return type of the operation */
    const Type* m_resultType;
    
  public:
    /**
	 * Constructor, creates a new abstract operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
    AbstractOperation() {}
    /**
     * Constructor, creates a new abstract operation.
     * 
     * @param resultType type of the result of this operation
     */
    AbstractOperation (const Type* resultType) : m_resultType (resultType) {}

    friend class boost::serialization::access;
	template<class Archive>
	void save(Archive & ar, const unsigned int) const
	{
		ar << m_resultType;
	}

	template<class Archive>
	void load(Archive & ar, const unsigned int)
	{
		Type* resultType;
		ar >> resultType;
		m_resultType = (const Type*) resultType;
	}

	template<class Archive>
	void serialize(Archive & ar, const unsigned int version)
	{
		boost::serialization::split_member(ar, *this, version);
	}
};
    
    
/**
 * Base class for all unary operations.
 */
class AbstractUnaryOperation : public AbstractOperation, public UnaryOperation {
  private:
	/*! Operand of this unary operation */
    std::auto_ptr<Operand> m_op;
    
  public:
    /**
	 * Constructor, creates a new abstract unary operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
    AbstractUnaryOperation() : AbstractOperation() {}
    /**
	 * Constructor, creates a new abstract unary operation.
	 *
	 * @param resultType type of the result of this operation
	 * @param op operand of the operation
	 */
    AbstractUnaryOperation (const Type* resultType, std::auto_ptr<Operand> op) 
        : AbstractOperation (resultType), m_op (op) {}
    
    /**
     * Obtain an operand of this unary operation.
     *
     * @return operand of this operation
     */
    const Operand* getOperand () const
    { return m_op.get (); }

    friend class boost::serialization::access;
	template<class Archive>
	void save(Archive & ar, const unsigned int) const
	{
		ar << boost::serialization::base_object<AbstractOperation>(*this);
		Operand* op = m_op.get();
		ar << op;
	}

	template<class Archive>
	void load(Archive & ar, const unsigned int)
	{
		ar >> boost::serialization::base_object<AbstractOperation>(*this);
		Operand* op;
		ar >> op;
		m_op = std::auto_ptr<Operand> (op);
	}

	template<class Archive>
	void serialize(Archive & ar, const unsigned int version)
	{
		boost::serialization::split_member(ar, *this, version);
	}
};


/**
 * Base class for all binary operations.
 */
class AbstractBinaryOperation : public AbstractOperation, public BinaryOperation {
  private:
	/*! First operand of this binary operation */
    std::auto_ptr<Operand> m_op1;
    /*! Second operand of this binary operation */
    std::auto_ptr<Operand> m_op2;
  public:
    /**
	 * Constructor, creates a new abstract binary operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
    AbstractBinaryOperation () : AbstractOperation() {}
    /**
	 * Constructor, creates a new abstract binary operation.
	 *
	 * @param resultType type of the result of this operation
	 * @param op1 first operand of this operation
	 * @param op2 second operand of this operation
	 */
    AbstractBinaryOperation (const Type* resultType, 
                     std::auto_ptr<Operand> op1, std::auto_ptr<Operand> op2) 
        : AbstractOperation (resultType), m_op1 (op1), m_op2 (op2) {}
    
    /**
     * Obtain the first operand of this operation.
     *
     * @return the first operand of this operation
     */
    const Operand* getOperand1 () const
    { return m_op1.get (); }
    
    /**
	 * Obtain the second operand of this operation.
	 *
	 * @return the second operand of this operation
	 */
    const Operand* getOperand2 () const
    { return m_op2.get (); }

    friend class boost::serialization::access;
	template<class Archive>
	void save(Archive & ar, const unsigned int) const
	{
		ar << boost::serialization::base_object<AbstractOperation>(*this);
		Operand* op1 = m_op1.get();
		ar << op1;
		Operand* op2 = m_op2.get();
		ar << op2;
	}

	template<class Archive>
	void load(Archive & ar, const unsigned int)
	{
		ar >> boost::serialization::base_object<AbstractOperation>(*this);
		Operand* op1;
		ar >> op1;
		m_op1 = std::auto_ptr<Operand> (op1);
		Operand* op2;
		ar >> op2;
		m_op2 = std::auto_ptr<Operand> (op2);
	}

	template<class Archive>
	void serialize(Archive & ar, const unsigned int version)
	{
		boost::serialization::split_member(ar, *this, version);
	}
};


/**
 * Helper class adding cardinality function to the unary operation.
 * It also initializes serialization.
 */
class UnaryOperationImpl : public AbstractUnaryOperation {
  public:
	/**
	 * Constructor, creates a new unary operation implementation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	UnaryOperationImpl () : AbstractUnaryOperation() {}
    /**
     * Constructor, creates a new unary operation implementation.
     * 
     * @param resultType type of the result of the operation
     * @param op operand of the operation
     */
    UnaryOperationImpl (const Type* resultType, 
                         std::auto_ptr<Operand> op)
    : AbstractUnaryOperation (resultType, op) {}
    
    /**
     * Get a type of the result of this operation
     *
     * @return type of the result of this operation
     */
    const Type* getResultType () const
	{ return m_resultType; }

    /**
     * Get a cardinality of this operation (i.e. 1 for unary operation).
     *
     * @return cardinality of this operation
     */
    int cardinality() const
	{ return 1; }

    template<class Archive>
	void serialize(Archive & ar, const unsigned int)
	{
		boost::serialization::void_cast_register(static_cast<UnaryOperationImpl *>(NULL), static_cast<Operation *>(NULL));
		boost::serialization::void_cast_register(static_cast<NegateOperation *>(NULL), static_cast<Operation *>(NULL));
		boost::serialization::void_cast_register(static_cast<AbsOperation *>(NULL), static_cast<Operation *>(NULL));
		boost::serialization::void_cast_register(static_cast<FixTruncOperation *>(NULL), static_cast<Operation *>(NULL));
		boost::serialization::void_cast_register(static_cast<FloatOperation *>(NULL), static_cast<Operation *>(NULL));
		boost::serialization::void_cast_register(static_cast<BitNotOperation *>(NULL), static_cast<Operation *>(NULL));
		boost::serialization::void_cast_register(static_cast<TruthNotOperation *>(NULL), static_cast<Operation *>(NULL));
		boost::serialization::void_cast_register(static_cast<ParenOperation *>(NULL), static_cast<Operation *>(NULL));
		boost::serialization::void_cast_register(static_cast<NonLvalueOperation *>(NULL), static_cast<Operation *>(NULL));
		boost::serialization::void_cast_register(static_cast<NoOperation *>(NULL), static_cast<Operation *>(NULL));
		boost::serialization::void_cast_register(static_cast<ConvertOperation *>(NULL), static_cast<Operation *>(NULL));
		boost::serialization::void_cast_register(static_cast<FloatOperation *>(NULL), static_cast<Operation *>(NULL));
		ar & boost::serialization::base_object<AbstractUnaryOperation>(*this);
	}
};


/**
 * Helper class adding cardinality function to the binary operation.
 * It also initializes serialization.
 */
class BinaryOperationImpl : public AbstractBinaryOperation {
  public:
	/**
	 * Constructor, creates a new binary operation implementation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	BinaryOperationImpl () : AbstractBinaryOperation() {}

    /**
     * Constructor, creates a new binary operation implementation.
     * 
     * @param resultType type of the result of the operation
     * @param op1 first operand of the operation
     * @param op2 second operand of the operation
     */
    BinaryOperationImpl (const Type* resultType, 
                         std::auto_ptr<Operand> op1, std::auto_ptr<Operand> op2)
    : AbstractBinaryOperation (resultType, op1, op2) {}
    
    /**
	 * Get the type of the result of this operation
	 *
	 * @return type of the result of this operation
	 */
    const Type* getResultType () const
	{ return m_resultType; }

    /**
	 * Get a cardinality of this operation (i.e. 1 for unary operation).
	 *
	 * @return cardinality of this operation
	 */
    int cardinality() const
	{ return 2; }

    template<class Archive>
	void serialize(Archive & ar, const unsigned int)
	{
		boost::serialization::void_cast_register(static_cast<BinaryOperationImpl *>(NULL), static_cast<Operation *>(NULL));
		boost::serialization::void_cast_register(static_cast<PlusOperation *>(NULL), static_cast<Operation *>(NULL));
		boost::serialization::void_cast_register(static_cast<MinusOperation *>(NULL), static_cast<Operation *>(NULL));
		boost::serialization::void_cast_register(static_cast<MultOperation *>(NULL), static_cast<Operation *>(NULL));
		boost::serialization::void_cast_register(static_cast<RdivOperation *>(NULL), static_cast<Operation *>(NULL));
		boost::serialization::void_cast_register(static_cast<TruncDivOperation *>(NULL), static_cast<Operation *>(NULL));
		boost::serialization::void_cast_register(static_cast<FloorDivOperation *>(NULL), static_cast<Operation *>(NULL));
		boost::serialization::void_cast_register(static_cast<CeilDivOperation *>(NULL), static_cast<Operation *>(NULL));
		boost::serialization::void_cast_register(static_cast<RoundDivOperation *>(NULL), static_cast<Operation *>(NULL));
		boost::serialization::void_cast_register(static_cast<TruncModOperation *>(NULL), static_cast<Operation *>(NULL));
		boost::serialization::void_cast_register(static_cast<FloorModOperation *>(NULL), static_cast<Operation *>(NULL));
		boost::serialization::void_cast_register(static_cast<CeilModOperation *>(NULL), static_cast<Operation *>(NULL));
		boost::serialization::void_cast_register(static_cast<RoundModOperation *>(NULL), static_cast<Operation *>(NULL));
		boost::serialization::void_cast_register(static_cast<ExactDivOperation *>(NULL), static_cast<Operation *>(NULL));
		boost::serialization::void_cast_register(static_cast<MinOperation *>(NULL), static_cast<Operation *>(NULL));
		boost::serialization::void_cast_register(static_cast<MaxOperation *>(NULL), static_cast<Operation *>(NULL));
		boost::serialization::void_cast_register(static_cast<LshiftOperation *>(NULL), static_cast<Operation *>(NULL));
		boost::serialization::void_cast_register(static_cast<RshiftOperation *>(NULL), static_cast<Operation *>(NULL));
		boost::serialization::void_cast_register(static_cast<LrotateOperation *>(NULL), static_cast<Operation *>(NULL));
		boost::serialization::void_cast_register(static_cast<RrotateOperation *>(NULL), static_cast<Operation *>(NULL));
		boost::serialization::void_cast_register(static_cast<BitIorOperation *>(NULL), static_cast<Operation *>(NULL));
		boost::serialization::void_cast_register(static_cast<BitXorOperation *>(NULL), static_cast<Operation *>(NULL));
		boost::serialization::void_cast_register(static_cast<BitAndOperation *>(NULL), static_cast<Operation *>(NULL));
		boost::serialization::void_cast_register(static_cast<TruthAndOperation *>(NULL), static_cast<Operation *>(NULL));
		boost::serialization::void_cast_register(static_cast<TruthOrOperation *>(NULL), static_cast<Operation *>(NULL));
		boost::serialization::void_cast_register(static_cast<TruthXorOperation *>(NULL), static_cast<Operation *>(NULL));
		boost::serialization::void_cast_register(static_cast<LtOperation *>(NULL), static_cast<Operation *>(NULL));
		boost::serialization::void_cast_register(static_cast<LeOperation *>(NULL), static_cast<Operation *>(NULL));
		boost::serialization::void_cast_register(static_cast<GtOperation *>(NULL), static_cast<Operation *>(NULL));
		boost::serialization::void_cast_register(static_cast<GeOperation *>(NULL), static_cast<Operation *>(NULL));
		boost::serialization::void_cast_register(static_cast<EqOperation *>(NULL), static_cast<Operation *>(NULL));
		boost::serialization::void_cast_register(static_cast<NeOperation *>(NULL), static_cast<Operation *>(NULL));
		boost::serialization::void_cast_register(static_cast<OrderedOperation *>(NULL), static_cast<Operation *>(NULL));
		boost::serialization::void_cast_register(static_cast<UnorderedOperation *>(NULL), static_cast<Operation *>(NULL));
		boost::serialization::void_cast_register(static_cast<UnLtOperation *>(NULL), static_cast<Operation *>(NULL));
		boost::serialization::void_cast_register(static_cast<UnLeOperation *>(NULL), static_cast<Operation *>(NULL));
		boost::serialization::void_cast_register(static_cast<UnGtOperation *>(NULL), static_cast<Operation *>(NULL));
		boost::serialization::void_cast_register(static_cast<UnGeOperation *>(NULL), static_cast<Operation *>(NULL));
		boost::serialization::void_cast_register(static_cast<UnEqOperation *>(NULL), static_cast<Operation *>(NULL));
		boost::serialization::void_cast_register(static_cast<LtGtOperation *>(NULL), static_cast<Operation *>(NULL));
		boost::serialization::void_cast_register(static_cast<WidenSumOperation *>(NULL), static_cast<Operation *>(NULL));
		boost::serialization::void_cast_register(static_cast<WidenMultOperation *>(NULL), static_cast<Operation *>(NULL));
		ar & boost::serialization::base_object<AbstractBinaryOperation>(*this);
	}
};

}

/**
 * Class representing GIMPLE @c NEGATE_EXPR.
 * It is an unary negation of the single operand, for both integer and floating-point types.
 * The type of negation can be determined by looking at the type of the expression.
 */
class NegateOperation : public detail::UnaryOperationImpl
{
public:
	/**
	 * Constructor, creates a new negation operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	NegateOperation() {}

	/**
	 * Constructor, creates a new negation operation.
     *
     * @param resultType type of the result of the operation
     * @param op operand of the operation
     */
	NegateOperation(const Type* resultType, std::auto_ptr<Operand> op)
	: detail::UnaryOperationImpl (resultType, op) {}

	void accept (Visitor& visitor) const
	{ visitor.visit (*this); }

};

/**
 * Class representing GIMPLE @c ABS_EXPR.
 * ABS_EXPR represents the absolute value of the single operand,
 * for both integer and floating-point types.
 * Inside GCC, it is typically used to implement the abs, labs and llabs builtins for integer types,
 * and the fabs, fabsf and fabsl builtins for floating point types.
 * The type of abs operation can be determined by looking at the type of the expression.
 */
class AbsOperation : public detail::UnaryOperationImpl
{
public:
	/**
	 * Constructor, creates a new absolute value operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	AbsOperation() {}

	/*
	 * Constructor, creates a new absolute value operation.
	 *
     * @param resultType type of the result of the operation
     * @param op operand of the operation
     */
	AbsOperation(const Type* resultType, std::auto_ptr<Operand> op)
	: detail::UnaryOperationImpl (resultType, op) {}

	void accept (Visitor& visitor) const
	    { visitor.visit (*this); }

};

/**
 * Class representing GIMPLE @c FIX_TRUNC_EXPR.
 * It represents conversion of a floating-point value to an integer.
 * The single operand will have a floating-point type,
 * while the complete expression will have an integral (or boolean) type.
 * The operand is rounded towards zero.
 */
class FixTruncOperation : public detail::UnaryOperationImpl
{
public:
	/**
	 * Constructor, creates a new fix trunc operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	FixTruncOperation() {}

	/**
	 * Constructor, creates a new fix trunc operation.
	 *
     * @param resultType type of the result of the operation
     * @param op operand of the operation
     */
	FixTruncOperation(const Type* resultType, std::auto_ptr<Operand> op)
	: detail::UnaryOperationImpl (resultType, op) {}

	void accept (Visitor& visitor) const
	{ visitor.visit (*this); }

};

/**
 * Class representing GIMPLE @c FLOAT_EXPR.
 * It represents conversion of an integral (or boolean) value to a floating-point value.
 * The single operand will have integral type,
 * while the complete expression will have a floating-point type.
 */
class FloatOperation : public detail::UnaryOperationImpl
{
public:
	/**
	 * Constructor, creates a new float operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	FloatOperation() {}

	/**
	 * Constructor, creates a new float operation.
	 *
     * @param resultType type of the result of the operation
     * @param op operand of the operation
     */
	FloatOperation(const Type* resultType, std::auto_ptr<Operand> op)
	: detail::UnaryOperationImpl (resultType, op) {}

	void accept (Visitor& visitor) const
	{ visitor.visit (*this); }

};

/**
 * Class representing GIMPLE @c BIT_NOT_EXPR.
 * It represents bitwise complement, and will always have integral type.
 * The only operand is the value to be complemented.
 */
class BitNotOperation : public detail::UnaryOperationImpl
{
public:
	/**
	 * Constructor, creates a new bit not operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	BitNotOperation() {}

	/**
	 * Constructor, creates a new bit not operation.
	 *
     * @param resultType type of the result of the operation
     * @param op operand of the operation
     */
	BitNotOperation(const Type* resultType, std::auto_ptr<Operand> op)
	: detail::UnaryOperationImpl (resultType, op) {}

	void accept (Visitor& visitor) const
	{ visitor.visit (*this); }

};

/**
 * Class representing GIMPLE @c TRUTH_NOT_EXPR.
 * It represents logical negation, and will always have integral (or boolean) type.
 * The operand is the value being negated. The type of the operand and that of
 * the result are always of boolean or integer type.
 */
class TruthNotOperation : public detail::UnaryOperationImpl
{
public:
	/**
	 * Constructor, creates a new truth not operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	TruthNotOperation() {}

	/**
	 * Constructor, creates a new truth not operation.
	 *
     * @param resultType type of the result of the operation
     * @param op operand of the operation
     */
	TruthNotOperation(const Type* resultType, std::auto_ptr<Operand> op)
	: detail::UnaryOperationImpl (resultType, op) {}

	void accept (Visitor& visitor) const
	{ visitor.visit (*this); }

};

/**
 * Class representing GIMPLE @c PAREN_EXPR.
 * It represents a re-association barrier for floating point expressions
 * like explicit parenthesis in fortran.
 * Currently, it is threated in GMC same as @c NoOperation.
 *
 * @see NoOperation
 */
class ParenOperation : public detail::UnaryOperationImpl
{
public:
	/**
	 * Constructor, creates a new parentheses operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	ParenOperation() {}

	/**
	 * Constructor, creates a new parentheses operation.
	 *
     * @param resultType type of the result of the operation
     * @param op operand of the operation
     */
	ParenOperation(const Type* resultType, std::auto_ptr<Operand> op)
	: detail::UnaryOperationImpl (resultType, op) {}

	void accept (Visitor& visitor) const
	{ visitor.visit (*this); }

};

/**
 * Class representing GIMPLE @c NON_LVALUE_EXPR.
 * It indicates that its one and only operand is not an lvalue.
 * In GCC, a back end can then treat these identically to the single operand.
 * In GMC, these nodes are threated as @c NoOperation.
 *
 * @see NoOperation
 */
class NonLvalueOperation : public detail::UnaryOperationImpl
{
public:
	/**
	 * Constructor, creates a new non lvalue operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	NonLvalueOperation() {}

	/**
	 * Constructor, creates a new non lvalue operation.
	 *
     * @param resultType type of the result of the operation
     * @param op operand of the operation
     */
	NonLvalueOperation(const Type* resultType, std::auto_ptr<Operand> op)
	: detail::UnaryOperationImpl (resultType, op) {}

	void accept (Visitor& visitor) const
	    { visitor.visit (*this); }

};

/**
 * Class representing GIMPLE @c NOP_EXPR.
 *
 * In GCC, these nodes are used to represent conversions that do not require any code-generation.
 * For example, conversion of a char* to an int* does not require any code be generated;
 * such a conversion is represented by a @c NOP_EXPR. The single operand is the expression to be converted.
 * The conversion from a pointer to a reference is also represented with a @c NOP_EXPR.
 *
 * GMC simply extracts the operand without and uses it as a result of the operation.
 */
class NoOperation : public detail::UnaryOperationImpl
{
public:
	/**
	 * Constructor, creates a new no operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	NoOperation() {}

	/**
	 * Constructor, creates a new no operation.
	 *
     * @param resultType type of the result of the operation
     * @param op operand of the operation
     */
	NoOperation(const Type* resultType, std::auto_ptr<Operand> op)
	: detail::UnaryOperationImpl (resultType, op) {}

	void accept (Visitor& visitor) const
	{ visitor.visit (*this); }

};

/**
 * Class representing GIMPLE @c CONVERT_EXPR.
 *
 * In GCC, these nodes are similar to NOP_EXPRs, but are used in those situations where code may
 * need to be generated. For example, if an int* is converted to an int code may need to be
 * generated on some platforms. These nodes are never used for C++-specific conversions,
 * like conversions between pointers to different classes in an inheritance hierarchy.
 * Any adjustments that need to be made in such cases are always indicated explicitly.
 * Similarly, a user-defined conversion is never represented by a CONVERT_EXPR;
 * instead, the function calls are made explicit.
 *
 * GMC uses this class for most of the types of conversions.
 */
class ConvertOperation : public detail::UnaryOperationImpl
{
public:
	/**
	 * Constructor, creates a new convert operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	ConvertOperation() {}

	/**
	 * Constructor, creates a new convert operation.
	 *
     * @param resultType type of the result of the operation
     * @param op operand of the operation
     */
	ConvertOperation(const Type* resultType, std::auto_ptr<Operand> op)
	: detail::UnaryOperationImpl (resultType, op) {}

	void accept (Visitor& visitor) const
	    { visitor.visit (*this); }

};

/**
 * Class representing GIMPLE @c PLUS_EXPR and @c POINTER_PLUS_EXPR.
 * In the first case, it represents binary arithmetic addition.
 * Its operands may have either integral or floating type,
 * but there will never be case in which one operand is of floating type
 * and the other is of integral type.
 * In the second case, it represents pointer arithmetic.
 * The first operand is always a pointer type and the second an unsigned integer.
 */
class PlusOperation : public detail::BinaryOperationImpl
{
public:
	/**
	 * Constructor, creates a new plus operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	PlusOperation() {}

	/**
	 * Constructor, creates a new plus operation.
	 *
	 * @param resultType type of the result of the operation
	 * @param op1 first operand of the operation
	 * @param op2 second operand of the operation
	 */
	PlusOperation(const Type* resultType, std::auto_ptr<Operand> op1, std::auto_ptr<Operand> op2)
	: detail::BinaryOperationImpl (resultType, op1, op2) {}

	void accept (Visitor& visitor) const
	    { visitor.visit (*this); }
};

/**
 * Class representing GIMPLE @c MINUS_EXPR.
 * It represents binary arithmetic subtraction.
 * Its operands may have either integral or floating type,
 * but there will never be case in which one operand is of floating type
 * and the other is of integral type.
 */
class MinusOperation : public detail::BinaryOperationImpl
{
public:
	/**
	 * Constructor, creates a new minus operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	MinusOperation() {}

	/**
	 * Constructor, creates a new minus operation.
	 *
	 * @param resultType type of the result of the operation
	 * @param op1 first operand of the operation
	 * @param op2 second operand of the operation
	 */
	MinusOperation(const Type* resultType, std::auto_ptr<Operand> op1, std::auto_ptr<Operand> op2)
	: detail::BinaryOperationImpl (resultType, op1, op2) {}

	void accept (Visitor& visitor) const
	{ visitor.visit (*this); }
};

/**
 * Class representing GIMPLE @c MULT_EXPR.
 * It represents binary arithmetic multiplication.
 * Its operands may have either integral or floating type,
 * but there will never be case in which one operand is of floating type
 * and the other is of integral type.
 */
class MultOperation : public detail::BinaryOperationImpl
{
public:
	/**
	 * Constructor, creates a new mult operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	MultOperation() {}

	/**
	 * Constructor, creates a new mult operation.
	 *
	 * @param resultType type of the result of the operation
	 * @param op1 first operand of the operation
	 * @param op2 second operand of the operation
	 */
	MultOperation(const Type* resultType, std::auto_ptr<Operand> op1, std::auto_ptr<Operand> op2)
	: detail::BinaryOperationImpl (resultType, op1, op2) {}

	void accept (Visitor& visitor) const
	{ visitor.visit (*this); }
};

/**
 * Class representing GIMPLE @c RDIV_EXPR.
 * It represents a floating point division operation - division for real result.
 */
class RdivOperation : public detail::BinaryOperationImpl
{
public:
	/**
	 * Constructor, creates a new rdiv operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	RdivOperation() {}

	/**
	 * Constructor, creates a new rdiv operation.
	 *
	 * @param resultType type of the result of the operation
	 * @param op1 first operand of the operation
	 * @param op2 second operand of the operation
	 */
	RdivOperation(const Type* resultType, std::auto_ptr<Operand> op1, std::auto_ptr<Operand> op2)
	: detail::BinaryOperationImpl (resultType, op1, op2) {}

	void accept (Visitor& visitor) const
	{ visitor.visit (*this); }
};

/**
 * Class representing GIMPLE @c TRUNC_DIV_EXPR.
 * It represents integer division operation that returns an integer result.
 * This operation rounds towards zero.
 * Integer division in C and C++ is truncating, i.e. @c TRUNC_DIV_EXPR.
 */
class TruncDivOperation : public detail::BinaryOperationImpl
{
public:
	/**
	 * Constructor, creates a new trunc div operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	TruncDivOperation() {}

	/**
	 * Constructor, creates a new trunc div operation.
	 *
	 * @param resultType type of the result of the operation
	 * @param op1 first operand of the operation
	 * @param op2 second operand of the operation
	 */
	TruncDivOperation(const Type* resultType, std::auto_ptr<Operand> op1, std::auto_ptr<Operand> op2)
	: detail::BinaryOperationImpl (resultType, op1, op2) {}

	void accept (Visitor& visitor) const
	    { visitor.visit (*this); }
};

/**
 * Class representing GIMPLE @c FLOOR_DIV_EXPR
 * It represents integer division operation that returns an integer result.
 * This operation rounds towards negative infinity.
 */
class FloorDivOperation : public detail::BinaryOperationImpl
{
public:
	/**
	 * Constructor, creates a new floor div operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	FloorDivOperation() {}

	/**
	 * Constructor, creates a new floor div operation.
	 *
	 * @param resultType type of the result of the operation
	 * @param op1 first operand of the operation
	 * @param op2 second operand of the operation
	 */
	FloorDivOperation(const Type* resultType, std::auto_ptr<Operand> op1, std::auto_ptr<Operand> op2)
	: detail::BinaryOperationImpl (resultType, op1, op2) {}

	void accept (Visitor& visitor) const
	{ visitor.visit (*this); }
};

/**
 * Class representing GIMPLE @c CEIL_DIV_EXPR.
 * It represents integer division operation that returns an integer result.
 * This operation rounds towards positive infinity.
 */
class CeilDivOperation : public detail::BinaryOperationImpl
{
public:
	/**
	 * Constructor, creates a new ceil div operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	CeilDivOperation() {}

	/**
	 * Constructor, creates a new ceil div operation.
	 *
	 * @param resultType type of the result of the operation
	 * @param op1 first operand of the operation
	 * @param op2 second operand of the operation
	 */
	CeilDivOperation(const Type* resultType, std::auto_ptr<Operand> op1, std::auto_ptr<Operand> op2)
	: detail::BinaryOperationImpl (resultType, op1, op2) {}

	void accept (Visitor& visitor) const
	{ visitor.visit (*this); }
};

/**
 * Class representing GIMPLE @c ROUND_DIV_EXPR
 * It represents integer division operation that returns an integer result.
 * This operation rounds to the closest integer.
 */
class RoundDivOperation : public detail::BinaryOperationImpl
{
public:
	/**
	 * Constructor, creates a new round div operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	RoundDivOperation() {}

	/**
	 * Constructor, creates a new round div operation.
	 *
	 * @param resultType type of the result of the operation
	 * @param op1 first operand of the operation
	 * @param op2 second operand of the operation
	 */
	RoundDivOperation(const Type* resultType, std::auto_ptr<Operand> op1, std::auto_ptr<Operand> op2)
	: detail::BinaryOperationImpl (resultType, op1, op2) {}

	void accept (Visitor& visitor) const
	{ visitor.visit (*this); }
};

/**
 * Class representing GIMPLE @c TRUNC_MOD_EXPR.
 * It represents the integer remainder or modulus operation. The integer modulus of
 * two operands a and b is defined as a - (a/b)*b where the division in this case is
 * @c TruncDivOperation - division rounding towards zero.
 * Integer remainder in C and C++ uses truncating division, i.e. @c TRUNC_MOD_EXPR.
 *
 * @see TruncDivOperation
 */
class TruncModOperation : public detail::BinaryOperationImpl
{
public:
	/**
	 * Constructor, creates a new trunc mod operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	TruncModOperation() {}

	/**
	 * Constructor, creates a new trunc mod operation.
	 *
	 * @param resultType type of the result of the operation
	 * @param op1 first operand of the operation
	 * @param op2 second operand of the operation
	 */
	TruncModOperation(const Type* resultType, std::auto_ptr<Operand> op1, std::auto_ptr<Operand> op2)
	: detail::BinaryOperationImpl (resultType, op1, op2) {}

	void accept (Visitor& visitor) const
	    { visitor.visit (*this); }
};

/**
 * Class representing GIMPLE @c FLOOR_MOD_EXPR.
 * It represents the integer remainder or modulus operation. The integer modulus of
 * two operands a and b is defined as a - (a/b)*b where the division in this case is
 * @c FloorDivOperation - division rounding towards negative infinity.
 *
 * @see FloorDivOperation
 */
class FloorModOperation : public detail::BinaryOperationImpl
{
public:
	/**
	 * Constructor, creates a new floor mod operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	FloorModOperation() {}

	/**
	 * Constructor, creates a new floor mod operation.
	 *
	 * @param resultType type of the result of the operation
	 * @param op1 first operand of the operation
	 * @param op2 second operand of the operation
	 */
	FloorModOperation(const Type* resultType, std::auto_ptr<Operand> op1, std::auto_ptr<Operand> op2)
	: detail::BinaryOperationImpl (resultType, op1, op2) {}

	void accept (Visitor& visitor) const
	    { visitor.visit (*this); }
};

/**
 * Class representing GIMPLE @c CEIL_MOD_EXPR.
 * It represents the integer remainder or modulus operation. The integer modulus of
 * two operands a and b is defined as a - (a/b)*b where the division in this case is
 * @c CeilDivOperation - division rounding towards positive infinity.
 *
 * @see CeilDivOperation
 */
class CeilModOperation : public detail::BinaryOperationImpl
{
public:
	/**
	 * Constructor, creates a new ceil mod operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	CeilModOperation() {}

	/**
	 * Constructor, creates a new ceil mod operation.
	 *
	 * @param resultType type of the result of the operation
	 * @param op1 first operand of the operation
	 * @param op2 second operand of the operation
	 */
	CeilModOperation(const Type* resultType, std::auto_ptr<Operand> op1, std::auto_ptr<Operand> op2)
	: detail::BinaryOperationImpl (resultType, op1, op2) {}

	void accept (Visitor& visitor) const
	    { visitor.visit (*this); }
};

/**
 * Class representing GIMPLE @c ROUND_MOD_EXPR.
 * It represents the integer remainder or modulus operation. The integer modulus of
 * two operands a and b is defined as a - (a/b)*b where the division in this case is
 * @c RoundDivOperation - division rounding to the nearest integer.
 *
 * @see RoundDivOperation
 */
class RoundModOperation : public detail::BinaryOperationImpl
{
public:
	/**
	 * Constructor, creates a new round mod operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	RoundModOperation() {}

	/**
	 * Constructor, creates a new round mod operation.
	 *
	 * @param resultType type of the result of the operation
	 * @param op1 first operand of the operation
	 * @param op2 second operand of the operation
	 */
	RoundModOperation(const Type* resultType, std::auto_ptr<Operand> op1, std::auto_ptr<Operand> op2)
	: detail::BinaryOperationImpl (resultType, op1, op2) {}

	void accept (Visitor& visitor) const
	    { visitor.visit (*this); }
};

/**
 * Class representing GIMPLE @c EXACT_DIV_EXPR.
 *
 * In GCC, @c EXACT_DIV_EXPR is used to represent integer divisions where the numerator
 * is known to be an exact multiple of the denominator. This allows the backend to
 * choose between the faster of @c TRUNC_DIV_EXPR, @c CEIL_DIV_EXPR and @c FLOOR_DIV_EXPR
 * for the current target.
 *
 * In GMC, @c TruncDivOperation is used to implement this operation.
 */
class ExactDivOperation : public detail::BinaryOperationImpl
{
public:
	/**
	 * Constructor, creates a new exact div operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	ExactDivOperation() {}

	/**
	 * Constructor, creates a new exact div operation.
	 *
	 * @param resultType type of the result of the operation
	 * @param op1 first operand of the operation
	 * @param op2 second operand of the operation
	 */
	ExactDivOperation(const Type* resultType, std::auto_ptr<Operand> op1, std::auto_ptr<Operand> op2)
	: detail::BinaryOperationImpl (resultType, op1, op2) {}

	void accept (Visitor& visitor) const
	    { visitor.visit (*this); }
};

/**
 * Class representing GIMPLE @c MIN_EXPR.
 *
 * It represents minimum value of the two operands.
 * In GCC, When used with floating point, if both operands are zeros,
 * or if either operand is NaN, then it is unspecified
 * which of the two operands is returned as the result.
 * In GMC, the second operand is returned in this case.
 */
class MinOperation : public detail::BinaryOperationImpl
{
public:
	/**
	 * Constructor, creates a new min operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	MinOperation() {}

	/**
	 * Constructor, creates a new min operation.
	 *
	 * @param resultType type of the result of the operation
	 * @param op1 first operand of the operation
	 * @param op2 second operand of the operation
	 */
	MinOperation(const Type* resultType, std::auto_ptr<Operand> op1, std::auto_ptr<Operand> op2)
	: detail::BinaryOperationImpl (resultType, op1, op2) {}

	void accept (Visitor& visitor) const
	{ visitor.visit (*this); }
};

/**
 * Class representing GIMPLE @c MAX_EXPR.
 * It represents miaximum value of the two operands.
 * In GCC, When used with floating point, if both operands are zeros,
 * or if either operand is NaN, then it is unspecified
 * which of the two operands is returned as the result.
 * In GMC, the second operand is returned in this case.
 */
class MaxOperation : public detail::BinaryOperationImpl
{
public:
	/**
	 * Constructor, creates a new max operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	MaxOperation() {}

	/**
	 * Constructor, creates a new max operation.
	 *
	 * @param resultType type of the result of the operation
	 * @param op1 first operand of the operation
	 * @param op2 second operand of the operation
	 */
	MaxOperation(const Type* resultType, std::auto_ptr<Operand> op1, std::auto_ptr<Operand> op2)
	: detail::BinaryOperationImpl (resultType, op1, op2) {}

	void accept (Visitor& visitor) const
	{ visitor.visit (*this); }
};

/**
 * Class representing GIMPLE @c LSHIFT_EXPR.
 * It represents left shift. The first operand is the value to shift;
 * it is always of integral type. The second operand is the number
 * of bits by which to shift.
 */
class LshiftOperation : public detail::BinaryOperationImpl
{
public:
	/**
	 * Constructor, creates a new left shift operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	LshiftOperation() {}

	/**
	 * Constructor, creates a new left shift operation.
	 *
	 * @param resultType type of the result of the operation
	 * @param op1 first operand of the operation
	 * @param op2 second operand of the operation
	 */
	LshiftOperation(const Type* resultType, std::auto_ptr<Operand> op1, std::auto_ptr<Operand> op2)
	: detail::BinaryOperationImpl (resultType, op1, op2) {}

	void accept (Visitor& visitor) const
	{ visitor.visit (*this); }
};

/**
 * Class representing GIMPLE @c RSHIFT_EXPR.
 * It represents right shift. The first operand is the value to shift;
 * it is always of integral type. The second operand is the number
 * of bits by which to shift.
 * Right shift is be treated as arithmetic, i.e., the high-order bits should be zero-filled when
 * the expression has unsigned type and filled with the
 * sign bit when the expression has signed type.
 */
class RshiftOperation : public detail::BinaryOperationImpl
{
public:
	/**
	 * Constructor, creates a new right shift operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	RshiftOperation() {}

	/**
	 * Constructor, creates a new right shift operation.
	 *
	 * @param resultType type of the result of the operation
	 * @param op1 first operand of the operation
	 * @param op2 second operand of the operation
	 */
	RshiftOperation(const Type* resultType, std::auto_ptr<Operand> op1, std::auto_ptr<Operand> op2)
	: detail::BinaryOperationImpl (resultType, op1, op2) {}

	void accept (Visitor& visitor) const
	{ visitor.visit (*this); }
};

/**
 * Class representing GIMPLE @c LROTATE_EXPR
 * It represents left rotation. It is defined for integer types only.
 * The first operand is the value to rotate;
 * it is always of integral type. The second operand is the number
 * of bits by which to rotate.
 */
class LrotateOperation : public detail::BinaryOperationImpl
{
public:
	/**
	 * Constructor, creates a new left rotate operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	LrotateOperation() {}

	/**
	 * Constructor, creates a new left rotate operation.
	 *
	 * @param resultType type of the result of the operation
	 * @param op1 first operand of the operation
	 * @param op2 second operand of the operation
	 */
	LrotateOperation(const Type* resultType, std::auto_ptr<Operand> op1, std::auto_ptr<Operand> op2)
	: detail::BinaryOperationImpl (resultType, op1, op2) {}

	void accept (Visitor& visitor) const
	{ visitor.visit (*this); }
};

/**
 * Class representing GIMPLE @c RROTATE_EXPR
 * It represents right rotation. It is defined for integer types only.
 * The first operand is the value to rotate;
 * it is always of integral type. The second operand is the number
 * of bits by which to rotate.
 */
class RrotateOperation : public detail::BinaryOperationImpl
{
public:
	/**
	 * Constructor, creates a new right rotate operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	RrotateOperation() {}

	/**
	 * Constructor, creates a new right rotate operation.
	 *
	 * @param resultType type of the result of the operation
	 * @param op1 first operand of the operation
	 * @param op2 second operand of the operation
	 */
	RrotateOperation(const Type* resultType, std::auto_ptr<Operand> op1, std::auto_ptr<Operand> op2)
	: detail::BinaryOperationImpl (resultType, op1, op2) {}

	void accept (Visitor& visitor) const
	{ visitor.visit (*this); }
};

/**
 * Class representing GIMPLE @c BIT_IOR_EXPR.
 * It represents bitwise inclusive or. Both operands do always have integral type.
 */
class BitIorOperation : public detail::BinaryOperationImpl
{
public:
	/**
	 * Constructor, creates a new bit ior operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	BitIorOperation() {}

	/**
	 * Constructor, creates a new bit ior operation.
	 *
	 * @param resultType type of the result of the operation
	 * @param op1 first operand of the operation
	 * @param op2 second operand of the operation
	 */
	BitIorOperation(const Type* resultType, std::auto_ptr<Operand> op1, std::auto_ptr<Operand> op2)
	: detail::BinaryOperationImpl (resultType, op1, op2) {}

	void accept (Visitor& visitor) const
	{ visitor.visit (*this); }
};

/**
 * Class representing GIMPLE @c BIT_XOR_EXPR.
 * It represents bitwise exclusive or. Both operands do always have integral type.
 */
class BitXorOperation : public detail::BinaryOperationImpl
{
public:
	/**
	 * Constructor, creates a new bit xor operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	BitXorOperation() {}

	/**
	 * Constructor, creates a new bit xor operation.
	 *
	 * @param resultType type of the result of the operation
	 * @param op1 first operand of the operation
	 * @param op2 second operand of the operation
	 */
	BitXorOperation(const Type* resultType, std::auto_ptr<Operand> op1, std::auto_ptr<Operand> op2)
	: detail::BinaryOperationImpl (resultType, op1, op2) {}

	void accept (Visitor& visitor) const
	{ visitor.visit (*this); }
};

/**
 * Class representing GIMPLE @c BIT_AND_EXPR.
 * It represents bitwise and. Both operands do always have integral type.
 */
class BitAndOperation : public detail::BinaryOperationImpl
{
public:
	/**
	 * Constructor, creates a new bit and operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	BitAndOperation() {}

	/**
	 * Constructor, creates a new bit and operation.
	 *
	 * @param resultType type of the result of the operation
	 * @param op1 first operand of the operation
	 * @param op2 second operand of the operation
	 */
	BitAndOperation(const Type* resultType, std::auto_ptr<Operand> op1, std::auto_ptr<Operand> op2)
	: detail::BinaryOperationImpl (resultType, op1, op2) {}

	void accept (Visitor& visitor) const
	{ visitor.visit (*this); }
};

/**
 * Class representing GIMPLE @c TRUTH_AND_EXPR.
 * It represents logical and. It is strict; both arguments are always evaluated.
 * There are no corresponding operators in C or C++, but the front end will sometimes
 * generate these expressions anyhow, if it can tell that strictness does not matter.
 * The type of the operands and that of the result are always of boolean or integer type.
 */
class TruthAndOperation : public detail::BinaryOperationImpl
{
public:
	/**
	 * Constructor, creates a new truth and operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	TruthAndOperation() {}

	/**
	 * Constructor, creates a new truth and operation.
	 *
	 * @param resultType type of the result of the operation
	 * @param op1 first operand of the operation
	 * @param op2 second operand of the operation
	 */
	TruthAndOperation(const Type* resultType, std::auto_ptr<Operand> op1, std::auto_ptr<Operand> op2)
	: detail::BinaryOperationImpl (resultType, op1, op2) {}

	void accept (Visitor& visitor) const
	{ visitor.visit (*this); }
};

/**
 * Class representing GIMPLE @c TRUTH_OR_EXPR.
 * It represents logical or. It is strict; both arguments are always evaluated.
 * There are no corresponding operators in C or C++, but the front end will sometimes
 * generate these expressions anyhow, if it can tell that strictness does not matter.
 * The type of the operands and that of the result are always of boolean or integer type.
 */
class TruthOrOperation : public detail::BinaryOperationImpl
{
public:
	/**
	 * Constructor, creates a new truth or operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	TruthOrOperation() {}

	/**
	 * Constructor, creates a new truth or operation.
	 *
	 * @param resultType type of the result of the operation
	 * @param op1 first operand of the operation
	 * @param op2 second operand of the operation
	 */
	TruthOrOperation(const Type* resultType, std::auto_ptr<Operand> op1, std::auto_ptr<Operand> op2)
	: detail::BinaryOperationImpl (resultType, op1, op2) {}

	void accept (Visitor& visitor) const
	{ visitor.visit (*this); }
};

/**
 * Class representing GIMPLE @c TRUTH_XOR_EXPR.
 * It represents logical exclusive or. It is strict; both arguments are always evaluated.
 * There are no corresponding operators in C or C++, but the front end will sometimes
 * generate these expressions anyhow, if it can tell that strictness does not matter.
 * The type of the operands and that of the result are always of boolean or integer type.
 */
class TruthXorOperation : public detail::BinaryOperationImpl
{
public:
	/**
	 * Constructor, creates a new truth xor operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	TruthXorOperation() {}

	/**
	 * Constructor, creates a new truth xor operation.
	 *
	 * @param resultType type of the result of the operation
	 * @param op1 first operand of the operation
	 * @param op2 second operand of the operation
	 */
	TruthXorOperation(const Type* resultType, std::auto_ptr<Operand> op1, std::auto_ptr<Operand> op2)
	: detail::BinaryOperationImpl (resultType, op1, op2) {}

	void accept (Visitor& visitor) const
	{ visitor.visit (*this); }
};

/**
 * Class representing GIMPLE @c LT_EXPR.
 * It represents the less than comparison operator. The first and second
 * operands are both of integral type or both of floating type.
 * The result type of this operation is of boolean type.
 *
 * For floating point comparisons, if we honor IEEE NaNs and either
 * operand is NaN, this operand returns false.
 */
class LtOperation : public detail::BinaryOperationImpl
{
public:
	/**
	 * Constructor, creates a new less than operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	LtOperation() {}

	/**
	 * Constructor, creates a new less than operation.
	 *
	 * @param resultType type of the result of the operation
	 * @param op1 first operand of the operation
	 * @param op2 second operand of the operation
	 */
	LtOperation(const Type* resultType, std::auto_ptr<Operand> op1, std::auto_ptr<Operand> op2)
	: detail::BinaryOperationImpl (resultType, op1, op2) {}

	void accept (Visitor& visitor) const
	{ visitor.visit (*this); }
};

/**
 * Class representing GIMPLE @c LE_EXPR.
 * It represents the less than or equal comparison operator. The first and second
 * operands are both of integral type or both of floating type.
 * The result type of this operation is of boolean type.
 *
 * For floating point comparisons, if we honor IEEE NaNs and either
 * operand is NaN, this operand returns false.
 */
class LeOperation : public detail::BinaryOperationImpl
{
public:
	/**
	 * Constructor, creates a new less than or equal operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	LeOperation() {}

	/**
	 * Constructor, creates a new less than or equal operation.
	 *
	 * @param resultType type of the result of the operation
	 * @param op1 first operand of the operation
	 * @param op2 second operand of the operation
	 */
	LeOperation(const Type* resultType, std::auto_ptr<Operand> op1, std::auto_ptr<Operand> op2)
	: detail::BinaryOperationImpl (resultType, op1, op2) {}

	void accept (Visitor& visitor) const
	{ visitor.visit (*this); }
};

/**
 * Class representing GIMPLE @c GT_EXPR.
 * It represents the greater than comparison operator. The first and second
 * operands are both of integral type or both of floating type.
 * The result type of this operation is of boolean type.
 *
 * For floating point comparisons, if we honor IEEE NaNs and either
 * operand is NaN, this operand returns false.
 */
class GtOperation : public detail::BinaryOperationImpl
{
public:
	/**
	 * Constructor, creates a new greater than operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	GtOperation() {}

	/**
	 * Constructor, creates a new greater than operation.
	 *
	 * @param resultType type of the result of the operation
	 * @param op1 first operand of the operation
	 * @param op2 second operand of the operation
	 */
	GtOperation(const Type* resultType, std::auto_ptr<Operand> op1, std::auto_ptr<Operand> op2)
	: detail::BinaryOperationImpl (resultType, op1, op2) {}

	void accept (Visitor& visitor) const
	{ visitor.visit (*this); }
};

/**
 * Class representing GIMPLE @c GE_EXPR
 * It represents the greater than or equal comparison operator. The first and second
 * operands are both of integral type or both of floating type.
 * The result type of this operation is of boolean type.
 *
 * For floating point comparisons, if we honor IEEE NaNs and either
 * operand is NaN, this operand returns false.
 */
class GeOperation : public detail::BinaryOperationImpl
{
public:
	/**
	 * Constructor, creates a new greater than or equal operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	GeOperation() {}

	/**
	 * Constructor, creates a new greater than or equal operation.
	 *
	 * @param resultType type of the result of the operation
	 * @param op1 first operand of the operation
	 * @param op2 second operand of the operation
	 */
	GeOperation(const Type* resultType, std::auto_ptr<Operand> op1, std::auto_ptr<Operand> op2)
	: detail::BinaryOperationImpl (resultType, op1, op2) {}

	void accept (Visitor& visitor) const
	{ visitor.visit (*this); }
};

/**
 * Class representing GIMPLE @c EQ_EXPR
 * It represents the equal operator. The first and second
 * operands are both of integral type or both of floating type.
 * The result type of this operation is of boolean type.
 *
 * For floating point comparisons, if we honor IEEE NaNs and either
 * operand is NaN, this operand returns false.
 */
class EqOperation : public detail::BinaryOperationImpl
{
public:
	/**
	 * Constructor, creates a new equal to operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	EqOperation() {}

	/**
	 * Constructor, creates a new equal to operation.
	 *
	 * @param resultType type of the result of the operation
	 * @param op1 first operand of the operation
	 * @param op2 second operand of the operation
	 */
	EqOperation(const Type* resultType, std::auto_ptr<Operand> op1, std::auto_ptr<Operand> op2)
	: detail::BinaryOperationImpl (resultType, op1, op2) {}

	void accept (Visitor& visitor) const
	{ visitor.visit (*this); }
};

/**
 * Class representing GIMPLE @c NE_EXPR
 * It represents the not equal operator. The first and second
 * operands are both of integral type or both of floating type.
 * The result type of this operation is of boolean type.
 *
 * For floating point comparisons, if we honor IEEE NaNs and either
 * operand is NaN, this operand returns true.
 */
class NeOperation : public detail::BinaryOperationImpl
{
public:
	/**
	 * Constructor, creates a new not equal to operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	NeOperation() {}

	/**
	 * Constructor, creates a new not equal to operation.
	 *
	 * @param resultType type of the result of the operation
	 * @param op1 first operand of the operation
	 * @param op2 second operand of the operation
	 */
	NeOperation(const Type* resultType, std::auto_ptr<Operand> op1, std::auto_ptr<Operand> op2)
	: detail::BinaryOperationImpl (resultType, op1, op2) {}

	void accept (Visitor& visitor) const
	{ visitor.visit (*this); }
};

/**
 * Class representing GIMPLE @c ORDERED_EXPR.
 * It represents non-trapping ordered comparison operators. It takes two
 * floating point operands and determine whether they are ordered relative
 * to each other. If either operand is an IEEE NaN, their comparison
 * is defined to be unordered, otherwise the comparison is defined to be ordered.
 * The result type of these expressions is of boolean type.
 *
 * @see UnorderedOperation
 */
class OrderedOperation : public detail::BinaryOperationImpl
{
public:
	/**
	 * Constructor, creates a new ordered operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	OrderedOperation() {}

	/**
	 * Constructor, creates a new ordered operation.
	 *
	 * @param resultType type of the result of the operation
	 * @param op1 first operand of the operation
	 * @param op2 second operand of the operation
	 */
	OrderedOperation(const Type* resultType, std::auto_ptr<Operand> op1, std::auto_ptr<Operand> op2)
	: detail::BinaryOperationImpl (resultType, op1, op2) {}

	void accept (Visitor& visitor) const
	{ visitor.visit (*this); }
};

/**
 * Class representing GIMPLE @c UNORDERED_EXPR.
 * It represents non-trapping unordered comparison operators. It takes two
 * floating point operands and determine whether they are unordered relative
 * to each other. If either operand is an IEEE NaN, their comparison
 * is defined to be unordered, otherwise the comparison is defined to be ordered.
 * The result type of these expressions is of boolean type.
 *
 * @see OrderedOperation
 */
class UnorderedOperation : public detail::BinaryOperationImpl
{
public:
	/**
	 * Constructor, creates a new unordered operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	UnorderedOperation() {}

	/**
	 * Constructor, creates a new unordered operation.
	 *
	 * @param resultType type of the result of the operation
	 * @param op1 first operand of the operation
	 * @param op2 second operand of the operation
	 */
	UnorderedOperation(const Type* resultType, std::auto_ptr<Operand> op1, std::auto_ptr<Operand> op2)
	: detail::BinaryOperationImpl (resultType, op1, op2) {}

	void accept (Visitor& visitor) const
	{ visitor.visit (*this); }
};

/**
 * Class representing GIMPLE @c UNLT_EXPR.
 * It represents the unordered or less than comparison operator.
 * It takes two floating point operands and determine whether the operands are
 * unordered or are less than.
 * @c UnLtOperation returns true if either operand is an IEEE NaN or the first operand
 * is less than the second.
 * The result type is boolean type.
 */
class UnLtOperation : public detail::BinaryOperationImpl
{
public:
	/**
	 * Constructor, creates a new unordered less than operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	UnLtOperation() {}

	/**
	 * Constructor, creates a new unordered less than operation.
	 *
	 * @param resultType type of the result of the operation
	 * @param op1 first operand of the operation
	 * @param op2 second operand of the operation
	 */
	UnLtOperation(const Type* resultType, std::auto_ptr<Operand> op1, std::auto_ptr<Operand> op2)
	: detail::BinaryOperationImpl (resultType, op1, op2) {}

	void accept (Visitor& visitor) const
	{ visitor.visit (*this); }
};

/**
 * Class representing GIMPLE @c UNLE_EXPR.
 * It represents the unordered or less than or equal comparison operator.
 * It takes two floating point operands and determine whether the operands are
 * unordered or are less than or equal.
 * @c UnLeOperation returns true if either operand is an IEEE NaN or the first operand
 * is less than or equal to the second.
 * The result type is boolean type.
 */
class UnLeOperation : public detail::BinaryOperationImpl
{
public:
	/**
	 * Constructor, creates a new unordered less than or equal operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	UnLeOperation() {}

	/**
	 * Constructor, creates a new unordered less than or equal operation.
	 *
	 * @param resultType type of the result of the operation
	 * @param op1 first operand of the operation
	 * @param op2 second operand of the operation
	 */
	UnLeOperation(const Type* resultType, std::auto_ptr<Operand> op1, std::auto_ptr<Operand> op2)
	: detail::BinaryOperationImpl (resultType, op1, op2) {}

	void accept (Visitor& visitor) const
	{ visitor.visit (*this); }
};

/**
 * Class representing GIMPLE @c UNGT_EXPR.
 * It represents the unordered or greater than comparison operator.
 * It takes two floating point operands and determine whether the operands are
 * unordered or are greater than.
 * @c UnGtOperation returns true if either operand is an IEEE NaN or the first operand
 * is grater than the second.
 * The result type is boolean type.
 */
class UnGtOperation : public detail::BinaryOperationImpl
{
public:
	/**
	 * Constructor, creates a new unordered greater than operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	UnGtOperation() {}

	/**
	 * Constructor, creates a new unordered greater than operation.
	 *
	 * @param resultType type of the result of the operation
	 * @param op1 first operand of the operation
	 * @param op2 second operand of the operation
	 */
	UnGtOperation(const Type* resultType, std::auto_ptr<Operand> op1, std::auto_ptr<Operand> op2)
	: detail::BinaryOperationImpl (resultType, op1, op2) {}

	void accept (Visitor& visitor) const
	{ visitor.visit (*this); }
};

/**
 * Class representing GIMPLE @c UNGE_EXPR.
 * It represents the unordered or greater than or equal comparison operator.
 * It takes two floating point operands and determine whether the operands are
 * unordered or are greater than or equal.
 * @c UnGtOperation returns true if either operand is an IEEE NaN or the first operand
 * is grater than or equal to the second.
 * The result type is boolean type.
 */
class UnGeOperation : public detail::BinaryOperationImpl
{
public:
	/**
	 * Constructor, creates a new unordered greater than or equal operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	UnGeOperation() {}

	/**
	 * Constructor, creates a new unordered greater than or equal operation.
	 *
	 * @param resultType type of the result of the operation
	 * @param op1 first operand of the operation
	 * @param op2 second operand of the operation
	 */
	UnGeOperation(const Type* resultType, std::auto_ptr<Operand> op1, std::auto_ptr<Operand> op2)
	: detail::BinaryOperationImpl (resultType, op1, op2) {}

	void accept (Visitor& visitor) const
	{ visitor.visit (*this); }
};

/**
 * Class representing GIMPLE @c UNEQ_EXPR.
 * It represents the unordered or equal comparison operator.
 * It takes two floating point operands and determine whether the operands are
 * unordered or equal.
 * @c UnEqOperation returns true if either operand is an IEEE NaN or the first operand
 * is equal to the second.
 * The result type is boolean type.
 */
class UnEqOperation : public detail::BinaryOperationImpl
{
public:
	/**
	 * Constructor, creates a new unordered equal to operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	UnEqOperation() {}

	/**
	 * Constructor, creates a new unordered equal to operation.
	 *
	 * @param resultType type of the result of the operation
	 * @param op1 first operand of the operation
	 * @param op2 second operand of the operation
	 */
	UnEqOperation(const Type* resultType, std::auto_ptr<Operand> op1, std::auto_ptr<Operand> op2)
	: detail::BinaryOperationImpl (resultType, op1, op2) {}

	void accept (Visitor& visitor) const
	{ visitor.visit (*this); }
};

/**
 * Class representing GIMPLE @c LTGT_EXPR.
 * It represents the unordered or less than or greater than operator.
 * It is a reverse of @c UnExOperation.
 * It takes two floating point operands and determine whether the operands are
 * unordered or or less than or greater than.
 * @c UnEqOperation returns true if either operand is an IEEE NaN or the first operand
 * is less than or greater than the second.
 * The result type is boolean type.
 */
class LtGtOperation : public detail::BinaryOperationImpl
{
public:
	/**
	 * Constructor, creates a new unordered less than or greater than operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	LtGtOperation() {}

	/**
	 * Constructor, creates a new unordered less than or greater than operation.
	 *
	 * @param resultType type of the result of the operation
	 * @param op1 first operand of the operation
	 * @param op2 second operand of the operation
	 */
	LtGtOperation(const Type* resultType, std::auto_ptr<Operand> op1, std::auto_ptr<Operand> op2)
	: detail::BinaryOperationImpl (resultType, op1, op2) {}

	void accept (Visitor& visitor) const
	{ visitor.visit (*this); }
};

/**
 * Class representing GIMPLE @c WIDEN_SUM_EXPR.
 * It represents Widening summation.
 * The first argument is of type t1.
 * The second argument is of type t2, such that t2 is at least twice
 * the size of t1. The type of the entire expression is also t2.
 * @c WidenSumOperation is equivalent to first widening (promoting)
 * the first argument from type t1 to type t2, and then summing it
 * with the second argument.
 */
class WidenSumOperation : public detail::BinaryOperationImpl
{
public:
	/**
	 * Constructor, creates a new widen sum operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	WidenSumOperation() {}

	/**
	 * Constructor, creates a new widen sum operation.
	 *
	 * @param resultType type of the result of the operation
	 * @param op1 first operand of the operation
	 * @param op2 second operand of the operation
	 */
	WidenSumOperation(const Type* resultType, std::auto_ptr<Operand> op1, std::auto_ptr<Operand> op2)
	: detail::BinaryOperationImpl (resultType, op1, op2) {}

	void accept (Visitor& visitor) const
	{ visitor.visit (*this); }
};

/**
 * Class representing GIMPLE @c WIDEN_MULT_EXPR.
 * It represents widening multiplication.
 * The two arguments are of type t1.
 * The result is of type t2, such that t2 is at least twice
 * the size of t1. @c WidenMultOperation is equivalent to first widening (promoting)
 * the arguments from type t1 to type t2, and then multiplying them.
 */
class WidenMultOperation : public detail::BinaryOperationImpl
{
public:
	/**
	 * Constructor, creates a new widen mult sum operation.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	WidenMultOperation() {}

	/**
	 * Constructor, creates a new widen mult sum operation.
	 *
	 * @param resultType type of the result of the operation
	 * @param op1 first operand of the operation
	 * @param op2 second operand of the operation
	 */
	WidenMultOperation(const Type* resultType, std::auto_ptr<Operand> op1, std::auto_ptr<Operand> op2)
	: detail::BinaryOperationImpl (resultType, op1, op2) {}

	void accept (Visitor& visitor) const
	{ visitor.visit (*this); }
};

}

#endif /* GIT_OPERATIONS_H */
