// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file PhiStmt.h
 * Class representing the GIMPLE_PHI statement.
 */
#ifndef GIT_PHI_STATEMENT_H
#define GIT_PHI_STATEMENT_H

#include "Stmt.h"
#include "Operand.h"

#include <boost/noncopyable.hpp>
#include <boost/ptr_container/ptr_vector.hpp>
#include <boost/serialization/serialization.hpp>
#include <boost/ptr_container/serialize_ptr_vector.hpp>
#include <memory>

namespace giterator {

/**
 * Class representing the @c GIMPLE_PHI statement.
 * GIMPLE_PHI looks like this:
 * GIMPLE_PHI <RESULT, ARG1, ..., ARGN>, where
 *  RESULT = PHI <ARG1, ..., ARGN>.
 *  RESULT is the name created by this node.
 *
 * ARG1 ... ARGN are the arguments to the PHI node.  N must be
 * exactly the same as the number of incoming edges to the basic block
 * holding the PHI node.
 *
 * Sometimes, flow of control makes it impossible to determine what is the most
 * recent version of a variable. In these cases, the compiler inserts an artificial
 * definition for that variable called PHI function or PHI node.
 * This new definition merges all the incoming versions of the variable to create
 * a new name for it.
 * PHI nodes mean “one of these operands. I don't know which”.
 *
 * @see gimple_phi_to_git_stmt() in gimplexx-bridge.c to see how
 * a phi statement gets created from GIMPLE tree node.
 */
class PhiStmt : public Stmt, private boost::noncopyable {
  private:
	/*! Holder of a type of the versions of a variable. */
	typedef boost::ptr_vector<Operand> Arguments;

	/*! The result merged variable. */
    std::auto_ptr<Operand> m_result;
    /*! All the merged versions of a variable. */
    Arguments m_arguments;
  public:
      typedef Arguments::const_iterator ArgumentsIterator;

      /**
       * Constructor, creates a new phi statement.
       * Empty constructor exists for serialization purposes,
       * use nonempty for normal use.
       */
      PhiStmt() {}

      /**
       * Constructor, creates a new phi statement.
       */
      PhiStmt(std::auto_ptr<Operand> result) : m_result(result) {}

      void accept (Stmt::Visitor& visitor) const
      { visitor.visit (*this); }

      /**
       * Get a result merged variable of this operation.
       *
       * @return result merged variable for this operation
       */
      const Operand* getResult() const
      { return m_result.get(); }

      /**
       * Obtain all the versions of a variable merged by this class.
       *
       * @return pair of iterators, the first to the beginning of an array
       * of the variable versions, the second to the end.
       */
      std::pair<ArgumentsIterator, ArgumentsIterator>
      arguments () const
      { return std::make_pair(m_arguments.begin(), m_arguments.end()); }

      /**
       * Add a variable to variables merged by this operation.
       *
       * @param arg new variable version to merge.
       */
      void addArgument(std::auto_ptr<Operand> arg)
      {
    	  m_arguments.push_back(arg);
      }

      friend class boost::serialization::access;
      template<class Archive>
      void save(Archive & ar, const unsigned int) const
      {
    	  Operand* result = m_result.get();
    	  ar << result;
    	  ar << m_arguments;
      }

      template<class Archive>
      void load(Archive & ar, const unsigned int)
      {
    	  Operand* result;
    	  ar >> result;
    	  m_result = std::auto_ptr<Operand> (result);
    	  ar >> m_arguments;
      }

      template<class Archive>
      void serialize(Archive & ar, const unsigned int version)
      {
    	  boost::serialization::void_cast_register(static_cast<PhiStmt *>(NULL), static_cast<Stmt *>(NULL));
    	  boost::serialization::split_member(ar, *this, version);
      }
};

}

#endif /* GIT_PHI_STATEMENT_H */

