// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file PtrType.h
 * Class representing the type of a pointer.
 */
#ifndef GIT_PTR_TYPE_H
#define GIT_PTR_TYPE_H

#include "AbstractType.h"
//#include "Serialization.h"

namespace giterator { 
    
/**
 * Class representing the type of a pointer.
 * This type gets created from GIMPLE @c POINTER_TYPE
 * tree nodes.
 * This type stores an information about the type
 * of a values the pointers of this type point to.
 *
 * * @see create_git_type_from_tree() in gimplexx-bridge.c
 * to see how pointer types get created from GIMPLE types
 */
class PtrType : public detail::AbstractType {
  private:
    /*! the type of a values the pointers of this type point to */
    const Type* m_pointingToType;

  public:
    /**
     * Constructor, creates a new pointer type.
     * Empty constructor exists for serialization purposes,
     * use nonempty for normal use.
     */
    PtrType() : detail::AbstractType(), m_pointingToType(NULL) {}
    /**
     * Constructor, creates a new pointer type.
     * 
     * @param name name of the type
     * @param bitSize size of the type in bits
     * @param pointingToType the type to which the pointers of the new type point to
     */
    PtrType (const std::string& name, 
             size_t bitSize, const Type* pointingToType) 
        : detail::AbstractType (name, bitSize),
    	m_pointingToType (pointingToType) {}

    /**
     * Return the type the pointers of this type point to.
     * 
     * @return the type the pointers of this type point to
     */
    const Type* getPointingToType() const 
    {
      return m_pointingToType;
    }

    /**
     * Type comparison.
     * Compares the type to another pointer type.
     * Two pointer types are the same if they point to the same
     * type.
     *
     * @param type the type this type is compared to
     * @return true if the types are the same, false otherwise
     */
    bool isSame (const Type& type) const 
    {
      const PtrType* ptrType = dynamic_cast<const PtrType*>(&type);
      if (ptrType == 0)
        return false;

      return getPointingToType() == ptrType->getPointingToType();
    }

    friend class boost::serialization::access;
	template<class Archive>
	void save(Archive & ar, const unsigned int) const
	{
		ar << boost::serialization::base_object<AbstractType>(*this);
		ar << m_pointingToType;
	}

	template<class Archive>
	void load(Archive & ar, const unsigned int)
	{
		ar >> boost::serialization::base_object<AbstractType>(*this);
		Type* type;
		ar >> type;
		m_pointingToType = (const Type*) type;
	}

	template<class Archive>
	void serialize(Archive & ar, const unsigned int version)
	{
		boost::serialization::void_cast_register(static_cast<PtrType *>(NULL), static_cast<TreeProduct *>(NULL));
		boost::serialization::void_cast_register(static_cast<PtrType *>(NULL), static_cast<Type *>(NULL));
		boost::serialization::split_member(ar, *this, version);
	}
};

}

#endif /* GIT_PTR_TYPE_H */

