// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file RealConstant.h
 * Class representing a real constant.
 */

#ifndef GIT_REAL_CONSTANT_H
#define GIT_REAL_CONSTANT_H

#include "Constant.h"
#include "Type.h"
#include "Serialization.h"

namespace giterator {

/**
 * Class representing a real constant.
 * Real constants get created from GIMPLE @c REAL_CST nodes.
 * The value of real constant can be retrieved from GIMPLE by
 * @c real_to_decimal_for_mode() function from real.h
 * @see create_git_constant_from_tree() in gimplexx-bridge.c to
 * see how real constants get created from GIMPLE constants
 */
class RealConstant : public Constant {
  private:
    /*! type of this constant */
    const Type* m_type;

	 /*! raw value of this constant */
	long double m_value;

  public:
	/**
	 * Constructor, creates a new real constant.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
    RealConstant() {}
    /**
     * Constructor, creates a new real constant.
     *
     * @param type the type of the constant
     * @param value raw value of the constant.
     *      All values are passed in as a @c long double
     *      The precision and the unsigned flag in the @c type specify, how will be
     *      the raw value represented.
     */
    RealConstant (const Type* type, long double value)
        : m_type (type), m_value(value) {}

    virtual void accept (Visitor& visitor) const
    { visitor.visit (*this); }

    /**
     * Obtain the type of this constant.
     */
    const Type* getType () const
    { return m_type; }

    /**
     * Returns the raw value of this constant.
     *
     * @return raw value of this constant
     */
    long double getValue () const
    { return m_value; }

	friend class boost::serialization::access;
	template<class Archive>
	void save(Archive & ar, const unsigned int) const
	{
		ar << m_type;
		ar << m_value;
	}

	template<class Archive>
	void load(Archive & ar, const unsigned int)
	{
		Type* type;
		ar >> type;
		m_type = (const Type*) type;
		ar >> m_value;
	}

	template<class Archive>
	void serialize(Archive & ar, const unsigned int version)
	{
		boost::serialization::void_cast_register(static_cast<RealConstant *>(NULL), static_cast<TreeProduct *>(NULL));
		boost::serialization::void_cast_register(static_cast<RealConstant *>(NULL), static_cast<Constant *>(NULL));
		boost::serialization::split_member(ar, *this, version);
	}
};

}

#endif /* GIT_REAL_CONSTANT_H */
