// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file RealType.h
 * Class representing float and double type.
 */
#ifndef GIT_REAL_TYPE_H
#define GIT_REAL_TYPE_H

#include "AbstractType.h"

#include <boost/serialization/serialization.hpp>

namespace giterator {

/**
 * Class representing float and double type.
 * Real types get created from GIMPLE @c REAL_TYPE
 * nodes. Each such node has @c TYPE_PRECISION determining
 * number of bits used for representation (not the same
 * as @c TYPE_SIZE).
 *
 * @see create_git_type_from_tree() in gimplexx-bridge.c to see
 * how real types get created from GIMPLE types
 *
 * @see giterator::Type
 */
class RealType : public detail::AbstractType {
  private:
    /*! type precision in bits */
    size_t m_bitPrecision;

  public:
    /**
     * Constructor, creates a new real type.
     * Empty constructor exists for serialization purposes,
     * use nonempty for normal use.
     */
    RealType() : detail::AbstractType() {}
    /**
     * Constructor, creates a new real type.
     *
     * @param name name of the type
     * @param bitSize size of the type in bits
     * @param bitPrecision the precision of the new type
     */
    RealType (const std::string& name,
                 size_t bitSize, size_t bitPrecision)
        : detail::AbstractType (name, bitSize),
          m_bitPrecision (bitPrecision) {}

    /**
     * Returns the precision of this type.
     *
     * @return the precision of this type in bits
     */
    size_t getBitPrecision() const
    {
      return m_bitPrecision;
    }

    /** Type comparison.
     * Compare the type to another type.
     * Two real types are the same if their bit size and bit
     * precisions are the same.
     *
     * @param type the type this type is compared to.
     * @return true if the types are the same,
     * false otherwise
     */
    bool isSame (const Type& type) const
    {
      const RealType* realType = dynamic_cast<const RealType*>(&type);
      if (realType == 0)
        return false;

      return getBitSize() == realType->getBitSize()
              && getBitPrecision() == realType->getBitPrecision();
    }

    friend class boost::serialization::access;
	template<class Archive>
	void serialize(Archive & ar, const unsigned int)
	{
		boost::serialization::void_cast_register(static_cast<RealType *>(NULL), static_cast<TreeProduct *>(NULL));
		boost::serialization::void_cast_register(static_cast<RealType *>(NULL), static_cast<Type *>(NULL));
		ar & boost::serialization::base_object<AbstractType>(*this);
		ar & m_bitPrecision;
	}
};

}

#endif /* GIT_REAL_TYPE_H */
