// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file RecordConstructorOperand.h
 * Operand, that is a constructor of a record.
 */
#ifndef GIT_RECORD_CONSTRUCTOR_OPERAND_H
#define GIT_RECORD_CONSTRUCTOR_OPERAND_H

#include "ConstructorOperand.h"
#include "FieldDecl.h"

#include <boost/ptr_container/ptr_map.hpp>
#include <boost/serialization/serialization.hpp>


namespace giterator {

/**
 * The operand that is a constructor of a record (struct/union).
 * For example in: 'struct p p1 = {"xx", {5, "xx"}, 1, 2};'
 * {"xx", {5, "xx"}, 1, 2} is the initializer.
 * These operands get created from GIMPLE @c CONSTRUCTOR
 * tree nodes, that have @c TREE_TYPE @c RECORD_TYPE or @c UNION_TYPE.
 * This class stores initialization value for each array index
 * that will be initialized.
 *
 * @see create_git_operand_from_tree() in gimplexx-brideg.c to see
 * how record constructor operands get created from GIMPLE operands
 */
class RecordConstructorOperand : public ConstructorOperand {
  private:
	/*! type of the container for mapping record fields to initialization values. */
	typedef boost::ptr_map<const FieldDecl*, Operand> RecordConstructors;

	/*! Field mapping record field declarations to their initialization values.*/
	RecordConstructors m_constructors;

  public:
	/** iterator for field declaration pointer->initializer mapper
	 * @see m_constructors
	 */
	typedef RecordConstructors::const_iterator RecordConstructorsIterator;

	/**
	 * Constructor, creates a new record constructor operand.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
    RecordConstructorOperand () {}

    /**
     * Constructor, creates a new record constructor operand.
     *
     * @param type the constructed type
     */
    RecordConstructorOperand (const Type* type)
    : ConstructorOperand(type) {}

    virtual void accept (Visitor& v) const
    { v.visit (*this); }

    /**
	 * Obtain the record constructors held by this operand.
	 *
	 * @return pair of values - beginning of constructors map, end of constructors map
	 */
    std::pair<RecordConstructorsIterator, RecordConstructorsIterator>
	constructors () const
	{ return std::make_pair (m_constructors.begin (), m_constructors.end ()); }

    /**
	 * Add a initialization value for a specific field to the constructor map.
	 *
	 * @param field field in the record for the initialization value
	 * @param value the initialization value for specified field
	 */
    void addConstructor(const FieldDecl* field, std::auto_ptr<Operand> value)
    {
    	assert(m_constructors.find(field) == m_constructors.end());
    	m_constructors.insert(field, value);
    }

    friend class boost::serialization::access;
	template<class Archive>
	void save(Archive & ar, const unsigned int) const
	{
		ar << boost::serialization::base_object<detail::TypeOperand>(*this);

		size_t size = m_constructors.size();
		ar << size;
		for (RecordConstructorsIterator i1 = m_constructors.begin(); i1 != m_constructors.end(); ++i1)
		{
			ar << i1->first;
			Operand* op = (Operand*) i1->second;
			ar << op;
		}
	}

	template<class Archive>
	void load(Archive & ar, const unsigned int)
	{
		ar >> boost::serialization::base_object<detail::TypeOperand>(*this);

		size_t size;
		ar >> size;
		for (size_t i = 0; i < size; ++i)
		{
			FieldDecl* fd;
			ar >> fd;
			Operand* op;
			ar >> op;
			m_constructors.insert((const FieldDecl*) fd, std::auto_ptr<Operand> (op));
		}
	}

	template<class Archive>
	void serialize(Archive & ar, const unsigned int version)
	{
		boost::serialization::void_cast_register(static_cast<RecordConstructorOperand *>(NULL), static_cast<Operand *>(NULL));
		boost::serialization::split_member(ar, *this, version);
	}
};

}

#endif /* GIT_RECORD_CONSTRUCTOR_OPERAND_H */
