// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file RecordDecl.h
 * Class representing declaration of a struct/union.
 */
#ifndef GIT_RECORD_DECL_H
#define GIT_RECORD_DECL_H

#include "OffsetDecl.h"
#include "Type.h"
#include "RecordType.h"
#include <boost/serialization/serialization.hpp>

namespace giterator {

/**
 * Class representing declaration of a struct or union.
 * Record declarations are equivalents of variable declarations
 * (@c VarDecl). This class is created instead of @c VarDecl
 * from GIMPLE @c VAR_DECL nodes if the created variable
 * declaration has record or union type (@c RecordType).
 *
 * @see create_git_decl_from_tree() in gimplexx-bridge.c
 * to see how declarations of structs/unions get created
 *  from GIMPLE declarations
 *
 * @see giterator::VarDecl
 */
class RecordDecl : public OffsetDecl {
  private:
	/*! Whether this is local ol global declaration. */
	bool m_local;
  public:
	/**
	 * Constructor, creates a new record declaration.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	RecordDecl () : OffsetDecl() {}
	/**
	 * Constructor, creates a new record declaration.
	 *
     * @param uid unique ID of the new declaration
     * @param name name of the new declaration
     * @param type type of the new declaration
     * @param local whether this is local or global declaration
     */
	RecordDecl (unsigned int uid, const std::string& name, const Type* type, bool local) :
                OffsetDecl (uid, name, type), m_local(local) {}

    void accept (Visitor& visitor) const
    { visitor.visit (*this); }

    /**
     * Find out whether this is local or global declaration.
     *
     * @return true if this is local declaration,
     * false otherwise.
     */
    bool getLocal() const
    { return m_local; }

    friend class boost::serialization::access;
	template<class Archive>
	void serialize(Archive & ar, const unsigned int version)
	{
		boost::serialization::void_cast_register(static_cast<RecordDecl *>(NULL), static_cast<TreeProduct *>(NULL));
		boost::serialization::void_cast_register(static_cast<RecordDecl *>(NULL), static_cast<Decl *>(NULL));
		ar & boost::serialization::base_object<OffsetDecl>(*this);
		ar & m_local;
	}
};

}

#endif /* GIT_RECORD_DECL_H */
