// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file RecordType.h
 * Class representing type of a struct/union.
 */
#ifndef GIT_RECORD_TYPE_H
#define GIT_RECORD_TYPE_H

#include "AbstractType.h"
#include "Decl.h"

#include <map>
#include <string>
#include <cassert>
#include <boost/serialization/serialization.hpp>

namespace giterator {

/**
 * Class representing struct or union type.
 * Record types get created from GIMPLE @c RECORD_TYPE
 * or @c UNION_TYPE nodes. Each such node has @c TYPE_FIELDS
 * containing declarations of all the fields.
 * Pointers to these are stored inside this class.
 *
 * @see create_git_type_from_tree() in gimplexx-bridge.c
 * to see how struct/union types get created from GIMPLE types
 */
class RecordType : public detail::AbstractType {
  private:
	/*! Type of the container holding fields of this type.*/
	typedef std::map<std::string, const Decl*> Fields;
	/*! Type of the container holding local declarations of this type (for future use). */
	typedef std::map<std::string, const Decl*> LocalDecls;

	/*! Fields of this type */
	Fields m_fields;
	/*! Local declarations of this type (for future use). */
	LocalDecls m_localDecls;

  public:
	/*! Iterator over a fields container. */
	typedef std::map<std::string, const Decl*>::const_iterator FieldsIterator;
	/*! Iterator over a local declarations container. */
	typedef std::map<std::string, const Decl*>::const_iterator LocalDeclsIterator;

	/**
	 * Constructor, creates a new record type.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
    RecordType() : detail::AbstractType() {}
    /**
     * Constructor, creates a new record type.
     *
     * @param name name of the type
     * @param bitSize size of the type in bits
     */
    RecordType (const std::string& name, size_t bitSize)
        : detail::AbstractType (name, bitSize) {}

    /**
     * Type comparison.
     * Compare this type to another type.
     * Two record types are the same if their fields, local declarations
     * and bit sizes are the same.
     *
     * @param type the type this type is compared to
     * @return true if the types are the same, false otherwise
     */
    bool isSame (const Type& type) const
    {
      const RecordType* recType = dynamic_cast<const RecordType*>(&type);
      if (recType == 0)
        return false;

      return fieldsEqualWith(recType)
    		  && localDeclsEqualWith(recType)
    		  && getBitSize() == recType->getBitSize();
    }

    /**
     * Add a field declaration of this type to a container
     * holding field declarations.
     *
     * @param name name of the added field
     * @param value declaration of a field of this type with name @c name
     */
    void addField(std::string name, const Decl* value)
    {
    	assert(!containsField(name));
		m_fields[name] = value;
    }

    /**
     * Get a field declaration of this type with a specified name.
     *
     * @param name of the field to get from the container holding
     * field declarations of this type
     * @return declaration of a field of this type with name @c name
     */
    const Decl* getField(std::string name) const
	{
    	assert(containsField(name));
    	return m_fields.find(name)->second;
	}

    /**
	 * Add a local declaration for this type to a container
	 * holding local declarations.
	 *
	 * @param name name of the added local declaration
	 * @param value local declaration for this type with name @c name
	 */
    void addLocalDecl(std::string name, const Decl* value)
    {
    	assert(!containsLocalDecl(name));
    	m_localDecls[name] = value;
    }

    /**
	 * Get a local declaration for this type with a specified name.
	 *
	 * @param name of the field to get from the container holding
	 * local declarations of this type
	 * @return local declaration for this type with name @c name
	 */
    const Decl* getLocalDecl(std::string name) const
	{
    	assert(containsLocalDecl(name));
    	return m_localDecls.find(name)->second;
	}

    /**
     * Check whether this type contains field with specified name.
     *
     * @param name name of the field to check
     * @return true if this type contains field with name @c name,
     * false otherwise
     */
    bool containsField(std::string name) const
    {
    	return m_fields.find(name) != m_fields.end();
    }

    /**
	 * Check whether this type contains local declaration with specified name.
	 *
	 * @param name name of the local declaration to check
	 * @return true if this type contains local declaration with name @c name,
	 * false otherwise
	 */
    bool containsLocalDecl(std::string name) const
	{
		return m_localDecls.find(name) != m_localDecls.end();
	}

    /**
   	 * Obtain the fields of this type.
   	 *
   	 * @return pair of values - beginning of fields container, end of fields container
   	 */
    std::pair<FieldsIterator, FieldsIterator>
	fields () const
	{ return std::make_pair (m_fields.begin (), m_fields.end ()); }

    /**
	 * Obtain the fields for this type.
	 *
	 * @return pair of values - beginning of local declar. container, end of local declar. container
	 */
    std::pair<LocalDeclsIterator, LocalDeclsIterator>
	localDecls () const
	{ return std::make_pair (m_localDecls.begin (), m_localDecls.end ()); }

    friend class boost::serialization::access;
	template<class Archive>
	void save(Archive & ar, const unsigned int) const
	{
		ar << boost::serialization::base_object<AbstractType>(*this);

		size_t size =  m_fields.size();
		ar << size;
		for (Fields::const_iterator i1 = m_fields.begin(); i1 != m_fields.end(); ++i1)
		{
			std::string name = i1->first;
			ar << name;
			const Decl* value = (const Decl*) i1->second;
			ar << value;
		}

		size =  m_localDecls.size();
		ar << size;
		for (LocalDecls::const_iterator i1 = m_localDecls.begin(); i1 != m_localDecls.end(); ++i1)
		{
			std::string name = i1->first;
			ar << name;
			const Decl* value = (const Decl*) i1->second;
			ar << value;
		}
	}

	template<class Archive>
	void load(Archive & ar, const unsigned int)
	{
		ar >> boost::serialization::base_object<AbstractType>(*this);

		size_t size;
		ar >> size;
		for (size_t i = 0; i < size; ++i)
		{
			std::string name;
			Decl* value;
			ar >> name;
			ar >> value;
			m_fields[name] = value;
		}

		ar >> size;
		for (size_t i = 0; i < size; ++i)
		{
			std::string name;
			Decl* value;
			ar >> name;
			ar >> value;
			m_localDecls[name] = value;
		}
	}

	template<class Archive>
	void serialize(Archive & ar, const unsigned int version)
	{
		boost::serialization::void_cast_register(static_cast<RecordType *>(NULL), static_cast<TreeProduct *>(NULL));
		boost::serialization::void_cast_register(static_cast<RecordType *>(NULL), static_cast<Type *>(NULL));
		boost::serialization::split_member(ar, *this, version);
	}

  private:
	/**
	 * Check whether fields of this type equal with fields of another type.
	 *
	 * @param type type to check the fields equality with
	 * @return true if fields of this type equal with fields of type @n type,
	 * false otherwise
	 *
	 * @see isSame()
	 */
	bool fieldsEqualWith(const RecordType* type) const
	{
		for (Fields::const_iterator i1 = m_fields.begin(); i1 != m_fields.end(); ++i1)
		{
			if (!type->containsField(i1->first))
				return false;

			if (!type->getField(i1->first)->isSame(*i1->second))
				return false;
		}

		return true;
	}

	/**
	 * Check whether local declarations of this type equal with local declarations of another type.
	 *
	 * @param type type to check the local declarations equality with
	 * @return true if local declarations of this type equal with fields of type @n type,
	 * false otherwise
	 *
	 * @see isSame()
	 */
	bool localDeclsEqualWith(const RecordType* type) const
	{
		for (LocalDecls::const_iterator i1 = m_localDecls.begin(); i1 != m_localDecls.end(); ++i1)
		{
			if (!type->containsLocalDecl(i1->first))
				return false;

			if (!type->getLocalDecl(i1->first)->isSame(*i1->second))
				return false;
		}

		return true;
	}
};

}

#endif /* GIT_RECORD_TYPE_H */

