// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file ReturnStmt.h
 * Class representing the GIMPLE_RETURN statement.
 */
#ifndef GIT_RETURN_STATEMENT_H
#define GIT_RETURN_STATEMENT_H

#include <memory>

#include "Stmt.h"
#include "Operand.h"

#include <boost/serialization/serialization.hpp>

namespace giterator {
    
/**
 * Class representing the @c GIMPLE_RETURN statement.
 * GIMPLE_RETURN looks like this:
 * GIMPLE_RETURN <RETVAL>, where
 * RETVAL is the value to return or NULL.
 *
 * @see gimple_return_to_git_stmt() in gimplexx-bridge.c to see
 * how a return statement gets created from GIMPLE tree node.
 */
class ReturnStmt : public Stmt {
  private:
    /*! operand representing the returned value */
    std::auto_ptr<Operand> m_returnValue;

  public:
    /**
     * Constructor, creates a new return statement.
     * Empty constructor exists for serialization purposes,
     * use nonempty for normal use.
     */
    ReturnStmt() {}
    /**
     * Constructor, creates a new return statement.
     * 
     * @param returnValue operand representing the returned value
     */
    ReturnStmt (std::auto_ptr<Operand> returnValue) 
    	: m_returnValue (returnValue) {}

    void accept (Stmt::Visitor& visitor) const
    { visitor.visit (*this); }

    /**
     * Returns the operand representing the value returned by this statement.
     * 
     * @return the operand representing the value returned by this statement
     */
    const Operand* getReturnValue () const 
    { return m_returnValue.get (); }

    friend class boost::serialization::access;
	template<class Archive>
	void save(Archive & ar, const unsigned int) const
	{
		Operand* returnValue = (Operand *) m_returnValue.get();
		ar << returnValue;
	}

	template<class Archive>
	void load(Archive & ar, const unsigned int)
	{
		Operand* returnValue;
		ar >> returnValue;
		m_returnValue = std::auto_ptr<Operand> (returnValue);
	}

	template<class Archive>
	void serialize(Archive & ar, const unsigned int version)
	{
		boost::serialization::void_cast_register(static_cast<ReturnStmt *>(NULL), static_cast<Stmt *>(NULL));
		boost::serialization::split_member(ar, *this, version);
	}
};

}

#endif /* GIT_RETURN_STATEMENT_H */

