// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file Stmt.h
 * Interface for classes that represent GIMPLE statements.
 */
#ifndef GIT_STATEMENT_H
#define GIT_STATEMENT_H

namespace giterator {
    
class CallStmt;
class ReturnStmt;
class AssignStmt;
class CondStmt;
class GotoStmt;
class LabelStmt;
class NopStmt;
class PhiStmt;
class SwitchStmt;
    
/**
 * Interface for classes that represent GIMPLE statements.
 * GIMPLE statements are the instruction set of GMC.
 * Subclasses of this interface represent concrete GIMPLE
 * statements.
 * GIMPLE statements get extracted from GIMPLE basic blocks.
 * They store all of their operands and other data needed
 * to interpret the statement.
 *
 * @see create_git_decl_from_tree() in gimplexx_bridge.c
 * to see how declarations get created from GIMPLE constants.
 */
class Stmt {
  public:
    /**
     * Visitor class for sub-classes of @c Stmt class, that are meant to be used 
     * by the user.
     */
    class Visitor {
      public:
        virtual ~Visitor () {}
    
      public:
        virtual void visit (const CallStmt& stmt) = 0;
        virtual void visit (const ReturnStmt& stmt) = 0;
        virtual void visit (const AssignStmt& stmt) = 0;
        virtual void visit (const CondStmt& stmt) = 0;
        virtual void visit (const GotoStmt& stmt) = 0;
        virtual void visit (const LabelStmt& stmt) = 0;
        virtual void visit (const NopStmt& stmt) = 0;
        virtual void visit (const PhiStmt& stmt) = 0;
        virtual void visit (const SwitchStmt& stmt) = 0;
    };
    
  public:
    /**
     * Destructor.
     */
    virtual ~Stmt () {}

    /**
     * Calls @c visitor.accept().
     * Classes implementing this interface implement this method to call @c visitor.accept()
     * with @c *this as a parameter.
     * 
     * @param visitor visitor on which to call the @c accept method
     */
    virtual void accept (Stmt::Visitor& visitor) const = 0;
};

}

#endif /* GIT_STATEMENT_H */

