// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file StringConstant.h
 * Class representing a string constant.
 */
#ifndef GIT_STRING_CONSTANT_H
#define GIT_STRING_CONSTANT_H

#include "Constant.h"
#include "Type.h"
#include "Serialization.h"

#include <string>

namespace giterator { 
    
/**
 * The class representing a string constant.
 * String constants get created from GIMPLE @c STRING_CST nodes.
 * Each such node has @c TREE_STRING_LENGTH holding the length
 * of the constant ant @c TREE_STRING_POINTER holding the
 * constant itself.
 *
 * @see create_git_constant_from_tree() in gimplexx-bridge.c to
 * see how string constants get created from GIMPLE constants
 */
class StringConstant : public Constant {
  private:
    /*! type of this constant */
    const Type* m_type;
    /*! the value of this constant */
    std::string m_str;
    
  public:
    /**
     * Constructor, creates a new string constant.
     * Empty constructor exists for serialization purposes,
     * use nonempty for normal use.
     */
    StringConstant () {}

    /**
     * Constructor, creates a new string constant.
     * 
     * @param type type of the constant
     * @param str raw value of the constant - repesented as @c std::string
     */
    StringConstant (const Type* type, const std::string& str) 
        : m_type (type), m_str (str) {}
    
    /**
     * Returns the value of this constant.
     * 
     * @return value of this constant
     */
    const std::string& getValue () const
    { return m_str; }
    
    void accept (Visitor& visitor) const
    { visitor.visit (*this); }
    
    /**
     * Get the type of this constant.
     *
     * @return the type of this constant
     */
    const Type* getType () const
    { return m_type; }

    friend class boost::serialization::access;
	template<class Archive>
	void save(Archive & ar, const unsigned int) const
	{
		ar << m_type;
		ar << m_str;
	}

	template<class Archive>
	void load(Archive & ar, const unsigned int)
	{
		Type* type;
		ar >> type;
		m_type = (const Type*) type;
		ar >> m_str;
	}

	template<class Archive>
	void serialize(Archive & ar, const unsigned int version)
	{
		boost::serialization::void_cast_register(static_cast<StringConstant *>(NULL), static_cast<TreeProduct *>(NULL));
		boost::serialization::void_cast_register(static_cast<StringConstant *>(NULL), static_cast<Constant *>(NULL));
		boost::serialization::split_member(ar, *this, version);
	}
};

}

#endif /* GIT_STRING_CONSTANT_H */

