// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file StringConstant.h
 * Class representing a string constant.
 */
/** @file SwitchStmt.h
 * Class representing GIMPLE_SWITCH statement.
 */
#ifndef GIT_SWITCH_STATEMENT_H
#define GIT_SWITCH_STATEMENT_H

#include "Stmt.h"
#include "Operand.h"

#include <boost/noncopyable.hpp>
#include <boost/ptr_container/ptr_vector.hpp>
#include <boost/serialization/serialization.hpp>
#include <boost/ptr_container/serialize_ptr_vector.hpp>
#include <memory>

namespace giterator {

/**
 * Class representing the @c GIMPLE_SWITCH statement.
 * It is a multiway branch.
 * GIMPLE_SWITCH looks like this:
 * GIMPLE_SWITCH <INDEX, DEFAULT_LAB, LAB1, ..., LABN> ,
 * which is an equivalent of
 * @code
 *  switch (INDEX)
 *  {
 *    case LAB1: ...; break;
 *    ...
 *    case LABN: ...; break;
 *    default: ...
 *  }
 *  @endcode
 *
 *  INDEX is the variable evaluated to decide which label to jump to.
 *
 *  DEFAULT_LAB, LAB1 ... LABN are the tree nodes representing case labels.
 *  They must be CASE_LABEL_EXPR nodes.
 *
 * @see gimple_switch_to_git_stmt() in gimplexx-bridge.c to see how
 * a switch statement gets created from GIMPLE tree node.
 */
class SwitchStmt : public Stmt, private boost::noncopyable {
  private:

	/*! Container for the labels of this switch. */
    typedef boost::ptr_vector<Operand> Labels;

    /*! Variable to decide which label to jump to. */
    std::auto_ptr<Operand> m_index;

    /*! The labels of this switch. */
    Labels m_labels;
  public:
    /*! Iterator for the labels of this switch. */
    typedef Labels::const_iterator LabelsIterator;

	/**
	* Constructor, creates a new switch statement.
	* Empty constructor exists for serialization purposes,
	* use nonempty for normal use.
	*/
	SwitchStmt() {}

	/**
	* Constructor, creates a new switch statement.
	*
	* @param index variable to decide which label to jump to
	*/
	SwitchStmt(std::auto_ptr<Operand> index) : m_index(index) {}

	void accept (Stmt::Visitor& visitor) const
	{ visitor.visit (*this); }

	/**
	 * Get a variable to decide which label to jump to
	 *
	 * @return variable to decide which label to jump to
	 */
	const Operand* getIndex() const
	{ return m_index.get(); }

	/**
	 * Get all the labels of this switch
	 *
	 * @return pair of values: beginning of the labels container,
	 * end of the labels container
	 */
	std::pair<LabelsIterator, LabelsIterator>
	labels () const
	{ return std::make_pair(m_labels.begin(), m_labels.end()); }

	/**
	 * Add a label to labels of this switch.
	 * This object owns the value.
	 *
	 * @param label a label to add to labels of this switch
	 */
	void addLabel(std::auto_ptr<Operand> label)
	{
	  m_labels.push_back(label);
	}

	friend class boost::serialization::access;
	template<class Archive>
	void save(Archive & ar, const unsigned int) const
	{
	  Operand* result = m_index.get();
	  ar << result;
	  ar << m_labels;
	}

	template<class Archive>
	void load(Archive & ar, const unsigned int)
	{
	  Operand* result;
	  ar >> result;
	  m_index = std::auto_ptr<Operand> (result);
	  ar >> m_labels;
	}

	template<class Archive>
	void serialize(Archive & ar, const unsigned int version)
	{
	  boost::serialization::void_cast_register(static_cast<SwitchStmt *>(NULL), static_cast<Stmt *>(NULL));
	  boost::serialization::split_member(ar, *this, version);
	}
};

}

#endif /* GIT_SWITCH_STATEMENT_H */

