// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file ThreadConfigParser.cpp
 * Definitions for class for parsing thread configuration file.
 */
#include "ThreadConfigParser.h"

#include <fstream>
#include <cstdlib>
#include <exception>

namespace giterator
{
	/// <summary>
	/// Error message displayed when unable to open configuration file.
	/// </summary>
	#define CONFIG_FILE_OPEN_ERROR_STR "Unable to open thread config file. Location: "

	/// <summary>
	/// Error message when unexpected error during parsing occured.
	/// </summary>
	#define UNEXPECTED_PARSER_ERROR_STR "Unexpected error while parsing thread config file: "

	/// <summary>
	/// Char marking comment line in configuration file.
	/// </summary>
	#define COMMENT_START_CHAR '#'

	/// <summary>
	/// Name of item in config file defining thread creation.
	/// </summary>
	/// <see cref="CREATE_NAME_LENGTH" />
	#define CREATE_NAME "CREATE"

	/// <summary>
	/// Length of item definig thread creation.
	/// </summary>
	/// <see cref="CREATE_NAME" />
	#define CREATE_NAME_LENGTH 6

	/// <summary>
	/// Name of item in config file definig thread join.
	/// </summary>
	/// <see cref="JOIN_NAME_LENGTH" />
	#define JOIN_NAME "JOIN"

	/// <summary>
	/// Length of item defining command for thread join.
	/// </summary>
	/// <see cref="JOIN_NAME" />
	#define JOIN_NAME_LENGTH 4

	/// <summary>
	/// Name of item in config file defining thread exit.
	/// </summary>
	/// <see cref="EXIT_NAME_LENGTH" />
	#define EXIT_NAME "EXIT"

	/// <summary>
	/// Length of item defining thread exit.
	/// </summary>
	/// <see cref="EXIT_NAME" />
	#define EXIT_NAME_LENGTH 4

	/// <summary>
	/// Name of item in config file defining type of thread.
	/// </summary>
	/// <see cref="TYPE_NAME_LENGTH" />
	#define TYPE_NAME "TYPE"

	/// <summary>
	/// Length of item defining type of thread.
	/// </summary>
	/// <see cref="TYPE_NAME" />
	#define TYPE_NAME_LENGTH 4

	/// <summary>
	/// Name of item in config file defining mutex create.
	/// </summary>
	/// <see cref="MUTEX_INIT_NAME_LENGTH" />
	#define MUTEX_INIT_NAME "MUTEX_INIT"

	/// <summary>
	/// Length of item defining type of thread.
	/// </summary>
	/// <see cref="MUTEX_INIT_NAME" />
	#define MUTEX_INIT_NAME_LENGTH 10

	/// <summary>
	/// Name of item in config file defining mutex destroy.
	/// </summary>
	/// <see cref="MUTEX_DESTROY_NAME_LENGTH" />
	#define MUTEX_DESTROY_NAME "MUTEX_DESTROY"

	/// <summary>
	/// Length of item defining mutex destroy.
	/// </summary>
	/// <see cref="MUTEX_DESTROY_NAME" />
	#define MUTEX_DESTROY_NAME_LENGTH 13

	/// <summary>
	/// Name of item in config file defining mutex lock.
	/// </summary>
	/// <see cref="MUTEX_LOCK_NAME_LENGTH" />
	#define MUTEX_LOCK_NAME "MUTEX_LOCK"

	/// <summary>
	/// Length of item defining mutex lock.
	/// </summary>
	/// <see cref="MUTEX_LOCK_NAME" />
	#define MUTEX_LOCK_NAME_LENGTH 10

	/// <summary>
	/// Name of item in config file defining mutex try lock.
	/// </summary>
	/// <see cref="MUTEX_TRY_LOCK_NAME_LENGTH" />
	#define MUTEX_TRY_LOCK_NAME "MUTEX_TRY_LOCK"

	/// <summary>
	/// Length of item defining mutex mutex try lock.
	/// </summary>
	/// <see cref="MUTEX_TRY_LOCK_NAME" />
	#define MUTEX_TRY_LOCK_NAME_LENGTH 14

	/// <summary>
	/// Name of item in config file defining mutex unlock.
	/// </summary>
	/// <see cref="MUTEX_UNLOCK_NAME_LENGTH" />
	#define MUTEX_UNLOCK_NAME "MUTEX_UNLOCK"

	/// <summary>
	/// Length of item defining mutex unlock.
	/// </summary>
	/// <see cref="MUTEX_UNLOCK_NAME" />
	#define MUTEX_UNLOCK_NAME_LENGTH 12

	/// <summary>
	/// Name of item in config file defining mutex type.
	/// </summary>
	/// <see cref="MUTEX_TYPE_NAME_LENGTH" />
	#define MUTEX_TYPE_NAME "MUTEX_TYPE"

	/// <summary>
	/// Length of item defining mutex type.
	/// </summary>
	/// <see cref="MUTEX_TYPE_NAME" />
	#define MUTEX_TYPE_NAME_LENGTH 10

	/// <summary>
	/// Name of item in config file defining condition init.
	/// </summary>
	/// <see cref="COND_INIT_NAME_LENGTH" />
	#define COND_INIT_NAME "COND_INIT"

	/// <summary>
	/// Length of item defining condition init.
	/// </summary>
	/// <see cref="COND_INIT_NAME" />
	#define COND_INIT_NAME_LENGTH 9

	/// <summary>
	/// Name of item in config file defining condition destroy.
	/// </summary>
	/// <see cref="COND_DESTROY_NAME_LENGTH" />
	#define COND_DESTROY_NAME "COND_DESTROY"

	/// <summary>
	/// Length of item defining condition destroy.
	/// </summary>
	/// <see cref="COND_DESTROY_NAME" />
	#define COND_DESTROY_NAME_LENGTH 12

	/// <summary>
	/// Name of item in config file defining condition signal.
	/// </summary>
	/// <see cref="COND_SIGNAL_NAME_LENGTH" />
	#define COND_SIGNAL_NAME "COND_SIGNAL"

	/// <summary>
	/// Length of item defining condition signal.
	/// </summary>
	/// <see cref="COND_SIGNAL_NAME" />
	#define COND_SIGNAL_NAME_LENGTH 11

	/// <summary>
	/// Name of item in config file defining condition wait.
	/// </summary>
	/// <see cref="COND_WAIT_NAME_LENGTH" />
	#define COND_WAIT_NAME "COND_WAIT"

	/// <summary>
	/// Length of item defining condition wait.
	/// </summary>
	/// <see cref="COND_WAIT_NAME" />
	#define COND_WAIT_NAME_LENGTH 9

	/// <summary>
	/// Name of item in config file defining condition type.
	/// </summary>
	/// <see cref="COND_TYPE_NAME_LENGTH" />
	#define COND_TYPE_NAME "COND_TYPE"

	/// <summary>
	/// Length of item defining condition type.
	/// </summary>
	/// <see cref="COND_TYPE_NAME" />
	#define COND_TYPE_NAME_LENGTH 9

	/// <summary>
	/// Error message displayed when parsing error occurred.
	/// </summary>
	#define PARSING_ERROR_STR "Error while parsing config file: "

	/// <summary>
	/// Error message displayed when parsing error occurred while parsing concrete item.
	/// </summary>
	#define UNABLE_TO_PARSE_ERROR_STR "Unable to parse item: "

	/// <summary>
	/// String to report successful parsing.
	/// </summary>
	#define PARSER_SUCCESS_STR "Everything is OK!"

	void DeleteWhitespaceFromStr(string& str)
	{
		size_t found;

		while ((found = str.find(' ')) != string::npos)
		{
			str.replace(found, 1, "");
		}

		while ((found = str.find('\t')) != string::npos)
		{
			str.replace(found, 1, "");
		}
	}

	ThreadConfigParser::ThreadConfigParser(ostream* logger)
	{
		if (logger == NULL)
		{
			m_Logger = &cerr;
		}
		else
		{
			m_Logger = logger;
		}
	}

	ThreadConfigContent ThreadConfigParser::ParseThreadConfig(string filename)
	{
		ThreadConfigContent outputContent;

		bool createFound = false;
		bool joinFound = false;
		bool exitFound = false;
		bool typeFound = false;
		bool mutexInitFound = false;
		bool mutexDestroyFound = false;
		bool mutexLockFound = false;
		bool mutexTryLockFound = false;
		bool mutexUnlockFound = false;
		bool mutexTypeFound = false;
		bool condInitFound = false;
		bool condDestroyFound = false;
		bool condSignalFound = false;
		bool condWaitFound = false;
		bool condTypeFound = false;

		try
		{
			ifstream i(filename.c_str(), ifstream::in);

			if (i.fail())
			{
				*m_Logger << CONFIG_FILE_OPEN_ERROR_STR << filename << endl;
			}
			else
			{
				string line_str;

				while (i.good())
				{
					getline(i, line_str);

					//DeleteWhitespaceFromStr(line_str);

					// Skip comment lines
					if (line_str[0] == COMMENT_START_CHAR)
						continue;

					//	Skip empty lines
					if (line_str.compare("") == 0)
						continue;

					size_t found;
					size_t spacePos;

					// Parse CREATE
					if (!createFound && (found = line_str.find(CREATE_NAME)) != string::npos)
					{
						found += CREATE_NAME_LENGTH + 1;

						spacePos = line_str.find(" ", found);
						outputContent.createName = line_str.substr(found, spacePos - found);
						found = spacePos + 1;

						spacePos = line_str.find(" ", found);
						string threadArgnumStr = line_str.substr(found, spacePos - found);
						outputContent.createThreadArgnum = atoi(threadArgnumStr.c_str());
						found = spacePos + 1;

						spacePos = line_str.find(" ", found);
						string funcArgnumStr = line_str.substr(found, spacePos - found);
						outputContent.createFuncArgnum = atoi(funcArgnumStr.c_str());
						found = spacePos + 1;

						string dataArgnumStr = line_str.substr(found, line_str.size() - found);
						outputContent.createDataArgnum = atoi(dataArgnumStr.c_str());

						createFound = true;
						continue;
					}

					// Parse JOIN
					if (!joinFound && (found = line_str.find(JOIN_NAME)) != string::npos)
					{
						found += JOIN_NAME_LENGTH + 1;

						spacePos = line_str.find(" ", found);
						outputContent.joinName = line_str.substr(found, spacePos - found);
						found = spacePos + 1;

						string threadArgnumStr = line_str.substr(found, line_str.size() - found);
						outputContent.joinThreadArgnum = atoi(threadArgnumStr.c_str());

						joinFound = true;
						continue;
					}

					//Parse EXIT
					if (!exitFound && (found = line_str.find(EXIT_NAME)) != string::npos)
					{
						found += EXIT_NAME_LENGTH + 1;

						outputContent.exitName = line_str.substr(found, line_str.size() - found);

						exitFound = true;
						continue;
					}

					//Parse TYPE
					if (!typeFound && (found = line_str.find(TYPE_NAME)) != string::npos)
					{
						found += TYPE_NAME_LENGTH + 1;

						outputContent.typeName = line_str.substr(found, line_str.size() - found);

						typeFound = true;
						continue;
					}

					//Parse MUTEX_INIT
					if (!mutexInitFound && (found = line_str.find(MUTEX_INIT_NAME)) != string::npos)
					{
						found += MUTEX_INIT_NAME_LENGTH + 1;

						spacePos = line_str.find(" ", found);
						outputContent.mutexInitName = line_str.substr(found, spacePos - found);
						found = spacePos + 1;

						string mutexInitArgnumStr = line_str.substr(found, line_str.size() - found);
						outputContent.mutexInitArgnumStr = atoi(mutexInitArgnumStr.c_str());

						mutexInitFound = true;
						continue;
					}

					//Parse MUTEX_DESTROY
					if (!mutexDestroyFound && (found = line_str.find(MUTEX_DESTROY_NAME)) != string::npos)
					{
						found += MUTEX_DESTROY_NAME_LENGTH + 1;

						spacePos = line_str.find(" ", found);
						outputContent.mutexDestroyName = line_str.substr(found, spacePos - found);
						found = spacePos + 1;

						string mutexDestroyArgnumStr = line_str.substr(found, line_str.size() - found);
						outputContent.mutexDestroyArgnumStr = atoi(mutexDestroyArgnumStr.c_str());

						mutexDestroyFound = true;
						continue;
					}

					//Parse MUTEX_LOCK
					if (!mutexLockFound && (found = line_str.find(MUTEX_LOCK_NAME)) != string::npos)
					{
						found += MUTEX_LOCK_NAME_LENGTH + 1;

						spacePos = line_str.find(" ", found);
						outputContent.mutexLockName = line_str.substr(found, spacePos - found);
						found = spacePos + 1;

						string mutexLockArgnumStr = line_str.substr(found, line_str.size() - found);
						outputContent.mutexLockArgnumStr = atoi(mutexLockArgnumStr.c_str());

						mutexLockFound = true;
						continue;
					}

					//Parse MUTEX_TRY_LOCK
					if (!mutexTryLockFound && (found = line_str.find(MUTEX_TRY_LOCK_NAME)) != string::npos)
					{
						found += MUTEX_TRY_LOCK_NAME_LENGTH + 1;

						spacePos = line_str.find(" ", found);
						outputContent.mutexTryLockName = line_str.substr(found, spacePos - found);
						found = spacePos + 1;

						string mutexTryLockArgnumStr = line_str.substr(found, line_str.size() - found);
						outputContent.mutexTryLockArgnumStr = atoi(mutexTryLockArgnumStr.c_str());

						mutexTryLockFound = true;
						continue;
					}

					//Parse MUTEX_UNLOCK
					if (!mutexUnlockFound && (found = line_str.find(MUTEX_UNLOCK_NAME)) != string::npos)
					{
						found += MUTEX_UNLOCK_NAME_LENGTH + 1;

						spacePos = line_str.find(" ", found);
						outputContent.mutexUnlockName = line_str.substr(found, spacePos - found);
						found = spacePos + 1;

						string mutexUnlockArgnumStr = line_str.substr(found, line_str.size() - found);
						outputContent.mutexUnlockArgnumStr = atoi(mutexUnlockArgnumStr.c_str());

						mutexUnlockFound = true;
						continue;
					}

					//Parse MUTEX_TYPE
					if (!mutexTypeFound && (found = line_str.find(MUTEX_TYPE_NAME)) != string::npos)
					{
						found += MUTEX_TYPE_NAME_LENGTH + 1;

						outputContent.mutexTypeName = line_str.substr(found, line_str.size() - found);

						mutexTypeFound = true;
						continue;
					}

					//Parse COND_INIT
					if (!condInitFound && (found = line_str.find(COND_INIT_NAME)) != string::npos)
					{
						found += COND_INIT_NAME_LENGTH + 1;

						spacePos = line_str.find(" ", found);
						outputContent.condInitName = line_str.substr(found, spacePos - found);
						found = spacePos + 1;

						string condInitArgnumStr = line_str.substr(found, line_str.size() - found);
						outputContent.condInitArgnumStr = atoi(condInitArgnumStr.c_str());

						condInitFound = true;
						continue;
					}

					//Parse COND_DESTROY
					if (!condDestroyFound && (found = line_str.find(COND_DESTROY_NAME)) != string::npos)
					{
						found += COND_DESTROY_NAME_LENGTH + 1;

						spacePos = line_str.find(" ", found);
						outputContent.condDestroyName = line_str.substr(found, spacePos - found);
						found = spacePos + 1;

						string condDestroyArgnumStr = line_str.substr(found, line_str.size() - found);
						outputContent.condDestroyArgnumStr = atoi(condDestroyArgnumStr.c_str());

						condDestroyFound = true;
						continue;
					}

					//Parse COND_SIGNAL
					if (!condSignalFound && (found = line_str.find(COND_SIGNAL_NAME)) != string::npos)
					{
						found += COND_SIGNAL_NAME_LENGTH + 1;

						spacePos = line_str.find(" ", found);
						outputContent.condSignalName = line_str.substr(found, spacePos - found);
						found = spacePos + 1;

						string condSignalArgnumStr = line_str.substr(found, line_str.size() - found);
						outputContent.condSignalArgnumStr = atoi(condSignalArgnumStr.c_str());

						condSignalFound = true;
						continue;
					}

					//Parse COND_WAIT
					if (!condWaitFound && (found = line_str.find(COND_WAIT_NAME)) != string::npos)
					{
						found += COND_WAIT_NAME_LENGTH + 1;

						spacePos = line_str.find(" ", found);
						outputContent.condWaitName = line_str.substr(found, spacePos - found);
						found = spacePos + 1;

						string condWaitArgnumStr = line_str.substr(found, line_str.size() - found);
						outputContent.condWaitArgnumStr = atoi(condWaitArgnumStr.c_str());

						condWaitFound = true;
						continue;
					}

					//Parse COND_TYPE
					if (!condTypeFound && (found = line_str.find(COND_TYPE_NAME)) != string::npos)
					{
						found += COND_TYPE_NAME_LENGTH + 1;

						spacePos = line_str.find(" ", found);
						outputContent.condTypeName = line_str.substr(found, spacePos - found);

						condTypeFound = true;
						continue;
					}
				}

				i.close();
			}
		}
		catch(const exception &e)
		{
			*m_Logger << UNEXPECTED_PARSER_ERROR_STR << e.what() << endl;
		}

		ReportNotFound(createFound, joinFound, exitFound, typeFound,
				mutexInitFound, mutexDestroyFound, mutexLockFound,
				mutexTryLockFound, mutexUnlockFound, mutexTypeFound,
				condInitFound, condDestroyFound,
				condSignalFound, condWaitFound, condTypeFound);

		return (outputContent);
	}

	void ThreadConfigParser::ReportNotFound(bool createFound, bool joinFound,
											bool exitFound,	bool typeFound,
											bool mutexInitFound, bool mutexDestroyFound,
											bool mutexLockFound,bool mutexTryLockFound,
											bool mutexUnlockFound, bool mutexTypeFound,
											bool condInitFound, bool condDestroyFound,
											bool condSignalFound, bool condWaitFound,
											bool condTypeFound)
	{
		bool allOk = true;

		if (!createFound)
		{
			allOk = false;
			*m_Logger << PARSING_ERROR_STR << UNABLE_TO_PARSE_ERROR_STR << CREATE_NAME << endl;
		}
		if (!joinFound)
		{
			allOk = false;
			*m_Logger << PARSING_ERROR_STR << UNABLE_TO_PARSE_ERROR_STR << JOIN_NAME << endl;
		}
		if (!exitFound)
		{
			allOk = false;
			*m_Logger << PARSING_ERROR_STR << UNABLE_TO_PARSE_ERROR_STR << EXIT_NAME << endl;
		}
		if (!typeFound)
		{
			allOk = false;
			*m_Logger << PARSING_ERROR_STR << UNABLE_TO_PARSE_ERROR_STR << TYPE_NAME << endl;
		}
		if (!mutexInitFound)
		{
			allOk = false;
			*m_Logger << PARSING_ERROR_STR << UNABLE_TO_PARSE_ERROR_STR << MUTEX_INIT_NAME << endl;
		}
		if (!mutexDestroyFound)
		{
			allOk = false;
			*m_Logger << PARSING_ERROR_STR << UNABLE_TO_PARSE_ERROR_STR << MUTEX_DESTROY_NAME << endl;
		}
		if (!mutexLockFound)
		{
			allOk = false;
			*m_Logger << PARSING_ERROR_STR << UNABLE_TO_PARSE_ERROR_STR << MUTEX_LOCK_NAME << endl;
		}
		if (!mutexTryLockFound)
		{
			allOk = false;
			*m_Logger << PARSING_ERROR_STR << UNABLE_TO_PARSE_ERROR_STR << MUTEX_TRY_LOCK_NAME << endl;
		}
		if (!mutexUnlockFound)
		{
			allOk = false;
			*m_Logger << PARSING_ERROR_STR << UNABLE_TO_PARSE_ERROR_STR << MUTEX_UNLOCK_NAME << endl;
		}
		if (!mutexTypeFound)
		{
			allOk = false;
			*m_Logger << PARSING_ERROR_STR << UNABLE_TO_PARSE_ERROR_STR << MUTEX_TYPE_NAME << endl;
		}
		if (!condInitFound)
		{
			allOk = false;
			*m_Logger << PARSING_ERROR_STR << UNABLE_TO_PARSE_ERROR_STR << COND_INIT_NAME << endl;
		}
		if (!condDestroyFound)
		{
			allOk = false;
			*m_Logger << PARSING_ERROR_STR << UNABLE_TO_PARSE_ERROR_STR << COND_DESTROY_NAME << endl;
		}
		if (!condSignalFound)
		{
			allOk = false;
			*m_Logger << PARSING_ERROR_STR << UNABLE_TO_PARSE_ERROR_STR << COND_SIGNAL_NAME << endl;
		}
		if (!condWaitFound)
		{
			allOk = false;
			*m_Logger << PARSING_ERROR_STR << UNABLE_TO_PARSE_ERROR_STR << COND_WAIT_NAME << endl;
		}
		if (!condTypeFound)
		{
			allOk = false;
			*m_Logger << PARSING_ERROR_STR << UNABLE_TO_PARSE_ERROR_STR << COND_TYPE_NAME << endl;
		}


		if (allOk)
		{
			*m_Logger << PARSER_SUCCESS_STR << endl;
		}
	}
}
