// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file ThreadConfigParser.h
 * Declaration of class for parsing thread configuration file.
 */
#ifndef GIT_THREAD_CONFIG_PARSER_H
#define GIT_THREAD_CONFIG_PARSER_H

#include <iostream>

using namespace std;

namespace giterator
{
	/**
	 * Struct to store information stored in thread configuration file.
	 */
	struct ThreadConfigContent
	{
		/**
		 * Check whether a specified name was in thread configuration file
		 *
		 * @param name the name to check
		 * @return true if @c name was in thread config file, false otherwise
		 */
		bool containsName(string name)
		{
			if (createName.compare(name) == 0 ||
					joinName.compare(name) == 0 ||
					exitName.compare(name) == 0 ||
					typeName.compare(name) == 0 ||
					mutexInitName.compare(name) == 0 ||
					mutexDestroyName.compare(name) == 0 ||
					mutexLockName.compare(name) == 0 ||
					mutexTryLockName.compare(name) == 0 ||
					mutexUnlockName.compare(name) == 0 ||
					mutexTypeName.compare(name) == 0 ||
					condInitName.compare(name) == 0 ||
					condDestroyName.compare(name) == 0 ||
					condWaitName.compare(name) == 0 ||
					condSignalName.compare(name) == 0 ||
					condTypeName.compare(name) == 0)
				return true;

			return false;
		}

		/** @name Names of various supported thread functions supported by GMC.
		 * Order of the arguments of the functions is also specified
		 * by the *Argnum variables.
		 */
		/*@{*/
		string createName, joinName, exitName, typeName;
		string mutexInitName, mutexDestroyName, mutexLockName, mutexTryLockName, mutexUnlockName, mutexTypeName;
		string condInitName, condDestroyName, condWaitName, condSignalName, condTypeName;

		int createThreadArgnum, createFuncArgnum, createDataArgnum;
		int joinThreadArgnum;
		int mutexInitArgnumStr, mutexDestroyArgnumStr, mutexLockArgnumStr, mutexTryLockArgnumStr, mutexUnlockArgnumStr;
		int condInitArgnumStr, condDestroyArgnumStr, condWaitArgnumStr, condSignalArgnumStr;
		/*@}*/
	};

	/**
	 * Class for parsing thread configuration file.
	 * Configuration file stores some information about supported threads inside GMC.
	 * Class is written for easy extension.
	 */
	class ThreadConfigParser
	{
	public:
		/**
		 * Constructor, initializes logger.
		 * @param logger Determines where to write error and output messages.
		 * Argument is optional, if not set, std::cerr is used.
		 */
		ThreadConfigParser(ostream* logger = NULL);

		/**
		 * Parses thread configuration file.
		 * Searches for items that should be defined in configuration file.
		 * If any item is not found, reports warning.
		 *
		 * param filename Input thread configuration file to parse.
		 * @return Structure filled with information stored in thread configuration file.
		 */
		ThreadConfigContent ParseThreadConfig(string filename);
	private:

		/**
		 * Report what items, that should have been defined in config file were not found.
		 * </suumary>
		 * @param createFound Was create config file path  item found in config file?
		 * @param joinFound Was join item found in config file?
		 * @param exitFound Was exit item found in config file?
		 * @param typeFound Was type item found in config file?
		 * @param mutexInitFound Was mutex create item found in config file?
		 * @param mutexDestroyFound Was mutex destroy item found in config file?
		 * @param mutexLockFound Was mutex lock item found in config file?
		 * @param mutexTryLockFound Was mutex try lock item found in config file?
		 * @param mutexUnlockFound Was mutex unlock item found in config file?
		 * @param mutexTypeFound Was mutex type item found in config file?
		 * @param condInitFound Was thread condition init item found in config file?
		 * @param condDestroyFound Was thread condition destroy item found in config file?
		 * @param condSignalFound Was thread conditon signal item found in config file?
		 * @param condWaitFound Was thread conditon wait item found in config file?
		 * @param condTypeFound Was thread condition type item found in config file?
		 */
		void ReportNotFound(bool createFound,
							bool joinFound,
							bool exitFound,
							bool typeFound,
							bool mutexInitFound,
							bool mutexDestroyFound,
							bool mutexLockFound,
							bool mutexTryLockFound,
							bool mutexUnlockFound,
							bool mutexTypeFound,
							bool condInitFound,
							bool condDestroyFound,
							bool condSignalFound,
							bool condWaitFound,
							bool condTypeFound);

		/*! Determines where to write error and output messages.*/
		ostream* m_Logger;
	};
}


#endif /* GIT_THREAD_CONFIG_PARSER_H */
