// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file Type.h
 * Interface for classes representing types.
 */
#ifndef GIT_TYPE_H
#define GIT_TYPE_H

#include "TreeProduct.h"
#include "../mm/fwd.h"

#include <string>

namespace giterator {
    
/**
 * Interface for classes representing types
 * Types get created from GIMPLE @c *_TYPE nodes.
 * Subclasses of this interface represent concrete
 * GIMPLE types.
 *
 * Each ..._TYPE node has:
 * TYPE_SIZE which is the number of bits representing the type.
 * TYPE_NAME which stores name of the type - but only in small
 * amount of cases (often is NULL), therefore GMC has its own
 * mechanism of assigning names to types.
 * This information is stored inside every type.
 * Type has to be capable of computing its own unique hash for
 * state equivalence checking purposes in model checking.
 *
 * @see create_git_type_from_tree() in gimplexx-bridge.c
 * to see how types get created from GIMPLE types
 */
class Type : public detail::TreeProduct {
  private:
    /*! number of bits in a byte */
    static const size_t bitsPerByte = 8;
    
  public:
    /**
     * Destructor.
     */
    virtual ~Type() {}

    /**
     * Returns the size of the values of this type in bits.
     * 
     * @return the size of the values of this type in bits
     */
    virtual size_t getBitSize() const = 0;
    
    /**
     * Returns the size of the values of this type in bytes.
     * The returned value is counted from @c getBitSize(). 
     * 
     * @return the size of the values of this type in bytes
     */
    virtual size_t getUnitSize() const
    { 
      return (getBitSize()+bitsPerByte-1)/bitsPerByte; 
    }

    /**
     * Returns the hash of this type.
     * This is used for state equivalence checking purposes
     * in model checking.
     * 
     * @return hash of this type
     */
    virtual mmodule::Hash getHash () const = 0;
    
    /**
     * The name of the type.
     * 
     * @return name of the type
     */
    virtual std::string getName() const = 0;
    
    /**
     * Checks if this type is the same type as @c type.
     * 
     * @return @c true if this type is the same type as @c type, @c false otherwise
     */
    virtual bool isSame (const Type& type) const = 0;
};

/**
 * Checks if two types are the same.
 * 
 * @param t1 first type to compare
 * @param t2 second type to compare
 */
inline bool 
compareTypes (const Type* t1, const Type* t2)
{
    return t1->isSame (*t2);
}

}


#endif  /* GIT_TYPE_H */

