// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file TypeDecl.h
 * Class representing declaration of a type(typedef).
 */
#ifndef GIT_TYPE_DECL_H
#define GIT_TYPE_DECL_H

#include "AbstractDecl.h"
#include "Type.h"

#include <boost/serialization/serialization.hpp>

namespace giterator {

/**
 * Class representing declaration of a type (typedef).
 * Type declarations get created from GIMPLE @c TYPE_DECL nodes.
 *
 * @see create_git_decl_from_tree() to see how declarations of
 * types get created from GIMPLE declarations
 */
class TypeDecl : public detail::AbstractDecl {
  private:
	/*! whether the declaration is local or global. */
	bool m_isLocal;
  public:
	/**
	 * Constructor, creates a new type declaration.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	TypeDecl () : detail::AbstractDecl() {}

	/**
	 * Constructor, creates a new type declaration.
	 *
     * @param uid unique ID of the new declaration
     * @param name name of the new declaration
     * @param type type of the new declaration
     * @param isLocal whether the declaration is local or global
     */
    TypeDecl (unsigned int uid, const std::string& name, const Type* type, bool isLocal) :
                detail::AbstractDecl (uid, name, type), m_isLocal(isLocal) {}

    void accept (Visitor& visitor) const
    { visitor.visit (*this); }

    /**
     * Find out whether the declaration is local or global.
     *
     * @return true if the declaration is local, false otherwise
     */
    bool getLocal () const
	{ return m_isLocal; }

    friend class boost::serialization::access;
	template<class Archive>
	void save(Archive & ar, const unsigned int) const
	{
		ar << boost::serialization::base_object<AbstractDecl>(*this);
		ar << m_isLocal;
	}

	template<class Archive>
	void load(Archive & ar, const unsigned int)
	{
		ar >> boost::serialization::base_object<AbstractDecl>(*this);
		ar >> m_isLocal;
	}

	template<class Archive>
	void serialize(Archive & ar, const unsigned int version)
	{
		boost::serialization::void_cast_register(static_cast<TypeDecl *>(NULL), static_cast<TreeProduct *>(NULL));
		boost::serialization::void_cast_register(static_cast<TypeDecl *>(NULL), static_cast<Decl *>(NULL));
		boost::serialization::split_member(ar, *this, version);
	}
};

}

#endif /* GIT_TYPE_DECL_H */
