// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file VarDecl.h
 * Class representing declaration of a variable.
 */
#ifndef GIT_VAR_DECL_H
#define GIT_VAR_DECL_H

#include "OffsetDecl.h"
#include "Type.h"

#include <boost/serialization/serialization.hpp>

namespace giterator { 

class OffsetMapper;    

/**
 * Class representing declaration of a variable.
 *
 * Variable declarations get created from GIMPLE @c VAR_DECL
 * nodes. These nodes represent variables with block scope,
 * as well as static data members. The DECL_SIZE is the number of
 * bits representing the declaration.
 * If this variable is global and initialized (but does not require a constructor),
 * its constructor or initialization value is appended to the beginning of the first
 * block of main function.
 *
 * @see create_git_decl_from_tree() in gimplexx-bridge.c
 * to see how declarations of variables get created
 * from GIMPLE declarations
 */
class VarDecl : public OffsetDecl {
  private:
	/*! Whether the declaration is local or global. */
	bool m_local;
  public:
	/**
	 * Constructor, creates a new variable declaration.
	 * Empty constructor exists for serialization purposes,
	 * use nonempty for normal use.
	 */
	VarDecl () : OffsetDecl() {}

	/**
	 * Constructor, creates a new variable declaration.
	 *
     * @param uid unique ID of the new declaration
     * @param name name of the new declaration
     * @param type type of the new declaration
     * @param local whether the declaration is local or global
     */
    VarDecl (unsigned int uid, const std::string& name, const Type* type, bool local)
    : OffsetDecl (uid, name, type), m_local(local) {}
    
    void accept (Visitor& visitor) const
    { visitor.visit (*this); }

    /**
     * Find out whether the declaration is local or global.
     *
     * @return true if the declaration is local, false otherwise
     */
    bool getLocal() const
    { return m_local; }

    friend class boost::serialization::access;

	template<class Archive>
	void serialize(Archive & ar, const unsigned int version)
	{
		boost::serialization::void_cast_register(static_cast<VarDecl *>(NULL), static_cast<TreeProduct *>(NULL));
		boost::serialization::void_cast_register(static_cast<VarDecl *>(NULL), static_cast<Decl *>(NULL));
		ar & boost::serialization::base_object<OffsetDecl>(*this);
		ar & m_local;
	}
};

}

#endif /* GIT_VAR_DECL_H */

