// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file AbstractValue.cpp
 * Defintions for base class for classes representing values.
 */
#include "AbstractValue.h"

#include "../Error.h"
#include "../git/Type.h"

#include <ostream>
   
namespace ginterpreter {
    
static void opNiError (const char* opStr, const Value& op1, const Value& op2) ATTRIBUTE_NORETURN;
static void opNiError (const char* opStr, const Value& op) ATTRIBUTE_NORETURN;

static void
opNiError (const char* opStr, const Value& op1, const Value& op2) 
{
    ni_error ("operator `%s' on values `%s' and `%s'", opStr, 
              op1.getType()->getName().c_str(), op2.getType()->getName().c_str()); 
}

static void
opNiError (const char* opStr, const Value& op)
{
    ni_error ("operator `%s' on value `%s'", opStr,
              op.getType()->getName().c_str());
}


/**
 * Prints this value.
 * The default implementation prints the address of the object representing the value.
 */
void
AbstractValue::print (std::ostream& os) const
{ 
    os << "->" << ((void*)this) << "<-"; 
}


bool 
AbstractValue::operator==(const Value& v) const
{ 
  opNiError ("==", *this, v); 
}

    
bool 
AbstractValue::operator<(const Value& v) const 
{ 
  opNiError ("<", *this, v); 
}


bool
AbstractValue::operator<=(const Value& v) const
{ 
  return !(v < *this); 
}

    
bool 
AbstractValue::operator>(const Value& v) const
{ 
  return v < *this; 
}

   
bool 
AbstractValue::operator>=(const Value& v) const
{ 
  return !(*this < v); 
}


bool 
AbstractValue::operator!=(const Value& v) const
{ 
  return !(*this == v); 
}

bool
AbstractValue::orderedWith(const Value& v) const
{
	opNiError("orderedWith", *this, v);
}

    

std::auto_ptr<Value> 
AbstractValue::addOp (const Value& v) const
{
    opNiError ("add", *this, v);
}


std::auto_ptr<Value> 
AbstractValue::subOp (const Value& v) const
{
    opNiError ("sub", *this, v);
}


std::auto_ptr<Value> 
AbstractValue::multOp (const Value& v) const
{
    opNiError ("mult", *this, v);
}

        
std::auto_ptr<Value>
AbstractValue::truncModOp (const Value& v) const
{
    opNiError ("truncMod", *this, v);
}

std::auto_ptr<Value>
AbstractValue::ceilModOp (const Value& v) const
{
    opNiError ("ceilMod", *this, v);
}

std::auto_ptr<Value>
AbstractValue::floorModOp (const Value& v) const
{
    opNiError ("floorMod", *this, v);
}

std::auto_ptr<Value>
AbstractValue::roundModOp (const Value& v) const
{
    opNiError ("roundMod", *this, v);
}




std::auto_ptr<Value> 
AbstractValue::truncDivOp (const Value& v) const
{
    opNiError ("truncDiv", *this, v);
}

std::auto_ptr<Value>
AbstractValue::ceilDivOp (const Value& v) const
{
    opNiError ("ceilDiv", *this, v);
}

std::auto_ptr<Value>
AbstractValue::floorDivOp (const Value& v) const
{
    opNiError ("floorDiv", *this, v);
}

std::auto_ptr<Value>
AbstractValue::roundDivOp (const Value& v) const
{
    opNiError ("roundDiv", *this, v);
}

std::auto_ptr<Value>
AbstractValue::rDivOp (const Value& v) const
{
    opNiError ("rDiv", *this, v);
}

std::auto_ptr<Value>
AbstractValue::lShiftOp (const Value& v) const
{
    opNiError ("lShiftOp", *this, v);
}
std::auto_ptr<Value>
AbstractValue::rShiftOp (const Value& v) const
{
    opNiError ("rShiftOp", *this, v);
}
std::auto_ptr<Value>
AbstractValue::lRotateOp (const Value& v) const
{
    opNiError ("lRotateOp", *this, v);
}
std::auto_ptr<Value>
AbstractValue::rRotateOp (const Value& v) const
{
    opNiError ("rRotateOp", *this, v);
}
std::auto_ptr<Value>
AbstractValue::bitNotOp () const
{
    opNiError ("bitNotOp", *this);
}
std::auto_ptr<Value>
AbstractValue::bitIorOp (const Value& v) const
{
    opNiError ("bitIorOp", *this, v);
}
std::auto_ptr<Value>
AbstractValue::bitXorOp (const Value& v) const
{
    opNiError ("bitIorOp", *this, v);
}
std::auto_ptr<Value>
AbstractValue::bitAndOp (const Value& v) const
{
    opNiError ("bitAndOp", *this, v);
}

std::auto_ptr<Value>
AbstractValue::truthNotOp () const
{
    opNiError ("truthNotOp", *this);
}
std::auto_ptr<Value>
AbstractValue::truthOrOp (const Value& v) const
{
    opNiError ("truthIorOp", *this, v);
}
std::auto_ptr<Value>
AbstractValue::truthXorOp (const Value& v) const
{
    opNiError ("truthIorOp", *this, v);
}
std::auto_ptr<Value>
AbstractValue::truthAndOp (const Value& v) const
{
    opNiError ("truthAndOp", *this, v);
}

std::auto_ptr<Value>
AbstractValue::negateOp () const
{
    opNiError ("negateOp", *this);
}

std::auto_ptr<Value>
AbstractValue::absOp () const
{
    opNiError ("abs", *this);
}

    
mmodule::Hash 
AbstractPtrValue::getTypeHash () const
{
    return getType ()->getHash ();
}

}


