// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file AbstractValue.h
 * Declaration of base class for classes representing values.
 */
#ifndef INT_ABSTRACT_VALUE_H
#define INT_ABSTRACT_VALUE_H

#include "Value.h"

#include "../mm/Content.h"
#include "../git/Type.h"

#include <memory>

namespace ginterpreter {

/**
 * The base class for classes that represent values.
 * It implements all the operators from @c Value
 * class to report @c ni_error. Subclasses should
 * re-implement those operands that are defined
 * on the value the subclass represents.
 */
class AbstractValue : public Value {
  private:
    /*! type of this value */
    const giterator::Type* m_type;

  protected:
    /**
     * Constructor.
     * 
     * @param type type of the value
     */
    AbstractValue (const giterator::Type* type) : m_type (type) {}

  public:
    /**
     * Returns the type of this value.
     */
    const giterator::Type* getType() const 
    { return m_type; }
    
    /**
     * Returns the size of this value in bytes.
     *
     * @return the size of this value in bytes
     */
    size_t getUnitSize () const
    { return getType ()->getUnitSize (); }
    
    /**
     * Write the value to a stream.
     *
     * @param os the stream to write the value to
     */
    void print (std::ostream& os) const;
    
    /** @name Relation operators
     * The relation operators on value.
     * All the operator are implemented to call @c ni_error().
     */
    //@{
    virtual bool operator==(const Value& v) const;
    virtual bool operator<(const Value& v) const;

    virtual bool operator<=(const Value& v) const;
    virtual bool operator>(const Value& v) const;
    virtual bool operator>=(const Value& v) const;
    virtual bool operator!=(const Value& v) const;    
    virtual bool orderedWith(const Value& v) const;
    //@}
    
    /** @name Arithmetic and logical operators
     * The arithmetic operators and logical on the value.
     * These operators correspond to the arithmetic and
     * logical operators used in GIMPLE.
     * All the operator are implemented to call @c ni_error().
     */
    //@{
    virtual std::auto_ptr<Value> addOp (const Value& v) const;
    virtual std::auto_ptr<Value> subOp (const Value& v) const;
    virtual std::auto_ptr<Value> multOp (const Value& v) const;
    virtual std::auto_ptr<Value> truncModOp (const Value& v) const;
	virtual std::auto_ptr<Value> floorModOp (const Value& v) const;
	virtual std::auto_ptr<Value> ceilModOp (const Value& v) const;
	virtual std::auto_ptr<Value> roundModOp (const Value& v) const;
	virtual std::auto_ptr<Value> truncDivOp (const Value& v) const;
	virtual std::auto_ptr<Value> floorDivOp (const Value& v) const;
	virtual std::auto_ptr<Value> ceilDivOp (const Value& v) const;
	virtual std::auto_ptr<Value> roundDivOp (const Value& v) const;
    virtual std::auto_ptr<Value> rDivOp (const Value& v) const;
    virtual std::auto_ptr<Value> lShiftOp (const Value& v) const;
	virtual std::auto_ptr<Value> rShiftOp (const Value& v) const;
	virtual std::auto_ptr<Value> lRotateOp (const Value& v) const;
	virtual std::auto_ptr<Value> rRotateOp (const Value& v) const;
	virtual std::auto_ptr<Value> bitNotOp () const;
	virtual std::auto_ptr<Value> bitIorOp (const Value& v) const;
	virtual std::auto_ptr<Value> bitXorOp (const Value& v) const;
	virtual std::auto_ptr<Value> bitAndOp (const Value& v) const;
	virtual std::auto_ptr<Value> truthNotOp () const;
	virtual std::auto_ptr<Value> truthOrOp (const Value& v) const;
	virtual std::auto_ptr<Value> truthXorOp (const Value& v) const;
	virtual std::auto_ptr<Value> truthAndOp (const Value& v) const;
	virtual std::auto_ptr<Value> negateOp () const;
    virtual std::auto_ptr<Value> absOp () const;
    //@}
};


/**
 * Base class for all classes representing pointer values.
 */
struct AbstractPtrValue : AbstractValue, mmodule::PtrContent {
	/**
	 * Constructor.
	 *
	 * @param type the type of the new value
	 */
    AbstractPtrValue (const giterator::Type* type) : AbstractValue (type) {}
    
    /**
     * Get the hash of the type of the value.
     * For memory state equivalency purposes.
     */
    mmodule::Hash getTypeHash () const;
    
    /**
     * Destructor.
     */
    virtual ~AbstractPtrValue () {}
};


/**
 * Base class for all classes representing non-pointer values.
 */
struct AbstractNonPtrValue : AbstractValue, mmodule::NonPtrContent {
	/**
	 * Constructor.
	 *
	 * @param type the type of the new value
	 */
    AbstractNonPtrValue (const giterator::Type* type) : AbstractValue (type) {}
    
    /**
	 * Destructor.
	 */
    virtual ~AbstractNonPtrValue () {}
};

}

#endif /* INT_ABSTRACT_VALUE_H */
