// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file AddrResolver.h
 * Declaration of class finding addresses of values represented by operand.
 */
#ifndef INT_ADDR_RESOLVER_H
#define INT_ADDR_RESOLVER_H

#include "../git/git_fwd.h"
#include "../git/Operand.h"
#include "../git/Decl.h"
#include "../mm/fwd.h"
#include "../mm/Ptr.h"

namespace ginterpreter {
    
class FunctionValue;
class ConstantsManager;
class GlobalsManager;

/**
 * This class finds the address of the value that is represented by the given operand.
 * To get the address of the value represented by operand @c o, an object of this class 
 * is supplied as a parameter to the @c Operand::accept() method of object @c o.
 * The found address and the type of the value can be acquired 
 * by call to @c getPtr() and @c getTargetType().
 * This class implements the Visitor interfaces for Operand and Declaration class.
 * The implemented visitor methods realize the search for the addresses.
 */
class AddrResolver : public giterator::Operand::Visitor, public giterator::Decl::Visitor 
{
  private:
    /*! Current function */
    const FunctionValue* m_func;
    /*! Interface to the simulated memory */
    const mmodule::MemoryState* m_memoryState;
    /*! Constants and function declarations of the program */
    const ConstantsManager* m_cstManager;
    /*! Global declarations of the program */
    const GlobalsManager* m_globalsManager;
    /*! pointer to the current stack frame of the current function */
    mmodule::Ptr m_frame;
    /*! the type of the value which address was found */
    const giterator::Type* m_type;
    /*! pointer to the found value */
    mmodule::Ptr m_ptr;
    /** Offset of the value in the function stack frame for local variables
     * or in the global pool for global variables.
     */
    size_t m_declOffset;
    /**
     * For declarations, the found declaration.
     */
    const giterator::OffsetDecl* m_decl;
  public:

    /**
     * Constructor, creates a new address resolver.
     *
     * @param func currently interpreted function of the program, where to look
     * for local values
     * @param ms interface to the simulated memory
     * @param cstManager constants and function declarations of the program
     * @param globalsManager global declarations of the program
     * @param frame pointer to the stack frame of the current function
     */
    AddrResolver (const FunctionValue* func, const mmodule::MemoryState* ms, 
                  const ConstantsManager* cstManager, const GlobalsManager* globalsManager,
                  mmodule::Ptr frame);
    /**
     * Returns the type of the value which address was found. 
     *
     * @return the type of the value which address was found
     */
    const giterator::Type* getTargetType () const
    { return m_type; }
    /**
     * Returns the address of the found value.
     *
     * @return the address of the found value
     */
    mmodule::Ptr getPtr () const
    { return m_ptr; }
    /**
     * Get offset of the found value in the function stack frame for
     * local variables or in the global pool for global variables.
     *
     * @return offset of the found value
     */
    size_t getDeclOffset() const
    { return m_declOffset; }

    /** @name Visitor methods
     * Methods for resolving addresses of operands and declarations
     */
    /*@{*/
    void visit (const giterator::AddrOperand&);
    void visit (const giterator::DeclOperand&);
    void visit (const giterator::LabelExprOperand&);
    void visit (const giterator::IndirectRefOperand&);
    void visit (const giterator::ArrayRefOperand&);
    void visit (const giterator::ComponentRefOperand&);
    void visit (const giterator::CstOperand&);
    void visit (const giterator::ArrayConstructorOperand&);
	void visit (const giterator::RecordConstructorOperand&);
                 
    void visit (const giterator::ParamDecl& decl);
    void visit (const giterator::VarDecl& decl);
    void visit (const giterator::RecordDecl& decl);
    void visit (const giterator::FunctionDecl& decl);
    void visit (const giterator::FieldDecl& decl);
	void visit (const giterator::LabelDecl& decl);
	void visit (const giterator::ResultDecl& decl);
	void visit (const giterator::TypeDecl& decl);
	void visit (const giterator::ConstDecl& decl);
	/*@}*/
  private:
	/**
	 * Resolve the address of a local declaration.
	 *
	 * @param decl the declaration for which to resolve
	 * the address
	 */
    void getLocalPtr_ (const giterator::Decl& decl);
    /**
	 * Resolve the address of a global declaration.
	 *
	 * @param decl the declaration for which to resolve
	 * the address
	 */
    void getGlobalPtr_ (const giterator::Decl& decl);
};

}

#endif /* INT_ADDR_RESOLVER_H */
