// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file ArrayConstructorValue.h
 * Class representing simulated value of an array constructor.
 */
#ifndef INT_ARRAY_CONSTRUCTOR_VALUE_H
#define INT_ARRAY_CONSTRUCTOR_VALUE_H

#include "AbstractValue.h"

#include <boost/ptr_container/ptr_map.hpp>
#include <boost/ptr_container/ptr_vector.hpp>
#include <memory>

namespace ginterpreter {

/**
 * Class representing a simulated array constructor.
 * This class is a regular simulated value, so it can be stored in the simulated memory.
 * In GMC however, ArrayConstructorValue is not stored to the simulated memory.
 * It is only a helper class that is returned from @c ValueReader
 * when reading an @c ArrayConstructorOperand. It is composed
 * of the values of the constructors for the indexes of the array.
 * @c ValueWriter decomposes it back into the constructors for the indexes
 * and stores them into simulated memory one by one.
 */
class ArrayConstructorValue : public AbstractNonPtrValue {
  private:
	/*! type of the container holding the values of constructors for indexes of an array. */
    typedef boost::ptr_map<const Value*, Value> Values;
    /*! type of value holder container */
    typedef boost::ptr_vector<Value> ValueOwners;
    /*! values of constructors for indexes of an array. */
    Values m_values;
    /*! owners of some values that are not owned anywhere else */
    ValueOwners m_valueOwners;
  public:
    /*! iterator for the constructors values container */
    typedef Values::const_iterator ValuesIterator;
    /**
     * Constructor.
     *
     * @param type GIMPLE++ representation of the type of the constructed
     * array
     */
    ArrayConstructorValue (const giterator::Type* type);

    /**
     * Clone this value.
     *
     * @returns an exact copy of this value.
     */
    Value* clone () const;

    /**
     * Get a hash of this value.
     * The hash is combined of the hashes of the sub-values.
     * For memory state equivalency check purposes.
     *
     * @return hash of this value.
     */
    mmodule::Hash getHash () const;

    /**
     * Add a value of a constructor for an index of an array to this value.
     *
     * @param index of array which the added value constructs
     * @param constructor @c constructor of the index @c index
     * in an array
     */
    void addValue(const Value* index, std::auto_ptr<Value> constructor);

    /**
     * Set a value to be owned by this class.
     *
     * @param value value to be owned by this class
     */
    void addValueOwner(std::auto_ptr<Value> value);

    /**
     * Obtain all the values of this constructor.
     * i.e. value of constructors for all the indexes
     * of the array.
     *
     * @return pair iterators <first value iterator,
     * 						   last value iterator>
     */
    std::pair<ValuesIterator, ValuesIterator> values() const;
};

}

#endif /* INT_ARRAY_CONSTRUCTOR_VALUE_H */
