// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file BuiltinFunctionValue.h
 * Declaration of class representing simulated builtin function.
 */
#ifndef INT_BUILTIN_FUNCTION_VALUE_H
#define INT_BUILTIN_FUNCTION_VALUE_H

#include "AbstractValue.h"

#include "../git/FunctionInfo.h"
#include "../git/CallStmt.h"    // TODO remove this include

namespace ginterpreter {

class ConstantsManager;
class GlobalsManager;
class FunctionValue;

/**
 * Class representing simulated built-in function.
 * This class is a regular simulated value, so it can be stored in the memory.
 * It stores all the information needed to interpret the function as well
 * as pointer to the method implementing the function.
 */
class BuiltinFunctionValue : public AbstractNonPtrValue {
  public:
    /*! type of the pointer to the function, that implements this simulated function */
    typedef std::auto_ptr<Value> (*FuncMethod) (
        const FunctionValue* currFunc, mmodule::MemoryState* memoryState, 
        ConstantsManager* cstManager, GlobalsManager* globalsManager,
        mmodule::Ptr topFrame,
        std::pair<giterator::CallStmt::ParameterIterator,     
                  giterator::CallStmt::ParameterIterator> parameters, 
        const giterator::Type* returnType);
  private:
    /*! the code of this built-in function */
    enum giterator::function_code m_code;
    /*! pointer to the function implementing this built-in function */
    FuncMethod m_method;
  public:
	/**
	* Constructor.
	*
	* @param code code of the built-in function
	* @param type type of the built-in function
	* @param method pointer to the function implementing the built-in function
	*/
	BuiltinFunctionValue (enum giterator::function_code code,
						const giterator::FunctionType* type,
						FuncMethod method);
	/**
	 * Clone this value.
	 * Aborts, function values should not be copied.
	 */
    Value* clone() const;
    /**
	 * Get a hash of this value.
	 * For memory state equivalence check purposes.
	 * Hashes the code of this function.
	 *
	 * @return hash of this value.
	 */
    mmodule::Hash getHash () const;
    /**
     * Get the code of this function.
     *
     * @return code of this function
     */
    enum giterator::function_code getCode () const;
    /**
     * Get the type of this function.
     *
     * @return the type of this function.
     */
    const giterator::FunctionType* getFunctionType () const;
    /**
     * Get the pointer to the function implementing this built-in function.
     *
     * @return pointer to the function implementing this built-in function
     */
    FuncMethod getMethod () const;
};

}

#endif /* INT_BUILTIN_FUNCTION_VALUE_H */
