// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file BuiltinFunctions.cpp
 * This file contains definitions of functions implementing simulated built-in functions.
 */
#include "BuiltinFunctions.h"

#include "ValueReader.h"
#include "ValueWriter.h"
#include "PtrValue.h"
#include "IntegerValue.h"
#include "StringValue.h"
#include "RealValue.h"
#include "FunctionValue.h"
#include "NameResolver.h"

#include "Thread.h"
#include "Program.h"

#include "InterpretExceptions.h"
#include "ModelCheckerExceptions.h"

#include "../mm/MemoryState.h"
#include "../mm/MemoryExceptions.h"

#include "../git/PtrType.h"
#include "../git/IntegerType.h"
#include "../git/Function.h"

#include <string>
#include <sstream>
#include <iostream>
#include <cctype>

using namespace mmodule;
using namespace giterator;


namespace ginterpreter {

const Value* readRawParam(giterator::CallStmt::ParameterIterator param, ValueReader& valReader)
{
	param->accept (valReader);

	return valReader.getValue ();
}

const IntegerValue* readIntParam(giterator::CallStmt::ParameterIterator param, ValueReader& valReader)
{
	param->accept (valReader);

	const IntegerValue* intVal = dynamic_cast<const IntegerValue*> (valReader.getValue ());
	//assert (intVal);

	return intVal;
}

const PtrValue* readPtrParam(giterator::CallStmt::ParameterIterator param, ValueReader& valReader)
{
	param->accept (valReader);

	const PtrValue* ptrVal = dynamic_cast<const PtrValue*> (valReader.getValue ());
	//assert (ptrVal);

	return ptrVal;
}

const IntegerValue* readIntPtrParam(giterator::CallStmt::ParameterIterator param, ValueReader& valReader, MemoryState* memoryState)
{
	param->accept (valReader);

	const PtrValue* ptrVal = dynamic_cast<const PtrValue*> (valReader.getValue ());
	//assert (ptrVal);

	const IntegerValue* intVal = dynamic_cast<const IntegerValue*> (memoryState->load(ptrVal->getPtr()));
	//assert(intVal);

	return intVal;
}

const StringValue* readStrParam(giterator::CallStmt::ParameterIterator param, ValueReader& valReader, MemoryState* memoryState)
{
	param->accept (valReader);

	const PtrValue* strPtrValue = dynamic_cast<const PtrValue*> (valReader.getValue ());
	//assert (strPtrValue);

	const StringValue* strValue = dynamic_cast<const StringValue*> (
			memoryState->load (strPtrValue->getPtr ()));
	//assert (strValue);

	return strValue;
}

/**
 * Sets the values of parameters in the stack frame of called function.
 *
 * @param memoryState inteface to the simulated memory
 * @param cstManager where the constants of the program are stored
 * @param globalsManager where the global declarations of the program are stored
 * @param calleeFunc function that is beeing called
 * @param val interator pointing to the first parameter value
 * @param decl iterator pointing to the first parameter declaration
 * @param calleeFrame pointer to the beginning of the stack frame of the callee
 */
static void
setParameter (MemoryState* memoryState, ConstantsManager* cstManager, GlobalsManager* globalsManager,
              const FunctionValue* calleeFunc, Value* val, const Decl* decl, Ptr calleeFrame)
{
	const FunctionVarsMapper* mapper = calleeFunc->getVarsMapper ();

	//ValueReader reader (callerFunc, memoryState, cstManager, globalsManager, callersFrame);
	//val->accept (reader);

	 Ptr varPtr = calleeFrame + mapper->getVarOffset (decl->getUid ());

	 ValueWriter writer (calleeFunc, memoryState, cstManager, globalsManager, calleeFrame, std::auto_ptr<Value> (val));
	 writer.storeViaPointer(varPtr);
}

/**
 * Finds next conversion specifier in the format string of @c printf() function.
 * 
 * @param str format string
 * @param pos position where to start the search 
 * @return index of the '%' character int the next format string
 */
static size_t
findNextConvSpec (const std::string& str, size_t pos = 0)
{
    size_t splitAt = pos;
    do {
        splitAt = str.find ('%', splitAt);
        if (splitAt == str.npos)
            return str.length ();
        
        rt_assert (splitAt+1 < str.length (), "%% as last character in format of print ()");
    } while (str[splitAt+1] == '%');
    
    return splitAt;
}


/**
 * Finds the last character of conversion specifier.
 * 
 * @param str string beginning with conversion specifier
 * @return index of the first character after the conversion specifier
 */
static size_t
findConvSpecEnd (const std::string& str)
{
    std::string::const_iterator b = str.begin (), e = str.end ();
    
    for (;b != e; b++)
    {
        if (isalpha (*b) && *b != 'l' && *b != 'L' && *b != 'h')
        {
            b++;
            break;
        }
    }
    
    return b - str.begin ();
}


/**
 * Helper function that interprets @c printf() function.
 * 
 * @param func value of the interpreted function
 * @param memoryState interface to the simulated memory
 * @param cstManager where the constants of the program are stored
 * @param globalsManager where the global declarations of the program are stored
 * @param frame
 * @param format format string passed in the @c printf() function
 * @param begin iterator pointing to the second parameter
 * @param end iterator pointing right after the last parameter
 * @param printedCount number of printed values so far
 * @see builtinPrintf
 */
template <typename It> 
static int
doPrintf (const FunctionValue* func, MemoryState* memoryState,
		 ConstantsManager* cstManager, GlobalsManager* globalsManager,
         Ptr frame, const std::string& format, It begin, It end, int printedCount)
{
    if (begin == end)
    {
        rt_assert (format.empty (), "too few parameters for printf ()");
        return printedCount;
    }
    
    rt_assert (!format.empty (), "too many parameters for print ()");
        
    size_t convSpecEnd = findConvSpecEnd (format);
    size_t nextConvSpec = findNextConvSpec (format, convSpecEnd);
    
    std::string beginFormat = format.substr (convSpecEnd, nextConvSpec-convSpecEnd);
    
    std::string subFormat = format.substr (nextConvSpec);
    
    std::stringstream strStream;
    ValueReader valReader (func, memoryState, cstManager, globalsManager, frame);
    begin->accept (valReader);
    
    //
    const Value* valueToPrint = valReader.getValue();

    const PtrValue* strPtrValue = dynamic_cast<const PtrValue*> (valReader.getValue ());
    if (strPtrValue)
    {
		const StringValue* strValue = dynamic_cast<const StringValue*> (
		memoryState->load (strPtrValue->getPtr ()));
		if (strValue)
		{
			valueToPrint = strValue;
		}
    }
    //

    //strStream << *(valReader.getValue ()) << beginFormat;
    strStream << *(valueToPrint) << beginFormat;
    printedCount += strStream.str ().length ();
    std::cout << strStream.str ();
    
    begin++;
    return doPrintf (func, memoryState, cstManager, globalsManager,
    				 frame, subFormat, begin, end, printedCount);
}


/**
 * Interprets built-in @c printf() function.
 * The interpreted function completely ignores the conversion specifier in the format string.
 * The values are printed according to their real type.
 * 
 * @param func current function
 * @param memoryState memory state
 * @param cstManager constants manager
 * @param frame pointer to the beginning of the current stack frame
 * @param params range of function parameters
 * @param returnType type of return value
 * @return result of the @c printf() function
 */
std::auto_ptr<Value>
builtinPrintf (const FunctionValue* func, MemoryState* memoryState, 
               ConstantsManager* cstManager, GlobalsManager* globalsManager,
               Ptr frame,
               std::pair<giterator::CallStmt::ParameterIterator,     
                         giterator::CallStmt::ParameterIterator> params,
               const Type* returnType)
{
    assert (params.first != params.second);
    
    std::stringstream strStream;
    ValueReader valReader (func, memoryState, cstManager, globalsManager, frame);
    params.first->accept (valReader);
    
    const PtrValue* strPtrValue = dynamic_cast<const PtrValue*> (valReader.getValue ());
    assert (strPtrValue);
    
    const StringValue* strValue = dynamic_cast<const StringValue*> (
            memoryState->load (strPtrValue->getPtr ()));
    assert (strValue);

    std::string str = strValue->getValue ();
    LOG_DEBUG ("format = `%s'", str.c_str ());
    
    size_t splitAt = findNextConvSpec (str);
    int printed = printf ("%s", str.substr (0, splitAt).c_str ());
    
    printed += doPrintf (func, memoryState, cstManager, globalsManager, frame,
             str.substr (splitAt), ++(params.first), params.second, printed);
    
    return std::auto_ptr<Value> (new IntegerValue (returnType, static_cast<uint64_t> (printed)));
}


/**
 * Interprets built-in @c puts() function.
 */
std::auto_ptr<Value>
builtinPuts (const FunctionValue* func, MemoryState* memoryState, 
             ConstantsManager* cstManager, GlobalsManager* globalsManager,
             Ptr frame,
             std::pair<giterator::CallStmt::ParameterIterator,     
                       giterator::CallStmt::ParameterIterator> params,
             const Type* returnType)
{
    assert (params.first+1 == params.second);
    
    ValueReader valReader (func, memoryState, cstManager, globalsManager, frame);
    params.first->accept (valReader);
    
    const PtrValue* strPtrValue = dynamic_cast<const PtrValue*> (valReader.getValue ());
    assert (strPtrValue);
    
    const StringValue* strValue = dynamic_cast<const StringValue*> (
            memoryState->load (strPtrValue->getPtr ()));
    assert (strValue);

    std::cout << strValue->getValue () << '\n';
    
    return std::auto_ptr<Value> (new IntegerValue (returnType, static_cast<uint64_t> (1)));
}

/**
 * Interprets built-in @c putchar() function.
 */
std::auto_ptr<Value>
builtinPutchar (const FunctionValue* func, MemoryState* memoryState,
             ConstantsManager* cstManager, GlobalsManager* globalsManager,
             Ptr frame,
             std::pair<giterator::CallStmt::ParameterIterator,
                       giterator::CallStmt::ParameterIterator> params,
             const Type* returnType)
{
    assert (params.first+1 == params.second);

    ValueReader valReader (func, memoryState, cstManager, globalsManager, frame);
    params.first->accept (valReader);

    const IntegerValue* intValue = dynamic_cast<const IntegerValue*> (valReader.getValue ());
    assert (intValue);

    std::cout << (char) intValue->getValue ();

    return std::auto_ptr<Value> (new IntegerValue (returnType, static_cast<uint64_t> (1)));
}


/**
 * Reads the integer value from the user and stores in into the memory.
 * Helper function for the @c builtinScanf() function.
 * @see builtinScanf()
 * 
 * @param op operand representing a pointer to the address, where to store the read value
 * @param func current function
 * @param memoryState memory state
 * @param cstManager constants manager
 * @param frame pointer to the beginning of the current stack frame
 * @param index index of the @c scanf() parameter this call will read
 */
void
scanfParam (const Operand* op, const FunctionValue* func, MemoryState* memoryState, 
            ConstantsManager* cstManager, GlobalsManager* globalsManager
            , Ptr frame, int index)
{
    ValueReader valReader (func, memoryState, cstManager, globalsManager, frame);
    op->accept (valReader);
    
    const PtrValue* paramPtr = dynamic_cast<const PtrValue*> (valReader.getValue ());
    rt_assert (paramPtr, "%d. parameter of scanf () is not a pointer", index);
    
    const PtrType* paramPtrType = dynamic_cast<const PtrType*> (op->getType ());
    assert (paramPtrType);
    
    const Type* paramType = paramPtrType->getPointingToType ();
    if (typeid (*paramType) == typeid (const IntegerType))
    {
        const IntegerType* it = static_cast<const IntegerType*> (paramType);
        std::auto_ptr<Content> ret;
        if (it->isUnsigned ())
        {
            uint64_t in;
            std::cin >> in;
            ret.reset (new IntegerValue (paramType, in));
        }
        else
        {
            int64_t in;
            std::cin >> in;
            ret.reset (new IntegerValue (paramType, in));
        }
        
        memoryState->store (paramPtr->getPtr (), ret);
    }
    else
        ni_error ("reading values of type %s", paramType->getName ().c_str ());
}


/**
 * Interprets built-in @c scanf() function.
 * The format string is completely ignored.
 * The interpreted function reads @c n - 1 integer values, where @c n is the number
 * of parameters passed to the @c scanf() function.
 */
std::auto_ptr<Value>
builtinScanf (const FunctionValue* func, MemoryState* memoryState, 
              ConstantsManager* cstManager, GlobalsManager* globalsManager
              , Ptr frame,
              std::pair<giterator::CallStmt::ParameterIterator,     
                        giterator::CallStmt::ParameterIterator> params,
              const Type* returnType)
{
    CallStmt::ParameterIterator b = params.first, e = params.second;
    assert (b != e);
    b++;
    
    int i = 1;
    for (;b != e; b++, i++ )
    {
        scanfParam (&*b, func, memoryState, cstManager, globalsManager, frame, i);
    }
    
    return std::auto_ptr<Value> (new IntegerValue (returnType, i-1));
}


/**
 * Interprets built-in @c malloc() function.
 */
std::auto_ptr<Value>
builtinMalloc (const FunctionValue* func, MemoryState* memoryState, 
               ConstantsManager* cstManager, GlobalsManager* globalsManager
               , Ptr frame,
               std::pair<giterator::CallStmt::ParameterIterator,     
                         giterator::CallStmt::ParameterIterator> params, 
               const Type* returnType)
{
    assert (params.first+1 == params.second);
    const Operand* op = &*params.first;
    
    ValueReader valReader (func, memoryState, cstManager, globalsManager, frame);
    op->accept (valReader);
    
    const IntegerValue* sizeValue = dynamic_cast<const IntegerValue*> (valReader.getValue ());
    assert (sizeValue);
    size_t size = (size_t)sizeValue->getValue ();
    
    Ptr ptr;
    if (size > 0)
        ptr = memoryState->alloc (size);
    
    std::auto_ptr<Value> ret (new PtrValue (returnType, ptr));
    return ret;
}


/**
 * Interprets built-in @c free function.
 */
std::auto_ptr<Value>
builtinFree (const FunctionValue* func, MemoryState* memoryState, 
             ConstantsManager* cstManager, GlobalsManager* globalsManager
             , Ptr frame,
             std::pair<giterator::CallStmt::ParameterIterator,     
                       giterator::CallStmt::ParameterIterator> params,
             const Type*)
{
    assert (params.first+1 == params.second);
    const Operand* op = &*params.first;
    
    ValueReader valReader (func, memoryState, cstManager, globalsManager, frame);
    op->accept (valReader);
    
    const PtrValue* ptrValue = dynamic_cast<const PtrValue*> (valReader.getValue ());
    assert (ptrValue);
    Ptr ptr = ptrValue->getPtr ();
    
    if (!ptr.isNull ())
        memoryState->free (ptr);
    
    return std::auto_ptr<Value> ();
}

std::auto_ptr<Value>
builtinAbort(const FunctionValue* func, mmodule::MemoryState* memoryState,
        ConstantsManager* cstManager, GlobalsManager* globalsManager,
        mmodule::Ptr frame,
        std::pair<giterator::CallStmt::ParameterIterator,
                   giterator::CallStmt::ParameterIterator> params,
        const giterator::Type* returnType)
{
	assert(params.first == params.second);
	throw AbortException();
	return std::auto_ptr<Value> ();
}

std::auto_ptr<Value>
builtinAbs(const FunctionValue* func, mmodule::MemoryState* memoryState,
        ConstantsManager* cstManager, GlobalsManager* globalsManager,
        mmodule::Ptr frame,
        std::pair<giterator::CallStmt::ParameterIterator,
                   giterator::CallStmt::ParameterIterator> params,
        const giterator::Type* returnType)
{
	assert (params.first+1 == params.second);

	ValueReader valReader (func, memoryState, cstManager, globalsManager, frame);

	const IntegerValue* intVal = readIntParam(params.first, valReader);
	return intVal->absOp();
}

std::auto_ptr<Value>
builtinAtof(const FunctionValue* func, mmodule::MemoryState* memoryState,
        ConstantsManager* cstManager, GlobalsManager* globalsManager,
        mmodule::Ptr frame,
        std::pair<giterator::CallStmt::ParameterIterator,
                   giterator::CallStmt::ParameterIterator> params,
        const giterator::Type* returnType)
{
	assert(params.first+1 == params.second);

	ValueReader valReader (func, memoryState, cstManager, globalsManager, frame);
	const StringValue* strValue = readStrParam(params.first, valReader, memoryState);

	double result = strtod(strValue->getValue().c_str(), NULL);
	return std::auto_ptr<Value> (new RealValue(returnType, result));
}

std::auto_ptr<Value>
builtinAtoi(const FunctionValue* func, mmodule::MemoryState* memoryState,
        ConstantsManager* cstManager, GlobalsManager* globalsManager,
        mmodule::Ptr frame,
        std::pair<giterator::CallStmt::ParameterIterator,
                   giterator::CallStmt::ParameterIterator> params,
        const giterator::Type* returnType)
{
	assert(params.first+1 == params.second);

	ValueReader valReader (func, memoryState, cstManager, globalsManager, frame);
	const StringValue* strValue = readStrParam(params.first, valReader, memoryState);

	int result = strtol(strValue->getValue().c_str(), (char **) NULL, 10);
	return std::auto_ptr<Value> (new IntegerValue(returnType, static_cast<uint64_t> (result)));
}

std::auto_ptr<Value>
builtinAtol(const FunctionValue* func, mmodule::MemoryState* memoryState,
        ConstantsManager* cstManager, GlobalsManager* globalsManager,
        mmodule::Ptr frame,
        std::pair<giterator::CallStmt::ParameterIterator,
                   giterator::CallStmt::ParameterIterator> params,
        const giterator::Type* returnType)
{
	assert(params.first+1 == params.second);

	ValueReader valReader (func, memoryState, cstManager, globalsManager, frame);
	const StringValue* strValue = readStrParam(params.first, valReader, memoryState);

	long result = strtol(strValue->getValue().c_str(), (char **) NULL, 10);
	return std::auto_ptr<Value> (new IntegerValue(returnType, static_cast<uint64_t> (result)));
}

std::auto_ptr<Value>
builtinAtoll(const FunctionValue* func, mmodule::MemoryState* memoryState,
        ConstantsManager* cstManager, GlobalsManager* globalsManager,
        mmodule::Ptr frame,
        std::pair<giterator::CallStmt::ParameterIterator,
                   giterator::CallStmt::ParameterIterator> params,
        const giterator::Type* returnType)
{
	assert(params.first+1 == params.second);

	ValueReader valReader (func, memoryState, cstManager, globalsManager, frame);
	const StringValue* strValue = readStrParam(params.first, valReader, memoryState);

	long long result = strtoll(strValue->getValue().c_str(), (char **) NULL, 10);
	return std::auto_ptr<Value> (new IntegerValue(returnType, static_cast<uint64_t> (result)));
}

std::auto_ptr<Value>
builtinExit(const FunctionValue* func, mmodule::MemoryState* memoryState,
        ConstantsManager* cstManager, GlobalsManager* globalsManager,
        mmodule::Ptr frame,
        std::pair<giterator::CallStmt::ParameterIterator,
                   giterator::CallStmt::ParameterIterator> params,
        const giterator::Type* returnType)
{
	assert(params.first+1 == params.second);

	ValueReader valReader (func, memoryState, cstManager, globalsManager, frame);
	const IntegerValue* exitStatus = readIntParam(params.first, valReader);
	throw ExitException(exitStatus);
	return std::auto_ptr<Value> ();
}

std::auto_ptr<Value>
builtinRand(const FunctionValue* func, mmodule::MemoryState* memoryState,
        ConstantsManager* cstManager, GlobalsManager* globalsManager,
        mmodule::Ptr frame,
        std::pair<giterator::CallStmt::ParameterIterator,
                   giterator::CallStmt::ParameterIterator> params,
        const giterator::Type* returnType)
{
	assert(params.first == params.second);

	int randVal = rand();
	return std::auto_ptr<Value> (new IntegerValue(returnType, static_cast<uint64_t> (randVal)));
}

std::auto_ptr<Value>
builtinSrand(const FunctionValue* func, mmodule::MemoryState* memoryState,
        ConstantsManager* cstManager, GlobalsManager* globalsManager,
        mmodule::Ptr frame,
        std::pair<giterator::CallStmt::ParameterIterator,
                   giterator::CallStmt::ParameterIterator> params,
        const giterator::Type* returnType)
{
	assert(params.first+1 == params.second);

	ValueReader valReader (func, memoryState, cstManager, globalsManager, frame);
	const IntegerValue* randSeed = readIntParam(params.first, valReader);
	assert(randSeed->getIntegerType()->isUnsigned());

	srand(static_cast<unsigned int>(randSeed->getValue()));
	return std::auto_ptr<Value> ();
}

std::auto_ptr<Value>
builtinStrtod(const FunctionValue* func, mmodule::MemoryState* memoryState,
        ConstantsManager* cstManager, GlobalsManager* globalsManager,
        mmodule::Ptr frame,
        std::pair<giterator::CallStmt::ParameterIterator,
                   giterator::CallStmt::ParameterIterator> params,
        const giterator::Type* returnType)
{
	assert(params.first+2 == params.second);

	ValueReader valReader (func, memoryState, cstManager, globalsManager, frame);
	const StringValue* strValue = readStrParam(params.first, valReader, memoryState);
	const PtrValue* ptrValue = readPtrParam(params.first + 1, valReader);

	double result;

	if (ptrValue && !ptrValue->getPtr().isNull())
	{
		WARNING("At the moment, usage of the second argument is not implemented. It will be ignored.");
	}

	result = strtod(strValue->getValue().c_str(), NULL);

	return std::auto_ptr<Value> (new RealValue(returnType, result));
}

std::auto_ptr<Value>
builtinStrtol(const FunctionValue* func, mmodule::MemoryState* memoryState,
        ConstantsManager* cstManager, GlobalsManager* globalsManager,
        mmodule::Ptr frame,
        std::pair<giterator::CallStmt::ParameterIterator,
                   giterator::CallStmt::ParameterIterator> params,
        const giterator::Type* returnType)
{
	assert(params.first+3 == params.second);

	ValueReader valReader (func, memoryState, cstManager, globalsManager, frame);
	const StringValue* strValue = readStrParam(params.first, valReader, memoryState);
	const PtrValue* ptrValue = readPtrParam(params.first + 1, valReader);
	const IntegerValue* intValue = readIntParam(params.first + 2, valReader);

	long int result;

	if (ptrValue && !ptrValue->getPtr().isNull())
	{
		WARNING("At the moment, usage of the second argument is not implemented. It will be ignored.");
	}

	result = strtol(strValue->getValue().c_str(), NULL, static_cast<int> (intValue->getValue()));

	return std::auto_ptr<Value> (new IntegerValue(returnType, result));
}

std::auto_ptr<Value>
builtinMemcpy(const FunctionValue* func, mmodule::MemoryState* memoryState,
        ConstantsManager* cstManager, GlobalsManager* globalsManager,
        mmodule::Ptr frame,
        std::pair<giterator::CallStmt::ParameterIterator,
                   giterator::CallStmt::ParameterIterator> params,
        const giterator::Type* returnType)
{
	assert(params.first + 3 == params.second);

	ValueReader valReader (func, memoryState, cstManager, globalsManager, frame);
	const PtrValue* dest = readPtrParam(params.first, valReader);
	Ptr destPtr = dest->getPtr();
	const PtrValue* src = readPtrParam(params.first + 1, valReader);
	const IntegerValue* n = readIntParam(params.first + 2, valReader);

	std::pair<MemoryState::ContentIterator, MemoryState::ContentIterator> values =
			memoryState->loadOverlapping(src->getPtr(), (size_t) n->getValue());

	for (MemoryState::ContentIterator i1 = values.first; i1 != values.second; ++i1)
	{
		const Value* storeVal = dynamic_cast<const Value*> (i1->second);
		memoryState->store(destPtr, std::auto_ptr<Content>(storeVal->clone()));
		destPtr += storeVal->getUnitSize();
	}

	// TODO: overlapping areas (throw an exception).

	//return std::auto_ptr<Value> (new PtrValue(returnType, destPtr));
	return std::auto_ptr<Value> ();
}

std::auto_ptr<Value>
builtinMemmove(const FunctionValue* func, mmodule::MemoryState* memoryState,
        ConstantsManager* cstManager, GlobalsManager* globalsManager,
        mmodule::Ptr frame,
        std::pair<giterator::CallStmt::ParameterIterator,
                   giterator::CallStmt::ParameterIterator> params,
        const giterator::Type* returnType)
{
	ValueReader valReader (func, memoryState, cstManager, globalsManager, frame);
	const PtrValue* dest = readPtrParam(params.first, valReader);
	Ptr destPtr = dest->getPtr();
	const PtrValue* src = readPtrParam(params.first + 1, valReader);
	const IntegerValue* n = readIntParam(params.first + 2, valReader);

	std::pair<MemoryState::ContentIterator, MemoryState::ContentIterator> values =
			memoryState->loadOverlapping(src->getPtr(), (size_t) n->getValue());

	std::map<size_t, Value*> newValues;

	size_t offset = 0;
	for (MemoryState::ContentIterator i1 = values.first; i1 != values.second; ++i1)
	{
		const Value* storeVal = dynamic_cast<const Value*> (i1->second);
		newValues[offset] = storeVal->clone();
		offset += storeVal->getUnitSize();
	}

	for (std::map<size_t, Value*>::const_iterator i1 = newValues.begin(); i1 != newValues.end(); ++i1)
	{
		memoryState->store(destPtr + i1->first, std::auto_ptr<Content>(i1->second));
	}

	//return std::auto_ptr<Value> (new PtrValue(returnType, destPtr));
	return std::auto_ptr<Value> ();
}

std::auto_ptr<Value>
builtinStrcat(const FunctionValue* func, mmodule::MemoryState* memoryState,
        ConstantsManager* cstManager, GlobalsManager* globalsManager,
        mmodule::Ptr frame,
        std::pair<giterator::CallStmt::ParameterIterator,
                   giterator::CallStmt::ParameterIterator> params,
        const giterator::Type* returnType)
{
	// this may be either strcat or strncat
	assert(params.first + 3 >= params.second);

	//it seems that strcat() is realized by memcpy() inside gcc,
	// so this body is pointless

	return std::auto_ptr<Value> ();
}

std::auto_ptr<Value>
builtinStrcpy(const FunctionValue* func, mmodule::MemoryState* memoryState,
        ConstantsManager* cstManager, GlobalsManager* globalsManager,
        mmodule::Ptr frame,
        std::pair<giterator::CallStmt::ParameterIterator,
                   giterator::CallStmt::ParameterIterator> params,
        const giterator::Type* returnType)
{
	return builtinMemcpy(func, memoryState, cstManager, globalsManager, frame, params, returnType);
}

std::auto_ptr<Value>
builtinStrcmp(const FunctionValue* func, mmodule::MemoryState* memoryState,
        ConstantsManager* cstManager, GlobalsManager* globalsManager,
        mmodule::Ptr frame,
        std::pair<giterator::CallStmt::ParameterIterator,
                   giterator::CallStmt::ParameterIterator> params,
        const giterator::Type* returnType)
{
	assert(params.first + 2 == params.second);

	ValueReader valReader (func, memoryState, cstManager, globalsManager, frame);
	const StringValue* s1 = readStrParam(params.first, valReader, memoryState);
	const StringValue* s2 = readStrParam(params.first + 1, valReader, memoryState);

	const char* s1Str = s1->getValue().c_str();
	const char* s2Str = s2->getValue().c_str();

	int ret = strcmp(s1Str, s2Str);

	return std::auto_ptr<Value> (new IntegerValue(returnType, ret));
}

std::auto_ptr<Value>
builtinStrlen(const FunctionValue* func, mmodule::MemoryState* memoryState,
        ConstantsManager* cstManager, GlobalsManager* globalsManager,
        mmodule::Ptr frame,
        std::pair<giterator::CallStmt::ParameterIterator,
                   giterator::CallStmt::ParameterIterator> params,
        const giterator::Type* returnType)
{
	assert(params.first + 1 == params.second);

	ValueReader valReader (func, memoryState, cstManager, globalsManager, frame);
	const StringValue* s = readStrParam(params.first, valReader, memoryState);

	const char* ss = s->getValue().c_str();
	size_t length = strlen(ss);

	return std::auto_ptr<Value> (new IntegerValue(returnType, length));
}

std::auto_ptr<Value>
builtinAssert(const FunctionValue* func, mmodule::MemoryState* memoryState,
        ConstantsManager* cstManager, GlobalsManager* globalsManager,
        mmodule::Ptr frame,
        std::pair<giterator::CallStmt::ParameterIterator,
                   giterator::CallStmt::ParameterIterator> params,
        const giterator::Type* returnType)
{
	assert(params.first+4 == params.second);

	ValueReader valReader (func, memoryState, cstManager, globalsManager, frame);
	const StringValue* assertion = readStrParam(params.first, valReader, memoryState);
	const StringValue* file = readStrParam(params.first + 1, valReader, memoryState);
	const IntegerValue* line = readIntParam(params.first + 2, valReader);
	const StringValue* function = readStrParam(params.first + 3, valReader, memoryState);
	assert(line->getIntegerType()->isUnsigned());

	throw AssertException(assertion->getValue(),
							  file->getValue(),
							  static_cast<unsigned int> (line->getValue()),
							  function->getValue());

	return std::auto_ptr<Value> ();
}

std::auto_ptr<Value>
niFunction(const FunctionValue* func, mmodule::MemoryState* memoryState,
        ConstantsManager* cstManager, GlobalsManager* globalsManager,
        mmodule::Ptr frame,
        std::pair<giterator::CallStmt::ParameterIterator,
                   giterator::CallStmt::ParameterIterator> params,
        const giterator::Type* returnType)
{
	ni_error("Not implemented function.");
	return std::auto_ptr<Value> ();
}

/**
 * Creates a new stack frame for function @c func.
 */
Ptr
createFrame (MemoryState* memoryState, const FunctionValue* func)
{
  /*
   * Create and init stack frame of called function
  */
  size_t frameSize = func->getVarsMapper ()->getSize ();

  if (frameSize != 0)
	  return memoryState->alloc (frameSize);
  else
	  return Ptr();
}

std::auto_ptr<Value>
builtinSleep(const FunctionValue* func, mmodule::MemoryState* memoryState,
        ConstantsManager* cstManager, GlobalsManager* globalsManager,
        mmodule::Ptr frame,
        std::pair<giterator::CallStmt::ParameterIterator,
                   giterator::CallStmt::ParameterIterator> params,
        const giterator::Type* returnType,
        Program* program)
{
	assert(params.first + 1 == params.second);

	ValueReader valReader(func, memoryState, cstManager, globalsManager, frame);
	const IntegerValue* iv = readIntParam(params.first, valReader);
	assert(iv);

	program->getCurrentThread()->sleep((unsigned int) iv->getValue(), program->getRunningMode());

	return std::auto_ptr<Value> (new IntegerValue(returnType, 0));
}


std::auto_ptr<Value>
builtinThreadCreate(const FunctionValue* func, MemoryState* memoryState,
					ConstantsManager* cstManager, GlobalsManager* globalsManager,
					Ptr frame,
					std::pair<giterator::CallStmt::ParameterIterator,
							  giterator::CallStmt::ParameterIterator> params,
					const Type* returnType,
					Program* program)
{
	assert(params.first + 4 == params.second);

	ValueReader valReader (func, memoryState, cstManager, globalsManager, frame);
	const PtrValue* pv = readPtrParam(params.first, valReader);

	const PtrType* pvPtrType = dynamic_cast<const PtrType*> (pv->getType ());
	assert (pvPtrType);

	const IntegerType* it = dynamic_cast<const IntegerType*> (pvPtrType->getPointingToType ());
	assert (it);

	ValueReader valReader2 (func, memoryState, cstManager, globalsManager, frame);
	const PtrValue* ptrValue = readPtrParam(params.first + 2, valReader2);
	assert (ptrValue);

	const FunctionValue* funcValue = dynamic_cast<const FunctionValue*> (memoryState->load(ptrValue->getPtr()));//valReader.getValue ());
	assert (funcValue);

	// Handle data argument
	ValueReader valReader3 (func, memoryState, cstManager, globalsManager, frame);
	const Value* dataValue = readRawParam(params.first + 3, valReader3);
	assert(dataValue);

	Ptr threadFuncFrame = createFrame (memoryState, funcValue);

	const Function::Declarations& paramDecls = funcValue->getFunction ()->getParamDeclarations ();

	size_t nParams = paramDecls.size();

	if (nParams == 0)
		; //TODO: some kind of warning?
	else if (nParams == 1)
		setParameter (memoryState, cstManager, globalsManager, funcValue,
					  dataValue->clone(), *paramDecls.begin (), threadFuncFrame);
	else
		assert(false);

	Thread::ID ID = program->createNewThread(funcValue, threadFuncFrame);

	// resolve name of the thread variable and store the information inside program
	NameResolver nameRes;
	params.first->accept(nameRes);
	program->addThreadName(ID, nameRes.getName());

	// return the new thread ID
	std::auto_ptr<Content> res(new IntegerValue(it, ID));
	memoryState->store (pv->getPtr (), res);

	return std::auto_ptr<Value> (new IntegerValue(returnType, 0));
}

std::auto_ptr<Value>
builtinThreadJoin(const FunctionValue* func, MemoryState* memoryState,
					ConstantsManager* cstManager, GlobalsManager* globalsManager
					, Ptr frame,
					std::pair<giterator::CallStmt::ParameterIterator,
							  giterator::CallStmt::ParameterIterator> params,
					const Type* returnType,
					Program* program)
{
	assert(params.first + 2 == params.second);
	ValueReader valReader( func, memoryState, cstManager, globalsManager, frame);
	const IntegerValue* iv = readIntParam(params.first, valReader);
	assert(iv);

	ThreadPool* tp = program->getThreadPool();
	int ret = tp->joinThreads(tp->getCurrentThreadID(), (Thread::ID) iv->getValue());

	return std::auto_ptr<Value> (new IntegerValue(returnType, ret));
}

std::auto_ptr<Value>
builtinThreadExit(const FunctionValue* func, MemoryState* memoryState,
					ConstantsManager* cstManager, GlobalsManager* globalsManager
					, Ptr frame,
					std::pair<giterator::CallStmt::ParameterIterator,
							  giterator::CallStmt::ParameterIterator> params,
					const Type* returnType,
					Program* program)
{
	assert(params.first + 1 == params.second);

	ThreadPool* tp = program->getThreadPool();
	Thread::ID currId = tp->getCurrentThreadID();

	if (tp->getThreadCount() == 0) // there are no more threads
		throw LastThreadExitException(currId);
	else
		throw ThreadExitException(currId);

	return std::auto_ptr<Value> ();
}

std::auto_ptr<Value>
builtinThreadMutexInit(const FunctionValue* func, mmodule::MemoryState* memoryState,
							ConstantsManager* cstManager, GlobalsManager* globalsManager,
							mmodule::Ptr frame,
							std::pair<giterator::CallStmt::ParameterIterator,
									   giterator::CallStmt::ParameterIterator> params,
							const giterator::Type* returnType,
							Program* program)
{
	assert(params.first + 2 == params.second);

	ValueReader valReader (func, memoryState, cstManager, globalsManager, frame);
	const PtrValue* pv = readPtrParam(params.first, valReader);
	assert(pv);

	Mutex::ID mutexID;

	try
	{
		// if this throws UndefinedMemoryLoadException,  mutex is not initialized yet.
		const IntegerValue* intVal = dynamic_cast<const IntegerValue*> (memoryState->load(pv->getPtr()));
		assert(intVal);

		mutexID = (Mutex::ID) intVal->getValue();
	}
	catch (UndefinedMemoryLoadException&)
	{
		mutexID = -1;
	}

	int ret = program->createNewMutex(mutexID);
	//TODO: type
	std::auto_ptr<Content> newID(new IntegerValue(returnType, mutexID));
	memoryState->store(pv->getPtr(), newID);

	return std::auto_ptr<Value> (new IntegerValue(returnType, ret));
}


std::auto_ptr<Value>
builtinThreadMutexDestroy(const FunctionValue* func, mmodule::MemoryState* memoryState,
							ConstantsManager* cstManager, GlobalsManager* globalsManager,
							mmodule::Ptr frame,
							std::pair<giterator::CallStmt::ParameterIterator,
									   giterator::CallStmt::ParameterIterator> params,
							const giterator::Type* returnType,
							Program* program)
{
	assert(params.first + 1 == params.second);

	ValueReader valReader( func, memoryState, cstManager, globalsManager, frame);
	const IntegerValue* iv = readIntPtrParam(params.first, valReader, memoryState);
	assert(iv);
	Mutex::ID mutexID = (Mutex::ID) iv->getValue();

	int ret = program->destroyMutex(mutexID);

	return std::auto_ptr<Value> (new IntegerValue(returnType, ret));
}


std::auto_ptr<Value>
builtinThreadMutexLock(const FunctionValue* func, mmodule::MemoryState* memoryState,
						ConstantsManager* cstManager, GlobalsManager* globalsManager,
						mmodule::Ptr frame,
						std::pair<giterator::CallStmt::ParameterIterator,
								   giterator::CallStmt::ParameterIterator> params,
						const giterator::Type* returnType,
						Program* program)
{
	assert(params.first + 1 == params.second);

	ValueReader valReader( func, memoryState, cstManager, globalsManager, frame);
	const IntegerValue* iv = readIntPtrParam(params.first, valReader, memoryState);
	assert(iv);
	Mutex::ID mutexID = (Mutex::ID) iv->getValue();

	int ret = program->lockMutex(mutexID);

	return std::auto_ptr<Value> (new IntegerValue(returnType, ret));
}


std::auto_ptr<Value>
builtinThreadMutexTryLock(const FunctionValue* func, mmodule::MemoryState* memoryState,
							ConstantsManager* cstManager, GlobalsManager* globalsManager,
							mmodule::Ptr frame,
							std::pair<giterator::CallStmt::ParameterIterator,
									   giterator::CallStmt::ParameterIterator> params,
							const giterator::Type* returnType,
							Program* program)
{
	assert(params.first + 1 == params.second);

	ValueReader valReader( func, memoryState, cstManager, globalsManager, frame);
	const IntegerValue* iv = readIntPtrParam(params.first, valReader, memoryState);
	assert(iv);
	Mutex::ID mutexID = (Mutex::ID) iv->getValue();

	int ret = program->tryLockMutex(mutexID);

	return std::auto_ptr<Value> (new IntegerValue(returnType, ret));
}

std::auto_ptr<Value>
builtinThreadMutexUnlock(const FunctionValue* func, mmodule::MemoryState* memoryState,
							ConstantsManager* cstManager, GlobalsManager* globalsManager,
							mmodule::Ptr frame,
							std::pair<giterator::CallStmt::ParameterIterator,
									   giterator::CallStmt::ParameterIterator> params,
							const giterator::Type* returnType,
							Program* program)
{
	assert(params.first + 1 == params.second);

	ValueReader valReader( func, memoryState, cstManager, globalsManager, frame);
	const IntegerValue* iv = readIntPtrParam(params.first, valReader, memoryState);
	assert(iv);
	Mutex::ID mutexID = (Mutex::ID) iv->getValue();

	int ret = program->unlockMutex(mutexID);

	return std::auto_ptr<Value> (new IntegerValue(returnType, ret));
}

std::auto_ptr<Value>
builtinThreadCondInit(const FunctionValue* func, mmodule::MemoryState* memoryState,
						ConstantsManager* cstManager, GlobalsManager* globalsManager,
						mmodule::Ptr frame,
						std::pair<giterator::CallStmt::ParameterIterator,
								   giterator::CallStmt::ParameterIterator> params,
						const giterator::Type* returnType,
						Program* program)
{
	assert(params.first + 2 == params.second);

	ValueReader valReader (func, memoryState, cstManager, globalsManager, frame);
	const PtrValue* pv = readPtrParam(params.first, valReader);
	assert(pv);

	Condition::ID condID;

	try
	{
		// if this throws UndefinedMemoryLoadException,  condition is not initialized yet.
		const IntegerValue* intVal = dynamic_cast<const IntegerValue*> (memoryState->load(pv->getPtr()));
		assert(intVal);

		condID = (Condition::ID) intVal->getValue();
	}
	catch (UndefinedMemoryLoadException&)
	{
		condID = -1;
	}

	int ret = program->createNewCondition(condID);
	//TODO: type
	std::auto_ptr<Content> newID(new IntegerValue(returnType, condID));
	memoryState->store(pv->getPtr(), newID);

	return std::auto_ptr<Value> (new IntegerValue(returnType, ret));
}

std::auto_ptr<Value>
builtinThreadCondDestroy(const FunctionValue* func, mmodule::MemoryState* memoryState,
							ConstantsManager* cstManager, GlobalsManager* globalsManager,
							mmodule::Ptr frame,
							std::pair<giterator::CallStmt::ParameterIterator,
									   giterator::CallStmt::ParameterIterator> params,
							const giterator::Type* returnType,
							Program* program)
{
	assert(params.first + 1 == params.second);

	ValueReader valReader( func, memoryState, cstManager, globalsManager, frame);
	const IntegerValue* iv = readIntPtrParam(params.first, valReader, memoryState);
	assert(iv);
	Condition::ID condID = (Condition::ID) iv->getValue();

	int ret = program->destroyCondition(condID);

	return std::auto_ptr<Value> (new IntegerValue(returnType, ret));
}

std::auto_ptr<Value>
builtinThreadCondSignal(const FunctionValue* func, mmodule::MemoryState* memoryState,
						ConstantsManager* cstManager, GlobalsManager* globalsManager,
						mmodule::Ptr frame,
						std::pair<giterator::CallStmt::ParameterIterator,
								   giterator::CallStmt::ParameterIterator> params,
						const giterator::Type* returnType,
						Program* program)
{
	assert(params.first + 1 == params.second);

	ValueReader valReader( func, memoryState, cstManager, globalsManager, frame);
	const IntegerValue* iv = readIntPtrParam(params.first, valReader, memoryState);
	assert(iv);
	Condition::ID condID = (Condition::ID) iv->getValue();

	int ret = program->signalCondition(condID);

	return std::auto_ptr<Value> (new IntegerValue(returnType, ret));
}

std::auto_ptr<Value>
builtinThreadCondWait(const FunctionValue* func, mmodule::MemoryState* memoryState,
							ConstantsManager* cstManager, GlobalsManager* globalsManager,
							mmodule::Ptr frame,
							std::pair<giterator::CallStmt::ParameterIterator,
									   giterator::CallStmt::ParameterIterator> params,
							const giterator::Type* returnType,
							Program* program)
{
	assert(params.first + 2 == params.second);

	ValueReader valReader( func, memoryState, cstManager, globalsManager, frame);
	const IntegerValue* iv = readIntPtrParam(params.first, valReader, memoryState);
	assert(iv);
	Condition::ID condID = (Condition::ID) iv->getValue();

	const IntegerValue* iv2 = readIntPtrParam(params.first + 1, valReader, memoryState);
	assert(iv2);
	Mutex::ID mutexID = (Mutex::ID) iv2->getValue();

	int ret = program->waitCondition(condID, mutexID);

	return std::auto_ptr<Value> (new IntegerValue(returnType, ret));
}


}
