// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file BuiltinFunctions.h
 * This file contains declarations of functions implementing simulated 
 * built-in functions.
 */
#ifndef INT_BUILTIN_FUNCTIONS_H
#define INT_BUILTIN_FUNCTIONS_H

#include "../mm/fwd.h"
#include "../mm/Ptr.h"
#include "../git/CallStmt.h"

namespace ginterpreter {
    
class Value;
class FunctionValue;
class ConstantsManager;
class GlobalsManager;
class Thread;
class Program;


/** @name Built-in and standard functions
 * Functions realizing interpretation of a built-in and standard functions
 * supported by GMC.
 * These methods gets assigned to the call statement for the corresponding functions
 * inside @c GimpleListener class.
 *
 * @param func value of the interpreted function
 * @param memoryState interface to the simulated memory
 * @param cstManager where the constants of the program are stored
 * @param globalsManager where the global declarations of the program are stored
 * @param frame current stack frame
 * @param parameters parameters of the function
 * @param returnType the type of the value to return from this function
 * @return value representing return value of the interpreted built-in or standard function
 */
/*@{*/
std::auto_ptr<Value>
builtinPrintf (const FunctionValue* func, mmodule::MemoryState* memoryState, 
               ConstantsManager* cstManager, GlobalsManager* globalsManager,
               mmodule::Ptr frame, 
               std::pair<giterator::CallStmt::ParameterIterator,     
                        giterator::CallStmt::ParameterIterator> parameters,
               const giterator::Type* returnType);
               
std::auto_ptr<Value>
builtinScanf (const FunctionValue* func, mmodule::MemoryState* memoryState, 
              ConstantsManager* cstManager, GlobalsManager* globalsManager,
              mmodule::Ptr frame,
              std::pair<giterator::CallStmt::ParameterIterator,     
                        giterator::CallStmt::ParameterIterator> parameters,
              const giterator::Type* returnType);
              
std::auto_ptr<Value>
builtinMalloc (const FunctionValue* func, mmodule::MemoryState* memoryState, 
               ConstantsManager* cstManager, GlobalsManager* globalsManager,
               mmodule::Ptr frame, 
               std::pair<giterator::CallStmt::ParameterIterator,     
                        giterator::CallStmt::ParameterIterator> parameters,
               const giterator::Type* returnType);

std::auto_ptr<Value>
builtinFree (const FunctionValue* func, mmodule::MemoryState* memoryState, 
             ConstantsManager* cstManager, GlobalsManager* globalsManager,
             mmodule::Ptr frame, 
             std::pair<giterator::CallStmt::ParameterIterator,     
                        giterator::CallStmt::ParameterIterator> parameters,
             const giterator::Type* returnType);
              
std::auto_ptr<Value>
builtinPuts (const FunctionValue* func, mmodule::MemoryState* memoryState, 
             ConstantsManager* cstManager, GlobalsManager* globalsManager,
             mmodule::Ptr frame, 
             std::pair<giterator::CallStmt::ParameterIterator,     
                        giterator::CallStmt::ParameterIterator> parameters,
             const giterator::Type* returnType);
std::auto_ptr<Value>
builtinPutchar (const FunctionValue* func, mmodule::MemoryState* memoryState,
             ConstantsManager* cstManager, GlobalsManager* globalsManager,
             mmodule::Ptr frame,
             std::pair<giterator::CallStmt::ParameterIterator,
                        giterator::CallStmt::ParameterIterator> parameters,
             const giterator::Type* returnType);

// TODO: void
std::auto_ptr<Value>
builtinAbort(const FunctionValue* func, mmodule::MemoryState* memoryState,
        ConstantsManager* cstManager, GlobalsManager* globalsManager,
        mmodule::Ptr frame,
        std::pair<giterator::CallStmt::ParameterIterator,
                   giterator::CallStmt::ParameterIterator> parameters,
        const giterator::Type* returnType);

std::auto_ptr<Value>
builtinAbs(const FunctionValue* func, mmodule::MemoryState* memoryState,
        ConstantsManager* cstManager, GlobalsManager* globalsManager,
        mmodule::Ptr frame,
        std::pair<giterator::CallStmt::ParameterIterator,
                   giterator::CallStmt::ParameterIterator> parameters,
        const giterator::Type* returnType);

std::auto_ptr<Value>
builtinAtof(const FunctionValue* func, mmodule::MemoryState* memoryState,
        ConstantsManager* cstManager, GlobalsManager* globalsManager,
        mmodule::Ptr frame,
        std::pair<giterator::CallStmt::ParameterIterator,
                   giterator::CallStmt::ParameterIterator> parameters,
        const giterator::Type* returnType);

std::auto_ptr<Value>
builtinAtoi(const FunctionValue* func, mmodule::MemoryState* memoryState,
        ConstantsManager* cstManager, GlobalsManager* globalsManager,
        mmodule::Ptr frame,
        std::pair<giterator::CallStmt::ParameterIterator,
                   giterator::CallStmt::ParameterIterator> parameters,
        const giterator::Type* returnType);

std::auto_ptr<Value>
builtinAtol(const FunctionValue* func, mmodule::MemoryState* memoryState,
        ConstantsManager* cstManager, GlobalsManager* globalsManager,
        mmodule::Ptr frame,
        std::pair<giterator::CallStmt::ParameterIterator,
                   giterator::CallStmt::ParameterIterator> parameters,
        const giterator::Type* returnType);

std::auto_ptr<Value>
builtinAtoll(const FunctionValue* func, mmodule::MemoryState* memoryState,
        ConstantsManager* cstManager, GlobalsManager* globalsManager,
        mmodule::Ptr frame,
        std::pair<giterator::CallStmt::ParameterIterator,
                   giterator::CallStmt::ParameterIterator> parameters,
        const giterator::Type* returnType);

//TODO: void
std::auto_ptr<Value>
builtinExit(const FunctionValue* func, mmodule::MemoryState* memoryState,
        ConstantsManager* cstManager, GlobalsManager* globalsManager,
        mmodule::Ptr frame,
        std::pair<giterator::CallStmt::ParameterIterator,
                   giterator::CallStmt::ParameterIterator> parameters,
        const giterator::Type* returnType);

std::auto_ptr<Value>
builtinRand(const FunctionValue* func, mmodule::MemoryState* memoryState,
        ConstantsManager* cstManager, GlobalsManager* globalsManager,
        mmodule::Ptr frame,
        std::pair<giterator::CallStmt::ParameterIterator,
                   giterator::CallStmt::ParameterIterator> parameters,
        const giterator::Type* returnType);

// TODO: void
std::auto_ptr<Value>
builtinSrand(const FunctionValue* func, mmodule::MemoryState* memoryState,
        ConstantsManager* cstManager, GlobalsManager* globalsManager,
        mmodule::Ptr frame,
        std::pair<giterator::CallStmt::ParameterIterator,
                   giterator::CallStmt::ParameterIterator> parameters,
        const giterator::Type* returnType);

std::auto_ptr<Value>
builtinStrtod(const FunctionValue* func, mmodule::MemoryState* memoryState,
        ConstantsManager* cstManager, GlobalsManager* globalsManager,
        mmodule::Ptr frame,
        std::pair<giterator::CallStmt::ParameterIterator,
                   giterator::CallStmt::ParameterIterator> parameters,
        const giterator::Type* returnType);

std::auto_ptr<Value>
builtinStrtol(const FunctionValue* func, mmodule::MemoryState* memoryState,
        ConstantsManager* cstManager, GlobalsManager* globalsManager,
        mmodule::Ptr frame,
        std::pair<giterator::CallStmt::ParameterIterator,
                   giterator::CallStmt::ParameterIterator> parameters,
        const giterator::Type* returnType);

std::auto_ptr<Value>
builtinMemcpy(const FunctionValue* func, mmodule::MemoryState* memoryState,
        ConstantsManager* cstManager, GlobalsManager* globalsManager,
        mmodule::Ptr frame,
        std::pair<giterator::CallStmt::ParameterIterator,
                   giterator::CallStmt::ParameterIterator> parameters,
        const giterator::Type* returnType);

std::auto_ptr<Value>
builtinMemmove(const FunctionValue* func, mmodule::MemoryState* memoryState,
        ConstantsManager* cstManager, GlobalsManager* globalsManager,
        mmodule::Ptr frame,
        std::pair<giterator::CallStmt::ParameterIterator,
                   giterator::CallStmt::ParameterIterator> parameters,
        const giterator::Type* returnType);

std::auto_ptr<Value>
builtinStrcat(const FunctionValue* func, mmodule::MemoryState* memoryState,
        ConstantsManager* cstManager, GlobalsManager* globalsManager,
        mmodule::Ptr frame,
        std::pair<giterator::CallStmt::ParameterIterator,
                   giterator::CallStmt::ParameterIterator> parameters,
        const giterator::Type* returnType);

std::auto_ptr<Value>
builtinStrcpy(const FunctionValue* func, mmodule::MemoryState* memoryState,
        ConstantsManager* cstManager, GlobalsManager* globalsManager,
        mmodule::Ptr frame,
        std::pair<giterator::CallStmt::ParameterIterator,
                   giterator::CallStmt::ParameterIterator> parameters,
        const giterator::Type* returnType);

std::auto_ptr<Value>
builtinStrcmp(const FunctionValue* func, mmodule::MemoryState* memoryState,
        ConstantsManager* cstManager, GlobalsManager* globalsManager,
        mmodule::Ptr frame,
        std::pair<giterator::CallStmt::ParameterIterator,
                   giterator::CallStmt::ParameterIterator> parameters,
        const giterator::Type* returnType);

std::auto_ptr<Value>
builtinStrlen(const FunctionValue* func, mmodule::MemoryState* memoryState,
        ConstantsManager* cstManager, GlobalsManager* globalsManager,
        mmodule::Ptr frame,
        std::pair<giterator::CallStmt::ParameterIterator,
                   giterator::CallStmt::ParameterIterator> parameters,
        const giterator::Type* returnType);

std::auto_ptr<Value>
builtinAssert(const FunctionValue* func, mmodule::MemoryState* memoryState,
        ConstantsManager* cstManager, GlobalsManager* globalsManager,
        mmodule::Ptr frame,
        std::pair<giterator::CallStmt::ParameterIterator,
                   giterator::CallStmt::ParameterIterator> parameters,
        const giterator::Type* returnType);
/*@}*/

/**
 * For every not implemented function, this method is called (and error is reported).
 */
std::auto_ptr<Value>
niFunction(const FunctionValue* func, mmodule::MemoryState* memoryState,
        ConstantsManager* cstManager, GlobalsManager* globalsManager,
        mmodule::Ptr frame,
        std::pair<giterator::CallStmt::ParameterIterator,
                   giterator::CallStmt::ParameterIterator> parameters,
        const giterator::Type* returnType);

/**
 * @name Thread function
 * Functions realizing interpretation of thread functions supported by GMC.
 * These methods gets assigned to the call statement for the corresponding functions
 * inside @c GimpleListener class.
 * Function sleep() is considered to be thread function inside GMC since it
 * puts a thread to sleep.
 * Apart from the built-in and standard functions, these function take as an input pointer
 * to the simulated program to be able to manipulate threads.
 *
 * @param func value of the interpreted function
 * @param memoryState interface to the simulated memory
 * @param cstManager where the constants of the program are stored
 * @param globalsManager where the global declarations of the program are stored
 * @param frame current stack frame
 * @param parameters parameters of the function
 * @param returnType the type of the value to return from this function
 * @param program the interpreted program
 * @return value representing return value of the interpreted built-in or standard function
 */
/*@{*/
std::auto_ptr<Value>
builtinSleep(const FunctionValue* func, mmodule::MemoryState* memoryState,
        ConstantsManager* cstManager, GlobalsManager* globalsManager,
        mmodule::Ptr frame,
        std::pair<giterator::CallStmt::ParameterIterator,
                   giterator::CallStmt::ParameterIterator> parameters,
        const giterator::Type* returnType,
        Program* program);

std::auto_ptr<Value>
builtinThreadCreate(const FunctionValue* func, mmodule::MemoryState* memoryState,
					ConstantsManager* cstManager, GlobalsManager* globalsManager,
					mmodule::Ptr frame,
					std::pair<giterator::CallStmt::ParameterIterator,
							   giterator::CallStmt::ParameterIterator> parameters,
					const giterator::Type* returnType,
					Program* program);

std::auto_ptr<Value>
builtinThreadJoin(const FunctionValue* func, mmodule::MemoryState* memoryState,
					ConstantsManager* cstManager, GlobalsManager* globalsManager,
					mmodule::Ptr frame,
					std::pair<giterator::CallStmt::ParameterIterator,
							   giterator::CallStmt::ParameterIterator> parameters,
					const giterator::Type* returnType,
					Program* program);

std::auto_ptr<Value>
builtinThreadExit(const FunctionValue* func, mmodule::MemoryState* memoryState,
					ConstantsManager* cstManager, GlobalsManager* globalsManager,
					mmodule::Ptr frame,
					std::pair<giterator::CallStmt::ParameterIterator,
							   giterator::CallStmt::ParameterIterator> parameters,
					const giterator::Type* returnType,
					Program* program);

std::auto_ptr<Value>
builtinThreadMutexInit(const FunctionValue* func, mmodule::MemoryState* memoryState,
						ConstantsManager* cstManager, GlobalsManager* globalsManager,
						mmodule::Ptr frame,
						std::pair<giterator::CallStmt::ParameterIterator,
								   giterator::CallStmt::ParameterIterator> parameters,
						const giterator::Type* returnType,
						Program* program);

std::auto_ptr<Value>
builtinThreadMutexDestroy(const FunctionValue* func, mmodule::MemoryState* memoryState,
							ConstantsManager* cstManager, GlobalsManager* globalsManager,
							mmodule::Ptr frame,
							std::pair<giterator::CallStmt::ParameterIterator,
									   giterator::CallStmt::ParameterIterator> parameters,
							const giterator::Type* returnType,
							Program* program);

std::auto_ptr<Value>
builtinThreadMutexLock(const FunctionValue* func, mmodule::MemoryState* memoryState,
						ConstantsManager* cstManager, GlobalsManager* globalsManager,
						mmodule::Ptr frame,
						std::pair<giterator::CallStmt::ParameterIterator,
								   giterator::CallStmt::ParameterIterator> parameters,
						const giterator::Type* returnType,
						Program* program);

std::auto_ptr<Value>
builtinThreadMutexTryLock(const FunctionValue* func, mmodule::MemoryState* memoryState,
							ConstantsManager* cstManager, GlobalsManager* globalsManager,
							mmodule::Ptr frame,
							std::pair<giterator::CallStmt::ParameterIterator,
									   giterator::CallStmt::ParameterIterator> parameters,
							const giterator::Type* returnType,
							Program* program);

std::auto_ptr<Value>
builtinThreadMutexUnlock(const FunctionValue* func, mmodule::MemoryState* memoryState,
							ConstantsManager* cstManager, GlobalsManager* globalsManager,
							mmodule::Ptr frame,
							std::pair<giterator::CallStmt::ParameterIterator,
									   giterator::CallStmt::ParameterIterator> parameters,
							const giterator::Type* returnType,
							Program* program);

std::auto_ptr<Value>
builtinThreadCondInit(const FunctionValue* func, mmodule::MemoryState* memoryState,
						ConstantsManager* cstManager, GlobalsManager* globalsManager,
						mmodule::Ptr frame,
						std::pair<giterator::CallStmt::ParameterIterator,
								   giterator::CallStmt::ParameterIterator> parameters,
						const giterator::Type* returnType,
						Program* program);

std::auto_ptr<Value>
builtinThreadCondDestroy(const FunctionValue* func, mmodule::MemoryState* memoryState,
							ConstantsManager* cstManager, GlobalsManager* globalsManager,
							mmodule::Ptr frame,
							std::pair<giterator::CallStmt::ParameterIterator,
									   giterator::CallStmt::ParameterIterator> parameters,
							const giterator::Type* returnType,
							Program* program);

std::auto_ptr<Value>
builtinThreadCondSignal(const FunctionValue* func, mmodule::MemoryState* memoryState,
						ConstantsManager* cstManager, GlobalsManager* globalsManager,
						mmodule::Ptr frame,
						std::pair<giterator::CallStmt::ParameterIterator,
								   giterator::CallStmt::ParameterIterator> parameters,
						const giterator::Type* returnType,
						Program* program);

std::auto_ptr<Value>
builtinThreadCondWait(const FunctionValue* func, mmodule::MemoryState* memoryState,
							ConstantsManager* cstManager, GlobalsManager* globalsManager,
							mmodule::Ptr frame,
							std::pair<giterator::CallStmt::ParameterIterator,
									   giterator::CallStmt::ParameterIterator> parameters,
							const giterator::Type* returnType,
							Program* program);
/*@}*/
}

#endif /* INT_BUILTIN_FUNCTIONS_H */
