// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file Condition.cpp
 * Definitions for class for thread condition.
 */

#include "Condition.h"

using namespace mmodule;

namespace ginterpreter
{

Ref<ConditionInfo>
Condition::create(MemoryState* memoryState)
{
	Ref<ConditionInfo> ci = memoryState->allocExpanding<ConditionInfo>();
	memoryState->store(ci, &ConditionInfo::condID, Thread::NO_ID);
	memoryState->store(ci, &ConditionInfo::condState, COND_UNINITIALIZED);
	memoryState->store(ci, &ConditionInfo::unlockedMutexID, Thread::NO_ID);

	Ref<ConditionWaitingThreadsInfo> cwti = memoryState->allocExpanding<ConditionWaitingThreadsInfo>();
	memoryState->store(ci, &ConditionInfo::waitingThreads, cwti);
	memoryState->store(ci, &ConditionInfo::waitingThreadsCount,(size_t) 0);

	return ci;
}

Condition::Condition(ID condID, mmodule::MemoryState* memoryState, Ref<ConditionInfo> conditionInfoRef)
: m_ID (condID), m_state(COND_UNINITIALIZED), m_unlockedMutex(Thread::NO_ID), m_memoryState(memoryState)
{
	reset(conditionInfoRef);
}

Condition::~Condition()
{
}

void
Condition::flush()
{
	m_memoryState->store(m_conditionInfoRef, &ConditionInfo::condID, m_ID);
	m_memoryState->store(m_conditionInfoRef, &ConditionInfo::condState, m_state);
	m_memoryState->store(m_conditionInfoRef, &ConditionInfo::unlockedMutexID, m_unlockedMutex);
	m_memoryState->store(m_conditionInfoRef, &ConditionInfo::waitingThreadsCount, m_waitingThreads.size());

	Ref<ConditionWaitingThreadsInfo> cwti = m_memoryState->load(m_conditionInfoRef, &ConditionInfo::waitingThreads);

	int i = 0;
	for (WaitingThreadsConstIterator i1 = m_waitingThreads.begin(); i1 != m_waitingThreads.end(); ++i1)
	{
		m_memoryState->storeIndexed(cwti, &ConditionWaitingThreadsInfo::waitingThreadID, i++, *i1);
	}

	m_memoryState->store(m_conditionInfoRef, &ConditionInfo::waitingThreads, cwti);
}

void
Condition::reset(Ref<ConditionInfo> conditionInfoRef)
{
	m_conditionInfoRef = conditionInfoRef;

	m_ID = m_memoryState->load(m_conditionInfoRef, &ConditionInfo::condID);
	m_state = m_memoryState->load(m_conditionInfoRef, &ConditionInfo::condState);
	m_unlockedMutex = m_memoryState->load(m_conditionInfoRef, &ConditionInfo::unlockedMutexID);

	m_waitingThreads.clear();

	size_t wtcount = m_memoryState->load(m_conditionInfoRef, &ConditionInfo::waitingThreadsCount);
	Ref<ConditionWaitingThreadsInfo> cwti = m_memoryState->load(m_conditionInfoRef, &ConditionInfo::waitingThreads);

	for (size_t i = 0; i < wtcount; ++i)
	{
		m_waitingThreads.push_back(m_memoryState->loadIndexed(cwti, &ConditionWaitingThreadsInfo::waitingThreadID, i));
	}
}

void
Condition::init()
{
	assert(m_state == COND_UNINITIALIZED);
	m_state = COND_INITIALIZED;
}

void
Condition::destroy()
{
	assert(m_state != COND_UNINITIALIZED);
	m_state = COND_UNINITIALIZED;
	m_unlockedMutex = -1;

	assert(m_waitingThreads.empty());

	/*if (!m_waitingThreads.empty())
		m_waitingThreads.clear();
	else
		throw UninitNonemptyThreadCond(m_ID);*/
}

Condition::ID
Condition::getID() const
{
	return m_ID;
}

void
Condition::addWaitingThread(Mutex::ID mutexID, Thread::ID threadID)
{
	assert(m_unlockedMutex == mutexID);
	m_waitingThreads.push_back(threadID);
}

Thread::ID
Condition::getFrontWaitingThread()
{
	if (m_waitingThreads.empty())
		return -1;

	Thread::ID threadID = m_waitingThreads.front();
	m_waitingThreads.pop_front();

	return threadID;
}

void
Condition::setUnlockedMutex(Mutex::ID mutexID)
{
	//assert(m_unlockedMutex == -1);
	m_unlockedMutex = mutexID;
}

Mutex::ID
Condition::getUnlockedMutex() const
{
	assert(m_unlockedMutex != -1);
	return m_unlockedMutex;
}

bool Condition::isInUse() const
{
	if (m_state == COND_UNINITIALIZED)
		return false;

	return m_waitingThreads.size() != 0;
}

}
