// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file Condition.h
 * Declaration of a class for thread condition.
 */

#ifndef INT_CONDITION_H
#define INT_CONDITION_H

#include "ConditionState.h"
#include "Mutex.h"
#include "Thread.h"
#include "ConditionInfo.h"

#include "../mm/MemoryState.h"

#include <deque>
#include <boost/noncopyable.hpp>

namespace ginterpreter
{

/**
 * Class representing thread condition.
 *
 * Thread condition can be used as a way to synchronize threads.
 * While mutexes implement synchronization by controlling thread access to data,
 * condition variables allow threads to synchronize based upon the actual value of
 * data. It is always used in conjunction with a mutex variable.
 * Condition operations include wait and signal (described in the ThreadPool class).
 *
 * This is a dumb class, everything important is done in class ThreadPool.
 * It is more of a structure holding the thread condition ID,
 * threads waiting for the condition, and state of the condition.
 * Thread conditions are represented in the simulated memory
 * by the ConditionInfo class. This class stores a reference to
 * this structure in the simulated memory
 *
 * @see ThreadPool, Mutex, ConditionInfo
 */
class Condition : public boost::noncopyable
{
public:
	/*! type of the ID of condition */
	typedef int ID;
private:
	/*! container for threads waiting on this condition. */
	typedef std::deque<Thread::ID> WaitingThreads;
	/*! Iterator for container for threads waiting on this condition. */
	typedef WaitingThreads::const_iterator WaitingThreadsConstIterator;
	/*! unique ID of this condition. */
	ID m_ID;
	/*! current state of this condition. */
	ConditionState m_state;
	/*! Thread waiting for this condition. */
	WaitingThreads m_waitingThreads;
	/*! ID of mutex, that has been unlocked by thread(s) waiting for this condition. */
	Mutex::ID m_unlockedMutex;
	/*! Reference structure representing this condition in the simulated memory. */
	mmodule::Ref<ConditionInfo> m_conditionInfoRef;
	/*! Interface to the simulated memory. */
	mmodule::MemoryState* m_memoryState;
public:
	/**
	 * Initializes condition memory structures to be stored in the simulated memory.
	 *
	 * @param memoryState interface to the simulated memory
	 * @return initialized structure representing condition in the simulated memory
	 */
	static mmodule::Ref<ConditionInfo> create (mmodule::MemoryState* memoryState);
	/**
	 * Constructor, creates a new thread condition.
	 * Condition::create() should be called before this to get a structure representing
	 * the condition in the simulated memory.
	 *
	 * @param condID unique ID of the new condition
	 * @param memoryState interface to the simulated memory
	 * @param conditionInfoRef initialized structure representing condition in the simulated memory
	 * created by previous call to Condition::create()
	 *
	 * @see Condition::create()
	 */
	Condition(ID condID, mmodule::MemoryState* memoryState, mmodule::Ref<ConditionInfo> conditionInfoRef);
	/**
	 * Destructor.
	 */
	~Condition();
	/**
	 * Flushes all the condition memory structures in the simulated memory.
	 */
	void flush();
	/**
	 * Reset all the the memory structures according to a condition simulated
	 * memory representation.
	 * To be used after backtracking, when previous states of the program
	 * get reset and the condition structures are no more actual.
	 *
	 * All the information from ConditionInfo structure will
	 * be read and the condition data structures will be updated.
	 *
	 * @param conditionInfoRef info structure according to which
	 * reset the condition
	 *
	 * @see mmodule::MemoryState::backtrack()
	 */
	void reset(mmodule::Ref<ConditionInfo> conditionInfoRef);
	/**
	 * Initialize a condition for use, it has to be uninitialized.
	 * After initialization, the state of the condition becomes unlocked.
	 */
	void init();
	/**
	 * Uninitialize a condition, it cannot be uninitialized and no threads can
	 * wait for the condition.
	 * A destroyed condition object can be reinitialized using init().
	 * Signal/wait functions cannot be used on uninitialized condition.
	 *
	 */
	void destroy();
	/** Acquire the unique ID of the condition.
	 *
	 * @return unique ID of this condition
	 */
	Condition::ID getID() const;
	/**
	 * Store ID of thread waiting for this condition it to a queue of such threads.
	 * Mutex associated with this conditon has to be already set by @c setUnlockedMutex()
	 * and is has to be the same as the first argument of this function.
	 *
	 * @param mutexID ID of the unlocked mutex (used only for comparison with stored mutex ID)
	 * @param threadID ID of the thread that unlocked mutex with ID @c mutexID
	 *
	 * @see setUnlockedMutex()
	 */
	void addWaitingThread(Mutex::ID mutexID, Thread::ID threadID);
	/**
	 * Obtain an ID of one of the threads (first in queue) waiting for current condition.
	 * -1 if there is no such thread.
	 * The information gets dequeued from queue storing such information using @c addWaitingThread().
	 *
	 * @return ID of one of the threads waiting for current condition, or -1 if there is no such thread
	 */
	Thread::ID getFrontWaitingThread();
	/**
	 * Set ID of mutex, that has been unlocked by thread(s) waiting for this condition.
	 * If such ID is already set, the value is overwritten.
	 *
	 * @param mutexID ID of mutex, that has been unlocked by thread(s) waiting for this condition
	 */
	void setUnlockedMutex(Mutex::ID mutexID);
	/**
	 * Obtain ID of mutex, that has been unlocked by thread(s) waiting for this condition.
	 * The mutex ID has to be set.
	 *
	 * @return ID of mutex, that has been unlocked by thread(s) waiting for this condition
	 */
	Mutex::ID getUnlockedMutex() const;
	/**
	 * Find out whether this condition is currently used by some condition wait function.
	 * The condition cannot be uninitialized.
	 *
	 * @return true if this condition is currently used by some condition wait function,
	 * false otherwise
	 *
	 * @see ThreadPool::waitCondition
	 */
	bool isInUse() const;
};

}

#endif /* INT_CONDITION_H */
