// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file ConstantsManager.h
 * Declaration of class that manages constants in the interpreted program.
 */
#ifndef INT_CONSTANTS_MANAGER_H
#define INT_CONSTANTS_MANAGER_H

#include "../git/git_fwd.h"
#include "../mm/fwd.h"
#include "../mm/Ptr.h"

#include <map>
#include <memory>
#include <boost/noncopyable.hpp>

namespace ginterpreter {
    
/**
 * Class that manages constants in the interpreted program.
 * This class gets filled during creation of GIMPLE++ representation
 * of a program inside @c GimpleListener class passed
 * as a constructor to @c Gimple class that holds the GIMPLE++
 * representation.
 *
 * For each constant and function definition in a GIMPLE++ representation
 * of the program, @c addCst() or @c addFunc() functions get called
 * on @c ConstantsManager, which saves the constant or function
 * definition into simulated memory.
 * They can be later retrieved during the program execution simulation
 * by calling @c getCst() and @c getFunc() functions.
 *
 * @see GimpleListener
 */
class ConstantsManager : private boost::noncopyable {
  public:
	/*! iterator for the container holding pointers to declarations to the simulated memory */
	typedef std::map<const giterator::Decl*, mmodule::Ptr>::const_iterator DeclPtrsIterator;
  private:
	/*! iterator for the container holding pointers to constants to the simulated memory */
    typedef std::map<const giterator::Constant*, mmodule::Ptr> CstPtrs;
    /*! iterator for the container holding pointers to declarations to the simulated memory */
    typedef std::map<const giterator::Decl*, mmodule::Ptr> DeclPtrs;
    /*! interface to the simulated memory */
    mmodule::MemoryState* m_memoryState;
    /*! map containing pointers for stored constants */
    CstPtrs m_cstPtrs;
    /*! map containing pointers for stored functions */
    DeclPtrs m_declPtrs;
    /*! pointer to the beginning of the read-only area */
    mmodule::Ptr m_roAreaPtr;
    /*! the next free address in the read-only area */
    mmodule::Ptr m_nextFreeRoAreaPtr;
  public:
    /** Constructor.
     *
     * @param memoryState interface to the simulated memory.
     */
    ConstantsManager (mmodule::MemoryState* memoryState);
    
    /**
     * Returns a pointer to the constant within the simulated memory.
     *
     * @param constant GIMPLE++ representation of the constant we are
     * looking for
     * @return pointer to the constant @c constant within simulated memory
     */
    mmodule::Ptr getCstPtr (const giterator::Constant* constant) const;

    /**
     * Stores a constant in the simulated memory.
     *
     * @param constant constant to be stored
     * @return pointer to the constant just added into the simulated memory
     */
    mmodule::Ptr addCst (const giterator::Constant* constant);
    
    /**
     * Returns a pointer to the function.
     *
     * @param decl declaration of the function
     * @return pointer to the function declared by @c decl
     */
    mmodule::Ptr getFuncPtr (const giterator::Decl* decl) const;

    /**
     * Stores function in the simulated memory.
     *
     * @param decl declaration of the function
     * @param value the value holding the body of the function
     * @return pointer to the just stored function
     */
    mmodule::Ptr addFunc (const giterator::Decl* decl, 
                          std::auto_ptr<mmodule::Content> value);

    /**
     * Returns pointer to the beginning of the read-only area containing all
     * the constants and functions.
     */
    mmodule::Ptr getRoAreaPtr () const;

    /**
     * Get all the function stored inside this class.
     *
     * @return pair of iterators <iter. to first function , iter. beyond last function>
     */
    std::pair<DeclPtrsIterator, DeclPtrsIterator>
    funcs() const;
};

}

#endif /* INT_CONSTANTS_MANAGER_H */




