// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file EnumeralValue.cpp
 * Definitions for class representing simulated enumeral value.
 */
#include "EnumeralValue.h"

#include <typeinfo>
#include <ostream>
#include <boost/functional/hash.hpp>

namespace ginterpreter
{

EnumeralValue::EnumeralValue (const giterator::Type* type, uint64_t value)
	: IntegerValue (type, value)
{}

Value* EnumeralValue::clone () const
{
  return new EnumeralValue (*this);
}

std::auto_ptr<Value> EnumeralValue::addOp (const Value& v) const
{
  const EnumeralValue& iv = checkedCast (v);
  return std::auto_ptr<Value> (new EnumeralValue (getType(), m_value + iv.m_value));
}

std::auto_ptr<Value> EnumeralValue::subOp (const Value& v) const
{
  const EnumeralValue& iv = checkedCast (v);
  return std::auto_ptr<Value> (new EnumeralValue (getType (), m_value - iv.m_value));
}

std::auto_ptr<Value> EnumeralValue::truncDivOp (const Value& v) const
{
  const EnumeralValue& iv = checkedCast (v);
  return std::auto_ptr<Value> (new EnumeralValue (getType (), m_value / iv.m_value));
}

std::auto_ptr<Value> EnumeralValue::truncModOp (const Value& v) const
{
  const EnumeralValue& iv = checkedCast (v);
  return std::auto_ptr<Value> (new EnumeralValue (getType (), m_value % iv.m_value));
}

std::auto_ptr<Value> EnumeralValue::multOp (const Value& v) const
{
  const EnumeralValue& iv = checkedCast (v);
  return std::auto_ptr<Value> (new EnumeralValue (getType (), m_value * iv.m_value));
}

std::auto_ptr<Value> EnumeralValue::absOp () const
{
	if (m_value < 0)
		return std::auto_ptr<Value> (new EnumeralValue (getType (), -m_value));
	else
		return std::auto_ptr<Value> (new EnumeralValue (getType (), m_value ));
}

bool EnumeralValue::operator== (const Value& v) const
{
  checkType (v);
  const EnumeralValue& iv = static_cast<const EnumeralValue&> (v);
  return m_value == iv.m_value;
}

bool EnumeralValue::operator< (const Value& v) const
{
  checkType (v);
  const EnumeralValue& iv = static_cast<const EnumeralValue&> (v);
  return m_value < iv.m_value;
}

void EnumeralValue::print (std::ostream& os) const
{
	if (getEnumeralType ()->isUnsigned ())
		os << m_value;
	else
		os << (int64_t)m_value;
}

const giterator::EnumType* EnumeralValue::getEnumeralType () const
{
	assert (typeid (*getType ()) == typeid (giterator::EnumType));
	return static_cast<const giterator::EnumType*> (getType ());
}

void EnumeralValue::checkType (const Value& e) const
{
	assert (getType()->isSame(*e.getType()));     (void)e;
}

const EnumeralValue& EnumeralValue::checkedCast (const Value& v) const
{
	checkType (v);
	return static_cast<const EnumeralValue&> (v);
}

}
