// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file GimpleDumper.cpp
 * Definitions for class for dumping GIMPLE++ representation of a program into a file or stream.
 */

#include "GimpleDumper.h"

#include "../git/Function.h"
#include "../git/FunctionDecl.h"
#include "FunctionValue.h"
#include "ConstantsManager.h"
#include "../mm/MemoryState.h"
#include "SymbolicPrinter.h"

#include <fstream>

using namespace mmodule;
using namespace giterator;

namespace ginterpreter
{

GimpleDumper::GimpleDumper(ConstantsManager* cstManager, MemoryState* memoryState)
: m_cstManager(cstManager), m_memoryState(memoryState) {}

void
GimpleDumper::dump(const char* filename)
{
	std::ofstream file(filename);

	dump(file);
	file.close();
}

void GimpleDumper::dump(std::ostream& outstream)
{
	assert(outstream.good());

	std::pair<ConstantsManager::DeclPtrsIterator, ConstantsManager::DeclPtrsIterator> funcs =
			m_cstManager->funcs();

	for (ConstantsManager::DeclPtrsIterator i1 = funcs.first; i1 != funcs.second; ++i1)
	{
		const FunctionValue* fv = dynamic_cast<const FunctionValue*> (m_memoryState->load(i1->second));
		// we are interested only in user-defined functions
		if (fv == NULL)
			continue;

		// dump function type and header
		const FunctionDecl* fd = dynamic_cast<const FunctionDecl*> (i1->first);
		assert(fd);
		outstream << fd->getFunctionType()->getReturnType()->getName();
		//

		// dump function parameter declarations
		outstream << " " << fd->getName() << " (";
		const Function* func = fv->getFunction();
		const Function::Declarations& paramDecls = func->getParamDeclarations();
		for (Function::Declarations::const_iterator i2 = paramDecls.begin();
				i2 != paramDecls.end();
				++i2)
		{
			outstream << (*i2)->getType()->getName() << " " << (*i2)->getName();

			if (i2 + 1 != paramDecls.end())
				outstream << ",";
		}
		outstream << ")" << std::endl;
		//

		// dump function body
		outstream << "{" << std::endl;
		ginterpreter::SymbolicPrinter sp(outstream);

		// first dump local declarations
		const Function::Declarations& localDecls = func->getLocalDeclarations();
		for (Function::Declarations::const_iterator i1 = localDecls.begin(); i1 != localDecls.end(); ++i1)
		{
			outstream << "\t" << (*i1)->getType()->getName() << " " << (*i1)->getName() << ";" << std::endl;
		}
		outstream << std::endl;
		//

		// second dump all the blocks of statements
		const Function::Blocks& blocks = func->getBlocks();
		for (Function::Blocks::const_iterator i1 = blocks.begin(); i1 != blocks.end(); ++i1)
		{
			std::pair<Block::StmtIterator, Block::StmtIterator> stmts
					= (*i1).stmts();

			for (Block::StmtIterator i2 = stmts.first; i2 != stmts.second; ++i2)
			{
				outstream << "\t";
				i2->accept(sp);
				outstream << std::endl;
			}
		}
		//

		outstream << "}" << std::endl;
	}
}

}
