// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file GimpleListener.cpp
 * Definitions for listener for the Gimple class.
 */
#include "GimpleListener.h"

#include "ConstantsManager.h"
#include "GlobalsManager.h"
#include "BuiltinFunctionValue.h"
#include "ThreadFunctionValue.h"
#include "BuiltinFunctions.h"
#include "FunctionValue.h"

#include "../git/ThreadConfigParser.h"
#include "../git/Decl.h"
#include "../git/FunctionDecl.h"
#include "../git/VarDecl.h"
#include "../git/RecordDecl.h"

#include "../mm/Ptr.h"

#include "../Error.h"


using namespace mmodule;
using namespace giterator;


namespace ginterpreter {

GimpleListener::GimpleListener (ConstantsManager* cstManager, GlobalsManager* globalsManager)
: m_cstManager (cstManager), m_globalsManager(globalsManager)
{
	giterator::ThreadConfigParser tcp;

	m_threadContent = tcp.ParseThreadConfig("thread.config");
}

void 
GimpleListener::onCst (const Constant* cst)
{
    Ptr cstPtr = m_cstManager->getCstPtr (cst);
    if (cstPtr.isNull ())
    {
        m_cstManager->addCst (cst);
    }
}

void 
GimpleListener::onDecl (const Decl* decl)
{
    class DeclVisitor : public Decl::Visitor {
      private:
        ConstantsManager* m_cstManager;
        GlobalsManager* m_globalsManager;

        giterator::ThreadConfigContent m_threadContent;
            
      public:
        DeclVisitor (ConstantsManager* cstManager,
        			 GlobalsManager* globalsManager,
        			 giterator::ThreadConfigContent threadContent)
        : m_cstManager (cstManager),
          m_globalsManager (globalsManager),
          m_threadContent(threadContent)
        {}
        
        void handleThreadFunc(const FunctionDecl& decl)
        {
        	ThreadFunctionValue::FuncMethod funcMethod = 0;
        	thread_function_code tfc;

        	if (decl.getName().compare(m_threadContent.createName) == 0)
        	{
        		funcMethod = &builtinThreadCreate;
        		tfc = THREAD_CREATE;
        	}
        	else if (decl.getName().compare(m_threadContent.joinName) == 0)
        	{
        		funcMethod = &builtinThreadJoin;
        		tfc = THREAD_JOIN;
        	}
        	else if (decl.getName().compare(m_threadContent.exitName) == 0)
        	{
        		funcMethod = &builtinThreadExit;
        		tfc = THREAD_EXIT;
        	}
        	else if (decl.getName().compare(m_threadContent.mutexInitName) == 0)
        	{
        		funcMethod = &builtinThreadMutexInit;
        		tfc = THREAD_MUTEX_INIT;
        	}
        	else if (decl.getName().compare(m_threadContent.mutexDestroyName) == 0)
			{
        		funcMethod = &builtinThreadMutexDestroy;
        		tfc = THREAD_MUTEX_DESTROY;
			}
        	else if (decl.getName().compare(m_threadContent.mutexLockName) == 0)
			{
				funcMethod = &builtinThreadMutexLock;
				tfc = THREAD_MUTEX_LOCK;
			}
        	else if (decl.getName().compare(m_threadContent.mutexTryLockName) == 0)
			{
				funcMethod = &builtinThreadMutexTryLock;
				tfc = THREAD_MUTEX_TRYLOCK;
			}
        	else if (decl.getName().compare(m_threadContent.mutexUnlockName) == 0)
			{
				funcMethod = &builtinThreadMutexUnlock;
				tfc = THREAD_MUTEX_UNLOCK;
			}
        	else if (decl.getName().compare(m_threadContent.condInitName) == 0)
			{
				funcMethod = &builtinThreadCondInit;
				tfc = THREAD_COND_INIT;
			}
			else if (decl.getName().compare(m_threadContent.condDestroyName) == 0)
			{
				funcMethod = &builtinThreadCondDestroy;
				tfc = THREAD_COND_DESTROY;
			}
			else if (decl.getName().compare(m_threadContent.condSignalName) == 0)
			{
				funcMethod = &builtinThreadCondSignal;
				tfc = THREAD_COND_SIGNAL;
			}
			else if (decl.getName().compare(m_threadContent.condWaitName) == 0)
			{
				funcMethod = &builtinThreadCondWait;
				tfc = THREAD_COND_WAIT;
			}
			else if (decl.getName().compare(BUILT_IN_SLEEP_NAME) == 0)
			{
				funcMethod = &builtinSleep;
				tfc = THREAD_SLEEP;
			}
        	else
        	{
        		ni_error ("thread function %s", decl.getName ().c_str ());
            }

            LOG_DEBUG ("adding thread function %s", decl.getName ().c_str ());
            std::auto_ptr<Content> funcPtr (
                    new ThreadFunctionValue (tfc, decl.getFunctionType (), funcMethod));

            m_cstManager->addFunc (&decl, funcPtr);
        }

        void visit (const FunctionDecl& decl)
        {
        	// If function definintion already exists, we don't
        	// have to do anything
            if (!m_cstManager->getFuncPtr (&decl).isNull ())
                return;
    
            BuiltinFunctionValue::FuncMethod funcMethod = 0;
    
            switch (decl.getCode ())
            {
            	case NOT_BUILT_IN_FUNC:
            		if (m_threadContent.containsName(decl.getName ()))
            			handleThreadFunc(decl);
            		return;
            	case GCC_BUILT_IN_SLEEP:
            		// We handle sleep() as thread function, since it puts to sleep current thread,
					// therefore it works with thread pool.
            		handleThreadFunc(decl);
            		return;
                case GCC_BUILT_IN_PRINTF:
                    funcMethod = &builtinPrintf;
                    break;
                case GCC_BUILT_IN_PUTS:
                    funcMethod = &builtinPuts;
                    break;
                case GCC_BUILT_IN_PUTCHAR:
                    funcMethod = &builtinPutchar;
                    break;
                case GCC_BUILT_IN_SCANF:
                    funcMethod = &builtinScanf;
                    break;
                case GCC_BUILT_IN_MALLOC:
                    funcMethod = &builtinMalloc;
                    break;
                case GCC_BUILT_IN_FREE:
                    funcMethod = &builtinFree;
                    break;
                case GCC_BUILT_IN_ABORT:
                	funcMethod = &builtinAbort;
                	break;
                case GCC_BUILT_IN_ABS:
                case GCC_BUILT_IN_LABS:
                case GCC_BUILT_IN_LLABS:
                	funcMethod = &builtinAbs;
                	break;
                case GCC_BUILT_IN_ATOF:
					funcMethod = &builtinAtof;
					break;
				case GCC_BUILT_IN_ATOI:
					funcMethod = &builtinAtoi;
				case GCC_BUILT_IN_ATOL:
					funcMethod = &builtinAtol;
				case GCC_BUILT_IN_ATOLL:
					funcMethod = &builtinAtoll;
					break;
				case GCC_BUILT_IN_EXIT:
					funcMethod = &builtinExit;
					break;
				case GCC_BUILT_IN_RAND:
				case GCC_BUILT_IN_RANDOM:
					funcMethod = &builtinRand;
					break;
				case GCC_BUILT_IN_SRAND:
				case GCC_BUILT_IN_SRANDOM:
					funcMethod = &builtinSrand;
					break;
				case GCC_BUILT_IN_STRTOD:
				case GCC_BUILT_IN_STRTOLD:
				case GCC_BUILT_IN_STRTOF:
                	funcMethod = &builtinStrtod;
                	break;
                case GCC_BUILT_IN_STRTOL:
                case GCC_BUILT_IN_STRTOLL:
                case GCC_BUILT_IN_STRTOUL:
                case GCC_BUILT_IN_STRTOULL:
                	funcMethod = &builtinStrtol;
                	break;
                case GCC_BUILT_IN_MEMCPY:
                	funcMethod = &builtinMemcpy;
                	break;
                case GCC_BUILT_IN_MEMMOVE:
                	funcMethod = &builtinMemmove;
                	break;
                case GCC_BUILT_IN_STRCAT:
                case GCC_BUILT_IN_STRNCAT:
                	funcMethod = &builtinStrcat;
                	break;
                case GCC_BUILT_IN_STRCPY:
                case GCC_BUILT_IN_STRCNPY:
                	funcMethod = &builtinStrcpy;
                	break;
                case GCC_BUILT_IN_STRCMP:
                case GCC_BUILT_IN_STRNCMP:
                	funcMethod = &builtinStrcmp;
                	break;
                case GCC_BUILT_IN_STRLEN:
                	funcMethod = &builtinStrlen;
                	break;
                case GCC_BUILT_IN_ASSERT:
                	funcMethod = &builtinAssert;
                	break;
                	/* Moved to thread functions
                case GCC_BUILT_IN_SLEEP:
                	funcMethod = &builtinSleep;
                	break;
                	*/
                default:
                    ni_error ("built-in function %s", decl.getName ().c_str ());
            }

            LOG_DEBUG ("adding built-in function %s", decl.getName ().c_str ());
            std::auto_ptr<Content> funcPtr (
                    new BuiltinFunctionValue (decl.getCode (), 
                                              decl.getFunctionType (), funcMethod));
        
            m_cstManager->addFunc (&decl, funcPtr);
        }
        
        void visit (const ParamDecl&) {}

        void visit (const VarDecl& decl)
        {
        	if (decl.getLocal())
        		return;

        	Ptr declPtr = m_globalsManager->getGlobalDeclPtr (&decl);
        	if (!declPtr.isNull ())
        		return;

        	m_globalsManager->addGlobalDecl (&decl);
        }

		void visit (const RecordDecl& decl)
		{
			Ptr declPtr = m_globalsManager->getGlobalDeclPtr (&decl);
			if (!declPtr.isNull ())
				return;

			const RecordType* type = dynamic_cast<const RecordType*> (decl.getType());
			if (type)
			{
				m_globalsManager->addGlobalDecl (&decl);
			}
		}

        void visit (const FieldDecl&) {}
		void visit (const LabelDecl&) {}
		void visit (const ResultDecl&) {}
		void visit (const TypeDecl&) {}
		void visit (const ConstDecl&) {}
    };
    
    DeclVisitor dv (m_cstManager, m_globalsManager, m_threadContent);
    
    decl->accept (dv);
}


void 
GimpleListener::onFuncDef (const FunctionDecl* decl, const Function* func)
{
    std::auto_ptr<Content> cont (new FunctionValue (func));
    m_cstManager->addFunc (decl, cont);
}

void
GimpleListener::allocGlobalArea()
{
	m_globalsManager->allocArea();
}

}
