// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file GimpleListener.h
 * Declaration of listener for Gimple class.
 */
#ifndef INT_GIMPLE_LISTENER_H
#define INT_GIMPLE_LISTENER_H

#include "../git/Gimple.h"
#include "../git/ThreadConfigParser.h"

namespace ginterpreter {
    
class ConstantsManager;
class GlobalsManager;

/**
 * Listener for the @c Gimple class.
 *
 * During the creation of a GIMPLE++ representation of the program,
 * @c GimpleListener class is passed to a constructor of @c giterator::Gimple
 * class. @c Gimple+ calls the appropriate @c GimpleListener methods when
 * anything interesting for the Interpreter happens. For each constant and
 * function definition in a GIMPLE++ representation of the program, @c addCst()
 * or @c addFunc() functions get called on @c ConstantsManager, which saves
 * the constant or function definition into simulated memory. They can be later
 * retrieved during the program execution simulation by calling @c getCst() and
 * @c getFunc() functions. The same principle is applied for global variables,
 * that are maintained by @c GlobalsManager(). Function definition holds all
 * the statements and local declarations inside the function.
 *
 * @c GimpleListener ensures that the simulated memory is properly initialized with
 * all required data needed to start interpreting or model checking of the program.
 *
 * @see ConstantsManager, GlobalsManager
 */
class GimpleListener : public giterator::Gimple::Listener {
  private:
    /*! Constants and function declarations of the program  */
    ConstantsManager* m_cstManager;
    /*! Global declarations of the program */
    GlobalsManager* m_globalsManager;
    /*! content of the thread configuration file */
    giterator::ThreadConfigContent m_threadContent;
  public:
    /**
     * Constructor, creates a new gimple listener.
     *
     * @param cstManager where to store constants and function declarations of the program
     * @param globalsManager where to store global declarations of the program
     */
    GimpleListener (ConstantsManager* cstManager, GlobalsManager* globalsManager);
    /**
     * Stores constant in the simulated memory.
     *
     * @param cst constant to be stored in the simulated memory
     */
    void onCst (const giterator::Constant* cst);

    /**
     * If @c decl is a declaration of a supported built-in or standard function,
     * or a declaration of a supported thread function,
     * this function stores it in the simulated memory.
     * @c BuiltinFunctionValue or @c ThreadFunctionValue gets created
     * and stored into simulated memory
     *
     * @param decl declaration to (possibly) store into simulated memory
     * @see BuiltinFunctionValue, ThreadFunctionValue
     */
    void onDecl (const giterator::Decl* decl);
    /**
     * Stores the user-defined function into the simulated memory.
     * @c FunctionValue gets created according to parameters passed to
     * this function and stored in the simulated memory.
     *
     * @param decl declaration of the stored function
     * @param func definition of the stored function
     *
     * @see FunctionValue
     */
    void onFuncDef (const giterator::FunctionDecl* decl, const giterator::Function* func);
    /**
     * Allocate an area for global declarations inside
     * @c GlobalsManager.
     */
    void allocGlobalArea();
};

}

#endif /* INT_GIMPLE_LISTENER_H */
