// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file GimpleStmtInterpreter.cpp
 * Definitions for class that interprets GIMPLE++ statements.
 */
#include "GimpleStmtInterpreter.h"

#include "Steps.h"
#include "OperationInterpreter.h"
#include "ValueWriter.h"
#include "ValueReader.h"
#include "FunctionVarsMapper.h"
#include "NameResolver.h"

#include "BoolValue.h"
#include "PtrValue.h"
#include "IntegerValue.h"
#include "StringValue.h"
#include "RecordValue.h"
#include "FunctionValue.h"
#include "BuiltinFunctionValue.h"
#include "ThreadFunctionValue.h"
#include "VoidPtrValue.h"
#include "ArrayConstructorValue.h"

#include "../git/Gimple.h"
#include "../git/Decl.h"
#include "../git/Function.h"
#include "../git/FunctionDecl.h"
#include "../git/FunctionType.h"
#include "../git/PtrType.h"
#include "../git/Type.h"
#include "../git/Stmts.h"
#include "../git/DeclOperand.h"
#include "../git/AddrOperand.h"
#include "../git/LabelExprOperand.h"
#include "../git/ArrayType.h"

#include "../mm/MemoryState.h"

#include "../Error.h"

#include <memory>
#include <typeinfo>
#include <iostream>
#include <sstream>

using namespace mmodule;
using namespace giterator;


namespace ginterpreter {

/**
 * Sets the values of parameters in the stack frame of called function.
 * 
 * @tparam ValIt type of the iterator through parameter values
 * @tparam DeclIt type of the iterator through parameter declarations
 * @param memoryState interface to the simulated memory
 * @param cstManager constants and function declarations of the program
 * @param globalsManager global declarations of the program
 * @param callerFunc caller function
 * @param calleeFunc function that is beeing called
 * @param valBegin interator pointing to the first parameter value
 * @param valEnd interator pointing right after the last parameter value
 * @param callersFrame pointer to the beginning of the stack frame of the caller
 * @param declBegin iterator pointing to the first parameter declaration
 * @param declEnd iterator pointing right after the last parameter declaration
 * @param calleeFrame pointer to the beginning of the stack frame of the callee
 */
template <typename ValIt, typename DeclIt>
static void
setParameters (MemoryState* memoryState, ConstantsManager* cstManager, GlobalsManager* globalsManager,
               const FunctionValue* callerFunc, const FunctionValue* calleeFunc,  
               ValIt valBegin, ValIt valEnd, Ptr callersFrame,
               DeclIt declBegin, DeclIt declEnd, Ptr calleeFrame)
{  
    const FunctionVarsMapper* mapper = calleeFunc->getVarsMapper ();
    
    ValIt vi;
    DeclIt di;
    for (vi = valBegin, di = declBegin; di != declEnd; vi++, di++) 
    {
        assert (vi != valEnd);
        
        ValueReader reader (callerFunc, memoryState, cstManager, globalsManager, callersFrame);
        vi->accept (reader);

        Ptr varPtr = calleeFrame + mapper->getVarOffset ((*di)->getUid ());

        ValueWriter writer (calleeFunc, memoryState, cstManager, globalsManager, calleeFrame, reader.acquireValue());
		writer.storeViaPointer(varPtr);
    }
    
    ni_assert (vi == valEnd, "variable function parameter count");
}



GimpleStmtInterpreter::GimpleStmtInterpreter (Gimple* gimple, const FunctionValue* func,
    MemoryState* memoryState, ConstantsManager* cstManager, GlobalsManager* globalsManager,
    Program* program, Ptr topFrame) :
        m_gimple (gimple), m_memoryState (memoryState), m_cstManager (cstManager), m_globalsManager(globalsManager),
                  m_program (program), m_topFrame (topFrame), m_func (func) {}

std::auto_ptr<Step> 
GimpleStmtInterpreter::acquireStep ()
{ 
    assert (m_step.get ());
    return m_step;
}

/*
ArrayConstructorValue* convertStrToArrayInitializer(const StringValue* strVal, const ArrayType* arrType)
{
	const Type* elementType = arrType->getElementType();
	std::string str = strVal->getValue();
	ArrayConstructorValue* acv = new ArrayConstructorValue(arrType);
	for(size_t index = 0; index < str.size(); ++index)
	{
		// TODO: first type
		IntegerValue* indexVal = new IntegerValue(elementType, index);
		IntegerValue* valueVal = new IntegerValue(elementType, (uint64_t) str[index]);
		acv->addValue(indexVal, std::auto_ptr<Value> (valueVal));
		acv->addValueOwner(std::auto_ptr<Value> (indexVal));
	}

	return acv;
}*/


/**
 * Interprets the GIMPLE++ assign statement.
 */
void 
GimpleStmtInterpreter::visit (const AssignStmt& stmt)
{
    OperationInterpreter opInt (m_func, m_memoryState, m_cstManager, m_globalsManager, m_gimple, m_topFrame);
    
    const giterator::Operation* oper = stmt.getOperation ();
    oper->accept (opInt);
  
    std::auto_ptr<Value> res = opInt.acquireResult ();
    
    // Check if assigning integer into pointer
    const giterator::Operand* lhs = stmt.getLhs ();
    if (typeid (*lhs->getType ()) == typeid (PtrType))
    {
        const IntegerValue* integerVal = dynamic_cast<const IntegerValue*> (res.get ());
        if (integerVal && integerVal->isZero ())
            res.reset (new PtrValue (lhs->getType (), Ptr ()));
    }

    // check if assigning void pointer
    if (typeid(*res.get()) == typeid(VoidPtrValue))//(*oper->getResultType()) == typeid(const PtrType))
    {
    	const VoidPtrValue* voidPtrVal = dynamic_cast<const VoidPtrValue*> (res.get());
    	const Value* val = dynamic_cast<const Value*> (m_memoryState->load(voidPtrVal->getPtr()));
    	res.reset(val->clone());
    }

    // check if initializing a char array with a string constant
    /*if (typeid (*lhs->getType ()) == typeid (ArrayType))
    {
    	const StringValue* strVal = dynamic_cast<const StringValue*> (res.get ());
    	if (strVal)
    	{
    		const giterator::ArrayType* arrType = dynamic_cast<const giterator::ArrayType*> (lhs->getType ());
			assert(arrType);

			res.reset(convertStrToArrayInitializer(strVal, arrType));
    	}
    }*/

    ValueWriter writer (m_func, m_memoryState, m_cstManager, m_globalsManager, m_topFrame, res);
    lhs->accept (writer);

    m_step.reset (new NextStep ());
}

/**
 * Interprets the GIMPLE++ cond statement.
 */
void 
GimpleStmtInterpreter::visit (const CondStmt& stmt)
{
  const Operation* oper = stmt.getOperation ();
  OperationInterpreter opInt (m_func, m_memoryState, m_cstManager, m_globalsManager, m_gimple, m_topFrame);
  oper->accept (opInt);
  
  // TODO This is ugly.
  std::auto_ptr<Value> res (opInt.acquireResult ());
  BoolValue* resBool = dynamic_cast<BoolValue*>(res.get ());
  assert (resBool);

  m_step.reset (new CondStep (resBool->getValue ()));
}


/**
 * Interprets the GIMPLE++ return statement.
 */
void
GimpleStmtInterpreter::visit (const ReturnStmt& stmt)
{
  if (stmt.getReturnValue ())
  {
    ValueReader valReader (m_func, m_memoryState, m_cstManager, m_globalsManager, m_topFrame);
    stmt.getReturnValue ()->accept (valReader);
  
    std::auto_ptr<Content> res = static_cast<std::auto_ptr<Content> >(valReader.acquireValue ());
  
    m_memoryState->store (m_topFrame + m_func->getVarsMapper ()->getRetValOffset (), res);
  }
  
  m_step.reset (new ReturnStep ());
}


/**
 * Interprets the GIMPLE++ call statement.
 */
void 
GimpleStmtInterpreter::visit (const CallStmt& stmt)
{
    // TODO Load pointer to function as any other value
    ValueReader reader (m_func, m_memoryState, m_cstManager, m_globalsManager, m_topFrame);
    stmt.getFunction ()->accept (reader);
    
    const PtrValue* funcPtr = dynamic_cast<const PtrValue*> (reader.getValue ());

    if (funcPtr->getPtr().isNull())
    {
    	NameResolver nameRes;
    	stmt.getFunction()->accept(nameRes);
    	ni_error("Function '%s' is not implemented in GMC.", nameRes.getName().c_str());
    }

    const Content* content = m_memoryState->load (funcPtr->getPtr ());
    
    if (typeid (*content) == typeid (const BuiltinFunctionValue))
    {
        const BuiltinFunctionValue* bi = dynamic_cast<const BuiltinFunctionValue*> (content);
        
        std::auto_ptr<Value> res = bi->getMethod () (m_func, m_memoryState, m_cstManager, m_globalsManager,
                m_topFrame, stmt.parameters (), bi->getFunctionType ()->getReturnType ());
        
        if (stmt.getLhs ())
        {
            ValueWriter writer (m_func, m_memoryState, m_cstManager, m_globalsManager, m_topFrame, res);
            stmt.getLhs ()->accept (writer);
        }
        
        m_step.reset (new NextStep ());
        return;
    }
    else if (typeid (*content) == typeid (const ThreadFunctionValue))
    {
        const ThreadFunctionValue* th = dynamic_cast<const ThreadFunctionValue*> (content);

        std::auto_ptr<Value> res = th->getMethod () (m_func, m_memoryState, m_cstManager, m_globalsManager,
                m_topFrame, stmt.parameters (), th->getFunctionType ()->getReturnType (), m_program);

        if (stmt.getLhs ())
        {
            ValueWriter writer (m_func, m_memoryState, m_cstManager, m_globalsManager, m_topFrame, res);
            stmt.getLhs ()->accept (writer);
        }

        m_step.reset (new NextStep ());
        return;
    }
    
    const FunctionValue* funcValue = dynamic_cast<const FunctionValue*> (content);
    assert (funcValue);
                
    Ptr callersFrame = m_topFrame;
    Ptr topFrame = createFrame (m_memoryState, funcValue);
    
    const Function::Declarations& paramDecls = funcValue->getFunction ()->getParamDeclarations ();
    setParameters (m_memoryState, m_cstManager, m_globalsManager, m_func, funcValue,
                    stmt.parameters ().first, stmt.parameters ().second, callersFrame, 
                    paramDecls.begin (), paramDecls.end (), topFrame);
    
    m_step.reset (new CallStep (funcValue, topFrame));
}

void
GimpleStmtInterpreter::visit (const GotoStmt&)
{
	// connected by single edge to the label destination
	//const giterator::Operand* dest = stmt.getDest ();
	m_step.reset (new JumpStep ());
}

void
GimpleStmtInterpreter::visit (const LabelStmt&)
{
	// don't have to do anything, see GotoStmt
	m_step.reset (new NextStep ());
}

void
GimpleStmtInterpreter::visit (const NopStmt&)
{
	m_step.reset (new NextStep ());
}

void
GimpleStmtInterpreter::visit (const PhiStmt&)
{
	m_step.reset (new NextStep ());
}

void
GimpleStmtInterpreter::visit (const SwitchStmt& stmt)
{
	ValueReader valReader (m_func, m_memoryState, m_cstManager, m_globalsManager, m_topFrame);
	stmt.getIndex ()->accept (valReader);
	std::auto_ptr<Value> indexValue = valReader.acquireValue ();

	const giterator::Operand* usedLabel = NULL;

	for(SwitchStmt::LabelsIterator i1 = stmt.labels ().first; i1 != stmt.labels ().second; ++i1)
	{
		const giterator::LabelExprOperand* leo = dynamic_cast<const giterator::LabelExprOperand*> (&(*i1));
		assert(leo);

		if (leo->getIsDefault() && usedLabel == NULL)
		{
			usedLabel = &(*i1);
		}
		else
		{
			leo->getLower ()->accept(valReader);
			std::auto_ptr<Value> lowerValue = valReader.acquireValue ();

			bool cond;
			if (leo->getIsRange())
			{
				leo->getUpper ()->accept(valReader);
				std::auto_ptr<Value> upperValue = valReader.acquireValue ();

				cond = (*indexValue >= *lowerValue) && (*indexValue <= *upperValue);
			}
			else
				cond = *indexValue == *lowerValue;

			if (cond)
			{
				//we found the right case label
				usedLabel = &(*i1);
				break;
			}
		}
	}

	m_step.reset (new JumpStep (usedLabel));
}

Ptr
GimpleStmtInterpreter::createFrame (MemoryState* memoryState, const FunctionValue* func)
{
  /*
   * Create and init stack frame of called function
  */
  size_t frameSize = func->getVarsMapper ()->getSize ();

  if (frameSize != 0)
	  return memoryState->alloc (frameSize);
  else
	  return Ptr();
}

}
