// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file GimpleStmtInterpreter.h
 * Declaration of class that interprets GIMPLE++ statements.
 */
#ifndef INT_GIMPLE_STATEMENT_INTERPRETER_H
#define INT_GIMPLE_STATEMENT_INTERPRETER_H

#include "Steps.h"

#include "../mm/fwd.h"
#include "../mm/Ptr.h"
#include "../git/Stmt.h"
#include "../git/git_fwd.h"

#include <memory>
#include <boost/noncopyable.hpp>


namespace ginterpreter {
    
class FunctionValue;
class ConstantsManager;
class GlobalsManager;
class Program;

/**
 * Class that can interpret GIMPLE++ statements.
 *
 * @c GimpleStmtInterpreter is capable of interpret GIMPLE++
 * statements that are representations of GIMPLE statements.
 * For each GIMPLE++ statement, it implements a visitor
 * method that interprets this statement.
 * It is instantiated inside @c Thread::step() to interpret
 * the statement that the current thread's instruction pointer
 * points to.
 *
 * @see Thread::step()
 */
class GimpleStmtInterpreter : public giterator::Stmt::Visitor, private boost::noncopyable {
  private:    
    /*! GIMPLE++ representation of the program */
    giterator::Gimple* m_gimple;
    /*! Interface to the simulated memory */
    mmodule::MemoryState* m_memoryState;
    /*! Constants and function declarations of the program */
    ConstantsManager* m_cstManager;
    /*! Global declarations of the program */
    GlobalsManager* m_globalsManager;
    /*! The simulated program */
    Program* m_program;
    /*! pointer to the beginning of the current stack frame */
    mmodule::Ptr m_topFrame;
    /*! current function */
    const FunctionValue* m_func;
    /*! temporary holding the next step */
    std::auto_ptr<Step> m_step;

  public:
    /**
     * Constructor.
     *
     * @param gimple GIMPLE++ representation of the program
     * @param func current function
     * @param memoryState interface to the simulated memory
     * @param cstManager where the constants of the program are stored
     * @param globalsManager where the global declarations of the program are stored
     * @param program the program this thread is part of
     * @param topFrame pointer to the beginning of the current stack frame
     */
    GimpleStmtInterpreter (giterator::Gimple* gimple, const FunctionValue* func, 
                           mmodule::MemoryState* memoryState,
                           ConstantsManager* cstManager,
                           GlobalsManager* globalsManager,
                           Program* program,
                           mmodule::Ptr topFrame);

    /**
     * Acquires the next step.
     *
     * @returns the next step
     *
     * @see Step
     */
    std::auto_ptr<Step> acquireStep ();                           

    /**
     * Creates a new stack frame for function @c func.
     *
     * @param memoryState interface to the simulated memory
     * @param func fucntion for which to create a new stack frame
     */
    static mmodule::Ptr createFrame (mmodule::MemoryState* memoryState,
                                     const FunctionValue* func);
  protected:
    void visit (const giterator::AssignStmt& stmt);
    void visit (const giterator::CondStmt& stmt);
    void visit (const giterator::CallStmt& stmt);
    void visit (const giterator::ReturnStmt& stmt);
    void visit (const giterator::GotoStmt& stmt);
    void visit (const giterator::LabelStmt& stmt);
    void visit (const giterator::NopStmt& stmt);
    void visit (const giterator::PhiStmt& stmt);
    void visit (const giterator::SwitchStmt& stmt);
};

}

#endif /* INT_GIMPLE_STATEMENT_INTERPRETER_H */
