// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file GlobalsManager.h
 * Declaration of class that manages globals in the interpreted program.
 */
#ifndef INT_GLOBALS_MANAGER_H
#define INT_GLOBALS_MANAGER_H

#include "../git/git_fwd.h"
#include "../mm/fwd.h"
#include "../mm/Ptr.h"

#include <map>
#include <boost/noncopyable.hpp>

namespace ginterpreter {

/**
 * Class that manages globals in the interpreted program.
 * This class gets filled during creation of GIMPLE++ representation
 * of a program inside @c GimpleListener class passed
 * as a constructor to @c Gimple class that holds the GIMPLE++
 * representation.
 *
 * For each global declaration in a GIMPLE++ representation
 * of the program, @c addGlobalDecl() function get called
 * on @c GlobalsManager, which saves the global declaration
 * into simulated memory.
 * It can be later retrieved during the program execution simulation
 * by calling @c getGlobalDeclPtr().
 *
 * @see GimpleListener
 */
class GlobalsManager : private boost::noncopyable {
  private:
	/** type of the container holding offsets of the global
	 * declarations within the global memory area
	 */
    typedef std::map<const giterator::Decl*, size_t> DeclPtrs;
    /*! interface to the simulated memory */
    mmodule::MemoryState* m_memoryState;
    /*! pointer to the beginning of the gloal memory area */
    mmodule::Ptr m_globAreaPtr;
    /*! the next free address in the global memory area */
    size_t m_nextFreeOffset;
    /*! whether the area containing global variables was already created */
    bool m_areaCreated;
    /*! offsets of the global declarations within the global memory area */
    DeclPtrs m_globalDeclPtrs;
  public:
    /**
     * Constructor
     *
     * @param memoryState interface to the simulated memory
     */
    GlobalsManager (mmodule::MemoryState* memoryState);
    /**
	 * Returns a pointer to the global declaration within the simulated memory.
	 *
	 * @param decl GIMPLE++ representation of the declaration we are
	 * looking for
	 * @return pointer to the declaration @c declaration within simulated memory
	 */
    mmodule::Ptr getGlobalDeclPtr (const giterator::Decl* decl) const;
    /**
     * Get an offset of the global declaration within the global memory area.
     *
     * @param decl GIMPLE++ representation of the declaration we are
	 * looking for
	 * @return offset of the global declaration within the global memory area
     */
    size_t getGlobalDeclOffset (const giterator::Decl* decl) const;

    /**
	 * Stores a global declaration in the simulated memory.
	 *
	 * @param decl declaration to be stored
	 */
    void addGlobalDecl (const giterator::Decl* decl);

    /**
     * Returns pointer to the beginning of the memory area containing all
     * the global declarations.
     *
     * @return the described pointer
     */
    mmodule::Ptr getGlobAreaPtr () const;

    /**
	 * Allocate an area for global declarations.
	 */
    void allocArea ();
};

}

#endif /* INT_GLOBALS_MANAGER_H */
