// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file InstructionLogger.h
 * Declaration of a class for logging executed instructions.
 */

#ifndef INT_INSTRUCTION_LOGGER_H
#define INT_INSTRUCTION_LOGGER_H

#include "../git/Stmt.h"
#include "SymbolicPrinter.h"

#include <deque>
#include <iostream>

namespace ginterpreter
{

/**
 * Class storing pointers to already interpreted instructions
 * (GIMPLE++ statements).
 * Used for output purposes.
 * Instance of this class is passed to Program::step() and
 * Program::simpleStep() classes and the stepped instruction
 * is stored inside it.
 * Instance of this class is also passed to Rescheduler::reschedule()
 * during model checking, which may change it when backtracking
 * (remove the backtracked instructions).
 * Used for output purposes.
 */
class InstructionLogger
{
private:
	/*! Type of container storing instructions. */
	typedef std::deque<const giterator::Stmt*> Instructions;
	/*! Type of container storing names of threads of stored instructions. */
	typedef std::deque<std::string> ThreadNames;
	typedef Instructions::const_iterator InstructionsIterator;

	/*! Stored instructions */
	Instructions m_instructions;
	/*! Names of threads of stored instructions. */
	ThreadNames m_threadNames;
public:
	/**
	 * Constructor
	 */
	InstructionLogger() {}

	/**
	 * Add an instruction to instructions maintained by this class.
	 *
	 * @param name name of the thread in which is the instruction
	 * executed
	 * @param stmt the stored instruction
	 */
	void push(std::string name, const giterator::Stmt* stmt)
	{
		m_instructions.push_back(stmt);
		m_threadNames.push_back(name);
	}

	/**
	 * Remove a last stored instruction from instructions managed
	 * by this class.
	 * There have to be some stored instructions.
	 */
	void pop()
	{
		m_instructions.pop_back();
		m_threadNames.pop_back();
	}

	/**
	 * Change a size of the container storing the instructions.
	 * There have to be some stored instructions.
	 * Use for decreasing the number of instructions,
	 * resulting in popping some las stored instructions.
	 *
	 * @param size the new size of the container storing the
	 * instructions
	 */
	void resize(int size)
	{
		assert(size >= 0);
		m_instructions.resize(size);
		m_threadNames.resize(size);
	}

	/**
	 * Print information about the stored instructions to a
	 * specified stream.
	 *
	 * @param stream the stream where print the information
	 */
	void logToStream(std::ostream& stream)
	{
		stream << "Logger: Instructions executed:" << std::endl;
		SymbolicPrinter sp(stream);
		//for(InstructionsIterator i1 = m_instructions.begin(); i1 != m_instructions.end(); ++i1)
		for (size_t i = 0; i < m_instructions.size(); ++i)
		{
			stream << m_threadNames[i] << ": ";
			//(*i1)->accept(sp);
			m_instructions[i]->accept(sp);
			stream << std::endl;
		}
	}
};

}

#endif /* INT_INSTRUCTION_LOGGER_H */
