// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file InstructionPtr.cpp
 * Definitions for class representing instruction pointer.
 */
#include "InstructionPtr.h"

#include "FunctionValue.h"
#include "../git/Function.h"

//OND:
#include "../git/LabelDecl.h"
#include "../git/LabelStmt.h"
#include "../git/DeclOperand.h"
#include "../git/LabelExprOperand.h"
#include "ValueReader.h"

using namespace giterator;


namespace ginterpreter {

/**
 * Creates an IP pointing to the first statement of function @c func.
 */
InstructionPtr::InstructionPtr (const FunctionValue* func)
    : m_func (func)
{
    m_block = m_func->getFunction ()->getFirstBlock ();
    m_stmtIter = m_block->stmts ().first;
    while ((m_stmtIter = m_block->stmts ().first) == m_block->stmts ().second)
	{
		m_block = m_block->getFallthruEdge ()->getDest ();
	}
}


/**
 * Creates a null IP.
 */
InstructionPtr::InstructionPtr () : m_func (0), m_block (0)
{
}
    
    
/**
 * Returns the function to which this IP points.
 */
const FunctionValue* 
InstructionPtr::getFunction () const 
{
    assert (m_func);
    return m_func;
}


/**
 * Returns the block to which this IP points.
 */
const Block* 
InstructionPtr::getBlock () const
{
    assert (m_func);
    return m_block;
}


/**
 * Returns the statement to which this IP points.
 */
const Stmt*
InstructionPtr::getStmt () const
{
    assert (m_func);
    return &*m_stmtIter;
}


/**
 * Moves this IP to the next statement.
 * If the statement this IP points to is the last statement in the block, 
 * there must be a fallthru edge starting in the block.
 */
void 
InstructionPtr::step ()
{
    assert (m_func);
    m_stmtIter++;
    while (m_stmtIter == m_block->stmts ().second)
    {
        m_block = m_block->getFallthruEdge ()->getDest ();
        m_stmtIter = m_block->stmts ().first;
    }
}

void InstructionPtr::jump ()
{
	assert (m_func);
	m_stmtIter++;
	assert (m_stmtIter == m_block->stmts ().second);

	while (m_stmtIter == m_block->stmts ().second)
	{
		m_block = m_block->getFallthruEdge ()->getDest ();
		m_stmtIter = m_block->stmts ().first;
	}
}

void InstructionPtr::jumpToLabel (const giterator::Operand* label)
{
	assert (m_func);
	m_stmtIter++;
	assert (m_stmtIter == m_block->stmts ().second);

    for (giterator::Block::SuccEdgesIterator i1 = m_block->succEdges ().first; i1 != m_block->succEdges ().second; ++i1)
    {
		giterator::Block::StmtIterator stmtIter = i1->getDest ()->stmts ().first;
		const giterator::LabelStmt* ls = dynamic_cast<const giterator::LabelStmt*> (&(*stmtIter));

		if (!ls)
			continue;

		const giterator::DeclOperand* labeld = dynamic_cast<const giterator::DeclOperand*> (ls->getLabel ());
		assert(labeld);
		const giterator::LabelExprOperand* labeld2 = dynamic_cast<const giterator::LabelExprOperand*> (label);
		assert(labeld2);

		if (labeld->getDecl ()->getName ().compare(labeld2->getDecl ()->getName ()) == 0)
		{
			m_stmtIter = stmtIter;
			m_block = i1->getDest ();
			break;
		}
    }
}

/**
 * Moves this IP pointer to the first statement of the block 
 * to which leads the true or false branch edge.
 * 
 * @param cond condition that specifies whether to take the true or false branch
 */
void 
InstructionPtr::takeBranch (bool cond)
{
    assert (m_func);
 
    if (cond)
        m_block = m_block->getTrueBranchEdge ()->getDest ();
    else
        m_block = m_block->getFalseBranchEdge ()->getDest ();
    
    while ((m_stmtIter = m_block->stmts ().first) == m_block->stmts ().second)
    {
        m_block = m_block->getFallthruEdge ()->getDest ();
    }
}

bool InstructionPtr::isNull () const
{ return m_func == 0; }
    
}
