// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file InstructionSafetyResolver.cpp
 * Definitions for class to distinguish safe and unsafe instructions
 */

#include "InstructionSafetyResolver.h"
#include "GlobalsResolver.h"

#include "../git/Stmts.h"
#include "../git/Operation.h"
#include "../git/AddrOperand.h"
#include "../git/DeclOperand.h"
#include "../git/FunctionDecl.h"
#include "../git/FunctionInfo.h"

using namespace giterator;

namespace ginterpreter
{

InstructionSafetyResolver::InstructionSafetyResolver() : m_unsafe(false)
{}

bool InstructionSafetyResolver::getUnsafe() const
{
	return m_unsafe;
}

void InstructionSafetyResolver::visit (const AssignStmt& stmt)
{
	resolveOperand(stmt.getLhs());

	if (!m_unsafe)
		resolveOperation(stmt.getOperation());
}

void InstructionSafetyResolver::visit (const CondStmt& stmt)
{
	resolveOperation(stmt.getOperation());
}

void InstructionSafetyResolver::visit (const CallStmt& stmt)
{
	// Call statement is considered unsafe only if result of a call is stored
	// to an unsafe place, if we are calling function with
	// unsafe parameter, or if it is a call to built-in sleep() function

	resolveOperand(stmt.getLhs());

	//TODO: do this only for not user defined function
	for (CallStmt::ParameterIterator i1 = stmt.parameters().first; i1 != stmt.parameters().second; ++i1)
	{
		resolveOperand(&*i1);
		if (m_unsafe)
			break;
	}

	const AddrOperand* addrFuncDeclOp = dynamic_cast<const AddrOperand*> (stmt.getFunction());
	assert(addrFuncDeclOp);
	const DeclOperand* funcDeclOp = dynamic_cast<const DeclOperand*> (addrFuncDeclOp->getSubOperand());
	assert(funcDeclOp);
	const FunctionDecl* funcDecl = dynamic_cast<const FunctionDecl*> (funcDeclOp->getDecl());
	assert(funcDecl);

	if (funcDecl->getName().compare(BUILT_IN_SLEEP_NAME) == 0)
	{
		m_unsafe = true;
	}
}

void InstructionSafetyResolver::visit (const ReturnStmt& stmt)
{
	resolveOperand(stmt.getReturnValue());
}

void InstructionSafetyResolver::visit (const GotoStmt& stmt)
{
	m_unsafe = false;
}

void InstructionSafetyResolver::visit (const LabelStmt& stmt)
{
	m_unsafe = false;
}

void InstructionSafetyResolver::visit (const NopStmt& stmt)
{
	m_unsafe = false;
}

void InstructionSafetyResolver::visit (const PhiStmt& stmt)
{
	m_unsafe = false;
}

void InstructionSafetyResolver::visit (const SwitchStmt& stmt)
{
	resolveOperand(stmt.getIndex());

	if (!m_unsafe)
	{
		std::pair<SwitchStmt::LabelsIterator, SwitchStmt::LabelsIterator> labels;

		for (SwitchStmt::LabelsIterator i1 = labels.first; i1 != labels.second; ++i1)
		{
			resolveOperand(&*i1);
			if (m_unsafe)
				break;
		}
	}
}

void InstructionSafetyResolver::resolveOperand(const Operand* op)
{
	if (op == NULL)
	{
		m_unsafe = false;
		return;
	}

	GlobalsResolver globRes;
	op->accept(globRes);
	if (globRes.isValid() && globRes.isGlobal())
		m_unsafe = true;
}

void InstructionSafetyResolver::resolveOperation(const Operation* op)
{
	if (op == NULL)
	{
		m_unsafe = false;
		return;
	}

	if (op->cardinality() == 1)
	{
		const UnaryOperation* uop = dynamic_cast<const UnaryOperation*> (op);
		assert(uop);

		resolveOperand(uop->getOperand());
	}
	else if (op->cardinality() == 2)
	{
		const BinaryOperation* bop = dynamic_cast<const BinaryOperation*> (op);
		assert(bop);

		resolveOperand(bop->getOperand1());

		if (!m_unsafe)
			resolveOperand(bop->getOperand2());
	}
}

}
