// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file InstructionSafetyResolver.h
 * Declaration of a class to distinguish safe and unsafe instructions
 */

#ifndef INT_INSTRUCTION_SAFETY_RESOLVER_H
#define INT_INSTRUCTION_SAFETY_RESOLVER_H

#include "../git/Stmt.h"

namespace giterator
{
class Operation;
class Operand;
}

namespace ginterpreter
{

/**
 * Class for clasifying instructions (GIMPLE++ statements)
 * to safe or unsafe.
 *
 * This class marks marks an instruction as safe, when its
 * result cannot depend on the different scheduling policies
 * of threads. That means all instructions that manipulate
 * with local variables are marked as safe. On the other hand,
 * unsafe instructions can be affected by different scheduling
 * policies. These are the instructions that manipulate global
 * variables.
 */
class InstructionSafetyResolver : public giterator::Stmt::Visitor
{
private:
	/*! Whether the classified instruction is safe or unsafe */
	bool m_unsafe;
public:
	/**
	 * Constructior
	 */
	InstructionSafetyResolver();

	/**
	 * Find out whether the classified instruction is safe or unsafe.
	 *
	 * @return true if the instruction is unsafe, false otherwise.
	 */
	bool getUnsafe() const;

    void visit (const giterator::AssignStmt& stmt);
    void visit (const giterator::CondStmt& stmt);
    void visit (const giterator::CallStmt& stmt);
    void visit (const giterator::ReturnStmt& stmt);
    void visit (const giterator::GotoStmt& stmt);
    void visit (const giterator::LabelStmt& stmt);
    void visit (const giterator::NopStmt& stmt);
    void visit (const giterator::PhiStmt& stmt);
    void visit (const giterator::SwitchStmt& stmt);
private:
    /**
     * Find out whether an operand is safe or unsafe,
     * i.e. whether it is global or local variable.
     * Store the result in the m_unsafe instance variable
     * of this class.
     *
     * @param op operand to resolve
     */
    void resolveOperand(const giterator::Operand* op);

    /**
	 * Find out whether an operand is safe or unsafe,
	 * i.e. whether either of the operands is unsafe.
	 * Store the result in the m_unsafe instance variable
	 * of this class.
	 *
	 * @param op opertion to resolve
	 * @see resolveOperand()
	 */
    void resolveOperation(const giterator::Operation* op);
};
}


#endif /* INT_INSTRUCTION_SAFETY_RESOLVER_H */
