// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file IntegerValue.cpp
 * Declaration of Class representing simulated integer value.
 */
#include "IntegerValue.h"
#include "BoolValue.h"
#include "RealValue.h"

#include <typeinfo>
#include <ostream>
#include <boost/functional/hash.hpp>
#include <cmath>


namespace ginterpreter
{

IntegerValue::IntegerValue (const giterator::Type* type, uint64_t value)
	: AbstractNonPtrValue (type), m_value (value)
{}

Value* IntegerValue::clone () const
{
  return new IntegerValue (*this);
}

mmodule::Hash IntegerValue::getHash () const
{
	return boost::hash<uint64_t>() (m_value);
}


std::auto_ptr<Value> IntegerValue::addOp (const Value& v) const
{
  const IntegerValue& iv = checkedCast (v);
  return std::auto_ptr<Value> (new IntegerValue (getType(), m_value + iv.m_value));
}

std::auto_ptr<Value> IntegerValue::subOp (const Value& v) const
{
  const IntegerValue& iv = checkedCast (v);
  return std::auto_ptr<Value> (new IntegerValue (getType (), m_value - iv.m_value));
}

std::auto_ptr<Value> IntegerValue::multOp (const Value& v) const
{
  const IntegerValue& iv = checkedCast (v);
  return std::auto_ptr<Value> (new IntegerValue (getType (), m_value * iv.m_value));
}

std::auto_ptr<Value> IntegerValue::truncDivOp (const Value& v) const
{
  const IntegerValue& iv = checkedCast (v);
  return std::auto_ptr<Value> (new IntegerValue (getType (), m_value / iv.m_value));
}

std::auto_ptr<Value> IntegerValue::ceilDivOp (const Value& v) const
{
  const IntegerValue& iv = checkedCast (v);
  long double d1 = (long double) getValue();
  long double d2 = (long double) iv.getValue();
  return std::auto_ptr<Value> (new IntegerValue (getType (), ceill(d1 / d2)));
}

std::auto_ptr<Value> IntegerValue::floorDivOp (const Value& v) const
{
  const IntegerValue& iv = checkedCast (v);
  long double d1 = (long double) getValue();
  long double d2 = (long double) iv.getValue();
  return std::auto_ptr<Value> (new IntegerValue (getType (), floorl(d1 / d2)));
}

std::auto_ptr<Value> IntegerValue::roundDivOp (const Value& v) const
{
  const IntegerValue& iv = checkedCast (v);
  long double d1 = (long double) getValue();
  long double d2 = (long double) iv.getValue();
  return std::auto_ptr<Value> (new IntegerValue (getType (), roundl(d1 / d2)));
}

std::auto_ptr<Value> IntegerValue::truncModOp (const Value& v) const
{
  const IntegerValue& iv = checkedCast (v);
  return std::auto_ptr<Value> (new IntegerValue (getType (), m_value % iv.m_value));
}

std::auto_ptr<Value> IntegerValue::ceilModOp (const Value& v) const
{

	// Modulus is defined as a - (a/b)*b
	return this->subOp(*((this->ceilDivOp(v))->multOp(v)));
}

std::auto_ptr<Value> IntegerValue::floorModOp (const Value& v) const
{
	// Modulus is defined as a - (a/b)*b
	return this->subOp(*((this->floorDivOp(v))->multOp(v)));
}

std::auto_ptr<Value> IntegerValue::roundModOp (const Value& v) const
{
	// Modulus is defined as a - (a/b)*b
	return this->subOp(*((this->roundDivOp(v))->multOp(v)));
}

std::auto_ptr<Value> IntegerValue::lShiftOp (const Value& v) const
{
  const IntegerValue& iv = checkedCast (v);

  uint64_t newValue;

  if (getIntegerType()->isUnsigned())
    newValue = m_value << iv.m_value;
  else
    newValue = ((int64_t) m_value) << iv.m_value;

  return std::auto_ptr<Value> (new IntegerValue (getType (), newValue));
}

std::auto_ptr<Value> IntegerValue::rShiftOp (const Value& v) const
{
  const IntegerValue& iv = checkedCast (v);

  uint64_t newValue;

  if (getIntegerType()->isUnsigned())
	  newValue = m_value >> iv.m_value;
  else
	  newValue = ((int64_t) m_value) >> iv.m_value;

  return std::auto_ptr<Value> (new IntegerValue (getType (), newValue));
}

std::auto_ptr<Value> IntegerValue::lRotateOp (const Value& v) const
{
	const IntegerValue& iv = checkedCast (v);

	uint64_t rotateBy = iv.m_value;
	if (rotateBy < 0)
	{
		//WARNING ("Invalid left rotation, ignored");
		return std::auto_ptr<Value> (this->clone());
	}

	uint64_t returnValue = m_value;
	for (uint64_t i = 0; i < rotateBy; ++i)
	{
		returnValue = lRotateByOne(returnValue);
	}

	return std::auto_ptr<Value> (new IntegerValue (getType (), returnValue));
}

std::auto_ptr<Value> IntegerValue::rRotateOp (const Value& v) const
{
  const IntegerValue& iv = checkedCast (v);

  uint64_t rotateBy = iv.m_value;
  if (rotateBy < 0)
  {
  	  //WARNING ("Invalid right rotation, ignored");
  	  return std::auto_ptr<Value> (this->clone());
  }

  uint64_t returnValue = m_value;
  for (uint64_t i = 0; i < rotateBy; ++i)
  {
	  returnValue = rRotateByOne(returnValue, this->getUnitSize());
  }

  return std::auto_ptr<Value> (new IntegerValue (getType (), returnValue));
}

std::auto_ptr<Value> IntegerValue::bitNotOp () const
{
  return std::auto_ptr<Value> (new IntegerValue (getType (), ~m_value));
}

std::auto_ptr<Value> IntegerValue::bitIorOp (const Value& v) const
{
  const IntegerValue& iv = checkedCast (v);
  return std::auto_ptr<Value> (new IntegerValue (getType (), m_value | iv.m_value));
}

std::auto_ptr<Value> IntegerValue::bitXorOp (const Value& v) const
{
  const IntegerValue& iv = checkedCast (v);
  return std::auto_ptr<Value> (new IntegerValue (getType (), m_value ^ iv.m_value));
}

std::auto_ptr<Value> IntegerValue::bitAndOp (const Value& v) const
{
  const IntegerValue& iv = checkedCast (v);
  return std::auto_ptr<Value> (new IntegerValue (getType (), m_value & iv.m_value));
}

std::auto_ptr<Value> IntegerValue::truthNotOp () const
{
  return std::auto_ptr<Value> (new IntegerValue (getType (), !m_value));
}

std::auto_ptr<Value> IntegerValue::truthOrOp (const Value& v) const
{
  const IntegerValue& iv = checkedCast (v);
  return std::auto_ptr<Value> (new IntegerValue (getType (), m_value || iv.m_value));
}

std::auto_ptr<Value> IntegerValue::truthXorOp (const Value& v) const
{
  const IntegerValue& iv = checkedCast (v);
  return std::auto_ptr<Value> (new IntegerValue (getType (), (m_value || iv.m_value) && (!(m_value && iv.m_value))));
}

std::auto_ptr<Value> IntegerValue::truthAndOp (const Value& v) const
{
  const IntegerValue& iv = checkedCast (v);
  return std::auto_ptr<Value> (new IntegerValue (getType (), m_value && iv.m_value));
}

std::auto_ptr<Value> IntegerValue::negateOp () const
{
	return std::auto_ptr<Value> (new IntegerValue(getType(), !m_value));
}

std::auto_ptr<Value> IntegerValue::absOp () const
{
	if (m_value < 0)
		return std::auto_ptr<Value> (new IntegerValue (getType (), -m_value));
	else
		return std::auto_ptr<Value> (new IntegerValue (getType (), m_value ));
}

bool IntegerValue::operator== (const Value& v) const
{
  checkType (v);

  if (typeid(v) == typeid(BoolValue))
  {
	  const BoolValue& bv = dynamic_cast<const BoolValue&> (v);
	  return m_value == bv.getValue();
  }
  else if (typeid(v) == typeid(IntegerValue))
  {
	  const IntegerValue& iv = dynamic_cast<const IntegerValue&> (v);
	  return m_value == iv.getValue();
  }
  else if (typeid(v) == typeid(RealValue))
  {
	  const RealValue& rv = dynamic_cast<const RealValue&> (v);
	  return m_value == rv.getValue();
  }
  else
	  return false;
}

bool IntegerValue::operator< (const Value& v) const
{
  checkType (v);
  const IntegerValue& iv = dynamic_cast<const IntegerValue&> (v);
  return m_value < iv.m_value;
}

void IntegerValue::print (std::ostream& os) const
{
	if (getIntegerType ()->isUnsigned ())
		os << m_value;
	else
		os << (int64_t)m_value;
}

uint64_t IntegerValue::getValue () const
{
  return m_value;
}

const giterator::IntegerType* IntegerValue::getIntegerType () const
{
	assert (typeid (*getType ()) == typeid (giterator::IntegerType));
	return static_cast<const giterator::IntegerType*> (getType ());
}

bool IntegerValue::isZero () const
{ return m_value == 0; }

void IntegerValue::checkType (const Value& i) const
{
	assert (getType()->isSame(*i.getType()));     (void)i;
}

const IntegerValue& IntegerValue::checkedCast (const Value& v) const
{
	checkType (v);
	return static_cast<const IntegerValue&> (v);
}

uint64_t IntegerValue::lRotateByOne(uint64_t value) const
{
	int highbit;
	if(value & (0x100000000LL)) // the high bit only
		highbit = 1;
	else
		highbit = 0;
	// Left shift (bottom bit becomes 0):

	if (getIntegerType()->isUnsigned())
		value <<= 1;
	else
		value = ((int64_t) value) << 1;
	// Rotate the high bit onto the bottom:
	value |= highbit;
	return value;
}

uint64_t IntegerValue::rRotateByOne(uint64_t value, size_t valueUnitSize) const
{
	uint64_t lowbit;
	if(value & 1) // Check the low bit
		lowbit = 1;
	else
		lowbit = 0;

	if (getIntegerType()->isUnsigned())
		value >>= 1; // Right shift by one position
	else
		value = ((int64_t) value) >> 1;
	// Rotate the low bit onto the top:
	value |= (lowbit << valueUnitSize);
	return value;
}

}
