// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file IntegerValue.h
 * Declaration of Class representing simulated integer value.
 */
#ifndef INT_INTEGER_VALUE_H
#define INT_INTEGER_VALUE_H

#include "AbstractValue.h"

#include "../git/IntegerType.h"

namespace ginterpreter {

/**
 * Class representing simulated integer value.
 * The value is stored as an uint64_t,
 * it is therefore capable of storing all the
 * integer types.
 */
class IntegerValue : public AbstractNonPtrValue {
protected:
    /*! the value of this simulated value */
    uint64_t m_value;
  public:
    /**
     * Creates new integer simulated value.
     * 
     * @param type type of the simulated value
     * @param value value of the simulated value
     */
    IntegerValue (const giterator::Type* type, uint64_t value);
    
    /**
	 * Clone this value.
	 *
	 * @returns an exact copy of this value.
	 */
    Value* clone () const;
    
    /**
	 * Get a hash of this value.
	 * For memory state equivalency check purposes.
	 *
	 * @return hash of this value.
	 */
    mmodule::Hash getHash () const;


    /**
	 * @name Arithmetic and logical operators
	 * The arithmetic and logical operators on the value.
	 */
	//@{
    std::auto_ptr<Value> addOp (const Value& v) const;
    std::auto_ptr<Value> subOp (const Value& v) const;
    std::auto_ptr<Value> multOp (const Value& v) const;

    std::auto_ptr<Value> truncModOp (const Value& v) const;
	std::auto_ptr<Value> floorModOp (const Value& v) const;
	std::auto_ptr<Value> ceilModOp (const Value& v) const;
	std::auto_ptr<Value> roundModOp (const Value& v) const;
	std::auto_ptr<Value> truncDivOp (const Value& v) const;
	std::auto_ptr<Value> floorDivOp (const Value& v) const;
	std::auto_ptr<Value> ceilDivOp (const Value& v) const;
	std::auto_ptr<Value> roundDivOp (const Value& v) const;

    std::auto_ptr<Value> lShiftOp (const Value& v) const;
	std::auto_ptr<Value> rShiftOp (const Value& v) const;
	std::auto_ptr<Value> lRotateOp (const Value& v) const;
	std::auto_ptr<Value> rRotateOp (const Value& v) const;

	std::auto_ptr<Value> bitNotOp () const;
	std::auto_ptr<Value> bitIorOp (const Value& v) const;
	std::auto_ptr<Value> bitXorOp (const Value& v) const;
	std::auto_ptr<Value> bitAndOp (const Value& v) const;

	std::auto_ptr<Value> truthNotOp () const;
	std::auto_ptr<Value> truthOrOp (const Value& v) const;
	std::auto_ptr<Value> truthXorOp (const Value& v) const;
	std::auto_ptr<Value> truthAndOp (const Value& v) const;

	std::auto_ptr<Value> negateOp () const;
    std::auto_ptr<Value> absOp () const;
    //@}

    /** @name Relation operators
	 * The relation operators on the value.
	 */
	//@{
    bool operator== (const Value& v) const;
    bool operator< (const Value& v) const;
    //@}

    /**
	 * Print the value to a stream.
	 *
	 * @param os stream to print the value to
	 */
    void print (std::ostream& os) const;

    /**
     * Returns the raw value of this simulated value.
     *
     * @return the raw value of this simulated value
     */
    uint64_t getValue () const;
    
    /**
     * Returns the integer type of this simulated value.
     *
     * @return the integer type of this simulated value
     */
    const giterator::IntegerType* getIntegerType () const;
    
    /**
     * Checks if this simulated value is zero.
     *
     * @return thrue if this value is zero, false otherwise
     */
    bool isZero () const;
  protected:
	/**
	* Checks if the type of value @c i and the type of this value are the same.
	*
	* @param i the value to check if the type is the same
	*
	* @see giterator::IntegerType::isSame()
	*/
	void checkType (const Value& i) const;

	/**
	* Casts the reference to the value @c v to the reference to the integer value
	* and checks, if the value @c v is of the same type as this value.
	*
	* @param v the value to cast
	* @return the casted reference
	*/
	const IntegerValue& checkedCast (const Value& v) const;

	/**
	 * Helper function to implement left rotation operator.
	 * This function rotates the value left by one bit.
	 *
	 * @param value the value to rotate left
	 * @return value @c value rotated left by one bit
	 *
	 * @see lRotateOp()
	 */
	uint64_t lRotateByOne(uint64_t value) const;

	/**
	 * Helper function to implement right rotation operator.
	 * This function rotates the value right by one bit.
	 *
	 * @param value the value to rotate right
	 * @param valueUnitSize size of the value to properly
	 * set the rotated bit
	 * @return value @c value rotated right by one bit
	 *
	 * @see rRotateOp()
	 */
	uint64_t rRotateByOne(uint64_t value, size_t valueUnitSize) const;
};

}

#endif /* INT_INTEGER_VALUE_H */

