// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

 /** @file Interpret.cpp
 * Definitions of functions for GIMPLE interpretation callable from C.
 */
#include "Interpret.h"

#include "Program.h"
#include "Value.h"
#include "ConstantsManager.h"
#include "GlobalsManager.h"
#include "FunctionValue.h"
#include "BuiltinFunctionValue.h"
#include "GimpleListener.h"
#include "BuiltinFunctions.h"
#include "InterpretExceptions.h"
#include "ExceptionMessages.h"
#include "InstructionLogger.h"
#include "GimpleDumper.h"

#include "../git/Serializer.h"
#include "../mm/MemoryState.h"
#include "../mm/MemoryAccessListener.h"
#include "../mm/MemoryExceptions.h"

#include <iostream>
#include <set>
#include <map>

/**
 * This namespace contains the implementation of the C++ part of GIMPLE Interpreter.
 */
namespace ginterpreter {
    
class ProgramInfo;

using namespace mmodule;

#define DUMPFILE "GIMPLE++.dump"

/**
 * Class that reports unreachable areas.
 * This is passed as the visitor to the @c mmodule::MemoryState class.
 */
class InterpreterMMListener : public MemoryAccessListener {
  public:
    virtual ~InterpreterMMListener () {}
    
    void unreachableArea (unsigned int)
    {
        std::cout << UNREACHABLE_AREA_MSG << std::endl;
    }
};


/**
 * All the objects needed by the interpreter.
 */
struct InterpreterData {
    /*! visitor for @c mmodule::MemoryState class */
    InterpreterMMListener listener;
    /*! Interface to the simulated memory */
    MemoryState memoryState;
    /*! Constants and function declarations of the simulated program */
    ConstantsManager cstManager;
    /*! Global declarations of the simulated program */
    GlobalsManager globalsManager;
    /*! listener for @c giterator::Gimple class */
    GimpleListener gimpleListener;
    /*! GIMPLE++ representation of the program */
    giterator::Gimple gimple;
    /*! flags passed to the interpreter */
    struct interpreter_flags interpreterFlags;
    /*! set of hashes of the explored states */
    std::set<Hash> visitedStates;
    /*! Storage for interpreted instructions */
    InstructionLogger iLogger;

    /**
     * Constructor.
     */
    InterpreterData ()
        : listener (),
          memoryState (&listener),
          cstManager (&memoryState),
          globalsManager (&memoryState),
          gimpleListener (&cstManager, &globalsManager),
          gimple (&gimpleListener)
    {}
};



/**
 * Function that returns the reference to the structure
 * containing the interpreter data.
 */
static InterpreterData&
intData ()
{
    static std::auto_ptr<InterpreterData> interpreterData_;
    
    if (!interpreterData_.get ())
        interpreterData_.reset (new InterpreterData ());
    
    return *interpreterData_.get ();
}


/**
 * Asks the user about how many steps to backtrack.
 * 
 * @return number of steps to backtrack
 */
static int
askBacktrace ()
{
    int btCount = 0;
    MemoryState& memoryState = intData ().memoryState;
    
    if (memoryState.savedStateCount () > 0)
    {
        std::cerr << "<";
        do {
            std::cerr << "bt: ";
            std::cin >> btCount;
        } while (btCount < 0 || memoryState.savedStateCount () < (size_t)btCount);
        
        std::cerr << ">" << std::endl;
    }
    
    return btCount;
}


/**
 * Reports an exception to the standard error output.
 *
 * @param description message to report
 */
static void
reportException (const std::string& description)
{
    std::cout << OUTPUT_SEPARATOR << "Exception encountered:" << std::endl
    		  << description << std::endl;
}


/**
 * Checks if program is at a state that has been visited before.
 * If it is, prints a message on the standard error output.
 * 
 * @return @c true if program already visited the current state, @c false otherwise
 */
static bool
checkReentry (Hash hash, std::set<Hash>& visitedStates)
{
    bool added = visitedStates.insert (hash).second;
    
    if (!added)
    {
        // state reentered
        std::cout << "<state re-entered>" << std::endl;
        return true;
    }
    
    return false;
}

/**
 * Print what was the return value of the interpreted program.
 *
 * @param retVal the return value or NULL if there wasn't any
 */
void returnFromProgram(const Value* retVal)
{
	std::cerr << "Program returned: ";
	if (retVal != NULL)
		std::cerr << *retVal << std::endl;
	else
		std::cerr << "no return value." << std::endl;
}

/**
 * Starts the interpreter.
 */
int
doInterpret (struct interpreter_flags flags, const char* gimplexx_file)
{
    giterator::Gimple& gimple = intData ().gimple;
    MemoryState& memoryState = intData ().memoryState;
    ConstantsManager& cstManager = intData ().cstManager;
    GlobalsManager& globalsManager = intData ().globalsManager;
    std::set<Hash>& visitedStates = intData ().visitedStates;
    InstructionLogger& iLogger = intData ().iLogger;
    flags.state_reentry_check |= flags.state_reentry_check_fine;

	if (giterator::deserialize(gimple, gimplexx_file) == -1)
	{
		return -1;
	}
	if (!gimple.entryPoint ())
	{
		std::cerr << "interpreter: program entry point not found" << std::endl;
		return -1;
	}

	LOG_DEBUG ("Interpreting");
        
    Ref<ProgramInfo> programRef = Program::init (&gimple, &memoryState, &cstManager, &globalsManager);
  
    std::auto_ptr<Program> program (new Program (&gimple, &memoryState, 
                                    &cstManager, &globalsManager, programRef, MODE_INTERPRETER));

    size_t stepCount = 0;
    bool reenteredState = false;
    bool bException;
    bool exit = false;
    while (!program->isFinished ())
    {
        if ((flags.state_reentry_check && program->aboutToReadInput ()) ||
            (flags.state_reentry_check_fine && program->atCondStmt ()))
        {
            LOG_DEBUG ("checking state re-entry");
            
            program->flush ();
            memoryState.push ();
            Hash hash = memoryState.hash ();
            memoryState.pop ();
            
            reenteredState = checkReentry (hash, visitedStates);
            if (reenteredState)
                break;
        }
        
        if (flags.backtrack && program->aboutToReadInput ())
        {
            int btCount = askBacktrace ();
            if (0 < btCount)
            {
                for (int i = 0; i < btCount-1; i++)
                    memoryState.pop ();
        
                program.reset ();
                memoryState.backtrack ();
                program.reset (new Program (&gimple, &memoryState, &cstManager, &globalsManager, programRef, MODE_INTERPRETER));
            }
            else
            {
                program->flush ();
                memoryState.push ();
            }
        }
            
        bException = true;
        try {
            program->step (&iLogger);
            bException = false;
        } catch (FreedAreaAccessException&) {
            reportException (FREED_AREA_ACCESS_EXC_MSG);
        } catch (NullPointerException&) {
            reportException (NULL_POINTER_EXC_MSG);
        } catch (NotAreaBeginException&) {
            reportException (NOT_AREA_BEGIN_EXC_MSG);
        } catch (UndefinedMemoryLoadException&) {
            reportException (UNDEFINED_MEMORY_LOAD_EXC_MSG);
        } catch (OutOfBoundsException&) {
            reportException (OUT_OF_BOUNDS_EXC_MSG);
        } catch (UndefinedPtrOperationException&) {
            reportException (UNDEFINED_PTR_OPERATION_EXC_MSG);
        } catch (AbortException&) {
        	reportException(ABORT_EXC_MSG);
        } catch (AssertException& e) {
        	reportException(ASSERT_EXC_MSG + e.getMessage());
        } catch (InvalidPtrConvertException& e) {
        	reportException(INVALID_PTR_EXC_MSG);
        } catch (OverlappingAreasException& e) {
			reportException(OVERLAPPING_AREAS_EXC_MSG);
		} catch (IDeadlockException& e) {
        	reportException(DEADLOCK_EXC_MSG);
        } catch (ExitException& e) {
         	// This is not an error or abnormal behavior.
           	returnFromProgram(e.m_status);
           	exit = true;
        }
        
        if (bException)
            break;
        
        stepCount++;
    }
    
    if (!exit)
    {
		if (reenteredState)
			std::cerr << "state re-entry detected" << std::endl;

		else if (bException)
		{
			std::cout << OUTPUT_SEPARATOR;
			iLogger.logToStream(std::cout);
			std::cerr << "exception caught" << std::endl;
		}
		else
		{
			returnFromProgram(program->getReturnValue ());
		}
    }
    std::cerr << "steps:        " << stepCount << std::endl;
    std::cerr << "saved states: " << memoryState.savedStateCount () << std::endl;
    std::cerr << "undos:        " << memoryState.undoCount () << std::endl;
    
    GimpleDumper gd(&cstManager, &memoryState);
    gd.dump(DUMPFILE);

    return 0;
}

}
