// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file InterpretExceptions.h
 * Exceptions that can be thrown by interpret.
 */
#ifndef INT_INTERPRET_EXCEPTIONS_H
#define INT_INTERPRET_EXCEPTIONS_H

#include <exception>

namespace ginterpreter {

/**
 * Common base class of all exceptions that can be thrown when
 * working with the interpret.
 */
struct InterpretException : public std::exception {
};

/**
 * Exception that is thrown when builtin function abort is interpreted.
 */
struct AbortException : public InterpretException {
    /**
     * Constructor.
     */
	AbortException() {}

    const char* what () const throw ()
    {
        return "Builtin function abort called";
    }
};

/**
 * Exception that is thrown when builtin function exit is interpreted.
 */
struct ExitException : public InterpretException {
    /**
     * Constructor.
     *
     * @param status value of exit parameter
     */
	ExitException(const Value* status) : m_status(status) {}

    const char* what () const throw ()
    {
        return "Builtin function exit called";
    }

    const Value* m_status;
};

/**
 * Exception that is thrown when assertion failed.
 */
struct AssertException : public InterpretException {
    /**
     * Constructor.
     */
	AssertException(std::string assertion, std::string file,
					unsigned int line, std::string function)
	: m_assertion(assertion), m_file(file),
	  m_line(line), m_function(function)
	{}

    const char* what () const throw ()
    {
    	std::string message = m_file + ": " + m_function + ": " +
    			"Assertion `" + m_assertion + "' failed.";
        return message.c_str();
    }

    std::string getMessage() const
    {
    	std::string message = m_file + ": " + m_function + ": " +
				"Assertion `" + m_assertion + "' failed.";
		return message;
    }

    virtual ~AssertException() throw () {}

	std::string m_assertion;
	std::string m_file;
	unsigned int m_line;
	std::string m_function;
};

/**
 * Exception that is thrown when invalid pointer conversion occured
 * (e.g. assigning nonzero integer to pointer).
 */
struct InvalidPtrConvertException : public InterpretException {
    /**
     * Constructor.
     */
	InvalidPtrConvertException() {}

    const char* what () const throw ()
    {
        return "Invalid pointer conversion";
    }
};

/**
 * Exception that is thrown when two areas that should not overlap
 * do overlap (used in interpretation of strcpy and memcpy functions).
 */
struct OverlappingAreasException : public InterpretException {
    /**
     * Constructor.
     */
	OverlappingAreasException() {}

    const char* what () const throw ()
    {
        return "Overlapping areas";
    }
};

/**
 * Exception that is thrown when deadlock occured.
 * This is the case when interpreter detected the deadlock
 * by checking for cycles in wait-for graph.
 * There are other deadlocks detected by model checker
 *
 * @see MCDeadlockException
 */
struct IDeadlockException : public InterpretException {
    /**
     * Constructor.
     */
	IDeadlockException() {}

    const char* what () const throw ()
    {
        return "Deadlock!";
    }
};

}

#endif /* INT_INTERPRET_EXCEPTIONS_H */
