// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file LockManager.h
 * Declaration of a class that manages acquiring and releasing locks.
 */

#ifndef INT_LOCK_MANAGER_H
#define INT_LOCK_MANAGER_H

//TODO: most likely delete this file

#include "../mm/Ptr.h"

#include <deque>
#include <boost/ptr_container/ptr_map.hpp>

namespace ginterpreter
{

	class Lock
	{
	public:
		typedef std::deque<int> Queue;
	private:
		bool m_isReadonly;
		unsigned int m_ownerID;
		int m_count;
		bool m_isDirty;
		Queue m_waitQueue;
		Queue m_readyQueue;
	public:

		int getOwner () const
		{ return m_ownerID; }

		void setOwner (int value)
		{
			m_isDirty |= m_ownerID != value;
			m_ownerID = value;
		}

		int getCount() const
		{ return m_count; }

		void setCount(int value)
		{
			if (value >= 0)
			{
				m_isDirty |= m_count != value;
				m_count = value;
			}
			else
				assert(false);
		}

		Queue& getWaitQueue ()
		{
			m_isDirty = true;
			return m_waitQueue;
		}

		Queue& getReadyQueue ()
		{
			m_isDirty = true;
			return m_readyQueue;
		}

		bool equals (Lock& l) const
		{
			return  m_ownerID == l.m_ownerID &&
					m_count == l.m_count &&
					queuesEqual(m_readyQueue, l.m_readyQueue) &&
					queuesEqual(m_waitQueue, l.m_waitQueue);

		}

		bool getReadOnly () const
		{
			return m_isReadonly;
		}

		void setReadOnly (bool value)
		{
			if (m_isReadonly)
				assert(false); //TODO

			m_isReadonly = value;
		}

		Lock* clone() const
		{
			return new Lock(m_ownerID, m_count);
		}

		void clean()
		{
			m_isDirty = false;
		}

		bool getIsDirty() const
		{
			return m_isDirty;
		}

		void destroy()
		{
			m_ownerID = -1;
			m_count = 0;
			m_isDirty = true;
			m_waitQueue.clear();
			m_readyQueue.clear();
		}

		Lock() : m_ownerID(-1), m_count(0), m_isDirty(true) {}

	protected:
		Lock(int owner, int count) : m_ownerID(owner), m_count(count), m_isDirty(true) {}
	private:
		bool queuesEqual(const Queue& q1, const Queue& q2) const
		{
			if (q1.size() != q2.size())
				return false;

			size_t i = 0;
			while (i < q1.size() && i < q2.size())
			{
				if (q1[i] != q2[i])
					return false;

				++i;
			}

			return true;
		}
	};

	class LockManager
	{
	private:

		Thread::ID m_threadID;

		typedef boost::ptr_map<mmodule::Ptr, Lock> Locks;

		Locks m_locks;
	public:
		bool isLocked(mmodule::Ptr ptr) const
		{
			return m_locks.find(ptr) != m_locks.end();
		}

		Lock* getLock(mmodule::Ptr ptr)
		{
			if (m_locks.find(ptr) == m_locks.end())
			{
				// TODO:
				Lock* newLock = new Lock();
				m_locks.insert(ptr, std::auto_ptr<Lock> (newLock));
				return newLock;
			}
			else
				return NULL;
		}

		bool acquireLock(mmodule::Ptr ptr, Thread::ID threadID)
		{
			Lock* l = getLock(ptr);
			bool retval = l->getOwner() == threadID || l->getOwner() == -1;
			if (retval)
			{
				l->setOwner(threadID);
				l->setCount(l->getCount() + 1);
			}
			return retval;
		}

		bool IsAcquireable(mmodule::Ptr ptr, Thread::ID threadID) //const
		{
			Lock* l = getLock(ptr);
			return l->getOwner() == threadID || l->getOwner() == -1;
		}

		void Pulse(mmodule::Ptr ptr, bool all)
		{
			Lock* l = getLock(ptr);
			if (l->getWaitQueue().size() != 0)
			{
				Lock::Queue& waitQueue = l->getWaitQueue();
				Lock::Queue& readyQueue = l->getReadyQueue();

				int max = all ? waitQueue.size() : 1;
				while (max-- > 0 && waitQueue.size() > 0)
				{
					readyQueue.push_back(waitQueue.front());
					waitQueue.pop_front();
				}
			}
		}

		void Release(mmodule::Ptr ptr)
		{
			Lock* l = getLock(ptr);
			l->setCount(l->getCount() - 1);

			if (l->getCount() == 0)
			{
				if (l->getReadyQueue().size() != 0)
				{
					// Let first in line in the ready queue acquire the lock.
					l->setOwner(l->getReadyQueue().front());
					l->getReadyQueue().pop_front();
					l->setCount(1);
				}
				else
					l->setOwner(-1);
			}
		}

	};
}

#endif /* INT_LOCK_MANAGER_H */
