// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file
 * Definitions of functions realizing the model checking algorithm.
 */

#include "ModelChecker.h"

#include "Rescheduler.h"
#include "SimpleInstructionExecutor.h"
#include "InterpretExceptions.h"
#include "ModelCheckerExceptions.h"

#include "Program.h"
#include "Value.h"
#include "ConstantsManager.h"
#include "GlobalsManager.h"
#include "GimpleListener.h"
#include "SymbolicPrinter.h"
#include "ExceptionMessages.h"
#include "InstructionLogger.h"

#include "../git/Serializer.h"
#include "../mm/MemoryState.h"
#include "../mm/MemoryAccessListener.h"
#include "../mm/MemoryExceptions.h"

#include <iostream>
#include <stack>
#include <set>

namespace ginterpreter
{

using namespace mmodule;

typedef std::stack<ThreadPool::RunnableThreads> BacktrackStack;
typedef std::stack<const giterator::Stmt*> LoggerStack;

/**
 * Class that reports unreachable areas.
 * This is passed as the visitor to the @c mmodule::MemoryState class.
 */
class ModelCheckerMMListener : public MemoryAccessListener {
  public:
    virtual ~ModelCheckerMMListener () {}

    void unreachableArea (unsigned int)
    {
        std::cout << UNREACHABLE_AREA_MSG << std::endl;
    }
};


/**
 * All the objects needed by the model checker.
 */
struct ModelCheckerData {
    /*! visitor for @c mmodule::MemoryState class */
    ModelCheckerMMListener listener;
    /*! memory state */
    MemoryState memoryState;
    /*! constants manager */
    ConstantsManager cstManager;
    /*! globals manager */
    GlobalsManager globalsManager;
    /*! listener for @c giterator::Gimple class */
    GimpleListener gimpleListener;
    /*! GIMPLE++ representation of the program */
    giterator::Gimple gimple;
    /*! set of hashes of the explored states */
    std::set<Hash> visitedStates;
    /*! stack of runnable threads used for backtracking */
    BacktrackStack backtrackStack;
    /**
     * Constructor.
     */
    ModelCheckerData ()
        : listener (),
          memoryState (&listener),
          cstManager (&memoryState),
          globalsManager (&memoryState),
          gimpleListener (&cstManager, &globalsManager),
          gimple (&gimpleListener)
    {}
};



/**
 * Class that returns the reference to the structure
 * containing the model checker data.
 */
static ModelCheckerData&
mcData ()
{
    static std::auto_ptr<ModelCheckerData> ModelCheckerData_;

    if (!ModelCheckerData_.get ())
        ModelCheckerData_.reset (new ModelCheckerData ());

    return *ModelCheckerData_.get ();
}

/**
 * Reports an exception to the standard error output.
 *
 * @param description message to report
 */
static void
reportException (const std::string& description)
{
	std::cout << OUTPUT_SEPARATOR << "Exception encountered:" << std::endl
			  << description << std::endl;
}

/**
 * Print what was the return value of the model checked program.
 *
 * @param retVal the return value or NULL if there wasn't any
 */
void mcReturnFromProgram(const Value* retVal)
{
	std::cerr << "Program returned: ";
	if (retVal != NULL)
		std::cerr << *retVal << std::endl;
	else
		std::cerr << "0" << std::endl;
}

int
doModelCheck_ (InstructionLogger& iLogger, const char* gimplexx_file);

int
doModelCheck (const char* gimplexx_file)
{
	bool exception = true;
    InstructionLogger iLogger;

    int ret;
    try
    {
    	ret = doModelCheck_ (iLogger, gimplexx_file);
    	exception = false;
    } catch (ExitException& e) {
		// This is not an error or abnormal behavior.
		mcReturnFromProgram(e.m_status);
		exception = false;
    } catch (LastThreadExitException& e) {
		mcReturnFromProgram(NULL);
		exception = false;
	} catch (FreedAreaAccessException&) {
		reportException (FREED_AREA_ACCESS_EXC_MSG);
	} catch (NullPointerException&) {
		reportException (NULL_POINTER_EXC_MSG);
	} catch (NotAreaBeginException&) {
		reportException (NOT_AREA_BEGIN_EXC_MSG);
	} catch (UndefinedMemoryLoadException&) {
		reportException (UNDEFINED_MEMORY_LOAD_EXC_MSG);
	} catch (OutOfBoundsException&) {
		reportException (OUT_OF_BOUNDS_EXC_MSG);
	} catch (UndefinedPtrOperationException&) {
		reportException (UNDEFINED_PTR_OPERATION_EXC_MSG);
	} catch (AbortException&) {
		reportException(ABORT_EXC_MSG);
	} catch (AssertException& e) {
		reportException(ASSERT_EXC_MSG + e.getMessage());
	} catch (InvalidPtrConvertException& e) {
    	reportException(INVALID_PTR_EXC_MSG);
    } catch (OverlappingAreasException& e) {
		reportException(OVERLAPPING_AREAS_EXC_MSG);
	} catch (IDeadlockException& e) {
		reportException(DEADLOCK_EXC_MSG);
	} catch (MCDeadlockException& e) {
		reportException(DEADLOCK_EXC_MSG);
	}

	if (exception || ret)
	{
		std::cout << OUTPUT_SEPARATOR;
		iLogger.logToStream(std::cout);
		return -1;
	}
	else
	{
		std::cout << OUTPUT_SEPARATOR;
		std::cout << "No errors detected." << std::endl;
		return ret;
	}
}

int
doModelCheck_ (InstructionLogger& iLogger, const char* gimplexx_file)
{
	giterator::Gimple& gimple = mcData ().gimple;
	MemoryState& memoryState = mcData ().memoryState;
	ConstantsManager& cstManager = mcData ().cstManager;
	GlobalsManager& globalsManager = mcData ().globalsManager;
	std::set<Hash>& visitedStates = mcData ().visitedStates;
	BacktrackStack& backtrackStack = mcData ().backtrackStack;

	int fail = giterator::deserialize(gimple, gimplexx_file);

	if (fail)
		return -1;

	if (!gimple.entryPoint ())
	{
		std::cerr << "ModelChecker: program entry point not found" << std::endl;
		return -1;
	}

	LOG_DEBUG ("Model Checking");

	Ref<ProgramInfo> programRef = Program::init (&gimple, &memoryState, &cstManager, &globalsManager);

	std::auto_ptr<Program> program (new Program (&gimple, &memoryState,
									&cstManager, &globalsManager, programRef, MODE_MODELCHECKER));

	program->flush();
	memoryState.push();

	bool doReschedule = false;
	bool doExecute = false;
	bool skipNextReschedule = false;
	Rescheduler rescheduler(&gimple, &memoryState, &cstManager, &globalsManager, programRef);
	ReschedulerData resData;

	int stepCount = 0;

	do
	{
		Thread* currentThread = program->getCurrentThread();

		if (currentThread->isRunnable())
		{
			doExecute = true;
			doReschedule = !skipNextReschedule &&
							currentThread->isCurrentInstructionUnsafe();
		}
		else
		{
			doExecute = false;
			doReschedule = true;
			skipNextReschedule = true;
			if (currentThread->isRunning())
				program->getThreadPool()->terminateThread(currentThread->getID());
		}

		if (doReschedule)
		{
			resData = rescheduler.reschedule(program, visitedStates, backtrackStack, iLogger, stepCount);
			skipNextReschedule = skipNextReschedule || resData.backtrackCount > 0;
			doExecute = doExecute &&
						resData.backtrackCount == 0 &&
						resData.continueExploration;
		}

		if (doExecute)
		{
			skipNextReschedule = false;

			++stepCount;

			SimpleInstructionExecutor sie;
			sie.execute(program.get(), &memoryState, &iLogger);
		}

		if (doReschedule)
		{
			if (resData.continueExploration)
			{
				assert(resData.nextThread != Thread::NO_ID);
				program->getThreadPool()->setCurrentThreadID(resData.nextThread);
			}
		}

	} while (resData.continueExploration);

	return 0;
}

}
