// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file
 * Declaration of functions realizing the model checking algorithm.
 */

#ifndef INT_MODEL_CHECKER_H
#define INT_MODEL_CHECKER_H

#ifdef __cplusplus
extern "C" {
namespace ginterpreter {
#endif

/**
 * Runs the model checker.
 * This functions runs an algorithm that explores all possible ways of program execution.
 * The algorithm in pseudocode looks like this:
 * @code
 * program.init();
 * memoryState.init();
 * memoryState.store(program);
 * resch_ret <- empty;
 * execute <- false
 * reschedule <- true;
 *
 * do
 * {
 *	currentThread <- program.currentThread();
 *
 *	if (currentThread.runnable())
 *	{
 *		execute <- true;
 *		reschedule <- currentThread.isCurrentInstructionUnsafe();
 *	}
 *		execute <- false;
 *		reschedule <- true;
 *		if (currentThread.running())
 *		{
 *			currentThread.terminate();
 *		}
 *	}
 *
 *	if (reschedule)
 *	{
 *		resch_ret <- reschedule(program, memoryState);
 *		execute <- execute && resch_ret.backtrackCount >= 0  && resch_ret.continue;
 *	}
 *
 *	if (execute)
 *	{
 *		program.step();
 *	}
 *
 *	if (reschedule)
 *	{
 *		program.currentThread <- resch_ret.next;
 *	}
 * }
 * while{resch_ret.continue != false}
 * @endcode

 * It is a depth-first search based algorithm, that runs until all possible program execution paths
 * have been explored, a deadlock was encountered or a Memory Module exception was thrown.
 * Each iteration in the main loop corresponds to a step forward or a sequence of steps backward.
 * While stepping forward means execution of instruction(s) and eventual rescheduling,
 * stepping backward means restoring a previous memory state and rescheduling
 * (i.e. backtracking in the depth-first search).
 * The rescheduler (called by @c reschedule()) compares the current memory state with
 * the previously stored ones, and if it finds that it is in a previously visited state,
 * it backtracks to the previous state. Additionally, it may select a different thread
 * as the running thread of the program for the the next iteration.
 *
 * In the beginning the algorithm itializes the model checked program and stores its initial state
 * in the simulated memory. It initializes the structure that will be returned from rescheduler.
 * This structure contains the ID of next thread that rescheduler selected to run (@c resch_ret.next),
 * number of steps the rescheduler backtracked (@c resch_ret.backtrackCount) and whether there are
 * any more non-executed possible paths of the program to explore (@c resch_ret.continue).
 *
 * If the current thread has any more instructions to be executed and the thread is not waiting for an
 * event that is not available at the moment (e.g. waiting for another thread to join), step of
 * the program is executed and it is checked whether the rescheduler should be called.
 * Rescheduler is called if the executed instruction is unsafe. This process is repeated until
 * the whole state space of the program has been explored or a deadlock has occurred.
 *
 * A deadlock occurs when none of the threads are runnable but not all of the threads are terminated.
 * This is an undesirable situation, usually the cause of incorrect locking or synchronization
 * and is often hard to detect by classical testing. Checking for deadlocks is implemented inside
 * the rescheduler.
 *
 * Function @c isCurrentInstructionUnsafe() marks an instruction as safe, when its result cannot
 * depend on the different scheduling policies of threads. That means all instructions that manipulate
 * with local variables are marked as safe. On the other hand, unsafe instructions can be affected
 * by different scheduling policies. These are the instructions that manipulate global variables.
 *
 *
 * @param gimplexx_file name of a file where the GIMPLE++ representation of a program is
 * @return
 * @see ginterpreter::Rescheduler, ginterpreter::InstructionSafetyResolver
 */
extern int doModelCheck(const char* gimplexx_file);

#ifdef __cplusplus
}}
#endif

#endif /* INT_MODEL_CHECKER_H */
