// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file Mutex.cpp
 * Definitions for a class for mutual exclusion.
 */

#include "Mutex.h"

using namespace mmodule;

namespace ginterpreter
{

Ref<MutexInfo>
Mutex::create(MemoryState* memoryState)
{
	Ref<MutexInfo> mi = memoryState->allocExpanding<MutexInfo>();
	memoryState->store(mi, &MutexInfo::mutexID, Thread::NO_ID);
	memoryState->store(mi, &MutexInfo::ownerID, Thread::NO_ID);
	memoryState->store(mi, &MutexInfo::mutexState, MUTEX_UNINITIALIZED);
	memoryState->store(mi, &MutexInfo::isReferenced, false);

	Ref<MutexWaitingThreadsInfo> mwti = memoryState->allocExpanding<MutexWaitingThreadsInfo>();
	memoryState->store(mi, &MutexInfo::waitingThreads, mwti);
	memoryState->store(mi, &MutexInfo::waitingThreadsCount, (size_t) 0);

	return mi;
}

Mutex::Mutex(ID mutexID, mmodule::MemoryState* memoryState, Ref<MutexInfo> mutexInfoRef)
: m_ID (mutexID), m_ownerID(Thread::NO_ID), m_state(MUTEX_UNINITIALIZED), m_isReferenced(false), m_memoryState(memoryState)
{
	reset(mutexInfoRef);
}

Mutex::~Mutex()
{
}

void Mutex::flush()
{
	m_memoryState->store(m_mutexInfoRef, &MutexInfo::mutexID, m_ID);
	m_memoryState->store(m_mutexInfoRef, &MutexInfo::ownerID, m_ownerID);
	m_memoryState->store(m_mutexInfoRef, &MutexInfo::mutexState, m_state);
	m_memoryState->store(m_mutexInfoRef, &MutexInfo::isReferenced, m_isReferenced);

	m_memoryState->store(m_mutexInfoRef, &MutexInfo::waitingThreadsCount, m_waitingThreads.size());

	Ref<MutexWaitingThreadsInfo> mwti = m_memoryState->load(m_mutexInfoRef, &MutexInfo::waitingThreads);

	int i = 0;
	for (WaitingThreadsConstIterator i1 = m_waitingThreads.begin(); i1 != m_waitingThreads.end(); ++i1)
	{
		m_memoryState->storeIndexed(mwti, &MutexWaitingThreadsInfo::waitingThreadID, i++, *i1);
	}

	m_memoryState->store(m_mutexInfoRef, &MutexInfo::waitingThreads, mwti);
}

void Mutex::reset(Ref<MutexInfo> mutexInfoRef)
{
	m_mutexInfoRef = mutexInfoRef;

	m_ID = m_memoryState->load(m_mutexInfoRef, &MutexInfo::mutexID);
	m_ownerID = m_memoryState->load(m_mutexInfoRef, &MutexInfo::ownerID);
	m_state = m_memoryState->load(m_mutexInfoRef, &MutexInfo::mutexState);
	m_isReferenced = m_memoryState->load(m_mutexInfoRef, &MutexInfo::isReferenced);

	m_waitingThreads.clear();

	size_t wtcount = m_memoryState->load(m_mutexInfoRef, &MutexInfo::waitingThreadsCount);
	Ref<MutexWaitingThreadsInfo> mwti = m_memoryState->load(m_mutexInfoRef, &MutexInfo::waitingThreads);

	for (size_t i = 0; i < wtcount; ++i)
	{
		m_waitingThreads.push_back(m_memoryState->loadIndexed(mwti, &MutexWaitingThreadsInfo::waitingThreadID, i));
	}
}

void
Mutex::init()
{
	assert(m_state == MUTEX_UNINITIALIZED);
	m_state = MUTEX_UNLOCKED;
}

void
Mutex::destroy()
{
	assert(m_state != MUTEX_UNINITIALIZED);
	m_state = MUTEX_UNINITIALIZED;
	m_ownerID = Thread::NO_ID;
	m_isReferenced = false;
}


void
Mutex::lock(Thread::ID threadID)
{
	assert(m_state != MUTEX_UNINITIALIZED);

	assert (m_state != MUTEX_LOCKED);
	m_state = MUTEX_LOCKED;
	assert(m_ownerID == Thread::NO_ID);
	m_ownerID = threadID;
}

void
Mutex::unlock()
{
	assert (m_state != MUTEX_UNINITIALIZED);

	m_state = MUTEX_UNLOCKED;

	//assert(m_ownerID != -1);
	if (m_ownerID == Thread::NO_ID)
		WARNING("Mutex with ID %d is being unlocked, but is not locked.", m_ID);
	m_ownerID = Thread::NO_ID;
}

bool
Mutex::isLocked() const
{
	return m_state == MUTEX_LOCKED;
}

Mutex::ID
Mutex::getID() const
{
	return m_ID;
}

void
Mutex::addWaitingThread(Thread::ID threadID)
{
	m_waitingThreads.push_back(threadID);
}

Thread::ID
Mutex::getFrontWaitingThread()
{
	if (m_waitingThreads.empty())
		return Thread::NO_ID;

	Thread::ID retID = m_waitingThreads.front();
	m_waitingThreads.pop_front();

	return retID;
}

Mutex::ID
Mutex::getOwnerID() const
{
	return m_ownerID;
}

void
Mutex::setReferenced(bool value)
{
	assert(m_state != MUTEX_UNINITIALIZED);

	m_isReferenced = value;
}

bool Mutex::isReferenced() const
{
	return m_isReferenced;
}

}
