// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file Mutex.h
 * Declaration of a class for mutual exclusion.
 */

#ifndef INT_MUTEX_H
#define INT_MUTEX_H

#include "MutexState.h"
#include "Thread.h"
#include "MutexInfo.h"

#include "../mm/MemoryState.h"

#include <deque>
#include <boost/noncopyable.hpp>

namespace ginterpreter
{

/**
 * Class representing a mutex variable.
 *
 * Mutex is designed to be used in concurrent programming to avoid
 * simultaneous use of a common resource, such as a global variable,
 * by two or more critical sections. Mutex, as a shortcut for mutual
 * exclusion is a program object that negotiates mutual exclusion
 * among threads. Threads can lock, unlock, and try to lock a mutex
 * (described in the ThreadPool class). It can be locked only by one
 * thread at a time, other thread trying to lock the same mutex
 * have to wait until the owner of the mutex unlocks it.
 * Using mutexes may result in a deadlock.
 *
 * This is a dumb class, everything important is done in class ThreadPool.
 * It is more of a structure holding the mutex ID,
 * threads threads waiting for this mutex, ID of the owner of the mutex
 * and state of the mutex.
 * Mutexes are represented in the simulated memory
 * by the MutexInfo class. This class stores a reference to
 * this structure in the simulated memory
 *
 * @see ThreadPool, Condition, MutexInfo
 */
class Mutex : boost::noncopyable
{
public:
	/*! type of the ID of mutex */
	typedef int ID;
private:
	/*! container for threads waiting for this mutex. */
	typedef std::deque<Thread::ID> WaitingThreads;
	/*! Iterator for container for threads waiting for this mutex. */
	typedef WaitingThreads::const_iterator WaitingThreadsConstIterator;
	/*! unique ID of this mutex. */
	ID m_ID;
	/* ID of threads waiting for this mutex */
	WaitingThreads m_waitingThreads;
	/*! ID of the thread that is owning the mutex. */
	Thread::ID m_ownerID;
	/*! current state of this mutex. */
	MutexState m_state;
	/*! whether this mutex is referenced (in use) by some thread condition */
	bool m_isReferenced;
	/*! Reference structure representing this mutex in the simulated memory. */
	mmodule::Ref<MutexInfo> m_mutexInfoRef;
	/*! Interface to the simulated memory */
	mmodule::MemoryState* m_memoryState;
public:
	/**
	 * Initializes mutex memory structures to be stored in the simulated memory.
	 *
	 * @param memoryState interface to the simulated memory
	 * @return initialized structure representing mutex in the simulated memory
	 */
	static mmodule::Ref<MutexInfo> create(mmodule::MemoryState* memoryState);
	/**
	 * Constructor, creates a new mutex variable.
	 * Mutex::create() should be called before this to get a structure representing
	 * the mutex variable in the simulated memory.
	 *
	 * @param mutexID unique ID of the new mutex
	 * @param memoryState interface to the simulated memory
	 * @param mutexInfoRef initialized structure representing mutex variable
	 * in the simulated memory created by previous call to Mutex::create()
	 *
	 * @see Mutex::create()
	 */
	Mutex(ID mutexID, mmodule::MemoryState* memoryState, mmodule::Ref<MutexInfo> mutexInfoRef);
	/**
	 * Destructor.
	 */
	~Mutex();
	/**
	 * Flushes all the mutex memory structures in the simulated memory.
	 */
	void flush();
	/**
	 * Reset all the the memory structures according to a mutex simulated
	 * memory representation.
	 * To be used after backtracking, when previous states of the program
	 * get reset and the mutex structures are no more actual.
	 *
	 * All the information from MutexInfo structure will
	 * be read and the mutex data structures will be updated.
	 *
	 * @param mutexInfoRef info structure according to which
	 * reset the mutex
	 *
	 * @see mmodule::MemoryState::backtrack()
	 */
	void reset(mmodule::Ref<MutexInfo> mutexInfoRef);
	/**
	 * Initialize a mutex for use, it has to be uninitialized.
	 * After initialization, the state of the mutex becomes unlocked.
	 */
	void init();
	/**
	 * Uninitialize a mutex, it cannot be already uninitialized.
	 * A destroyed mutex object can be reinitialized using init().
	 * Lock functions cannot be used on uninitialized mutex.
	 */
	void destroy();
	/**
	 * Acquires the mutex.
	 * Mutex cannot be uninitialized.
	 * Mutex cannot be held by another thread.
	 * All the intelligence of locking is implemented in ThreadPool class.
	 *
	 * @param threadID unique ID of the thread that is locking the mutex
	 *
	 * @see ThreadPool::lockMutex, ThreadPool::tryLockMutex
	 */
	 void lock(Thread::ID threadID);
	 /**
	  * Releases the mutex, sets its state to @c MUTEX_UNLOCKED.
	  * Mutex cannot be uninitialized.
	  * All the intelligence of unlocking is implemented in ThreadPool class.
	  *
	  * @see ThreadPool::unlockMutex
	  */
	 void unlock();
	 /**
	  * Whether the mutex is currently locked.
	  *
	  * @return true if the mutex is currently locked, false otherwise.
	  */
	 bool isLocked() const;
	 /** Acquire the unique ID of the mutex.
	  *
	  * @return unique ID of this mutex
	  */
	 Mutex::ID getID() const;
	 /**
	  * Add a thread into the queue of threads waiting on this mutex.
	  *
	  * @param threadID ID of thread waiting for this mutex
	  */
	 void addWaitingThread(Thread::ID threadID);
	 /**
	 * Obtain an ID of one of the threads (first in queue) waiting for current condition.
	 * -1 if there is no such thread.
	 * The information gets dequeued from queue storing such information using @c addWaitingThread().
	 *
	 * @return ID of one of the threads waiting for current condition, or -1 if there is no such thread
	 */
	 Thread::ID getFrontWaitingThread();
	 /**
	  * Acquire the ID of the owner of the mutex, or -1 if there is no owner.
	  *
	  * @return ID of owner of mutex, or -1 if there is no owner
	  */
	 Mutex::ID getOwnerID() const;
	 /**
	  * Set a flag determining whether this mutex is used by a thread condition.
	  * The mutex cannot be uninitialized.
	  *
	  * @param value true if it is used by tread condition, false otherwise.
	  *
	  * @see ThreadPool::waitCondition()
	  */
	 void setReferenced(bool value);
	 /**
	  * Find out whether this mutex is used by a thread condition.
	  *
	  * @return true if it is used by tread condition, false otherwise.
	  *
	  * @see ThreadPool::waitCondition()
	  */
	 bool isReferenced() const;
};
}


#endif /* INT_MUTEX_H */
