// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file OperationInterpreter.cpp
 * Definitions for class that interprets GIMPLE++ operations.
 */
#include "OperationInterpreter.h"

#include "ValueReader.h"
#include "Values.h"
#include "InterpretExceptions.h"
#include "VoidPtrValue.h"
#include "AddrResolver.h"

#include "../mm/MemoryState.h"

#include "../git/Gimple.h"
#include "../git/Operations.h"
#include "../git/Types.h"

#include "../Error.h"

/*
 * This must be macro, because pointer returned from ValueReader 
 * must not be used after destruction of reader object.
 */
#define READ_VALUES                             \
ValueReader op1Reader (m_func, m_memoryState, m_cstManager, m_globalsManager, m_frame); \
op.getOperand1 ()->accept (op1Reader);          \
const Value* op1Val = op1Reader.getValue ();    \
                                                \
ValueReader op2Reader (m_func, m_memoryState, m_cstManager, m_globalsManager, m_frame); \
op.getOperand2 ()->accept (op2Reader);          \
const Value* op2Val = op2Reader.getValue ();    


using namespace mmodule;
using namespace giterator;


namespace ginterpreter {

/**
 * Constructor.
 */
OperationInterpreter::OperationInterpreter (const FunctionValue* func, mmodule::MemoryState* memoryState,
					  ConstantsManager* cstManager,
					  GlobalsManager* globalsManager,
					  giterator::Gimple* gimple,
					  mmodule::Ptr frame)
:	m_func (func), m_memoryState (memoryState),
	m_cstManager (cstManager), m_globalsManager(globalsManager),
	m_gimple (gimple), m_frame (frame)
{}

/**
 * Acquires the result of the operation.
 */
std::auto_ptr<Value> OperationInterpreter::acquireResult ()
{ return m_result; }

void OperationInterpreter::visit (const NegateOperation& op)
{
	ValueReader reader (m_func, m_memoryState, m_cstManager, m_globalsManager, m_frame);
	op.getOperand()->accept(reader);

	m_result = reader.getValue()->negateOp();
}

void OperationInterpreter::visit (const AbsOperation& op)
{
	ValueReader reader (m_func, m_memoryState, m_cstManager, m_globalsManager, m_frame);
	op.getOperand ()->accept (reader);

	m_result = reader.getValue ()->absOp() ;
}

void OperationInterpreter::visit (const FixTruncOperation& op)
{
	ValueReader reader (m_func, m_memoryState, m_cstManager, m_globalsManager, m_frame);
	op.getOperand ()->accept (reader);

	const RealValue* realVal = dynamic_cast<const RealValue*> (reader.getValue());
	assert(realVal);

	m_result.reset(new IntegerValue(op.getResultType(), (uint64_t) realVal->getValue()));
}

void OperationInterpreter::visit (const FloatOperation& op)
{
	ValueReader reader (m_func, m_memoryState, m_cstManager, m_globalsManager, m_frame);
	op.getOperand ()->accept (reader);

	const IntegerValue* intVal = static_cast<const IntegerValue*> (reader.getValue());
	assert(intVal);
	m_result.reset(new RealValue(op.getResultType(), (long double) intVal->getValue()));
}

void OperationInterpreter::visit (const BitNotOperation& op)
{
	ValueReader reader (m_func, m_memoryState, m_cstManager, m_globalsManager, m_frame);
	op.getOperand ()->accept (reader);

	m_result = reader.getValue ()->bitNotOp() ;
}

void OperationInterpreter::visit (const TruthNotOperation& op)
{
	ValueReader reader (m_func, m_memoryState, m_cstManager, m_globalsManager, m_frame);
	op.getOperand ()->accept (reader);

	m_result = reader.getValue ()->truthNotOp() ;
}

void OperationInterpreter::visit (const ParenOperation& op)
{
	// TODO: this does not do anything really, is it ok?
	ValueReader reader (m_func, m_memoryState, m_cstManager, m_globalsManager, m_frame);
	op.getOperand ()->accept (reader);
	m_result = reader.acquireValue ();
}

void OperationInterpreter::visit (const NonLvalueOperation& op)
{
	ValueReader reader (m_func, m_memoryState, m_cstManager, m_globalsManager, m_frame);
	op.getOperand ()->accept (reader);
	m_result = reader.acquireValue ();
}

void OperationInterpreter::visit (const NoOperation& op)
{
	ValueReader reader (m_func, m_memoryState, m_cstManager, m_globalsManager, m_frame);
	op.getOperand ()->accept (reader);
	m_result = reader.acquireValue ();
}

Value* convertToInteger(const Type* resultType, const Value* valueToConvert)
{
	uint64_t rawValue = 0;
	if (typeid (*valueToConvert) == typeid (const IntegerValue))
	{
		const IntegerValue* intVal = static_cast<const IntegerValue*> (valueToConvert);
		rawValue = intVal->getValue ();
	}
	else if (typeid (*valueToConvert) == typeid (const PtrValue))
	{
		const PtrValue* ptrVal = static_cast<const PtrValue*> (valueToConvert);
		rawValue = ptrVal->getPtr ().getOffset ();
	}
	else if (typeid (*valueToConvert) == typeid (const RealValue))
	{
		const RealValue* realVal = static_cast<const RealValue*> (valueToConvert);
		rawValue = (uint64_t) realVal->getValue();
	}
	else
		ni_error ("converting to int");

	return new IntegerValue(resultType, rawValue);
}

Value* convertToReal(const Type* resultType, const Value* valueToConvert)
{
	long double rawValue = 0;
	if (typeid (*valueToConvert) == typeid (const IntegerValue))
	{
		const IntegerValue* intVal = static_cast<const IntegerValue*> (valueToConvert);
		rawValue = (long double) intVal->getValue ();
	}
	else if (typeid (*valueToConvert) == typeid (const PtrValue))
	{
		const PtrValue* ptrVal = static_cast<const PtrValue*> (valueToConvert);
		rawValue = (long double) ptrVal->getPtr ().getOffset ();
	}
	else if (typeid (*valueToConvert) == typeid (const RealValue))
	{
		const RealValue* realVal = static_cast<const RealValue*> (valueToConvert);
		rawValue = realVal->getValue();
	}
	else
		ni_error ("converting to real");

	return new RealValue(resultType, rawValue);
}

Value* convertToBool(const Type* resultType, const Value* valueToConvert)
{
	bool rawValue = 0;
	if (typeid (*valueToConvert) == typeid (const IntegerValue))
	{
		const IntegerValue* intVal = static_cast<const IntegerValue*> (valueToConvert);
		rawValue = !intVal->isZero();
	}
	else if (typeid (*valueToConvert) == typeid (const PtrValue))
	{
		const PtrValue* ptrVal = static_cast<const PtrValue*> (valueToConvert);
		rawValue = !ptrVal->getPtr ().isNull();
	}
	else if (typeid (*valueToConvert) == typeid (const RealValue))
	{
		const RealValue* realVal = static_cast<const RealValue*> (valueToConvert);
		rawValue = !realVal->isZero();
	}
	else
		ni_error ("converting to bool");

	return new BoolValue(resultType, rawValue);
}

Value* convertToEnum(const Type* resultType, const Value* valueToConvert)
{
	uint64_t rawValue = 0;
	if (typeid (*valueToConvert) == typeid (const IntegerValue))
	{
		const IntegerValue* intVal = static_cast<const IntegerValue*> (valueToConvert);
		rawValue = intVal->getValue ();
	}
	else if (typeid (*valueToConvert) == typeid (const RealValue))
	{
		const RealValue* realVal = static_cast<const RealValue*> (valueToConvert);
		rawValue = (uint64_t) realVal->getValue();
	}
	else
		ni_error ("converting to enum");

	return new EnumeralValue(resultType, rawValue);
}

Value* convertToPtr(const Type* resultType, const Value* valueToConvert)
{
	if (typeid (*valueToConvert) == typeid (const IntegerValue))
	{
		const IntegerValue* intVal = static_cast<const IntegerValue*> (valueToConvert);
		if (intVal->isZero())
			return new PtrValue(resultType, Ptr());
		else
			throw InvalidPtrConvertException();
	}
	else if (typeid (*valueToConvert) == typeid (const PtrValue))
	{
		const PtrValue* ptrVal = static_cast<const PtrValue*> (valueToConvert);
		return new PtrValue(resultType, ptrVal->getPtr());
	}
	else if (typeid (*valueToConvert) == typeid (const RealValue))
	{
		const RealValue* realVal = static_cast<const RealValue*> (valueToConvert);
		if (realVal->isZero())
			return new PtrValue(resultType, Ptr());
		else
			throw InvalidPtrConvertException();
	}
	else
		ni_error ("converting to ptr");
}

Value* convertToVoidPtr(const Type* resultType, mmodule::Ptr valPtr)
{
	return new VoidPtrValue(resultType, valPtr);
}

void OperationInterpreter::visit (const ConvertOperation& op)
{
	ValueReader reader (m_func, m_memoryState, m_cstManager, m_globalsManager, m_frame);
	op.getOperand ()->accept (reader);

	std::auto_ptr<Value> value = reader.acquireValue ();

	//assert (!op.getResultType ()->isSame (*value->getType ()));

	const Type* resType = op.getResultType();

	// conversion from void ptr type
	if (typeid(*value.get()) == typeid(const VoidPtrValue))
	{
		const VoidPtrValue* voidPtrVal = dynamic_cast<const VoidPtrValue*> (value.get());
		const Value* result = dynamic_cast<const Value*> (m_memoryState->load(voidPtrVal->getPtr()));
		assert(result);
		m_result.reset(result->clone());
		return;
	}

	Value* resValue = NULL;

	if (typeid(*resType) == typeid(const IntegerType))
		resValue = convertToInteger(resType, value.get());
	else if (typeid(*resType) == typeid(const RealType))
		resValue = convertToReal(resType, value.get());
	else if (typeid(*resType) == typeid(const BoolType))
		resValue = convertToBool(resType, value.get());
	else if (typeid(*resType) == typeid(const EnumType))
		resValue = convertToEnum(resType, value.get());
	else if (typeid(*resType) == typeid(const PtrType))
	{
		const PtrType* ptrType = dynamic_cast<const PtrType*> (resType);
		if (typeid(*ptrType->getPointingToType()) == typeid(const VoidType))
		{
			AddrResolver addRes(m_func, m_memoryState, m_cstManager, m_globalsManager, m_frame);
			op.getOperand ()->accept (addRes);
			assert(!addRes.getPtr().isNull());
			resValue = convertToVoidPtr(resType, addRes.getPtr());
		}
		else
			resValue = convertToPtr(resType, value.get());
	}
	else
		ni_error("invalid conversion");

	m_result.reset (resValue);
}

void OperationInterpreter::visit (const PlusOperation& op)
{
    READ_VALUES;
    m_result = op1Val->addOp (*op2Val);
}

void OperationInterpreter::visit (const MinusOperation& op)
{
    READ_VALUES;
    m_result = op1Val->subOp (*op2Val);
}

void OperationInterpreter::visit (const MultOperation& op)
{
	READ_VALUES;
	m_result = op1Val->multOp (*op2Val);
}

void OperationInterpreter::visit (const RdivOperation& op)
{
	READ_VALUES;
	m_result = op1Val->rDivOp (*op2Val);
}

void OperationInterpreter::visit (const TruncDivOperation& op)
{
    READ_VALUES;
    m_result = op1Val->truncDivOp (*op2Val);
}

void OperationInterpreter::visit (const FloorDivOperation& op)
{
	READ_VALUES;
	m_result = op1Val->floorDivOp(*op2Val);
}

void OperationInterpreter::visit (const CeilDivOperation& op)
{
	READ_VALUES;
	m_result = op1Val->ceilDivOp(*op2Val);
}

void OperationInterpreter::visit (const RoundDivOperation& op)
{
	READ_VALUES;
	m_result = op1Val->roundDivOp(*op2Val);
}

void OperationInterpreter::visit (const TruncModOperation& op)
{
	READ_VALUES;
	m_result = op1Val->truncModOp (*op2Val);
}

void OperationInterpreter::visit (const FloorModOperation& op)
{
	READ_VALUES;
	m_result = op1Val->floorModOp (*op2Val);
}

void OperationInterpreter::visit (const CeilModOperation& op)
{
	READ_VALUES;
	m_result = op1Val->ceilModOp (*op2Val);
}

void OperationInterpreter::visit (const RoundModOperation& op)
{
	READ_VALUES;
	m_result = op1Val->roundModOp (*op2Val);
}

void OperationInterpreter::visit (const ExactDivOperation& op)
{
	READ_VALUES;
	m_result = op1Val->truncDivOp (*op2Val);
}

void OperationInterpreter::visit (const MinOperation& op)
{
	READ_VALUES;
	if (*op1Val <= *op2Val)
		m_result = op1Reader.acquireValue ();
	else
		m_result = op2Reader.acquireValue ();
}

void OperationInterpreter::visit (const MaxOperation& op)
{
	READ_VALUES;
	if (*op1Val >= *op2Val)
		m_result = op1Reader.acquireValue ();
	else
		m_result = op2Reader.acquireValue ();
}

void OperationInterpreter::visit (const LshiftOperation& op)
{
	READ_VALUES;
	m_result = op1Val->lShiftOp (*op2Val);
}

void OperationInterpreter::visit (const RshiftOperation& op)
{
	READ_VALUES;
	m_result = op1Val->rShiftOp (*op2Val);
}

void OperationInterpreter::visit (const LrotateOperation& op)
{
	READ_VALUES;
	m_result = op1Val->lRotateOp (*op2Val);
}

void OperationInterpreter::visit (const RrotateOperation& op)
{
	READ_VALUES;
	m_result = op1Val->rRotateOp (*op2Val);
}

void OperationInterpreter::visit (const BitIorOperation& op)
{
	READ_VALUES;
	m_result = op1Val->bitIorOp (*op2Val);
}

void OperationInterpreter::visit (const BitXorOperation& op)
{
	READ_VALUES;
	m_result = op1Val->bitXorOp (*op2Val);
}

void OperationInterpreter::visit (const BitAndOperation& op)
{
	READ_VALUES;
	m_result = op1Val->bitAndOp (*op2Val);
}

void OperationInterpreter::visit (const TruthAndOperation& op)
{
	READ_VALUES;
	m_result = op1Val->truthAndOp (*op2Val);
}

void OperationInterpreter::visit (const TruthOrOperation& op)
{
	READ_VALUES;
	m_result = op1Val->truthOrOp (*op2Val);
}

void OperationInterpreter::visit (const TruthXorOperation& op)
{
	READ_VALUES;
	m_result = op1Val->truthXorOp (*op2Val);
}

void OperationInterpreter::visit (const LtOperation& op)
{
	READ_VALUES;
	m_result.reset (new BoolValue (m_gimple->getBoolType (), *op1Val < *op2Val));
}

void OperationInterpreter::visit (const LeOperation& op)
{
	READ_VALUES;
	m_result.reset (new BoolValue (m_gimple->getBoolType (), *op1Val <= *op2Val));
}

void OperationInterpreter::visit (const GtOperation& op)
{
	READ_VALUES;
	m_result.reset (new BoolValue (m_gimple->getBoolType (), *op1Val > *op2Val));
}

void OperationInterpreter::visit (const GeOperation& op)
{
	READ_VALUES;
	m_result.reset (new BoolValue (m_gimple->getBoolType (), *op1Val >= *op2Val));
}

void OperationInterpreter::visit (const EqOperation& op)
{
	READ_VALUES;
	m_result.reset (new BoolValue (m_gimple->getBoolType (), *op1Val == *op2Val));
}

void OperationInterpreter::visit (const NeOperation& op)
{
	READ_VALUES;
	m_result.reset (new BoolValue (m_gimple->getBoolType (), *op1Val != *op2Val));
}

void OperationInterpreter::visit (const OrderedOperation& op)
{
	READ_VALUES;

	bool ordered = op1Val->orderedWith(*op2Val);
	m_result.reset (new BoolValue (m_gimple->getBoolType (), ordered));
}

void OperationInterpreter::visit (const UnorderedOperation& op)
{
	READ_VALUES;

	bool unordered = !op1Val->orderedWith(*op2Val);
	m_result.reset (new BoolValue (m_gimple->getBoolType (), unordered));
}

void OperationInterpreter::visit (const UnLtOperation& op)
{
	READ_VALUES;

	bool returnValue;
	if (!op1Val->orderedWith(*op2Val))
		returnValue = true;
	else
		returnValue = *op1Val < *op2Val;

	m_result.reset (new BoolValue (m_gimple->getBoolType (), returnValue));
}

void OperationInterpreter::visit (const UnLeOperation& op)
{
	READ_VALUES;

	bool returnValue;
	if (!op1Val->orderedWith(*op2Val))
		returnValue = true;
	else
		returnValue = *op1Val <= *op2Val;

	m_result.reset (new BoolValue (m_gimple->getBoolType (), returnValue));
}

void OperationInterpreter::visit (const UnGtOperation& op)
{
	READ_VALUES;

	bool returnValue;
	if (!op1Val->orderedWith(*op2Val))
		returnValue = true;
	else
		returnValue = *op1Val > *op2Val;

	m_result.reset (new BoolValue (m_gimple->getBoolType (), returnValue));
}

void OperationInterpreter::visit (const UnGeOperation& op)
{
	READ_VALUES;

	bool returnValue;
	if (!op1Val->orderedWith(*op2Val))
		returnValue = true;
	else
		returnValue = *op1Val >= *op2Val;

	m_result.reset (new BoolValue (m_gimple->getBoolType (), returnValue));
}

void OperationInterpreter::visit (const UnEqOperation& op)
{
	READ_VALUES;

	bool returnValue;
	if (!op1Val->orderedWith(*op2Val))
		returnValue = true;
	else
		returnValue = *op1Val == *op2Val;

	m_result.reset (new BoolValue (m_gimple->getBoolType (), returnValue));
}

void OperationInterpreter::visit (const LtGtOperation& op)
{
	READ_VALUES;

	bool returnValue;
	if (!op1Val->orderedWith(*op2Val))
		returnValue = true;
	else
		returnValue = *op1Val != *op2Val;

	m_result.reset (new BoolValue (m_gimple->getBoolType (), returnValue));
}

void OperationInterpreter::visit (const WidenSumOperation& op)
{
	READ_VALUES;

	assert(op2Val->getUnitSize() >= 2 * op1Val->getUnitSize());

	m_result = op2Val->addOp (*op1Val);
}

void OperationInterpreter::visit (const WidenMultOperation& op)
{
	READ_VALUES;

	assert(op2Val->getUnitSize() >= 2 * op1Val->getUnitSize());

	m_result = op2Val->multOp (*op1Val);
}

}
