// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file PrinterBase.cpp
 * Definitions for base class for printers that can print GIMPLE++ classes.
 */
#include "PrinterBase.h"

#include "../Error.h"

#include "../git/Constants.h"
#include "../git/Operands.h"
#include "../git/Operations.h"
#include "../git/Type.h"
#include "../git/Stmts.h"

#include <ostream>
#include <boost/foreach.hpp>

using namespace giterator;


namespace ginterpreter {
    
PrinterBase::PrinterBase(std::ostream& os)
    : m_os (os)
{
}


void
PrinterBase::visit (const IntegerConstant& cst)
{
    m_os << cst.getValue ();
}

void
PrinterBase::visit (const EnumeralConstant& cst)
{
    m_os << cst.getValue ();
}

void
PrinterBase::visit (const RealConstant& cst)
{
    m_os << cst.getValue ();
}

void
PrinterBase::visit (const StringConstant& cst)
{
    m_os << '"' << cst.getValue () << '"';
}


void
PrinterBase::visit(const AddrOperand& op)
{
    m_os << "&";
    op.getSubOperand ()->accept (*this);
}
            

void
PrinterBase::visit(const IndirectRefOperand& op)
{
    m_os << "*";
    op.getSubOperand ()->accept (*this);
}
            

void
PrinterBase::visit(const ArrayRefOperand& op)
{
    op.getArray ()->accept (*this);
    m_os << "[";
    op.getIndex ()->accept (*this);
    m_os << "]";
}


void
PrinterBase::visit(const ComponentRefOperand& op)
{
    op.getRecord ()->accept (*this);
    m_os << ".";
    op.getField ()->accept (*this);
}
            
void 
PrinterBase::visit(const CstOperand& op)
{
    op.getConstant ()->accept (*this);
}

void PrinterBase::visit (const LabelExprOperand& op)
{
	if (op.getIsDefault ())
	{
		m_os << "default:";
	}
	else
	{
		m_os << "case ";
		if (op.getIsRange ())
		{
			op.getLower ()->accept (*this);
			m_os << "..";
			op.getUpper ()->accept (*this);
		}
		else
			op.getLower ()->accept (*this);
		m_os << ":";
	}
}

void PrinterBase::visit (const ArrayConstructorOperand& op)
{
	m_os << "{";

	bool hasPrinted = false;
	for (ArrayConstructorOperand::ArrayConstructorsIterator i1 = op.constructors().first; i1 != op.constructors().second; ++i1)
	{
		if (hasPrinted)
			m_os << ", ";
		else
			hasPrinted = true;

		m_os << "(";
		m_os << "index: ";
		i1->first->accept (*this);
		m_os << ", value: ";
		i1->second->accept(*this);
		m_os << ")";
	}

	m_os << "}";
}

/*
void PrinterBase::visit (const ArrayFieldConstructorOperand& op)
{
	op.getValue ()-> accept(*this);
}*/

void PrinterBase::visit (const RecordConstructorOperand& op)
{
	m_os << "{";

	bool hasPrinted = false;
	for (RecordConstructorOperand::RecordConstructorsIterator i1 = op.constructors().first; i1 != op.constructors().second; ++i1)
	{
		if (hasPrinted)
			m_os << ", ";
		else
			hasPrinted = true;

		m_os << "(";
		i1->second->accept(*this);
		m_os << ")";
	}

	m_os << "}";
}

/*
void PrinterBase::visit (const RecordFieldConstructorOperand& op)
{
	op.getValue ()->accept (*this);
}*/


static void
printUnOp(std::ostream& os, Operand::Visitor& printer,
		  const UnaryOperation& op, const std::string& opStr)
{
	os << opStr;
	op.getOperand ()->accept (printer);
}

static void
printBinOp (std::ostream& os, Operand::Visitor& printer,
            const BinaryOperation& op, const std::string& opStr)
{
    op.getOperand1 ()->accept (printer);
    os << opStr;
    op.getOperand2 ()->accept (printer);
}

void PrinterBase::visit (const NegateOperation& op)
{
	ni_error("NegateOperation (PrinterBase)");
}

void
PrinterBase::visit (const AbsOperation& op)
{
	m_os << "abs(" << op.getOperand () << ")";
}

void
PrinterBase::visit (const FixTruncOperation& op)
{
    m_os << "fix_trunc: ";
    op.getOperand ()->accept (*this);
}

void
PrinterBase::visit (const FloatOperation& op)
{
    m_os << "(float) ";
    op.getOperand ()->accept (*this);
}

void PrinterBase::visit (const BitNotOperation& op)
{
	printUnOp (m_os, *this, op, "~ ");
}

void PrinterBase::visit (const TruthNotOperation& op)
{
	printUnOp (m_os, *this, op, "(truth)~ ");
}

void PrinterBase::visit (const ParenOperation& op)
{
	ni_error("ParenOperation (PrinterBase)");
}

void PrinterBase::visit (const NonLvalueOperation& op)
{
	ni_error("NonLvalueOperation (PrinterBase)");
}

void
PrinterBase::visit (const NoOperation& op)
{
    op.getOperand ()->accept (*this);
}

void
PrinterBase::visit (const ConvertOperation& op)
{
    m_os << "(" << op.getResultType ()->getName () << ")";
    op.getOperand ()->accept (*this);
}

void
PrinterBase::visit (const PlusOperation& op)
{
    printBinOp (m_os, *this, op, " + ");
}

void
PrinterBase::visit (const MinusOperation& op)
{
    printBinOp (m_os, *this, op, " - ");
}

void
PrinterBase::visit (const MultOperation& op)
{
    printBinOp (m_os, *this, op, " * ");
}

void PrinterBase::visit (const RdivOperation& op)
{
	printBinOp (m_os, *this, op, " / ");
}

void
PrinterBase::visit (const TruncDivOperation& op)
{
    printBinOp (m_os, *this, op, " / ");
}

void PrinterBase::visit (const FloorDivOperation& op)
{
	ni_error("FloorDivOperation (PrinterBase)");
}

void PrinterBase::visit (const CeilDivOperation& op)
{
	ni_error("CeilDivOperation (PrinterBase)");
}

void PrinterBase::visit (const RoundDivOperation& op)
{
	ni_error("RoundDivOperation (PrinterBase)");
}

void
PrinterBase::visit (const TruncModOperation& op)
{
    printBinOp (m_os, *this, op, " % ");
}

void PrinterBase::visit (const FloorModOperation& op)
{
	ni_error("FloorModOperation (PrinterBase)");
}

void PrinterBase::visit (const CeilModOperation& op)
{
	ni_error("CeilModOperation (PrinterBase)");
}

void PrinterBase::visit (const RoundModOperation& op)
{
	ni_error("RoundModOperation (PrinterBase)");
}

void PrinterBase::visit (const ExactDivOperation& op)
{
	ni_error("ExactDivOperation (PrinterBase)");
}

void
PrinterBase::visit (const MinOperation& op)
{
	m_os << "min(" << op.getOperand1 () << ", " << op.getOperand2 () << ")";
}

void
PrinterBase::visit (const MaxOperation& op)
{
	m_os << "max(" << op.getOperand1 () << ", " << op.getOperand2 () << ")";
}

void PrinterBase::visit (const LshiftOperation& op)
{
	printBinOp (m_os, *this, op, " << ");
}

void PrinterBase::visit (const RshiftOperation& op)
{
	printBinOp (m_os, *this, op, " >> ");
}

void PrinterBase::visit (const LrotateOperation& op)
{
	ni_error("LrotateOperation (PrinterBase)");
}

void PrinterBase::visit (const RrotateOperation& op)
{
	ni_error("RrotateOperation (PrinterBase)");
}

void PrinterBase::visit (const BitIorOperation& op)
{
	printBinOp (m_os, *this, op, " | ");
}

void PrinterBase::visit (const BitXorOperation& op)
{
	printBinOp (m_os, *this, op, " ^ ");
}

void PrinterBase::visit (const BitAndOperation& op)
{
	printBinOp (m_os, *this, op, " & ");
}

void PrinterBase::visit (const TruthAndOperation& op)
{
	printBinOp (m_os, *this, op, " && ");
}

void PrinterBase::visit (const TruthOrOperation& op)
{
	printBinOp (m_os, *this, op, " || ");
}

void PrinterBase::visit (const TruthXorOperation& op)
{
	printBinOp (m_os, *this, op, " XOR ");
}

void
PrinterBase::visit (const LtOperation& op)
{
    printBinOp (m_os, *this, op, " < ");
}

void
PrinterBase::visit (const LeOperation& op)
{
    printBinOp (m_os, *this, op, " <= ");
}

void
PrinterBase::visit (const GtOperation& op)
{
    printBinOp (m_os, *this, op, " > ");
}

void
PrinterBase::visit (const GeOperation& op)
{
    printBinOp (m_os, *this, op, " >= ");
}

void
PrinterBase::visit (const EqOperation& op)
{
    printBinOp (m_os, *this, op, " == ");
}

void
PrinterBase::visit (const NeOperation& op)
{
    printBinOp (m_os, *this, op, " != ");
}

void PrinterBase::visit (const OrderedOperation& op)
{
	ni_error("OrderedOperation (PrinterBase)");
}

void PrinterBase::visit (const UnorderedOperation& op)
{
	ni_error("UnorderedOperation (PrinterBase)");
}

void PrinterBase::visit (const UnLtOperation& op)
{
	ni_error("createUnLtOperation (PrinterBase)");
}

void PrinterBase::visit (const UnLeOperation& op)
{
	ni_error("UnLeOperation (PrinterBase)");
}

void PrinterBase::visit (const UnGtOperation& op)
{
	ni_error("UnGtOperation (PrinterBase)");
}

void PrinterBase::visit (const UnGeOperation& op)
{
	ni_error("UnGeOperation (PrinterBase)");
}

void PrinterBase::visit (const UnEqOperation& op)
{
	ni_error("UnEqOperation (PrinterBase)");
}

void PrinterBase::visit (const LtGtOperation& op)
{
	ni_error("LtGtOperation (PrinterBase)");
}

void PrinterBase::visit (const WidenSumOperation& op)
{
	ni_error("WidenSumOperation (PrinterBase)");
}

void PrinterBase::visit (const WidenMultOperation& op)
{
	ni_error("WidenMultOperation (PrinterBase)");
}
   
   
void 
PrinterBase::visit (const CallStmt& stmt)
{
    if (stmt.getLhs ())
    {
        stmt.getLhs ()->accept (*this);
        m_os << " = ";
    }
    stmt.getFunction ()->accept (*this);
    m_os << " (";
    
    bool hasPrinted = false;
    BOOST_FOREACH (const Operand& op, stmt.parameters ())
    {
        if (hasPrinted)
            m_os << ", ";
        else
            hasPrinted = true;
        
        op.accept (*this);
    }
    
    m_os << ")";
}
    

void 
PrinterBase::visit (const ReturnStmt& stmt)
{
    m_os << "return";
    if (stmt.getReturnValue ())
    {
        m_os << " ";
        stmt.getReturnValue ()->accept (*this);
    }
}

    
void 
PrinterBase::visit (const AssignStmt& stmt)
{
    stmt.getLhs ()->accept (*this);
    m_os << " = ";
    stmt.getOperation ()->accept (*this);
}

    
void 
PrinterBase::visit (const CondStmt& stmt)
{
    m_os << "if (";
    stmt.getOperation ()->accept (*this);
    m_os << ") ";
}

void
PrinterBase::visit (const GotoStmt& stmt)
{
	m_os << "goto";
	stmt.getDest ()->accept (*this);
}
void
PrinterBase::visit (const LabelStmt& stmt)
{
	stmt.getLabel ()->accept (*this);
	m_os << ":";
}

void
PrinterBase::visit (const NopStmt&)
{
	m_os << "NopStmt";
}

void
PrinterBase::visit (const PhiStmt& stmt)
{
	stmt.getResult ()->accept (*this);
	m_os << " = ";

	m_os << " (";
	bool hasPrinted = false;
	BOOST_FOREACH(const Operand& op, stmt.arguments())
	{
		if (hasPrinted)
			m_os << ", ";
		else
			hasPrinted = true;

		op.accept (*this);
	}

	m_os << ")";
}

void
PrinterBase::visit (const SwitchStmt& stmt)
{
	m_os << "switch (";
	stmt.getIndex ()->accept (*this);
	m_os << ")";

	m_os << "cases: (";
	bool hasPrinted = false;
	BOOST_FOREACH(const Operand& op, stmt.labels())
	{
		if (hasPrinted)
			m_os << ", ";
		else
			hasPrinted = true;

		op.accept (*this);
	}

	m_os << ")";
}

}
