// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file PrinterBase.h
 * Declaration of base class for printers that can print GIMPLE++ classes.
 */
#ifndef INT_PRINTER_BASE_H
#define INT_PRINTER_BASE_H

#include "../git/git_fwd.h"
#include "../git/Constant.h"
#include "../git/Operand.h"
#include "../git/Operation.h"
#include "../git/Operations_fwd.h"
#include "../git/Stmt.h"

#include <iosfwd>

namespace ginterpreter {
    
/**
 * Base class for printers that can print constants, operands, operations and statements.
 */
class PrinterBase 
    : public giterator::Constant::Visitor,
      public giterator::Operand::Visitor,
      public giterator::Operation::Visitor,
      public giterator::Stmt::Visitor
{
  protected:
    std::ostream& m_os;
    
  public:
    /**
     * Constructor.
     *
     * @param os output stream where to print
     */
    PrinterBase (std::ostream& os);

    virtual ~PrinterBase () {}

    /** @name Constant visitors
     * These methods are used to print the constants.
     */
    //@{
    void visit (const giterator::IntegerConstant& cst);
    void visit (const giterator::EnumeralConstant& cst);
    void visit (const giterator::RealConstant& cst);
    void visit (const giterator::StringConstant& cst);
    //@}
    
    /** @name Operand visitors
     * These methods are used to print operands.
     */
    //@{
    void visit(const giterator::AddrOperand& op);
    void visit(const giterator::IndirectRefOperand& op);
    void visit(const giterator::ArrayRefOperand& op);
    void visit(const giterator::ComponentRefOperand& op);
    void visit(const giterator::CstOperand& op);
    void visit(const giterator::LabelExprOperand& op);
    void visit (const giterator::ArrayConstructorOperand& op);
    void visit (const giterator::RecordConstructorOperand& op);
    //@}

    /** @name Operation visitors
     * These methods are used to print operations.
     */
    //@{
    void visit (const giterator::NegateOperation&);
	void visit (const giterator::AbsOperation&);
	void visit (const giterator::FixTruncOperation&);
	void visit (const giterator::FloatOperation&);
	void visit (const giterator::BitNotOperation&);
	void visit (const giterator::TruthNotOperation&);
	void visit (const giterator::ParenOperation&);
	void visit (const giterator::NonLvalueOperation&);
	void visit (const giterator::NoOperation&);
	void visit (const giterator::ConvertOperation&);

	void visit (const giterator::PlusOperation&);
	void visit (const giterator::MinusOperation&);
	void visit (const giterator::MultOperation&);
	void visit (const giterator::RdivOperation&);
	void visit (const giterator::TruncDivOperation&);
	void visit (const giterator::FloorDivOperation&);
	void visit (const giterator::CeilDivOperation&);
	void visit (const giterator::RoundDivOperation&);
	void visit (const giterator::TruncModOperation&);
	void visit (const giterator::FloorModOperation&);
	void visit (const giterator::CeilModOperation&);
	void visit (const giterator::RoundModOperation&);
	void visit (const giterator::ExactDivOperation&);
	void visit (const giterator::MinOperation&);
	void visit (const giterator::MaxOperation&);
	void visit (const giterator::LshiftOperation&);
	void visit (const giterator::RshiftOperation&);
	void visit (const giterator::LrotateOperation&);
	void visit (const giterator::RrotateOperation&);
	void visit (const giterator::BitIorOperation&);
	void visit (const giterator::BitXorOperation&);
	void visit (const giterator::BitAndOperation&);
	void visit (const giterator::TruthAndOperation&);
	void visit (const giterator::TruthOrOperation&);
	void visit (const giterator::TruthXorOperation&);
	void visit (const giterator::LtOperation&);
	void visit (const giterator::LeOperation&);
	void visit (const giterator::GtOperation&);
	void visit (const giterator::GeOperation&);
	void visit (const giterator::EqOperation&);
	void visit (const giterator::NeOperation&);
	void visit (const giterator::OrderedOperation&);
	void visit (const giterator::UnorderedOperation&);
	void visit (const giterator::UnLtOperation&);
	void visit (const giterator::UnLeOperation&);
	void visit (const giterator::UnGtOperation&);
	void visit (const giterator::UnGeOperation&);
	void visit (const giterator::UnEqOperation&);
	void visit (const giterator::LtGtOperation&);
	void visit (const giterator::WidenSumOperation&);
	void visit (const giterator::WidenMultOperation&);
    //@}
    
    /** @name Statement visitors
     * These methods are used to print statements.
     */
    //@{
    void visit (const giterator::CallStmt& stmt);
    void visit (const giterator::ReturnStmt& stmt);
    void visit (const giterator::AssignStmt& stmt);
    void visit (const giterator::CondStmt& stmt);
    void visit (const giterator::GotoStmt& stmt);
    void visit (const giterator::LabelStmt& stmt);
    void visit (const giterator::NopStmt& stmt);
    void visit (const giterator::PhiStmt& stmt);
    void visit (const giterator::SwitchStmt& stmt);
    //@}
};

}

#endif /* INT_PRINTER_BASE_H */
