// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file Program.cpp
 * Definitions for class that interprets program.
 */
#include "Program.h"

#include "Thread.h"
#include "ProgramInfo.h"
#include "ThreadInfo.h"
#include "ConstantsManager.h"
#include "GlobalsManager.h"
#include "FunctionValue.h"
#include "ModelCheckerExceptions.h"
#include "InstructionLogger.h"

#include "../git/FunctionDecl.h"
#include "../git/Gimple.h"

using namespace mmodule;
using namespace giterator;


namespace ginterpreter {

#define MAIN_THREAD_ID 0
#define MAIN_THREAD_NAME "main"

Ref<ProgramInfo> 
Program::init (Gimple* gimple, MemoryState* memoryState,
			   ConstantsManager* cstManager, GlobalsManager* globalsManager)
{
    Ptr mainFuncPtr = cstManager->getFuncPtr (gimple->entryPoint ());
    assert (!mainFuncPtr.isNull ());
    
    const FunctionValue* mainFunc = 
            dynamic_cast<const FunctionValue*> (memoryState->load (mainFuncPtr));

    Ref<ThreadInfo> threadInfoRef = Thread::create (mainFunc, memoryState, Ptr());
    
    Ref<ThreadsInfo> threadsInfoRef = memoryState->allocExpanding<ThreadsInfo>();
	memoryState->storeIndexed(threadsInfoRef, &ThreadsInfo::threads, MAIN_THREAD_ID, threadInfoRef);

	Ref<MutexesInfo> mutexesInfoRef = memoryState->allocExpanding<MutexesInfo>();
	Ref<ConditionsInfo> conditionsInfoRef = memoryState->allocExpanding<ConditionsInfo>();

    Ref<ThreadPoolInfo> threadPoolInfoRef = memoryState->allocExpanding<ThreadPoolInfo>();
    memoryState->store (threadPoolInfoRef, &ThreadPoolInfo::threadCount, (size_t) 1);
    memoryState->store (threadPoolInfoRef, &ThreadPoolInfo::threads, threadsInfoRef);
    memoryState->store (threadPoolInfoRef, &ThreadPoolInfo::mutexes, mutexesInfoRef);
    memoryState->store (threadPoolInfoRef, &ThreadPoolInfo::conditions, conditionsInfoRef);

    Ref<ProgramInfo> progInfoRef = memoryState->alloc<ProgramInfo> ();
    memoryState->store (progInfoRef, &ProgramInfo::stepCount, 0);
    memoryState->store (progInfoRef, &ProgramInfo::threadPool, threadPoolInfoRef);
    memoryState->store (progInfoRef, &ProgramInfo::roAreaPtr, cstManager->getRoAreaPtr ());
    memoryState->store (progInfoRef, &ProgramInfo::globalsAreaPtr, globalsManager->getGlobAreaPtr());
    memoryState->setRootArea (progInfoRef);
    
    return progInfoRef;
}

Program::Program (Gimple* gimple, MemoryState* memoryState,
				  ConstantsManager* cstManager, GlobalsManager* globalsManager,
                  Ref<ProgramInfo> progInfoRef, RunningMode mode)
    : m_gimple (gimple), m_memoryState (memoryState), 
      m_cstManager (cstManager), m_globalsManager(globalsManager),
      m_progInfoRef (progInfoRef),
      m_threads(memoryState->load(progInfoRef, &ProgramInfo::threadPool), memoryState),
      m_mode(mode),
      m_stepCount(memoryState->load (progInfoRef, &ProgramInfo::stepCount))
{
		m_threads.newThread(m_gimple, m_memoryState, m_cstManager, m_globalsManager,
							this, getMainThread_ (), MAIN_THREAD_ID);

		m_threads.setMainThreadID(MAIN_THREAD_ID);
		m_threads.setCurrentThreadID(MAIN_THREAD_ID);

		m_threads.addThreadName(MAIN_THREAD_ID, MAIN_THREAD_NAME);
}

void
Program::reset()
{
	Ref<ThreadPoolInfo> tpi = m_memoryState->load(m_progInfoRef, &ProgramInfo::threadPool);
	m_stepCount = m_memoryState->load (m_progInfoRef, &ProgramInfo::stepCount);

	m_threads.reset(tpi, m_gimple, m_memoryState, m_cstManager, m_globalsManager, this);
}

void
Program::step (InstructionLogger* iLogger)
{
	bool programFinished = isFinished();

	if (programFinished)
		return;

	size_t sleepingThreadCount = m_threads.getThreadCount(THREAD_SLEEP);
	size_t runningThreadCount = m_threads.getThreadCount(THREAD_RUNNING);

	if ( sleepingThreadCount != 0 && runningThreadCount == 0)
	{
		unsigned int interval = m_threads.getMinRemainingSleepInterval();

		sleep(interval);
	}

	Thread* currentThread = m_threads.getCurrentThread();

	if (currentThread == NULL || !currentThread->isRunnable())
	{
		m_threads.pickAndSetRunnableThread();
		currentThread = m_threads.getCurrentThread();
	}

	try
	{
		if (iLogger != NULL)
			iLogger->push(m_threads.getThreadName(currentThread->getID()),
						  currentThread->getCurrentStmt());

		currentThread->step();

		if (currentThread->isFinished())
		{
			m_threads.terminateThread(currentThread->getID());
		}
	}
	catch (ThreadExitException& e)
	{
		m_threads.terminateThread(e.getId());
	}
}

void
Program::simpleStep(InstructionLogger* iLogger)
{
	Thread* currentThread = m_threads.getCurrentThread();
	assert(currentThread != NULL && currentThread->isRunnable() && !currentThread->isFinished());

	try
	{
		if (iLogger != NULL)
			iLogger->push(m_threads.getThreadName(currentThread->getID()),
						  currentThread->getCurrentStmt());

		currentThread->step();
	}
	catch (ThreadExitException&)
	{
		currentThread->setState(THREAD_STOPPED);
	}
	catch (LastThreadExitException&)
	{
		currentThread->setState(THREAD_STOPPED);
	}
}

void
Program::flush () 
{
	m_threads.flush();
	m_memoryState->store(m_progInfoRef, &ProgramInfo::stepCount, m_stepCount);
}


bool 
Program::isFinished () const
{
	//return m_threads.getMainThread()->isFinished();
	return m_threads.getUnfinishedThreadCount() == 0;
}

bool 
Program::aboutToReadInput () const
{
	return m_threads.getCurrentThread()->aboutToReadInput();
}

bool
Program::atCondStmt () const
{
	return m_threads.getCurrentThread()->atCondStmt();
}

/**
 * Gets program return value.
 */
const Value* 
Program::getReturnValue () const
{
	if (m_threads.getMainThreadID() == Thread::EXITED_ID) // main thread was terminated;
		return NULL;
	else
		return m_threads.getMainThread()->getReturnValue();
}

Thread::ID
Program::createNewThread(const FunctionValue* funcVal, Ptr funcFrame)
{
	assert(!funcFrame.isNull());

	Ref<ThreadInfo> threadInfoRef = Thread::create (funcVal, m_memoryState, funcFrame);
	Thread::ID newThreadID = m_threads.newThread(m_gimple, m_memoryState, m_cstManager,
												 m_globalsManager, this, threadInfoRef);

	return newThreadID;
}

int
Program::createNewMutex(Mutex::ID& mutexID)
{
	Ref<MutexInfo> mutexInfoRef = Mutex::create(m_memoryState);
	int ret = m_threads.newMutex(mutexID, mutexInfoRef);

	if (ret == 0)
		m_threads.initMutex(mutexID);

	return ret;
}

int
Program::destroyMutex(unsigned int mutexID)
{
	return m_threads.destroyMutex(mutexID);
}

int
Program::lockMutex(unsigned int mutexID)
{
	return m_threads.lockMutex(mutexID);
}

int
Program::tryLockMutex(unsigned int mutexID)
{
	return m_threads.tryLockMutex(mutexID);
}

int
Program::unlockMutex(unsigned int mutexID)
{
	return m_threads.unlockMutex(mutexID);
}

int
Program::createNewCondition(Condition::ID& condID)
{
	Ref<ConditionInfo> conditionInfoRef = Condition::create(m_memoryState);
	int ret = m_threads.newCondition(condID, conditionInfoRef);

	if (ret == 0)
		m_threads.initCondition(condID);

	return ret;
}

int
Program::destroyCondition(unsigned int condID)
{
	return m_threads.destroyCondition(condID);
}

int
Program::signalCondition(unsigned int condID)
{
	return m_threads.signalCondition(condID);
}

int
Program::waitCondition(unsigned int condID, unsigned int mutexID)
{
	return m_threads.waitCondition(condID, mutexID);
}

size_t
Program::getNThreads() const
{
	return m_threads.getThreadCount();
}

ThreadPool::ThreadsNonOwning
Program::runnableThreads()
{
	return m_threads.threads(THREAD_RUNNING);
}

ThreadPool*
Program::getThreadPool()
{
	return &m_threads;
}

Thread*
Program::getCurrentThread()
{
	return m_threads.getCurrentThread();
}

void
Program::setCurrentThread(Thread::ID threadID)
{
	m_threads.setCurrentThreadID(threadID);
}

int Program::getStepCount() const
{
	return m_stepCount;
}

void
Program::addThreadName(Thread::ID threadID, std::string name)
{
	m_threads.addThreadName(threadID, name);
}

RunningMode
Program::getRunningMode() const
{
	return m_mode;
}


Ref<ThreadInfo>
Program::getThread_ (size_t index) const
{
	Ref<ThreadPoolInfo> tpi = m_memoryState->load(m_progInfoRef, &ProgramInfo::threadPool);
	Ref<ThreadsInfo> ti = m_memoryState->load(tpi, &ThreadPoolInfo::threads);

	return m_memoryState->loadIndexed(ti, &ThreadsInfo::threads, index);
}

Ref<ThreadInfo>
Program::getMainThread_ () const
{
	Ref<ThreadPoolInfo> tpi = m_memoryState->load(m_progInfoRef, &ProgramInfo::threadPool);
	Ref<ThreadsInfo> ti = m_memoryState->load(tpi, &ThreadPoolInfo::threads);

	return m_memoryState->loadIndexed(ti, &ThreadsInfo::threads, MAIN_THREAD_ID);
}

}
