// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file Program.h
 * Declaration of class that interprets program.
 */
#ifndef INT_PROGRAM_H
#define INT_PROGRAM_H

#include "../mm/MemoryState.h"
#include "../git/git_fwd.h"

#include "ThreadPool.h"
#include "RunningMode.h"

#include <boost/noncopyable.hpp>

namespace ginterpreter {

class ProgramInfo;
class Value;
class ConstantsManager;
class GlobalsManager;
class InstructionLogger;


/**
 * Class that has access to all information required
 * to interpret or model check program.
 *
 * Class @c Program serves for interpreting the simulated program.
 * It stores information about all threads in the simulated program.
 * Calling method @c step() results in the interpretation of a single
 * GIMPLE instruction, which is done by calling @c step() on a thread
 * currently selected to run (i.e. the main thread at the beginning).
 * Afterwards, by calling method @c flush(), the current state of the
 * memory can be stored into simulated memory. This state can later
 * be restored when needed during model checking by calling @c reset().
 * Threads of the program are managed by an instance of @c ThreadPool class.
 * Class @c Program holds a pointer to the structure that represents the program
 * in the simulated memory. It also holds pointers to @c ConstantsManager,
 * @c GlobalsManager classes, which manage constants and globals of the program,
 * respectively. Finally it holds pointer to @c mmodule::MemoryState,
 * which is an interface into the simulated memory.
 *
 * @see ThreadPool, Thread, mmodule::MemoryState, ProgramInfo,
 * ConstantsManager, GlobalsManager
 */
class Program : boost::noncopyable {
  private:
    /*! GIMPLE++ representation of the program */
    giterator::Gimple* m_gimple;
    /*! Interface to the simulated memory */
    mmodule::MemoryState* m_memoryState;
    /*! Constants and function declarations of the program */
    ConstantsManager* m_cstManager;
    /*! Global declarations of the program */
	GlobalsManager* m_globalsManager;
    /*! Pointer to the simulated memory where the program is represented. */
    mmodule::Ref<ProgramInfo> m_progInfoRef;
    /*! Manager of the threads of the program */
    ThreadPool m_threads;
    /*! Mode in which this class is used. */
    RunningMode m_mode;
    /*! Counter of steps of the program already executed */
    int m_stepCount;
  public:
    /**
     * Function initializing program representation to be stored in the simulated memory.
     * Main thread ID and current thread ID remain unset,
     * they are set in program constructor.
     *
     * @param gimple GIMPLE++ representation of the program
     * @param memoryState interface to the simulated memory
     * @param cstManager where the constants of the program are stored
     * @param globalsManager where the global declarations of the program are stored
     * @return initialized structure representing program in the simulated memory
     */
    static mmodule::Ref<ProgramInfo> init (giterator::Gimple* gimple, 
                                           mmodule::MemoryState* memoryState, 
                                           ConstantsManager* cstManager,
                                           GlobalsManager* globalsManager);
    
    /** Constructor, creates a new program.
     * Before creating instance of program, Program::init() should be called
     * to initialize program representation in the simulated memory.
     * This representation is then passed to this constructor as an argument.
     * After program is constructed, ThreadPool contains single, main thread
     * of the program ready to be interpreted.
     *
     * @param gimple GIMPLE++ representation of the program
     * @param memoryState interface to the simulated memory
     * @param cstManager where the constants of the program are stored
     * @param globalsManager where the global declarations of the program are stored
     * @param progInfoRef structure representing program in the simulated memory
     * (initialized by Program::init())
     * @param mode mode in which the program will be used
     *
     * @see Program::init()
     */
    Program (giterator::Gimple* gimple, mmodule::MemoryState* memoryState, 
             ConstantsManager* cstManager, GlobalsManager* globalsManager,
             mmodule::Ref<ProgramInfo> progInfoRef, RunningMode mode);

    /**
     * Reset all the program memory structures according to current content
     * of the simulated memory.
     * To be used after backtracking, when previous states of the program
	 * get reset and the program structures are no more actual.
     * All the information from ProgramInfo structure for this class will
     * be read and the program data structures will be updated.
     *
     * @see mmodule::MemoryState::backtrack()
     */
    void reset();

    /**
     * Perform a single step of the program.
     * To be used in interpreter mode, simpleStep() is there
     * for model checking mode.
     * This function manages the selection of another thread to run
     * if current thread is already finished or suspended.
     * If current thread is already finished, it is terminated.
     * Calling this function repeatedly results in interpretation of
     * the whole program, as it manages the scheduling policy.
     *
     * @param iLogger where to write information about the stepped
     * instruction
     *
     * @see void simpleStep(InstructionLogger* iLogger);
     */
    void step (InstructionLogger* iLogger);

    /**
	 * Perform a 'dumb' single step of the program.
	 * To be used in model checking mode, step() is there
	 * for interpreter mode.
	 * This function does not manage the selection of another
	 * thread to run if current thread is already finished or suspended,
	 * neither it does manage terminition of finished thread.
	 * This is done by rescheduler in model checking mode.
	 * Current thread has to be set, unfinished and runnable.
	 *
	 * @param iLogger where to write information about the stepped
	 * instruction
	 *
	 * @see void step(InstructionLogger* iLogger);
	 */
    void simpleStep(InstructionLogger* iLogger);

    /**
     * Flush all the program memory structures into the simulated
     * memory.
     */
    void flush ();

    /**
     * Check whether the program is finished.
     * Program is finished when all of the threads are finished.
     *
     * @return true if the program si finished, false otherwise
     *
     * @see Thread::isFinished()
     */
    bool isFinished () const;

    /**
     * Checks if the program is about to read input from the user.
     * In GMC used to find out whether program is about to call
     * scanf() function.
     * For backtracking purposes, not used at the moment.
     *
     * @see atCondStmt()
     */
    bool aboutToReadInput () const;

    /**
	 * Find out whether program is about to interpret if statement.
	 * For backtracking purposes, not used at the moment.
	 *
	 * @see aboutToReadInput()
	 */
    bool atCondStmt () const;

    /**
     * Obtain the return value of the program -
     * the value returned from main thread.
     * @pre program must be finished.
     *
     * @returns value returned from main thread
     */
    const Value* getReturnValue () const;
    
    /**
     * Wrapper function for creating threads inside ThreadPool class.
     *
     * @param funcVal the function where the thread execution begins
     * @param funcFrame stack frame of the @c funcVal function
     *
     * @see ThreadPool::newThread()
     */
    Thread::ID createNewThread(const FunctionValue* funcVal, mmodule::Ptr funcFrame);

    /** @name Wrapper functions for ThreadPool functions
     * Functions for manipulating mutexes and conditions.
     */
    /*@{*/
    int createNewMutex(Mutex::ID& mutexID);
    int destroyMutex(unsigned int mutexID);
    int lockMutex(unsigned int mutexID);
    int tryLockMutex(unsigned int mutexID);
    int unlockMutex(unsigned int mutexID);

    int createNewCondition(Condition::ID& condID);
    int destroyCondition(unsigned int condID);
    int signalCondition(unsigned int condID);
    int waitCondition(unsigned int condID, unsigned int mutexID);
    /*@}*/

    /**
     * Get number of threads of the program.
     * Threads that are not finished.
     *
     * @return number of threas of the program
     *
     * @see ThreadPool::getThreadCount()
     */
	size_t getNThreads () const;

	/**
	 * Get a list of runnable threads of the program.
	 *
	 * @return runnable threads of the program
	 *
	 * @see Thread::isRunnable()
	 */
	ThreadPool::ThreadsNonOwning runnableThreads();

	/**
	 * Get a pointer to ThreadPool of the program
	 *
	 * @return ThreadPool of the program
	 */
	ThreadPool* getThreadPool();

	/**
	 * Get the thread that is currently selected to run.
	 *
	 * @return the thread that is currently selected to run.
	 */
	Thread* getCurrentThread();

	/**
	 * Set the thread to be currently selected to run.
	 *
	 * @param threadID ID of the thread to be currently selected to run.
	 */
	void setCurrentThread(Thread::ID threadID);

	/**
	 * Get the count of steps of the program already executed.
	 *
	 * @return  count of steps of the program already executed
	 */
	int getStepCount() const;

	/**
	 * Save a name of a thread with specified ID. It can be
	 * later overwritten. Used for output purposes
	 *
	 * @param threadID ID of the thread to set the name
	 * @param name name of the thread with ID @c threadID
	 */
	void addThreadName(Thread::ID threadID, std::string name);

	/**
	 * Get the mode in which this class is used.
	 * Used at the moment only beceause of different behavior
	 * of function @c sleep() in both modes.
	 *
	 * @return the mode in which this class is used
	 */
	RunningMode getRunningMode() const;
  private:
	/**
	 * Get a simulated memory representation of a thread with specified
	 * index in the array of thread representations.
	 * @deprecated
	 *
	 * @param index of the thread to get a memory representation of
	 * @return simulated memory representation of a thread with index @c index
	 */
	mmodule::Ref<ThreadInfo> getThread_ (size_t index) const;

	/**
	 * Get a simulated memory representation of the main thread.
	 *
	 * @return simulated memory representation of the main thread
	 */
    mmodule::Ref<ThreadInfo> getMainThread_ () const;
};

}

#endif /* INT_PROGRAM_H */

