// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file PtrValue.h
 * Declaration of class representing simulated pointer value.
 */
#ifndef INT_PTR_VALUE_H
#define INT_PTR_VALUE_H

#include "AbstractValue.h"

#include "../mm/Ptr.h"

namespace ginterpreter {

class IntegerValue;
    
/**
 * Class representing a simulated pointer value.
 * It stores the pointer represented by this value.
 */
class PtrValue : public AbstractPtrValue {
  private:
    /*! pointer represented by this value */
    mmodule::Ptr m_ptr;
  public:
    /**
     * Constructor.
     * 
     * @param type type of the value
     * @param ptr pointer that the value represents
     */
    PtrValue(const giterator::Type* type, mmodule::Ptr ptr);

    /**
	 * Clone this value.
	 *
	 * @returns an exact copy of this value.
	 */
    virtual Value* clone() const;

    /**
     * Returns the pointer represented by this value.
     *
     * @return the pointer represented by this value
     */
    mmodule::Ptr getPtr() const;

    /**
     * Returns the pointer type of this value.
     *
     * @return the pointer type of this value
     */
    const giterator::PtrType* getPtrType() const;

    /**
	 * @name Arithmetic operators
	 * The arithmetic operators on the value.
	 */
	//@{
    std::auto_ptr<Value> addOp (const Value& v) const;
    //@}

    /** @name Relation operators
	 * The relation operators on the value.
	 */
	//@{
    bool operator< (const Value& v) const;
    bool operator== (const Value& v) const;
    //@}

    /**
     * Whether this is pointer value (always true).
     *
     * @return whether this is pointer value (true).
     */
    bool isPtr () const;

    /**
	 * Print the value to a stream.
	 *
	 * @param os stream to print the value to
	 */
    void print (std::ostream& os) const;
  private:
    /**
     * Checked casts of @c Value& to the @c IntegerValue&.
     *
     * @param v the value to cast
     * @return the casted value
     */
    const IntegerValue& checkedIntegerCast (const Value& v) const;
    
    /**
     * Checked cast of @c Value& to the @c PtrValue&.
     *
     * @param v the value to cast
     * @return the casted value
     */
    const PtrValue& checkedPtrCast (const Value& v) const;
};

}

#endif /* INT_PTR_VALUE_H */

