// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file RealValue.cpp
 * Definitions for class representing simulated real value.
 */
#include "RealValue.h"
#include "IntegerValue.h"
#include "BoolValue.h"

#include "../git/RealType.h"

#include <memory>
#include <typeinfo>
#include <ostream>
#include <boost/functional/hash.hpp>

namespace ginterpreter
{

RealValue::RealValue (const giterator::Type* type, long double value)
	: AbstractNonPtrValue (type), m_value (value)
{}

Value* RealValue::clone () const
{
  return new RealValue (*this);
}

mmodule::Hash RealValue::getHash () const
{
	return boost::hash<long double>() (m_value);
}


std::auto_ptr<Value> RealValue::addOp (const Value& v) const
{
  const RealValue& iv = checkedCast (v);
  return std::auto_ptr<Value> (new RealValue (getType(), m_value + iv.m_value));
}

std::auto_ptr<Value> RealValue::subOp (const Value& v) const
{
  const RealValue& iv = checkedCast (v);
  return std::auto_ptr<Value> (new RealValue (getType (), m_value - iv.m_value));
}

std::auto_ptr<Value> RealValue::divOp (const Value& v) const
{
  const RealValue& iv = checkedCast (v);
  return std::auto_ptr<Value> (new RealValue (getType (), m_value / iv.m_value));
}

std::auto_ptr<Value> RealValue::rDivOp (const Value& v) const
{
	const RealValue& iv = checkedCast (v);
	return std::auto_ptr<Value> (new RealValue (getType (), m_value / iv.m_value));
}

std::auto_ptr<Value> RealValue::multOp (const Value& v) const
{
  const RealValue& iv = checkedCast (v);
  return std::auto_ptr<Value> (new RealValue (getType (), m_value * iv.m_value));
}

std::auto_ptr<Value> RealValue::absOp () const
{
	if (m_value < 0)
		return std::auto_ptr<Value> (new RealValue (getType (), -m_value));
	else
		return std::auto_ptr<Value> (new RealValue (getType (), m_value ));
}

std::auto_ptr<Value> RealValue::negateOp () const
{
	return std::auto_ptr<Value> (new RealValue(getType(), !m_value));
}

bool RealValue::operator== (const Value& v) const
{
	checkType (v);

	if (typeid(v) == typeid(BoolValue))
	{
		const BoolValue& bv = dynamic_cast<const BoolValue&> (v);
		return m_value == bv.getValue();
	}
	else if (typeid(v) == typeid(IntegerValue))
	{
		const IntegerValue& iv = dynamic_cast<const IntegerValue&> (v);
		return m_value == iv.getValue();
	}
	else if (typeid(v) == typeid(RealValue))
	{
		const RealValue& rv = dynamic_cast<const RealValue&> (v);

		if (this->isNan() || rv.isNan())
			  return false;

		return m_value == rv.getValue();
	}
	else
	{
		//TODO:
		return false;
	}
}

bool RealValue::operator< (const Value& v) const
{
  checkType (v);
  const RealValue& rv = dynamic_cast<const RealValue&> (v);

  if (this->isNan() || rv.isNan())
	  return false;

  return m_value < rv.m_value;
}

bool RealValue::orderedWith(const Value& v) const
{
	checkType (v);
	const RealValue& rv = dynamic_cast<const RealValue&> (v);

	if (this->isNan() || rv.isNan())
		return false;
	else
		return true;
}

void RealValue::print (std::ostream& os) const
{
	os << m_value;
}


long double RealValue::getValue () const
{
  return m_value;
}

const giterator::RealType* RealValue::getRealType () const
{
	assert (typeid (*getType ()) == typeid (giterator::RealType));
	return static_cast<const giterator::RealType*> (getType ());
}

bool RealValue::isZero () const
{
	return m_value == 0;
}

void RealValue::checkType (const Value& r) const
{
	assert (getType()->isSame(*r.getType()));     (void)r;
}

const RealValue& RealValue::checkedCast (const Value& v) const
{
	checkType (v);
	return static_cast<const RealValue&> (v);
}

bool RealValue::isNan() const
{
	volatile long double d = m_value;
	return d != d;
}

}
