// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file RealValue.h
 * Declaration of class representing simulated real value.
 */
#ifndef INT_REAL_VALUE_H
#define INT_REAL_VALUE_H

#include "AbstractValue.h"

namespace ginterpreter {

/**
 * Class representing simulated real value.
 * The value is stored as a long dobule,
 * it is therefore capable of storing all the
 * real types.
 */
class RealValue : public AbstractNonPtrValue {
  private:
    /*! the value of this simulated value */
    long double m_value;

  public:
    /**
     * Creates new simulated value.
     *
     * @param type type of the simulated value
     * @param value value of the simulated value
     */
    RealValue (const giterator::Type* type, long double value);

    /**
	 * Clone this value.
	 *
	 * @returns an exact copy of this value.
	 */
    Value* clone () const;

    /**
	 * Get a hash of this value.
	 * For memory state equivalency check purposes.
	 *
	 * @return hash of this value.
	 */
    mmodule::Hash getHash () const;

    /**
	 * @name Arithmetic and logical operators
	 * The arithmetic and logical operators on the value.
	 */
	//@{
    std::auto_ptr<Value> addOp (const Value& v) const;
    std::auto_ptr<Value> subOp (const Value& v) const;
    std::auto_ptr<Value> divOp (const Value& v) const;
    std::auto_ptr<Value> rDivOp (const Value& v) const;
    std::auto_ptr<Value> multOp (const Value& v) const;
    std::auto_ptr<Value> absOp () const;
    std::auto_ptr<Value> negateOp () const;
    //@}

    /** @name Relation operators
	 * The relation operators on the value.
	 */
	//@{
    bool operator== (const Value& v) const;
    bool operator< (const Value& v) const;
    bool orderedWith(const Value& v) const;
    //@}

    /**
	 * Print the value to a stream.
	 *
	 * @param os stream to print the value to
	 */
    void print (std::ostream& os) const;

    /**
     * Returns the raw value of this simulated value.
     *
     * @return the raw value of this simulated value
     */
    long double getValue () const;

    /**
     * Returns the real type of this simulated value.
     *
     * @return the real type of this simulated value
     */
    const giterator::RealType* getRealType () const;

    /**
	 * Checks if this simulated value is zero.
	 *
	 * @return true if this value is zero,
	 * false otherwise
	 */
	bool isZero () const;

	/**
	 * Check if the simulated value is an IEEE NaN (Not a Number).
	 *
	 * @return true if the value is NaN, false otherwise
	 */
	bool isNan() const;
  protected:
	/**
	* Checks if the type of value @c i and the type of this value are the same.
	*
	* @param r the value to check if the type is the same
	*
	* @see giterator::RealType::isSame()
	*/
	void checkType (const Value& r) const;

	/**
	* Casts the reference to the value @c v to the reference to the real value
	* and checks, if the value @c v is of the same type as this value.
	*
	* @param v the value to cast
	* @return the casted reference
	*/
	const RealValue& checkedCast (const Value& v) const;
};

}

#endif /* INT_REAL_VALUE_H */
