// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file RecordConstructorValue.h
 * Declaration of class representing simulated record constructor.
 */
#ifndef INT_RECORD_CONSTRUCTOR_VALUE_H
#define INT_RECORD_CONSTRUCTOR_VALUE_H

#include "AbstractValue.h"

#include <boost/ptr_container/ptr_map.hpp>
#include <memory>


namespace ginterpreter {

/**
 * Class representing a simulated record constructor.
 * This class is a regular simulated value, so it can be stored in the simulated memory.
 * In GMC however, RecordConstructorValue is not stored to the simulated memory.
 * It is only a helper class that is returned from @c ValueReader
 * when reading an @c RecordConstructorOperand. It is composed
 * of the values of the constructors for the fields of the record.
 * @c ValueWriter decomposes it back into the constructors for the fields
 * and stores them into simulated memory one by one.
 */
class RecordConstructorValue : public AbstractNonPtrValue {
  private:
	/*! type of the container holding the values of constructors for fields of a record. */
    typedef boost::ptr_map<const giterator::FieldDecl*, Value> Values;
    /*!  values of constructors for fields of a record. */
    Values m_values;
  public:
    /*! iterator for the constructors values container */
    typedef Values::const_iterator ValuesIterator;
    /**
     * Constructor.
     *
     * @param type GIMPLE++ representation of the type of the constructed
     * record
     */
    RecordConstructorValue (const giterator::Type* type);

    /**
	 * Clone this value.
	 *
	 * @returns an exact copy of this value.
	 */
    Value* clone () const;

    /**
	 * Get a hash of this value.
	 * The hash is combined of the hashes of the sub-values.
	 * For memory state equivalency check purposes.
	 *
	 * @return hash of this value.
	 */
    mmodule::Hash getHash () const;


    /**
	 * Add a value of a constructor for a field of a record to this value.
	 *
	 * @param field field of a record the added value constructs
	 * @param constructor @c constructor of the field @c field
	 * in a record
	 */
    void addValue(const giterator::FieldDecl* field, std::auto_ptr<Value> constructor);

    /**
	 * Obtain all the values of this constructor.
	 * i.e. value of constructors for all the fields
	 * of the structure.
	 *
	 * @return pair iterators <first value iterator,
	 * 						   last value iterator>
	 */
    std::pair<ValuesIterator, ValuesIterator> values() const;
};

}

#endif /* INT_RECORD_CONSTRUCTOR_VALUE_H */
