// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file RecordValue.h
 * Declaration of class representing a simulated struct/union.
 */
#ifndef INT_RECORD_VALUE_H
#define INT_RECORD_VALUE_H

#include "AbstractValue.h"

namespace ginterpreter {

/**
 * Class representing a simulated struct/union.
 * This class is a regular simulated value, so it can be stored in the simulated memory.
 * In GMC however, RecordValue is not stored in the simulated memory.
 * It is only a helper class that is returned from @c ValueReader.
 * It holds the pointer to the beginning of a memory area where fields
 * of a record are stored as well as the size of the record
 * and the offset of the record.
 * @c ValueWriter reads all the values overlapping over area <ptr, ptr+size>
 * and stores them one by one to the simulated memory.
 */
class RecordValue : public AbstractPtrValue {
  private:
	/*! pointer to the beginning of the record fields. */
	mmodule::Ptr m_ptr;
	/*! size of the record */
	size_t m_size;
	/*! offset of the record */
	size_t m_offset;

  public:
    /**
     * Constructor.
     *
     * @param type GIMPLE++ representation of the type of the record
     * @param recordPtr pointer to the beginning of the fields of the record
     * @param unitSize size of the record starting at @c ptr
     * @param offset offset of the record
     */
	RecordValue(const giterator::Type* type, mmodule::Ptr recordPtr,
				size_t unitSize, size_t offset);

	/**
	 * Clone this value.
	 *
	 * @returns an exact copy of this value.
	 */
    Value* clone () const;

    /**
	 * Get a pointer to the beginning of the record fields.
	 *
	 * @return pointer to the beginning of the record fields
	 */
    mmodule::Ptr getPtr () const;

    /**
	 * Get a size of the record in bytes.
	 *
	 * @return size of the record
	 */
    size_t getUnitSize () const;

    /**
     * Get an offset of the record.
     *
     * @return an offset of the record
     */
    size_t getOffset() const;

    // mmodule::Hash getHash () const
};

}

#endif /* INT_RECORD_VALUE_H */
