// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file Rescheduler.cpp
 * Definitions for a class that reschedules thread execution order during modelchecking.
 */

#include "Rescheduler.h"
#include "Program.h"
#include "InstructionLogger.h"
#include "ModelCheckerExceptions.h"

namespace ginterpreter
{

using namespace giterator;
using namespace mmodule;

Rescheduler::Rescheduler (Gimple* gimple, MemoryState* memoryState,
				  	  	  ConstantsManager* cstManager, GlobalsManager* globalsManager,
				  	  	  Ref<ProgramInfo> progInfoRef)
: m_gimple (gimple), m_memoryState (memoryState),
  m_cstManager (cstManager), m_globalsManager(globalsManager), m_progInfoRef (progInfoRef)
{}

ReschedulerData Rescheduler::reschedule(std::auto_ptr<Program>& program,
										std::set<Hash>& visitedStates,
										std::stack<ThreadPool::RunnableThreads>& backtrackStack,
										InstructionLogger& iLogger,
										int& stepCount)
{
	ReschedulerData returnData(Thread::NO_ID, 0, true);
	ThreadPool::RunnableThreads runnableThreads;

	program->flush();
	m_memoryState->push();

	if (!checkReentry(program.get(), visitedStates))
	{
		runnableThreads = program->getThreadPool()->runnableThreads();

		if (checkDeadlock(program.get()))
			throw MCDeadlockException();
			//return ReschedulerData(-1, 0, false);
	}

	while(runnableThreads.size() == 0 && !backtrackStack.empty())
	{
		m_memoryState->pop ();

		runnableThreads = backtrackStack.top();
		backtrackStack.pop();
		//iLogger.pop();

		++returnData.backtrackCount;
	}

	if (returnData.backtrackCount > 0)
	{
		assert(m_memoryState->hasMarkedState());
		m_memoryState->backtrack ();
		program->reset();
		iLogger.resize(program->getStepCount());
	}

	returnData.continueExploration = returnData.continueExploration && !runnableThreads.empty();

	if (!runnableThreads.empty())
	{
		returnData.nextThread = runnableThreads.front();
		runnableThreads.pop_front();

		backtrackStack.push(runnableThreads);
	}

	return returnData;
}

bool
Rescheduler::checkReentry (Program* program, std::set<Hash>& visitedStates) //const
{
	Hash hash = m_memoryState->hash ();

	bool added = visitedStates.insert (hash).second;

	if (!added)
	{
		LOG_DEBUG("state reentered");
		return true;
	}

	return false;
}

bool
Rescheduler::checkDeadlock(Program* program) //const
{
	bool possible = false;

	//ThreadsConstIterator
	for (ThreadPool::ThreadsIterator i1 = program->getThreadPool()->threads().first;
			i1 != program->getThreadPool()->threads().second;
			++i1)
	{
		if (i1->second->isRunnable())
			return false;
		else
			possible = possible || !i1->second->isTerminated();
	}

	return possible;
}

}
