// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file Rescheduler.h
 * Declaration of a class that reschedules thread execution order during modelchecking.
 */

#ifndef INT_RESCHEDULER_H
#define INT_RESCHEDULER_H

#include "ThreadPool.h"

#include "../mm/MemoryState.h"
#include "../git/git_fwd.h"

#include <memory>
#include <set>
#include <stack>

namespace ginterpreter
{

class Program;
class ProgramInfo;
class Thread;
class ConstantsManager;
class GlobalsManager;
class InstructionLogger;

/**
 * Data returned from the rescheduler.
 *
 * @see Rescheduler::reschedule
 */
struct ReschedulerData
{
	/** Constructor */
	ReschedulerData() : nextThread(-1), backtrackCount(0), continueExploration(true) {}
	/** Constructor, initializing the structure to be returned from rescheduler. */
	ReschedulerData(Thread::ID next, int bc, bool ce)
	: nextThread(next), backtrackCount(bc), continueExploration(ce) {}

	/*! Thread selected to run in the next iteration of model checker. */
	Thread::ID nextThread;
	/*! Number of steps the rescheduler backtracked the program */
	int backtrackCount;
	/*! Whether to continue exploring the state space (there is more to explore) */
	bool continueExploration;
};

/**
 * Class managing the thread scheduling during model checking.
 *
 * The rescheduler @c reschedule() compares the current memory state with the
 * previously stored ones, and if it finds that it is in a previously visited state,
 * it backtracks to the previous state. Additionally, it may select a different
 * thread as the running thread of the program for the the next iteration.
 */
class Rescheduler
{
private:
    /*! GIMPLE++ representation of the program */
    giterator::Gimple* m_gimple;
    /*! Interface to the simulated memory */
    mmodule::MemoryState* m_memoryState;
    /*! Constants and function declarations of the program  */
    ConstantsManager* m_cstManager;
    /*! Global declarations of the program */
	GlobalsManager* m_globalsManager;
    /*! Pointer to the simulated memory where the simulated program is represented. */
    mmodule::Ref<ProgramInfo> m_progInfoRef;
public:
     /**
     * Constructor, creates a new rescheduler.
     *
     * @param gimple GIMPLE++ representation of the program
     * @param memoryState interface to the simulated memory
     * @param cstManager constants and function declarations of the program
     * @param globalsManager global declarations of the program
     * @param progInfoRef pointer to the simulated memory where the simulated program is represented.
     */
    Rescheduler (giterator::Gimple* gimple, mmodule::MemoryState* memoryState,
    	             ConstantsManager* cstManager, GlobalsManager* globalsManager,
    	             mmodule::Ref<ProgramInfo> progInfoRef);

    /**
     * This function compares the current memory state with the
     * previously stored ones, and if it finds that it is in a previously visited state,
     * it backtracks to the previous state.
     * It sets the thread to run in the next iteration of model checker.
     *
     * @param program the simulated program
     * @param visitedStates structure holding hashes of already visited states
     * @param backtrackStack stack holding runnable threads in states of program
     * @param iLogger where the information about the stepped instructions
     * is stored, it can be changed by rescheduler after backtracking
     * @param stepCount steps of program already excetued, it can be changed
     * by rescheduler after backtracking
     */
	ReschedulerData reschedule(std::auto_ptr<Program>& program,
							   std::set<mmodule::Hash>& visitedStates,
							   std::stack<ThreadPool::RunnableThreads>& backtrackStack,
							   InstructionLogger& iLogger,
							   int& stepCount);
private:
	/**
	 * Checks if program is at a state that has been visited before.
	 * If it is, prints a message on the standard error output.
	 *
	 * @return @c true if program already visited the current state, @c false otherwise
	 */
	bool checkReentry (Program* program, std::set<mmodule::Hash>& visitedStates); //const;

	/**
	 * Check for deadlocks in the current state of program.
	 * Deadlock is detected when not all the threads are finished
	 * but no thread is runnable.
	 *
	 * @param program the simulated program
	 * @throws MCDeadlockException
	 */
	bool checkDeadlock(Program* program);// const;
};
}


#endif /* INT_RESCHEDULER_H */
