// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file Steps.h
 * All classes representing instruction pointer steps.
 */
#ifndef INT_STEPS_H
#define INT_STEPS_H

#include "../mm/Ptr.h"

//OND
namespace giterator
{
class Operand;
}

namespace ginterpreter {
    
class NextStep;
class ReturnStep;
class CondStep;
class CallStep;
class JumpStep;
class FunctionValue;


/**
 * Class representing an instruction pointer step.
 */
struct Step {
	/**
	 * Visitor for sub-classes of @c Step class,
	 * that are meant to be used directly by the user.
	 */
    struct Visitor {
        virtual ~Visitor () {}
        
        virtual void visit (const NextStep&) = 0;
        virtual void visit (const ReturnStep&) = 0;
        virtual void visit (const CondStep&) = 0;
        virtual void visit (const CallStep&) = 0;
        virtual void visit (const JumpStep&) = 0;
    };
    
    /**
     * Destructor
     */
    virtual ~Step () {}
    
    /**
	 * Calls @c visitor.accept().
	 * Classes implementing this interface implement this method to call @c visitor.accept()
	 * with @c *this as a parameter.
	 *
	 * The parameter is visitor on which to call the @c accept method.
	 */
    virtual void accept (Visitor&) const = 0;
};


/**
 * Class representing a step to the next statement.
 */
struct NextStep : public Step {
    void accept (Visitor& v) const
    { v.visit (*this); }
};


/**
 * Class representing a step that returns from the current function.
 */
struct ReturnStep : public Step {
    void accept (Visitor& v) const
    { v.visit (*this); }
};


/**
 * Class representing a conditional step.
 */
struct CondStep : public Step {
    /**
     * Constructor.
     * 
     * @param condResult result of the condition
     */
    CondStep (bool condResult) : m_condResult (condResult) {}
    
    void accept (Visitor& v) const
    { v.visit (*this); }
    
    bool getCondResult () const
    { return m_condResult; }
    
  private:
    bool m_condResult;
};


/**
 * Class representing a call step.
 */
struct CallStep : public Step {
    /**
     * Constructor.
     * 
     * @param func function that should be called
     * @param funcLocalData pointer to the beginning of the stack frame of the function
     *      that should be called
     */
    CallStep (const FunctionValue* func, mmodule::Ptr funcLocalData) : 
            m_func (func), m_funcLocalData (funcLocalData) {}
    
    void accept (Visitor& v) const
    { v.visit (*this); }
    
    const FunctionValue* getFunction () const
    { return m_func; }
    
    mmodule::Ptr getFuncLocalData () const
    { return m_funcLocalData; }
    
  private:
    const FunctionValue* m_func;
    mmodule::Ptr m_funcLocalData;
};

/**
 * Class representing a jump step.
 */
struct JumpStep : public Step {
	JumpStep () : m_dest(0) {}
	JumpStep(const giterator::Operand* dest) : m_dest(dest) {}
	void accept (Visitor& v) const
	{ v.visit (*this); }
	const giterator::Operand* getDest() const
	{ return m_dest; }
private:
	const giterator::Operand* m_dest;
};

}

#endif /* INT_STEPS_H */
