// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file Thread.cpp
 * Definitions for class that represents simulated thread.
 */
#include "Thread.h"

#include "Value.h"
#include "GimpleStmtInterpreter.h"
#include "FrameInfo.h"
#include "FunctionValue.h"
#include "BuiltinFunctionValue.h"
#include "PtrValue.h"
#include "ValueWriter.h"
#include "ValueReader.h"
#include "SymbolicPrinter.h"
#include "MemoryContentPrinter.h"
#include "InstructionSafetyResolver.h"

#include "../mm/MemoryState.h"
#include "../git/Gimple.h"
#include "../git/CallStmt.h"
#include "../git/CondStmt.h"

#include <iostream>

using namespace mmodule;
using namespace giterator;


namespace ginterpreter {

const Thread::ID Thread::NO_ID = -1;
const Thread::ID Thread::EXITED_ID = -2;
    
/**
 * Returns the top frame of a thread.
 * 
 * @param ms memory state
 * @param threadInfo the info area for the thread
 * @return reference to the infor area of the top frame stac of thread @c threadInfo
 */
static Ref<FrameInfo> 
getTopFrame_ (MemoryState* ms, Ref<ThreadInfo> threadInfo)
{
    Ref<FrameInfo> frame = ms->load (threadInfo, &ThreadInfo::bottomFrame);
    Ref<FrameInfo> oldFrame;
    do {
        oldFrame = frame;
        frame = ms->load (frame, &FrameInfo::nextFrame);
    } while (!frame.isNull ());
    
    return oldFrame;
}

Ref<ThreadInfo>
Thread::create (const FunctionValue* func, MemoryState* memoryState, Ptr frameData)
{
	if (frameData.isNull())
		frameData = GimpleStmtInterpreter::createFrame (memoryState, func);

    Ref<FrameInfo> frameInfoRef = memoryState->alloc<FrameInfo> ();
    memoryState->store (frameInfoRef, &FrameInfo::prevFrame, Ref<FrameInfo> ());
    memoryState->store (frameInfoRef, &FrameInfo::nextFrame, Ref<FrameInfo> ());
    memoryState->store (frameInfoRef, &FrameInfo::returnAddress, InstructionPtr ());
    memoryState->store (frameInfoRef, &FrameInfo::functionLocalData, frameData);

    Ref<ThreadInfo> threadInfoRef = memoryState->alloc<ThreadInfo> ();
    memoryState->store (threadInfoRef, &ThreadInfo::bottomFrame, frameInfoRef);
    memoryState->store (threadInfoRef, &ThreadInfo::currentIp, InstructionPtr (func));
    memoryState->store (threadInfoRef, &ThreadInfo::threadID, Thread::NO_ID);
    memoryState->store (threadInfoRef, &ThreadInfo::threadState, THREAD_RUNNING);
    memoryState->store (threadInfoRef, &ThreadInfo::waitingFor, Thread::NO_ID);
    memoryState->store (threadInfoRef, &ThreadInfo::isBeingJoined, false);

    return threadInfoRef;
}

Thread::Thread (Gimple* gimple, MemoryState* memoryState,
				ConstantsManager* cstManager, GlobalsManager* globalsManager,
		Program* program, Ref<ThreadInfo> threadInfoRef, unsigned long int ID)
    : m_gimple (gimple), m_memoryState (memoryState), 
      m_cstManager (cstManager), m_globalsManager(globalsManager),
      m_program (program)
{
    reset(threadInfoRef);
}

Thread::~Thread ()
{
    //flush ();
}

void
Thread::reset(Ref<ThreadInfo> threadInfoRef)
{
	m_threadInfoRef = threadInfoRef;
	m_topFrameRef = getTopFrame_ (m_memoryState, m_threadInfoRef);
	m_topFrameLocalData = m_memoryState->load (m_topFrameRef, &FrameInfo::functionLocalData);
	m_currentIp = m_memoryState->load (m_threadInfoRef, &ThreadInfo::currentIp);
	m_ID = m_memoryState->load(m_threadInfoRef, &ThreadInfo::threadID);
	m_state = m_memoryState->load(m_threadInfoRef, &ThreadInfo::threadState);
	m_waitingFor = m_memoryState->load(m_threadInfoRef, &ThreadInfo::waitingFor);
	m_isBeingJoined = m_memoryState->load(m_threadInfoRef, &ThreadInfo::isBeingJoined);
}

void
Thread::step ()
{
#ifndef NDEBUG
    SymbolicPrinter sPrinter (std::cerr);
    MemoryContentPrinter mcPrinter (std::cerr, m_currentIp.getFunction (), m_memoryState,
                                    m_cstManager, m_globalsManager, m_topFrameLocalData);
    std::cerr << "interpreting: ";
    m_currentIp.getStmt ()->accept (sPrinter);
    std::cerr << "\t (";
    //m_currentIp.getStmt ()->accept (mcPrinter);
    std::cerr << ')' << std::endl;
#endif
    
    GimpleStmtInterpreter interpreter (m_gimple, m_currentIp.getFunction(),
                                       m_memoryState, m_cstManager, m_globalsManager,
                                       m_program, m_topFrameLocalData);
    m_currentIp.getStmt ()->accept (interpreter);
    
    interpreter.acquireStep ()->accept (*this);
}

void
Thread::flush ()
{
    m_memoryState->store (m_threadInfoRef, &ThreadInfo::currentIp, m_currentIp);
    m_memoryState->store (m_threadInfoRef, &ThreadInfo::threadID, m_ID);
    m_memoryState->store (m_threadInfoRef, &ThreadInfo::threadState, m_state);
    m_memoryState->store (m_threadInfoRef, &ThreadInfo::waitingFor, m_waitingFor);
    m_memoryState->store (m_threadInfoRef, &ThreadInfo::isBeingJoined, m_isBeingJoined);
}

bool
Thread::aboutToReadInput () const
{
    const CallStmt* callStmt = dynamic_cast<const CallStmt*> (m_currentIp.getStmt ());
    if (!callStmt)
        return false;

    ValueReader reader (m_currentIp.getFunction (), m_memoryState,
    					m_cstManager, m_globalsManager,
                        m_topFrameLocalData);
    callStmt->getFunction ()->accept (reader);
    
    const PtrValue* funcPtr = dynamic_cast<const PtrValue*> (reader.getValue ());
    if (!funcPtr)
        return false;
    
    const BuiltinFunctionValue* biF = 
            dynamic_cast<const BuiltinFunctionValue*> (m_memoryState->load (funcPtr->getPtr ()));
    if (!biF)
        return false;
    
    return biF->getCode () == GCC_BUILT_IN_SCANF;
}

bool
Thread::atCondStmt () const
{
    return typeid (*m_currentIp.getStmt ()) == typeid (CondStmt);
}

const Value*
Thread::getReturnValue () const
{
    assert (isFinished ());

    if(m_topFrameLocalData.isNull())
    	return NULL;

    std::pair<MemoryState::ContentIterator, MemoryState::ContentIterator> ret = 
            m_memoryState->loadOverlapping (m_topFrameLocalData, 1);
    //assert (ret.first != ret.second);
    if (ret.first == ret.second)
    	return NULL;
    
    const Value* retVal = dynamic_cast<const Value*> (ret.first->second);
    assert (retVal);
    return retVal;
}

void
Thread::destroy ()
{    
    while (!m_topFrameRef.isNull ()) {
        m_memoryState->free (m_topFrameLocalData);
        
        Ref<FrameInfo> oldTopFrameRef = m_topFrameRef;
        m_topFrameRef = m_memoryState->load (oldTopFrameRef, &FrameInfo::prevFrame);
        m_memoryState->free (oldTopFrameRef);
        
        if (m_topFrameRef.isNull())
        	break;

        m_topFrameLocalData = m_memoryState->load (m_topFrameRef,
                                                   &FrameInfo::functionLocalData);
    }
}

bool
Thread::isFinished () const
{
	return m_currentIp.isNull ();
}

bool
Thread::isAlive() //const
{
	return  getState() != THREAD_STOPPED;
}

bool
Thread::isRunning() //const
{
	return getState() == THREAD_RUNNING;
}

bool
Thread::isRunnable() //const
{
	return  !isFinished() && getState() == THREAD_RUNNING;
}

bool
Thread::isTerminated() //const
{
	return getState() == THREAD_STOPPED;
}

void
Thread::awaken()
{
	m_state = THREAD_RUNNING;
	m_waitingFor = NO_ID;
}

void
Thread::setWaitingFor(ID threadID)
{
	assert(m_waitingFor == NO_ID);
	m_waitingFor = threadID;
}

Thread::ID
Thread::getWaitingFor() const
{
	return m_waitingFor;
}

ThreadState
Thread::getState()
{
	if (m_timer.isStarted() && m_timer.elapsed())
	{
		m_timer.clear();
		assert(m_state == THREAD_SLEEP);
		m_state = THREAD_RUNNING;
	}
	return m_state;
}

void
Thread::setState(ThreadState newState)
{
	m_state = newState;
}

Thread::ID
Thread::getID() const
{
	return m_ID;
}

void
Thread::sleep(unsigned int interval, RunningMode mode)
{
	if (mode == MODE_INTERPRETER)
	{
		m_timer.setInterval(interval);
		m_timer.start();
		setState(THREAD_SLEEP);
	}

	// Do nothing in model checker mode.
	// Sleep is considered only as a backtracking
	// spot, i.e. an equivalent of an unsafe
	// instruction
}

unsigned int
Thread::getRemainingTime() const
{
	return m_timer.getRemainingTime();
}

void
Thread::setIsBeingJoined(bool value)
{
	m_isBeingJoined = value;
}

bool
Thread::getIsBeingJoined() const
{
	return m_isBeingJoined;
}

void
Thread::returnFromFunction ()
{
    Ref<FrameInfo> prevFrame = m_memoryState->load (m_topFrameRef, &FrameInfo::prevFrame);
    InstructionPtr retIp = m_memoryState->load (m_topFrameRef, &FrameInfo::returnAddress);
    
    
    if (!retIp.isNull ())
    {
        // Copy return value
        Ptr prevFrameLocalData = m_memoryState->load (prevFrame,
                                                      &FrameInfo::functionLocalData);
        
        const CallStmt* callStmt = dynamic_cast<const CallStmt*> (retIp.getStmt ());
        if (callStmt->getLhs () && !prevFrame.isNull ())
        {
            const Value* val = dynamic_cast<const Value*> (m_memoryState->load (
                    m_topFrameLocalData +
                    m_currentIp.getFunction ()->getVarsMapper ()->getRetValOffset ()));
            
            std::auto_ptr<Value> value (val->clone ());
            ValueWriter writer (retIp.getFunction (), m_memoryState,
            					m_cstManager, m_globalsManager,
                                prevFrameLocalData, value);
            callStmt->getLhs ()->accept (writer);
        }
        
        if (!m_topFrameLocalData.isNull())
        	m_memoryState->free (m_topFrameLocalData);
        m_memoryState->free (m_topFrameRef);
        m_memoryState->store (prevFrame, &FrameInfo::nextFrame, Ref<FrameInfo> ());
        
        m_topFrameRef = prevFrame;
        m_topFrameLocalData = prevFrameLocalData;
    }

    m_currentIp = retIp;
}

void 
Thread::callFunction (const FunctionValue* func, Ptr frame)
{    
    Ref<FrameInfo> newTopFrameRef = m_memoryState->alloc<FrameInfo> ();
    m_memoryState->store (newTopFrameRef, &FrameInfo::prevFrame, m_topFrameRef);
    m_memoryState->store (newTopFrameRef, &FrameInfo::nextFrame, Ref<FrameInfo> ());
    m_memoryState->store (newTopFrameRef, &FrameInfo::returnAddress, m_currentIp);
    m_memoryState->store (newTopFrameRef, &FrameInfo::functionLocalData, frame);
    m_memoryState->store (m_topFrameRef, &FrameInfo::nextFrame, newTopFrameRef);
    
    m_currentIp = InstructionPtr (func);
    m_topFrameRef = newTopFrameRef;
    m_topFrameLocalData = frame;
}


void
Thread::visit (const NextStep&)
{
    m_currentIp.step ();
}
    
    
void
Thread::visit (const CondStep& step)
{
    m_currentIp.takeBranch (step.getCondResult ());
}
    
    
void 
Thread::visit (const ReturnStep&)
{
    returnFromFunction ();
        
    if (!m_currentIp.isNull ())
        m_currentIp.step ();
}
    
void 
Thread::visit (const CallStep& step)
{
    callFunction (step.getFunction (), step.getFuncLocalData ());
}

void
Thread::visit (const JumpStep& step)
{
	if (step.getDest() == 0)
		m_currentIp.jump ();
	else
		m_currentIp.jumpToLabel(step.getDest());
}

bool
Thread::isCurrentInstructionUnsafe() const
{
	InstructionSafetyResolver isr;
	m_currentIp.getStmt()->accept(isr);

	return isr.getUnsafe();
}

const Stmt*
Thread::getCurrentStmt() const
{
	return m_currentIp.getStmt();
}

}
