// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file Thread.h
 * Declaration of class that represents simulated thread.
 */
#ifndef INT_THREAD_H
#define INT_THREAD_H

#include "ThreadState.h"
#include "ThreadInfo.h"
#include "Steps.h"
#include "Value.h"
#include "Timer.h"
#include "RunningMode.h"

#include "../mm/MemoryState.h"
#include "../mm/Ptr.h"
#include "../git/git_fwd.h"

#include <boost/noncopyable.hpp>

namespace ginterpreter {
    
class Program;
class Value;
class FrameInfo;
class FunctionValue;
class ConstantsManager;
class GlobalsManager;

/**
 * Class that represents a simulated thread.
 *
 * Thread of execution can be defined as a unit of processing that
 * can be scheduled by an operating system (by GMC in this case).
 * It generally results from a fork of a program into two or
 * more concurrently running tasks. Threads share some resources
 * of the program (its instructions and global memory).
 *
 * Class @c Thread represents a single thread of the program.
 * Like class @c Program, it has methods @c flush(), @c reset()
 * and @c step(), the last one to interpret GIMPLE++ instructions.
 * This is done via an instance of @c GimpleStmtInterpreter class,
 * which is capable of executing GIMPLE++ instructions. During the
 * program execution, there are different states that thread may be in:
 * Running - currently selected to run.
 * Stopped - already terminated.
 * Wait for lock - waiting to acquire a lock.
 * Wait for condition - Waiting on a thread condition.
 * Sleep - sleeping (only in interpreting mode, not model checking mode).
 * Wait to join - waiting for another thread to terminate.
 *
 * Same as class @c Program, @c Thread holds a pointer to its representation
 * in the simulated memory.
 *
 * @see ThreadPool, ThreadInfo
 */
class Thread : public Step::Visitor, boost::noncopyable{
public:
	typedef int ID;
	static const ID NO_ID;
	static const ID EXITED_ID;
private:
	/*! GIMPLE++ representation of the program */
    giterator::Gimple* m_gimple;
    /*! Interface to the simulated memory */
    mmodule::MemoryState* m_memoryState;
    /*! Constants and function declarations of the program */
    ConstantsManager* m_cstManager;
    /*! Global declarations of the program */
    GlobalsManager* m_globalsManager;
    /*! Program that this thread is a part of */
    Program* m_program;
    /*! Pointer to the simulated memory where the thread is represented.*/
    mmodule::Ref<ThreadInfo> m_threadInfoRef;
    /*! Pointer to the simulated memory where the top stack frame of this thread is represented. */
    mmodule::Ref<FrameInfo> m_topFrameRef;
    /*! pointer to the beginning of the top stack frame of this thread */
    mmodule::Ptr m_topFrameLocalData;
    /*! current instruction pointer */
    InstructionPtr m_currentIp;
    /*! Unique ID of this thread */
    ID m_ID;
    /*! Current state of this thread */
    ThreadState m_state;
    /*! ID of a thread this thread is waiting for (or -1 if for no). */
    ID m_waitingFor;
    /*! Computer of thread sleep time period. */
    Timer m_timer;
    /*! Whether there is join called on this thread */
    bool m_isBeingJoined;

    public:
    /**
     * Function initializing thread representation to be stored in the simulated memory.
     *
     * @param func function where the execution of the thread begins
     * @param memoryState interface to the simulated memory
     * @param frameData pointer to the beginning of the top stack frame of this thread
     * @return initialized structure representing thread in the simulated memory
     */
    static mmodule::Ref<ThreadInfo> create (const FunctionValue* func,
                                            mmodule::MemoryState* memoryState,
                                            mmodule::Ptr frameData);
    /**
     * Constructor, creates a new thread.
     * Before creating instance of thread, Thraed::create() should be called
     * to initialize thread representation in the simulated memory.
     * This representation is then passed to this constructor as an argument.
     *
     * @param gimple GIMPLE++ representation of the program
     * @param memoryState interface to the simulated memory
     * @param cstManager where the constants of the program are stored
     * @param globalsManager where the global declarations of the program are stored
     * @param program the program this thread is part of
     * @param threadInfoRef reference to the info structure of the thread, that will be simulated
     *      by this object
     * @param ID ID of the thread (currently unused)
     */
    Thread (giterator::Gimple* gimple, mmodule::MemoryState* memoryState,
            ConstantsManager* cstManager, GlobalsManager* globalsManager,
            Program* program,
            mmodule::Ref<ThreadInfo> threadInfoRef, unsigned long int ID);
    /**
     * Destructor.
     */
    ~Thread ();
    /**
	 * Reset all the the memory structures according to a thread simulated
	 * memory representation.
	 * To be used after backtracking, when previous states of the program
	 * get reset and the thread structures are no more actual.
	 *
	 * All the information from ThreadInfo structure will
	 * be read and the thread data structures will be updated.
	 *
	 * @param threadInfoRef info structure according to which reset the thread
	 *
	 * @see mmodule::MemoryState::backtrack()
	 */
    void reset(mmodule::Ref<ThreadInfo> threadInfoRef);
    /**
	 * Flush all the thread memory structures into the simulated memory.
	 */
    void flush ();
    /**
     * Performs a step.
     * Interprets the statement to which points the current IP of the thread
     * and moves the current IP to the next statement.
     * IP - instruction pointer.
     */
    void step ();
    /**
     * Returns if this thread is about to read standard input.
     * Unused at the moment.
     *
     * @return @c true if next call to @c step() will read from standard input; @c false otherwise
     */
    bool aboutToReadInput () const;
    /**
     * Returns if this thread is about to execute @c CondStmt statement.
     * Unused at the moment.
     *
     * @return @c true if next call to @c step() will execute @c CondStmt statement;
     *         @c false otherwise
     */
    bool atCondStmt () const;
    /**
     * Obtain the return value of this thread.
     * @pre the thread must be finished.
     *
     * @returns the value returned by this thread.
     */
    const Value* getReturnValue () const;
    /**
     * Clears all thread memory structures from the simulated memory.
     */
    void destroy ();
    /**
     * Whether there are no more instructions to execute
     * (IP is null).
     *
     * @return true if the thread is finished, false otherwise
     */
    bool isFinished () const;
    /**
     * Whether the thread state is not stopped (state != THREAD_STOPPED).
     *
     * @return true if the thread is not stopped,
     * false otherwise
     */
    bool isAlive(); //const
    /**
     * Whether the thread is currently selected to run
     * (state == THREDAD_RUNNING).
     *
     * @return true if the thread is selected to run,
     * false otherwise
     */
    bool isRunning(); //const
    /**
	 * Whether the thread is running and not finished
	 * (state == THREDAD_RUNNING).
	 *
	 * @return true if the thread is selected to run,
	 * false otherwise
	 *
	 * @see isRunning(), isFinished()
	 */
    bool isRunnable(); //const
    /**
	 * Whether the thread state is stopped (state == THREAD_STOPPED).
	 *
	 * @return true if the thread is stopped,
	 * false otherwise
	 */
    bool isTerminated(); //const
    /**
     * Set a state of the thread to running (THREAD_RUNNING).
     */
    void awaken();
    /**
     * Set a thread this thread is waiting for.
     *
     * @param threadID ID of the thread this thread is waiting for
     */
    void setWaitingFor(ID threadID);
    /**
	 * Get a thread this thread is waiting for.
	 *
	 * @return threadID ID of the thread this thread is waiting for
	 * or -1 if there is no such thread
	 */
    ID getWaitingFor() const;
    /**
     * Get a state of the thread.
     * This is not a constant function, because in every call it checks
     * whether a potential timer that put the thread to sleep for a
     * specified number of second elapsed.
     *
     * @return current state of the thread
     */
    ThreadState getState();
    /**
     * Set a state of the thread.
     *
     * @param newState new state for this thread
     */
    void setState(ThreadState newState);
    /**
     * Obtain an ID of this thread.
     *
     * @return unique ID of this thread
     */
    Thread::ID getID() const;
    /**
     * Put a thread sleep (state THREAD_SLEEP) for specified number of seconds.
     * This counts only for interpreter mode, in model checking mode
     * no such sleeping is performed.
     *
     * @param interval how many seconds should the thread sleep
     * @param mode mode in which is the program simulated
     */
    void sleep(unsigned int interval, RunningMode mode);
    /**
     * Get time that is remining on the sleep timer for the thread to wake up.
     *
     * @return remaining time on sleep timer or 0 if the timer is not even started
     */
    unsigned int getRemainingTime() const;
    /**
     * Inform the thread that whether there is join called on it.
     *
     * @param value whether there is join called on this thred
     */
    void setIsBeingJoined(bool value);
    /**
     * Find out whether there is join called on this thread
     */
    bool getIsBeingJoined() const;
    /**
     * Find out whether current instruction of the thread is safe.
     * Instruction is when its result cannot depend on the different scheduling policies of threads.
     * That means all instructions that manipulate with local variables are marked as safe.
     * On the other hand, unsafe instructions can be affected by different scheduling policies.
	 * These are the instructions that manipulate global variables.
	 *
	 * @return true if current instruction is not safe, false otherwise
	 *
	 * @see InstructionSafetyResolver
     */
    bool isCurrentInstructionUnsafe() const;
    /**
     * Obtain current statement - the statement that will be interpreted
     * in the very next call to step()
     */
    const giterator::Stmt* getCurrentStmt() const;
protected:
    /** @name Stepping methods
     * These methods move the current instruction pointer of this thread.
     */
    //@{
    void visit (const NextStep& step);    
    void visit (const CondStep& step);
    void visit (const ReturnStep& step);
    void visit (const CallStep& step);
    void visit (const JumpStep& step);
    //@}
private:
    /**
     * Performs the return from current function.
     *
     * It copies the return value to the right place
     * and resets the callee function. Stack frame
     * of the function is destroyed
     */
	void returnFromFunction ();
	/**
	 * Performs a function call.
	 *
	 * @param func function to call
	 * @param frame pointer to the beginning of the stack frame of function @c func
	 */
	void callFunction (const FunctionValue* func, mmodule::Ptr frame);
};

}

#endif /* INT_THREAD_H */
