// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file ThreadFunctionValue.h
 * Declaration of class representing simulated thread function.
 */
#ifndef INT_THREAD_FUNCTION_VALUE_H
#define INT_THREAD_FUNCTION_VALUE_H

#include "AbstractValue.h"

#include "../git/FunctionInfo.h"
#include "../git/CallStmt.h"    // TODO remove this include

namespace ginterpreter {

class ConstantsManager;
class GlobalsManager;
class FunctionValue;
class Program;


/**
 * Class representing simulated thread function.
 * This class is a regular simulated value, so it can be stored in the memory.
 * It stores all the information needed to interpret the function as well
 * as pointer to the method implementing the function.
 * It is similar to the @c BuiltinFunctionValue, except funcitons
 * that implement thread functions require one more parameter -
 * pointer to the simulated program.
 */
class ThreadFunctionValue : public AbstractNonPtrValue {
  public:
    /*! type of the pointer to the function, that implements this simulated function */
    typedef std::auto_ptr<Value> (*FuncMethod) (
        const FunctionValue* currFunc, mmodule::MemoryState* memoryState,
        ConstantsManager* cstManager, GlobalsManager* globalsManager,
        mmodule::Ptr topFrame,
        std::pair<giterator::CallStmt::ParameterIterator,
                  giterator::CallStmt::ParameterIterator> parameters,
        const giterator::Type* returnType,
        Program* program);

  private:
    /*! the code of this thread function */
    enum giterator::thread_function_code m_code;
    /*! pointer to the function implementing this built-in function */
    FuncMethod m_method;

  public:
	/**
	* Constructor.
	*
	* @param code code of the thread function
	* @param type type of the thread function
	* @param method pointer to the function implementing the built-in function
	*/
	ThreadFunctionValue (enum giterator::thread_function_code code,
						const giterator::FunctionType* type,
						FuncMethod method);

	/**
	 * Clone this value.
	 * Aborts, function values should not be copied.
	 */
	Value* clone() const;

	/**
	 * Get a hash of this value.
	 * For memory state equivalence check purposes.
	 * Hashes the code of this function.
	 *
	 * @return hash of this value.
	 */
	mmodule::Hash getHash () const;
	/**
	 * Returns the code of this thread function.
	 *
	 * @return the code of this thread function
	 */
	enum giterator::thread_function_code getCode () const;

	/**
	 * Returns the function type of this function.
	 *
	 * @return the function type of this function
	 */
	const giterator::FunctionType* getFunctionType () const;

	/**
	 * Returns the pointer to the function implementing this function.
	 *
	 * @return the pointer to the function implementing this function
	 */
	FuncMethod getMethod () const;
};

}

#endif /* INT_THREAD_FUNCTION_VALUE_H */
