// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file ThreadPool.cpp
 * Definitions for a class that manages threads.
 */

#include "ThreadPool.h"

#include "ModelCheckerExceptions.h"
#include "InterpretExceptions.h"

#include <climits>
#include <cerrno>

using namespace mmodule;

namespace ginterpreter
{

ThreadPool::ThreadPool(Ref<ThreadPoolInfo> threadPoolInfoRef, MemoryState* memoryState)
: m_currentThreadID(Thread::NO_ID),
  m_mainThreadID(Thread::NO_ID),
  m_threadPoolInfoRef(threadPoolInfoRef),
  m_memoryState(memoryState)
{
	m_memoryState->store (m_threadPoolInfoRef, &ThreadPoolInfo::mainThreadID, m_mainThreadID);
	m_memoryState->store (m_threadPoolInfoRef, &ThreadPoolInfo::currentThreadID, m_currentThreadID);
}

ThreadPool::ThreadPool(Thread::ID currentThreadID,
					   Thread::ID mainThreadID,
					   Ref<ThreadPoolInfo> threadPoolInfo,
					   MemoryState* memoryState)
: m_currentThreadID(currentThreadID),
  m_mainThreadID(mainThreadID),
  m_threadPoolInfoRef(threadPoolInfo),
  m_memoryState(memoryState)
{
	m_memoryState->store (m_threadPoolInfoRef, &ThreadPoolInfo::mainThreadID, m_mainThreadID);
	m_memoryState->store (m_threadPoolInfoRef, &ThreadPoolInfo::currentThreadID, m_currentThreadID);
}

void
ThreadPool::flush()
{
	m_memoryState->store (m_threadPoolInfoRef, &ThreadPoolInfo::mainThreadID, getMainThreadID());
	m_memoryState->store (m_threadPoolInfoRef, &ThreadPoolInfo::currentThreadID, getCurrentThreadID());
	m_memoryState->store (m_threadPoolInfoRef, &ThreadPoolInfo::threadCount, getThreadCount());

	// Threads
	unsigned int ids = 0;
	for (ThreadsIterator i1 = m_threads.begin(); i1 != m_threads.end(); ++i1)
	{
		ids = 10 * ids + i1->second->getID() + 1;
		i1->second->flush();
	}

	m_memoryState->store(m_threadPoolInfoRef, &ThreadPoolInfo::threadIds, ids);

	// Mutexes
	ids = 0;
	for (MutexesIterator i1 = m_mutexes.begin(); i1 != m_mutexes.end(); ++i1)
	{
		ids = 10 * ids + i1->second->getID() + 1;
		i1->second->flush();
	}

	m_memoryState->store(m_threadPoolInfoRef, &ThreadPoolInfo::mutexIds, ids);

	// Conditions
	ids = 0;
	for (ConditionsIterator i1 = m_conditions.begin(); i1 != m_conditions.end(); ++i1)
	{
		ids = 10 * ids + i1->second->getID() + 1;
		i1->second->flush();
	}

	m_memoryState->store(m_threadPoolInfoRef, &ThreadPoolInfo::conditionIds, ids);

}

void ThreadPool::reset(Ref<ThreadPoolInfo> threadPoolInfoRef,
					   giterator::Gimple* gimple, MemoryState* memoryState,
		 	 	 	   ConstantsManager* cstManager, GlobalsManager* globalsManager,
		 	 	 	   Program* program)
{
	clear();

	m_threadPoolInfoRef = threadPoolInfoRef;

	Thread::ID mainID = m_memoryState->load(m_threadPoolInfoRef, &ThreadPoolInfo::mainThreadID);
	Thread::ID currentID = m_memoryState->load(m_threadPoolInfoRef, &ThreadPoolInfo::currentThreadID);

	setMainThreadID(mainID);
	setCurrentThreadID(currentID);

	//size_t NThreads = m_memoryState->load(m_threadPoolInfoRef, &ThreadPoolInfo::threadCount);

	// Threads
	Ref<ThreadsInfo> tsi = m_memoryState->load(m_threadPoolInfoRef, &ThreadPoolInfo::threads);
	int threadIds = m_memoryState->load(m_threadPoolInfoRef, &ThreadPoolInfo::threadIds);

	while (threadIds != 0)
	{
		Thread::ID id = threadIds % 10 - 1;
		Ref<ThreadInfo> ti = m_memoryState->loadIndexed(tsi, &ThreadsInfo::threads, id);
		newThread(gimple, memoryState, cstManager, globalsManager, program, ti, id);
		threadIds = threadIds / 10;
	}

	// Mutexes
	Ref<MutexesInfo> msi = m_memoryState->load(m_threadPoolInfoRef, &ThreadPoolInfo::mutexes);
	int mutexIds = m_memoryState->load(m_threadPoolInfoRef, &ThreadPoolInfo::mutexIds);

	while (mutexIds != 0)
	{
		Mutex::ID id = mutexIds % 10 - 1;
		Ref<MutexInfo> mi = m_memoryState->loadIndexed(msi, &MutexesInfo::mutexes, id);
		newMutex(id, mi);
		mutexIds = mutexIds / 10;
	}

	// Conditions
	Ref<ConditionsInfo> csi = m_memoryState->load(m_threadPoolInfoRef, &ThreadPoolInfo::conditions);
	int conditionIds = m_memoryState->load(m_threadPoolInfoRef, &ThreadPoolInfo::conditionIds);

	while (conditionIds != 0)
	{
		Condition::ID id = conditionIds % 10 - 1;
		Ref<ConditionInfo> ci = m_memoryState->loadIndexed(csi, &ConditionsInfo::conditions, id);
		newCondition(id, ci);
		conditionIds = conditionIds / 10;
	}
}

void
ThreadPool::clear()
{
	m_threads.clear();
	m_mutexes.clear();
	m_conditions.clear();
	m_currentThreadID = Thread::NO_ID;
	m_mainThreadID = Thread::NO_ID;
}

std::pair<ThreadPool::ThreadsConstIterator, ThreadPool::ThreadsConstIterator>
ThreadPool::threads() const
{
	return std::make_pair (m_threads.begin(), m_threads.end());
}

std::pair<ThreadPool::ThreadsIterator, ThreadPool::ThreadsIterator>
ThreadPool::threads()
{
	return std::make_pair (m_threads.begin(), m_threads.end());
}

ThreadPool::ThreadsNonOwning
ThreadPool::threads(ThreadState state)
{
	ThreadsNonOwning outThreads;

	for (ThreadsIterator i1 = m_threads.begin(); i1 != m_threads.end(); ++i1)
	{
		if (i1->second->getState() == state)
			outThreads.push_back(i1->second);
	}

	return outThreads;
}

ThreadPool::RunnableThreads
ThreadPool::runnableThreads()
{
	RunnableThreads outThreads;

	for (ThreadsIterator i1 = m_threads.begin(); i1 != m_threads.end(); ++i1)
	{
		if (i1->second->isRunnable())
			outThreads.push_back(i1->second->getID());
	}

	return outThreads;
}

size_t
ThreadPool::getThreadCount() const
{
	return m_threads.size();
}

size_t
ThreadPool::getThreadCount(ThreadState state) //const
{
	size_t outCount = 0;

	// ThreadsConstIterator
	for (ThreadsIterator i1 = m_threads.begin(); i1 != m_threads.end(); ++i1)
	{
		if (i1->second->getState() == state)
			++outCount;
	}

	return outCount;
}

size_t
ThreadPool::getUnfinishedThreadCount() const
{
	size_t outCount = 0;

	for (ThreadsConstIterator i1 = m_threads.begin(); i1 != m_threads.end(); ++i1)
	{
		if (i1->second->isFinished() == false)
			++outCount;
	}

	return outCount;
}

Thread::ID
ThreadPool::getCurrentThreadID() const
{
	assert(m_currentThreadID != Thread::NO_ID);
	return m_currentThreadID;
}

void
ThreadPool::setCurrentThreadID(Thread::ID threadID)
{
	m_currentThreadID = threadID;
}

Thread*
ThreadPool::getCurrentThread()
{
	assert(m_currentThreadID != Thread::NO_ID);
	if (m_currentThreadID == Thread::EXITED_ID)
		return NULL;
	ThreadsIterator i1 = m_threads.find(m_currentThreadID);
	assert (i1 != m_threads.end());
	return i1->second;
}

const Thread*
ThreadPool::getCurrentThread() const
{
	assert(m_currentThreadID != Thread::NO_ID);
	if (m_currentThreadID == Thread::EXITED_ID)
		return NULL;
	ThreadsConstIterator i1 = m_threads.find(m_currentThreadID);
	assert (i1 != m_threads.end());
	return i1->second;
}

Thread::ID
ThreadPool::getMainThreadID() const
{
	assert(m_mainThreadID != Thread::NO_ID);
	return m_mainThreadID;
}

void
ThreadPool::setMainThreadID(Thread::ID id)
{
	assert(m_mainThreadID == Thread::NO_ID);
	m_mainThreadID = id;
}

const Thread*
ThreadPool::getMainThread() const
{
	assert(m_mainThreadID != Thread::NO_ID && m_mainThreadID != Thread::EXITED_ID);
	ThreadsConstIterator i1 = m_threads.find(m_mainThreadID);
	assert(i1 != m_threads.end());
	return i1->second;
}

Thread::ID
ThreadPool::newThread(giterator::Gimple* gimple, MemoryState* memoryState,
					 ConstantsManager* cstManager, GlobalsManager* globalsManager,
					 Program* program, Ref<ThreadInfo> threadInfoRef, Thread::ID threadID)
{
	if (threadID == Thread::NO_ID)
		threadID = m_threads.size();
	m_memoryState->store(threadInfoRef, &ThreadInfo::threadID, threadID);
	Thread* thread = new Thread(gimple, memoryState, cstManager, globalsManager, program, threadInfoRef, threadID);

	assert(m_threads.find(thread->getID()) == m_threads.end());
	m_threads.insert(thread->getID(), std::auto_ptr<Thread> (thread));

	m_memoryState->store(m_threadPoolInfoRef, &ThreadPoolInfo::threadCount, m_threads.size());

	Ref<ThreadsInfo> tsi = m_memoryState->load(m_threadPoolInfoRef, &ThreadPoolInfo::threads);
	m_memoryState->storeIndexed(tsi, &ThreadsInfo::threads, threadID, threadInfoRef);
	m_memoryState->store(m_threadPoolInfoRef, &ThreadPoolInfo::threads, tsi);

	return threadID;
}

void
ThreadPool::deleteThread(Thread::ID threadID)
{
	assert(m_threads.find(threadID) != m_threads.end());
	m_threads.erase(threadID);
	//m_threads.insert(threadID, std::auto_ptr<Thread> (NULL));
}

void
ThreadPool::terminateThread(Thread::ID threadID)
{
	assert(m_threads.find(threadID) != m_threads.end());

	if (threadID == m_mainThreadID)
		m_mainThreadID = Thread::EXITED_ID;
	if (threadID == m_currentThreadID)
		m_currentThreadID = Thread::EXITED_ID;

	Thread& toTerminate = m_threads.at(threadID);
	toTerminate.setState(THREAD_STOPPED);

	// Wake up threads that called Join on this thread.
	for (ThreadsIterator i1 = m_threads.begin(); i1 != m_threads.end(); ++i1)
	{
		Thread* t = i1->second;
		if (t->getState() == THREAD_JOIN && t->getWaitingFor() == threadID)
		{
			t->awaken();
			t->setWaitingFor(Thread::NO_ID);
		}
	}

	toTerminate.destroy();
	deleteThread(threadID);

	LOG_DEBUG("Thread with ID %d terminated.", threadID);
}

int
ThreadPool::joinThreads(Thread::ID blocking, Thread::ID toTerminate)
{
	assert(m_threads.find(blocking) != m_threads.end());

	if (m_threads.find(toTerminate) == m_threads.end()) // thread has already exited
	{
		WARNING("Thread with ID %d does not exist.", toTerminate);
		return ESRCH;
	}
	if (blocking == toTerminate)
	{
		WARNING("Join called with current thread with ID %d.", blocking);
		return EDEADLK;
	}

	Thread* blockingThread = &m_threads.at(blocking);
	assert(blockingThread->getState() == THREAD_RUNNING);

	Thread* toTerminateThread = &m_threads.at(toTerminate);
	if (toTerminateThread->getIsBeingJoined() == true)
	{
		WARNING("Another thread is already waiting to join with this thread with ID %d.",
				toTerminate);
		return EINVAL;
	}

	toTerminateThread->setIsBeingJoined(true);

	/*if (!toTerminateThread->isAlive())
		LOG_DEBUG("Thread with id %d is not waiting for %d since %d is not alive",
				  blocking, toTerminate, blocking);*/

	blockingThread->setWaitingFor(toTerminate);
	blockingThread->setState(THREAD_JOIN);

	LOG_DEBUG("Thread with ID %d is now waiting for thread %d.", blocking, toTerminate);

	//Perform early deadlock detection by cycle detection in the waiting for graph.
	checkDeadlockFrom(blocking);

	return 0;
}

void ThreadPool::checkDeadlockFrom(Thread::ID threadID) const
{
	Thread::ID currentID = threadID;
	std::map<Thread::ID, bool> threadSeen;
	//bool threadSeen[m_threads.size()];

	for (ThreadsConstIterator i1 = m_threads.begin(); i1 != m_threads.end(); ++i1)
		threadSeen[i1->first] = false;

	while (currentID != Thread::NO_ID && !threadSeen[currentID])
	{
		// is some thread waiting for a null thread ?
		//assert(m_threads.at(currentID) != NULL);
		threadSeen[currentID] = true;

		currentID = m_threads.at(currentID).getWaitingFor();
	}

	if (currentID != Thread::NO_ID)
	{
		throw IDeadlockException();
	}
}

void ThreadPool::pickAndSetRunnableThread()
{
	m_currentThreadID = Thread::EXITED_ID;
	for (ThreadsIterator i1 = m_threads.begin(); i1 != m_threads.end(); ++i1)
	{
		if (i1->second->isRunnable())
		{
			m_currentThreadID = i1->second->getID();
			return;
		}
	}
}

int ThreadPool::newMutex(Mutex::ID& mutexID, Ref<MutexInfo> mutexInfoRef)
{
	if (mutexID == Thread::NO_ID) // never initialized mutex, create a new ID
		mutexID = m_mutexes.size();
	else
	{
		if (m_mutexes.find(mutexID) != m_mutexes.end())
		{
			WARNING ("Attempted to initialize an already initialized mutex with ID %d.",
					mutexID);
			return EINVAL;
		}
	}

	LOG_DEBUG ("###(Current Thread %d) Init mutex: %d", m_currentThreadID, mutexID);

	m_memoryState->store(mutexInfoRef, &MutexInfo::mutexID, mutexID);
	Mutex* newMutex = new Mutex(mutexID, m_memoryState, mutexInfoRef);
	m_mutexes.insert(mutexID, std::auto_ptr<Mutex> (newMutex));


	Ref<MutexesInfo> msi = m_memoryState->load(m_threadPoolInfoRef, &ThreadPoolInfo::mutexes);
	m_memoryState->storeIndexed(msi, &MutexesInfo::mutexes, mutexID, mutexInfoRef);
	m_memoryState->store(m_threadPoolInfoRef, &ThreadPoolInfo::mutexes, msi);

	return 0;
}

void ThreadPool::initMutex(Mutex::ID mutexID)
{
	assert(m_mutexes.find(mutexID) != m_mutexes.end());
	m_mutexes.at(mutexID).init();
}

int ThreadPool::destroyMutex(Mutex::ID mutexID)
{
	LOG_DEBUG ("###(Current Thread %d) Destroy mutex: %d", m_currentThreadID, mutexID);

	MutexesIterator i1 = m_mutexes.find(mutexID);
	if (i1 == m_mutexes.end())
	{
		WARNING("Attempted to destroy uninitialized mutex with ID %d.", mutexID);
		return EINVAL;
	}

	Mutex* toDestroy = i1->second;

	if (toDestroy->isLocked() || toDestroy->isReferenced())
	{
		WARNING("Attempted to destroy locked mutex or mutex referenced by thread condition with ID %d", mutexID);
		return EBUSY;
	}

	toDestroy->destroy();
	m_mutexes.erase(mutexID);
	return 0;
}

int ThreadPool::lockMutex(Mutex::ID mutexID)
{
	LOG_DEBUG ("###(Current Thread %d) Lock mutex: %d", m_currentThreadID, mutexID);

	MutexesIterator i1 = m_mutexes.find(mutexID);
	if (i1 == m_mutexes.end())
	{
		WARNING("Attempted to lock uninitialized mutex with ID %d.", mutexID);
		return EINVAL;
	}

	Mutex* toLock = i1->second;

	if (toLock->isLocked())
	{
		Thread::ID lockOwnerID = toLock->getOwnerID();

		if (lockOwnerID == m_currentThreadID)
		{
			WARNING("Attempted to lock an already lock mutex with ID %d by the same thread", mutexID);
			return EDEADLK;
		}
		else
		{
			toLock->addWaitingThread(m_currentThreadID);
			//Thread* lockOwner = m_threads.get(lockOwnerID);
			Thread& currentThread = m_threads.at(m_currentThreadID);

			currentThread.setWaitingFor(lockOwnerID);
			currentThread.setState(THREAD_WAITLOCK);

			LOG_DEBUG("Thread with ID %d is now waiting for thread %d.", m_currentThreadID, lockOwnerID);

			//Perform early deadlock detection by cycle detection in the waiting for graph.
			checkDeadlockFrom(m_currentThreadID);
		}
	}
	else
		toLock->lock(m_currentThreadID);

	return 0;
}

int ThreadPool::tryLockMutex(Mutex::ID mutexID)
{
	LOG_DEBUG ("###(Trylock Thread %d) Unlock mutex: %d", m_currentThreadID, mutexID);

	MutexesIterator i1 = m_mutexes.find(mutexID);
	if (i1 == m_mutexes.end())
	{
		WARNING("Attempted to try to lock uninitialized mutex with ID %d.", mutexID);
		return EINVAL;
	}

	Mutex* toTryLock = i1->second;

	if (toTryLock->isLocked())
		return EBUSY;
	else
	{
		toTryLock->lock(m_currentThreadID);
		return 0;
	}
}

int ThreadPool::unlockMutex(Mutex::ID mutexID)
{
	LOG_DEBUG ("###(Current Thread %d) Unlock mutex: %d", m_currentThreadID, mutexID);

	MutexesIterator i1 = m_mutexes.find(mutexID);
	if (i1 == m_mutexes.end())
	{
		WARNING("Attempted to unlock uninitialized mutex with ID %d.", mutexID);
		return EINVAL;
	}

	Mutex* toUnlock = i1->second;
	Thread::ID toUnlockOwnerID = toUnlock->getOwnerID();

	if (toUnlockOwnerID != m_currentThreadID)
	{
		WARNING("Attempted to unlock mutex with ID %d that is not owned by current thread.", mutexID);
		return EPERM;
	}

	toUnlock->unlock();

	Thread::ID frontID = toUnlock->getFrontWaitingThread();
	if (frontID != Thread::NO_ID)
	{
		assert(m_threads.find(frontID) != m_threads.end());
		Thread& front = m_threads.at(frontID);
		front.awaken();
		toUnlock->lock(frontID);
	}
	/*
	// Wake up threads that waited for this mutex.
	for (ThreadsIterator i1 = m_threads.begin(); i1 != m_threads.end(); ++i1)
	{
		Thread* t = i1->second;
		if (t->getState() == THREAD_WAITLOCK && t->getWaitingFor() == toUnlockOwnerID)
			t->awaken();
	}*/

	return 0;
}

int
ThreadPool::newCondition(Condition::ID& condID, Ref<ConditionInfo> conditionInfoRef)
{
	if (condID == Thread::NO_ID) // never initialized condition, create a new ID
		condID = m_conditions.size();
	else
	{
		if (m_conditions.find(condID) != m_conditions.end())
		{
			WARNING ("Attempted to initialize an already initialized condition with ID %d.",
					 condID);
			return EINVAL;
		}
	}

	LOG_DEBUG ("###(Current Thread %d) Init condition: %d", m_currentThreadID, condID);

	m_memoryState->store(conditionInfoRef, &ConditionInfo::condID, condID);
	Condition* newCondition = new Condition(condID, m_memoryState, conditionInfoRef);
	m_conditions.insert(condID, std::auto_ptr<Condition> (newCondition));

	Ref<ConditionsInfo> csi = m_memoryState->load(m_threadPoolInfoRef, &ThreadPoolInfo::conditions);
	m_memoryState->storeIndexed(csi, &ConditionsInfo::conditions, condID, conditionInfoRef);
	m_memoryState->store(m_threadPoolInfoRef, &ThreadPoolInfo::conditions, csi);

	return 0;
}

void ThreadPool::initCondition(Condition::ID condID)
{
	assert(m_conditions.find(condID) != m_conditions.end());
	m_conditions.at(condID).init();
}

int
ThreadPool::destroyCondition(Condition::ID condID)
{
	LOG_DEBUG ("###(Current Thread %d) Destroy condition: %d", m_currentThreadID, condID);

	ConditionsIterator i1 = m_conditions.find(condID);
	if (i1 == m_conditions.end())
	{
		WARNING("Attempted to destroy uninitialized condition with ID %d.", condID);
		return EINVAL;
	}

	Condition* toDestroy = i1->second;

	if (toDestroy->isInUse())
	{
		WARNING("Attempted to destroy condition with ID %d that is in use (by some condition wait function).", condID);
		return EBUSY;
	}

	i1->second->destroy();
	m_conditions.erase(condID);

	return 0;
}

int
ThreadPool::signalCondition(Condition::ID condID)
{
	LOG_DEBUG ("###(Current Thread %d) Signal condition: %d", m_currentThreadID, condID);

	ConditionsIterator i1 = m_conditions.find(condID);
	if (i1 == m_conditions.end())
	{
		WARNING("Attempted to signal uninitialized condition with ID %d.", condID);
		return EINVAL;
	}

	Condition* toSignal = i1->second;

	Thread::ID unlockedWhileWaitID = toSignal->getFrontWaitingThread();

	if (unlockedWhileWaitID == Thread::NO_ID)
	{
		// this is not a mistake
		LOG_DEBUG("Signal from condition with ID %d does not wake any thread because no thread is waiting for this condition",
				  condID);
		return 0;
	}

	Mutex::ID unlockedMutexID = toSignal->getUnlockedMutex();

	assert(m_mutexes.find(unlockedMutexID) != m_mutexes.end());
	assert(m_threads.find(unlockedWhileWaitID) != m_threads.end());

	Mutex& unlockedMutex = m_mutexes.at(unlockedMutexID);
	assert(unlockedMutex.isLocked() == false);
	unlockedMutex.lock(unlockedWhileWaitID);
	unlockedMutex.setReferenced(false);

	Thread& unlockedWhileWait = m_threads.at(unlockedWhileWaitID);
	assert(unlockedWhileWait.getState() == THREAD_WAITCOND);
	unlockedWhileWait.awaken();

	//this->unlockMutex(unlockedWhileWait.first);

	return 0;
}

int
ThreadPool::waitCondition(Condition::ID condID, Mutex::ID mutexID)
{
	ConditionsIterator i2 = m_conditions.find(condID);
	MutexesIterator i1 = m_mutexes.find(mutexID);

	if (i2 == m_conditions.end() || i1 == m_mutexes.end())
	{
		WARNING("Either condition with ID %d or mutex with ID %d is uninitialized in condition wait.", condID, mutexID);
		return EINVAL;
	}

	LOG_DEBUG ("###(Current Thread %d) Wait condition: %d (mutex %d)", m_currentThreadID, condID, mutexID);

	Mutex* blocked = i1->second;
	bool bblocked = blocked->isLocked();

	if (!bblocked)
	{
		// this behaviour is not mentioned in pthread_cond_wait man page, but the function behaves like this
		WARNING("Mutex is not blocked at all, but wait continues as it was locked by current thread.");
	}
	else if (blocked->getOwnerID() != m_currentThreadID)
	{
		//throw ThreadCondWaitMutexNotBlockedException(condID, mutexID);
		WARNING("Trying to call condition wait (condition ID: %d) function on a mutex with ID %d that is not locked by current thread with ID %d (or not locked at all).",
				  condID, mutexID, m_currentThreadID);
		return EPERM;
	}

	Condition* toWait = i2->second;

	if (toWait->isInUse())
	{
		if (toWait->getUnlockedMutex() != mutexID)
		{
			WARNING("Different mutexes were supplied for concurrent pthread_cond_wait() operations on the same condition variable with id %d.", condID);
			return EINVAL;
		}
	}
	else
		toWait->setUnlockedMutex(mutexID);

	blocked->unlock();
	blocked->setReferenced(true);

	toWait->addWaitingThread(mutexID, m_currentThreadID);

	Thread& currentThread = m_threads.at(m_currentThreadID);

	//currentThread.setWaitingFor(lockOwnerID);
	currentThread.setState(THREAD_WAITCOND);

	LOG_DEBUG("Thread with ID %d that locked mutex %d is now waiting for condition %d.",
			  m_currentThreadID, mutexID, condID);

	//Perform early deadlock detection by cycle detection in the waiting for graph.
	checkDeadlockFrom(m_currentThreadID);

	return 0;
}

unsigned int
ThreadPool::getMinRemainingSleepInterval()
{
	unsigned int min = UINT_MAX;

	for (ThreadsIterator i1 = m_threads.begin(); i1 != m_threads.end(); ++i1)
	{
		ThreadState tstate = i1->second->getState();
		if (tstate == THREAD_SLEEP)
		{
			unsigned int remainingTime = i1->second->getRemainingTime();

			if (remainingTime < min)
				min = remainingTime;
		}
		else if (tstate == THREAD_RUNNING)
			return 0;
	}

	assert(min != UINT_MAX);
	return min;
}

void
ThreadPool::addThreadName(Thread::ID threadID, std::string name)
{
	m_threadNames.insert(threadID, name);
}

std::string
ThreadPool::getThreadName(Thread::ID threadID) const
{
	return m_threadNames.getName(threadID);
}

void
ThreadPool::print() //const
{

	for (ThreadsIterator i1 = m_threads.begin(); i1 != m_threads.end(); ++i1)
	{
		Thread* t = i1->second;
		LOG_DEBUG("Thread, ID: %d, state: %d, waiting for: %d, being joined: %d",
				  t->getID(), t->getState(), t->getWaitingFor(), t->getIsBeingJoined());
	}

	LOG_DEBUG("Main thread: %d, current thread: %d", m_mainThreadID, m_currentThreadID);
}

}
