// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file ThreadPool.h
 * Declaration of a class that manages threads.
 */

#ifndef INT_THREADPOOL_H
#define INT_THREADPOOL_H

#include "Thread.h"
#include "Mutex.h"
#include "Condition.h"
#include "ThreadState.h"
#include "ThreadPoolInfo.h"
#include "ThreadNameHolder.h"

#include "../mm/MemoryState.h"

#include <memory>
#include <boost/ptr_container/ptr_map.hpp>
#include <vector>
#include <deque>

namespace ginterpreter
{

class ConstantsManager;
class GlobalsManager;
class Program;

/**
 * Class that manages all the threads in the simulated program.
 *
 * Class @c ThreadPool is capable of creating a thread, destroying it and joining two threads.
 * @c ThreadPool also manages mutexes and thread conditions of the program.
 * @c ThreadPool thus manages creation, deletion, locking and  unlocking of mutexes as well as
 * creation, deletion, signaling and waiting for thread conditions. The mechanism by which
 * the mutexes and conditions work with the simulated memory is the same as for Thread class
 * (usage of ThreadPoolInfo structure). Same as @c Program, @c ThreadPool is also able to
 * @c flush() itself to simulated memory or @c reset() according to the current state of memory.
 *
 * @see Thread, Mutex, Condition, ThreadPoolInfo
 */
class ThreadPool
{
private:
	/*! type of the container with threads that owns thread pointers. */
	typedef boost::ptr_map<Thread::ID, Thread> Threads;
	/*! type of the container with mutexes that owns mutex pointers. */
	typedef boost::ptr_map<Mutex::ID, Mutex> Mutexes;
	/*! type of the container with thread conditions that owns thred conditions pointers. */
	typedef boost::ptr_map<Condition::ID, Condition> Conditions;
	/*! threads managed by this class. */
	Threads m_threads;
	/*! mutexes managed by this class. */
	Mutexes m_mutexes;
	/*! conditions managed by this class. */
	Conditions m_conditions;
	/*! ID of thread currently selected to run. */
	Thread::ID m_currentThreadID;
	/*! ID of main thread of a program */
	Thread::ID m_mainThreadID;
	/*! reference to the info structure of this thread pool */
	mmodule::Ref<ThreadPoolInfo> m_threadPoolInfoRef;
	/*! holder of the names of threads (only for output purposes). */
	ThreadNameHolder m_threadNames;
	/*! memory state */
	mmodule::MemoryState* m_memoryState;
public:
	/*! type of the container with threads that does not own the thread pointers. */
	typedef std::vector<Thread*> ThreadsNonOwning;
	/*! type of the container for runnable threads that does not own the thread pointers. */
	typedef std::deque<Thread::ID> RunnableThreads;
	/*! iterator for the container with threads that owns thread pointers. */
	typedef Threads::const_iterator ThreadsConstIterator;
	/*! iterator for the container with threads that owns thread pointers. */
	typedef Threads::iterator ThreadsIterator;
	/*! iterator for the container with threads that does not own thread pointers. */
	typedef ThreadsNonOwning::const_iterator ThreadsNonOwningIterator;
	/*! iterator for the container with mutexes that owns mutex pointers. */
	typedef Mutexes::const_iterator MutexesConstIterator;
	/*! iterator for the container with mutexes that owns mutex pointers. */
	typedef Mutexes::iterator MutexesIterator;
	/*! iterator for the container with thread conditions that owns mutex pointers. */
	typedef Conditions::const_iterator ConditionsConstIterator;
	/*! iterator for the container with thread conditions that owns mutex pointers. */
	typedef Conditions::iterator ConditionsIterator;
	/**
	 * Constructor, creates a new thread pool.
	 *
	 * @param threadPoolInfoRef reference to the info structure of this thread pool
	 * @param memoryState interface to the simulated memory
	 */
	ThreadPool(mmodule::Ref<ThreadPoolInfo> threadPoolInfoRef, mmodule::MemoryState* memoryState);

	/**
	 * Constructor, creates a new thread pool.
	 *
	 * @param currentThreadID ID of the thread that will be selected to run
	 * @param mainThreadID ID of main thread
	 * @param threadPoolInfo reference to the info structure of this thread pool
	 * @param memoryState interface to the simulated memory
	 */
	ThreadPool(Thread::ID currentThreadID,
			   Thread::ID mainThreadID,
			   mmodule::Ref<ThreadPoolInfo> threadPoolInfo,
			   mmodule::MemoryState* memoryState);

	/**
	 * Empties all the data structures of thread pool.
	 */
	void clear();
	/**
	 * Flushes all the information cached in thread pool into simulated memory.
	 */
	void flush();
	/**
	 * Reset all the the memory structures of this class according to a thread
	 * simulated memory representation.
	 * To be used after backtracking, when previous states of the program
	 * get reset and the thread pool structures are no more actual.
	 *
	 * All the information from ThreadPoolInfo structure will
	 * be read and the thread pool data structures will be updated.
	 *
	 * After call to this function, current thread is not set (== -1).
	 *
	 * @param threadPoolInfoRef info structure according to which reset the thread pool
	 * @param gimple GIMPLE++ representation of the program
     * @param memoryState interface to the simulated memory
     * @param cstManager where the constants of the program are stored
     * @param globalsManager where the global declarations of the program are stored
     * @param program the program this thread pool is part of
	 *
	 * @see mmodule::MemoryState::backtrack()
	 */
	void reset(mmodule::Ref<ThreadPoolInfo> threadPoolInfoRef,
			   giterator::Gimple* gimple, mmodule::MemoryState* memoryState,
	 	 	   ConstantsManager* cstManager, GlobalsManager* globalsManager,
	 	 	   Program* program);
	/**
	 * Obtain the threads managed by this class.
	 *
	 * @returns threads managed by this class
	 */
	std::pair<ThreadsConstIterator, ThreadsConstIterator>
	threads() const;
	/**
	 * Obtain the threads managed by this class.
	 *
	 * @returns threads managed by this class
	 */
	std::pair<ThreadsIterator, ThreadsIterator>
	threads();
	/**
	 * Obtain the threads managed by this class in specified state.
	 *
	 * @param state state of threads to obtain
	 * @returns threads managed by this class in state @c state
	 */
	ThreadsNonOwning threads(ThreadState state);
	/**
	 * Obtain ID's of runnable threads managed by this class.
	 *
	 * @returns ID's runnable threads managed by this class
	 */
	RunnableThreads runnableThreads();
	/**
	 * Get the number of threads managed by this class.
	 *
	 * @return number of threads managed by this class
	 */
	size_t getThreadCount() const;
	/**
	 * Get the number of threads managed by this class in specified state.
	 *
	 * @param state state in which are counted threads
	 * @return number of threads managed by this class in state @c state
	 *
	 * TODO: const removed, because it uses Thread::getState(), which is
	 * not const either (it wakes up sleeping threads).
	 */
	size_t getThreadCount(ThreadState state);// const;
	/**
	 * Get the number of threads managed by this class that are not finished.
	 *
	 * @return number of threads managed by this class that are not finished.
	 */
	size_t getUnfinishedThreadCount() const;
	/**
	 * Obtain an ID of a thread currently selected to run.
	 *
	 * @return ID of a thread currently selected to run
	 */
	Thread::ID getCurrentThreadID() const;
	/**
	 * Set which thread is currently selected to run.
	 *
	 * @param threadID unique ID of setted thread
	 */
	void setCurrentThreadID(Thread::ID threadID);
	/**
	 * Obtain a thread currently selected to run.
	 *
	 * @return thread currently selected to run
	 */
	Thread* getCurrentThread();
	/**
	 * Obtain an constant thread currently selected to run.
	 *
	 * @return constant thread currently selected to run
	 */
	const Thread* getCurrentThread() const;
	/**
	 * Obtain an ID of the main thread of the program.
	 *
	 * @return ID of the main thread of the program.
	 */
	Thread::ID getMainThreadID() const;
	/**
	 * Set main thread of the program.
	 * This may be done only once
	 *
	 * @param id ID of the main thread of the program
	 */
	void setMainThreadID(Thread::ID id);
	/**
	 * Obtain a main thread of the program.
	 *
	 * @return main thread of the program
	 */
	const Thread* getMainThread() const;
	/**
	 * Create and add a new thread into threads managed by this class.
	 *
	 * @param gimple holder of the GIMPLE++ representation of the progra
	 * @param memoryState simulated memory
	 * @param cstManager manager of constants of the program
	 * @param globalsManager manager of global variables of the progra
	 * @param program the intepreted program
	 * @param threadInfoRef reference to the info structure for thread to create add to managed threads
	 * @param threadID unique ID of the new thread or -1, in which case an unique ID will be assigned
	 * to newly created thread and returned by this function
	 * @return unique ID of the newly created thread
	 */
	Thread::ID newThread(giterator::Gimple* gimple, mmodule::MemoryState* memoryState,
			 	   	     ConstantsManager* cstManager, GlobalsManager* globalsManager,
			 	   	     Program* program, mmodule::Ref<ThreadInfo> threadInfoRef,
			 	   	     Thread::ID threadID = -1);
	/**
	 * Terminate thread with specified ID.
	 *
	 * @param threadID ID of thread to be terminated
	 */
	void terminateThread(Thread::ID threadID);
	/**
	 * Join two threads.
	 *
	 * Joining two threads means block execution of one thread until the joined thread is finished.
	 * If joined thread has already terminated, function returns immediately with error code.
	 * Deadlock may occur, which is detected and exception is thrown when deadlock detected.
	 *
	 * @param blocking ID of the thread (current thread), that is joining thread with ID @c toTerminate
	 * @param toTerminate the joined thread
	 * @return 0 if successful, ESRCH when trying to join with non-existing thread, EDEADLK when thread
	 * calls join on itself, or EINVAL when another thread is already waiting to join with this thread
	 * @throws DeadlockException, when deadlock occured = there is a cycle in graph of waiting threads
	 *
	 * @note pthread_join() returns EDEADLK also when two threads tried to join with each other, but is
	 * unable to detect more complex deadlocks. This function throws @c DeadlockException in this case
	 * as well as the more complex deadlocks.
	 */
	int joinThreads(Thread::ID blocking, Thread::ID toTerminate);
	/**
	 * Picks first runnable thread from threads managed by this class
	 * and sets it as current thread.
	 */
	void pickAndSetRunnableThread();
	/**
	 * Create and new mutex and add it to mutexes managed by this class.
	 *
	 * @param mutexID of the created mutex. If -1, new ID gets created and
	 * the new ID value is stored to this variable, otherwise the value is used
	 * @param mutexInfoRef reference to the info structure of this mutex
	 * @return 0 if successful, or EINVAL when trying to
	 * initialize an already initialized mutex
	 *
	 * @see Mutex::Mutex, initMutex
	 */
	int newMutex(Mutex::ID& mutexID, mmodule::Ref<MutexInfo> mutexInfoRef);
	/**
	 * Set a mutex state from uninitialized to unlocked.
	 *
	 * @param mutexID unique ID of the mutex to initialize
	 *
	 * @see Mutex::init
	 *
	 */
	void initMutex(Mutex::ID mutexID);
	/**
	 * Destroy mutex managed by this class.
	 * When value other than zero is returned, warning message is reported.
	 *
	 * @param mutexID unique ID of the mutex to destroy
	 * @return 0 if successful, EINVAL when trying to destroy uninitialized mutex,
	 * EBUSY when trying to destroy locked or used by thread condition mutex
	 *
	 * @see Mutex::destroy
	 */
	int destroyMutex(Mutex::ID mutexID);
	/**
	 * Lock a mutex managed by this class.
	 * If the mutex is currently held by another thread this call blocks and waits until the mutex
	 * becomes available. A thread should not attempt to lock a Mutex that it already owns.
	 * When value other than zero is returned, warning message is reported.
	 *
	 * @param mutexID unique ID of the mutex to lock
	 * @returns 0 if successful, EINVAL if trying to lock uninitialized mutex,
	 * or EDEADLK when trying to lock a locked mutex by the same thread
	 *
	 * @see Mutex::lock
	 */
	int lockMutex(Mutex::ID mutexID);
	/**
	 * Attempt to lock mutex managed by this class.
	 * When value other than zero is returned, warning message is reported.
	 *
	 * @param mutexID unique ID of the mutex to try to lock
	 * @return 0 if the locking was successful,
	 * EINVAL if trying to lock uninitialized mutex,
	 * or EBUSY if the mutex is locked
	 *
	 * @see lockMutex
	 */
	int tryLockMutex(Mutex::ID mutexID);
	/**
	 * Unlock mutex managed by this class.
	 * When value other than zero is returned, warning message is reported.
	 *
	 * @param mutexID unique ID of the mutex to unlock
	 * @return 0 if successful, EINVAL if trying to unlock uninitialized mutex,
	 * EPERM when trying to unlock a mutex that is not owned by current thread
	 *
	 * @see Mutex::unlock
	 */
	int unlockMutex(Mutex::ID mutexID);
	/**
	 * Create a new thread condition and add it to
	 * thread conditions managed by this class.
	 *
	 * @param condID of the created condition. If -1, new ID gets created and
	 * the new ID value is stored to this variable, otherwise
	 * the value is used
	 * @param conditionInfoRef reference to the info structure of this condition
	 * @return 0 if successful, or EINVAL when trying to
	 * initialize an already initialized condition
	 *
	 * @see Condition::Condition, initCondition
	 */
	int newCondition(Condition::ID& condID, mmodule::Ref<ConditionInfo> conditionInfoRef);
	/**
	 * Set a condition state from uninitialized to initialized.
	 *
	 * @param condID unique ID of the condition to initialize
	 * @see Condition::init
	 */
	void initCondition(Condition::ID condID);
	/**
	 * Destroy thread condition managed by this class.
	 *
	 * @param condID unique ID of the thread condition to destroy
	 * @return 0 if successful, EINVAL when trying to destroy uninitialized condition,
	 * EBUSY when trying to destroy condition that is in use (by some condition wait function)
	 *
	 * @see Condition::destroy
	 */
	int destroyCondition(Condition::ID condID);
	/**
	 * Wake up a thread that waits on a thread condition variable (if there is any).
	 * Should be called after mutex is locked, and must unlock mutex for @c waitCondition()
	 * to complete.
	 *
	 * @param condID unique ID of the thread condition to wake up a thread
	 * @return 0 if successful, or EINVAL when trying to signal uninitialized condition.
	 *
	 * @see waitCondition
	 */
	int signalCondition(Condition::ID condID);
	/**
	 * Block a thread until specified condition is signalled.
	 * Should be called while mutex is locked.
	 * Release mutex while waiting.
	 * After signal recieved and thread awakened,
	 * mutex will be automatically locked for use by thread.
	 *
	 * @param condID condition that blocks the thread
	 * @param mutexID ID of the associated mutex
	 * @return 0 if successful, EINVAL when either condition with ID @c condID
	 * or mutex with ID @c mutexID is uninitialized, EINVAL also when different mutexes
	 * were supplied for concurrent condtion wait operations on the same condition variable,
	 * or EPERM when mutex with ID @c mutexID is not locked by current thread (or not locked at all).
	 *
	 * @see signalCondition
	 */
	int waitCondition(Condition::ID condID, Mutex::ID mutexID);
	/**
	 * From all sleeping threads, return the smallest time interval until the thread will wake up.
	 * (If there are two threads T1 T2, T1 sleeps for 5 seconds, T2 for 3, 1 second is already gone,
	 * this fction will return 2).
	 *
	 * @return the smallest time interval until thread that is closest to wake up will wake up
	 */
	unsigned int getMinRemainingSleepInterval();
	/**
	 * Set a name of a thread with a specified ID (name of thread variable holding the thread, or 'main').
	 *
	 * @param threadID unique ID of the thread to set the name
	 * @param name name of the thread with ID @c threadID
	 */
	void addThreadName(Thread::ID threadID, std::string name);
	/**
	 * Obtain a name of a thread with a specified ID (name of thread variable holding the thread, or 'main').
	 *
	 * @param threadID unique ID of the thread to obtain a name
	 * @return name name of the thread with ID @c threadID
	 */
	std::string getThreadName(Thread::ID threadID) const;
	/**
	 * Print information about threads managed by this class into LOG_DEBUG
	 */
	void print();// const;
private:
	/**
	 * Delete a thread with specified ID from threads managed by this class.
	 *
	 * @param threadID ID of thread to be deleted
	 */
	void deleteThread(Thread::ID threadID);
	/**
	 * Perform early deadlock detection by cycle detection in
	 * the waiting for graph, starting with specified thread.
	 *
	 * @param threadID thread to begin the detection with
	 * @throws IDeadlockException when deadlock occurs
	 */
	void checkDeadlockFrom(Thread::ID threadID) const;
};

}
#endif /* INT_THREADPOOL_H */
