// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file Timer.h
 * Declaration of a class counting specified periods of time.
 */

#ifndef INT_TIMER_H
#define INT_TIMER_H

#include <ctime>

namespace ginterpreter
{

/**
 * Class for computing interval of time elapsed from a specific start time.
 * Member function elapsed() returns true after specified interval in seconds
 * elapsed since specified start time.
 * Used to interpret sleep() function for program in interpretation mode.
 */
class Timer
{
private:
	/*! Interval between start time of timer and end time. */
	unsigned int m_interval;
	/*! Start time of timer */
	time_t m_starttime;
	/*! If the timer was already started */
	bool m_started;
public:
	/**
	 * Constructor, creates a new unstarted timer.
	 */
	Timer();

	/**
	 * Set interval between start time of timer and end time.
	 * Timer returns elapsed() == true when current time >= start time + interval.
	 * Should never be called for already started timer.
	 *
	 * @param interval interval between start time and end time of timer.
	 */
	void setInterval(unsigned int interval);

	/**
	 * Start the timer. Set start time to current time.
	 * Should never be called for already started timer.
	 */
	void start();

	/**
	 * Find out whether the timer was already started.
	 *
	 * @return true if the timer was already started, false otherwise
	 */
	bool isStarted() const;

	/**
	 * Find out whether interval has elapsed since start time.
	 *
	 * @return true if interval has elapsed since start time, false if not
	 * or if the timer is not started
	 */
	bool elapsed() const;

	/**
	 * Get number of seconds remaining for timer to be elapsed.
	 * Remaining time = start time + interval - current time.
	 *
	 * @return number of seconds remaining for timer to be elapsed,
	 * 0 if the timer is not started
	 */
	unsigned int getRemainingTime() const;

	/**
	 * Uninitialize the timer.
	 * After this, the timer is not started.
	 * Should neved be called while interval is not elapsed.
	 */
	void clear();
};

}


#endif /* INT_TIMER_H */
