// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file Value.h
 * Base class of all classes representing values.
 */
#ifndef INT_VALUE_H
#define INT_VALUE_H

#include "../git/git_fwd.h"
#include "../mm/Content.h"

#include <iosfwd>
#include <memory>


namespace ginterpreter {

/**
 * Interface for all classes representing values.
 * It implements Memory Module's interface @c Content,
 * and it is used to store all kinds of values into the
 * simulated memory.
 * Subclasses of this class represent the concrete stored
 * values
 */
class Value : virtual public mmodule::Content {
  public:
	/**
	 * Destructor
	 */
    virtual ~Value() {}

    /**
     * Returns the type of the value.
     */
    virtual const giterator::Type* getType() const = 0;
    
    /**
     * Clones this value.
     */
    virtual Value* clone() const = 0;

    /**
     * Prints this value into @c os.
     */
    virtual void print (std::ostream& os) const = 0;
    
    /** @name Relation operators
     * The relation operators on value.
     */
    //@{
    virtual bool operator==(const Value& v) const = 0;
    virtual bool operator<(const Value& v) const = 0;

    virtual bool operator<=(const Value& v) const = 0;
    virtual bool operator>(const Value& v) const = 0;
    virtual bool operator>=(const Value& v) const = 0;
    virtual bool operator!=(const Value& v) const = 0;
    virtual bool orderedWith(const Value& v) const = 0;
    //@}
    
    /** @name Arithmetic and logic operators
     * The arithmetic and logic operators on the value.
     * These operators correspond to the arithmetic and
     * logical operators used in GIMPLE.
     */
    //@{
    virtual std::auto_ptr<Value> addOp (const Value& v) const = 0;
    virtual std::auto_ptr<Value> subOp (const Value& v) const = 0;
    virtual std::auto_ptr<Value> multOp (const Value& v) const = 0;
    virtual std::auto_ptr<Value> truncModOp (const Value& v) const = 0;
    virtual std::auto_ptr<Value> floorModOp (const Value& v) const = 0;
    virtual std::auto_ptr<Value> ceilModOp (const Value& v) const = 0;
    virtual std::auto_ptr<Value> roundModOp (const Value& v) const = 0;
    virtual std::auto_ptr<Value> truncDivOp (const Value& v) const = 0;
    virtual std::auto_ptr<Value> floorDivOp (const Value& v) const = 0;
    virtual std::auto_ptr<Value> ceilDivOp (const Value& v) const = 0;
    virtual std::auto_ptr<Value> roundDivOp (const Value& v) const = 0;
    virtual std::auto_ptr<Value> rDivOp (const Value& v) const = 0;
    virtual std::auto_ptr<Value> absOp () const = 0;
    virtual std::auto_ptr<Value> lShiftOp (const Value& v) const = 0;
    virtual std::auto_ptr<Value> rShiftOp (const Value& v) const = 0;
    virtual std::auto_ptr<Value> lRotateOp (const Value& v) const = 0;
	virtual std::auto_ptr<Value> rRotateOp (const Value& v) const = 0;
    virtual std::auto_ptr<Value> bitNotOp () const = 0;
    virtual std::auto_ptr<Value> bitIorOp (const Value& v) const = 0;
    virtual std::auto_ptr<Value> bitXorOp (const Value& v) const = 0;
    virtual std::auto_ptr<Value> bitAndOp (const Value& v) const = 0;
    virtual std::auto_ptr<Value> truthNotOp () const = 0;
	virtual std::auto_ptr<Value> truthOrOp (const Value& v) const = 0;
	virtual std::auto_ptr<Value> truthXorOp (const Value& v) const = 0;
	virtual std::auto_ptr<Value> truthAndOp (const Value& v) const = 0;
	virtual std::auto_ptr<Value> negateOp () const = 0;
    //@}
};

}

/**
 * Operator for printing values into @c ostream.
 *
 * @param os the stream to print the value to
 * @param value the printed value
 * @return the stream where the value was printed
 */
inline std::ostream& 
operator<< (std::ostream& os, const ginterpreter::Value& value)
{
    value.print (os);
    return os;
}

#endif /* INT_VALUE_H */

