// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file ValueReader.cpp
 * Definitions for class that allows to read values
 * represented by an operand from the simulated memory.
 */
#include "ValueReader.h"

#include "AddrResolver.h"
#include "ConstantsManager.h"

#include "PtrValue.h"
#include "RecordValue.h"
#include "ArrayValue.h"
#include "ArrayConstructorValue.h"
#include "RecordConstructorValue.h"

#include "../git/Operands.h"
#include "../git/Decl.h"
#include "../git/ArrayType.h"
#include "../git/RecordType.h"
#include "../git/IntegerConstant.h"

#include "../mm/MemoryState.h"

using namespace mmodule;
using namespace giterator;

namespace ginterpreter {

ValueReader::ValueReader (const FunctionValue* func, const MemoryState* memoryState, 
                          const ConstantsManager* cstManager, const GlobalsManager* globalsManager,
                          Ptr frame)
    : m_func (func), m_memoryState (memoryState),
      m_cstManager (cstManager), m_globalsManager (globalsManager),
      m_frame (frame), m_value (0) 
{
}

std::auto_ptr<Value> 
ValueReader::acquireValue ()
{
    if (m_valueOwner.get () == 0)
    {
        m_valueOwner.reset (m_value->clone ());
        m_value = m_valueOwner.get ();
    }
    
    return m_valueOwner;
}


void
ValueReader::visit(const AddrOperand& op)
{
    AddrResolver addRes (m_func, m_memoryState, m_cstManager, m_globalsManager, m_frame);
    
    op.getSubOperand()->accept (addRes);
    
    m_valueOwner.reset (new PtrValue (op.getType (), addRes.getPtr ()));
    m_value = m_valueOwner.get ();
}


void
ValueReader::visit(const DeclOperand& op)
{
    AddrResolver addrRes (m_func, m_memoryState, m_cstManager, m_globalsManager, m_frame);
    op.accept (addrRes);

    m_value = resolveValue(addrRes.getPtr(), op.getDecl()->getType(), addrRes.getDeclOffset());
}

void
ValueReader::visit(const LabelExprOperand& op)
{
	AddrResolver addrRes (m_func, m_memoryState, m_cstManager, m_globalsManager, m_frame);
	op.accept (addrRes);

	m_value = loadValue (addrRes.getPtr ());
}


void
ValueReader::visit(const IndirectRefOperand& op)
{
    AddrResolver addrRes (m_func, m_memoryState, m_cstManager, m_globalsManager, m_frame);
    op.accept (addrRes);
    
    m_value = resolveValue(addrRes.getPtr(), op.getType(), 0);
}
        


void
ValueReader::visit(const ArrayRefOperand& op)
{
    AddrResolver addrRes (m_func, m_memoryState, m_cstManager, m_globalsManager, m_frame);
    op.accept (addrRes);
    
    m_value = resolveValue(addrRes.getPtr(), op.getType(), 0);
}

void
ValueReader::visit(const ComponentRefOperand& op)
{
    AddrResolver addrRes (m_func, m_memoryState, m_cstManager, m_globalsManager, m_frame);
    op.accept (addrRes);

    m_value = resolveValue(addrRes.getPtr(), op.getField()->getType(), addrRes.getDeclOffset());
}
    

void
ValueReader::visit(const CstOperand& op)
{
    AddrResolver addrRes (m_func, m_memoryState, m_cstManager, m_globalsManager, m_frame);
    op.accept (addrRes);
    
    m_value = loadValue (addrRes.getPtr ());
}

void
ValueReader::visit (const giterator::ArrayConstructorOperand& op)
{
	ArrayConstructorValue* outValue = new ArrayConstructorValue(op.getType ());

	std::pair<ArrayConstructorOperand::ArrayConstructorsIterator,
			ArrayConstructorOperand::ArrayConstructorsIterator> c = op.constructors ();
		ArrayConstructorOperand::ArrayConstructorsIterator it;
	for (it = c.first; it != c.second; ++it)
	{
		// TODO: set ArrayConstructorOperand key to operand and read via addrresolver
		const Value* index = dynamic_cast<const Value*> (m_memoryState->load(m_cstManager->getCstPtr(it->first)));

		ValueReader valReader(m_func, m_memoryState, m_cstManager, m_globalsManager, m_frame);
		it->second->accept(valReader);

		outValue->addValue(index, valReader.acquireValue());
	}

	m_value = outValue;
}

void
ValueReader::visit (const giterator::RecordConstructorOperand& op)
{
	RecordConstructorValue* outValue = new RecordConstructorValue(op.getType ());

	std::pair<RecordConstructorOperand::RecordConstructorsIterator,
		RecordConstructorOperand::RecordConstructorsIterator> c = op.constructors ();
	RecordConstructorOperand::RecordConstructorsIterator it;
	for (it = c.first; it != c.second; ++it)
	{
		ValueReader valReader (m_func, m_memoryState, m_cstManager, m_globalsManager, m_frame);
		it->second->accept (valReader);

		outValue->addValue(it->first, valReader.acquireValue());
	}

	m_value = outValue;
}

const Value*
ValueReader::resolveValue (Ptr ptr, const Type* type, size_t offset)
{
	const ArrayType* arrType = dynamic_cast<const ArrayType*> (type);
	const RecordType* recType = dynamic_cast<const RecordType*> (type);
	if (recType)
	{
		m_valueOwner.reset(new RecordValue(recType, ptr,
										   recType->getUnitSize (), offset));
		return m_valueOwner.get();
	}
	else if (arrType)
	{
		m_valueOwner.reset(new ArrayValue(arrType, ptr, arrType->getUnitSize()));
		return m_valueOwner.get();
	}
	else
	{
		return loadValue(ptr);
	}
}

const Value*
ValueReader::loadValue (Ptr ptr)
{
    const Value* value = dynamic_cast<const Value*> (m_memoryState->load (ptr));
    assert (value);
    return value;
}

}
