// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file ValueReader.h
 * Declaration of class that allows to read values
 * represented by an operand from the simulated memory.
 */
#ifndef INT_VALUE_READER_H
#define INT_VALUE_READER_H

#include "../git/git_fwd.h"
#include "../git/Operand.h"
#include "../mm/fwd.h"
#include "../mm/Ptr.h"

#include "Value.h"

#include <memory>


namespace ginterpreter {
    
class Value;
class StackFrame;
class FunctionValue;
class ConstantsManager;
class GlobalsManager;


/**
 * Class that allows to read values represented by an operand from the simulated memory.
 * To get the value represented by operand @c o, an object of this class
 * is supplied as a parameter to the @c Operand::accept() method of object @c o.
 * The found value can be acquired  by call to @c getValue() and @c acquireValue().
 * This class implements the Visitor interface for Operand class.
 * The implemented visitor methods realize the search for the values.
 *
 * @see AddrResolver, ValueWriter
 */
class ValueReader : public giterator::Operand::Visitor {
  private:
	/*! Current function */
    const FunctionValue* m_func;
    /*! Interface to the simulated memory */
    const mmodule::MemoryState* m_memoryState;
    /*! Constants and function declarations of the program */
    const ConstantsManager* m_cstManager;
    /*! Global declarations of the program */
    const GlobalsManager* m_globalsManager;
    /*! pointer to the current stack frame of the current function */
    mmodule::Ptr m_frame;
    /*! the found value */
    const Value* m_value;
    /*! the found value, owned by this class */
    std::auto_ptr<Value> m_valueOwner;
        
  public:
    /**
     * Constructor.
     * 
     * @param func current function
     * @param memoryState interface to the simulated memory
     * @param cstManager constants and function declarations of the program
     * @param globalsManager global declarations of the program
     * @param frame pointer to the beginning of the current stack frame
     */
    ValueReader (const FunctionValue* func, const mmodule::MemoryState* memoryState, 
                 const ConstantsManager* cstManager, const GlobalsManager* globalsManager,
                 mmodule::Ptr frame);
    
    /**
     * Returns the value read by this reader.
     * @return value read by this reader; after destruction of this object, 
     *      returned pointer must not be used
     */
    const Value* getValue () 
    { return m_value; }
    
    /**
     * Returns the value read by this reader. The reader
     * no longer ownes the value.
     *
     * @return value read by this reader; after destruction of this object,
     *      returned pointer can be used
     */
    std::auto_ptr<Value> acquireValue ();

    void visit (const giterator::AddrOperand&);
    void visit (const giterator::DeclOperand&);
    void visit (const giterator::LabelExprOperand&);
    void visit (const giterator::IndirectRefOperand&);
    void visit (const giterator::ArrayRefOperand&);
    void visit (const giterator::ComponentRefOperand&);
    void visit (const giterator::CstOperand&);
    void visit (const giterator::ArrayConstructorOperand&);
    void visit (const giterator::RecordConstructorOperand&);

  private:
    /**
     * Load a value from a pointer to a simulated memory.
     *
     * @param ptr pointer to the simulated memory where to load the value
     * @return value at pointer @c ptr in the simulated memory
     */
    const Value* loadValue (mmodule::Ptr ptr);
    /**
	 * Load a value from a pointer to a simulated memory
	 * according to a type of the value.
	 *
	 * @param ptr pointer to the simulated memory where to load the value
	 * @param type the type of the loaded value
	 * @param offset the offset of the value
	 * @return value at pointer @c ptr in the simulated memory
	 */
    const Value* resolveValue (mmodule::Ptr ptr, const giterator::Type* type, size_t offset);
};

}

#endif /* INT_VALUE_READER_H */
