// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file ValueWriter.cpp
 * Definitions for class that allows to store a value using an operand.
 */
#include "ValueWriter.h"

#include "AddrResolver.h"
#include "../mm/MemoryState.h"

#include "../git/Operands.h"

#include "Value.h"
#include "IntegerValue.h"
#include "RecordValue.h"
#include "ArrayValue.h"
#include "ArrayConstructorValue.h"
#include "RecordConstructorValue.h"
#include "ConstantsManager.h"
#include "GlobalsManager.h"

using namespace mmodule;
using namespace giterator;


namespace ginterpreter {
    
ValueWriter::ValueWriter (const FunctionValue* func, MemoryState* memoryState, 
                          ConstantsManager* cstManager, GlobalsManager* globalsManager,
                          Ptr frame, std::auto_ptr<Value> value) :
        m_func (func), m_memoryState (memoryState),
        m_cstManager (cstManager), m_globalsManager (globalsManager),
        m_frame (frame), m_value (value)
{}

void
ValueWriter::storeViaPointer (mmodule::Ptr ptr)
{
	const ArrayValue* arrVal = dynamic_cast<const ArrayValue*> (m_value.get());
	const RecordValue* recVal = dynamic_cast<const RecordValue*> (m_value.get());
	const ArrayConstructorValue* constVal = dynamic_cast<const ArrayConstructorValue*> (m_value.get());
	const RecordConstructorValue* recConstVal = dynamic_cast<const RecordConstructorValue*> (m_value.get());
	if (arrVal)
	{
		std::pair<MemoryState::ContentIterator, MemoryState::ContentIterator> values =
				m_memoryState->loadOverlapping (arrVal->getPtr(), arrVal->getUnitSize());
		uint64_t index = 0;
		for (MemoryState::ContentIterator i1 = values.first; i1 != values.second; ++i1)
		{
			const Value* storeVal = dynamic_cast<const Value*> (i1->second);
			assert(storeVal);
			const giterator::ArrayType* arrayType =
								dynamic_cast<const giterator::ArrayType*> (arrVal->getType());
			assert (arrayType);
			const giterator::Type* elementType = arrayType->getElementType ();
			size_t elementSize = elementType->getUnitSize ();
			m_memoryState->store(ptr + index++ * elementSize, std::auto_ptr<Content> (storeVal->clone()));
		}
	}
	else if (recVal)
	{
		std::pair<MemoryState::ContentIterator, MemoryState::ContentIterator> values =
					m_memoryState->loadOverlapping (recVal->getPtr(), recVal->getUnitSize());
		std::map<size_t, const Value*> valueMap;

		for (MemoryState::ContentIterator i1 = values.first; i1 != values.second; ++i1)
		{
			const Value* storeVal = dynamic_cast<const Value*> (i1->second);
			assert(storeVal);
			valueMap[i1->first - recVal->getOffset()] = storeVal;
		}

		for (std::map<size_t, const Value*>::const_iterator i1 = valueMap.begin(); i1 != valueMap.end(); ++i1)
		{
			m_memoryState->store(ptr + i1->first, std::auto_ptr<Content> (i1->second->clone()));
		}
	}
	else if(constVal)
	{
		Ptr arrayPtr = ptr;

		const giterator::ArrayType* arrayType =
					dynamic_cast<const giterator::ArrayType*> (constVal->getType());
		assert (arrayType);
		const giterator::Type* elementType = arrayType->getElementType ();
		size_t elementSize = elementType->getUnitSize ();

		for (ArrayConstructorValue::ValuesIterator i1 = constVal->values().first; i1 != constVal->values().second; ++i1)
		{
			const IntegerValue* index = dynamic_cast<const IntegerValue*> (i1->first);
			assert (index);

			ValueWriter vw(m_func, m_memoryState, m_cstManager, m_globalsManager, m_frame, std::auto_ptr<Value> (i1->second->clone()));
			vw.storeViaPointer(arrayPtr + index->getValue () * elementSize);
		}
	}
	else if (recConstVal)
	{

		Ptr recPtr = ptr;
		for (RecordConstructorValue::ValuesIterator i1 = recConstVal->values().first; i1 != recConstVal->values().second; ++i1)
		{
			AddrResolver addRes (m_func, m_memoryState, m_cstManager, m_globalsManager, m_frame);
			i1->first->accept(addRes);
			ValueWriter vw(m_func, m_memoryState, m_cstManager, m_globalsManager, m_frame, std::auto_ptr<Value> (i1->second->clone()));
			vw.storeViaPointer(addRes.getPtr());
		}
	}
	else
	{
		m_memoryState->store (ptr, std::auto_ptr<Content> (m_value));
	}
}

void
ValueWriter::storeViaPointer (mmodule::Ptr ptr, const Decl* decl)
{
	const RecordConstructorValue* recConstVal = dynamic_cast<const RecordConstructorValue*> (m_value.get());
	if (recConstVal)
	{
		for (RecordConstructorValue::ValuesIterator i1 = recConstVal->values().first; i1 != recConstVal->values().second; ++i1)
		{
			const OffsetDecl* od = dynamic_cast<const OffsetDecl*> (decl);
			assert(od);
			size_t offset = od->getFieldsMapper()->getOffset(i1->first->getUid());
			ValueWriter vw(m_func, m_memoryState, m_cstManager, m_globalsManager, m_frame, std::auto_ptr<Value> (i1->second->clone()));
			vw.storeViaPointer(ptr + offset, i1->first);
		}
	}
	else
	{
		m_memoryState->store (ptr, std::auto_ptr<Content> (m_value));
	}
}

size_t resolveElementUnitSize(const giterator::ArrayType* type)
{
	const giterator::ArrayType* t = type;

	while(true)
	{
		const Type* elemType = t->getElementType();

		t = dynamic_cast<const giterator::ArrayType*> (elemType);

		if (t == NULL)
			return elemType->getUnitSize();
	}
}

void
ValueWriter::storeViaPointer (const Operand& op)
{
	const ArrayValue* arrVal = dynamic_cast<const ArrayValue*> (m_value.get());
	const RecordValue* recVal = dynamic_cast<const RecordValue*> (m_value.get());
	const ArrayConstructorValue* constVal = dynamic_cast<const ArrayConstructorValue*> (m_value.get());
	const RecordConstructorValue* recConstVal = dynamic_cast<const RecordConstructorValue*> (m_value.get());
	if (arrVal)
	{
		std::pair<MemoryState::ContentIterator, MemoryState::ContentIterator> values =
						m_memoryState->loadOverlapping (arrVal->getPtr(), arrVal->getUnitSize());
		Ptr ptr = getAddress(op);
		uint64_t index = 0;
		size_t elementSize = resolveElementUnitSize(dynamic_cast<const giterator::ArrayType*> (arrVal->getType()));
		for (MemoryState::ContentIterator i1 = values.first; i1 != values.second; ++i1)
		{
			const Value* storeVal = dynamic_cast<const Value*> (i1->second);
			assert(storeVal);

			/*const giterator::ArrayType* arrayType =
								dynamic_cast<const giterator::ArrayType*> (arrVal->getType());
			assert (arrayType);
			const giterator::Type* elementType = arrayType->getElementType ();
			size_t elementSize = elementType->getUnitSize ();*/

			m_memoryState->store(ptr + index++ * elementSize, std::auto_ptr<Content> (storeVal->clone()));
		}
	}
	else if (recVal)
	{
		std::pair<MemoryState::ContentIterator, MemoryState::ContentIterator> values =
					m_memoryState->loadOverlapping (recVal->getPtr(), recVal->getUnitSize());

		std::map<size_t, const Value*> valueMap;

		Ptr ptr = getAddress(op);

		for (MemoryState::ContentIterator i1 = values.first; i1 != values.second; ++i1)
		{
			const Value* storeVal = dynamic_cast<const Value*> (i1->second);
			assert(storeVal);
			valueMap[i1->first - recVal->getOffset()] = storeVal;
		}

		for (std::map<size_t, const Value*>::const_iterator i1 = valueMap.begin(); i1 != valueMap.end(); ++i1)
		{
			m_memoryState->store(ptr + i1->first, std::auto_ptr<Content> (i1->second->clone()));
		}
	}
	else if(constVal)
	{
		AddrResolver addRes (m_func, m_memoryState, m_cstManager, m_globalsManager, m_frame);
		op.accept (addRes);
		Ptr arrayPtr = addRes.getPtr ();

		const giterator::ArrayType* arrayType =
		            dynamic_cast<const giterator::ArrayType*> (constVal->getType());
		assert (arrayType);
		const giterator::Type* elementType = arrayType->getElementType ();
		size_t elementSize = elementType->getUnitSize ();

		for (ArrayConstructorValue::ValuesIterator i1 = constVal->values().first; i1 != constVal->values().second; ++i1)
		{
			const IntegerValue* index = dynamic_cast<const IntegerValue*> (i1->first);
			assert (index);

			ValueWriter vw(m_func, m_memoryState, m_cstManager, m_globalsManager, m_frame, std::auto_ptr<Value> (i1->second->clone()));
			vw.storeViaPointer(arrayPtr + index->getValue () * elementSize);
		}
	}
	else if (recConstVal)
	{
		for (RecordConstructorValue::ValuesIterator i1 = recConstVal->values().first; i1 != recConstVal->values().second; ++i1)
		{
			AddrResolver addRes (m_func, m_memoryState, m_cstManager, m_globalsManager, m_frame);
			op.accept(addRes);
			i1->first->accept(addRes);

			ValueWriter vw(m_func, m_memoryState, m_cstManager, m_globalsManager, m_frame, std::auto_ptr<Value> (i1->second->clone()));
			vw.storeViaPointer(addRes.getPtr(), i1->first);
		}
	}
	else
	{
		AddrResolver addrRes (m_func, m_memoryState, m_cstManager, m_globalsManager, m_frame);
		op.accept (addrRes);
		Ptr ptr = addrRes.getPtr ();
		assert(!ptr.isNull());

		m_memoryState->store (ptr, std::auto_ptr<Content> (m_value));
	}
}

mmodule::Ptr ValueWriter::getAddress(const Operand& op)
{
	AddrResolver addrRes (m_func, m_memoryState, m_cstManager, m_globalsManager, m_frame);
	op.accept (addrRes);
	return addrRes.getPtr ();
}


void
ValueWriter::visit (const AddrOperand&)
{
    abort ();
}


void
ValueWriter::visit (const DeclOperand& op)
{
	storeViaPointer (op);
}

void
ValueWriter::visit (const LabelExprOperand& op)
{
	storeViaPointer (op);
}


void
ValueWriter::visit (const IndirectRefOperand& op)
{
    storeViaPointer (op);
}


void
ValueWriter::visit (const ArrayRefOperand& op)
{
    storeViaPointer (op);
}

void
ValueWriter::visit (const ComponentRefOperand& op)
{
    storeViaPointer (op);
}

void
ValueWriter::visit (const CstOperand&)
{
    abort ();
}

void ValueWriter::visit (const ArrayConstructorOperand& op)
{
	// TODO:
	abort ();
}

void ValueWriter::visit (const RecordConstructorOperand& op)
{
	// TODO:
	abort ();
}

}
