// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/** @file ValueWriter.h
 * Declaration of class that allows to store a value using an operand.
 */
#ifndef INT_VALUE_WRITER_H
#define INT_VALUE_WRITER_H

#include "../git/Operand.h"
#include "../mm/fwd.h"
#include "../mm/Ptr.h"

#include <memory>

namespace ginterpreter {
    
class FunctionValue;
class ConstantsManager;
class GlobalsManager;
class Value;

/**
 * Class that allows to write values represented by an operand in the simulated memory.
 * To write a value represented by operand @c o, an object of this class
 * is supplied as a parameter to the @c Operand::accept() method of object @c o.
 * The value to write is passed to the constructor of this class.
 *
 * This class implements the Visitor interface for Operand class.
 * The implemented visitor methods realize the writing the values.
 *
 * TODO: This class is a mess, needs a cleanup.
 *
 * @see AddrResolver, ValueReader
 */
class ValueWriter : public giterator::Operand::Visitor {
  private:
	/*! Current function */
    const FunctionValue* m_func;
    /*! Interface to the simulated memory */
    mmodule::MemoryState* m_memoryState;
    /*! Constants and function declarations of the program */
    ConstantsManager* m_cstManager;
    /*! Global declarations of the program */
    GlobalsManager* m_globalsManager;
    /*! pointer to the current stack frame of the current function */
    mmodule::Ptr m_frame;
    /*! the value to store */
    std::auto_ptr<Value> m_value;
  public:
    /**
     * Constructor.
     *
     * @param func current function
     * @param memoryState interface to the simulated memory
     * @param cstManager constants and function declarations of the program
     * @param globalsManager global declarations of the program
     * @param frame pointer to the beginning of the current stack frame
     * @param value the value to store
     */
    ValueWriter (const FunctionValue* func, mmodule::MemoryState* memoryState, 
                 ConstantsManager* cstManager, GlobalsManager* globalsManager,
                 mmodule::Ptr frame, std::auto_ptr<Value> value);
    
    void visit (const giterator::AddrOperand&);
    void visit (const giterator::DeclOperand&);
    void visit (const giterator::LabelExprOperand&);
    void visit (const giterator::IndirectRefOperand&);
    void visit (const giterator::ArrayRefOperand&);
    void visit (const giterator::ComponentRefOperand&);
    void visit (const giterator::CstOperand&);
    void visit (const giterator::ArrayConstructorOperand&);
	void visit (const giterator::RecordConstructorOperand&);

	/**
	 * Stores a value @c m_value using a pointer.
	 *
	 * @param ptr the pointer at which to store
	 */
	void storeViaPointer (mmodule::Ptr ptr);
	/**
	 * Stores a value @c m_value using a pointer and declaration.
	 * If the stored value is @c RecordConstructorValue,
	 * the declaration of a struct has to be used for proper storage
	 *
	 * @param ptr the pointer at which to store
	 * @param decl the declaration of a struct associated with
	 * @c RecordConstructorValue
	 */
	void storeViaPointer (mmodule::Ptr ptr, const giterator::Decl* decl);
  private:
	/**
	 * Stores a value @c m_value using an operand.
	 *
	 * @param op the operand at which's address to store the value
	 */
    void storeViaPointer (const giterator::Operand& op);
    /**
     * Get an address of an operand in the simulated memory.
     *
     * @param op the operand to get the addres of
     */
    mmodule::Ptr getAddress(const giterator::Operand& op);
};

}

#endif /* INT_VALUE_WRITER_H */
