// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

#include "AreaEntry.h"

#include "Content.h"
#include "RefContentEntry.h"
#include "HashHolder.h"
#include "MemoryExceptions.h"
#include "../Error.h"

#include <typeinfo>


namespace mmodule { namespace detail {
    

/**
 * Constructor. Creates an expandable area.
 *
 * Complexity: O(1)
 */
AreaEntry::AreaEntry ()
    : m_bfsTreeLinkToParent (0),
      m_size (0), m_deleted (false), m_color (0),
      m_wasCreated (true), m_wasDeleted (false),
      m_expandable (true),
      m_unreachableReported (false)
{
}

/**
 * Constructor.
 * 
 * Complexity: O(1)
 * 
 * @param size size of the area in bytes
 */
AreaEntry::AreaEntry (size_t size) 
    : m_bfsTreeLinkToParent (0), 
      m_size (size), m_deleted (false), m_color (0), 
      m_wasCreated (true), m_wasDeleted (false),
      m_expandable (false),
      m_unreachableReported (false)
{
}

/**
  * Destructor.
  */
AreaEntry::~AreaEntry ()
{
    LOG_DEBUG("Clearing content list of size: %zu", m_content.size());
    m_content.clear ();
    m_inLinks.clear ();
    m_outLinks.clear ();
    m_modifiedContent.clear ();
}

/**
 * Comparator for searching in content entries according to offset.
 */
struct OffsetKeyComparator {
    bool operator() (size_t offset, const ContentEntry& e) const
    {
        return offset < e.getOffset ();
    }

    bool operator() (const ContentEntry& e, size_t offset) const
    {
        return e.getOffset () < offset;
    }
} comp;


/**
 * Adds content to modified list if it is not already there.
 * 
 * Complexity: O(1).
 * 
 * @param modList list to which to add the content
 * @param ce content 
 */
template <typename ModList> 
static void
addToModifiedList (ModList* modList, ContentEntry* ce)
{
    if (!ce->ModifiedContentBaseHook::is_linked ())
    {
        modList->push_back (*ce);
    }
}


/**
 * Adds content to this area.
 * This method adds @c content into the list of modified content.
 * 
 * Complexity: O(log(valuesCount ())).
 * 
 * @param offset offset at which to store the content
 * @param content content to add
 * @return @c ContentEntry corresponding to the newly added content
 */
ContentEntry*
AreaEntry::addContent (size_t offset, std::auto_ptr<Content> content)
{
    LOG_DEBUG ("adding content:   object %p, offset %zu, size %zu, target %p",
               this, offset, content->getUnitSize (), 
               dynamic_cast<PtrContent*> (content.get ()) ? 
                       reinterpret_cast<void*> (dynamic_cast<PtrContent*> (
                            content.get ())->getPtr ().getAreaId ()):
                       0);
    
    std::auto_ptr<ContentEntry> ce;
    
    {
        PtrContent* ptrContent = dynamic_cast<PtrContent*> (content.get ());
        if (ptrContent)
        {
            content.release ();
            std::auto_ptr<PtrContent> ptrContentAP (ptrContent);
            
            RefContentEntry* rce = new RefContentEntry (this, offset, ptrContentAP);
            ce.reset (rce);
        }
        else
            ce.reset (new ContentEntry (offset, content));
    }
    
    m_content.insert (*ce);
    ce->setWasAdded (true);
    
    addToModifiedList (&m_modifiedContent, ce.get ());
    
    return ce.release ();
}

    
/**
 * Helper class for clearing @c m_content.
 */
template <typename ModList>
class ContentDisposer_ {
    ModList* m_modList;
    
  public:
    ContentDisposer_ (ModList* modList) : m_modList (modList) {}
    
    /**
     * Marks value as removed and adds it into modified list.
     * 
     * Complexity: O(1)
     */
    void operator() (ContentEntry* ce)
    {
        ce->setWasRemoved (true);
        /* NOTE: New content is still held in the UndoSet!!! This should be
         * handled otherwise (e.g., by removing also the corresponding Undo?)
        if (ce->wasAdded ())
        {
            assert (ce->ModifiedContentBaseHook::is_linked ());
            m_modList->erase (ModList::s_iterator_to (*ce));
            delete  ce;
        }
        else*/
            addToModifiedList (m_modList, ce);
    }
};

/**
 * Removes content from this area.
 * This method adds @c ce into the list of modified content.
 * 
 * Complexity: Average O(1).
 * 
 * @param ce content to remove
 */
void 
AreaEntry::removeContent (ContentEntry* ce)
{
    LOG_DEBUG ("removing content: object %p, offset %zu", this, ce->getOffset ());
    
    ContentDisposer_<AreaEntry::ModifiedContent> cd (&m_modifiedContent);    
    m_content.erase_and_dispose (ContentEntries::s_iterator_to (*ce), cd);
}


/**
 * Un-removes content into this area.
 * This method <em>does not</em> add @c ce into the list of modified content.
 * 
 * Complexity: O(log(valuesCount ())).
 * 
 * @param ce content to un-remove
 */
void
AreaEntry::unremoveContent (ContentEntry* ce)
{
    m_content.insert (*ce);
}


/**
 * Un-adds content from this area.
 * This method <em>does not</em> add @c ce into the list of modified content.
 * 
 * Complexity: O(log(valuesCount ())).
 * 
 * @param ce content to un-add
 */
void 
AreaEntry::unaddContent (ContentEntry* ce)
{
    m_content.erase (ContentEntries::s_iterator_to (*ce));
}


/**
 * Removes all content from this area.
 * This method adds all removed content into the list of modified content.
 * 
 * @param undoSet @c UndoSet to be filled with undo info.
 * Complexity: O(valuesCount ()).
 */
void 
AreaEntry::clearContent (UndoSet* undoSet)
{
    LOG_DEBUG ("clearing content: object %p", this);
    
    BOOST_FOREACH (ContentEntry& ce, m_content)
        undoSet->addUndo(new RemoveContentUndo(this, &ce));

    m_content.clear_and_dispose (ContentDisposer_<AreaEntry::ModifiedContent> (&m_modifiedContent));
    
    OutLinks::iterator b = m_outLinks.begin (), e = m_outLinks.end ();
    while (b != e)
    {
        RefContentEntry* rce = &*b;
        b++;
        rce->removeLink ();
    }
}


/**
 * Informs this object that canonical address have changed. Adds all contents of this area and
 * all contents referencing this area into list of modified content of the areas they belong to.
 * 
 * Complexity: O(valuesCount () + inLinksCount ()) if canonical address has changed, 
 *      O(1) otherwise
 */
void 
AreaEntry::reportModifiedCanonAddress ()
{
    BOOST_FOREACH (ContentEntry& ce, m_content)
        addToModifiedList (&m_modifiedContent, &ce);

    BOOST_FOREACH (RefContentEntry& rce, m_inLinks)
        addToModifiedList (&(rce.getSource ()->m_modifiedContent), &rce);
}

/**
 * Changes current area canonical address. If the new value differs from
 * the old one, the method adds the corresponding undo into the
 * given @c UndoSet.
 *
 * @param cannonAddr new canonical address assigned to this area
 * @param undoSet the @c UndoSet, where the corresponding undo should be
 *          added.
 * @return @c true if the newly assigned address differs from the last one.
 */
bool
AreaEntry::changeCanonAddr (CanonAddr canonAddr, UndoSet* undoSet)
{
    if (canonAddr != m_canonAddr) {
        LOG_DEBUG ("canon address of %p updated from %u to %u", this,
            m_canonAddr, canonAddr);

        undoSet->addUndo(new CanonAddrChangedUndo(this, m_canonAddr));
        m_canonAddr = canonAddr;
        return true;
    }
    return false;
}

/**
 * Finds content entry stored on specified offset in this area.
 * 
 * Complexity: O(log(valuesCount ()))
 * 
 * @param offset offset of the content in this area
 * @return content stored at the specified offset
 * @throws UndefinedMemoryLoadException if there is no content stored at @c offset in this area
 */
ContentEntry* 
AreaEntry::findContentEntry (size_t offset)
{
    ContentEntries::iterator it = m_content.find (offset, comp);
    
    if (it == m_content.end ())
        throw UndefinedMemoryLoadException (Ptr (this, offset));
    
    return &(*it);
}
    
    
typedef AreaEntry::ContentEntryIterator I;

/**
 * Finds content in this area that overlaps with the specified range.
 * 
 * Complexity: O(log(valuesCount ()))
 *
 * @param offsetBegin beginning offset of the range (included)
 * @param offsetEnd end offset of the range (excluded)
 * @return iterator range with content entries that overlap 
 *      with <tt>&lt;offsetBegin; offsetEnd)</tt> range.
 */
std::pair<I, I>
AreaEntry::findContentEntries (size_t offsetBegin, size_t offsetEnd)
{
    I begin = m_content.lower_bound (offsetBegin, comp);
    if (begin != m_content.begin ())
    {
        begin--;
        if (begin->getOffset () + begin->getContent ()->getUnitSize () <= offsetBegin)
            begin++;
    }
    
    I end = m_content.lower_bound (offsetEnd, comp);
    
    return std::make_pair (begin, end);
}


/**
 * Adds link to the list of outgoing links.
 * Complexity: O(log(valuesCount ()))
 * 
 * @param link link
 */
void 
AreaEntry::addOutLink (RefContentEntry* link)
{
    bool ret = m_outLinks.insert (*link).second;
    assert (ret);   (void)ret;
}


/**
 * Adds link to the list of incoming links.
 * Complexity: O(1)
 * 
 * @param link link
 */
void 
AreaEntry::addInLink (RefContentEntry* link)
{
    m_inLinks.push_back (*link);
}
    

/**
 * Removes link from the list of outgoing links.
 * Complexity: Average O(1)
 * 
 * @param link link
 */
void
AreaEntry::removeOutLink (RefContentEntry* link)
{
    m_outLinks.erase (OutLinks::s_iterator_to (*link));
}
    

/**
 * Removes link from the list of incoming links.
 * Complexity: O(1)
 * 
 * @param link
 */
void
AreaEntry::removeInLink (RefContentEntry* link)
{
    m_inLinks.erase (InLinks::s_iterator_to (*link));
}


/**
 * Clears the list of modified content.
 * 
 * Complexity: O(s), where s is the number of modified values in this area
 */
void 
AreaEntry::clearModificationsList ()
{
    BOOST_FOREACH (ContentEntry& ce, m_modifiedContent)
    {
        ce.clearTmpFlags ();
    }
    m_modifiedContent.clear ();
}


#ifndef NDEBUG

/**
 * Checks if the relations (entries in @c m_outLinks) correspond to the content of this area.
 */
void
AreaEntry::checkRelations () const
{
    if (isDeleted ())
        assert (m_outLinks.empty ());
    else
    {
        size_t linkCount = 0;
        
        BOOST_FOREACH (const ContentEntry& ce, m_content)
        {
            const RefContentEntry* rce = dynamic_cast<const RefContentEntry*> (&ce);
            
            if (rce && rce->isValidLink () && rce->getSource () != rce->getTarget ())
            {
                assert (m_outLinks.find (*rce) != m_outLinks.end ());
                linkCount++;
            }
        }
        LOG_DEBUG("out links: %zu, link count: %zu", m_outLinks.size(), linkCount);
        assert (m_outLinks.size () == linkCount);
    }
}


/**
 * Recounts hash of this object and adds it to @c hashHolder.
 * 
 * @param hashHolder hash holder
 */
void
AreaEntry::recountHash (HashHolder* hashHolder) const
{
    if (isDeleted () || !isReachable ())
        assert (m_content.empty ());
    else
        hashHolder->addAreaHash (getCanonAddr (), getSize ());
    
    
    BOOST_FOREACH (const ContentEntry& ce, m_content)
        hashHolder->addContentHash (getCanonAddr () + ce.getOffset (),
                                    ce.getObjectHash ());
}
        
#endif


}}
