// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

#ifndef MM_OBJECTENTRY_H
#define MM_OBJECTENTRY_H

#include "RefContentEntry.h"
#include "MemoryAccessListener.h"
#include "Types.h"
#include "ContentEntry.h"
#include "undo/RemoveContentUndo.h"
#include "undo/AddContentUndo.h"
#include "undo/CanonAddrChangedUndo.h"
#include "undo/UndoSet.h"

#include <boost/intrusive/list.hpp>
#include <boost/intrusive/slist.hpp>
#include <boost/noncopyable.hpp>
#include <boost/foreach.hpp>


namespace mmodule { namespace detail {
    
class State;
class HashHolder;


class StateAreasTag {};
class ModifiedAreasTag {};

typedef boost::intrusive::list_base_hook<boost::intrusive::tag <StateAreasTag> >
        StateAreasBaseHook;
typedef boost::intrusive::list_base_hook<boost::intrusive::tag <ModifiedAreasTag> >
         ModifiedAreasBaseHook;
         
         
/**
 * Class that represents memory area.
 */
class AreaEntry : private boost::noncopyable, 
                    public StateAreasBaseHook,
                    public ModifiedAreasBaseHook {
  private:
    struct OffsetComparator {
        bool operator() (const ContentEntry& e1, const ContentEntry& e2) const
        {
            return e1.getOffset () < e2.getOffset ();
        }
    };

      
    
    template <typename T> struct BaseHookOp : boost::intrusive::base_hook<T> {};
    typedef boost::intrusive::constant_time_size<false> SizeOp;
    typedef boost::intrusive::cache_last<true> CacheOp;
    
    typedef boost::intrusive::set<ContentEntry, 
                                  boost::intrusive::compare<OffsetComparator>,
                                  BaseHookOp<ContentListContentBaseHook> > ContentEntries;
    typedef boost::intrusive::set<RefContentEntry,
                                  boost::intrusive::compare<OffsetComparator>, 
                                  BaseHookOp<OutLinksBaseHook> > OutLinks;
    typedef boost::intrusive::list<RefContentEntry,
                                  BaseHookOp<InLinksBaseHook> > InLinks;
    typedef boost::intrusive::list<ContentEntry, 
                BaseHookOp<ModifiedContentBaseHook>, SizeOp, CacheOp> ModifiedContent;

    
    /**
     * values saved in this memory area
     */
    ContentEntries m_content;
    
    /**
     * outgoing links; content entries from @c m_content referencing <em>other</em> areas
     */
    OutLinks m_outLinks;
    
    /**
     * incoming links; content entries referencing this area
     */
    InLinks m_inLinks;
    
    /**
     * list of modified content
     */
    ModifiedContent m_modifiedContent;
    
    /**
     * current canonical address of the area
     */
    CanonAddr m_canonAddr;

    /**
     * old canonical address of the area
     */
    CanonAddr m_oldCanonAddr;
    
    /**
     * link to parent area in the BFS tree
     */
    RefContentEntry* m_bfsTreeLinkToParent;
    
    /**
     * size of this area in bytes
     */
    size_t m_size;
    
    /*! @c true if this area is deleted, @c false otherwise */
    bool m_deleted : 1;
    /*! color of this area; used by @c AreaColorMap */
    int m_color : 3;
    
    /*! @c true if this area was just created, @c false otherwise */
    bool m_wasCreated : 1;
    /*! @c true if this area was just deleted, @c false otherwise */
    bool m_wasDeleted : 1;
    /*! @c true if this area is expandable (size grows as necessary and does
     * not influence state equality) */
    bool m_expandable : 1;
    
    /*! @c true if this area was already reported as unreachable */
    bool m_unreachableReported : 1;
        
  public:
    typedef ContentEntries::iterator ContentEntryIterator;
    typedef OutLinks::iterator OutLinksIterator;
    typedef InLinks::iterator InLinksIterator;
    typedef OutLinks::const_iterator OutLinksConstIterator;
    typedef InLinks::const_iterator InLinksConstIterator;
    
    
    AreaEntry ();
    AreaEntry (size_t size);
    ~AreaEntry ();
    
    std::pair<OutLinksIterator, OutLinksIterator> linkSources ();
    std::pair<InLinksIterator, InLinksIterator> linkTargets ();
    
    ContentEntry* findContentEntry (size_t offset);
    
    std::pair<ContentEntryIterator, ContentEntryIterator> 
    findContentEntries (size_t offsetBegin, size_t offsetEnd);
            
    /**
     * Returns range of content entries representing values stored in this area.
     * 
     * @return range of content entries representing values stored in this area
     */
    std::pair<ContentEntryIterator, ContentEntryIterator> 
    contentEntries ()
    { return std::make_pair (m_content.begin (), m_content.end ()); }
    
            
    /**
     * Returns size of this area.
     * 
     * @return size of this area in bytes
     */
    size_t getSize () const
    { return m_size; }
    
    /**
     * Returns pointer to the begining of the memory area represented by this area.
     * 
     * @return pointer to the begining of the memory area represented by this area
     */
    Ptr getBeginPtr () 
    { return Ptr (this, 0); }
    
    /**
     * Checks if this area is deleted.
     * 
     * @return @c true if this area is deleted, @c false otherwise
     * @see setDeleted()
     */
    bool isDeleted () const
    { return m_deleted; }

    /**
     * Checks if this area is expandable.
     *
     * @return @c true if this area is expandable, @c false otherwise
     */
    bool isExpandable () const
    { return m_expandable; }
    
    /**
     * @see isDeleted()
     */
    void setDeleted (bool deleted) 
    { m_deleted = deleted; }
    
    
    /**
     * Checks if this area was just deleted.
     * This flag is set only for a short time between area deletion and
     * next call to @c State::saveState_().
     * 
     * @return @c true if this area was just deleted, @c false otherwise; 
     * @see setWasDeleted()
     */
    bool wasDeleted () const
    { return m_wasDeleted; }
    
    /**
     * @see wasDeleted()
     */
    void setWasDeleted (bool wasDeleted) 
    { m_wasDeleted = wasDeleted; }
    
    /**
     * Checks if this area was just created.
     * This flag is set only for a short time between area creation and
     * next call to @c State::saveState_().
     * 
     * @return @c true if this area was just created, @c false otherwise; 
     * @see setWasCreated()
     */
    bool wasCreated () const
    { return m_wasCreated; }
    
    /**
     * @see wasCreated()
     */
    void setWasCreated (bool wasCreated)
    { m_wasCreated = wasCreated; }
    
    
    /**
     * Checks if this area is reachable.
     * 
     * @return @c true if this area is reachable from @c State::getRootArea() area, 
     *      @c false otherwise
     */
    bool isReachable () const
    { return m_canonAddr != CANON_UNREACHABLE; }
    
    /**
     * Returns unique ID of this memory area. Note that the ID is currently
     * derived from the address of this AreaEntry object in memory.
     * However, this behavior may change in the future.
     *
     * @return unique ID of this area
     */
    size_t getAreaId () const
    { return reinterpret_cast<size_t> (this); }
    
    /**
     * Returns link to parent area in the BFS tree.
     * 
     * @return link to parent area in the BFS tree
     */
    RefContentEntry* getBfsTreeLinkToParent () const
    { return m_bfsTreeLinkToParent; }
    
    /**
     * Sets link to parent area in the BFS tree.
     * 
     * @param bfsTreeLink link to parent area in the BFS tree
     */
    void setBfsTreeLinkToParent (RefContentEntry* bfsTreeLink) 
    { m_bfsTreeLinkToParent = bfsTreeLink; }
    
    
    
    /*! Gets color. */
    int getColor () const
    { return m_color; }
    
    /*! Sets color. */
    void setColor (int color)
    { m_color = color; }
    
    
    /**
     * Gets current area properties.
     *
     * @return current area properties
     */
    //const AreaEntryProperties& getProperties () const
    //{ return m_properties; }
    
    /**
     * Gets current area canonical address.
     *
     * @return current area canonical address
     */
    CanonAddr getCanonAddr () const
    { return m_canonAddr; }

    bool changeCanonAddr (CanonAddr canonAddr, UndoSet* undoSet);

    /**
     * Gets old area canonical address.
     *
     * @return old area canonical address
     */
    CanonAddr getOldCanonAddr () const
    { return m_oldCanonAddr; }


    /**
     * Returns range of content entries referencing this area.
     * 
     * @return range of content entries referencing this area
     */
    std::pair<InLinksConstIterator, InLinksConstIterator> inLinks () const
    { return std::make_pair (m_inLinks.begin (), m_inLinks.end ()); }
    
    /**
     * Returns range of content entries referencing this area.
     * 
     * @return range of content entries referencing this area
     */
    std::pair<InLinksIterator, InLinksIterator> inLinks ()
    { return std::make_pair (m_inLinks.begin (), m_inLinks.end ()); }
    
    /**
     * Returns range of content entries from @c m_content referencing <em>other</em> areas.
     * 
     * @return range of content entries from @c m_content referencing <em>other</em> areas
     */
    std::pair<OutLinksConstIterator, OutLinksConstIterator> outLinks () const
    { return std::make_pair (m_outLinks.begin (), m_outLinks.end ()); }
    
    /**
     * Returns range of content entries from @c m_content referencing <em>other</em> areas.
     * 
     * @return range of content entries from @c m_content referencing <em>other</em> areas
     */
    std::pair<OutLinksIterator, OutLinksIterator> outLinks ()
    { return std::make_pair (m_outLinks.begin (), m_outLinks.end ()); }
    
    
    /**
     * Returns number of content entries from @c m_content referencing <em>other</em> areas.
     * 
     * @return number of content entries from @c m_content referencing <em>other</em> areas
     */
    size_t outLinksCount () const
    { return m_outLinks.size (); }
    
    /**
     * Returns number of content entries referencing this area.
     * 
     * @return number of content entries referencing this area
     */
    size_t inLinksCount () const
    { return m_inLinks.size (); }
    
    /**
     * Returns the number of values stored int his area.
     * 
     * @return the number of values stored int his area
     */
    size_t valuesCount () const
    { return m_content.size (); }
    
    
    void clearModificationsList ();
    
    /**
     * Iterates through modified (added/removed) values of this area.
     * 
     * @param modVis visitor that is informed about the modified values
     */
    template <typename Visitor>
    void walkModifiedContent (Visitor& modVis)
    {
        BOOST_FOREACH (ContentEntry& ce, m_modifiedContent)
        {
            modVis.modifiedContent (this, &ce);
        }
    }
    
    
    /**
     * Clears temporary flags and values.
     */
    void clearTmpFlags ()
    {
        m_wasCreated = false;
        m_wasDeleted = false;

        m_oldCanonAddr = m_canonAddr;
    }

    /**
     * Reports this area as unreachable if it is unreachable and if it 
     * was not already reported before.
     * 
     * @param listener visitor 
     */
    void reportUnreachable (MemoryAccessListener* listener)
    {
        if (!m_unreachableReported)
        {
            m_unreachableReported = true;
            listener->unreachableArea (getAreaId());
        }
    }
    
    /**
     * Checks if this object was already reported as unreachable.
     * 
     * @return @c true if this object was already reported as unreachable, @c false otherwise
     * @see setReportedUnreachable()
     */
    bool wasReportedUnreachable () const
    { return m_unreachableReported; }
    
    /**
     * @see wasReportedUnreachable()
     */
    void setReportedUnreachable (bool unreachableReported)
    { m_unreachableReported = unreachableReported; }
    
#ifndef NDEBUG
    void checkRelations () const;
    void recountHash (HashHolder* hashHolder) const;
#endif
    
    
    bool operator< (const AreaEntry& ae) const
    { 
        return this < &ae; 
    }
    
  protected:
    void addOutLink (RefContentEntry* link);
    void addInLink (RefContentEntry* link);
    void removeOutLink (RefContentEntry* link);
    void removeInLink (RefContentEntry* link);

    ContentEntry* addContent (size_t offset, std::auto_ptr<Content> content);
    void removeContent (ContentEntry* content);
    void clearContent (UndoSet* undoSet);
    void reportModifiedCanonAddress ();
    
    void unremoveContent (ContentEntry* ce);
    void unaddContent (ContentEntry* content);
    void unchangeCanonAddr (CanonAddr canonAddr)
    {
        m_canonAddr = canonAddr;
        m_oldCanonAddr = canonAddr;
    }
    
    friend class RefContentEntry;
    friend class State;
    friend class RemoveContentUndo;
    friend class AddContentUndo;
    friend class CanonAddrChangedUndo;
    template <typename Vis, typename Areas> friend class UnreachableAreaCleaner;
};

}}

#endif
