// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

#ifndef MM_CACHINGFUNCTOR_H
#define MM_CACHINGFUNCTOR_H


#include <ext/hash_map>
#include <boost/functional/hash.hpp>

#include "../mm/Types.h"

namespace mmodule { namespace detail {

/**
 * Functor that caches results of underlying functor.
 * It is guaranteed, that the underlying functor will not be called more than once 
 * with a single argument.
 * 
 * @tparam Functor underlying functor type
 */
template <typename Functor>
class CachingFunctor {
  private:
    typedef typename Functor::Arg Arg;
    typedef typename Functor::Result Result;
        
    typedef __gnu_cxx::hash_map<Arg, CanonAddr, boost::hash<Arg> > Data_;
    
    /**
     * cached results
     */
    Data_ m_data;
        
    /**
     * reference to the underlying functor
     */
    Functor& m_functor;
    
  public:
    /**
     * Constructor.
     * 
     * @param functor functor to use as underlying functor
     */
    CachingFunctor (Functor& functor) : m_functor (functor) {}


    /**
     * Returns result of first call to underlying functor with argument @c arg.
     * 
     * Complexity: Average O(1).
     * 
     * @param arg arguments
     * @return result of first call to underlying functor with argument @c arg
     */
    Result operator() (const Arg& arg)
    {
        typename Data_::iterator it = m_data.find (arg);
        if (it != m_data.end ())
            return it->second;
        else
        {
            Result res = m_functor (arg);
            m_data.insert (std::make_pair (arg, res));
            return res;
        }
    }
};

}}

#endif 
