// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

#ifndef MM_CANON_H
#define MM_CANON_H


#include "CachingFunctor.h"
#include "Types.h"

#include <boost/functional/hash.hpp>

namespace mmodule { namespace detail {
    
    
/**
 * Functor implementing Canon function.
 */
class Canon {
  private:
    /**
     * Type for argument of Canon function.
     */
    struct Arg_ {
        /**
         * Constructor.
         * 
         * @param referenceAddr canonical address of the tree pointer
         * @param areaSize size of the area the pointer is pointing to (in bytes)
         */
        Arg_ (CanonAddr referenceAddr, size_t areaSize) 
            : refAddr (referenceAddr), size (areaSize) {}
        
        /**
         * canonical address of the tree pointer
         */
        CanonAddr refAddr;
        
        /**
         * size of the area the pointer is pointing to (in bytes)
         */
        size_t size;
        
        /**
         * Operator @c <.
         */
        bool operator< (const Arg_& a) const
        {
            if (refAddr < a.refAddr)
                return true;
            else if (refAddr > a.refAddr)
                return false;
            
            return size < a.size;
        }
        
        /**
         * Operator @c ==.
         */
        bool operator== (const Arg_& a) const
        { return refAddr == a.refAddr && size == a.size; }
    };
    
    
    /**
     * Functor generating canonical addresses.
     * Must not be called more than once for single argument.
     */
    class CanonGenerator {
        CanonAddr m_nextFreeAddr;
        
      public:
        typedef Arg_ Arg;
        typedef CanonAddr Result;
        
        CanonGenerator () : m_nextFreeAddr (4) {}
        
        Result operator() (const Arg& arg)
        {
            CanonAddr res = m_nextFreeAddr;
            m_nextFreeAddr += arg.size;
            assert (m_nextFreeAddr >= arg.size);
            
            return res;
        }
    };
    
    friend size_t hash_value (const Canon::Arg_& arg);
    
  private:
    CanonGenerator m_canonGenerator;
    CachingFunctor<CanonGenerator> m_cachingFunctor;

    
  public:
    /**
     * Constructor.
     */
    Canon () : m_canonGenerator (), m_cachingFunctor (m_canonGenerator) { }
    
    
    /**
     * Returns result of Canon function.
     * 
     * Complexity: Average O(1)
     * 
     * @param refAddr canonical address of the tree pointer
     * @param size size of the area the pointer is pointing to. It can be
     *          safely ignored, since it is taken into account when computing 
     *          the area hash.
     * @return function value; returned value is always greater than @c 0
     */
    CanonAddr get (const CanonAddr& refAddr, size_t size)
    {
        Arg_ arg (refAddr, size);
        
        return m_cachingFunctor (arg);
    }
};


inline size_t
hash_value (const Canon::Arg_& arg)
{
    std::size_t seed = 0;

    boost::hash_combine (seed, arg.refAddr);
    boost::hash_combine (seed, arg.size);
    
    return seed;
}

}}

#endif
