// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

#ifndef MM_CONTENT_H
#define MM_CONTENT_H

#include <cstddef>

#include "Ptr.h"
#include "Types.h"

namespace mmodule {
    
class PtrContent;
class NonPtrContent;

/**
 * Interface for classes representing content of the memory.
 * @warning Do not inherit directly from this class. Use @c PtrContent class
 * or @c NonPtrContent class instead.
 */
class Content {
  public:
    /**
     * Visitor type.
     */
    struct Visitor {
        virtual ~Visitor () {}
        
        virtual void visit (const PtrContent& decl)     = 0;
        virtual void visit (const NonPtrContent& decl)  = 0;
    };
        
  public:
    virtual ~Content () {}
    
    /**
     * Returns number of bytes occupied by the value represented by this object.
     * 
     * @return number of bytes occupied by the value represented by this object
     */
    virtual size_t getUnitSize () const = 0;
    
    /**
     * @warning Do not implement this method. If the compiler complains about this class
     * beeing pure, than inherit from @c PtrContent or @c NonPtrContent.
     */
    virtual void accept (Visitor& visitor) = 0;
};


/**
 * Interface for classes that represent values that are pointers.
 * Only objects with this interface will be interpreted by the canonicalization algorithm
 * as pointers.
 */
struct PtrContent : virtual public Content {
    virtual ~PtrContent () {}
    
    /**
     * Returns contained pointer.
     *
     * @return contained pointer
     */
    virtual Ptr getPtr () const = 0;
    
    /**
     * Returns the type hash.
     * The type hash should represent the type of the value represented by this object.
     *
     * @return type hash
     */
    virtual Hash getTypeHash () const = 0;
    
  private:
    virtual void accept (Visitor& visitor)
    { visitor.visit (*this); }
};


/**
 * Interface for classes that represent values that are not pointers.
 */
struct NonPtrContent : virtual public Content {
    virtual ~NonPtrContent () {}
    
    /**
     * Returns hash of this content.
     * 
     * @return hash of this content
     */
    virtual Hash getHash () const = 0;
    
  private:
    virtual void accept (Visitor& visitor)
    { visitor.visit (*this); }
};

}

#endif
