// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

#include "ContentEntry.h"

#include "AreaEntry.h"

#include <boost/functional/hash.hpp>

namespace mmodule { namespace detail {

/**
 * Returns hash of the @c Ptr object.
 * 
 * @param canonAddr canonical address of the area the pointer is pointing to
 * @param offset offset within the area the pointer is pointing to
 */
Hash
getPtrHash_ (CanonAddr canonAddr, size_t offset)
{
    std::size_t seed = 0;
    
    boost::hash_combine (seed, canonAddr);
    boost::hash_combine (seed, offset);
    
    return seed;
}


/**
 * Helper class for getting current object hash of the content.
 */
class HashGetter : public Content::Visitor {
    Hash m_hash;
    
    public:
        void visit (const PtrContent& c)
        {
            MultHash ret;
            AreaEntry* target = c.getPtr ().getTargetArea ();
            CanonAddr targetAddr = target ? 
                    target->getCanonAddr () : 0;
            ret = getPtrHash_ (targetAddr, c.getPtr ().getOffset ());
            ret += c.getTypeHash ();
            m_hash = ret % hashModul ();
        }
    
        void visit (const NonPtrContent& c)
        {
            m_hash = c.getHash ();
        }
    
        const Hash& getHash () const
        { return m_hash; }
};

/**
 * Returns current object hash of the value represented by this object.
 * 
 * @return current object hash of the value represented by this object
 */
Hash 
ContentEntry::getObjectHash () const
{ 
    HashGetter g;
    m_content->accept (g);
    return g.getHash ();
}
    

/**
 * Helper class for getting old object hash of the content.
 */
class OldHashGetter : public Content::Visitor {
    Hash m_hash;
    
    public:
        void visit (const PtrContent& c)
        {
            MultHash ret;
            AreaEntry* target = c.getPtr ().getTargetArea ();
            CanonAddr targetAddr = target ? 
                    target->getOldCanonAddr () : 0;
            ret = getPtrHash_ (targetAddr, c.getPtr ().getOffset ());
            ret += c.getTypeHash ();
            m_hash = ret % hashModul ();
        }
    
        void visit (const NonPtrContent& c)
        {
            m_hash = c.getHash ();
        }
    
        const Hash& getHash () const
        { return m_hash; }
};


/**
 * Returns old object hash of the content.
 * The old object hash differs from the current object hash, 
 * if the content is @c PtrContent and the canonical
 * address of the target area has just changed.
 * 
 * @return old hash of the content
 */
Hash 
ContentEntry::getOldObjectHash () const
{ 
    OldHashGetter g;
    m_content->accept (g);
    return g.getHash ();
}

}}
