// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

#ifndef MM_CONTENTENTRY_H
#define MM_CONTENTENTRY_H

#include "Content.h"

#include <memory>
#include <boost/intrusive/list.hpp>
#include <boost/intrusive/slist.hpp>
#include <boost/intrusive/set.hpp>
#include <boost/noncopyable.hpp>

namespace mmodule { namespace detail {
    
class ModifiedContentTag {};
class ContentListContentTag {};


typedef boost::intrusive::list_base_hook<boost::intrusive::tag<ModifiedContentTag> >
        ModifiedContentBaseHook;
        
typedef boost::intrusive::set_base_hook<boost::intrusive::tag<ContentListContentTag> >
        ContentListContentBaseHook;


/**
 * Class holding value stored in the memory.
 * Every content entry belongs to exactly one memory area.
 */
class ContentEntry 
    : private boost::noncopyable,
      public ModifiedContentBaseHook,
      public ContentListContentBaseHook
{
    /**
     * offset of the value within owning memory area
     */
    size_t m_offset;
    
    /**
     * the value
     */
    std::auto_ptr<Content> m_content;
    
    /*! @c true if this content was just removed, @c false otherwise */
    bool m_wasRemoved : 1;
    /*! @c true if this content was just added, @c false otherwise */
    bool m_wasAdded : 1;
    
  public:
    /**
     * Constructor.
     * 
     * @param offset offset of the content within the owning area
     * @param content value 
     */
    ContentEntry (size_t offset, std::auto_ptr<Content> content)
        : m_offset (offset), m_content (content), 
          m_wasRemoved (false), m_wasAdded (false)
    {
    }
    
    /**
     * Destructor.
     */
    virtual ~ContentEntry () {}
        
    
    /**
     * Returns the offset of this content within the owning area.
     * 
     * @return the offset of this content within the owning area
     */
    size_t getOffset () const
    { return m_offset; }
    
    /**
     * Returns the value stored in this content entry.
     * 
     * @return the value stored in this content entry
     */
    const Content* getContent () const
    { return m_content.get (); }
    
    Hash getObjectHash () const;    
    Hash getOldObjectHash () const;
    
    /**
     * Checks if this content was just removed.
     * 
     * @return @c true if this content was just removed, @c false otherwise
     * @see setWasRemoved()
     */
    bool wasRemoved () const
    { return m_wasRemoved; }
    
    /**
     * @see wasRemoved()
     */
    void setWasRemoved (bool wasRemoved) 
    { m_wasRemoved = wasRemoved; }
    
    /**
     * Checks if this content was just added.
     * 
     * @return @c true if this content was just added, @c false otherwise
     * @see setWasAdded()
     */
    bool wasAdded () const
    { return m_wasAdded; }
    
    /**
     * @see wasAdded()
     */
    void setWasAdded (bool wasAdded) 
    { m_wasAdded = wasAdded; }
    
    
    /**
     * Clears temporary flags and values.
     */
    void clearTmpFlags ()
    {
        m_wasRemoved = false;
        m_wasAdded = false;
    }
};

}}

#endif
