// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

#include "HashHolder.h"

#include "RandomValues.h"


namespace mmodule { namespace detail {

/**
 * Constructor.
 * 
 * @param areaRand functor returning random values for canonincal addresses 
 *      that are used in area partial hash computation
 * @param valueRand functor returning random values for canonincal addresses
 *      that area used in value partial hash computation
 */
HashHolder::HashHolder (RandomValues* areaRand, RandomValues* valueRand) 
    : m_hash (0), m_areaRand (areaRand), m_valueRand (valueRand)
{        
}


/**
 * Destructor.
 */
HashHolder::~HashHolder () 
{
}


/**
 * Computes content partial hash.
 * 
 * @param addr address of the content
 * @param hash hash of the contents value
 */
Hash
HashHolder::computeContentPartialHash (const CanonAddr& addr, const Hash& hash)
{
    assert (addr != 0);
    
    MultHash randVal = (*m_valueRand) (addr);
    MultHash contVal = hash;
    MultHash res = (randVal*contVal) % hashModul ();
    
    return res;
}

   
/**
 * Adds partial hash to the state hash.
 * 
 * @param partialHash partial hash to add to the state hash
 */
void
HashHolder::addPartialHash (Hash partialHash)
{
    MultHash res = partialHash;
    res += m_hash;
    res %= hashModul ();
    m_hash = res;
}


/**
 * Removes partial hash from the state hash.
 * 
 * @param partialHash partial hash to remove from the state hash
 */
void 
HashHolder::removePartialHash (Hash partialHash)
{
    addPartialHash (hashModul () - partialHash);
}



/**
 * Updates state hash as if new content was stored into the memory.
 * 
 * Complexity: Average O(1)
 * 
 * @param addr canonical address of the content
 * @param contentHash hash of the contents value
 */
void
HashHolder::addContentHash (const CanonAddr& addr, const Hash& contentHash)
{
    addPartialHash (computeContentPartialHash (addr, contentHash));
}
    

/**
 * Updates state hash as if the content was removed from the memory.
 * 
 * @param oldAddr canonical address of the content before it was removed
 * @param oldContentHash hash of the content's value before it was removed
 */
void
HashHolder::removeContentHash (const CanonAddr& oldAddr, const Hash& oldContentHash)
{
    removePartialHash (computeContentPartialHash (oldAddr, oldContentHash));
}


/**
 * Computes area partial hash.
 * 
 * @param beginAddr address of the begin of the area
 * @param size the size of the area in bytes
 */
Hash
HashHolder::computeAreaPartialHash (const CanonAddr& beginAddr, size_t size)
{
    assert (beginAddr != 0);
    
    return (*m_areaRand) (beginAddr) * size;
}


/**
 * Updates state hash as if the area was freed.
 * 
 * @param newAddr address of the begin of the removed area
 * @param size the size of the area in bytes
 */
void 
HashHolder::addAreaHash (const CanonAddr& newAddr, size_t size)
{
    addPartialHash (computeAreaPartialHash (newAddr, size));
}
    

/**
 * Updates state hash as if the area was allocated.
 * 
 * @param oldAddr address of the begin of the added area
 * @param size the size of the area in bytes
 */
void
HashHolder::removeAreaHash (const CanonAddr& oldAddr, size_t size)
{
    removePartialHash (computeAreaPartialHash (oldAddr, size));
}
    
    
}}
