// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

#ifndef MM_EXCEPTIONS_H
#define MM_EXCEPTIONS_H

#include "Ptr.h"

#include <exception>


namespace mmodule {

/**
 * Common base class of all exceptions that can be thrown when 
 * working with the state of the memory.
 */
struct MemoryException : public std::exception {
};


/**
 * Exception that is thrown when trying to access already deleted memory area.
 */
struct FreedAreaAccessException : public MemoryException {
    /**
     * Constructor.
     * 
     * @param p pointer that caused the exception
     */
    FreedAreaAccessException (Ptr p) : ptr (p) {}
    
    const char* what () const throw ()
    {
        return "access to freed memory area";
    }
    
    /*! pointer that caused the exception */
    Ptr ptr;
};


/**
 * Exception that is thrown when the pointer does not point to the beginning of memory area
 * but it should.
 */
struct NotAreaBeginException : public MemoryException {
    /**
     * Constructor.
     * 
     * @param p pointer that caused the invocation of the exception
     */
    NotAreaBeginException (Ptr p) : ptr (p) {}
    
    const char* what () const throw ()
    {
        return "pointer does not point to the beginning of area";
    }
    
    /*! pointer that caused the invocation of the exception */
    Ptr ptr;
};



/**
 * Exception that is thrown when trying to load an undefined value.
 * Value on address @c a is defined if @c MemoryModule::store() was called before 
 * to store some value at address @c a, and it was not overwritten (even partially).
 * In other case a value on address @c a is undefined.
 */
struct UndefinedMemoryLoadException : public MemoryException {
    /**
     * Constructor.
     * 
     * @param p pointer to the address from where it was attempted to load the value
     */
    UndefinedMemoryLoadException (Ptr p) : ptr (p) {}
    
    const char* what () const throw ()
    {
        return "loading content of undefined memory";
    }
    
    /*! the address from where it was attempted to load the value */
    Ptr ptr;
};


/**
 * Exception that is throws when trying to store/load a value that does not lie (wholly) 
 * within the memory area.
 */
struct OutOfBoundsException : public MemoryException {
    /**
     * Constructor.
     * 
     * @param p pointer to the address from where it was attempted to load or store the value
     */
    OutOfBoundsException (Ptr p) : ptr (p) {}
    
    const char* what () const throw ()
    {
        return "offset range out of bounds";
    }
    
    /** pointer to the address from where it was attempted to load the value */
    Ptr ptr;
};

}

#endif 
