// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

#include "MemoryState.h"

#include "State.h"
#include "Canon.h"
#include "AreaEntry.h"
#include "ContentEntry.h"
#include "MemoryExceptions.h"
#include "HashHolder.h"
#include "DefaultModificationsVisitor.h"
#include "Content.h"
#include "RandomValues.h"
#include "../Error.h"
#include "undo/AllocUndo.h"

#include <vector>
#include <boost/foreach.hpp>
#include <boost/utility.hpp>


namespace mmodule { 

using namespace detail;

void 
MemoryState::checkDeleted_ (AreaEntry* ae)
{
    if (ae->isDeleted ())
        throw FreedAreaAccessException (Ptr (ae, 0));
}


namespace detail {
    
static void 
checkAreaBegin_ (Ptr ptr)
{
    if (ptr.getOffset () != 0)
        throw NotAreaBeginException (ptr);
}


static void
checkNullPtr_ (AreaEntry* ae)
{
    if (!ae)
        throw NullPointerException ();
}

}
        

/**
 * Constructor.
 * 
 * @param listener visitor for reporting runtime problems
 */
MemoryState::MemoryState (MemoryAccessListener* listener)
    : m_currentUndoSet (new UndoSet ()), m_state (new State ()),
      m_areaRandom (new RandomValues ()), m_valueRandom (new RandomValues ()),
      m_hashHolder (new HashHolder (m_areaRandom.get (), m_valueRandom.get ())), 
      m_canon (new Canon ()), 
      m_listener (listener)
{
    LOG_INFO ("size overhead: area = %zu, nonPtrValue = %zu, ptrValue = %zu, "
               "areaUndoSet = %zu, areaContentUndo = %zu, undoSet = %zu, addContentUndo = %zu, "
               "allocUndo = %zu",
               sizeof (AreaEntry), sizeof (ContentEntry), sizeof (RefContentEntry),
               (size_t)64u, (size_t)40u, sizeof (UndoSet),
               sizeof (AddContentUndo), sizeof (AllocUndo));
}


/**
 * Destructor.
 */
MemoryState::~MemoryState ()
{
    m_state.reset ();
    m_markedStates.clear ();
    m_undoSets.reverse ();
    m_undoSets.clear_and_dispose (boost::checked_delete<UndoSet>);
    delete m_currentUndoSet;
}


/**
 * Returns the number of areas in the memory.
 * 
 * @return the number of areas in the memory including deleted areas, excluding unreachable areas
 */
size_t 
MemoryState::getAreaCount () const
{
    return m_state->getAreaCount ();
}


/**
 * Allocates memory area.
 * 
 * Complexity: O(1)
 * 
 * @param size size of the area in bytes
 * @return area entry representing the newly allocated area
 */
AreaEntry*
MemoryState::doAlloc (size_t size)
{
    AreaEntry* ae = m_state->createArea (size, m_currentUndoSet);

    return ae;
}


/**
 * Allocates memory area.
 * 
 * Complexity: O(1)
 * 
 * @param size size of the area in bytes
 * @return pointer pointing to the beginning of the newly allocated area
 */
Ptr 
MemoryState::alloc (size_t size)
{    
    return Ptr (doAlloc (size), 0);
}


/**
 * Allocates expanding memory area. The allocated area grows according to its 
 * contents. In other words, it precise size does not differentiate two states. 
 * This option is to be used to model system areas such as list of threads, 
 * thread call stack, etc.
 * 
 * Complexity: O(1)
 * 
 * @return area entry representing the newly allocated area
 */
AreaEntry*
MemoryState::doAllocExpanding ()
{
    AreaEntry* ae = m_state->createExpandingArea (m_currentUndoSet);

    return ae;
}
/**
 * Allocates expanding memory area. The allocated area grows according to its
 * contents. In other words, it precise size does not differentiate two states.
 * This option is to be used to model system areas such as list of threads,
 * thread call stack, etc.
 *
 * Complexity: O(1)
 *
 * @return pointer pointing to the beginning of the newly allocated area
 */
Ptr
MemoryState::allocExpanding ()
{
    return Ptr (doAllocExpanding (), 0);
}


/**
 * Frees memory area.
 * 
 * Complexity: O(area->valuesCount ())
 * 
 * @param ae area entry representing the area to be freed
 * @throws NullPointerException if @c area is null
 * @throws FreedAreaAccessException if the memory area represented 
 *      by @c area was already freed
 */
void 
MemoryState::doFree (AreaEntry* ae)
{
    checkNullPtr_ (ae);
    checkDeleted_ (ae);

    m_state->deleteArea (ae, m_currentUndoSet);
}
      
      
/**
 * Frees memory area.
 * 
 * Complexity: O(a), where a is the number of values in area @c ptr points to.
 * 
 * @param ptr pointer to the beginning of the memory area, that should be freed
 * @throws NullPointerException if @c ptr is null
 * @throws NotAreaBeginException if @c ptr does not point to the beginning 
 *      of the memory area
 * @throws FreedAreaAccessException if @c ptr points to already freed memory area
 */
void 
MemoryState::free (Ptr ptr)
{
    checkAreaBegin_ (ptr);
    
    AreaEntry* ae = ptr.getTargetArea ();
    
    doFree (ae);
}


/**
 * Sets root memory area.
 * This method or @c setRootArea(Ref<T>) must be called exactly once 
 * before call to @c push().
 * 
 * Complexity: O(1).
 * 
 * @param ptr root area; must not be null pointer
 */
void
MemoryState::setRootArea (Ptr ptr)   
{
    m_state->setRootArea (ptr.getTargetArea ());
}

      
/**
 * Returns size of the memory area.
 * 
 * Complexity: O(1)
 * 
 * @param ptr pointer to the memory area which size should be returned
 * @return size of the memory area in bytes
 * @throws NullPointerException if @c ptr is null pointer
 * @throws NotAreaBeginException if @c ptr does not point to the beginning 
 *      of the memory area
 * @throws FreedAreaAccessException if @c ptr points to already freed memory area
 */
size_t 
MemoryState::getSize (Ptr ptr) const
{
    AreaEntry* ae = ptr.getTargetArea ();
    
    checkNullPtr_ (ae);
    checkAreaBegin_ (ptr);
    checkDeleted_ (ae);
    
    return ae->getSize ();
}
          

/**
 * Loads content from memory area.
 * 
 * Complexity: O(log(n)) where n is the number of values in the @c ae area.
 * 
 * @param ae area entry representing memory area
 * @param offset offset of the content in the memory area
 * @return content starting at the specified offset in the memory area
 * @throws NullPointerException if @c ae is @c null
 * @throws FreedAreaAccessException if @c ae represents already freed memory area
 * @throws OutOfBoundsException if @c offset does not lie within the memory area
 * @throws UndefinedMemoryLoadException if there is no content stored in the memory area
 *      starting at the specified offset
 */
const Content* 
MemoryState::doLoad (AreaEntry* ae, size_t offset) const
{         
    checkNullPtr_ (ae);
    checkDeleted_ (ae);
    
    if (!ae->isExpandable() && !(offset < ae->getSize ()))
        throw OutOfBoundsException (Ptr (ae, offset));
    
    ContentEntry* ce = ae->findContentEntry (offset);
    
    if (!ce)
        throw UndefinedMemoryLoadException (Ptr (ae, offset));
    
    return ce->getContent ();
}


/**
 * Loads content from memory.
 * 
 * Complexity: O(log(n)) where n is the number of values in the area @c ptr points to.
 * 
 * @param ptr pointer to the content
 * @return content starting at the specified address; do not modify or delete returned object
 * @throws NullPointerException if @c ptr is null pointer
 * @throws FreedAreaAccessException if @c ptr points to already freed memory area
 * @throws OutOfBoundsException if @c ptr points to the valid memory area, 
 *      but the offset is out of bounds
 * @throws UndefinedMemoryLoadException if there is no content starting at the address 
 *      specified by @c ptr
 */
const Content*
MemoryState::load (Ptr ptr) const
{
    return doLoad (ptr.getTargetArea (), ptr.getOffset ());
}

/**
 * Returns iterator range of content, that overlaps with range <tt>&lt;from, from+size)</tt>.
 * 
 * Complexity: O(log(n)) where n is the number of values int the area @c from points to.
 * 
 * @param from begin address of the range
 * @param size size of the range in bytes
 * @returns iterator range of content that overlaps with specified range
 * @throws NullPointerException if @c from is null pointer
 * @throws FreedAreaAccessException if @c from points to already freed memory area
 * @throws OutOfBoundsException if the specified address range does not lie wholly within 
 *      the area
 */
std::pair<MemoryState::ContentIterator, MemoryState::ContentIterator> 
MemoryState::loadOverlapping (Ptr from, size_t size)
{
    AreaEntry* ae = from.getTargetArea ();
    size_t offset = from.getOffset ();
    
    checkNullPtr_ (ae);
    checkDeleted_ (ae);
    
    if (!ae->isExpandable() && !(offset <= offset+size && offset+size <= ae->getSize()))
        throw OutOfBoundsException (Ptr (ae, offset));
    
    return ae->findContentEntries (offset, offset+size);
}


/**
 * Stores content in the memory area.
 * 
 * Complexity: O(log(n)*(1+r)), where n == @c ae->getAreaCount () 
 * and r is the number of values overwritten by the inserted value.
 * 
 * @param ae area entry representing memory area
 * @param offset offset where should be the content stored
 * @param content value to store; 
 *      must inherit from @c PtrContent or @c NonPtrContent, must not inherit directly from @c Content
 *
 * @throws NullPointerException if @c ae is @c null
 * @throws FreedAreaAccessException if @c ae represents already freed memory area
 * @throws OutOfBoundsException if the content would not lie within the memory area 
 */
void 
MemoryState::doStore (AreaEntry* ae, size_t offset, std::auto_ptr<Content> content)
{
    typedef AreaEntry::ContentEntryIterator CI;
    
    assert (typeid (*content) != typeid (Content));
    
    size_t contentSize = content->getUnitSize ();

    checkNullPtr_ (ae);
    checkDeleted_ (ae);
    
    if (!ae->isExpandable() && !(offset <= offset+contentSize && offset+contentSize <= ae->getSize ()))
        throw OutOfBoundsException (Ptr (ae, offset));
    
    
    
    std::pair<CI, CI> entries = ae->findContentEntries (offset, offset+contentSize);
    if (entries.first != entries.second)    // if there are overlapping entries
    {        
        // remove them
        CI begin = entries.first;
        CI end = entries.second;
        while (begin != end)
        {
            CI oldB = begin;
            begin++;
            m_state->removeContent (ae, &(*oldB), m_currentUndoSet);
        }
    }

    m_state->addContent (ae, offset, content, m_currentUndoSet);
}


/**
 * Stores content in the memory area.
 * All values that overlap with the inserted value are removed.
 * 
 * Complexity: O(log(n)*(1+r)) where n is the number of values in the area @c ptr 
 * point to and r is the number of values overwritten by the inserted value.
 * 
 * @param ptr pointer to the address where should be the content stored
 * @param content content to store
 *
 * @throws NullPointerException if @c ptr is null
 * @throws FreedAreaAccessException if @c ptr points to already freed memory area
 * @throws OutOfBoundsException if the content would not lie within the memory area 
 */
void
MemoryState::store (Ptr ptr, std::auto_ptr<Content> content)
{
    doStore (ptr.getTargetArea (), ptr.getOffset (), content);
}


/**
 * Saves current state of the memory and pushes it onto the stack of marked states.
 * This method does not change current state of the memory.
 * 
 * Complexity: O(t) if links were not significantly modified, 
 *  average O(t+getAreaCount () + p) otherwise,
 *      where 
 *          - t is the time consumed by calls to @c alloc(), @c free()
 *            and @c store() methods (and their templated forms) between this call and previous 
 *            call to @c push() or @c backtrack()
 *          - p is the number of all non-null pointer pointing to other area than the one 
 *            they are stored in
 */
/* Complexity: If relations were not significantly updated, then
 *      average O(mv + sum(1+r(a)*log(sr(a)); for every modified area a)),  
 *    otherwise
 *      average O(getAreaCount () + p + cv + mv + 
 *      sum(r(a)*log(sr(a)); for every modified area a)); 
 *      where 
 *      - mv is the number of modified values
 *      - r(a) is the number of non-null pointers in area a, that were only added and are
 *        pointing to different area than a
 *      - sr(a) is the total number of such pointers in a
 *      - p is the number of all non-null pointers pointing to other area they are stored in
 *      - cv is the number of values 
 *        in all areas which canonical address has changed or which became unreachable.
 */
void 
MemoryState::push ()
{
    saveState_ ();
    
    assert (!m_undoSets.empty ());
    m_markedStates.push_back (m_undoSets.begin ());
    
    getMostRecentMarkedState_ ()->setHash (m_hashHolder->getHash ());
}
    

/**
 * Removes top state from the stack of marked states.
 * Number of calls to this method must not exceed number of calls to @c push().
 * This method does not change current state of the memory.
 * 
 * Complexity: O(1)
 */
void 
MemoryState::pop ()
{
    assert (!m_markedStates.empty ());
    
    m_markedStates.pop_back ();
}
    
/**
 * Checks if there are any states on the stack of marked states.
 * 
 * Complexity: O(1)
 * 
 * @return @c true if there are any states on the stack of marked states;
 */
bool 
MemoryState::hasMarkedState () const
{
    return !m_markedStates.empty ();
}
    

/**
 * Returns the number of undo sets.
 * 
 * Complexity: O(1)
 * 
 * @return the number of saved states; returned value 
 */
size_t 
MemoryState::undoCount () const
{
    return m_undoSets.size();
}


/**
 * Returns the number of saved states.
 * The number of saved states is equal to the number of states 
 * on the stack of marked states.
 * 
 * Complexity: O(1)
 * 
 * @return the number of states on the stack of marked states
 */
size_t 
MemoryState::savedStateCount () const
{
    return m_markedStates.size ();
}


/**
 * Restores the current state of the memory to the top state on the stack of marked states. 
 * Invalidates all pointers (@c Ptr) and references (@c Ref ) returned from calls to 
 * @c alloc(Ptr) and @c alloc(Ref<T>) after the state we are returning to was saved.
 * @pre @c hasMarkedStates() must return @c true
 * 
 * Complexity: O(t_m + t_p), where 
 *  - t_m is the time consumed by calls to @c alloc(), @c free()
 *    and @c store() methods (and their templated forms) between this call and previous 
 *    call to @c push() or @c backtrack()
 *  - t_p is the time consumed by calls to @c push() method which created the undo sets that 
 *    will be disposed by this call
 */
/* Complexity: O(sum(s->m_uncreateAreas.size () + s->m_undeleteAreas.size () 
 *                       + sum((av(a)+rv(a)) * log(a->m_area->valuesCount ()+rv(a)-av(a)); 
 *                             for every a in @c s->m_contentUndos )); for every s from 
 *                        range &lt;@c m_undoSets.begin(), @c getMostRecentMarkedState_() ),
 *     where 
 *       - av(a) is the number of values in unadd list of a
 *       - rv(a) is the number of values in unremove list of a
 */
void 
MemoryState::backtrack ()
{
    LOG_DEBUG ("-- rollback started");
    UndoSets_::iterator s = getMostRecentMarkedState_ ();
    
    pushUndoSet_ ();
    m_state->clearModificationsList ();
    undoToState_ (s);
    
    assert (!m_undoSets.empty ());
    Hash hash = m_undoSets.front ().getHash ();
    m_hashHolder->setHash (hash);
    
    LOG_DEBUG ("-- rollback ended");
}
    

/**
 * Returns the hash of the top state on the stack of marked states.
 * 
 * Complexity: O(1)
 * 
 * @return the hash of the top state on the stack of marked states
 * 
 * @warning this method does not return hash of the current state
 */
Hash 
MemoryState::hash ()
{
    return getMostRecentMarkedState_()->getHash ();
}


/**
 * Recounts hash of the current state and saves modifications between this state and last saved
 * state for later backtrack.
 * 
 * Complexity: If relations were not significantly updated, then
 *      O(mv + sum(r(a)*log(sr(a)); for every modified area a)),  
 *    otherwise
 *      Complexity: O(getAreaCount () + p + cv + mv + 
 *      sum(r(a)*log(sr(a)); for every modified area a)); 
 *      where 
 *      - mv is the number of modified values
 *      - r(a) is the number of non-null pointers in area a, that were only added and are
 *        pointing to different area than a
 *      - sr(a) is the total number of such pointers in a
 *      - p is the number of all pointers pointing to other area they are saved in
 *      - cv is the number of values 
 *        in all areas which canonical address has changed or which became unreachable.
 */
void 
MemoryState::saveState_ ()
{
    LOG_DEBUG ("saving state");
    
    if (!m_state->isModified ())
        return;
    
    bool relUpdated = updateRelations_ ();
    
    // if relations were significantly updated, or push() is called for the first time
    if (relUpdated || !hasMarkedState ())
    {
        m_state->updateCanonAddresses (m_canon.get (), m_currentUndoSet);
        m_state->findUnreachableAreas ();
    }
#ifndef NDEBUG
    else
        m_state->checkCanonAddresses (m_canon.get ());
#endif
    m_state->clearUnreachableAreas (m_listener, m_currentUndoSet);
    updateHash_ ();
    pushUndoSet_ ();
    m_state->clearModificationsList ();
}


/**
 * Saves modifications between current state of the memory and last saved state 
 * and pushes is onto undo-set stack.
 * 
 * Complexity: O(v), where v is the number of modified values in @c m_state.
 */
void
MemoryState::pushUndoSet_ ()
{
    if (!m_state->isModified ())
        return;

    m_undoSets.push_front(*m_currentUndoSet);
    m_currentUndoSet = new UndoSet();
}


/**
 * Returns top state on the stack of marked states.
 */
MemoryState::UndoSets_::iterator
MemoryState::getMostRecentMarkedState_ ()
{
    assert (!m_markedStates.empty ());
    
    return m_markedStates.back ();
}
    

/**
 * Rolls back all the saved modifications from the top of the undo-set stack till 
 * @c si (excluding).
 * 
 * Complexity: O(sum(s->m_uncreateAreas.size () + s->m_undeleteAreas.size () 
 *                       + sum((av(a)+rv(a)) * log(a->m_area->valuesCount ()+rv(a)-av(a)); 
 *                             for every a in @c s->m_contentUndos )); for every s from 
 *                        range &lt;@c m_undoSets.begin(), @c si ),
 *     where 
 *       - av(a) is the number of values in unadd list of a
 *       - rv(a) is the number of values in unremove list of a
 * 
 * @param si iterator to the first @c UndoSet, that will not be rolled back
 */
void 
MemoryState::undoToState_ (UndoSets_::iterator si)
{
    for (UndoSets_::iterator it = m_undoSets.begin (); it != si; )
    {  
        doOneUndo_ (&*it);
        it = m_undoSets.erase_and_dispose (it, boost::checked_delete<UndoSet>);
    }
    m_currentUndoSet = new UndoSet();
}
    
    
/**
 * Rolls back saved modifications @c s.
 * 
 * Complexity: O(s->m_uncreateAreas.size () + s->m_undeleteAreas.size () 
 *                       + sum((av(a)+rv(a)) * log(a->m_area->valuesCount ()+rv(a)-av(a)); 
 *                             for every a in @c s->m_contentUndos )),
 *     where av(a) is the number of values in unadd list of a,
 *          rv(a) is the number of values in unremove list of a
 */
void 
MemoryState::doOneUndo_ (UndoSet* s)
{
    assert (!m_state->isModified ());
    
    s->undo (*m_state.get ());
}


namespace detail {

/**
 * Modifications visitor that updates relations between areas.
 */
class RelationsUpdater {
    bool m_relUpdated;
    std::vector<RefContentEntry*> m_removedTreeLinks;
    
  public:
    RelationsUpdater () : m_relUpdated (false) {}
      
    /*
     * Complexity: O(m_removedTreeLinks.size ())
     */
    bool beforeModifiedArea (AreaEntry*)
    {
        m_removedTreeLinks.clear ();
        return true;
    }
    
    /*
     * Complexity: O(log(ae->outLinksCount ())) if @c ce is added @c RefContentEntry, 
     *               O(1) otherwise
     */
    void modifiedContent (AreaEntry* /* ae */, ContentEntry* ce)
    {
        RefContentEntry* rce = dynamic_cast<RefContentEntry*> (ce);
        if (!rce)
            return;
        if (!rce->getTarget ())
            return;
        
        if (rce->wasRemoved () && rce->wasAdded ())
            return;
        if (rce->getSource () == rce->getTarget ())
            return;
        
        else if (rce->wasRemoved ())
        {   // only removed
            rce->removeLink ();
            
            if (rce->getTarget ()->getBfsTreeLinkToParent () == rce)
            {
                rce->setRelationUpdateNeeded (true);
                m_removedTreeLinks.push_back (rce);
            }
        }
        else if (rce->wasAdded ())
        {   // only added
            rce->addLink ();    
            
            RefContentEntry* parentLink = rce->getTarget ()->getBfsTreeLinkToParent ();
            if (!parentLink || rce->getSource () != parentLink->getSource () ||
                rce->getOffset () < parentLink->getOffset ())
                m_relUpdated = true;
            else if (rce->getOffset () == parentLink->getOffset ())
                parentLink->setRelationUpdateNeeded (false);
        }
    }
    
    /*
     * Complexity: O(m_removedTreeLinks.size ())
     */
    void afterModifiedArea (AreaEntry*) 
    {
        if (m_relUpdated)
            return;
        
        BOOST_FOREACH (RefContentEntry* rce, m_removedTreeLinks)
        {
            if (rce->isRelationUpdateNeeded ())
                m_relUpdated = true;
        }
    }
    
    bool getRelUpdated () const
    { return m_relUpdated; }
};

}


/**
 * Updates relations between areas.
 * 
 * Complexity: O(v + sum(r(a)*log(sr(a)+r(a)); for every modified area a)), 
 *      where v is the number 
 *      of modified values, r(a) is the number of added links in area a, sr(a) is the number of
 *      all links in area a before this function is called
 * 
 * @return @c true if any link between areas was added or removed
 */
bool
MemoryState::updateRelations_ ()
{
    RelationsUpdater vis;
    m_state->walkModifications (vis);
    
#ifndef NDEBUG
    m_state->checkRelations ();
#endif
    
    return vis.getRelUpdated ();
}


namespace detail {

/**
 * Modifications visitor that updates hash of the state of the memory.
 */
class HashUpdater : public DefaultModificationsVisitor {
    HashHolder* m_hashHolder;
    
  public:
    HashUpdater (HashHolder* hashHolder) : m_hashHolder (hashHolder) {}
    
    /*
     * Complexity: Average O(1)
     */
    bool beforeModifiedArea (AreaEntry* ae)
    {
        if (ae->isDeleted () && !ae->wasDeleted ())
        {
            assert (ae->valuesCount () == 0);
            return false;
        }
        
        bool wasAdded = ae->wasCreated ();
        bool wasRemoved = ae->wasDeleted () || !ae->isReachable ();
        bool addHash = false;
        bool removeHash = false;
        
        if (wasAdded && wasRemoved)
        {
            assert (ae->valuesCount () == 0);
            return false;
        }
            
        if (wasAdded)
            addHash = true;
        else if (wasRemoved)
            removeHash = true;
        else
        {
            addHash = true;
            removeHash = true;
        }
        
        if (addHash)
        {
            LOG_DEBUG ("adding hash:   area %p", ae);
            m_hashHolder->addAreaHash (ae->getCanonAddr (), 
                                       ae->getSize ());
        }
        
        if (removeHash)
        {
            LOG_DEBUG ("removing hash: area %p", ae);
            m_hashHolder->removeAreaHash (ae->getOldCanonAddr (),
                                          ae->getSize ());
        }
        
        return true;
    }
    
    /*
     * Complexity: Average O(1)
     */
    void modifiedContent (AreaEntry* ae, ContentEntry* ce)
    {
        if (ce->wasRemoved () && ce->wasAdded ())
            return;
        
        bool addHash = false;
        bool removeHash = false;
        
        if (ce->wasAdded ())
            addHash = true;
        else if (ce->wasRemoved ())
            removeHash = true;
        else
        { 
            addHash = true;
            removeHash = true;
        }
                
        if (addHash)
        {
            LOG_DEBUG ("adding hash:   area %p, offset %zu", ae, ce->getOffset ());
            m_hashHolder->addContentHash (
                ae->getCanonAddr () + ce->getOffset (), 
                ce->getObjectHash ());
        }
        
        if (removeHash)
        {
            LOG_DEBUG ("removing hash: area %p, offset %zu", ae, ce->getOffset ());
            m_hashHolder->removeContentHash (
                ae->getOldCanonAddr () + ce->getOffset (),
                ce->getOldObjectHash ());
        }
    }
};

}


/**
 * Updates hash of current state of the memory.
 * 
 * Complexity: Average O(v), where v is the number of modified values in @c m_state
 */
void 
MemoryState::updateHash_ ()
{
    HashUpdater vis (m_hashHolder.get ());
    m_state->walkModifications (vis);
    
#ifndef NDEBUG
    HashHolder holder (m_areaRandom.get (), m_valueRandom.get ());
    m_state->recountHash (&holder);
    LOG_DEBUG("walked hash: %d, recounted hash: %d", m_hashHolder->getHash (), holder.getHash());
    assert (holder.getHash () == m_hashHolder->getHash ());
#endif 
}


}
