// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

#ifndef MEMORY_STATE_H
#define MEMORY_STATE_H

#include "Ptr.h"
#include "Ref.h"
#include "Types.h"
#include "ContentEntry.h"
#include "AreaEntry.h"
#include "undo/UndoSet.h"

#include <memory>
#include <vector>

#include <boost/intrusive/slist.hpp>
#include <boost/noncopyable.hpp>
#include <boost/iterator/transform_iterator.hpp>


/**
 * Namespace of the memory module. 
 * In this namespace are contained all classes, types and functions that are intended to 
 * be used by the user of the memory module. Internal classes, types and functions of memory module
 * can be found in namespace @c mmodule::detail.
 */
namespace mmodule {
    
namespace detail {
class State;
class Canon;
class HashHolder;
class RandomValues;
}

class MemoryAccessListener;
class Content;


/**
 * Class that represents state of program memory.
 * This class allows the user to manipulate current state of the memory and
 * to backtrack to previously saved states.
 * 
 * Current state of the memory can be changed by calling @c alloc(), @c free(), @c store()
 * and @c load() methods. These methods exist in two forms: first takes or returns @c Ptr
 * parameter, while the second takes or returns @c Ref parameter.
 * 
 * The first form methods are used to save and load values, whose type <em>is not</em> 
 * known at the compile time. These methods will be typically used for manipulation 
 * with variables of interpreted/model checked program.
 * 
 * The second form methods are used to save and load values, whose type @e is know
 * at the compile time. They exist to save the user of this class form dynamic casting 
 * of the returned @c Content objects and from packing every value that he wants to save
 * into some class derived from @c Content. These methods will be typically used for
 * manipulation with interpreter/model checker internal data, that are not accessible to
 * the interpreted/model checked program, but are also part of 
 * the state of the memory (current instruction pointer, stack of called functions, ...).
 * 
 * Functions @c pop(), @c push() and @c backtrack() are used for saving current state of the 
 * memory and for backtracking to previously saved states.
 */
class MemoryState : private boost::noncopyable { 
  private:
    typedef boost::intrusive::slist<detail::UndoSet> UndoSets_;
    typedef std::vector<UndoSets_::iterator> MarkedStates_;
    
    
    struct ContentGetter_ {
        typedef std::pair<size_t, const Content*> result_type;
        result_type operator() (const detail::ContentEntry& ce) const
        { return std::make_pair (ce.getOffset (), ce.getContent ()); }
    };
    
    typedef boost::transform_iterator<ContentGetter_, detail::AreaEntry::ContentEntryIterator>
            ContentIterator_;

    /**
     * Actual @c UndoSet which contains changes since last @c push().
     * This @c UndoSet is not yet on the @c m_undoSets list and must be managed
     * (deallocated) separatelly.
     */
    detail::UndoSet* m_currentUndoSet;

    /**
     * undo-set stack. contains modifications between saved states
     */
    UndoSets_ m_undoSets;
    
    /**
     * stack of marked states
     */
    MarkedStates_ m_markedStates;
    
    /**
     * current state of the memory
     */
    std::auto_ptr<detail::State> m_state;
    
    /**
     * random function used in area partial hash computation
     */
    std::auto_ptr<detail::RandomValues> m_areaRandom;
    
    /**
     * random function used in value partial hash computation
     */
    std::auto_ptr<detail::RandomValues> m_valueRandom;
    
    /**
     * hash of the current state.
     * the hash is up-to-date only immediately after call to @c saveState_().
     */
    std::auto_ptr<detail::HashHolder> m_hashHolder;
    
    /**
     * Canon function
     */
    std::auto_ptr<detail::Canon> m_canon;
    
    /**
     * listener
     */
    MemoryAccessListener* m_listener;
    
  public:
    /** 
     * type of iterator to the content of memory area.
     * the value is pair: @c first is the offset of the content within the area, 
     * @c second is the content itself
     */
    typedef ContentIterator_ ContentIterator;
    
    MemoryState (MemoryAccessListener* listener);
    ~MemoryState ();
            
    Ptr alloc (size_t size);
    Ptr allocExpanding ();
    void free (Ptr ptr);
    void setRootArea (Ptr ptr);
    
    void store (Ptr ptr, std::auto_ptr<Content> content);
    const Content* load (Ptr ptr) const;
    std::pair<ContentIterator, ContentIterator> loadOverlapping (Ptr from, size_t size);
    
    size_t getSize (Ptr ptr) const;
    
    template <typename T> Ref<T> alloc ();
    template <typename T> Ref<T> allocExpanding ();
    template <typename T> void free (Ref<T> ref);
    template <typename T> void setRootArea (Ref<T> ref);
    
    template <typename T, typename V> const V& load (Ref<T> ref, V T::* member) const;
    template <typename T, typename V> const V& loadIndexed (Ref<T> ref, V T::* member, size_t index) const;
    template <typename T, typename V> void store (Ref<T> ref, V T::* member, const V& value);
    template <typename T, typename V> void storeIndexed (Ref<T> ref, V T::* member, size_t index, const V& value);
    
    
    void push ();
    void pop ();
    void backtrack ();
    Hash hash ();
    
    bool hasMarkedState () const;
    size_t undoCount () const;
    size_t savedStateCount () const;
    
    size_t getAreaCount () const;
    
  private:    
    static void checkDeleted_ (detail::AreaEntry* ae);
    
    detail::AreaEntry* doAlloc (size_t size);
    detail::AreaEntry* doAllocExpanding ();
    void doFree (detail::AreaEntry* ae);
    
    const Content* doLoad (detail::AreaEntry* ae, size_t offset) const;
    void doStore (detail::AreaEntry* ae, size_t offset, std::auto_ptr<Content> content);
    
    void saveState_ ();
    UndoSets_::iterator getMostRecentMarkedState_ ();
    void undoToState_ (UndoSets_::iterator s);
    bool updateRelations_ ();
    void updateHash_ ();    
    void doOneUndo_ (detail::UndoSet* s);
    
    void pushUndoSet_ ();
};

}

#include "MemoryState_inline.h"
                 

#endif 
