// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

#include "State.h"
#include "Content.h"
#include "../mm/MemoryState.h"

#include <boost/functional/hash.hpp>
#include <typeinfo>


namespace mmodule {

/**
 * Allocates new memory area, that can hold values from structure @c T.
 * 
 * Complexity: O(1)
 * 
 * @tparam T POD structure
 * @warning
 *      there is no way how to check, that @c T is really a POD structure, 
 *      on compilation time, 
 *      so if @c T is not a POD structure, some strange compilation errors might occur, 
 *      or if everything compiles then the behaviour is undefined
 */
template <typename T> 
Ref<T> 
MemoryState::alloc ()
{
    return Ref<T> (doAlloc (sizeof (T)));
}
    

/**
 * Allocates new expanding memory area, that can hold values from
 * structure @c T.
 *
 * Complexity: O(1)
 *
 * @tparam T POD structure
 * @warning
 *      there is no way how to check, that @c T is really a POD structure,
 *      on compilation time,
 *      so if @c T is not a POD structure, some strange compilation errors might occur,
 *      or if everything compiles then the behaviour is undefined
 */
template <typename T>
Ref<T>
MemoryState::allocExpanding ()
{
    return Ref<T> (doAllocExpanding ());
}


/**
 * Frees memory area previously allocted by @c alloc<T>().
 * 
 * Complexity: O(a), where a is the number of values stored 
 * in the area @c ref points to.
 * 
 * @param ref reference to memory area that should be freed
 * @throws NullPointerException if @c ref is null
 * @throws FreedAreaAccessException if the memory area represented 
 *      by @c ref was already freed
 */
template <typename T> 
void
MemoryState::free (Ref<T> ref)
{
    doFree (ref.getAreaEntry ());
}


/**
 * Sets root memory area.
 * This method or @c setRootArea(Ptr) must be called exactly once before call to @c push().
 * 
 * Complexity: O(1)
 * 
 * @param ref reference to root memory area; must not be null reference
 */
template <typename T> 
void 
MemoryState::setRootArea (Ref<T> ref)
{
    m_state->setRootArea (ref.getAreaEntry ());
}
    
    
namespace detail {
    
/**
 * Returns offset to the member of structure.
 */
template <typename T, typename V>
static size_t
memberPtrToOffset_ (V T::* member)
{
    struct TmpStr_ {
        int8_t data[sizeof (T)];
    };
    
    static TmpStr_ val;
    static T* t = reinterpret_cast<T*> (&val);
    V* v = &(t->*member);
    int8_t* tv = (int8_t*)t;
    int8_t* vv = (int8_t*)v;

    return vv-tv;
}
    

/**
 * Base class for content classes.
 */
template <typename V>
class BaseContent_ : virtual public Content {
  private:
    V m_value;
    
  public:
    BaseContent_ (const V& v) : m_value (v) {}
    virtual ~BaseContent_ () {}
    
    // TODO fix to return right value for Ptr
    size_t getUnitSize () const
    { return sizeof (V); }
    
    const V& getValue () const
    { return m_value; }
};


/**
 * Content class for holding non-pointer values.
 */
template <typename V>
struct Content_ : BaseContent_<V>, NonPtrContent {    
    Content_ (const V& v) : BaseContent_<V> (v) {}
    virtual ~Content_ () {}
    
    Hash getHash () const
    { 
        std::size_t seed = 0;

        boost::hash_combine (seed, BaseContent_<V>::getValue ());
        boost::hash_combine (seed, &typeid (V));
    
        return seed;
    }
};


/**
 * Content class for holding @c Ptr values.
 */
template <>
struct Content_<Ptr> : BaseContent_<Ptr>, PtrContent {
    Content_ (const Ptr& ptr) : BaseContent_<Ptr> (ptr) {}
    virtual ~Content_ () {}
        
    Ptr getPtr () const
    { return getValue (); }
    
    Hash getTypeHash () const
    { return boost::hash<const std::type_info*> () (&typeid (Ptr)); }
};


/**
 * Content class for holding @c Ref values.
 */
template <typename T>
struct Content_<Ref<T> > : BaseContent_<Ref<T> >, PtrContent {
    Content_ (const Ref<T>& ref) : BaseContent_<Ref<T> > (ref) {}
    virtual ~Content_ () {}
    
    Ptr getPtr () const
    { return Ptr (this->getValue ().getAreaEntry (), 0); }
    
    Hash getTypeHash () const
    { return boost::hash<const std::type_info*> () (&typeid (Ref<T>)); }
};


/**
 * Creates content class for type @c V.
 * 
 * @param v value of the content
 * @return object holding value @c v; if @c V is @c Ptr or @c Ref 
 *      returns object of class derived from @c PtrContent, 
 *      else returns object of class derived from @c NonPtrContent.
 */
template <typename V>
std::auto_ptr<Content> 
createContent_ (const V& v)
{
    std::auto_ptr<Content> ret (new Content_<V> (v));
    return ret;
}

} /* namespace detail */


/**
 * Loads member of structure @c T.
 * 
 * Complexity: O(log(n)) where n is the number of values in the area @c ref references.
 * 
 * @param ref reference to the memory area, where is the member stored
 * @param member pointer to the member of the structure, that should be loaded
 * @return reference to stored member
 * @throws NullPointerException if @c ref is null reference
 * @throws FreedAreaAccessException if @c ref references already freed memory area
 * @throws UndefinedMemoryLoadException if there is no member stored in the memory area 
 *      (no corresponding @c store(Ref<T>) was called before call to this function)
 */
template <typename T, typename V> 
const V& 
MemoryState::load (Ref<T> ref, V T::* member) const
{
    const Content* content = doLoad (ref.getAreaEntry (), detail::memberPtrToOffset_ (member));
    const detail::BaseContent_<V>* baseContent = 
            dynamic_cast<const detail::BaseContent_<V>*> (content);
    assert (baseContent);
    return baseContent->getValue ();
}

template <typename T, typename V>
const V&
MemoryState::loadIndexed (Ref<T> ref, V T::* member, size_t index) const
{
    const Content* content = doLoad (ref.getAreaEntry (), detail::memberPtrToOffset_ (member) + (sizeof(V) * index));
    const detail::BaseContent_<V>* baseContent =
            dynamic_cast<const detail::BaseContent_<V>*> (content);
    assert (baseContent);
    return baseContent->getValue ();
}


/**
 * Stores member of structure @c T.
 * 
 * Complexity: O(log(n)+t_n) where 
 *  - n is the number of values in the area @c ref references
 *  - t_n is the time needed to create a copy of @c value
 * 
 * @tparam T type of the structure according to which are the values lied out
 * @tparam V type of the value stored; <tt>boost::hash&lt;V&gt; (value)</tt> must be 
 *      valid expression returning the hash of the @c value; specializations of
 *      <tt>boost::hash()</tt> function exist for @c Ref and @c Ptr classes.
 * @param ref reference to the memory area, where should be the member stored
 * @param member pointer to the member of the structure, that should be stored; 
 *      this function assumes, that saved value references a memory area 
 *      (and therefore must be checked when counting cannon addresses) if and only if  
 *      @c V is of @c Ptr or @c Ref type
 * @param value value of the member
 * @throws NullPointerException if @c ref is null reference
 * @throws FreedAreaAccessException if @c ref references already freed memory area
 */
template <typename T, typename V> 
void 
MemoryState::store (Ref<T> ref, V T::* member, const V& value)
{
    doStore (ref.getAreaEntry (), detail::memberPtrToOffset_ (member), 
             detail::createContent_<V> (value));
}

template <typename T, typename V>
void
MemoryState::storeIndexed (Ref<T> ref, V T::* member, size_t index, const V& value)
{
	LOG_DEBUG("size of T: %zu", sizeof(V));
	LOG_DEBUG("storing to offset %zu", detail::memberPtrToOffset_ (member) + (sizeof(V) * index));
	doStore (ref.getAreaEntry (), detail::memberPtrToOffset_ (member) + (sizeof(V) * index),
	             detail::createContent_<V> (value));
}

}   /* namespace mmodule */
