// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

#ifndef MM_PTR_H
#define MM_PTR_H

#include <exception>
#include <boost/operators.hpp>


namespace mmodule {
    
namespace detail {
class AreaEntry;
class RefContentEntry;
class HashGetter;
class OldHashGetter;
template <typename T> class Content_;
}

class UndefinedPtrOperationException;
struct NullPointerException;


/**
 * Class representing untyped pointer into memory.
 * 
 * Pointer is composed of two parts:
 * - area part
 * - offset part
 * 
 * <em>Area part</em> represents the memory object.
 * 
 * <em>Offset part</em> represents the offset to the content within the memory object.
 * 
 * @see Ref
 */
class Ptr : 
    boost::totally_ordered<Ptr, 
    boost::additive<Ptr, size_t> > {
  
  private:
    detail::AreaEntry* m_area;
    size_t m_offset;

  public:
    /**
     * Constructs null pointer.
     */
    Ptr () : m_area (0), m_offset (0) 
    {}
    
    
    const Ptr& operator+= (size_t i);
    const Ptr& operator-= (size_t i);
    
    /**
     * Checks if this pointer and @c ptr point to the same address.
     * @c boost::totally_ordered parent class implements the operator <tt>ptr1 != ptr2</tt>.
     * 
     * @param ptr pointer to compare with
     * @return @c true if this pointer and @c ptr point to the same address, @c false otherwise
     */
    bool operator== (const Ptr& ptr) const
    {
        return m_area == ptr.m_area && m_offset == ptr.m_offset;
    }
    
    /**
     * Checks if this pointer is null pointer (does not point to any valid address).
     * 
     * @return @c true if this pointer is null pointer, @c false otherwise
     */
    bool isNull () const
    {
        return m_area == 0;
    }
    
    /**
     * Returns offset part of this pointer.
     * 
     * @return offset part of this pointer
     */
    size_t getOffset () const 
    { return m_offset; }
    
    size_t getAreaId () const;
    size_t getAreaSize () const;
    bool isAreaDeleted () const;
    
  protected:
    /**
     * Constructor.
     * 
     * @param area area that the pointer should reference
     * @param offset offset in the area 
     */
    Ptr(detail::AreaEntry* area, size_t offset) : 
        m_area (area), m_offset (offset) {}
    
    /**
     * Returns object to which this pointer points.
     */
    detail::AreaEntry* getTargetArea () const
    { return m_area; }
    
    /**
     * Checks if this pointer points to the same area as @c ptr.
     */
    bool inSameArea (Ptr ptr) const 
    { return m_area == ptr.m_area; }

    bool isAreaExpandable () const;

  private:
    
    friend class MemoryState;
    friend class detail::RefContentEntry;
    template <typename T> friend class detail::Content_;
    friend class detail::HashGetter;
    friend class detail::OldHashGetter;
    friend class detail::AreaEntry;
};


/**
 * Common base class for all exceptions that can be thrown when working with pointers.
 */
struct PtrException : public std::exception {
};


/**
 * Exception that is thrown when trying to dereference null pointer.
 */
struct NullPointerException : public PtrException {
    const char* what () const throw ()
    {
        return "dereferencing null pointer";
    }
};


/**
 * Exception that is thrown when trying to do some undefined operation with pointers.
 */
struct UndefinedPtrOperationException : public PtrException {
    UndefinedPtrOperationException () {}
    
    const char* what () const throw ()
    {
        return "undefined pointer operation";
    }
};


/**
 * Moves this pointer to higher addresses.
 * @c boost::additive parent class implements the operators 
 * <tt>ptr + i</tt> and <tt>i + ptr</tt>.
 * 
 * @param i offset to add to this pointer
 * @throw NullPointerException if @c this is a null pointer
 * @throws UndefinedPtrOperationException if it is tried to move the pointer further then 
 *      to the byte just after the area
 */
inline const Ptr& 
Ptr::operator+= (size_t i) 
{
    /* 
     * the argument is really size_t because second argument of
     * si POINTER_PLUS_EXPR is always unsigned type compatible with sizetype.
     */
    if (isNull ())
        throw NullPointerException ();

    if (!isAreaExpandable() && (getAreaSize () < m_offset + i))
        throw UndefinedPtrOperationException ();
        
    m_offset += i;
    return *this;
}
    
    
/** 
 * Moves this pointer towards lower addresses.
 * @c boost::additive parent class implements the operator
 * <tt>ptr - i</tt>
 * 
 * @param i offset to subtract from this pointer
 * @throw NullPointerException if @c this is a null pointer
 * @throws UndefinedPtrOperationException if it is tried to move the pointer before the 
 *      beginning of the area
 */
inline const Ptr& 
Ptr::operator-= (size_t i)
{ 
    if (isNull ())
        throw NullPointerException ();
    if (m_offset < i)
        throw UndefinedPtrOperationException ();
        
    m_offset -= i;
    return *this;
}


/**
 * Compares two pointers by @c < operator.
 * @c boost::totally_ordered parent class implements the operators
 * @c >, @c <=, @c >=.
 * 
 * @param ptr1 first argument
 * @param ptr2 second argument
 * @return @c true if @c ptr1 represents lower address than @c ptr2, @c false otherwise;
 * @throws UndefinedPtrOperationException if @c ptr1 point to different area than @c ptr2
 */
inline bool 
operator< (const Ptr& ptr1, const Ptr& ptr2) 
{
    if (ptr1.getAreaId () != ptr2.getAreaId ())
        throw UndefinedPtrOperationException ();
    
    return ptr1.getOffset () < ptr2.getOffset ();
}


/**
 * Computes difference of two pointers.
 * 
 * @param ptr1 first argument
 * @param ptr2 second argument
 * @return address of @c ptr1 subtracted from address of @c ptr2
 * @throws UndefinedPtrOperationException if @c ptr1 point to different area than @c ptr2
 */
inline size_t
operator- (const Ptr& ptr1, const Ptr& ptr2) 
{
    if (ptr1.getAreaId () != ptr2.getAreaId ())
        throw UndefinedPtrOperationException ();
    
    return ptr1.getOffset () - ptr2.getOffset ();
}

}

#endif 


