// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

#include "State.h"

#include "AreaEntry.h"
#include "Canon.h"
#include "MemoryAccessListener.h"
#include "DefaultModificationsVisitor.h"
#include "../Error.h"
#include "undo/AllocUndo.h"
#include "undo/FreeUndo.h"
#include "undo/AddContentUndo.h"
#include "undo/RemoveContentUndo.h"
#include "undo/ClearUnreachableAreasUndo.h"

#include <boost/foreach.hpp>
#include <boost/graph/breadth_first_search.hpp>
#include <boost/graph/properties.hpp>


namespace mmodule { 

/**
 * Namespace that contains all internal classes, types and functions of memory module.
 */
namespace detail {

/**
 * Constructor.
 */
State::State () : m_root (0)
{
}


/**
 * Destructor.
 */
State::~State ()
{
    clearModificationsList ();
}


/**
 * Creates area.
 * The area is added into the modifications list.
 * 
 * Complexity: O(1)
 * 
 * @param size size of the area to create
 * @param undoSet @c UndoSet to be filled with undo info.
 * @return area entry representing the newly created area
 */
AreaEntry*
State::createArea (size_t size, UndoSet* undoSet)
{
    std::auto_ptr<AreaEntry> ae (new AreaEntry (size));
    m_areas.push_back (*ae);
    ae->setWasCreated (true);
    LOG_DEBUG ("created area %p", ae.get ());
                  
    addToModifiedAreas_ (ae.get ());
    undoSet->addUndo(new AllocUndo(ae.get()));
    return ae.release ();
}
    

/**
 * Creates an expanding area.
 * The area is added into the modifications list.
 *
 * Complexity: O(1)
 *
 * @param undoSet @c UndoSet to be filled with undo info.
 * @return area entry representing the newly created area
 */
AreaEntry*
State::createExpandingArea (UndoSet* undoSet)
{
    std::auto_ptr<AreaEntry> ae (new AreaEntry ());
    m_areas.push_back (*ae);
    ae->setWasCreated (true);
    LOG_DEBUG ("created expanding area %p", ae.get ());

    addToModifiedAreas_ (ae.get ());
    undoSet->addUndo(new AllocUndo(ae.get()));
    return ae.release ();
}


/**
 * Deletes area.
 * The area entry representing the area is no really deleted, it's only marked as deleted.
 * This is done because when backtracking the area will be un-deleted again and because
 * we want to safely check access to freed memory.
 * The area is added into the modifications list.
 * 
 * Complexity: O(area->valuesCount ()).
 * 
 * @param area area entry representing area that should be deleted
 * @param undoSet @c UndoSet to be filled with undo info.
 */
void
State::deleteArea (AreaEntry* area, UndoSet* undoSet)
{
    area->setDeleted (true);
    area->setWasDeleted (true);
    
    area->clearContent (undoSet);
    LOG_DEBUG ("deleted area %p", area);
    
    addToModifiedAreas_ (area);
    undoSet->addUndo(new FreeUndo(area));
}


/**
 * Un-deletes area.
 * The area <em>is not</em> added into the modifications list.
 * This is called when backtracking to some previously saved state.
 * 
 * Complexity: O(1)
 * 
 * @param area area entry representing area that should be un-deleted
 */
void 
State::undeleteArea (AreaEntry* area)
{
    area->setDeleted (false);
}
    

/**
 * Un-creates area.
 * The area <em>is not</em> added into the modifications list.
 * Complexity: O(1)
 * 
 * @param area area entry representing area that should be un-created
 */
void
State::uncreateArea (AreaEntry* area)
{
    m_areas.erase (Areas::s_iterator_to (*area));
}

/**
 * Un-clears unreachable areas (removed by \c clearUnreachableAreas method).
 *
 * The area <em>is not</em> added into the modifications list.
 * Complexity: O(1)
 *
 * @param area area entry representing area that should be un-cleaned
 */
void
State::unclearUnreachableAreas(AreaEntry* area)
{
    m_areas.push_back(*area);
}

/**
 * Adds area entry into modified area list if it is not already there.
 * Complexity: O(1)
 * 
 * @param area area entry representing the area that should be included 
 *      into modified area list
 */
void
State::addToModifiedAreas_ (AreaEntry* area)
{
    if (!area->ModifiedAreasBaseHook::is_linked ())
    {
        m_modifiedAreas.push_front (*area);
    }
}


/**
 * Adds content into area.
 * The area is added into the modifications list.
 * 
 * Complexity: O(log(area->valuesCount ())).
 * 
 * @param area area entry representing the area
 * @param offset the offset where should be the content stored
 * @param content content to store
 * @param undoSet @c UndoSet to be filled with undo info.
 */
void 
State::addContent (AreaEntry* area, size_t offset, std::auto_ptr<Content> content,
        UndoSet* undoSet)
{
    ContentEntry* ce = area->addContent (offset, content);
    undoSet->addUndo(new AddContentUndo(area, ce));
    addToModifiedAreas_ (area);
}
    

/**
 * Removes content from the area.
 * The area is added into the modifications list.
 * 
 * Complexity: Average O(1)
 * 
 * @param area area entry representing the area
 * @param content content to remove
 * @param undoSet @c UndoSet to be filled with undo info.
 */
void
State::removeContent (AreaEntry* area, ContentEntry* content, UndoSet* undoSet)
{
    undoSet->addUndo(new RemoveContentUndo(area, content));
    area->removeContent (content);
    addToModifiedAreas_ (area);
}


/**
 * Helper class for clearing of unreachable areas.
 */
template <typename Vis, typename Areas>
class UnreachableAreaCleaner : public DefaultModificationsVisitor 
{
    Vis* m_vis;
    Areas* m_areas;
    UndoSet* m_undoSet;
    
  public:
    UnreachableAreaCleaner (Vis* vis, Areas* areas, UndoSet* undoSet) : m_vis (vis), m_areas (areas), m_undoSet(undoSet) {}
    
    bool beforeModifiedArea (AreaEntry* ae)
    {
        if (!ae->isReachable ())
        {
            assert (!ae->wasReportedUnreachable ());
            assert (ae->StateAreasBaseHook::is_linked ());
            
            m_areas->erase (Areas::s_iterator_to (*ae));
            
            if (!ae->isDeleted ())
            {
                // unreachable non-deleted area
            	ae->clearContent (m_undoSet);
                ae->reportUnreachable (m_vis);
            }
            m_undoSet->addUndo( new ClearUnreachableAreasUndo(ae));
        }
        
        return false;
    }
};


/**
 * Clears content of all unreachable, not deleted areas and reports them to @c listener.
 * All unreachable, not deleted areas that were not reported before, 
 * are added into the modifications list.
 * 
 * Complexity: O(v + sum(a->valuesCount (); 
 *               a is non-deleted unreachable area not reported before)), 
 *      where a is the number of modified areas.
 * 
 * @param listener runtime visitor 
 * @param undoSet @c UndoSet to be filled with undo info.
 */
void
State::clearUnreachableAreas (MemoryAccessListener* listener, UndoSet* undoSet)
{                                    
    UnreachableAreaCleaner<MemoryAccessListener, Areas> cleaner (listener, &m_areas, undoSet);
    walkModifications (cleaner);
}


/**
 * Finds all unreachable areas and adds them into list of modified areas.
 * @warning This works only immediatelly after a call to
 * @c updateCanonAddresses, since the implementation relies on the fact
 * that color of the unreachable areas is left white.
 * 
 * Complexity: O(areasCount ())
 */
void
State::findUnreachableAreas ()
{
    BOOST_FOREACH (AreaEntry& ae, m_areas)
    {
        //if (!ae.isReachable ())
        if (ae.getColor() == boost::color_traits<AreaColorMap::value_type>::white())
            addToModifiedAreas_ (&ae);
    }
}


/**
 * Checks if the state was modified since last call to @c clearModificationsList().
 * 
 * @return @c true if was modified, @c false if not
 */
bool
State::isModified () const
{
    return !m_modifiedAreas.empty ();
}


/**
 * Clears modified areas list.
 * 
 * Complexity: O(s), where s is the number of all modified values
 */
void 
State::clearModificationsList ()
{
    BOOST_FOREACH (AreaEntry& ae, m_modifiedAreas)
    {
        ae.clearModificationsList ();
        ae.clearTmpFlags ();
    }
    m_modifiedAreas.clear ();
}


/**
 * Visitor for the BFS algorithm that computes the canonical addresses of the areas.
 */
class CanonCounterVisitor : public boost::default_bfs_visitor {
    Canon *m_canon;
    AreaEntry* m_root;
    UndoSet* m_undoSet;
    
    public:
        CanonCounterVisitor (Canon* canon, AreaEntry* root, UndoSet* undoSet)
    : m_canon (canon), m_root (root), m_undoSet (undoSet) {}
    
        void initialize_vertex (const AreaEntry* ae, const State& state)
        {
            AreaEntry* mae = const_cast<AreaEntry*> (ae);
            if (mae == m_root)
                updateCanonAddress (state, 0, mae, 0);
            //else
            //    mae->getProperties ().setCanonAddr (CANON_UNREACHABLE);
        }
    
        /*
         * Complexity: O(rce->getTarget ()->valuesCount () + rce->getTarget ()->inLinksCount ()) 
         * if canonical address of area @c rce->getTarget () has changed, O(1) otherwise
         */
        void tree_edge (RefContentEntry* rce, const State& state)
        {
            CanonAddr srcAddr = rce->getSource ()->getCanonAddr ();
            CanonAddr refAddr = srcAddr; //NOTE: The offset is superflous + rce->getOffset ();
        
            AreaEntry* target = rce->getTarget ();
            updateCanonAddress (state, refAddr, target, rce);
        }
        
    private:
        /*
         * Complexity: O(target->valuesCount () + target->inLinksCount ()) if canonical address 
         * of area @c target has changed, O(1) otherwise
         */
        void updateCanonAddress (const State& state, const CanonAddr& refAddr, 
                                 AreaEntry* target, RefContentEntry* rce)
        {
            CanonAddr tarAddr = m_canon->get (refAddr, target->getSize ());
        
            State& s = const_cast<State&> (state);
            s.updateAreaCanonAddress (target, tarAddr, m_undoSet);
            target->setBfsTreeLinkToParent (rce);
        }
};


/**
 * Computes canonical addresses of all areas in the state (including deleted areas).
 * 
 * Complexity: O(getAreaCount () + r + v), where r is the number of non-null poiners pointing 
 * to different area than they reside in
 * (not including pointers in deleted/unreachable areas), v is the number of values 
 * in all areas which canonical address has changed.
 *
 * @param canon the canonicalization function
 * @param undoSet the @c UndoSet to be filled with undos for canon address
 *          changes
 */
void 
State::updateCanonAddresses (Canon* canon, UndoSet* undoSet)
{
    LOG_DEBUG ("updating canonical addresses");
    assert (m_root);
    
    CanonCounterVisitor canonVisitor (canon, m_root, undoSet);
    boost::breadth_first_search (
            *this, m_root,
            boost::visitor (canonVisitor).color_map (getAreaColorMap ()));
}


/**
 * Updates canonical address stored in the area.
 * If the canonical address changed, the area is added into the modifications list.
 * 
 * Complexity: O(ae->valuesCount () + ae->inLinksCount ()) if canonical address of area @c ae
 *      has changed, O(1) otherwise
 * 
 * @param ae area entry representing the area
 * @param addr new canonical address
 * @param undoSet the @c UndoSet to be filled with undos for canon address
 *          changes
 */
void 
State::updateAreaCanonAddress (AreaEntry* ae, CanonAddr addr, UndoSet* undoSet)
{
    if (ae->changeCanonAddr (addr, undoSet))
    {
        ae->reportModifiedCanonAddress ();
        addToModifiedAreas_ (ae);
    }
}


#ifndef NDEBUG
/**
 * Checks if the area relations match with the area contents.
 * This function is called only when debuging is enabled.
 */
void
State::checkRelations ()
{
    LOG_DEBUG ("checking relations");
    BOOST_FOREACH (AreaEntry& ae, m_areas)
    {
        ae.checkRelations ();
    }
}


/**
 * Visitor for the BFS algorithm that checks the canonical addresses of the areas.
 */
class CanonCheckerVisitor : public boost::default_bfs_visitor {
    Canon *m_canon;
    
    public:
        CanonCheckerVisitor (Canon* canon) 
    : m_canon (canon) {}
        
        void examine_vertex (const AreaEntry* ae, const State&)
        {
            assert (ae->isReachable ());
        }
    
        void tree_edge (RefContentEntry* rce, const State&)
        {
            CanonAddr srcAddr = rce->getSource ()->getCanonAddr ();
            CanonAddr refAddr = srcAddr; //NOTE: The offset is superflous + rce->getOffset ();
        
            AreaEntry* target = rce->getTarget ();
            CanonAddr tarAddr = m_canon->get (refAddr, target->getSize ());
        
            assert (target->getCanonAddr () == tarAddr);
        }
};


/**
 * Checks values of canon addresses. 
 * Useful to check if the relations were really not significantly modified.
 */
void 
State::checkCanonAddresses (Canon* canon)
{
    assert (m_root);
    
    CanonCheckerVisitor canonVisitor (canon);
    boost::breadth_first_search (
            *this, m_root,
            boost::visitor (canonVisitor).color_map (getAreaColorMap ()));
}


/**
 * Calculates the hash of the state of the memory.
 * This function inspects every area and every content in the state.
 * This function is called only when debuging is enabled.
 * 
 * @param hashHolder hash holder
 */
void
State::recountHash (HashHolder* hashHolder) const
{
    LOG_DEBUG("Recounting hash of all %zu areas.", m_areas.size());
    BOOST_FOREACH (const AreaEntry& ae, m_areas)
        ae.recountHash (hashHolder);
}

#endif

}}
