// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

#ifndef MM_STATE_H
#define MM_STATE_H

#include "../Error.h"
#include "AreaColorMap.h"
#include "AreaEntry.h"
#include "undo/UndoSet.h"

#include <memory>
#include <boost/intrusive/list.hpp>
#include <boost/noncopyable.hpp>


namespace mmodule { 
        
class MemoryAccessListener;


namespace detail {

class AreaEntry;
class Canon;
class HashHolder;


/**
 * Represents the current state of the memory.
 * This class tracks all changes to the memory and saves them for later inquiry.
 * This class is a model of <tt>Vertex List Graph</tt> and
 * <tt>Incidence Graph</tt> concepts defined in @c boost::graph library.
 */
class State : private boost::noncopyable {
  private:    
    typedef boost::intrusive::constant_time_size<false> SizeOp;
    template <typename T> struct BaseHookOp : boost::intrusive::base_hook<T> {};
    
    typedef boost::intrusive::list<AreaEntry, BaseHookOp<StateAreasBaseHook> >
            Areas;
    typedef boost::intrusive::list<AreaEntry, BaseHookOp<ModifiedAreasBaseHook>, SizeOp >
            ModifiedAreas;
    
    
    /**
     * list of all areas in the state of the memory; contains also deleted areas
     */
    Areas m_areas;
        
    /**
     * root area where starts the canonicalization algorithm
     */
    AreaEntry* m_root;
      
    /**
     * list of modified areas;
     * the list owns all newly created areas
     */
    ModifiedAreas m_modifiedAreas;

  public:          
    /*! iterator through areas in the state */
    typedef Areas::iterator AreaEntryIterator;
    
    /*! constant iterator through areas in the state */
    typedef Areas::const_iterator AreaEntryConstIterator;
    
    
    State ();
    ~State ();
    
    AreaEntry* createArea (size_t size, UndoSet* undoSet);
    AreaEntry* createExpandingArea (UndoSet* undoSet);
    void deleteArea (AreaEntry* area, UndoSet* undoSet);
    
    void undeleteArea (AreaEntry* area);
    void uncreateArea (AreaEntry* area);
    void clearUnreachableAreas (MemoryAccessListener* listener, UndoSet* undoSet);
    void unclearUnreachableAreas(AreaEntry* area);
    
    /**
     * Un-makes area unreachable.
     * 
     * @param ae area to make unreachable
     */
    void unmakeUnreachableArea (AreaEntry* ae)
    { 
        m_areas.push_back (*ae); 
        ae->setReportedUnreachable (false);
    }
    
    void addContent (AreaEntry* area, size_t offset, std::auto_ptr<Content> content,
        UndoSet* undoSet);
    void removeContent (AreaEntry* area, ContentEntry* content, UndoSet* undoSet);
        
    /**
     * Returns the root area.
     * 
     * @return the root area; @c setRootArea() must be called before call to this method
     */
    AreaEntry* getRootArea () const
    { 
        assert (m_root);
        return m_root; 
    }
    
    /**
     * Sets root area. 
     * Must not be called more than once.
     * 
     * @param root root area; must not be @c null
     */
    void setRootArea (AreaEntry* root)
    { 
        assert (root);
        assert (!m_root);
        m_root = root; 
    }
    
    /**
     * Returns map for saving/loading colors in areas.
     * 
     * @return map for saving/loading colors in areas
     */
    AreaColorMap getAreaColorMap ()
    { return AreaColorMap (); }
    
    
    /**
     * Returns number of areas in the state including deleted areas, but excluding 
     * reported undeleted areas.
     * 
     * @return number of areas in the state including deleted areas
     */
    size_t getAreaCount () const
    { return m_areas.size (); }
    
    /**
     * Returns range with areas in the state including deleted areas, but excluding 
     * reported undeleted areas.
     * 
     * @return range with areas in the state including deleted areas
     */
    std::pair<AreaEntryConstIterator, AreaEntryConstIterator> areas () const
    { return std::make_pair (m_areas.begin (), m_areas.end ()); }
    
    /**
     * Returns range with areas in the state including deleted areas, but excluding 
     * reported undeleted areas.
     * 
     * @return range with areas in the state including deleted areas
     */
    std::pair<AreaEntryIterator, AreaEntryIterator> areas () 
    { return std::make_pair (m_areas.begin (), m_areas.end ()); }
    
    
    bool isModified () const;
    void clearModificationsList ();   
    void findUnreachableAreas ();
    void updateCanonAddresses (Canon* canon, UndoSet* undoSet);
    
    /**
     * Walks all the modifications since last call to @c clearModificationsList().
     * 
     * @param modVis visitor that is informed about the changes
     * @tparam Visitor type of the visitor that models Modifications Visitor concept; 
     * @see @c DefaultModificationsVisitor
     */
    template <typename Visitor>
    void walkModifications (Visitor& modVis)
    {
        LOG_DEBUG("Traversing %zu modified areas (%zu total).", m_modifiedAreas.size(), m_areas.size());
        BOOST_FOREACH (AreaEntry& ae, m_modifiedAreas)
        {
            if (modVis.beforeModifiedArea (&ae))
            {
                ae.walkModifiedContent (modVis);
                modVis.afterModifiedArea (&ae);
            }
        }
    }

    
#ifndef NDEBUG
    void checkRelations ();
    void checkCanonAddresses (Canon* canon);
    void recountHash (HashHolder* hashHolder) const;
#endif
  protected:
    void updateAreaCanonAddress (AreaEntry* oe, CanonAddr addr, UndoSet* undoSet);
    
    friend class CanonCounterVisitor;
    
  private:
    void addToModifiedAreas_ (AreaEntry* area);
};

}}

#include "StateGraphTraits.h"
                 

#endif 
