// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/**
 * @file StateGraphTraits.h
 * 
 * This file contains @c boost::graph_traits for @c State class.
 * The traits model <tt>Vertex List Graph</tt> and
 * <tt>Incidence Graph</tt> concepts and therefore
 * can be used in <tt>boost::breadt_first_search</tt> algorithm.
 */

#ifndef STATE_GRAPH_TRAITS_H
#define STATE_GRAPH_TRAITS_H

#include <boost/graph/graph_traits.hpp>
#include <boost/iterator/transform_iterator.hpp>

#include "State.h"

namespace boost {
    using namespace mmodule::detail;
    
    template <>
    struct graph_traits<State> 
    {
        template <typename T>
        struct PtrGetter {
            typedef T* result_type;
            
            T* operator() (T& t) const
            { return &t; }
        };
        
        typedef AreaEntry* vertex_descriptor;
        typedef RefContentEntry* edge_descriptor;
        typedef directed_tag directed_category;
        typedef allow_parallel_edge_tag edge_parallel_category;
        struct traversal_category : vertex_list_graph_tag, incidence_graph_tag {};
        
        typedef boost::transform_iterator<PtrGetter<AreaEntry>,
                                          State::AreaEntryIterator>
                vertex_iterator;
        typedef size_t vertices_size_type;
        
        typedef boost::transform_iterator<PtrGetter<RefContentEntry>,
                                          AreaEntry::OutLinksIterator> out_edge_iterator;
        typedef size_t degree_size_type;
        
        AreaEntry* null_vertex ()
        { return 0; }
    };
    
    
    inline size_t 
    num_vertices (const State& s)
    { 
        return s.getAreaCount (); 
    }
        
    inline std::pair<graph_traits<State>::vertex_iterator, graph_traits<State>::vertex_iterator>
    vertices (const State& s)
    { 
        return const_cast<State&> (s).areas (); 
    }
        
    inline graph_traits<State>::vertex_descriptor 
    source (graph_traits<State>::edge_descriptor e, const State&)
    { 
        return e->getSource (); 
    }
        
    inline graph_traits<State>::vertex_descriptor 
    target (graph_traits<State>::edge_descriptor e, const State&)
    { 
        return e->getTarget (); 
    }
        
    inline std::pair<graph_traits<State>::out_edge_iterator,
                     graph_traits<State>::out_edge_iterator>
    out_edges (graph_traits<State>::vertex_descriptor v, const State&)
    { 
        return v->outLinks (); 
    }
        
    inline size_t 
    out_degree (graph_traits<State>::vertex_descriptor v, const State&)
    { 
        return v->outLinksCount (); 
    }
}

#endif // STATE_GRAPH_TRAITS_H
