// 
// GMC Gimple Model Checker
// Copyright (C) 2009-2011 GMC team
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as 
// published by the Free Software Foundation, either version 3 of 
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public 
// License along with this program.  
// If not, see <http://www.gnu.org/licenses/>.

/*
 * File:   MemoryStateTest.cpp
 * Author: sery
 *
 * Created on Jul 8, 2010, 10:14:38 AM
 */

#include <memory>

#include "../Error.h"

#include "../git/ConstructionHelpers.h"
#include "../git/IntegerType.h"
#include "../int/IntegerValue.h"
#include "../mm/MemoryState.h"
#include "MemoryStateTest.h"

using namespace mmodule;
using namespace ginterpreter;
using namespace giterator;
using namespace giterator::detail;

CPPUNIT_TEST_SUITE_REGISTRATION(MemoryStateTest);

MemoryStateTest::MemoryStateTest() {
}

MemoryStateTest::~MemoryStateTest() {
}

void MemoryStateTest::setUp() {
}

void MemoryStateTest::tearDown() {
}

void MemoryStateTest::testMS01_Alloc() {
    MemoryState ms(NULL);

    Ptr p = ms.alloc(sizeof(int));
    ms.setRootArea(p);
    ms.push();

    LOG_DEBUG("Heap hash: %u", ms.hash());
}

void MemoryStateTest::testMS02_Store() {
    MemoryState ms(NULL);

    Ptr p = ms.alloc(sizeof(int));
    ms.setRootArea(p);
    ms.push();
    LOG_DEBUG("Heap hash: %u", ms.hash());

    git_type int32 = createIntegerType(32, false, 32);
    ms.store(p, std::auto_ptr<Content>(new IntegerValue(int32, 1)));
    ms.push();
    LOG_DEBUG("Heap hash: %u", ms.hash());
}

void MemoryStateTest::testMS03_StoreLoad() {
    MemoryState ms(NULL);

    Ptr p = ms.alloc(sizeof(int));
    ms.setRootArea(p);
    ms.push();
    LOG_DEBUG("Heap hash: %u", ms.hash());

    git_type int32 = createIntegerType(32, false, 32);
    IntegerValue* c = new IntegerValue(int32, 42);
    Hash h = c->getHash();
    ms.store(p, std::auto_ptr<Content>(c));
    ms.push();
    LOG_DEBUG("Heap hash: %u", ms.hash());

    const IntegerValue* iv = dynamic_cast<const IntegerValue*>(ms.load(p));

    CPPUNIT_ASSERT(int32 == iv->getType());
    CPPUNIT_ASSERT_EQUAL((uint64_t)42, iv->getValue());
    CPPUNIT_ASSERT(h == iv->getHash());
}


void MemoryStateTest::testMS04_HashTest() {
    MemoryState ms(NULL);
    Hash h1;
    Hash h2;
    Hash h3;

    Ptr p = ms.alloc(sizeof(int));
    ms.setRootArea(p);
    ms.push();
    LOG_DEBUG("Heap hash: %u", ms.hash());

    git_type int32 = createIntegerType(32, false, 32);
    IntegerValue* c = new IntegerValue(int32, 1);
    ms.store(p, std::auto_ptr<Content>(c));
    ms.push();
    h1 = ms.hash();
    LOG_DEBUG("Heap hash: %u", ms.hash());

    c = new IntegerValue(int32, 2);
    ms.store(p, std::auto_ptr<Content>(c));
    ms.push();
    h2 = ms.hash();
    LOG_DEBUG("Heap hash: %u", ms.hash());

    c = new IntegerValue(int32, 1);
    ms.store(p, std::auto_ptr<Content>(c));
    ms.push();
    h3 = ms.hash();
    LOG_DEBUG("Heap hash: %u", ms.hash());

    CPPUNIT_ASSERT_EQUAL(h1, h3);
    CPPUNIT_ASSERT_MESSAGE("This is not a strict requirement due to precision "
            "loss in hashing. However, colision in such marginally different "
            "heaps implies big troubles during model checking.", h1 != h2);

    ms.pop();
    ms.backtrack();
    h2 = ms.hash();
    ms.pop();
    ms.backtrack();
    h3 = ms.hash();
    CPPUNIT_ASSERT_EQUAL(h1, h3);
    CPPUNIT_ASSERT_MESSAGE("This is not a strict requirement due to precision "
            "loss in hashing. However, colision in such marginally different "
            "heaps implies big troubles during model checking.", h1 != h2);
    ms.pop();
    ms.backtrack();
}

void MemoryStateTest::testMS05_Backtrack() {
    MemoryState ms(NULL);
    git_type int32 = createIntegerType(32, false, 32);
    const uint32_t MAX = 10;
    Hash hb1;
    Hash hb2;
    Hash he1;
    Hash he2;

    Ptr p = ms.alloc(sizeof(int));
    ms.setRootArea(p);
    ms.push();
    hb1 = ms.hash();
    LOG_DEBUG("Heap hash: %u", ms.hash());

    for (uint32_t i = 0; i < MAX; i++) {
        LOG_DEBUG("Storing %u", i);
        IntegerValue* c = new IntegerValue(int32, i);
        ms.store(p, std::auto_ptr<Content>(c));
        ms.push();
    }
    he1 = ms.hash();
    LOG_DEBUG("Heap hash: %u", ms.hash());

    for (uint32_t i = MAX; i > 0; i--) {
        const IntegerValue* iv = dynamic_cast<const IntegerValue*>(ms.load(p));
        LOG_DEBUG("Loading %u (expected %u)", (uint32_t)iv->getValue(), i-1);
        CPPUNIT_ASSERT_EQUAL(i-1, (uint32_t)iv->getValue());
        ms.pop();
        ms.backtrack();
    }
    hb2 = ms.hash();
    LOG_DEBUG("Heap hash: %u", ms.hash());

    for (uint32_t i = 0; i < MAX; i++) {
        LOG_DEBUG("Storing %u", i);
        IntegerValue* c = new IntegerValue(int32, i);
        ms.store(p, std::auto_ptr<Content>(c));
        ms.push();
    }
    he2 = ms.hash();
    LOG_DEBUG("Heap hash: %u", ms.hash());

    CPPUNIT_ASSERT_EQUAL(hb1, hb2);
    CPPUNIT_ASSERT_EQUAL(he1, he2);
    CPPUNIT_ASSERT_MESSAGE("This is not a strict requirement due to precision "
            "loss in hashing. However, colision in such marginally different "
            "heaps implies big troubles during model checking.", hb1 != he1);
}

